#!/usr/bin/env python3

# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.




"""
NVSM CLI - Command-line interface for NVIDIA System Management tools.

This is documentation for the "nvsm" command itself. I want to include this
documentation in the nvsm(1) man page somehow.
"""

# Don't accidentally import readline module (might be GPL licensed)
import sys
sys.modules['readline'] = None

import argparse
import locale
import os
import traceback

from nvsm.color import colorize
from nvsm.backend.context import NVSMContext
from nvsm.backend.handle import NVSMHandle
import nvsm

recognized_options = [
    {
        'args': [ '--color' ],
        'kwargs': {
            'metavar': 'WHEN', 'default': 'auto',
            'choices': ['always', 'never', 'auto'],
            'help': '''Control colorization of output. Possible values for WHEN
                       are "always", "never", or "auto". Default value is
                       "auto".'''
        }
    },
    {
        'args': [ '-i', '--interactive' ],
        'kwargs': {
            'action': 'store_true',
            'help': '''When this option is given, run in interactive mode. The
                       default is automatic.'''
        }
    },
    {
        'args': [ '--ignore-privileges' ],
        'kwargs': {
            'action': 'store_true', 'help': argparse.SUPPRESS
        }
    },
    {
        'args': [ '--log-level' ],
        'kwargs': {
            'choices': [ 'debug', 'info', 'warning', 'error', 'critical' ],
            'metavar': 'LEVEL', 'default': 'warning',
            'help': '''Set the output logging level. Possible values for LEVEL
                       are "debug", "info", "warning", "error", and "critical".
                       The default value is "warning".'''
        }
    }
]

def get_arg_parser():
    parser = argparse.ArgumentParser(
            usage='%(prog)s [-h] [--color WHEN] [-i] [--] [<command>...]',
            description='NVIDIA System Management interface')
    parser.usage += "\nRun 'sudo nvsm stress-test --usage' for stress-test help message."
    for opt in recognized_options:
        parser.add_argument(*opt['args'], **opt['kwargs'])
    return parser

def main():
    # Use system default locale/encoding
    locale.setlocale(locale.LC_ALL, '')
    code = locale.getpreferredencoding()

    # Parse arguments passed by argv
    parser = get_arg_parser()
    args, remaining = parser.parse_known_args()

    # TODO: Print version

    # Map log level from input string
    import logging
    import logging.handlers
    args.log_level = {
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warning': logging.WARNING,
        'error': logging.ERROR,
        'critical': logging.CRITICAL
    }.get(args.log_level, logging.WARNING)

    # Initialize color
    nvsm.color.init(hint=args.color)

    # Determine if the shell is interactive
    interactive = False
    if args.interactive:
        # Force interactive
        interactive = True
    elif sys.stdout.isatty():
        interactive = True

    # Initialize logging (uses color if enabled)
    rootLogger = logging.getLogger()
    rootLogger.setLevel(args.log_level)
    if interactive:
        nvsmLogOut = sys.stdout
    else:
        nvsmLogOut = sys.stderr
    nvsm.log.init(
            level=args.log_level,
            out=nvsmLogOut)

    # Ensure superuser privileges
    if (os.geteuid() != 0) and not (args.ignore_privileges):
        nvsm.log.error('{0} must be run with root privileges',
                os.path.basename(sys.argv[0]))
        sys.exit(1)

    # Interpret SMASH commands passed directly through argv
    if remaining and remaining[0] == '--':
            remaining = remaining[1:]
    if remaining:
        nvsm_interpreter = nvsm.interpreter.Interpreter()
        for i in range(len(remaining)):
            remaining[i] = remaining[i].replace(' ', '\\ ')
        program = nvsm_interpreter.evaluate(' '.join(remaining))
        if not program:
            # Some error while constructing the program
            sys.exit(1)
        backend = NVSMHandle()
        backend.connect()
        with NVSMContext(backendHandle=backend) as context:
            try:
                program.run(context=context)
            except KeyboardInterrupt:
                nvsm.log.info('Keyboard interrupt; aborting command')
                context.invalidate()
                sys.exit(130)  # Exit code for "killed by keyboard interrupt"
            except (nvsm.exceptions.BackendError,
                    nvsm.exceptions.CommandError,
                    nvsm.exceptions.MissingProperty,
                    nvsm.exceptions.MissingTarget,
                    nvsm.exceptions.SchemaError) as e:
                nvsm.log.error('{0}', e)
                context.invalidate()
                sys.exit(1)
            except Exception as e:
                # For robustness, log unhandled exceptions
                nvsm.log.error('Unhandled exception: {}'.format(e))
                nvsm.log.error('Run nvsm again with "--log-level=debug" for details')
                nvsm.log.debug(traceback.format_exc())
                sys.exit(1)
        sys.exit(0)

    # Start REPL if the terminal interactive
    if interactive:
        nvsm.repl.main_loop()
    else:
        # TODO: Run command in batch mode (non-interactive)
        pass

if __name__ == '__main__':
    main()
