# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.


from nvsm.backend.context import Target, dict_to_properties

import nvsm.log


def filter_alerts(alerts, nodeIdentifier, parentIdentifier):
    # filter based on node and parent
    filtered_alerts = []
    for i in range(len(alerts)):
        a = alerts[i]
        if nodeIdentifier \
                and (nodeIdentifier.lower() != a.get('system_serial').lower()):
            continue
        elif parentIdentifier \
                and '+' in a.get('component_id') \
                and (parentIdentifier.lower() != a.get('component_id').split('+')[0].lower()):
            continue
        filtered_alerts.append(a)
    return filtered_alerts


class AlertTarget(Target):
    def __init__(self, ref, index, address, context, nodeIdentifier=None, parentIdentifier=None):
        super(AlertTarget, self).__init__(
            address=address, context=context)
        self.__ref = ref
        self.__index = index
        self.__properties = None
        self.__nodeIdentifier = nodeIdentifier
        self.__parentIdentifier = parentIdentifier

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        # Get contents of this alert from the NVSM backend
        response = self._context._queryBackend(self.__ref)
        if 'Alerts' not in response:
            nvsm.log.warning(
                'Could not find "Alerts" attribute in NVSM backend response')
            return
        alerts = response['Alerts']
        if alerts is None:
            alerts = []

        # filter based on node and parent
        alerts = filter_alerts(
            alerts, self.__nodeIdentifier, self.__parentIdentifier)

        if len(alerts) <= self.__index:
            nvsm.log.warning(
                'Could not find alert index {} in NVSM backend response',
                self.__index)
            return
        alert = alerts[self.__index]

        # FIXME +id: rework once the primary key is implemented to handle this case
        if '+' in alert['component_id']:
            alert['component_id'] = alert['component_id'].split('+')[-1]

        self.__properties = list(dict_to_properties(target=self, d=alert))
        default_properties = ['system_name',
                              'message_details',
                              'component_id',
                              'description',
                              'event_time',
                              'recommended_action',
                              'alert_id',
                              'system_serial',
                              'message',
                              'severity']

        def compare_properties(a):
            try:
                i = default_properties.index(a.name.lower())
            except ValueError:
                i = len(default_properties)
            return i
        self.__properties.sort(key=compare_properties)

    def __getitem__(self, key):
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties upon first access
            self.__initProperties()
        return self.__properties


class AlertsTarget(Target):
    def __init__(self, ref, address, context, nodeIdentifier=None, parentIdentifier=None):
        super(AlertsTarget, self).__init__(
            address=address, context=context)
        self.__ref = ref
        self.__targets = None
        self.__nodeIdentifier = nodeIdentifier
        self.__parentIdentifier = parentIdentifier

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}
        # Enumerate alerts with NVSM backend
        response = {}
        try:
            response = self._context._queryBackend(self.__ref)
        except:
            pass
        if 'Alerts' not in response:
            nvsm.log.warning(
                'Could not find "Alerts" attribute in NVSM backend response')
            return
        alerts = response['Alerts']
        if alerts is None:
            alerts = []

        # filter based on node and parent
        alerts = filter_alerts(
            alerts, self.__nodeIdentifier, self.__parentIdentifier)

        for i in range(len(alerts)):
            a = alerts[i]
            if (not self._context.showall) and (a.get('severity').lower() == 'cleared'):
                continue
            name = 'alert{}'.format(i)
            target = AlertTarget(
                ref=self.__ref,
                index=i,
                address=(self.address + name),
                context=self._context,
                nodeIdentifier=self.__nodeIdentifier,
                parentIdentifier=self.__parentIdentifier)
            self.__targets[name] = target

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?
