# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.


from nvsm.backend.context import Target, dict_to_properties
from nvsm.backend.environment import ChassisPowerTarget, ChassisThermalTarget
from nvsm.backend.networkadapters import NetworkAdapterCollectionTarget
from nvsm.backend.alerts import AlertsTarget
from nvsm.backend.policy import BladePolicyTarget
import nvsm.exceptions
import time
import sys
import signal
from nvsm.color import colorize

bad_status = ['warning', 'critical']
good_status = ['ok']


def colorize_status(value):
    if value == None:
        return "none"
    if value.lower() in good_status:
        return 'green'
    if value.lower() in bad_status:
        return 'red'
    return 'none'


class ChassisTarget(Target):

    def __init__(self, chassisref, address, context):
        super(ChassisTarget, self).__init__(
            address=address, context=context)
        self.__chassisref = chassisref
        self.__properties = None
        self.__targets = None

    def __initTargets(self):
        # Enumerate chassis with NVSM backend
        chassisidinfo = {}
        try:
            chassisidinfo = self._context._queryBackend(
                self.__chassisref, forceRefresh=True)
        except:
            pass
        self.__targets = {}
        power = chassisidinfo.get("Power", None)
        nodeIdentifier = chassisidinfo.get("SerialNumber", None)
        if not nodeIdentifier:
            try:
                # try to get the node serial from systems uri
                systemsidinfo = self._context._queryBackend(
                    "redfish/v1/Systems/1")
                nodeIdentifier = systemsidinfo["SerialNumber"]
            except:
                nodeIdentifier = "-1"  # some invalid serial number for alerts filtering

        if power is not None:
            chassisPowerTarget = ChassisPowerTarget(self.__chassisref + "/Power",
                                                    address=(
                                                        self.address + 'power'),
                                                    context=self._context,
                                                    nodeIdentifier=nodeIdentifier)
            self.__targets['power'] = chassisPowerTarget
        thermal = chassisidinfo.get("Thermal", None)
        if thermal is not None:
            chassisThermalTarget = ChassisThermalTarget(self.__chassisref + "/Thermal",
                                                        address=(
                                                            self.address + 'thermal'),
                                                        context=self._context,
                                                        nodeIdentifier=nodeIdentifier)
            self.__targets['thermal'] = chassisThermalTarget
        NetworkAdapters = chassisidinfo.get("NetworkAdapters", None)
        if NetworkAdapters is not None and len(NetworkAdapters) > 0:
            chassisNetworkAdaptorsTarget = NetworkAdapterCollectionTarget(self.__chassisref + "/NetworkAdapters",
                                                                          address=(
                                                                              self.address + 'NetworkAdapters'),
                                                                          context=self._context,
                                                                          nodeIdentifier=nodeIdentifier)
            self.__targets['NetworkAdapters'] = chassisNetworkAdaptorsTarget

        Links = chassisidinfo.get("Links", None)
        if not Links:
            contains = None
        else:
            contains = Links.get("Contains", None)
        if contains == None:
            contains = []
        for chassisref in contains:
            chassisref = chassisref.get('@odata.id', None)
            if not chassisref:
                continue
            index = chassisref.split('/')[-1]
            if index == "1":
                index = 'localhost'
            chassisid = ChassisTarget(
                chassisref=chassisref,
                address=(self.address + index),
                context=self._context)
            self.__targets[index] = chassisid

    def __getitem__(self, key):
        if self.__targets is None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    def __initProperties(self):
        assert self.__properties is None
        chassisinfo = self._context._queryBackend(
            self.__chassisref, forceRefresh=True)
        nvsm.log.debug('chassis response', chassisinfo)
        properties = ['oem_power_mode', 'assettag', 'indicatorled', 'manufacturer', 'serialnumber',
                      'oem_power_state', 'status_state', 'status_health', "powerstate"]
        properties_without_oem = ['power_mode', 'assettag', 'indicatorled', 'manufacturer', 'serialnumber',
                                  'power_state', 'status_state', 'status_health', "powerstate"]

        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            if str(value).lower() == 'n/a':
                return self._context.showall
            if name.lower() in properties:
                return True
            return self._context.showall

        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            if name.lower().startswith('oem_'):
                name = name[len('oem_'):]
            return (name, value, color)
        self.__properties = list(dict_to_properties(
            target=self, d=chassisinfo,
            criteria=filter_properties,
            mapping=map_properties))

        def compare_properties(a):
            try:
                i = properties_without_oem.index(a.name.lower())
            except ValueError:
                i = len(properties_without_oem)
            return i
        self.__properties.sort(key=compare_properties)

    @property
    def targets(self):
        if self.__targets is None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties is None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties

    def set(self, name, value):

        for x in ['PowerMode', 'Power_Mode', 'IndicatorLED', 'Power']:
            prop = name.lower()
            if name.lower() == x.lower():
                nvsm.log.debug('chassis endpoint ', self.__chassisref)
                if prop == 'powermode' or prop == 'power_mode':
                    response = self._context._queryBackend(self.__chassisref,
                                                           method='POST', body={'PowerMode': value})
                    if hasattr(response, 'get') and response.get('error'):
                        nvsm.log.error(response['error'])

                    last_update = ""
                    file = sys.stdout
                    with nvsm.progress.ProgressBar(file=file) as progress_bar:
                        progress_bar.update(
                            percent=0.0,
                            message="In Progress",
                            error=False)
                        retry = 0
                        while True:
                            response = self._context._queryBackend(
                                path=self.__chassisref, forceRefresh=True)

                            oem = response.get("Oem", None)
                            if oem is None:
                                break
                            oem_power = oem.get("Power", None)
                            if oem_power is None:
                                break
                            oem_power_mode = oem_power.get("Mode", None)
                            if oem_power_mode is None:
                                break
                            if str(value).upper() in str(oem_power_mode).upper():
                                last_update = "Success"
                                progress_bar.update(
                                    percent=100.0,
                                    message="Success",
                                    error=False)
                                break

                            if retry >= 4:
                                last_update = "Failed"
                                progress_bar.update(
                                    percent=0,
                                    message="Unable to confirm set powermode. Please use 'nvsm show powermode' to confirm.",
                                    error=True)
                                break

                            time.sleep(4)

                            retry += 1
                            progress_bar.update(
                                percent=25.0 * retry,
                                message="In Progress",
                                error=False)

                    print('Status: ', end='', file=file)
                    if last_update.lower() in ['done', 'finished', 'success']:
                        color = 'green'
                    else:
                        color = 'red'
                    with colorize(color):
                        print(last_update, file=file)

                if prop == 'power':
                    state = ''
                    uri = self.__chassisref + '/Actions/Chassis.Reset'
                    if value.lower() == 'on':
                        state = 'ForceOn'
                    if value.lower() == 'off':
                        state = 'ForceOff'
                    response = self._context._queryBackend(uri,
                                                           method='POST', body={'ResetType': state})
                    if hasattr(response, 'get') and response.get('error'):
                        nvsm.log.error(response['error'])
                if prop == 'indicatorled':
                    nvsm.log.debug('name {0} {1}.', name, value)
                    response = self._context._queryBackend(self.__chassisref,
                                                           method='POST', body={'IndicatorLED': value})

                    if hasattr(response, 'get') and response.get('error'):
                        nvsm.log.error(response['error'])
                    last_update = ""
                    file = sys.stdout
                    with nvsm.progress.ProgressBar(file=file) as progress_bar:
                        progress_bar.update(
                            percent=0.0,
                            message="In Progress",
                            error=False)
                        retry = 0
                        while True:
                            response = self._context._queryBackend(
                                path=self.__chassisref, forceRefresh=True)

                            led = response.get("IndicatorLED", None)
                            if led is None:
                                break
                            if str(value).upper() in str(led).upper():
                                last_update = "Success"
                                progress_bar.update(
                                    percent=100.0,
                                    message="Success",
                                    error=False)
                                break

                            if retry >= 5:
                                last_update = "Failed"
                                progress_bar.update(
                                    percent=0,
                                    message="Unable to confirm set IndicatorLED. Please use 'nvsm show led' to confirm.",
                                    error=True)
                                break

                            time.sleep(15)

                            retry += 1
                            progress_bar.update(
                                percent=20.0 * retry,
                                message="In Progress",
                                error=False)

                    print('Status: ', end='', file=file)
                    if last_update.lower() in ['done', 'finished', 'success']:
                        color = 'green'
                    else:
                        color = 'red'
                    with colorize(color):
                        print(last_update, file=file)

                return
        raise nvsm.exceptions.MissingProperty(
            'Target property "{0}" does not exist'.format(name))


# FIXME: The NVSM backend does not implement the "ChassisCollection" endpoint
#        at this time.
class ChassisCollectionTarget(Target):

    def __init__(self, address, context):
        super(ChassisCollectionTarget, self).__init__(
            address=address, context=context)
        self.__targets = None

    def __initTargets(self):
        # Enumerate chassis with NVSM backend
        self.__targets = {}
        chassisinfo = {}
        try:
            chassisinfo = self._context._queryBackend('/redfish/v1/Chassis')
        except:
            pass
        members = chassisinfo.get('Members', [])
        if members is None:
            members = []
        for chassisref in members:
            chassisref = chassisref.get('@odata.id', None)
            if not chassisref:
                continue
            index = chassisref.split('/')[-1]
            # import socket
            # system_host = socket.gethostname()
            if index == "1":
                index = 'localhost'

            chassis = ChassisTarget(
                chassisref=chassisref,
                address=(self.address + index),
                # address=(self.address + local_host),
                context=self._context)
            # self.__targets[index] = system
            self.__targets[index] = chassis

        try:
            # Assume every chassis has CMM alert targets (for DSHM alerts)
            alertsTarget = AlertsTarget(
                ref='/nvsm/v1/Alerts/Blade',
                address=(self.address + 'alerts'),
                context=self._context)
            if alertsTarget.targets:
                self.__targets['alerts'] = alertsTarget

            # Assume every chassis has CMM policy targets (for configuring DSHM)
            policyTarget = BladePolicyTarget(
                # FIXME: NVSM backend does not provide OData reference here
                policyref='/nvsm/v1/Policy/Blade',
                address=(self.address + 'policy'),
                context=self._context)
            self.__targets['policy'] = policyTarget
        except nvsm.exceptions.BackendError as e:
            # for non bladerunner systems
            nvsm.log.debug(e, exc_info=True)

    def __getitem__(self, key):
        if self.__targets is None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets is None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?
