# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



from nvsm.targetAddress import TargetAddress
import nvsm.context
import nvsm.exceptions
import nvsm.log
import nvsm.odata
from functools import reduce

class Property(object):
    def __init__(self, target, name, value, color='none'):
        self.__target = target
        self.__name = name
        self.__value = value
        self.__color = color

    @property
    def name(self):
        return self.__name

    @property
    def value(self):
        return self.__value

    @property
    def color(self):
        return self.__color


def dict_to_properties(
        target, d, s='',
        criteria=lambda name, value: True,
        mapping=lambda name, value: (name, value, 'none')):
    for k, v in list(d.items()):
        if s:
            name = '{0}_{1}'.format(s, k)
        else:
            name = k
        if hasattr(v, 'items'):
            for p in dict_to_properties(target, v, name, criteria, mapping):
                yield p
        elif ((type(v) is list) or (type(v) is nvsm.odata.ODataArray)) and \
            all([hasattr(x, 'items') for x in v]):
            v_len = len(v)
            for i in range(v_len):
                i_name = '{0}_{1}'.format(name, str(i))
                for p in dict_to_properties(target, v[i], i_name, criteria, mapping):
                    yield p
        elif criteria(name, v):
            (map_name, map_value, map_color) = mapping(name, v)
            p = Property(target=target, name=map_name, value=map_value, color=map_color)
            yield p


class Target(object):
    def __init__(self, address, context):
        super(Target, self).__init__()
        self.__address = address
        self.__context = context

    @property
    def address(self):
        return self.__address

    @property
    def _context(self):
        return self.__context

    @property
    def name(self):
        return self.__address.basename

    def __getitem__(self, key):
        raise NotImplementedError

    @property
    def targets(self):
        raise NotImplementedError

    @property
    def properties(self):
        raise NotImplementedError

    @property
    def verbs(self):
        result = []
        smash_verbs = [ 'create', 'delete', 'dump', 'help', 'load', 'reset',
                        'set', 'start', 'stop', 'enable', 'disable', 'lock', 
                        'unlock', 'export', 'erase', 'rotate', 'factoryreset' ]
        for v in smash_verbs:
            if hasattr(self, v):
                result.append(v)
        return result

class EmptyTarget(Target):
    def __init__(self, targetHierarchy, address, context):
        super(EmptyTarget, self).__init__(
                address=address, context=context)
        self.__targets = {}

        # Recursively construct the target hierarchy
        for key, value in list(targetHierarchy.items()):
            new_target = EmptyTarget(
                    targetHierarchy=value,
                    address=(self.address + key),
                    context=context)
            self.__targets[key] = new_target

    def _addTarget(self, target):
        self.__targets[target.name] = target

    def _addEpsilonTarget(self, target):
        targets = []
        targets.extend(target.targets)
        for t in targets:
            self.__targets[t.name] = t

    def __getitem__(self, key):
        return self.__targets[key]

    @property
    def targets(self):
        return list(self.__targets.values())

    @property
    def properties(self):
        return []


class NVSMContext(nvsm.context.SMASHContext):
    def __init__(self, backendHandle, workingTargetAddress=TargetAddress('/')):
        super(NVSMContext, self).__init__(
                workingTargetAddress=workingTargetAddress)

        self.__rootTarget = None
        self.__invalid = False
        self.__backend = backendHandle
        self.__backendCache = {}

    def __enter__(self):
        # self.__populateTargetHierarchy()
        return self

    def invalidate(self):
        nvsm.log.debug('Context invalidated; changes to NVSM backend' \
                       ' will be discarded')
        self.__invalid = True

    def __exit__(self, exc_type, exc_value, traceback):
        if not self.__invalid:
            pass  # TODO: Commit changes to NVSM backend?

    def _queryBackend(self, path, method='GET', body=None, stream=False, raw=False, forceRefresh=False):
        # Strip fragment identifier and slashes from path
        path = path.split('#', 1)[0]
        path = path.strip('/')

        self.__backendCache = {}  # XXX: Purge cache to get latest progress...

        if not stream:
            if method == 'GET':
                # Use cached response for repeated 'GET' queries to backend
                if path in self.__backendCache:
                    nvsm.log.debug('Using cached response to GET {}'.format(path))
                    return self.__backendCache[path]
            else:
                # Invalidate the cache when changes are made
                self.__backendCache.clear()
        response = self.__backend.query(path=path, method=method, body=body, stream=stream, raw=raw, forceRefresh=forceRefresh)
        # TODO: In a future refactor, we should construct OData object here

        if not stream:
            # Cache responses to 'GET' queries, to reduce duplicate queries
            if method == 'GET':
                self.__backendCache[path] = response

        return response

    def refresh(self):
        self.__populateTargetHierarchy()

    def __populateTargetHierarchy(self):
        # Construct the non-specialized targets at the root

        #import socket
        #chassis_host = socket.gethostname()
        chassis_host = 'localhost'

#        targetHierarchy = { 'chassis': { chassis_host: { } } }
        targetHierarchy = {}
        self.__rootTarget = EmptyTarget(
                targetHierarchy=targetHierarchy,
                address=nvsm.targetAddress.TargetAddress('/'),
                context=self)

        #Alerts Target
        from nvsm.backend.alerts import AlertsTarget
        alertsTarget = AlertsTarget(
            ref='/nvsm/v1/Alerts',
            address=TargetAddress('/alerts'),
            context=self)
        self.__rootTarget._addTarget(alertsTarget)

        #CallHome Target
        from nvsm.backend.callhome import CallHomeTarget
        callHomeTarget = CallHomeTarget(
                address=TargetAddress('/callhome'),
                context=self)
        self.__rootTarget._addTarget(callHomeTarget)

        #PagerDuty Target
        from nvsm.backend.pagerduty import PagerDutyTarget
        pagerDutyTarget = PagerDutyTarget(
                address=TargetAddress('/pagerduty'),
                context=self)
        self.__rootTarget._addTarget(pagerDutyTarget)

        # Assume we have global policy targets (for configuring DSHM)
        from nvsm.backend.policy import GlobalPolicyTarget
        policyTarget = GlobalPolicyTarget(
                address=TargetAddress('/policy'),
                context=self)
        self.__rootTarget._addTarget(policyTarget)

        # The systems collection is the initial entrypoint into NVSM
        from nvsm.backend.system import SystemCollectionTarget
        systemsTarget = SystemCollectionTarget(
                address=TargetAddress('/systems'),
                context=self)
        self.__rootTarget._addTarget(systemsTarget)

        # The systems collection is the initial entrypoint into NVSM
        from nvsm.backend.versions import VersionsTarget
        versionsTarget = VersionsTarget(
                address=TargetAddress('/versions'),
                context=self)
        self.__rootTarget._addTarget(versionsTarget)

        # NOTE: NVSM backend does not implement '/redfish/v1/chassis' or
        #       '/redfish/v1/chassis/1' at this time, so we "fake it" here.
        from nvsm.backend.chassis import ChassisCollectionTarget
#        # The chassis collection is adjacent to systems (according to Redfish)
        chassisTarget = ChassisCollectionTarget(
                address=TargetAddress('/chassis'),
                context=self)
        self.__rootTarget._addTarget(chassisTarget)

        try:
            # The systems collection is the initial entrypoint into NVSM
            from nvsm.backend.managers import ManagersCollectionTarget
            managersTarget = ManagersCollectionTarget(
                    address=TargetAddress('/managers'),
                    context=self)
            self.__rootTarget._addTarget(managersTarget)
        except nvsm.exceptions.BackendError as e:
            # for non bladerunner systems
            nvsm.log.debug(e, exc_info=True)

        # XXX
#        from nvsm.backend.environment import ChassisPowerTarget, ChassisThermalTarget
#        chassisPowerTarget = ChassisPowerTarget(
#                address=TargetAddress('/chassis/{0}/power'.format(chassis_host)),
#                context=self)
#        self.__rootTarget['chassis'][chassis_host]._addTarget(chassisPowerTarget)
#        chassisThermalTarget = ChassisThermalTarget(
#                address=TargetAddress('/chassis/{0}/thermal'.format(chassis_host)),
#                context=self)
#        self.__rootTarget['chassis'][chassis_host]._addTarget(chassisThermalTarget)

    def getTarget(self, address):
        if self.__rootTarget == None:
            self.__populateTargetHierarchy()

        target = self.__rootTarget
        for target_name in address.canonicalPath:
            try:
                next_target = target[target_name]
            except KeyError:
                raise nvsm.exceptions.MissingTarget(
                    'Could not find target at address "{0}"'.format(address))
            target = next_target
        return target

    def _listTargets(self, address):
        try:
            results = reduce(dict.get, address.canonicalPath, self.__targetHierarchy)
        except TypeError:
            raise nvsm.exceptions.MissingTarget(
                'Could not find target at address "{0}"'.format(address))
        results = [address + name for name in list(results.keys())]
        results = [addr.canonicalize() for addr in results]
        results = [self.getTarget(addr) for addr in results]
        return results
