# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.


from nvsm.backend.context import Target, dict_to_properties

from nvsm.backend.alerts import AlertsTarget
from nvsm.backend.policy import PowerPolicyTarget, ThermalPolicyTarget

import nvsm.log
from functools import reduce

bad_status = ['warning', 'critical']
good_status = ['ok']


def colorize_status(value):
    if value == None:
        return "none"
    if value.lower() in good_status:
        return 'green'
    if value.lower() in bad_status:
        return 'red'
    return 'none'


class ChassisPowerSupplyTarget(Target):

    def __init__(self, psuref, address, context):
        super(ChassisPowerSupplyTarget, self).__init__(
            address=address, context=context)
        self.__psuref = psuref
        self.__properties = None

    def __initProperties(self):
        self.__properties = []  # XXX?
        powerinfo = self._context._queryBackend(self.__psuref)
        # TODO: The queryBackend() routine should be able to cache the earlier
        #       call to "/redfish/v1/Chassis/1/Power" and also interpret the
        # "#/PowerSupplies/1" local reference. It does not at this time,
        #       so we do the work in this routine.
        psupath = self.__psuref.split('#')[-1].strip('/').split('/')

        def to_int(x):
            try:
                x = int(x)
            except:
                pass
            return x
        try:
            psuinfo = reduce(lambda x, y: x[to_int(y)], psupath, powerinfo)
        except (KeyError, TypeError):
            raise nvsm.exceptions.BackendError(
                'Could not find API object at path "{0}"'.format(
                    self.__psuref))

        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            return True

        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            return (name, value, color)
        self.__properties = list(dict_to_properties(
            target=self, d=psuinfo,
                criteria=filter_properties,
                mapping=map_properties))

    def __getitem__(self, key):
        # Chassis power supply has no targets at this time
        # (behaves as an empty dictionary)
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties is None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class ChassisPowerControlTarget(Target):

    def __init__(self, pcref, address, context):
        super(ChassisPowerControlTarget, self).__init__(
            address=address, context=context)
        self.__pcref = pcref
        self.__properties = None

    def __initProperties(self):
        self.__properties = []  # XXX?
        powerinfo = self._context._queryBackend(self.__pcref)
        # TODO: The queryBackend() routine should be able to cache the earlier
        #       call to "/redfish/v1/Chassis/1/Power" and also interpret the
        # "#/PowerSupplies/1" local reference. It does not at this time,
        #       so we do the work in this routine.
 
        try:
            for powerctrl in powerinfo.get("PowerControl", None):
                if powerctrl.get('@odata.id', None) == self.__pcref:
                    psuinfo = powerctrl
            if not psuinfo.get("PowerLimit", None):
                psuinfo["PowerLimit"]["LimitInWatts"] = "<not set>"
        except (KeyError, TypeError):
            raise nvsm.exceptions.BackendError(
                'Could not find API object at path "{0}"'.format(
                    self.__pcref))

        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            return True

        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            return (name, value, color)
        self.__properties = list(dict_to_properties(
            target=self, d=psuinfo,
                criteria=filter_properties,
                mapping=map_properties))

    def __getitem__(self, key):
        # Chassis power supply has no targets at this time
        # (behaves as an empty dictionary)
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties is None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


    def set(self, name, value):
        for x in [ 'Oem_Power_Control_Priority' , 'PowerLimit_LimitInWatts',
                   'Oem_Power_Control_SpecificInfo_PerGpuMaxPInWatts',
                  'Oem_Power_Control_SpecificInfo_PerGpuMaxQInWatts']:
            prop = name.lower()
            if name.lower() == x.lower():
                nvsm.log.debug('chassis endpoint ', self.__pcref)
                nvsm.log.debug('name {0} {1}.', name, value)
                if prop == 'Oem_Power_Control_Priority'.lower():
                    body = {
                        "PowerControl": [
                            {
                                "@odata.id": self.__pcref,
                                "Oem": {
                                    "Power_Control": {
                                        "Priority": value
                                    }
                                }
                            }
                        ]
                    }
                if prop == 'Oem_Power_Control_SpecificInfo_PerGpuMaxPInWatts'.lower():
                    body = {
                        "PowerControl": [
                            {
                                "@odata.id": self.__pcref,
                                "Oem": {
                                    "Power_Control": {
                                      "SpecificInfo": {
                                        "PerGpuMaxPInWatts": int(value),
                                      }
                                    }
                                }
                            }
                        ]
                    }
                if prop == 'Oem_Power_Control_SpecificInfo_PerGpuMaxQInWatts'.lower():
                    body = {
                        "PowerControl": [
                            {
                                "@odata.id": self.__pcref,
                                "Oem": {
                                    "Power_Control": {
                                      "SpecificInfo": {
                                        "PerGpuMaxQInWatts": int(value),
                                      }
                                    }
                                }
                            }
                        ]
                    }
                if prop == 'PowerLimit_LimitInWatts'.lower():
                    body = {
                        "PowerControl": [
                            {
                                "@odata.id": self.__pcref,
                                "PowerLimit": {
                                    "LimitInWatts": int(value)
                                }
                            }
                        ]
                    }
                response = self._context._queryBackend(self.__pcref,
                        method='PATCH', body=body)

                if hasattr(response, 'get') and response.get('error'):
                    nvsm.log.error(response['error'])

                return
        raise nvsm.exceptions.MissingProperty(
            'Target property "{0}" does not exist'.format(name))


class ChassisPowerTarget(Target):

    def __init__(self, powerref, address, context, nodeIdentifier):
        super(ChassisPowerTarget, self).__init__(
            address=address, context=context)
        self.__targets = None
        self.__powerref = powerref
        self.__nodeIdentifier = nodeIdentifier
        self.__properties = None

    def __initTargets(self):
        assert self.__targets is None
        self.__targets = {}
        nodeName = ""
        #If the node is Group don't add alerts as target
        try:
            nodeName = self.__powerref.split('/')[-2]
        except:
            pass
        if nodeName != "Group":
            # Assume every chassis has power alert targets (for DSHM alerts)
            alertsTarget = AlertsTarget(
                ref='/nvsm/v1/Alerts/Power',  # XXX
                address=(self.address + 'alerts'),
                context=self._context,
                nodeIdentifier=self.__nodeIdentifier)
            self.__targets['alerts'] = alertsTarget

        # Assume every chassis has power policy targets (for configuring DSHM)
        policyTarget = PowerPolicyTarget(
            # FIXME: NVSM backend does not provide OData reference here
                policyref='/nvsm/v1/Policy/Power',
                address=(self.address + 'policy'),
                context=self._context,
                nodeIdentifier=self.__nodeIdentifier)
        self.__targets['policy'] = policyTarget

        # Enumerate chassis power with NVSM backend
        powerinfo = {}
        try:
            powerinfo = self._context._queryBackend(self.__powerref)
        except:
            pass
        powerControl = powerinfo.get('PowerControl', [])
        if powerControl is None:
            powerControl = []
        index = 0
        for pc in powerControl:
            pref = pc.get('@odata.id', None)
            if not pref:
                nvsm.log.warning(
                    "NVSM backend did not give '@odata.id' for chassis power control")
                continue

            if str(index) != pref.strip('/').split('/')[-1]:
                nvsm.log.info(
                    "WARNING: NVSM Backend gave index '{0}' for power cotrol when '{1}' was expected",
                    pref.strip('/').split('/')[-1],
                    index)
            name = str(pc.get('Name', index))
            psu = ChassisPowerControlTarget(
                pcref=pref,
                    address=(self.address + name),
                    context=self._context)
            self.__targets[name] = psu
            index += 1

        powerSupplies = powerinfo.get('PowerSupplies', [])
        if powerSupplies is None:
            powerSupplies = []
        index = 0
        for psuinfo in powerSupplies:
            psuref = psuinfo.get('@odata.id', None)
            if not psuref:
                nvsm.log.warning(
                    "NVSM backend did not give '@odata.id' for chassis PSU")
                continue
            if str(index) != psuref.strip('/').split('/')[-1]:
                nvsm.log.info(
                    "WARNING: NVSM Backend gave index '{0}' for PSU when '{1}' was expected",
                    psuref.strip('/').split('/')[-1],
                    index)
            name = str(psuinfo.get('Name', index))
            psu = ChassisPowerSupplyTarget(
                psuref=psuref,
                    address=(self.address + name),
                    context=self._context)
            self.__targets[name] = psu
            index += 1

    def __getitem__(self, key):
        if self.__targets is None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets is None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    def __initProperties(self):
        self.__properties = []  # XXX?
        powerinfo = self._context._queryBackend(self.__powerref)
        if not powerinfo.get("Oem", None) or not powerinfo["Oem"].get("Power_Control", None):
            powerinfo["Oem"]["Power_Control"]["policy"] = "<not set>"

        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            if "PowerControl" in name:
                return False
            if "PowerSupplies" in name:
                return False
            return True

        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            return (name, value, color)
        self.__properties = list(dict_to_properties(
            target=self, d=powerinfo,
                criteria=filter_properties,
                mapping=map_properties))


    def set(self, name, value):
        for x in [ 'Oem_Power_Control_policy' , 'Oem_Power_Control_repeat_after_in_seconds',
                  'Oem_Power_Control_active']:
            prop = name.lower()
            if name.lower() == x.lower():
                nvsm.log.debug('chassis endpoint ', self.__powerref)
                nvsm.log.debug('name {0} {1}.', name, value)
                if prop == 'Oem_Power_Control_active'.lower():
                    body = {
                        "Oem": {
                            "Power_Control": {
                                "active": True if value.lower() == "true" else False
                            }
                        }
                    }
                if prop == 'Oem_Power_Control_policy'.lower():
                    body = {
                        "Oem": {
                            "Power_Control": {
                                "policy": value,
                            }
                        }
                    }
                if prop == 'Oem_Power_Control_repeat_after_in_seconds'.lower():
                    body = {
                        "Oem": {
                            "Power_Control": {
                                "repeat_after_in_seconds": int(value),
                            }
                        }
                    }
                response = self._context._queryBackend(self.__powerref,
                        method='PATCH', body=body)

                if hasattr(response, 'get') and response.get('error'):
                    nvsm.log.error(response['error'])

                return
        raise nvsm.exceptions.MissingProperty(
            'Target property "{0}" does not exist'.format(name))

    @property
    def properties(self):
        if self.__properties is None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties

class ChassisFanTarget(Target):

    def __init__(self, fanref, address, context):
        super(ChassisFanTarget, self).__init__(
            address=address, context=context)
        self.__fanref = fanref
        self.__properties = None

    def __initProperties(self):
        assert self.__properties is None
        self.__properties = []
        faninfo = self._context._queryBackend(self.__fanref)
        # XXX: Remove this once the _queryBackend() method is able to handle local
        #      @odata.id references.
        fanpath = self.__fanref.split('#')[-1].strip('/').split('/')

        def to_int(x):
            try:
                x = int(x)
            except:
                pass
            return x
        try:
            faninfo = reduce(lambda x, y: x[to_int(y)], fanpath, faninfo)
        except (KeyError, TypeError):
            raise nvsm.exceptions.BackendError(
                'Could not find API object at path "{0}"'.format(
                    self.__fanref))

        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            return True

        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            return (name, value, color)
        self.__properties = list(dict_to_properties(
            target=self, d=faninfo,
                criteria=filter_properties,
                mapping=map_properties))
        properties = [
            'status_state', 'status_health', 'name', 'memberid', 'readingunits',
                'lowerthresholdnoncritical', 'reading', 'lowerthresholdcritical']

        def compare_properties(a):
            try:
                i = properties.index(a.name.lower())
            except ValueError:
                i = len(properties)
            return i
        self.__properties.sort(key=compare_properties)

    @property
    def __getitem__(self, key):
        # Chassis fan has no targets at this time
        # (behaves as an empty dictionary)
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties is None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class ChassisFansTarget(Target):

    def __init__(self, fansref, address, context):
        super(ChassisFansTarget, self).__init__(
            address=address, context=context)
        self.__fansref = fansref
        self.__targets = None

    def __initTargets(self):
        # Enumerate chassis fans with NVSM backend
        assert self.__targets is None
        self.__targets = {}
        faninfo = {}
        try:
            faninfo = self._context._queryBackend(self.__fansref)
        except:
            pass
        # XXX: Remove this once the queryBackend() method can handle local
        # references
        faninfo = faninfo.get('Fans', [])
        if not faninfo:
            faninfo = []
        # Create a fan target for each fan in this chassis
        index = 0
        for fan in faninfo:
            # FIXME: The '@odata.id' field that the NVSM backend gives for fans
            #        is not suitable for lookup. We fudge an @odata.id here, but in
            # the future this should be obtained directly from the backend.
            fanref = '{0}/{1}'.format(self.__fansref, index)
            # FIXME: This is case-sensitive
            name = str(fan.get('Name', index))
            target = ChassisFanTarget(
                fanref=fanref,
                    address=(self.address + name),
                    context=self._context)
            self.__targets[name] = target
            index += 1

    def __getitem__(self, key):
        if self.__targets is None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets is None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?


class ChassisTemperatureTarget(Target):

    def __init__(self, temperatureref, address, context):
        super(ChassisTemperatureTarget, self).__init__(
            address=address, context=context)
        self.__temperatureref = temperatureref
        self.__properties = None

    def __initProperties(self):
        # Enumerate fan properties using NVSM backend
        assert self.__properties is None
        self.__properties = []
        temperatureinfo = self._context._queryBackend(self.__temperatureref)
        # XXX: Remove this once the _queryBackend() method is able to handle
        #      local @odata.id references.
        temperaturepath = self.__temperatureref.split(
            '#')[-1].strip('/').split('/')

        def to_int(x):
            try:
                x = int(x)
            except:
                pass
            return x
        try:
            temperatureinfo = reduce(
                lambda x,
                y: x[to_int(y)],
                temperaturepath,
                temperatureinfo)
        except (KeyError, TypeError):
            raise nvsm.exceptions.BackendError(
                'Could not find API object at path "{0}"'.format(
                    self.__temperatureref))

        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            return True

        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            return (name, value, color)
        self.__properties = list(dict_to_properties(
            target=self, d=temperatureinfo,
                criteria=filter_properties,
                mapping=map_properties))
        properties = [
            'status_state', 'status_health', 'name', 'physicalcontext', 'memberid',
                        'readingcelsius', 'upperthresholdnoncritical', 'sensornumber', 'lowerthresholdcritical']

        def compare_properties(a):
            try:
                i = properties.index(a.name.lower())
            except ValueError:
                i = len(properties)
            return i
        self.__properties.sort(key=compare_properties)

    @property
    def __getitem__(self, key):
        # Chassis fan has no targets at this time
        # (behaves as an empty dictionary)
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties is None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class ChassisTemperaturesTarget(Target):

    def __init__(self, temperaturesref, address, context):
        super(ChassisTemperaturesTarget, self).__init__(
            address=address, context=context)
        self.__temperaturesref = temperaturesref
        self.__targets = None

    def __initTargets(self):
        # Enumerate chassis temperatures with NVSM backend
        assert self.__targets is None
        self.__targets = {}
        temperaturesinfo = {}
        try:
            temperaturesinfo = self._context._queryBackend(self.__temperaturesref)
        except:
            pass
        # XXX: Remove this once the _queryBackend() method can handle local
        #      references
        temperaturesinfo = temperaturesinfo.get('Temperatures', [])
        if not temperaturesinfo:
            temperaturesinfo = []
        index = 0
        for temperature in temperaturesinfo:
            # FIXME: The '@odata.id' field that the NVSM backend gives for temperatures
            #        is not suitable for lookup. We fudge an @odata.id here, but in
            # the future this should be obtained directly from the backend.
            ref = '{0}/{1}'.format(self.__temperaturesref, index)
            # FIXME: This is case-sensitive
            name = str(temperature.get('Name', index))
            target = ChassisTemperatureTarget(
                temperatureref=ref,
                    address=(self.address + name),
                    context=self._context)
            self.__targets[name] = target
            index += 1

    def __getitem__(self, key):
        if self.__targets is None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets is None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?


class ChassisThermalTarget(Target):

    def __init__(self, thermalref, address, context, nodeIdentifier):
        super(ChassisThermalTarget, self).__init__(
            address=address, context=context)
        self.__targets = None
        self.__thermalref = thermalref
        self.__nodeIdentifier = nodeIdentifier

    def __initTargets(self):
        assert self.__targets is None
        self.__targets = {}

        # Assume every chassis has thermal alert targets (for DSHM alerts)
        alertsTarget = AlertsTarget(
            ref='/nvsm/v1/Alerts/Thermal',  # XXX
            address=(self.address + 'alerts'),
            context=self._context,
            nodeIdentifier=self.__nodeIdentifier)
        self.__targets['alerts'] = alertsTarget

        # Assume every chassis has thermal policy targets (for configuring
        # DSHM)
        policyTarget = ThermalPolicyTarget(
            # FIXME: NVSM backend does not provide OData reference here
                policyref='/nvsm/v1/Policy/Thermal',
                address=(self.address + 'policy'),
                context=self._context,
                nodeIdentifier=self.__nodeIdentifier)
        self.__targets['policy'] = policyTarget

        # Enumerate chassis thermals with NVSM backend
        # FIXME: This assumes chassis 1
        thermalinfo = {}
        try:
            thermalinfo = self._context._queryBackend(self.__thermalref)
        except:
            pass
        # FIXME: This is case-sensitive
        if 'Fans' in thermalinfo and thermalinfo.get('Fans') is not None:
            chassisFansTarget = ChassisFansTarget(
                # FIXME: This assumes chassis 1
                    fansref=self.__thermalref + '#Fans',
                    address=(self.address + 'fans'),
                    context=self._context)
            self.__targets['fans'] = chassisFansTarget
        # FIXME: This is case-sensitive
        if 'Temperatures' in thermalinfo and thermalinfo.get('Temperatures') is not None:
            chassisTemperaturesTarget = ChassisTemperaturesTarget(
                # FIXME: This assumes chassis 1
                    temperaturesref=self.__thermalref + '#Temperatures',
                    address=(self.address + 'temperatures'),
                    context=self._context)
            self.__targets['temperatures'] = chassisTemperaturesTarget

    def __getitem__(self, key):
        if self.__targets is None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets is None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?
