# Copyright (c) 2018-2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.




import time
import json
import os

from nvsm.exceptions import BackendError
from nvsm.backend.context import Target, Property, dict_to_properties
import nvsm.prompt
import nvsm.log

from nvsm.backend import line_buffered_stream

def find_fwuc(fwuc_name):
    #
    # Find given fwuc image in local repository. Every FWUC image should have matching
    # "fwuc_build" label
    cmd = "docker images --filter \"label=fwc_build\" --format \"{{.Repository}}:{{.Tag}}\""

    try:
        data = os.popen(cmd).read().strip().split('\n')
        for each in data:
            if fwuc_name == each:
                return True
    except:
        pass

    return False

def collect_fw_versions(context):
    versions = []

    # Get SBIOS version
    try:
        system_info = context._queryBackend('/redfish/v1/Systems/1')
        versions.append(('sbios_version', system_info['BiosVersion']))
    except:
        versions.append(('sbios_version', 'Unknown'))

    # Get VBIOS version
    try:
        gpu_info = context._queryBackend('/nvsm/v1/Systems/1/GPUs/GPU0')
        versions.append(('vbios_version', gpu_info['Versions']['VBIOSVersion']))
    except:
        versions.append(('vbios_version', 'Unknown'))

    # TODO: Get BMC FW version
    versions.append(('bmc_version', 'Unknown'))

    # TODO: Get NVMe firmware version (several vendors)
    versions.append(('nvme_version', 'Unknown'))

    # TODO: Get PSU FW version
    versions.append(('psu_version', 'Unknown'))

    return versions

def cancel_existing_prompt(ref, context):
    # Before starting any job on API server, the CLI will attempt
    # to cancel any existing job stuck in the "prompt" state. This will
    # prevent lockups whenever another API client starts a job but never
    # responds to the prompt.
    response = context._queryBackend(ref, method='GET')
    try:
        if response['Status']['State'].lower() == 'prompt':
            context._queryBackend(
                ref, method='PUT',
                body={
                    'State': 'Cancel',
                    'JobID': response['Status']['JobID']
                })
            time.sleep(2.0)
    except KeyError:
        pass

class FetchDockerTarget(Target):
    ''' Control the download process image. '''
    def __init__(self, ref, address, context):
        super(FetchDockerTarget, self).__init__(
                address=address, context=context)
        self.__ref = ref
        self.__properties = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        response = self._context._queryBackend(self.__ref)
        # Show properties related to Docker pull
        docker_properties = [
          'dockerusername',   # note using as lowercase below in compare_properties()
          'dockerpassword',
          'dockerimageref',
          'dockermode']
        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            #
            # DockerMode is internal flag used by FW CLI. No need to display
            if 'dockermode' in name.lower():
                return False

            if name.lower() not in docker_properties:    # <--- LOWER CASE COMPARE!
                return self._context.showall
            return True
        self.__properties.extend(list(dict_to_properties(
                    target=self, d=response,
                    criteria=filter_properties)))

        self.__properties.append(Property(
                    target=self,
                    name='ImageRefIsLocal',
                    value=find_fwuc(response.get('DockerImageRef', 'Unknown'))))

        # Display properties in order given by docker_properties array
        def compare_properties(a):
            try:
                i = docker_properties.index(a.name.lower())
            except ValueError:
                i = len(docker_properties)
            return i
        self.__properties.sort(key=compare_properties)
        # Show properties related to download progress
        response_status = response.get('Status', {})
        state = response_status.get('State', 'Unknown')
        message = 'N/A'
        progress = 'N/A'
        if state.lower() in [ 'downloading' ]:
            message = response_status.get('Message', 'Unknown')
            progress = response_status.get('Progress')
            if not progress:
                progress = 'Unknown'
            else:
                progress = '{:.2f}%'.format(progress)
        elif state.lower() in [ 'installing' ]:
            state = 'N/A (installing)'
        self.__properties.append(Property(
                    target=self,
                    name='State',
                    value=state))
        self.__properties.append(Property(
                    target=self,
                    name='Message',
                    value=message))
        self.__properties.append(Property(
                    target=self,
                    name='Progress',
                    value=progress))

    def __getitem__(self, key):
        raise KeyError  # No sub-targets

    @property
    def targets(self):
        return []  # No sub-targets

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties upon first access
            self.__initProperties()
        return self.__properties

    def set(self, name, value):
        for x in [ 'DockerUsername', 'DockerPassword', 'DockerImageRef', 'DockerMode' ]:
            if name.lower() == x.lower():
                self._context._queryBackend(self.__ref,
                        method='PUT', body={ x: value })
                return
        if name.lower() == 'state':
            body = { 'State': value }
            response = self._context._queryBackend(self.__ref)
            if 'JobID' in response.get('Status', {}):
                body['JobID'] = response['Status']['JobID']
            self._context._queryBackend(self.__ref, method='PUT', body=body)
            time.sleep(2.0)
            return
        if name.lower() in [ 'message', 'progress' ]:
            raise nvsm.exceptions.MissingProperty(
                'Verb "set" not supported for "{0}" target property'.format(name))
        raise nvsm.exceptions.MissingProperty(
            'Target property "{0}" does not exist'.format(name))

    def start(self, skip_prompt=True, wait=False):
        class DockerFetchJob(object):
            def __init__(self, target, statusref):
                super(DockerFetchJob, self).__init__()
                self.__target = target
                self.__statusref = statusref
                self.__status_stream = None

            @property
            def uninterruptible(self):
                # Fetch jobs can run in background
                return False

            @property
            def status_stream(self):
                if self.__status_stream == None:
                    # Obtain file handle for job status streaming API
                    stream = self.__target._context._queryBackend(
                            self.__statusref, stream=True)
                    # Apply line buffering to the status stream
                    self.__status_stream = line_buffered_stream(stream)
                return self.__status_stream

        cancel_existing_prompt(ref=self.__ref, context=self._context)

        # check to see that we have expected data -- name/password/image
        response = self._context._queryBackend(self.__ref)
        for key in [ 'DockerUsername', 'DockerPassword', 'DockerImageRef', 'DockerMode' ]:
            value = response.get(key, None)
            if value == None or value == "":
                raise nvsm.exceptions.MissingProperty("Missing or empty property for '%s'" %(key))

        # if local image (set this way in cli), then don't pull down any image
        # expect it to be local

        username = response.get('DockerUsername', None)
        if username == "local":
            nvsm.log.debug("'local' user name, assume image is already local, no fetch")
            return 0

        # making an assumption here: This code wil be executed for one of two reasons
        #
        #  1) it's being called from the nvsm-cli that set/valdidate everything
        #  2) it's being called because of a smash command from an "expert" user
        #
        # in both cases, we must assume that the image/username/password are good,
        # If they are wrong then no prompting as a secondary measure is
        # necessary -- the caller screwed up: simply indicate error and stop
        #
        # Note: Orignally this code use to prompt, don't do that any mode since
        #       code is smarter in cli, and we expect those who use the scripting
        #       smash interface 1) to be smart, and 2) not to want to prompt.

        # Ask API server to start FW download
        body = {'State': 'Fetch'}
        put_response = self._context._queryBackend(self.__ref, method='PUT', body=body)

        # Return handle to image fetch job
        return DockerFetchJob(target=self, statusref=put_response.get('Status'))

class InstallDockerTarget(Target):
    ''' Control the installation process of DGX system container. '''
    def __init__(self, ref, address, context):
        super(InstallDockerTarget, self).__init__(
                address=address, context=context)
        self.__ref = ref
        self.__properties = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        response = self._context._queryBackend(self.__ref)
        # Show Docker container image used for install (same as for fetch)
        self.__properties.append(Property(
                    target=self,
                    name='DockerImageRef',
                    value=response.get('DockerImageRef', 'Unknown')))
        self.__properties.append(Property(
                    target=self,
                    name='DockerUsername',
                    value=response.get('DockerUsername', 'Unknown')))
        self.__properties.append(Property(
                    target=self,
                    name='DockerPassword',
                    value=response.get('DockerPassword', 'Unknown')))

        self.__properties.append(Property(
                    target=self,
                    name='ImageRefIsLocal',
                    value=find_fwuc(response.get('DockerImageRef', 'Unknown'))))

        # Skip DockerMode
        #self.__properties.append(Property(
        #            target=self,
        #            name='DockerMode',
        #            value=response.get('DockerMode', 'Unknown')))
        # Show properties related to install progress
        response_status = response.get('Status', {})
        state = response_status.get('State', 'Unknown')
        message = 'N/A'
        progress = 'N/A'
        if state.lower() in [ 'installing' ]:
            message = response_status.get('Message', 'Unknown')
            progress = response_status.get('Progress')
            if not progress:
                progress = 'Unknown'
            else:
                progress = '{:.2f}%'.format(progress)
        elif state.lower() in [ 'downloading' ]:
            state = 'N/A (downloading)'
        self.__properties.append(Property(
                    target=self,
                    name='State',
                    value=state))
        self.__properties.append(Property(
                    target=self,
                    name='Message',
                    value=message))
        self.__properties.append(Property(
                    target=self,
                    name='Progress',
                    value=progress))
        # Show install flags
        self.__properties.append(Property(
                    target=self,
                    name='Flags',
                    value=response.get('Flags', '')))

    def __getitem__(self, key):
        # TODO: Add a target for viewing contianer installation logs?
        # NOTE: Docker installation logs are also stored in syslog
        raise KeyError  # No sub-targets

    @property
    def targets(self):
        return []  # No sub-targets

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties upon first access
            self.__initProperties()
        return self.__properties

    def set(self, name, value):
        for x in [ 'DockerUsername', 'DockerPassword', 'DockerImageRef', 'DockerMode', 'Flags' ]:
            if name.lower() == x.lower():
                self._context._queryBackend(self.__ref,
                        method='PUT', body={ x: value })
                return
        if name.lower() == 'state':
            body = { 'State': value }
            response = self._context._queryBackend(self.__ref)
            if 'JobID' in response.get('Status', {}):
                body['JobID'] = response['Status']['JobID']
            self._context._queryBackend(self.__ref, method='PUT', body=body)
            time.sleep(2.0)
            return
        if name.lower() in [ 'message', 'progress' ]:
            raise nvsm.exceptions.MissingProperty(
                'Verb "set" not supported for "{0}" target property'.format(name))
        raise nvsm.exceptions.MissingProperty(
            'Target property "{0}" does not exist'.format(name))

    def start(self, skip_prompt=False, wait=False):
        class DockerInstallJob(object):
            def __init__(self, target, ref, skip_prompt):
                super(DockerInstallJob, self).__init__()
                self.__target = target
                self.__ref = ref
                self.__skip_prompt = skip_prompt

                self.__statusref = None
                self.__status_stream = None
                self.__existing_job = False

            def start(self):
                # Check for existing job before starting
                response = self.__target._context._queryBackend(self.__ref)
                if response.get('Status', {}).get('State').lower() in [ 'running', 'installing' ]:
                    self.__existing_job = True

                # Ask API server to start FW install
                response = self.__target._context._queryBackend(
                    self.__ref, method='PUT', body={'State': 'Install'})
                self.__statusref = response.get('Status')

            @property
            def uninterruptible(self):
                # FW install jobs cannot run in background
                # (this is UX feature, only to avoid confusing the user)
                return True

            def __init_status_stream(self):
                assert self.__status_stream == None
                assert self.__statusref != None
                # Obtain file handle for job status streaming API
                stream = self.__target._context._queryBackend(
                        self.__statusref, stream=True)
                # Apply line buffering to the status stream
                buffered = line_buffered_stream(stream)
                # Do not prompt if FW install job is already running
                if self.__existing_job:
                    self.__status_stream = buffered
                    return
                # Prompt user to confirm/cancel contianer installation
                def status_stream_with_prompt(
                        buffered=buffered,
                        context=self.__target._context,
                        ref=self.__ref):
                    for line in buffered:
                        try:
                            data = json.loads(line)

                            # Bug 2831668 - Skip object marked for timeout
                            #
                            if data['State'].lower() == 'timeout':
                                continue
                            if data['State'].lower() == 'prompt':
                                if 'JobID' not in data:
                                    raise nvsm.exceptions.BackendError(
                                        'NVSM backend did not give job ID for prompt')
                                if 'Message' not in data:
                                    raise nvsm.exceptions.BackendError(
                                        'NVSM backend did not give message for prompt')
                                if self.__skip_prompt:
                                    context._queryBackend(
                                            ref, method='PUT',
                                            body={
                                                'State': 'Confirm',
                                                'JobID': data['JobID']
                                            })
                                    yield line
                                    continue
                                try:
                                    response = nvsm.prompt.Confirmation(
                                            message=data['Message'],
                                            question='Start installation?',
                                            formatted=False)
                                except EOFError:
                                    # Upon receiving EOF, cancel the job and leave quickly
                                    context._queryBackend(
                                            ref, method='PUT',
                                            body={
                                                'State': 'Cancel',
                                                'JobID': data['JobID']
                                            })
                                    raise
                                if response:
                                    context._queryBackend(
                                            ref, method='PUT',
                                            body={
                                                'State': 'Confirm',
                                                'JobID': data['JobID']
                                            })
                                else:
                                    context._queryBackend(
                                            ref, method='PUT',
                                            body={
                                                'State': 'Cancel',
                                                'JobID': data['JobID']
                                            })
                                continue
                        except KeyError:
                            yield line
                        else:
                            yield line
                self.__status_stream = status_stream_with_prompt()

            @property
            def status_stream(self):
                if self.__status_stream == None:
                    self.__init_status_stream()
                return self.__status_stream

        # Cancel any "zombie" prompt to avoid confusing the user
        # NOTE: These pre-existing prompts can occur if another user initiates
        #       a FW install job, but leaves the prompt idle without ever
        #       confirming the job.
        cancel_existing_prompt(ref=self.__ref, context=self._context)

        # Create and start install job
        job = DockerInstallJob(
                target=self,
                ref=self.__ref,
                skip_prompt=skip_prompt)
        job.start()

        # Return handle to  install job
        return job

class DockerTarget(Target):
    ''' Target to view, download, and install images. '''
    def __init__(self, systemref, address, context, ref='nvsm/v1/Systems/1/Firmware', showproperties=None):
        super(DockerTarget, self).__init__(
                address=address, context=context)
        self.__ref = ref
        self.__properties = None
        self.__targets = None
        self.__showproperties = showproperties

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}
        # Add a SMASH target to control the image download process
        fetchTarget = FetchDockerTarget(
                ref=self.__ref,
                address=(self.address + 'fetch'),
                context=self._context)
        self.__targets['fetch'] = fetchTarget
        # Add a SMASH target to control the installation process
        installTarget = InstallDockerTarget(
                ref=self.__ref,
                address=(self.address + 'install'),
                context=self._context)
        self.__targets['install'] = installTarget

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        if self.__showproperties:
            response = self._context._queryBackend(self.__ref)
            prop = self.__showproperties
            def filter_properties(name, value):
                if not prop:
                    return False
                if name.lower() in prop:
                    return True
                return False
            self.__properties.extend(list(dict_to_properties(
                        target=self, d=response,
                    criteria=filter_properties)))
#        # List installed firmware versions
#        for name, version in collect_fw_versions(self._context):
#            p = Property(target=self, name=name, value=version)
#            self.__properties.append(p)

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties upon first access
            self.__initProperties()
        return self.__properties
