# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



import nvsm.log
import nvsm.odata
from nvsm.backend.context import Property, Target, dict_to_properties

from nvsm.backend.alerts import AlertsTarget
from nvsm.backend.policy import GPUPolicyTarget
from nvsm.exceptions import BackendError


bad_status = [ 'warning', 'critical' ]
good_status = [ 'ok' ]
def colorize_status(value):
    if value == None:
        return "none"
    if str(value).lower() in good_status:
        return 'green'
    if str(value).lower() in bad_status:
        return 'red'
    return 'none'


class GPUHealthIncidentTarget(Target):
    def __init__(self, incident, address, context):
        super(GPUHealthIncidentTarget, self).__init__(
                address=address, context=context)
        self.__incident = incident
        self.__properties = None

    def __initProperties(self):
        def map_properties(name, value):
            color = 'none'
            if 'health' in name.strip().lower():
                color = colorize_status(value)
            return (name, value, color)
        self.__properties = list(dict_to_properties(
                target=self, d=self.__incident,
                mapping=map_properties))

    def __getitem__(self, key):
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties upon first access
            self.__initProperties()
        return self.__properties


class GPUHealthTarget(Target):
    def __init__(self, health, address, context):
        super(GPUHealthTarget, self).__init__(
                address=address, context=context)
        self.__health = health
        self.__targets = None
        self.__properties = None

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}
        # Create target for each GPU health incident
        incidents = self.__health.get('Incidents', [])
        for index in range(len(incidents)):
            name = 'incident{}'.format(index)
            target = GPUHealthIncidentTarget(
                    incident=incidents[index],
                    address=(self.address + name),
                    context=self._context)
            self.__targets[name] = target

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            if name.lower() in [ 'health' ]:
                return True
            return self._context.showall
        def map_properties(name, value):
            color = 'none'
            if 'health' in name.strip().lower():
                color = colorize_status(value)
            return (name, value, color)
        self.__properties = list(dict_to_properties(
                target=self, d=self.__health,
                criteria=filter_properties,
                mapping=map_properties))

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties upon first access
            self.__initProperties()
        return self.__properties


class GPUTarget(Target):
    def __init__(self, gpu, address, context):
        super(GPUTarget, self).__init__(address=address, context=context)
        self.__gpu = gpu
        self.__properties = None
        self.__targets = None

    def __initProperties(self):
        assert self.__properties == None
        # Squash hierarchies (SMASH does not support proprety hierarchies)
        default_properties = [
            'inventory_brandname',
            'inventory_modelname',
            'inventory_pciedeviceid',
            'inventory_pciesubsystemid',
            'inventory_partnumber',
            'inventory_serialnumber',
            'inventory_uuid',
            'specifications_computecapability',
            'specifications_maxpciegen',
            'specifications_maxpcielinkwidth',
            'specifications_maxspeeds_graphicsclock',
            'specifications_maxspeeds_memclock',
            'specifications_maxspeeds_smclock',
            'specifications_maxspeeds_videoclock',
            'specifications_memorysize',
            'versions_vbiosversion',
            'connections_pciegen',
            'connections_pcielinkwidth',
            'connections_pcielocation',
            'power_powerdraw',
            'stats_errorstats_eccmode',
            'stats_framebuffermemoryusage_free',
            'stats_framebuffermemoryusage_total',
            'stats_framebuffermemoryusage_used',
            'stats_pcierxthroughput',
            'stats_pcietxthroughput',
            'stats_performancestate',
            'stats_utildecoder',
            'stats_utilencoder',
            'stats_utilgpu',
            'stats_utilmemory',
            'status_health',
            'status_healthrollup',
            'status_state'
        ]
        hidden_properties = [
            'clockspeeds_defaultapplicationtimespeeds_graphicsclock',
            'clockspeeds_defaultapplicationtimespeeds_memclock',
            'clockspeeds_currentspeed_smclock',
            'clockspeeds_currentspeed_graphicsclock',
            'clockspeeds_currentspeed_memclock',
            'clockspeeds_currentspeed_videoclock',
            'clockspeeds_usersetapplicationstimespeeds_graphicsclock',
            'clockspeeds_usersetapplicationstimespeeds_memclock',
            'stats_throttlereasons',
            'stats_pciereplaycounter',
            'stats_usagememoryfb',
            'stats_bar1memoryusage_used',
            'stats_bar1memoryusage_free',
            'stats_bar1memoryusage_total',
            'stats_nvlinkcountersperlink_0_nvml_nvlink_error_dl_crc_flit',
            'stats_nvlinkcountersperlink_0_nvml_nvlink_error_dl_crc_data',
            'stats_nvlinkcountersperlink_0_dcgmcm_nvlink_counter_bytes',
            'stats_nvlinkcountersperlink_0_nvml_nvlink_error_dl_replay',
            'stats_nvlinkcountersperlink_0_nvml_nvlink_error_dl_recovery',
            'stats_nvlinkcountersperlink_1_nvml_nvlink_error_dl_crc_flit',
            'stats_nvlinkcountersperlink_1_nvml_nvlink_error_dl_crc_data',
            'stats_nvlinkcountersperlink_1_dcgmcm_nvlink_counter_bytes',
            'stats_nvlinkcountersperlink_1_nvml_nvlink_error_dl_replay',
            'stats_nvlinkcountersperlink_1_nvml_nvlink_error_dl_recovery',
            'stats_nvlinkcountersperlink_2_nvml_nvlink_error_dl_crc_flit',
            'stats_nvlinkcountersperlink_2_nvml_nvlink_error_dl_crc_data',
            'stats_nvlinkcountersperlink_2_dcgmcm_nvlink_counter_bytes',
            'stats_nvlinkcountersperlink_2_nvml_nvlink_error_dl_replay',
            'stats_nvlinkcountersperlink_2_nvml_nvlink_error_dl_recovery',
            'stats_nvlinkcountersperlink_3_nvml_nvlink_error_dl_crc_flit',
            'stats_nvlinkcountersperlink_3_nvml_nvlink_error_dl_crc_data',
            'stats_nvlinkcountersperlink_3_dcgmcm_nvlink_counter_bytes',
            'stats_nvlinkcountersperlink_3_nvml_nvlink_error_dl_replay',
            'stats_nvlinkcountersperlink_3_nvml_nvlink_error_dl_recovery',
            'stats_userprocesses',
            'stats_violations_thermal_referencetime',
            'stats_violations_thermal_violationtime',
            'stats_violations_power_referencetime',
            'stats_violations_power_violationtime',
            'stats_violations_sync_boost_referencetime',
            'stats_violations_sync_boost_violationtime',
            'stats_violations_boardlimit_referencetime',
            'stats_violations_boardlimit_violationtime',
            'stats_violations_reliability_referencetime',
            'stats_violations_reliability_violationtime',
            'stats_violations_totalappclocks_referencetime',
            'stats_violations_totalappclocks_violationtime',
            'stats_errorstats_errorslifetime_doublebiterrors_texturememory',
            'stats_errorstats_errorslifetime_doublebiterrors_l2cache',
            'stats_errorstats_errorslifetime_doublebiterrors_registerfile',
            'stats_errorstats_errorslifetime_doublebiterrors_devicememory',
            'stats_errorstats_errorslifetime_doublebiterrors_l1cache',
            'stats_errorstats_errorslifetime_doublebiterrors_textureshm',
            'stats_errorstats_errorslifetime_doublebiterrors_total',
            'stats_errorstats_errorslifetime_singlebiterrors_texturememory',
            'stats_errorstats_errorslifetime_singlebiterrors_l2cache',
            'stats_errorstats_errorslifetime_singlebiterrors_registerfile',
            'stats_errorstats_errorslifetime_singlebiterrors_devicememory',
            'stats_errorstats_errorslifetime_singlebiterrors_l1cache',
            'stats_errorstats_errorslifetime_singlebiterrors_textureshm',
            'stats_errorstats_errorslifetime_singlebiterrors_total',
            'stats_errorstats_retiredpages_pendingretirementcount',
            'stats_errorstats_retiredpages_duetodoublebiterrors_pageaddresses',
            'stats_errorstats_retiredpages_duetodoublebiterrors_pagecount',
            'stats_errorstats_retiredpages_duetomultiplesinglebiterrors_pageaddresses',
            'stats_errorstats_retiredpages_duetomultiplesinglebiterrors_pagecount',
            'stats_errorstats_errorssincedriverload_doublebiterrors_texturememory',
            'stats_errorstats_errorssincedriverload_doublebiterrors_l2cache',
            'stats_errorstats_errorssincedriverload_doublebiterrors_registerfile',
            'stats_errorstats_errorssincedriverload_doublebiterrors_devicememory',
            'stats_errorstats_errorssincedriverload_doublebiterrors_l1cache',
            'stats_errorstats_errorssincedriverload_doublebiterrors_textureshm',
            'stats_errorstats_errorssincedriverload_doublebiterrors_total',
            'stats_errorstats_errorssincedriverload_singlebiterrors_texturememory',
            'stats_errorstats_errorssincedriverload_singlebiterrors_l2cache',
            'stats_errorstats_errorssincedriverload_singlebiterrors_registerfile',
            'stats_errorstats_errorssincedriverload_singlebiterrors_devicememory',
            'stats_errorstats_errorssincedriverload_singlebiterrors_l1cache',
            'stats_errorstats_errorssincedriverload_singlebiterrors_textureshm',
            'stats_errorstats_errorssincedriverload_singlebiterrors_total',
            'stats_errorstats_eccmodepending',
            'thermals_fanspeed',
            'thermals_gputemperaturethresholdslow',
            'thermals_gputemperaturecurrent',
            'thermals_gputemperaturethresholdmax',
            'power_powerlimitenforced',
            'power_powerlimitmin',
            'power_powerlimit',
            'power_powerdraw',
            'power_powermanagementsupport',
            'power_powerlimitdefault',
            'power_powerlimitmax',
            'power_powerstate',
            'power_powemanagementsupport',
            'specifications_supportedspeeds_0_supportedgraphicsclocks',
            'specifications_supportedspeeds_0_formemclock',
            'specifications_numberofcores',
            'specifications_productfamily',
            'specifications_memorytype',
            'specifications_gpuarchitecture',
            'specifications_pciebridgechiptype',
            'connections_driverassignedboardid',
            'connections_displayattached',
            'connections_deviceminornumber',
            'connections_multigpuboard',
            'connections_cpuaffinitybitmap_cpus_64_127',
            'connections_cpuaffinitybitmap_cpus_128_191',
            'connections_cpuaffinitybitmap_cpus_0_63',
            'connections_cpuaffinitybitmap_cpus_192_255',
            'health_incidents',
            'health_health',
            'featureconfigurations_clockpolicy_autoboostbydefault',
            'featureconfigurations_clockpolicy_autoboost',
            'featureconfigurations_displayactive',
            'featureconfigurations_accountingbuffersize',
            'featureconfigurations_persistencemode',
            'featureconfigurations_virtualizationmode',
            'featureconfigurations_computemode',
            'featureconfigurations_displaymode',
            'featureconfigurations_operationmodeonnextboot',
            'featureconfigurations_accountingmode',
            'featureconfigurations_operationmode',
            'versions_inforomversion',
            'versions_inforompowerobjectversion',
            'versions_inforomimageversion',
            'versions_pciebridgechipfw',
            'versions_inforomchecksumvalidation',
            'versions_inforomchecksum',
            'versions_inforomeccobjectversion',
            'specifications_supportedspeeds',
            'stats_throttlereasons_swpowercap',
            'stats_throttlereasons_applicationsclockssetting',
            'stats_throttlereasons_hwslowdown',
            'stats_throttlereasons_swthermalslowdown',
            'stats_throttlereasons_syncboost',
            'stats_throttlereasons_hwpowerbrakeslowdown',
            'stats_throttlereasons_displayclocksetting',
            'stats_throttlereasons_hwthermalslowdown',
            'stats_throttlereasons_gpuidle',
            'stats_nvlinkcountersperlink',
            'stats_userprocesses',
            'stats_violations_syncboost_referencetime',
            'stats_violations_syncboost_violationtime',
            'stats_violations_lowutilization_referencetime',
            'stats_violations_lowutilization_violationtime',
            'stats_violations_totalbaseclocks_referencetime',
            'stats_violations_totalbaseclocks_violationtime',
        ]
        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            if name.lower() in default_properties:
                return True
            if name.lower() in hidden_properties:
                return self._context.showall  # TODO: Provide option to show hidden properties
            return self._context.showall
        def map_properties(name, value):
            color = 'none'
            if 'health' in name.strip().lower():
                color = colorize_status(value)
            elif 'supportedgraphicsclocks' in name.strip().lower():
                value = ', '.join(value)
                value = '[ {} ]'.format(value)
            return (name, value, color)
        self.__properties = list(dict_to_properties(
                target=self, d=self.__gpu,
                criteria=filter_properties,
                mapping=map_properties))
        # Display properties in order given by default_properties array
        def compare_properties(a):
            try:
                i = default_properties.index(a.name.lower())
            except ValueError:
                try:
                    i = len(default_properties) + hidden_properties.index(a.name.lower())
                except ValueError:
                    i = len(default_properties) + len(hidden_properties)
            return i
        self.__properties.sort(key=compare_properties)

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}
        # Optionally include GPU health target
        if 'Health' in self.__gpu:
            target = GPUHealthTarget(
                    health=self.__gpu['Health'],
                    address=(self.address + 'health'),
                    context=self._context)
            self.__targets['health'] = target

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties upon first access
            self.__initProperties()
        return self.__properties


class GPUCollectionTarget(Target):
    def __init__(self, systemref, address, context):
        super(GPUCollectionTarget, self).__init__(
                address=address, context=context)
        self.__systemref = systemref
        self.__targets = None
        self.__properties = None

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}

        try:
            gpuinfo = self._context._queryBackend(self.__systemref.replace("redfish", "nvsm")+"/GPUs")
            gpus = nvsm.odata.OData.from_value(gpuinfo, self._context)
        except BackendError as e:
            nvsm.log.error('{0}', e)
        else:
            # Enumerate GPUs with NVSM backend
            for gpu in gpus.get('Members', []):
                name = gpu.get('Id', default=None, forceRefresh=True)
                if not name:
                    continue
                address = self.address + name
                target = GPUTarget(
                        gpu=gpu,
                        address=address, context=self._context)
                self.__targets[name] = target
            # Optionally include GPU health target
            if 'Health' in gpus:
                target = GPUHealthTarget(
                        health=gpus['Health'],
                        address=(self.address + 'health'),
                        context=self._context)
                self.__targets['health'] = target

        # Assume every system has GPU alert targets (for DSHM alerts)
        alertsTarget = AlertsTarget(
            ref='/nvsm/v1/Alerts/GPUs',
            address=(self.address + 'alerts'),
            context=self._context)
        self.__targets['alerts'] = alertsTarget

        # Assume every system has GPU policy target (for configuring DSHM)
        policyTarget = GPUPolicyTarget(
                policyref='/nvsm/v1/Policy/GPUs',
                address=(self.address + 'policy'),
                context=self._context)
        self.__targets['policy'] = policyTarget

    def __initProperties(self):
        assert self.__properties == None
        gpuinfo = self._context._queryBackend('/nvsm/v1/Systems/1/GPUs')
        # Include overall GPU health status in properties
        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            if name.lower() in [ 'status_health', 'status_healthrollup',
                    'status_state' ]:
                return True
            return self._context.showall
        def map_properties(name, value):
            color = 'none'
            if 'health' in name.strip().lower():
                color = colorize_status(value)
            return (name, value, color)
        self.__properties = list(dict_to_properties(
                target=self, d=gpuinfo,
                criteria=filter_properties,
                mapping=map_properties))
        properties = [ 'status_healthrollup', 'status_health']
        def compare_properties(a):
            try:
                i = properties.index(a.name.lower())
            except ValueError:
                i = len(properties)
            return i
        self.__properties.sort(key=compare_properties)

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties upon first access
            self.__initProperties()
        return self.__properties
