# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



import http.client
import json
import yaml
import requests

import nvsm.exceptions
import nvsm.log
import urllib3
import os

urllib3.disable_warnings()

class NVSMHandle(object):
    def __init__(self,
            token='/etc/nvsm/nvsm-perpetual.jwt',
            nvsmapisconfigfile='/etc/nvsm/nvsm.config'):
        super(NVSMHandle, self).__init__()
        os.environ['NO_PROXY'] = 'localhost, 127.0.0.1'
        self.__tokenfile = os.environ.get('PERPETUAL_KEY_FILE', token)
        self.__nvsmapisconfigfile = os.environ.get('NVSM_APIS_CONFIG_FILE', nvsmapisconfigfile)
        self.__nvsmapisconfiginfo = None
        self.__baseUrl = None
        self.__ca_cert_file = None

        self.__parseNvsmApisConfigFile()
        self.__prepareToken()

    @property
    def configinfo(self):
        return self.__nvsmapisconfiginfo

    def __parseNvsmApisConfigFile(self):
        data = None
        try:
            with open(self.__nvsmapisconfigfile) as file_handle:
                data = file_handle.read()
        except IOError as e:
            raise nvsm.exceptions.BackendError(
                    'Could not read NVSM API nvsmapisconfig file: {0}'.format(e))

        self.__nvsmapisconfiginfo = yaml.safe_load(data)
        nvsm.log.debug('Successfully read NVSM API nvsmapisconfig file' )

        scheme = self.__nvsmapisconfiginfo.get('scheme', 'https')
        host = self.__nvsmapisconfiginfo.get('bindaddress', '127.0.0.1')
        port = str(self.__nvsmapisconfiginfo.get('servingport', 273))
        self.__ca_cert_file  = str(self.__nvsmapisconfiginfo.get('ca_cert', '/etc/nvsm/nvsm-ca.crt'))
        self.__baseUrl = '{0}://{1}:{2}/'.format(
                                (scheme != '' and scheme or 'https'),
                                (host != '' and host or '127.0.0.1'),
                                (port != '' and port or '273'))
        nvsm.log.debug('Base URL {0}', self.__baseUrl)

    def __prepareToken(self):
        self.__tokenstring = None
        nvsm.log.debug('Attempting to read NVSM API token file: {0}',
                self.__tokenfile)
        try:
            with open(self.__tokenfile, 'r') as jwt_file:
                self.__tokenstring = jwt_file.read()
        except IOError as e:
            raise nvsm.exceptions.BackendError(
                    'Could not read NVSM API token: {0}'.format(e))
        nvsm.log.debug('Successfully read NVSM API token')

    def connect(self):
        pass  # TODO: Check if we are able to connect to the NVSM backend

    def query(self, path, method='GET', body=None, stream=False, raw=False, forceRefresh=False):
        if method not in [ 'GET', 'POST', 'PUT' , 'PATCH']:
            raise ValueError('Unknown HTTP method "{}"'.format(method))
        url = "{0}{1}".format(self.__baseUrl, path.lstrip('/'))
        nvsm.log.debug('Query NVSM backend: {method} stream={stream} {path}\n{body}',
                path=path, method=method, body=str(body), stream=stream)

        # FIXME: Python requests API suggests closing streams if they are not
        #        read to the end.

#        from httplib import HTTPConnection
#        HTTPConnection.debuglevel = 1  # XXX

        # TODO: Select requests method based on 'method' parameter
        import urllib3
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)  # FIXME
        #Ignore warnings for Insecure requests from New Requets package versions starting 2.7.9
        import warnings
        try:
            from requests.packages.urllib3.exceptions import InsecureRequestWarning
        #Ignore import error for older requests package
        except ImportError as error:
            pass       

        headers = {
            'Authorization': 'Bearer {0}'.format(self.__tokenstring)
        }
        try:
            if forceRefresh:
                queryParams = {'ForceRefresh':'true'}
            else:
                queryParams = {'ForceRefresh':'false'}
            args = {
                'verify': self.__ca_cert_file,  # Check the SSL certificatee
                'headers': headers
            }
            if raw:
                args['data'] = body
            else:
                args['json'] = body
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=InsecureRequestWarning)
                if method == 'GET':
                    args['stream'] = stream
                    r = requests.get(url, params=queryParams, **args)
                elif method == 'POST':
                    r = requests.post(url, params=queryParams, **args)
                elif method == 'PUT':
                    r = requests.put(url, params=queryParams, **args)
                elif method == 'PATCH':
                    r = requests.patch(url, params=queryParams, **args)
        except requests.exceptions.ConnectionError as e:
            raise nvsm.exceptions.BackendError(
                    'Error connecting to NVSM backend: {0}'.format(e))

        if not stream:
            nvsm.log.debug('Response from NVSM backend:\n{0}', r.text)

        if r.status_code not in [ requests.codes.ok, requests.codes.created, requests.codes.accepted, requests.codes.no_content]:
            try:
                # If available, use message provided in JSON response
                data = r.json()
                msg = None

                if data is not None:
                    if 'Message' in data:
                        msg = data['Message']
                    elif 'error' in data:
                            if 'Message' in data['error']:
                                msg = data['error']['Message']
                            if 'message' in  data['error']:
                                msg = data['error']['message']
                    else:
                        msg = 'Bad HTTP status code "{0}" from NVSM backend: {1}'.format(
                        r.status_code,
                        http.client.responses.get(r.status_code, 'Unknown Status Code'))
            except:
                msg = 'Bad HTTP status code "{0}" from NVSM backend: {1}'.format(
                        r.status_code,
                        http.client.responses.get(r.status_code, 'Unknown Status Code'))
            raise nvsm.exceptions.BackendError(msg, status_code=r.status_code)
        if (method == 'GET') and stream:
            return r
        try:
            data = r.json()
            return data
        except ValueError as e:
            nvsm.log.debug('Could not decode JSON response from NVSM backend: {0}'.format(e))
            return r
