# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



from nvsm.backend.context import Target, Property, dict_to_properties

import nvsm.log
import json
import sys
import subprocess
import time
from contextlib import redirect_stdout

from nvsm.backend import line_buffered_stream
from nvsm.progress import ProgressBar
from nvsmhealth import handler
import logging
log_level_map = {
                  logging.CRITICAL: "critical",
                  logging.ERROR: "error",
                  logging.WARNING: "warning",
                  logging.INFO: "info",
                  logging.DEBUG: "debug",
                }
#FIXME: timeout to be set according to platform
TIMEOUT_FOR_PROGRESS_BAR = 300
def run_new_show_health(out_file=None, client_mode=False):
    ret_code=None
    status="healthy"
    try:
        progress_bar = ProgressBar(file=sys.stdout)
    except:
        progress_bar = None
        print("Progress bar is not supported for this terminal. Please wait for results")
    output, ret_code, BomCheckCollectionSummary = get_new_show_health_output(client_mode, progress_bar)
    if ret_code:
        status="Unhealthy"

    if out_file:
        import os
        out_file = os.path.realpath(out_file)
        print("Writing health output to {}".format(out_file))
        original_stdout = sys.stdout # Save a reference to the original standard output
        with open(out_file, 'w') as f:
            sys.stdout = f
            print(output)
            sys.stdout = original_stdout
            handler.progress_complete(status, progress_bar, out_file)
    else:
        print(output)
        handler.progress_complete(status, progress_bar)

    return ret_code, BomCheckCollectionSummary

# nvsm show health in client_mode shows cached results

def get_new_show_health_output(client_mode=False, progress_bar=None):
    output_str=""
    def append_to_output(str):
        return output_str+"\n"+str
    standalone_cmd_args = ['/usr/sbin/nvsm_core', '--mode=standalone', '--method=GET', '--nocolor=true','--uri=/nvsm/v1/BOMChecks']
    client_cmd_args = ['/usr/sbin/nvsm_core', '--mode=client', '--method=GET', '--nocolor=true','--uri=/nvsm/v1/BOMChecks']
    health_cmd =  standalone_cmd_args
    if client_mode:
        health_cmd = client_cmd_args

    printable_string = {"TimeStamp"           : "Timestamp",
                        "NvsmVersion"         : "Version",
                        "ProductName"         : "Product Name",
                        "Manufacturer"        : "Manufacturer",
                        "SerialNumber"        : "Serial Number",
                        "Uptime"              : "Uptime",
                        "BIOSVersion"         : "BIOS Version",
                        "FirmwareVersion"     : "Firmware Version",
                        "IPMIVersion"         : "IPMI Version",
                        "NVIDIADriverVersion" : "NVIDIA Driver Version",
                        "ProductName"         : "Product Name",
                        "VBIOSVersion(s)"     : "VBIOS Version(s)",
                        "BaseOSVersion"       : "BaseOS Version",
                        "KernelVersion"       : "Kernel Version",
                        "OSVersion"           : "OS Version",  
                        "Result"              : "",   
                        "Version"             : "Version",
                        "BmcBootSlot"         : "BMC Slot"
                       }

    #Checks where the output is printed as mixed alternate lines
    '''   
    VBIOS Version [GPU0 [00000000:34:00.0]] [88.00.6B.00.01].............
    Inforom Storage Version [GPU0 [00000000:34:00.0]] [G504.0200.00.03]..
    VBIOS Version [GPU1 [00000000:36:00.0]] [88.00.6B.00.01].............
    Inforom Storage Version [GPU1 [00000000:36:00.0]] [G504.0200.00.03].. 
    '''

    twin_checks = {"VBIOS Version":"Inforom Storage Version"
                  }

    NUMBER_OF_SECTIONS     = 5
    NUMBER_OF_SUB_SECTIONS = 5

    bomcheck_summary_map = {
    "1_section_str"    : "Info\n----",
    "1_section"        : "RunInfo",
    "1_section_keys"   : ["TimeStamp", "NvsmVersion"],

    "2_section_str"    : "\nChecks\n------",
    "2_section"        : "BomChecks",
    #Sort the BOMCHecks here and align them to retain the old order
    "2_section_keys"   :  ["Check Memory DIMMs devices information for consistency",
                           "Verify disk controllers",
                           "Verify Network Adapters",
                           "Verify Cluster NIC active link technology consistency",
                           "Verify Storage NIC active link technology consistency",
                           "Check GPUs PCIe devices information for consistency",
                           "Check VGA PCIe devices information for consistency",
                           "Check NVSwitch PCIe devices information for consistency",
                           "Check PCIe Switches PCIe devices information for consistency",
                           "Verify DIMM part number",
                           "Verify DIMM vendors",
                           "Verify DIMM vendors consistency",
                           "Verify chassis fan presence",
                           "Check FRU information for consistency",
                           "Verify GPUs VBIOS version consistency",
                           "VBIOS Version",
                           "Quick health check of GPU using DCGM",
                           "Number of logical CPU cores",
                           "Verify GRID License Status",
                           "Verify Network IP Reachability",
                           "Status of volumes",
                           "Status of software RAID disk superblocks",
                           "Check SMART status of NVMe devices",
                           "Verify Mellanox devices firmware version consistency",
                           "Verify GPU's identified using nvidia-smi",
                           "Checking GPUs p2p topology against the output of 'nvidia-smi topo -p2p rw",
                           "Verify installed NVMe devices",
                           "Verify chassis power supply presence",
                           "Check PSU Info (Vendor, Model) for Consistency",
                           #"DCC BMC Firmware Revision",
                           #"Check DCC BMC sensor thresholds",
                           "Verify installed disks",
                           "Check for SSD health",
                           "MegaRAID storcli utility installed",
                           "BaseOS support for storcli utility",
                           "MegaRAID virtual disk state",
                           "MegaRAID physical disk state",
                           "MegaRAID PHY link speed",
                           "MegaRAID SAS address",
                           "MegaRAID SAS port valid",
                           "Checking link speed and width",
                           "Checking GPUDirect topology against the output of 'nvidia-smi topo --matrix",
                           "NVIDIA Driver Version",
                           "Checking NVLink speed",
                           "Display DCC Information",
                           #"Check DCC CAN reachability Health",
                           "Verify DCC Display Config health",
                           "Verify DCC Display Sync health",
                           "Verify DCC Gpu health",
                           "Verify DCC Ethernet port health",
                           "Verify DCC CAN health",
                           "Verify DCC USB health",
                           "Verify DCC Chassis Fan health",
                           "Display ECU Information",
                           "Verify TegraA Application health",
                           "Verify TegraB Application health",
                           "Verify TegraA Hardware health",
                           "Verify TegraB Hardware health",
                           "Check BMC sensor thresholds",
                           "Check CEC Tag Auth Errors",
                           "BMC Firmware Revision",
                           "BaseOS Version",
                           "BIOS Version",
                           "Linux kernel version",
                           "System Uptime",
                           "Serial Number",
                           "Chassis Power Supply"
                          ],

    "3_section_str"    : "\nSystem Summary\n--------------",
    "3_section"        : "SystemDetails",
    "3_section_keys"   : ["ProductName", "Manufacturer", "SerialNumber", "Uptime"],

    "3_1_section_str"  : "\nMotherBoard:",
    "3_1_section"      : "MotherBoard",
    "3_1_section_keys" : ["BIOSVersion", "SerialNumber"],

    "3_2_section_str"  : "\nBMC:",
    "3_2_section"      : "BMC",
    "3_2_section_keys" : ["FirmwareVersion", "IPMIVersion"],

    "3_3_section_str"  : "\nGPU:",
    "3_3_section"      : "GPU",
    "3_3_section_keys" : ["NVIDIADriverVersion", "ProductName", "VBIOSVersion(s)"],

    "3_4_section_str"  : "\nSoftware:",
    "3_4_section"      : 'Software',
    "3_4_section_keys" : ["OSVersion",  "BaseOSVersion", "KernelVersion"],

    # optional subsection
    "3_5_section_str"  : "\nCEC:",
    "3_5_section"      : 'CEC',
    "3_5_section_keys" : ["Version", "EcFwTag0", "EcFwTag1", "BmcBootSlot", "BMCFwAuthState"],

    "4_section_str"    :  "\nHealth Summary\n--------------\n",
    "4_section"        : "BomCheckCollectionSummary",
    "4_section_keys"   : [],

    "5_section_str"    : "\n",
    "5_section"        :  "GeneralSummary",
    "5_section_keys"   : [],
    }
    systemhealth="Overall system status is healthy"
    # Subsystem supporting dynamic checks which are not present in bomcheckcatalogue
    subsys_with_dynamic_chks = ["DccChassis"]

    def get_check_keys_for_subsystem(subsystems, bomchecks):
        checks=[]
        for check_name in bomchecks:
            for check in bomchecks[check_name]:
                for subsystem in subsystems:
                    if subsystem == check["SubSystem"]:
                        checks.append(check["CheckName"])
        return checks

    show_health_process = subprocess.Popen(health_cmd,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE,
                                   shell=False)
    if progress_bar:
        progress_bar.updateProgress(percent=0)
    poll=show_health_process.poll()
    timeout_in_secs = TIMEOUT_FOR_PROGRESS_BAR
    start_time = time.time()
    exit_loop = False
    if client_mode:
        exit_loop = True
    perc_completed=0
    while not exit_loop:
        end_time = time.time()
        perc_completed_so_far = 100*(end_time-start_time)/timeout_in_secs
        if int(perc_completed_so_far) >= int(perc_completed)+1:
            perc_completed = perc_completed_so_far
            if progress_bar:
                progress_bar.updateProgress(percent=perc_completed)
        poll=show_health_process.poll()
        if poll is not None or end_time - start_time>=timeout_in_secs:
            exit_loop = True
        time.sleep(0.1) # poll after every 0.1sec to check execution status
    
    health_out, health_err = show_health_process.communicate()
    overall_system_state_healthy = True

    if show_health_process.returncode:
        overall_system_state_healthy = False
        if health_err:
             logging.error(
                "Error occurred during nvsm show health execution {}".format(health_err))
    else:
        if health_err:
            logging.warning(
                "Warning messages during nvsm show health execution {}".format(health_err))
    bom_checks = None
    if health_out:
        #logging.debug(health_out)
        X = health_out.split(b'}',1)
        X[0] = X[0] + b'}'
        bom_check_result = json.loads(X[0])
        #output_str = append_to_output(bom_checks)
        bom_checks = json.loads(X[1])
    
    # Check if URI output contains CEC section
    print_cec_section = False
    if "CEC" in bom_checks["SystemDetails"].keys():
        print_cec_section=True

    for i in range(NUMBER_OF_SECTIONS):
        i = i + 1
        print_str = "{}_section_str".format(i)
        section_details_key = "{}_section".format(i)

        if section_details_key not in bomcheck_summary_map:
            break

        output_str = append_to_output(bomcheck_summary_map[print_str])

        section_details = bom_checks[bomcheck_summary_map[section_details_key]]
        section_keys = []
        if "{}_section_keys".format(i) in bomcheck_summary_map:
            section_keys = bomcheck_summary_map["{}_section_keys".format(i)]

        for key in section_keys:
            if key not in printable_string:
                continue
            if i == 1:
                output_str = append_to_output("{:<20s}:     {}".format(printable_string[key],section_details[key]))
            else:
                if printable_string[key]:
                    output_str = append_to_output("    {0}:{1}".format(printable_string[key],section_details[key]))
                else:     
                    output_str = append_to_output("{0}".format(section_details[key]))
        if bomcheck_summary_map[section_details_key] == "BomChecks":
           #bomcheck_sorted_keys = bomcheck_summary_map["{}_section_keys".format(i)]
           # check if there are dynamic checks
           dynamic_check_keys = get_check_keys_for_subsystem(subsys_with_dynamic_chks, bom_checks["BomChecks"])
           if dynamic_check_keys:
               section_keys+=dynamic_check_keys
           for key in section_keys:
              if key not in section_details:
                  continue

              if key in twin_checks:
                  for check_1, check_2 in zip(section_details[key],section_details[twin_checks[key]]):
                      output_str = append_to_output(check_1["FormattedResult"])
                      output_str = append_to_output(check_2["FormattedResult"])
              else:
                  for check in section_details[key]:
                      output_str = append_to_output(check["FormattedResult"])
                      if "Healthy" not in check["Result"] and check['Messages'] is not None:
                          for msg in check['Messages']:
                              output_str= append_to_output("    "+msg)
        elif bomcheck_summary_map[section_details_key] == "BomCheckCollectionSummary":
            total_checks=section_details["TotalChecks"]
            for state in bom_checks["BomCheckCollectionSummary"]:
                if state !="TotalChecks":
                    check_count = bom_checks["BomCheckCollectionSummary"][state]
                    output_str= append_to_output("{0} out of {1} checks are {2}".format(check_count, total_checks, state.lower()))
            unhealthy_checks = bom_checks["BomCheckCollectionSummary"]["UnHealthy"]
            unknown_checks = bom_checks["BomCheckCollectionSummary"]["UnKnown"]
            if unhealthy_checks > 0 or unknown_checks > 0:
                overall_system_state_healthy = False
                systemhealth="Overall system status is unhealthy\n\nProblem detected."
            output_str= append_to_output(systemhealth)
        elif bomcheck_summary_map[section_details_key] == "GeneralSummary":
            if bom_checks["GeneralSummary"]:
                output_str= append_to_output(bom_checks["GeneralSummary"])

        else:
            for j in range(NUMBER_OF_SUB_SECTIONS):
                j = j+1
                sub_section_details_key = "{}_{}_section".format(i,j)
                sub_section_print_str = "{}_{}_section_str".format(i,j)

                if sub_section_details_key not in bomcheck_summary_map:
                    break
                if bomcheck_summary_map[sub_section_details_key] == "CEC":
                    if not print_cec_section:
                        continue

                sub_section_details = section_details.get(bomcheck_summary_map[sub_section_details_key], None)
                if sub_section_details is None:
                    continue
                output_str= append_to_output(bomcheck_summary_map[sub_section_print_str])
                for key in sub_section_details:
                    if key not in printable_string:
                        continue
                    output_str= append_to_output("    {0}:{1}".format(printable_string[key],sub_section_details[key]))
    # Returning non zero value in case overall system is unhealthy
    if not overall_system_state_healthy:
        return output_str, 1, bom_checks["BomCheckCollectionSummary"]
    return output_str, 0, None


def run_show_health(out_file=None):
        cmd = 'env LC_ALL=POSIX  nvsm-health --with-progress'

        if out_file:
            cmd += f' -o {out_file}'

        try:
            if log_level:
                cmd += f' --log-level={log_level_map[log_level]}'
        except:
            # its alright if the log level is not mapped
            pass

        import os
        rc = os.system(cmd)
        # on unix systems, higher 8bits are exit codes
        return rc >> 8

def run_new_dump_health(out_file_name=None, out_file_path=None, tags=None):
        body = {}

        if out_file_path:
            body["TarFilePath"] = out_file_path
        if out_file_name:
            body["TarFileName"] = out_file_name
        if not tags:
            body["Tags"] = []
        else:
            body["Tags"] = tags.split(",")
        cmd = "env LC_ALL=POSIX /usr/sbin/nvsm_dump --mode=standalone --body='{}'"

        cmd = cmd.format(json.dumps(body))
        #print (cmd)

        import os
        os.system(cmd)

def run_dump_health(out_file=None, log_level=None, sosreport_timeout=None, task_timeout=None):
        cmd = 'env LC_ALL=POSIX SOSREPORT_TIMEOUT={} TASK_TIMEOUT={} nvsm-health --dump --with-progress'

        if out_file:
            cmd += f' -o {out_file}'

        try:
            if log_level:
                cmd += f' --log-level={log_level_map[log_level]}'
        except:
            # its alright if the log level is not mapped
            pass

        cmd = cmd.format(sosreport_timeout, task_timeout)

        import os
        os.system(cmd)

class DumpTarget(Target):
    ''' Target to perform system health dump. '''
    def __init__(self, fwref, address, context):
        super(DumpTarget, self).__init__(
                address=address, context=context)
        self.__fwref = fwref
        self.__fwref_stream = None
        self.__properties = None
        self.__targets = None

    def __getitem__(self, key):
        raise KeyError

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        response = self._context._queryBackend(self.__fwref)
        # Show properties related to health running status
        state = response.get('State', 'Unknown')
        message = response.get('Message', 'Unknown').strip()
        progress = response.get('Progress', 'Unknown')
        result = response.get('Result', 'Unknown')
        try:
            result_summary = result['HealthSummary']['Result']
        except:
            result_summary = 'N/A'

        self.__properties.append(Property(
                    target=self,
                    name='State',
                    value=state))
        self.__properties.append(Property(
                    target=self,
                    name='Message',
                    value=message))
        self.__properties.append(Property(
                    target=self,
                    name='Progress',
                    value=progress))
        self.__properties.append(Property(
                    target=self,
                    name='Result',
                    value=result_summary))

    @property
    def targets(self):
        return [] # No sub-targets

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties upon first access
            self.__initProperties()
        return self.__properties

    # Stream the system health checks result
    def start(self, skip_prompt=False, wait=False):
        run_dump_health()

class HealthTarget(Target):
    ''' Target to perform, view and store the system health checks. '''
    def __init__(self, systemref, address, context):
        super(HealthTarget, self).__init__(
                address=address, context=context)
        self.__fwref = 'nvsm/v1/Systems/1/Health'
        self.__fwref_stream = None
        self.__properties = None
        self.__targets = None

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}
        # Add a SMASH target to handle health dump
        dumpTarget = DumpTarget(
                fwref=self.__fwref,
                address=(self.address + 'dump'),
                context=self._context)
        self.__targets['dump'] = dumpTarget

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        response = self._context._queryBackend(self.__fwref)
        # Show properties related to health running status
        state = response.get('State', 'Unknown')
        message = response.get('Message', 'Unknown')
        progress = response.get('Progress', 'Unknown')
        result = response.get('Result', 'Unknown')
        timestamp = response.get('TimeStamp', 'Unknown')
        try:
            result_summary = result['HealthSummary']['Result']
        except:
            result_summary = 'N/A'

        message = response.get('Message', 'Unknown').strip()
        progress = response.get('Progress')


        self.__properties.append(Property(
                    target=self,
                    name='State',
                    value=state))
        self.__properties.append(Property(
                    target=self,
                    name='TimeStamp',
                    value=timestamp))
        self.__properties.append(Property(
                    target=self,
                    name='Message',
                    value=message))
        self.__properties.append(Property(
                    target=self,
                    name='Progress',
                    value=progress))
        self.__properties.append(Property(
                    target=self,
                    name='Result',
                    value=result_summary))

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties upon first access
            self.__initProperties()
        return self.__properties

    # Stream the system health checks result
    def start(self, skip_prompt=False, wait=False, file=sys.stdout):
        run_show_health()
