# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



from nvsm.backend.context import Target, dict_to_properties

from nvsm.backend.alerts import AlertsTarget
from nvsm.backend.policy import CMMPolicyTarget

import nvsm.log

bad_status = [ 'warning', 'critical' ]
good_status = [ 'ok' ]
def colorize_status(value):
    if value == None:
        return "none"
    if value.lower() in good_status:
        return 'green'
    if value.lower() in bad_status:
        return 'red'
    return 'none'

class ManagerEthInterfaceTarget(Target):
    def __init__(self, ref, address, context):
        super(ManagerEthInterfaceTarget, self).__init__(
                address=address, context=context)
        self.__ref = ref
        self.__properties = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        info = self._context._queryBackend(self.__ref)
        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            return True
        self.__properties = list(dict_to_properties(
                    target=self, d=info,
                    criteria=filter_properties))
        self.__properties.sort(key=lambda x : x.name)

    def __getitem__(self, key):
        # System memory has no targets at this time
        # (behaves as an empty dictionary)
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties

class ManagerEthInterfaceCollectionTarget(Target):
    def __init__(self, ref, address, context):
        super(ManagerEthInterfaceCollectionTarget, self).__init__(
            address=address, context=context)
        self.__ref = ref
        self.__properties = None
        self.__targets = None

    def __initTargets(self):
        # Enumerate storage subsystems with NVSM backend
        self.__targets = {}
        info = self._context._queryBackend(
            '{0}'.format(self.__ref))
        members = info.get('Members', [])
        if members == None:
            members = []
        for eth in members:
            eth = eth.get('@odata.id', None)
            if not eth:
                continue
            index = eth.split('/')[-1]
            mgrEthInterface = ManagerEthInterfaceTarget(
                ref=eth,
                address=(self.address + index),
                context=self._context)
            # self.__targets[index] = storage
            self.__targets['ethinterface'] = mgrEthInterface

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?

class ManagerTarget(Target):
    def __init__(self, managerref, address, context):
        super(ManagerTarget, self).__init__(
                address=address, context=context)
        self.__managerref = managerref
        self.__properties = None
        self.__targets = {}

        ethTarget = ManagerEthInterfaceCollectionTarget(
            ref=self.__managerref+"/EthernetInterfaces",  # XXX
            address=(self.address + 'ethernetinterfaces'),
            context=self._context)
        self.__targets['ethernetinterfaces'] = ethTarget

    def __initProperties(self):
        self.__properties = []  # XXX?
        managerinfo = self._context._queryBackend(self.__managerref)

        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            if name.lower() == "description":
                return False
            return True
        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            return (name, value, color)
        self.__properties = list(dict_to_properties(
                target=self, d=managerinfo,
                criteria=filter_properties,
                mapping=map_properties))


    def __getitem__(self, key):
        return self.__targets[key]

    @property
    def targets(self):
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class ManagersCollectionTarget(Target):
    def __init__(self, address, context):
        super(ManagersCollectionTarget, self).__init__(
                address=address, context=context)
        self.__targets = None

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}
        # Enumerate network ports with NVSM backend
        memberinfo = {}
        try:
            memberinfo = self._context._queryBackend("/redfish/v1/Managers")
        except:
            pass
        members = memberinfo.get('Members', None)
        if members == None:
            members = []
        for member in members:
            ref = member.get('@odata.id', None)
            if not ref:
                nvsm.log.warning(
                    "NVSM backend did not give '@odata.id' for manager")
                continue
            index = ref.strip('/').split('/')[-1]
            fields = index.split('_')
            if index == "1":
                index = 'localhost'
            if len(fields) <= 2 :
                manager = ManagerTarget(
                        managerref=ref,
                        address=(self.address + index),
                        context=self._context)
                self.__targets[index] = manager
            else:
                submanager = ManagerSubTarget(
                        address=(self.address + fields[0]),
                        context=self._context)
                self.__targets[fields[0]] = submanager

        # populate alerts and policy targets only if there are
        # cmm targets
        if len(self.__targets):
            # Assume every blade has cmm alert targets (for DSHM alerts)
            alertsTarget = AlertsTarget(
                ref='/nvsm/v1/Alerts/CMM',
                address=(self.address + 'alerts'),
                context=self._context)
            self.__targets['alerts'] = alertsTarget

            # Assume every chassis has cmm policy targets (for configuring DSHM)
            policyTarget = CMMPolicyTarget(
                    # FIXME: NVSM backend does not provide OData reference here
                    policyref='/nvsm/v1/Policy/CMM',
                    address=(self.address + 'policy'),
                    context=self._context)
            self.__targets['policy'] = policyTarget

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?


class ManagerSubTarget(Target):
    def __init__(self, address, context):
        super(ManagerSubTarget, self).__init__(
                address=address, context=context)
        self.__targets = None

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}
        ref = "/redfish/v1/Managers"
        # Enumerate network ports with NVSM backend
        try:
            memberinfo = self._context._queryBackend(ref)
        except:
            pass
        members = memberinfo.get('Members', None)
        if members == None:
            members = []
        for member in members:
            ref = member.get('@odata.id', None)
            if not ref:
                nvsm.log.warning(
                    "NVSM backend did not give '@odata.id' for manager")
                continue
            index = ref.strip('/').split('/')[-1]
            index_fields = index.split('_')
# skip Managers like Enclosure0_CMM
            if len(index_fields) <= 2:
                continue
            address = self.address
            fields = address.string.split('/')
            basename = fields[-1]
# Cover the end Nodes in cases like Enclosure0_Node0_BMC
            if (basename == index_fields[0]) and len(index_fields) == 3:
                manager = ManagerTarget(
                        managerref=ref,
                        address=(self.address + index_fields[1]),
                        context=self._context)
                self.__targets[index_fields[1]] = manager
# Cover the Sub Manager cases like Enclosure0 and Blade0
            elif (basename == index_fields[0]) and (len(index_fields) == 4):
                manager = ManagerSubTarget(
                        address=(self.address + index_fields[1]),
                        context=self._context)
                if self.__targets.get(index_fields[1], None) is None:
                    self.__targets[index_fields[1]] = manager
# Cover the end nodes in cases like Node0_BMC
            elif (fields[-2] == index_fields[0]) and (index_fields[1] == fields[-1] ):
                manager = ManagerTarget(
                        managerref=ref,
                        address=(self.address + index_fields[2]),
                        context=self._context)
                self.__targets[index_fields[2]] = manager

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?
