# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



from nvsm.backend.context import Target, dict_to_properties

from nvsm.backend.alerts import AlertsTarget
from nvsm.backend.policy import MemoryPolicyTarget

import nvsm.log

bad_status = [ 'warning', 'critical' ]
good_status = [ 'ok' ]
def colorize_status(value):
    if value == None:
        return "none"
    if value.lower() in good_status:
        return 'green'
    if value.lower() in bad_status:
        return 'red'
    return 'none'

class MemoryTarget(Target):
    def __init__(self, memoryref, address, context):
        super(MemoryTarget, self).__init__(
                address=address, context=context)
        self.__memoryref = memoryref
        self.__properties = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        memoryinfo = self._context._queryBackend(self.__memoryref)
        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            if name.lower().startswith('oem_nvidia_errorsummary'):
                return True
            if name.lower().startswith('oem_'):
                return self._context.showall
            if name.lower() == 'schema':
                return False
            return True
        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            if name.lower().startswith('oem_nvidia_errorsummary'):
                name = name[len('oem_nvidia_errorsummary'):]
                name = "ErrorType" + name
            if name.lower().endswith('errortype'):
                name = name[:len('ErrorType_0_')] + "Type"
            if name.lower().endswith('errorcount'):
                name = name[:len('ErrorCount_0')] + "Count"
            return (name, value, color)
        self.__properties = list(dict_to_properties(
                    target=self, d=memoryinfo,
                    criteria=filter_properties,
                    mapping=map_properties))
        self.__properties.sort(key=lambda x : x.name)

    def __getitem__(self, key):
        # System memory has no targets at this time
        # (behaves as an empty dictionary)
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class MemoryCollectionTarget(Target):
    def __init__(self, systemref, address, context):
        super(MemoryCollectionTarget, self).__init__(
                address=address, context=context)
        self.__systemref = systemref
        self.__targets = None

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}

        # Assume every system has memory alert targets (for DSHM alerts)
        alertsTarget = AlertsTarget(
            ref='/nvsm/v1/Alerts/Memory',  # XXX
            address=(self.address + 'alerts'),
            context=self._context)
        self.__targets['alerts'] = alertsTarget

        # Assume every system has memory policy targets (for configuring DSHM)
        policyTarget = MemoryPolicyTarget(
                # FIXME: NVSM backend does not provide OData reference here
                policyref='/nvsm/v1/Policy/Memory',
                address=(self.address + 'policy'),
                context=self._context)
        self.__targets['policy'] = policyTarget

        # Enumerate memory targets with NVSM backend
        try:
            memoryinfo = self._context._queryBackend('{0}/Memory'.format(self.__systemref))
        except:
            pass
        # FIXME: This is case-sensitive
        memoryinfo = memoryinfo.get('Members', [])
        if not memoryinfo:
            memoryinfo = []
        for memoryref in memoryinfo:
            memoryref = memoryref.get('@odata.id', None)
            if not memoryref:
                nvsm.log.warning(
                    "NVSM backend did not give '@odata.id' for system memory")
                continue
            name = memoryref.strip('/').split('/')[-1]
            target = MemoryTarget(
                    memoryref=memoryref,
                    address=(self.address + name),
                    context=self._context)
            self.__targets[name] = target

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?
