# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



from nvsm.backend.context import Target, dict_to_properties
from nvsm.backend.networkports import NetworkPortCollectionTarget
from nvsm.backend.networkdevfuncs import NetworkDeviceFunctionsCollectionTarget
from nvsm.backend.alerts import AlertsTarget
from nvsm.backend.policy import NetworkPolicyTarget
import nvsm.exceptions

bad_status = ['warning', 'critical', 'disabled', 'down']
good_status = ['ok', 'enabled', 'up']


def colorize_status(value):
    if value == None:
        return "none"
    if value.lower() in good_status:
        return 'green'
    if value.lower() in bad_status:
        return 'red'
    return 'none'


class NetworkAdapterTarget(Target):
    def __init__(self, adapterref, address, context, nodeIdentifier):
        super(NetworkAdapterTarget, self).__init__(
            address=address, context=context)
        self.__adapterref = adapterref
        self.__properties = None
        self.__targets = None
        self.__nodeIdentifier = nodeIdentifier

    def __initTargets(self):
        # Get Network ports from NVSM backend
        try:
            adapterinfo = self._context._queryBackend(self.__adapterref)
        except:
            pass
        self.__targets = {}

        networkports = adapterinfo.get('NetworkPorts', None)
        if networkports:
            ref = networkports["@odata.id"]
            NetworkPortCollection = NetworkPortCollectionTarget(
                networkportsref=ref,
                address=(self.address + 'NetworkPorts'),
                context=self._context,
                nodeIdentifier=self.__nodeIdentifier,
                parentIdentifier=self.__adapterref.split('/')[-1])
            index = ref.strip('/').split('/')[-1]
            self.__targets[index] = NetworkPortCollection

        network_device_functions = adapterinfo.get(
            'NetworkDeviceFunctions', None)
        if network_device_functions:
            ref = network_device_functions["@odata.id"]
            NetworkDevFuncCollection = NetworkDeviceFunctionsCollectionTarget(
                nw_df_collection_ref=ref,
                address=(self.address + 'NetworkDeviceFunctions'),
                context=self._context,
                nodeIdentifier=self.__nodeIdentifier,
                parentIdentifier=self.__adapterref.split('/')[-1])
            index = ref.strip('/').split('/')[-1]
            self.__targets[index] = NetworkDevFuncCollection

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    def __initProperties(self):
        assert self.__properties == None
        adapterinfo = self._context._queryBackend(self.__adapterref)
        nvsm.log.debug('adapter respone {0}', adapterinfo)

        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            if name.lower() == 'description':
                return False
            return True

        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            return (name, value, color)
        self.__properties = list(dict_to_properties(
            target=self, d=adapterinfo,
            criteria=filter_properties,
            mapping=map_properties))

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class NetworkAdapterCollectionTarget(Target):
    def __init__(self, ref, address, context, nodeIdentifier):
        super(NetworkAdapterCollectionTarget, self).__init__(
            address=address, context=context)
        self.__targets = None
        self.__ref = ref
        self.__nodeIdentifier = nodeIdentifier

    def __initTargets(self):
        # Enumerate adapters with NVSM backend
        self.__targets = {}
        try:
            adaptersinfo = self._context._queryBackend(self.__ref)
        except:
            pass
        members = adaptersinfo.get('Members', [])
        if members == None:
            members = []
        for adapterref in members:
            ref = adapterref.get('@odata.id', None)
            if not ref:
                continue
            index = ref.split('/')[-1]
            adapter = NetworkAdapterTarget(
                adapterref=ref,
                address=(self.address + index),
                context=self._context,
                nodeIdentifier=self.__nodeIdentifier)
            self.__targets[index] = adapter

        # populate alerts and policy targets only if there are
        # network adapter targets
        if len(self.__targets):
            # Assume every chassis has network alert targets (for DSHM alerts)
            alertsTarget = AlertsTarget(
                ref='/nvsm/v1/Alerts/Network',  # XXX
                address=(self.address + 'alerts'),
                context=self._context,
                nodeIdentifier=self.__nodeIdentifier)
            self.__targets['alerts'] = alertsTarget

            # Assume every chassis has network policy targets (for configuring DSHM)
            policyTarget = NetworkPolicyTarget(
                # FIXME: NVSM backend does not provide OData reference here
                policyref='/nvsm/v1/Policy/Network',
                address=(self.address + 'policy'),
                context=self._context,
                nodeIdentifier=self.__nodeIdentifier)
            self.__targets['policy'] = policyTarget

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?
