# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



from nvsm.backend.context import Target, dict_to_properties

from nvsm.backend.alerts import AlertsTarget
from nvsm.backend.policy import NetworkDevFuncPolicyTarget

import nvsm.log

bad_status = ['warning', 'critical', 'disabled', 'down']
good_status = ['ok', 'enabled', 'up']


def colorize_status(value):
    if value == None:
        return "none"
    if value.lower() in good_status:
        return 'green'
    if value.lower() in bad_status:
        return 'red'
    return 'none'


class NetworkDeviceFunctionTarget(Target):
    def __init__(self, nw_df_ref, address, context):
        super(NetworkDeviceFunctionTarget, self).__init__(
            address=address, context=context)
        self.__nw_df_ref = nw_df_ref
        self.__targets = None
        self.__properties = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        nwportinfo = self._context._queryBackend(self.__nw_df_ref)

        def filter_properties(name, value):
            if '@odata' in name.lower():
                return False
            if name.lower() == 'links':
                return False
            return True

        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            if 'oem_nvidia' in name.lower():
                name = name.split('_')[-1]
            if 'oem_counters' in name.lower():
                name = name.split('Oem_Counters_')[-1]
            return (name, value, color)

        self.__properties = list(dict_to_properties(
            target=self, d=nwportinfo,
            criteria=filter_properties,
            mapping=map_properties))

    def __getitem__(self, key):
        # network interface has no targets at this time
        # (behaves as an empty dictionary)
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class NetworkDeviceFunctionsCollectionTarget(Target):
    def __init__(self, nw_df_collection_ref, address, context, nodeIdentifier, parentIdentifier=None):
        super(NetworkDeviceFunctionsCollectionTarget, self).__init__(
            address=address, context=context)
        self.__nw_df_collection_ref = nw_df_collection_ref
        self.__targets = None
        self.__properties = None
        self.__nodeIdentifier = nodeIdentifier
        self.__parentIdentifier = parentIdentifier

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}
        try:
            network_dfs_resp = self._context._queryBackend(
                self.__nw_df_collection_ref)
        except:
            pass
        network_dfs = []
        if network_dfs_resp:
            network_dfs = network_dfs_resp.get('Members', [])

        # Create sub-target for each network adapter.
        for nw_df in network_dfs:
            port_id = nw_df['@odata.id'].split('/')[-1]
            target = NetworkDeviceFunctionTarget(
                nw_df_ref=nw_df['@odata.id'],
                address=(self.address + port_id),
                context=self._context)
            self.__targets[port_id] = target

        # populate alerts and policy targets only if there are
        # network dev func targets
        if len(self.__targets):
            alertsTarget = AlertsTarget(
                ref='/nvsm/v1/Alerts/NetworkDevFunc',  # XXX
                address=(self.address + 'alerts'),
                context=self._context,
                nodeIdentifier=self.__nodeIdentifier,
                parentIdentifier=self.__parentIdentifier)
            self.__targets['alerts'] = alertsTarget

            # Assume every chassis has network port policy targets (for configuring DSHM)
            policyTarget = NetworkDevFuncPolicyTarget(
                # FIXME: NVSM backend does not provide OData reference here
                policyref='/nvsm/v1/Policy/NetworkDevFunc',
                address=(self.address + 'policy'),
                context=self._context,
                nodeIdentifier=self.__nodeIdentifier,
                parentIdentifier=self.__parentIdentifier)
            self.__targets['policy'] = policyTarget

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []
