# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



from nvsm.backend.context import Target, dict_to_properties
from nvsm.backend.networkports import NetworkPortCollectionTarget

import nvsm.log

bad_status = ['warning', 'critical', 'disabled', 'down']
good_status = ['ok', 'enabled', 'up']


def colorize_status(value):
    if value == None:
        return "none"
    if value.lower() in good_status:
        return 'green'
    if value.lower() in bad_status:
        return 'red'
    return 'none'


class NetworkInterfaceTarget(Target):
    def __init__(self, nw_if_ref, address, context):
        super(NetworkInterfaceTarget, self).__init__(
            address=address, context=context)
        self.__nw_if_ref = nw_if_ref
        self.__targets = None
        self.__properties = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        nwifinfo = self._context._queryBackend(self.__nw_if_ref)

        def filter_properties(name, value):
            if '@odata' in name.lower():
                return False
            if name.lower().startswith('links'):
                return False
            if name.lower().startswith('networkports'):
                return False
            return True

        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            return (name, value, color)

        self.__properties = list(dict_to_properties(
            target=self, d=nwifinfo,
            criteria=filter_properties,
            mapping=map_properties))

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}

        network_interfaces_resp = {}
        try:
            network_interfaces_resp = self._context._queryBackend(self.__nw_if_ref)
        except:
            pass
        # add networkports target only if ports are available
        network_ports_resp = None
        try:
            network_ports = network_interfaces_resp.get(
                'NetworkPorts', {}).get('@odata.id')
            network_ports_resp = self._context._queryBackend(network_ports)
        except:
            pass
        if network_interfaces_resp and network_ports_resp:
            network_ports = network_interfaces_resp.get(
                'NetworkPorts', {}).get('@odata.id')
            # get the parent adapter
            parentAdapter = network_ports_resp.get('@odata.id').split('/')[-2]

            NetworkPortCollection = NetworkPortCollectionTarget(
                networkportsref=network_ports,
                address=(self.address + 'NetworkPorts'),
                context=self._context,
                nodeIdentifier=None,
                parentIdentifier=parentAdapter,
                withAlerts=False)
            index = network_ports.strip('/').split('/')[-1]
            self.__targets[index] = NetworkPortCollection

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class NetworkInterfaceCollectionTarget(Target):
    def __init__(self, nw_intf_collection_ref, address, context):
        super(NetworkInterfaceCollectionTarget, self).__init__(
            address=address, context=context)
        self.__nw_intf_collection_ref = nw_intf_collection_ref
        self.__targets = None

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}
        network_interfaces_resp = {}
        try:
            network_interfaces_resp = self._context._queryBackend(
                self.__nw_intf_collection_ref)
        except:
            pass
        network_interfaces = []
        if network_interfaces_resp:
            network_interfaces = network_interfaces_resp.get('Members', [])

        # Create sub-target for each network interface.
        for nw_if in network_interfaces:
            if_id = nw_if['@odata.id'].split('/')[-1]
            target = NetworkInterfaceTarget(
                nw_if_ref=nw_if['@odata.id'],
                address=(self.address + if_id),
                context=self._context)
            self.__targets[if_id] = target

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []
