# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



from nvsm.backend.context import Target, dict_to_properties

from nvsm.backend.alerts import AlertsTarget
from nvsm.backend.policy import NetworkPortPolicyTarget

import nvsm.log

bad_status = ['warning', 'critical', 'disabled', 'down', 'fail']
good_status = ['ok', 'enabled', 'up', 'pass']


def colorize_status(value):
    if value == None:
        return "none"
    if value.lower() in good_status:
        return 'green'
    if value.lower() in bad_status:
        return 'red'
    return 'none'


class NetworkPortTarget(Target):
    def __init__(self, portref, address, context):
        super(NetworkPortTarget, self).__init__(
            address=address, context=context)
        self.__portref = portref
        self.__properties = None

    def __initProperties(self):
        self.__properties = []  # XXX?
        portinfo = self._context._queryBackend(self.__portref)

        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            if name.lower() == 'links':
                return False
            return True

        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            if name.lower().startswith('oem') and \
                    name.lower().endswith('hca_id'):
                name = 'HcaID'
            if '_test_' in name.strip().lower():
                color = colorize_status(value)
            if name.lower().startswith('supportedlinkcapabilities'):
                name = name.split('_')[-1]
            return (name, value, color)

        self.__properties = list(dict_to_properties(
            target=self, d=portinfo,
            criteria=filter_properties,
            mapping=map_properties))

    def __getitem__(self, key):
        # Network port has no targets
        # (behaves as an empty dictionary)
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class NetworkPortCollectionTarget(Target):
    def __init__(self, networkportsref, address, context, nodeIdentifier, parentIdentifier=None, withAlerts=True):
        super(NetworkPortCollectionTarget, self).__init__(
            address=address, context=context)
        self.__targets = None
        self.__networkportsref = networkportsref
        self.__nodeIdentifier = nodeIdentifier
        self.__parentIdentifier = parentIdentifier
        self.__withAlerts = withAlerts

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}

        # Enumerate network ports with NVSM backend
        portinfo ={}
        try:
            portinfo = self._context._queryBackend(self.__networkportsref)
        except:
            pass
        members = portinfo.get('Members', None)
        if members == None:
            members = []
        for port in members:
            portref = port.get('@odata.id', None)
            if not portref:
                nvsm.log.warning(
                    "NVSM backend did not give '@odata.id' for network port")
                continue
            index = portref.strip('/').split('/')[-1]
            port = NetworkPortTarget(
                portref=portref,
                address=(self.address + index),
                context=self._context)
            self.__targets[index] = port

        # populate alerts and policy targets only if there are
        # network ports targets
        if len(self.__targets) and self.__withAlerts:
            # Assume every chassis has network port alert targets (for DSHM alerts)
            alertsTarget = AlertsTarget(
                ref='/nvsm/v1/Alerts/NetworkPort',  # XXX
                address=(self.address + 'alerts'),
                context=self._context,
                nodeIdentifier=self.__nodeIdentifier,
                parentIdentifier=self.__parentIdentifier)
            self.__targets['alerts'] = alertsTarget

            # Assume every chassis has network port policy targets (for configuring DSHM)
            policyTarget = NetworkPortPolicyTarget(
                # FIXME: NVSM backend does not provide OData reference here
                policyref='/nvsm/v1/Policy/NetworkPort',
                address=(self.address + 'policy'),
                context=self._context,
                nodeIdentifier=self.__nodeIdentifier,
                parentIdentifier=self.__parentIdentifier)
            self.__targets['policy'] = policyTarget

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?
