# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import nvsm.log
import nvsm.odata
from nvsm.backend.context import Property, Target, dict_to_properties

from nvsm.backend.alerts import AlertsTarget
from nvsm.backend.policy import NVSwitchPolicyTarget
from nvsm.exceptions import BackendError

bad_status = [ 'warning', 'critical' ]
good_status = [ 'ok' ]

def colorize_status(value):
    if value == None:
        return "none"
    if str(value).lower() in good_status:
        return 'green'
    if str(value).lower() in bad_status:
        return 'red'
    return 'none'

class NVSwitchTarget(Target):
    def __init__(self, nvswitch, address, context):
        super(NVSwitchTarget, self).__init__(
            address=address, context=context)
        self.__nvswitch = nvswitch
        self.__properties = None

    def __initProperties(self):
        assert self.__properties == None
        nvswitchinfo = self._context._queryBackend(self.__nvswitch)
        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            if name.lower().startswith('oem_'):
                return self._context.showall
            if name == 'schema':
                return False
            return True
        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            return (name, value, color)
        self.__properties = list(dict_to_properties(
                    target=self, d=nvswitchinfo,
                    criteria=filter_properties,
                    mapping=map_properties))
        self.__properties.sort(key=lambda x : x.name)

    def __getitem__(self, key):
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class NVSwitchCollectionTarget(Target):
    def __init__(self, systemref, address, context):
        super(NVSwitchCollectionTarget, self).__init__(
            address=address, context=context)
        self.__systemref = systemref
        self.__targets = None
        self.__properties = None

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}

        try:
            nvswitchinfo = self._context._queryBackend('/nvsm/v1/Systems/1/NVSwitches')
            nvswitchrefs = nvswitchinfo.get('Members', [])
        except BackendError as e:
            nvsm.log.error('{0}', e)
        else:
            # Enumerate NVSwitches with NVSM backend
            for nvswitchref in nvswitchrefs:
                nvswitchref = nvswitchref.get('@odata.id', None)
                if not nvswitchref:
                    nvsm.log.warning(
                        "NVSM backend did not give '@odata.id' for system nvswitch")
                    continue
                name = nvswitchref.strip('/').split('/')[-1]
                target = NVSwitchTarget(
                            nvswitch=nvswitchref,
                            address=(self.address + name),
                            context=self._context)
                self.__targets[name] = target

        # Assume every system has NVSwitch alert targets (for DSHM alerts)
        alertsTarget = AlertsTarget(
            ref='/nvsm/v1/Alerts/NVSwitches',
            address=(self.address + 'alerts'),
            context=self._context)
        self.__targets['alerts'] = alertsTarget

        # Assume every system has NVSwitch policy target (for configuring DSHM)
        policyTarget = NVSwitchPolicyTarget(
                policyref='/nvsm/v1/Policy/NVSwitches',
                address=(self.address + 'policy'),
                context=self._context)
        self.__targets['policy'] = policyTarget

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []
