# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



from nvsm.backend.context import Target, dict_to_properties

from nvsm.backend.alerts import AlertsTarget
from nvsm.backend.policy import PCIePolicyTarget

import nvsm.log

bad_status = [ 'warning', 'critical' ]
good_status = [ 'ok' ]
def colorize_status(value):
    if value == None:
        return "none"
    if value.lower() in good_status:
        return 'green'
    if value.lower() in bad_status:
        return 'red'
    return 'none'

class PCIeTarget(Target):
    def __init__(self, pcieref, address, context):
        super(PCIeTarget, self).__init__(
                address=address, context=context)
        self.__pcieref = pcieref
        self.__properties = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        pcieinfo = self._context._queryBackend(self.__pcieref)
        properties = [ 'id', 'bdf', 'name', 'devicetype', 'isaccessible',
                       'lanesinuse', 'pcietype', 'maxlanes', 'maxpcietype',
                       'slot_slotnumber', 'slot_hotplugsupported', 'slot_presence',
                       'parentchild_@odata.id','parentchild_bdf','parentchild_devicetype',
                       'parentchild_maxlanes','parentchild_maxpcietype',
                       'error_correctablecount', 'error_correctabletimestamp',
                       'error_uncorrectablecount', 'error_uncorrectabletimestamp'
                       ]

        def filter_properties(name, value):
            if name.lower().startswith('@odata'):
                return False
            if name.lower().startswith('links'):
                return False
            return True

        def map_properties(name, value):
            color = 'none'
            if name.lower().startswith('status'):
                color = colorize_status(value)
            if name.lower().startswith('pciedeviceerrorreporting_'):
                name = name[len('pciedeviceerrorreporting_'):]
                name = 'ErrorReporting_'+name
            if name.lower().startswith('pcieinterface_'):
                name = name[len('pcieinterface_'):]
            if name.lower().startswith('pciedeviceerrorcount'):
                name = name[len('pciedeviceerrorcount'):]
                name = 'Error'+name
            if name.lower() in [ 'error_correctabletimestamp',
                                 'error_uncorrectabletimestamp' ]:
                if not value:
                    value = '[ ]'
                elif len(value) == 1:
                    value = '[ {} ]'.format(value[0])
                else:
                    indent = ' ' * 12
                    sep = ',\n{indent}'.format(indent=indent)
                    value = '[\n{indent}{value} ]'.format(
                            indent=indent, value=sep.join(value))
            return (name, value, color)

        def compare_properties(a):
            try:
                i = properties.index(a.name.lower())
            except ValueError:
                i = len(properties)
            return i

        self.__properties = list(dict_to_properties(
                    target=self, d=pcieinfo,
                    criteria=filter_properties,
                    mapping=map_properties))

        # Sort by name in ASCII descending order.
        self.__properties.sort(key=lambda x : x.name)

        # Have status_health, status_state fall in the end.
        self.__properties.sort(key=compare_properties)

    def __getitem__(self, key):
        # System memory has no targets at this time
        # (behaves as an empty dictionary)
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class PCIeSubTarget(Target):
    def __init__(self, pcieref, address, context):
        super(PCIeSubTarget, self).__init__(
                address=address, context=context)
        self.__pcieref = pcieref
        self.__targets = None
        self.__properties = None

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}

        refid = self.address.string.strip('/').split('/')[-1]

        # Hack: have SBDF subgroup PCIe domain of 0000
        if refid == 'SBDF':
            refid = '0000'

        # Enumerate PCIeDevices targets with NVSM backend
        try:
            pcieinfo = self._context._queryBackend(self.__pcieref)
        except:
            pass
        pcieinfo = pcieinfo.get('Members', [])
        if not pcieinfo:
            pcieinfo = []

        for pcieref in pcieinfo:
            pcieref = pcieref.get('@odata.id', None)
            if not pcieref:
                nvsm.log.warning(
                    "NVSM backend did not give '@odata.id' for system PCIe Devices")
                continue

            name = pcieref.strip('/').split('/')[-1]
            if name.lower().startswith(refid.lower()):
                target = PCIeTarget(
                        pcieref=pcieref,
                        address=(self.address + name),
                        context=self._context)
                self.__targets[name] = target

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets.get(key, None)

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []

class PCIeCollectionTarget(Target):
    def __init__(self, systemref, address, context):
        super(PCIeCollectionTarget, self).__init__(
                address=address, context=context)
        self.__systemref = systemref
        self.__targets = None
        #self.__properties = None

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}

        # Enumerate PCIeDevices targets with NVSM backend
        group = {}
        group_id = [ 'CPU', 'PCH', 'GPU', 'SW', 'NVSW', 'IB', 'NVME', 'M2',
                     'BMC', 'SLOT', 'ETH', '0000' ]
        try:
            pcieinfo = self._context._queryBackend('/nvsm/v1/Systems/1/PCIeDevices')
        except:
            pass
        pcieinfo = pcieinfo.get('Members', [])

        if not pcieinfo:
            pcieinfo = []

        for pcieref in pcieinfo:
            pcieref = pcieref.get('@odata.id', None)
            if not pcieref:
                nvsm.log.warning(
                    "NVSM backend did not give '@odata.id' for system PCIe Devices")
                continue

            # Group device's @odata.id reference according to device type.
            name = pcieref.strip('/').split('/')[-1]
            for gid in group_id:
                if not name.lower().startswith(gid.lower()):
                    continue
                if gid == '0000':
                    group['SBDF'] = '0000'
                else:
                    group[gid] = gid

        # Create sub-target for each grouped devices.
        for key, item in list(group.items()):
            target = PCIeSubTarget(
                pcieref='/nvsm/v1/Systems/1/PCIeDevices',
                address=(self.address + item),
                context=self._context)
            self.__targets[key] = target

        # Assume every system has PCIeDevices alert targets (for DSHM alerts)
        alertsTarget = AlertsTarget(
            ref='/nvsm/v1/Alerts/PCIeDevices',
            address=(self.address + 'alerts'),
            context=self._context)
        self.__targets['alerts'] = alertsTarget

        # Assume every system has PCIeDevices policy targets (for configuring DSHM)
        policyTarget = PCIePolicyTarget(
                policyref='/nvsm/v1/Policy/PCIeDevices',
                address=(self.address + 'policy'),
                context=self._context)
        self.__targets['policy'] = policyTarget

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []
