# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



import copy
import json
import re

from nvsm.backend.context import Target, Property, dict_to_properties

import nvsm.exceptions
import nvsm.log
import subprocess
import os
from functools import reduce

class DSHMPolicy(object):
    def __init__(self, policy, context):
        self.__policy = policy
        self.__context = context


class DSHMGlobalPolicy(DSHMPolicy):
    def __init__(self, policy, context):
        super(DSHMGlobalPolicy, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    @staticmethod
    def from_ref(policyref, context):
        # Check the DSHM version advertized in global policy
        policy = PolicyBlob(policyref, context=context)
        try:
            version = policy['version']
        except KeyError:
            # Assume version 1
            version = 1
        if version == 1:
            return DSHMGlobalPolicyV1(policy=policy, context=context)
        raise nvsm.exceptions.SchemaError(
            'Unknown DSHM global policy version: {}'.format(version))


class DSHMGPUPolicy(DSHMPolicy):
    def __init__(self, policy, context):
        super(DSHMGPUPolicy, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    @staticmethod
    def from_ref(policyref, context):
        # Check the DSHM version advertized in GPU policy
        policy = PolicyBlob(policyref, context=context)
        try:
            version = policy['version']
        except KeyError:
            # Assume version 1
            version = 1
        if version == 1:
            return DSHMGPUPolicyV1(policy=policy, context=context)
        raise nvsm.exceptions.SchemaError(
            'Unknown DSHM GPU policy version: {}'.format(version))


class DSHMPCIePolicy(DSHMPolicy):
    def __init__(self, policy, context):
        super(DSHMPCIePolicy, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    @staticmethod
    def from_ref(policyref, context):
        # Check the DSHM version advertized in memory policy
        policy = PolicyBlob(policyref, context=context)
        try:
            version = policy['version']
        except KeyError:
            # Assume version 1
            version = 1
        if version == 1:
            return DSHMPCIePolicyV1(policy=policy, context=context)
        raise nvsm.exceptions.SchemaError(
            'Unknown DSHM pcie policy version: {}'.format(version))


class DSHMMemoryPolicy(DSHMPolicy):
    def __init__(self, policy, context):
        super(DSHMMemoryPolicy, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    @staticmethod
    def from_ref(policyref, context):
        # Check the DSHM version advertized in memory policy
        policy = PolicyBlob(policyref, context=context)
        try:
            version = policy['version']
        except KeyError:
            # Assume version 1
            version = 1
        if version == 1:
            return DSHMMemoryPolicyV1(policy=policy, context=context)
        raise nvsm.exceptions.SchemaError(
            'Unknown DSHM memory policy version: {}'.format(version))


class DSHMPowerPolicy(DSHMPolicy):
    def __init__(self, policy, context):
        super(DSHMPowerPolicy, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    @staticmethod
    def from_ref(policyref, context):
        # Check the DSHM version advertized in power policy
        policy = PolicyBlob(policyref, context=context)
        try:
            version = policy['version']
        except KeyError:
            # Assume version 1
            version = 1
        if version == 1:
            return DSHMPowerPolicyV1(policy=policy, context=context)
        raise nvsm.exceptions.SchemaError(
            'Unknown DSHM power policy version: {}'.format(version))

class DSHMBladePolicy(DSHMPolicy):
    def __init__(self, policy, context):
        super(DSHMBladePolicy, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    @staticmethod
    def from_ref(policyref, context):
        # Check the DSHM version advertized in power policy
        policy = PolicyBlob(policyref, context=context)
        try:
            version = policy['version']
        except KeyError:
            # Assume version 1
            version = 1
        if version == 1:
            return DSHMBladePolicyV1(policy=policy, context=context)
        raise nvsm.exceptions.SchemaError(
            'Unknown DSHM Blade policy version: {}'.format(version))

class DSHMCMMPolicy(DSHMPolicy):
    def __init__(self, policy, context):
        super(DSHMCMMPolicy, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    @staticmethod
    def from_ref(policyref, context):
        # Check the DSHM version advertized in power policy
        policy = PolicyBlob(policyref, context=context)
        try:
            version = policy['version']
        except KeyError:
            # Assume version 1
            version = 1
        if version == 1:
            return DSHMCMMPolicyV1(policy=policy, context=context)
        raise nvsm.exceptions.SchemaError(
            'Unknown DSHM CMM policy version: {}'.format(version))

class DSHMNetworkPolicy(DSHMPolicy):
    def __init__(self, policy, context):
        super(DSHMNetworkPolicy, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    @staticmethod
    def from_ref(policyref, context):
        # Check the DSHM version advertized in power policy
        policy = PolicyBlob(policyref, context=context)
        try:
            version = policy['version']
        except KeyError:
            # Assume version 1
            version = 1
        if version == 1:
            return DSHMNetworkPolicyV1(policy=policy, context=context)
        raise nvsm.exceptions.SchemaError(
            'Unknown DSHM Network policy version: {}'.format(version))

class DSHMNetworkPortPolicy(DSHMPolicy):
    def __init__(self, policy, context):
        super(DSHMNetworkPortPolicy, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    @staticmethod
    def from_ref(policyref, context):
        # Check the DSHM version advertized in power policy
        policy = PolicyBlob(policyref, context=context)
        try:
            version = policy['version']
        except KeyError:
            # Assume version 1
            version = 1
        if version == 1:
            return DSHMNetworkPortPolicyV1(policy=policy, context=context)
        raise nvsm.exceptions.SchemaError(
            'Unknown DSHM Network Port policy version: {}'.format(version))

class DSHMNetworkDevFuncPolicy(DSHMPolicy):
    def __init__(self, policy, context):
        super(DSHMNetworkDevFuncPolicy, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    @staticmethod
    def from_ref(policyref, context):
        # Check the DSHM version advertized in  policy
        policy = PolicyBlob(policyref, context=context)
        try:
            version = policy['version']
        except KeyError:
            # Assume version 1
            version = 1
        if version == 1:
            return DSHMNetworkDevFuncPolicyV1(policy=policy, context=context)
        raise nvsm.exceptions.SchemaError(
            'Unknown DSHM Network Dev Func policy version: {}'.format(version))

class DSHMNVSwitchPolicy(DSHMPolicy):
    def __init__(self, policy, context):
        super(DSHMNVSwitchPolicy, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    @staticmethod
    def from_ref(policyref, context):
        # Check the DSHM version advertized in NVSwitch policy
        policy = PolicyBlob(policyref, context=context)
        try:
            version = policy['version']
        except KeyError:
            # Assume version 1
            version = 1
        if version == 1:
            return DSHMNVSwitchPolicyV1(policy=policy, context=context)
        raise nvsm.exceptions.SchemaError(
            'Unknown DSHM NVSwitch policy version: {}'.format(version))


class DSHMProcessorPolicy(DSHMPolicy):
    def __init__(self, policy, context):
        super(DSHMProcessorPolicy, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    @staticmethod
    def from_ref(policyref, context):
        # Check the DSHM version advertized in processor policy
        policy = PolicyBlob(policyref, context=context)
        try:
            version = policy['version']
        except KeyError:
            # Assume version 1
            version = 1
        if version == 1:
            return DSHMProcessorPolicyV1(policy=policy, context=context)
        raise nvsm.exceptions.SchemaError(
            'Unknown DSHM processor policy version: {}'.format(version))


class DSHMStoragePolicy(DSHMPolicy):
    def __init__(self, policy, context):
        super(DSHMStoragePolicy, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    @staticmethod
    def from_ref(policyref, context):
        # Check the DSHM version advertized in storage policy
        policy = PolicyBlob(policyref, context=context)
        try:
            version = policy['version']
        except KeyError:
            # Assume version 1
            version = 1
        if version == 1:
            return DSHMStoragePolicyV1(policy=policy, context=context)
        raise nvsm.exceptions.SchemaError(
            'Unknown DSHM storage policy version: {}'.format(version))


class DSHMThermalPolicy(DSHMPolicy):
    def __init__(self, policy, context):
        super(DSHMThermalPolicy, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    @staticmethod
    def from_ref(policyref, context):
        # Check the DSHM version advertized in thermal policy
        policy = PolicyBlob(policyref, context=context)
        try:
            version = policy['version']
        except KeyError:
            # Assume version 1
            version = 1
        if version == 1:
            return DSHMThermalPolicyV1(policy=policy, context=context)
        raise nvsm.exceptions.SchemaError(
            'Unknown DSHM thermal policy version: {}'.format(version))

def to_non_negative_integer(value, policy_target):
    try:
        value = int(value)
    except:
        msg = 'invalid int value: {}'.format(value)
        raise nvsm.exceptions.CommandError(msg)
    if value < 0:
        msg = 'value must be non-negative: {}'.format(value)
        raise nvsm.exceptions.CommandError(msg)
    return value

def pcie_device_list(metavar='device', relative_path='..'):
    def result(value, policy_target, metavar=metavar):
        device_ids = value.split(',')
        device_ids = [ e for e in device_ids if e != '']
        try:
            subsystem_address = policy_target.address + relative_path
            subsystem_target = policy_target._context.getTarget(subsystem_address)
            subsystem_targets = []
            subsystem_targets.extend(subsystem_target.targets)
            warn = set()
            # Filter Other Targets
            pci_subtargets_list = [x for x in subsystem_targets if isinstance(x, nvsm.backend.pcie.PCIeSubTarget)]

            for dev_id in device_ids:
                if dev_id in [ 'alerts', 'policy' ]:
                    warn.add(dev_id)
                    continue
                dev_id_target = [x for x in pci_subtargets_list if x[dev_id]]
                if not len(dev_id_target):
                    warn.add(dev_id)
            for dev_id in warn:
                nvsm.log.warning('Unknown {metavar} ID: {dev_id}',
                        metavar=metavar, dev_id=dev_id)
        except Exception as e:
            nvsm.log.warning('Could not validate {metavar} ID list: {e_type}: {e}',
                    metavar=metavar, e_type=type(e).__name__, e=e)
            import traceback
            nvsm.log.debug(traceback.format_exc())
        return device_ids
    return result


def to_device_list(metavar='device', relative_path='..'):
    def result(value, policy_target, metavar=metavar):
        # Device IDs parsed as comma separated list of strings
        device_ids = value.split(',')
        device_ids = [ e for e in device_ids if e != '']
        # Validate each device ID against the NVSM context
        # NOTE: Due to the transient nature of device IDs (e.g. devices added
        #       or removed from the system), the CLI cannot have perfect
        #       knowledge of which device IDs are "valid" or not. Because of
        #       this, the CLI allows for arbitrary device ID strings, and only prints a
        #       warning message if a given device ID is unknown at the time it
        #       is set.
        try:
            subsystem_address = policy_target.address + relative_path
            subsystem_target = policy_target._context.getTarget(subsystem_address)
            warn = set()
            for dev_id in device_ids:
                if dev_id in [ 'alerts', 'policy' ]:
                    warn.add(dev_id)
                    continue
                try:
                    subsystem_target[dev_id]
                except KeyError:
                    warn.add(dev_id)
            for dev_id in warn:
                nvsm.log.warning('Unknown {metavar} ID: {dev_id}',
                        metavar=metavar, dev_id=dev_id)
        except Exception as e:
            nvsm.log.warning('Could not validate {metavar} ID list: {e_type}: {e}',
                    metavar=metavar, e_type=type(e).__name__, e=e)
            import traceback
            nvsm.log.debug(traceback.format_exc())
        return device_ids
    return result

def to_recipients_list(value, policy_target):
    email_recipients = value.split(',')
    email_recipients = [ e for e in email_recipients if e != '']
    email_pattern = re.compile(r'^[^@]+@[^@]+\.[^@]+$')
    for addr in email_recipients:
        if not email_pattern.match(addr):
            nvsm.log.warning('Invalid Email address: {0}', addr)
    return email_recipients

def to_valid_email_address(addr, policy_target):
    email_pattern = re.compile(r'^[^@]+@[^@]+\.[^@]+$')
    if not email_pattern.match(addr):
        nvsm.log.warning('Invalid Email address: {0}', addr)
    return addr

def to_bool(value, policy_target):
    if value.lower() == "true":
        return True
    elif value.lower() == "false":
        return False
    else:
        nvsm.log.warning('Invalid boolean value: {0}', value)

    return False

def dump_supported():
    if subprocess.call("which nvsm-health".split()):
        nvsm.log.debug("nvsm-health is not installed on this machine.")
        return False
    return True

def to_callhome_bool_check(value, policy_target):
    warning_str = "\nEnabling the callhome feature allows NVSM to take following actions,\n" \
                     " 1.For, all existing crtical alerts and any new critical alerts raised in the system,\n" \
                        "   a.NVSM collects system dump.\n" \
                        "   b.Creates a ticket in NVES portal for the alert.\n" \
                        "   c.Adds the collected system dump to the ticket created in NVES portal.\n" \
                        "All the above steps get executed automatically and do not require any user input.\n\n" \
                        "For details about the system dump and its contents, please refer the NVSM user-guide at\n" \
                        "https://docs.nvidia.com/datacenter/nvsm/nvsm-user-guide\n"

    if dump_supported() == False:
        raise nvsm.exceptions.CommandError('callhome feature is not supported on this platform.')

    if value.lower() == "true":
        nvsm.log.warning(warning_str)
        return True
    elif value.lower() == "false":
        return False
    else:
        nvsm.log.warning('Invalid boolean value: {0}', value)

    return False

def to_offlinecallhome_bool_check(value, policy_target):
    if dump_supported() == False:
        raise nvsm.exceptions.CommandError('callhome feature is not supported on this platform.')
    if value.lower() == "true":
        return True
    elif value.lower() == "false":
        return False
    else:
        nvsm.log.warning('Invalid boolean value: {0}', value)
    return False

def to_offlinecallhome_destination_check(value, policy_target, retry=2):
    if retry == 0:
        raise nvsm.exceptions.CommandError("Invalid Path")
    warning_str = 'Invalid path for log destination: {}'
    if os.path.exists(value.lower()):
        return value.lower()
    else:
        response = nvsm.prompt.String(message=warning_str.format(value), prompt='Enter path for nvsm dump')
        if response == "" or not (os.path.exists(response)):
            return to_offlinecallhome_destination_check(value, policy_target, retry-1)
        else:
            return response

def to_alert_severity_check(value, policy_target):

    alert_severity_table = {0:'\tAlerts with severity set as Warning',
                            1:'\tAlerts with severity set as Critical' 
                           }

    alert_severity_map = {'warning':0,
                          'critical':1
                         }
                             

    alert_severity_help = "\n\nhelp:\n"\
                          "Severity level and severity as follows:\n" \
                          "\t 0 - Warning\n" \
                          "\t 1 - Critical\n" \
                          "Alerts at or numerically higher than the specified severity will get selected.\n" \
                          "\t Example: set severity as Warning to select alerts with severity as Warning and Critical.\n" 
                            
    display_str = 'Following Alerts will be processed and pushed by NVSM services to interfaces like Email,PagerDuty etc:\n'

    if value is None or not value.isalpha() or value.lower() not in alert_severity_map.keys():
        msg = 'invalid alert severity set: {}'.format(value)
        nvsm.log.warning(alert_severity_help)
        raise nvsm.exceptions.CommandError(msg)

    sev_value = alert_severity_map[value.lower()]

    for key in alert_severity_table.keys():
        if sev_value <= key:
            display_str = display_str + alert_severity_table[key]+ '\n'
           
    nvsm.log.warning(display_str + alert_severity_help)
    return value


class DSHMGlobalPolicyV1(DSHMGlobalPolicy):
    options = {
        'email_recipients': {
            'path': [ 'EmailConfig', 'recipients' ],
            'to_display': lambda x: ','.join(x),
            'to_blob': to_recipients_list,
        },
        'email_smtp_server_name': {
            'path': [ 'EmailConfig', 'smtp-server-name' ],
        },
        'email_smtp_server_port': {
            'path': [ 'EmailConfig', 'smtp-server-port' ],
            'to_blob': to_non_negative_integer,
        },
        'email_sender': {
            'path': [ 'EmailConfig', 'sender' ],
            'to_blob': to_valid_email_address,
        },
        'callhome_enable': {
            'path': [ 'CallHome', 'enable' ],
            'to_blob': to_callhome_bool_check,
        },
        'callhome_email_contact': {
            'path': [ 'CallHome', 'email_contact' ],
            'to_blob': to_valid_email_address,
        },
        'callhome_batch_interval': {
            'path': [ 'CallHome', 'batch_interval' ],
            'to_blob': to_non_negative_integer,
        },
        'offline_callhome_enable': {
            'path': ['OfflineCallHome', 'enable'],
            'to_blob': to_offlinecallhome_bool_check,
        },
        'offline_callhome_batch_interval':{
            'path': ['OfflineCallHome', 'batch_interval'],
            'to_blob': to_non_negative_integer,
        },
        'offline_callhome_dump_destination_location':{
            'path': ['OfflineCallHome', 'log_destination_location'],
            'to_blob': to_offlinecallhome_destination_check,
        },
        'offline_callhome_no_of_dumps_allowed':{
            'path': ['OfflineCallHome', 'no_of_dumps_allowed'],
            'to_blob': to_non_negative_integer,
        },
        'pagerduty_enable': {
            'path': [ 'PagerDuty', 'enable' ],
            'to_blob': to_bool,
        },
        'pagerduty_service_key': {
            'path': [ 'PagerDuty', 'service_key' ]
        },
        'pagerduty_batch_interval': {
            'path': [ 'PagerDuty', 'batch_interval' ],
            'to_blob': to_non_negative_integer,
         },
        'alert_severity': {
            'path': [ 'AlertFilter', 'alert_severity' ],
            'to_blob': to_alert_severity_check,
         }
    }

    def __init__(self, policy, context):
        super(DSHMGlobalPolicyV1, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    def __getitem__(self, key):
        return self.__policy[key]

    def __setitem__(self, key, value):
        self.__policy[key] = value


class DSHMGPUPolicyV1(DSHMGPUPolicy):
    options = {
        'mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='GPU device'),
            'path': [ 'Gpu', 'mute_monitoring' ],
        },
        'mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='GPU device'),
            'path': [ 'Gpu', 'mute_notification' ],
        }
    }

    def __init__(self, policy, context):
        super(DSHMGPUPolicyV1, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    def __getitem__(self, key):
        return self.__policy[key]

    def __setitem__(self, key, value):
        self.__policy[key] = value


class DSHMPCIePolicyV1(DSHMMemoryPolicy):
    options = {
        'mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': pcie_device_list(metavar='PCIe device'),
            'path': [ 'PCIe', 'mute_monitoring' ],
        },
        'mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': pcie_device_list(metavar='PCIe device'),
            'path': [ 'PCIe', 'mute_notification' ],
        }
    }

    def __init__(self, policy, context):
        super(DSHMPCIePolicyV1, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    def __getitem__(self, key):
        return self.__policy[key]

    def __setitem__(self, key, value):
        self.__policy[key] = value


class DSHMMemoryPolicyV1(DSHMMemoryPolicy):
    options = {
        'mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='DIMM device'),
            'path': [ 'Dimm', 'mute_monitoring' ],
        },
        'mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='DIMM device'),
            'path': [ 'Dimm', 'mute_notification' ],
        }
    }

    def __init__(self, policy, context):
        super(DSHMMemoryPolicyV1, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    def __getitem__(self, key):
        return self.__policy[key]

    def __setitem__(self, key, value):
        self.__policy[key] = value


class DSHMPowerPolicyV1(DSHMPowerPolicy):
    options = {
        'mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='PSU device'),
            'path': [ 'PowerSupplyUnit', 'mute_monitoring' ],
        },
        'mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='PSU device'),
            'path': [ 'PowerSupplyUnit', 'mute_notification' ],
        }
    }

    def __init__(self, policy, context):
        super(DSHMPowerPolicyV1, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    def __getitem__(self, key):
        return self.__policy[key]

    def __setitem__(self, key, value):
        self.__policy[key] = value

class DSHMBladePolicyV1(DSHMBladePolicy):
    options = {
        'mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='Blades'),
            'path': [ 'Blade', 'mute_monitoring' ],
        },
        'mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='Blades'),
            'path': [ 'Blade', 'mute_notification' ],
        }
    }

    def __init__(self, policy, context):
        super(DSHMBladePolicyV1, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    def __getitem__(self, key):
        return self.__policy[key]

    def __setitem__(self, key, value):
        self.__policy[key] = value

class DSHMCMMPolicyV1(DSHMCMMPolicy):
    options = {
        'mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='CMM'),
            'path': [ 'CMM', 'mute_monitoring' ],
        },
        'mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='CMM'),
            'path': [ 'CMM', 'mute_notification' ],
        }
    }

    def __init__(self, policy, context):
        super(DSHMCMMPolicyV1, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    def __getitem__(self, key):
        return self.__policy[key]

    def __setitem__(self, key, value):
        self.__policy[key] = value

class DSHMNetworkPolicyV1(DSHMNetworkPolicy):
    options = {
        'mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='Network'),
            'path': [ 'Network', 'mute_monitoring' ],
        },
        'mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='CMM'),
            'path': [ 'Network', 'mute_notification' ],
        }
    }

    def __init__(self, policy, context):
        super(DSHMNetworkPolicyV1, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    def __getitem__(self, key):
        return self.__policy[key]

    def __setitem__(self, key, value):
        self.__policy[key] = value

class DSHMNetworkPortPolicyV1(DSHMNetworkPortPolicy):
    options = {
        'mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='NetworkPort'),
            'path': [ 'NetworkPort', 'mute_monitoring' ],
        },
        'mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='NetworkPort'),
            'path': [ 'NetworkPort', 'mute_notification' ],
        }
    }

    def __init__(self, policy, context):
        super(DSHMNetworkPortPolicyV1, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    def __getitem__(self, key):
        return self.__policy[key]

    def __setitem__(self, key, value):
        self.__policy[key] = value

class DSHMNetworkDevFuncPolicyV1(DSHMNetworkDevFuncPolicy):
    options = {
        'rx_collision_threshold': {
            'path': [ 'NetworkDevFunc', 'rx_collision_threshold' ],
            'to_blob': to_non_negative_integer,
        },
        'rx_crc_threshold': {
            'path': [ 'NetworkDevFunc', 'rx_crc_threshold' ],
            'to_blob': to_non_negative_integer,
        },
        'tx_collision_threshold': {
            'path': [ 'NetworkDevFunc', 'tx_collision_threshold' ],
            'to_blob': to_non_negative_integer,
        },
        'mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='NetworkDevFunc'),
            'path': [ 'NetworkDevFunc', 'mute_monitoring' ],
        },
        'mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='NetworkDevFunc'),
            'path': [ 'NetworkDevFunc', 'mute_notification' ],
        }
    }

    def __init__(self, policy, context):
        super(DSHMNetworkDevFuncPolicyV1, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    def __getitem__(self, key):
        return self.__policy[key]

    def __setitem__(self, key, value):
        self.__policy[key] = value

class DSHMNVSwitchPolicyV1(DSHMNVSwitchPolicy):
    options = {
        'mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='NVSwitch device'),
            'path': [ 'NVSwitch', 'mute_monitoring' ],
        },
        'mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='NVSwitch device'),
            'path': [ 'NVSwitch', 'mute_notification' ],
        }
    }

    def __init__(self, policy, context):
        super(DSHMNVSwitchPolicyV1, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    def __getitem__(self, key):
        return self.__policy[key]

    def __setitem__(self, key, value):
        self.__policy[key] = value

class DSHMProcessorPolicyV1(DSHMProcessorPolicy):
    options = {
        'mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='CPU device'),
            'path': [ 'Cpu', 'mute_monitoring' ],
        },
        'mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='CPU device'),
            'path': [ 'Cpu', 'mute_notification' ],
        }
    }

    def __init__(self, policy, context):
        super(DSHMProcessorPolicyV1, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    def __getitem__(self, key):
        return self.__policy[key]

    def __setitem__(self, key, value):
        self.__policy[key] = value

def drive_device_list(value, policy_target, metavar='drive'):
    # Device IDs parsed as comma separated list of strings
    device_ids = value.split(',')
    device_ids = [ e for e in device_ids if e != '']
    # Validate each device ID against the NVSM context
    # NOTE: Due to the transient nature of device IDs (e.g. devices added
    #       or removed from the system), the CLI cannot have perfect
    #       knowledge of which device IDs are "valid" or not. Because of
    #       this, the CLI allows for arbitrary device ID strings, and only prints a
    #       warning message if a given device ID is unknown at the time it
    #       is set.
    try:
        subsystem_address = policy_target.address + '../drives'
        subsystem_target = policy_target._context.getTarget(subsystem_address)
        subsystem_targets = []
        subsystem_targets.extend(subsystem_target.targets)
        warn = set()
        for dev_id in device_ids:
            if dev_id in [ 'alerts', 'policy' ]:
                warn.add(dev_id)
                continue
            # Create a list of slot names using map, reduce, filter
            slots = [t.properties for t in subsystem_targets]
            slots = reduce(lambda x, y: x + y, slots)
            slots = [p for p in slots if p.name.strip().lower() == 'physicallocation_info']
            # Check if the given dev_id matches any drive slots
            if not any([str(p.value).lower() == dev_id.strip().lower() for p in slots]):
                warn.add(dev_id)
        for dev_id in warn:
            nvsm.log.warning('Unknown {metavar} ID: {dev_id}',
                    metavar=metavar, dev_id=dev_id)
    except Exception as e:
        nvsm.log.warning('Could not validate {metavar} ID list: {e_type}: {e}',
                metavar=metavar, e_type=type(e).__name__, e=e)
        import traceback
        nvsm.log.debug(traceback.format_exc())
    return device_ids

def controller_device_list(value, policy_target):
    # Device IDs parsed as comma separated list of strings
    device_ids = value.split(',')
    device_ids = [ e for e in device_ids if e != '']
    for dev_id in device_ids:
        if str(dev_id).lower() != 'c0':
            nvsm.log.warning('Unknown controller ID: {dev_id}', dev_id=dev_id)
    return device_ids

class DSHMStoragePolicyV1(DSHMProcessorPolicy):
    options = {
        'volume_mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='volume', relative_path='../volumes'),
            'path': [ 'Volume', 'mute_monitoring' ],
        },
        'volume_mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='volume', relative_path='../volumes'),
            'path': [ 'Volume', 'mute_notification' ],
        },
        'volume_poll_interval': {
            'path': [ 'Volume', 'poll_interval' ],
            'to_blob': to_non_negative_integer,
        },
        'drive_mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': drive_device_list,
            'path': [ 'Drive', 'mute_monitoring' ],
        },
        'drive_mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': drive_device_list,
            'path': [ 'Drive', 'mute_notification' ],
        },
        'drive_poll_interval': {
            'path': [ 'Drive', 'poll_interval' ],
            'to_blob': to_non_negative_integer,
        }
    }

    _controller_options = {
        'controller_mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': controller_device_list,
            'path': [ 'StorageController', 'mute_monitoring' ],
        },
        'controller_mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': controller_device_list,
            'path': [ 'StorageController', 'mute_notification' ],
        },
        'controller_poll_interval': {
            'path': [ 'StorageController', 'poll_interval' ],
            'to_blob': to_non_negative_integer,
        }
    }

    def __init__(self, policy, context):
        super(DSHMStoragePolicyV1, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context
        try:
            response = self.__context._queryBackend('/redfish/v1/Systems/1/Storage/1')
            controllers = response['StorageControllers']
        except:
            pass  # Not all systems have storage controller API (e.g. DGX-2)
        else:
            self.options.update(self._controller_options)

    def __getitem__(self, key):
        return self.__policy[key]

    def __setitem__(self, key, value):
        self.__policy[key] = value


class DSHMThermalPolicyV1(DSHMThermalPolicy):
    options = {
        'fan_mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='fan device', relative_path='../fans'),
            'path': [ 'Fan', 'mute_monitoring' ],
        },
        'fan_mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='fan device', relative_path='../fans'),
            'path': [ 'Fan', 'mute_notification' ],
        },
        'temperatures_mute_monitoring': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='temperature sensor', relative_path='../temperatures'),
            'path': [ 'Pdb', 'mute_monitoring' ],
        },
        'temperatures_mute_notification': {
            'to_display': lambda x: ','.join(x),
            'to_blob': to_device_list(metavar='temperature sensor', relative_path='../temperatures'),
            'path': [ 'Pdb', 'mute_notification' ],
        }
    }

    def __init__(self, policy, context):
        super(DSHMThermalPolicyV1, self).__init__(
                policy=policy, context=context)
        self.__policy = policy
        self.__context = context

    def __getitem__(self, key):
        return self.__policy[key]

    def __setitem__(self, key, value):
        self.__policy[key] = value


class PolicyBlob(object):
    def __init__(self, policyref, context):
        super(PolicyBlob, self).__init__()
        self.__ref = policyref
        self.__context = context
        self._blob = None

    def __getBlob(self):
        self._blob = {}
        response = self.__context._queryBackend(self.__ref)
        if 'Policy' not in response:
            raise nvsm.exceptions.SchemaError(
                    'Expected "Policy" attribute from NVSM backend')
        if type(response['Policy']) is list:
            if len(response['Policy']) == 0:
                blob = None
            else:
                blob = response['Policy'][0]
        else:
            blob = response['Policy']
        if not blob:
            nvsm.log.warning('Empty DSHM policy object')
            return
        nvsm.log.debug('Decoded DSHM policy blob: {0}', json.dumps(blob, indent=2))
        self._blob = blob

    def __putBlob(self, blob):
        self.__context._queryBackend(self.__ref, method='PUT', body=blob)
        # Fetch new blob (old blob probably out of date)
        self._blob = None
        self.__getBlob()
        self.__root = self._from_path(path=[])

    def _setPath(self, path, value):
        nvsm.log.debug(
            'Setting policy at path "{0}" to value "{1}"',
            path, value)
        new_blob = copy.deepcopy(self._blob)
        o = reduce(lambda x, y: x[y], path[:-1], new_blob)
        o[path[-1]] = value
        self.__putBlob(blob=new_blob)

    def _from_path(self, path):
        value = reduce(lambda x, y: x[y], path, self._blob)
        if type(value) is dict:
            return PolicyBlobObject(path=path, policy=self)
#        if type(value) is list:
#            return PolicyBlobArray(path=path, policy=self)
        return value

    def __getitem__(self, key):
        if self._blob == None:
            self.__getBlob()
            self.__root = self._from_path(path=[])
        return self.__root[key]

    def __setitem__(self, key, value):
        if self._blob == None:
            self.__getBlob()
            self.__root = self._from_path(path=[])
        self.__root[key] = value


class PolicyBlobObject(object):
    def __init__(self, path, policy):
        super(PolicyBlobObject, self).__init__()
        self.__path = path
        self.__policy = policy
        self.__cache = {}

    @property
    def __blob(self):
        return reduce(lambda x, y: x[y], self.__path, self.__policy._blob)

    def __getitem__(self, key):
        if key not in self.__blob:
            raise KeyError
        if key not in self.__cache:
            path = self.__path + [ key ]
            self.__cache[key] = self.__policy._from_path(path)
        return self.__cache[key]

    def __setitem__(self, key, value):
        path = self.__path + [ key ]
        self.__policy._setPath(path=path, value=value)


class PolicyBlobArray(object):
    pass  # TODO


class PolicyTarget(Target):
    def __init__(self, policyref, policyclass, address, context,
                 nodeIdentifier=None, parentIdentifier=None):
        super(PolicyTarget, self).__init__(
                address=address, context=context)
        self.__properties = None
        self.__policy = None
        self.__policyref = policyref
        self.__policyclass = policyclass
        self.__nodeIdentifier = nodeIdentifier
        self.__parentIdentifier = parentIdentifier

    def __initPolicy(self):
        # Use given policy class to abstract access to policy
        self.__policy = self.__policyclass.from_ref(
                policyref=self.__policyref,
                context=self._context)

    def __initProperties(self):
        self.__updateProperties()

    def _filter_policy(self, value):
        # filter based on node identifier
        filtered_values = []
        for v in value:
            if '+' in v:
                if self.__nodeIdentifier and v.split('+')[0] != self.__nodeIdentifier:
                    continue
                if self.__parentIdentifier and v.split('+')[1] != self.__parentIdentifier:
                    continue
            filtered_values.append(v)
        return filtered_values

    def __updateProperties(self):
        if self.__policy == None:
            self.__initPolicy()
        self.__properties = []
        options = type(self.__policy).options
        for name in sorted(options.keys()):
            option = options[name]
            try:
                value = reduce(lambda x, y: x[y], option['path'], self.__policy)
            except KeyError:
                nvsm.log.warning(
                    'Could not find policy: {}'.format(
                    name))
                continue
            if isinstance(value, list):
                value = self._filter_policy(value)
                value = [v.split('+')[-1] for v in value]
            if 'to_display' in option:
                value = option['to_display'](value)
            p = Property(target=self, name=name, value=value)
            self.__properties.append(p)

    def __getitem__(self, key):
        # Policy targets do not support subtargets at this time
        # (behaves as empty dict)
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties

    def set(self, name, value):
        if self.__policy == None:
            self.__initPolicy()
        options = type(self.__policy).options
        if name not in options:
            raise nvsm.exceptions.MissingProperty(
                'Target property "{0}" does not exist'.format(name))
        option = options[name]
        path = option['path']
        try:
            policyObject = reduce(lambda x, y: x[y], path[:-1], self.__policy)
        except KeyError:
            raise nvsm.exceptions.MissingProperty(
                'Target property "{0}" does not exist'.format(name))
        if 'to_blob' in option:
            value = option['to_blob'](value, policy_target=self)

        if isinstance(policyObject[path[-1]], list) \
            and path[-1] in ["mute_monitoring", "mute_notification"]:
            filtered_values = self._filter_policy(policyObject[path[-1]])
            # filter out alerts of this chassis
            filtered_out_values = [x for x in policyObject[path[-1]] if x not in filtered_values]
            # append serial number
            for i in range(len(value)):
                if self.__nodeIdentifier:
                    if self.__parentIdentifier:
                        value[i] = self.__parentIdentifier+"+"+value[i]
                    value[i] = self.__nodeIdentifier+"+"+value[i]
            value += filtered_out_values

        policyObject[path[-1]] = value
        self.__updateProperties()


class GPUPolicyTarget(PolicyTarget):
    def __init__(self, policyref, address, context):
        super(GPUPolicyTarget, self).__init__(
                policyref=policyref, policyclass=DSHMGPUPolicy,
                address=address, context=context)


class MemoryPolicyTarget(PolicyTarget):
    def __init__(self, policyref, address, context):
        super(MemoryPolicyTarget, self).__init__(
                policyref=policyref, policyclass=DSHMMemoryPolicy,
                address=address, context=context)

class PCIePolicyTarget(PolicyTarget):
    def __init__(self, policyref, address, context):
        super(PCIePolicyTarget, self).__init__(
                policyref=policyref, policyclass=DSHMPCIePolicy,
                address=address, context=context)


class PowerPolicyTarget(PolicyTarget):
    def __init__(self, policyref, address, context, nodeIdentifier):
        super(PowerPolicyTarget, self).__init__(
                policyref=policyref, policyclass=DSHMPowerPolicy,
                address=address, context=context, nodeIdentifier=nodeIdentifier)


class ProcessorPolicyTarget(PolicyTarget):
    def __init__(self, policyref, address, context):
        super(ProcessorPolicyTarget, self).__init__(
                policyref=policyref, policyclass=DSHMProcessorPolicy,
                address=address, context=context)


class StoragePolicyTarget(PolicyTarget):
    def __init__(self, policyref, address, context):
        super(StoragePolicyTarget, self).__init__(
                policyref=policyref, policyclass=DSHMStoragePolicy,
                address=address, context=context)


class ThermalPolicyTarget(PolicyTarget):
    def __init__(self, policyref, address, context, nodeIdentifier):
        super(ThermalPolicyTarget, self).__init__(
                policyref=policyref, policyclass=DSHMThermalPolicy,
                address=address, context=context, nodeIdentifier=nodeIdentifier)

class BladePolicyTarget(PolicyTarget):
    def __init__(self, policyref, address, context):
        super(BladePolicyTarget, self).__init__(
                policyref=policyref, policyclass=DSHMBladePolicy,
                address=address, context=context)

class CMMPolicyTarget(PolicyTarget):
    def __init__(self, policyref, address, context):
        super(CMMPolicyTarget, self).__init__(
                policyref=policyref, policyclass=DSHMCMMPolicy,
                address=address, context=context)

class NetworkPolicyTarget(PolicyTarget):
    def __init__(self, policyref, address, context, nodeIdentifier):
        super(NetworkPolicyTarget, self).__init__(
                policyref=policyref, policyclass=DSHMNetworkPolicy,
                address=address, context=context, nodeIdentifier=nodeIdentifier)

class NetworkPortPolicyTarget(PolicyTarget):
    def __init__(self, policyref, address, context, nodeIdentifier, parentIdentifier):
        super(NetworkPortPolicyTarget, self).__init__(
                policyref=policyref, policyclass=DSHMNetworkPortPolicy,
                address=address, context=context, nodeIdentifier=nodeIdentifier, parentIdentifier=parentIdentifier)

class NetworkDevFuncPolicyTarget(PolicyTarget):
    def __init__(self, policyref, address, context, nodeIdentifier, parentIdentifier):
        super(NetworkDevFuncPolicyTarget, self).__init__(
                policyref=policyref, policyclass=DSHMNetworkDevFuncPolicy,
                address=address, context=context, nodeIdentifier=nodeIdentifier, parentIdentifier=parentIdentifier)

class NVSwitchPolicyTarget(PolicyTarget):
    def __init__(self, policyref, address, context):
        super(NVSwitchPolicyTarget, self).__init__(
                policyref=policyref, policyclass=DSHMNVSwitchPolicy,
                address=address, context=context)

class GlobalPolicyTarget(PolicyTarget):
    def __init__(self, address, context):
        # XXX: NVSM backend does not give OData reference for policy resource
        policyref = '/nvsm/v1/Policy/Global'
        super(GlobalPolicyTarget, self).__init__(
                policyref=policyref, policyclass=DSHMGlobalPolicy,
                address=address, context=context)
