# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



from nvsm.backend.context import Target, dict_to_properties

from nvsm.backend.alerts import AlertsTarget
from nvsm.backend.policy import ProcessorPolicyTarget

import nvsm.log

bad_status = [ 'warning', 'critical' ]
good_status = [ 'ok' ]
def colorize_status(value):
    if value == None:
        return "none"
    if value.lower() in good_status:
        return 'green'
    if value.lower() in bad_status:
        return 'red'
    return 'none'

class ProcessorTarget(Target):
    def __init__(self, processorref, address, context):
        super(ProcessorTarget, self).__init__(
                address=address, context=context)
        self.__processorref = processorref
        self.__properties = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        processorinfo = self._context._queryBackend(self.__processorref)
        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            if name.lower().startswith('oem_'):
                return self._context.showall
            if name == 'schema':
                return False
            return True
        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            return (name, value, color)
        self.__properties = list(dict_to_properties(
                    target=self, d=processorinfo,
                    criteria=filter_properties,
                    mapping=map_properties))
        self.__properties.sort(key=lambda x : x.name)

    def __getitem__(self, key):
        # System processor has no targets at this time
        # (behaves as an empty dictionary)
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class ProcessorCollectionTarget(Target):
    def __init__(self, systemref, address, context):
        super(ProcessorCollectionTarget, self).__init__(
                address=address, context=context)
        self.__systemref = systemref
        self.__targets = None

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}

        # Assume every system has processor alert targets (for DSHM alerts)
        alertsTarget = AlertsTarget(
            ref='/nvsm/v1/Alerts/Processors',  # XXX
            address=(self.address + 'alerts'),
            context=self._context)
        self.__targets['alerts'] = alertsTarget

        # Assume every system has processor policy targets (for configuring DSHM)
        policyTarget = ProcessorPolicyTarget(
                # FIXME: NVSM backend does not provide OData reference here
                policyref='/nvsm/v1/Policy/Processors',
                address=(self.address + 'policy'),
                context=self._context)
        self.__targets['policy'] = policyTarget

        # Enumerate processor targets with NVSM backend
        try:
            processorinfo = self._context._queryBackend(
                    '{0}/Processors'.format(self.__systemref))
        except:
            pass
        processorinfo = processorinfo.get('Members', [])
        if not processorinfo:
            processorinfo = []
        for processorref in processorinfo:
            processorref = processorref.get('@odata.id', None)
            if not processorref:
                nvsm.log.warning(
                    "NVSM backend did not give '@odata.id' for system processor")
                continue
            name = processorref.strip('/').split('/')[-1]
            target = ProcessorTarget(
                    processorref=processorref,
                    address=(self.address + name),
                    context=self._context)
            self.__targets[name] = target

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?
