# Copyright (c) 2018-2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.




import time
import json
import os
import sys
from nvsm.exceptions import BackendError
from nvsm.backend.context import Target, Property, dict_to_properties
import nvsm.prompt
import nvsm.log
from nvsm.progress import ProgressBar
from datetime import timedelta
from datetime import datetime

class RdsDockerTarget(Target):
    ''' Control the download process image. '''
    def __init__(self, ref, address, context):
        super(RdsDockerTarget, self).__init__(
                address=address, context=context)
        self.__ref = ref.replace("redfish", "nvsm") +"/Actions/ComputerSystem.Oem.collectDump"
        self.__properties = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        response = self._context._queryBackend(self.__ref, forceRefresh=True)
        propertynotset ='<NOT_SET>'   
        defaultusername='$oauthtoken' 
        cmd_list = response.get('CmdList', propertynotset)
        docker_imageref = response.get('DockerImageRef', propertynotset)
        docker_username = response.get('DockerUsername', defaultusername)
        docker_password = {
            "value": response.get('DockerPassword', propertynotset),
            "promptvalue": propertynotset
        }

        def getPasswordPrompt():
            if docker_password["promptvalue"] == docker_password["value"]:
                return propertynotset
            else:
                return "**********"
            

        self.__properties.append(Property(
                    target=self,
                    name='CmdList',
                    value=cmd_list))
        self.__properties.append(Property(
                    target=self,
                    name='DockerImageRef',
                    value=docker_imageref))
        self.__properties.append(Property(
                    target=self,
                    name='DockerUsername',
                    value=docker_username))
        self.__properties.append(Property(
                    target=self,
                    name='DockerPassword',
                    value=getPasswordPrompt()))

    def __getitem__(self, key):
        raise KeyError  # No sub-targets

    @property
    def targets(self):
        return []  # No sub-targets

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties upon first access
            self.__initProperties()
        return self.__properties

    def set(self, name, value):
        for x in ['CmdList', 'DockerUsername', 'DockerPassword', 'DockerImageRef' ]:
            if name.lower() == x.lower():
                if name.lower() == 'cmdlist':
                    body=value.split(",")
                else:
                    body=value
                response=self._context._queryBackend(self.__ref,
                        method='POST', body={ x: body })
                return
        if name.lower() == 'state':
            body = { 'State': value }
            response = self._context._queryBackend(self.__ref)
            if 'JobID' in response.get('Status', {}):
                body['JobID'] = response['Status']['JobID']
            self._context._queryBackend(self.__ref, method='PUT', body=body)
            return
        if name.lower() in [ 'message', 'progress' ]:
            raise nvsm.exceptions.MissingProperty(
                'Verb "set" not supported for "{0}" target property'.format(name))
        raise nvsm.exceptions.MissingProperty(
            'Target property "{0}" does not exist'.format(name))

    def start(self, skip_prompt=True, wait=False):
        propertynotset ='N/A'
        class DockerFetchJob(object):
            def __init__(self, target, statusref):
                super(DockerFetchJob, self).__init__()
                self.__target = target
                self.__statusref = statusref
                self.stream = None
                self.__status_stream =None
                try:
                    self.progress_bar = ProgressBar(file=sys.stdout)
                except Exception as e:
                    self.progress_bar = None
                self.waitforJob()

            @property
            def uninterruptible(self):
                # Fetch jobs can run in background
                return False

            def waitforJob(self):
                if self.stream == None:
                    delete_TO_minutes = 120
                    wait_until = datetime.now() + timedelta(minutes=delete_TO_minutes)
                    break_loop = False
                    while not break_loop:
                        stream = self.__target._context._queryBackend(self.__statusref, method='GET')
                        # Apply line buffering to the status stream
                        stream["State"] = stream["TaskState"]
                        stream["Progress"] = stream["PercentComplete"]
                        if "Nvidia" in stream["Oem"]:
                            if "Message" in stream["Oem"]["Nvidia"]:
                                stream["msgs"]=stream["Oem"]["Nvidia"]["Message"]
                        if self.progress_bar:
                            self.progress_bar.updateProgress(percent=stream["Progress"])
                        if wait_until < datetime.now() or stream["State"] == "Completed" or stream["State"]=="Exception":
                            if "Nvidia" in stream["Oem"]:
                                if "Dump Location" in stream["Oem"]["Nvidia"]:
                                    print("Dump Location: "+stream["Oem"]["Nvidia"]["Dump Location"])
                                if stream["State"] == "Exception":
                                    raise nvsm.exceptions.BackendError(' '.join(stream["msgs"]))
                            break_loop = True  
                        self.stream = [json.dumps(stream)]
                    self.status_stream=self.stream
                    return self.status_stream

        #cancel_existing_prompt(ref=self.__ref, context=self._context)

        # check to see that we have expected data -- name/password/image
        response = self._context._queryBackend(self.__ref)
        dockerimageref=response.get('DockerImageRef', propertynotset)
        dockerpullrequired=False
        if dockerimageref == propertynotset:
            raise nvsm.exceptions.MissingProperty("Missing or empty property for 'DockerImageRef'")
        if dockerpullrequired:
            for key in [ 'DockerUsername', 'DockerPassword' ]:
                value = response.get(key, None)
                if value == None or value == "":
                    raise nvsm.exceptions.MissingProperty("Missing or empty property for '%s'" %(key))
        cmdlist = response.get('CmdList', propertynotset)
        if cmdlist == propertynotset:
            raise nvsm.exceptions.MissingProperty("'CmdList' is not set")

        # if local image (set this way in cli), then don't pull down any image
        # expect it to be local

        username = response.get('DockerUsername', None)
        if username == "local":
            nvsm.log.debug("'local' user name, assume image is already local, no fetch")
            return 0

        # making an assumption here: This code wil be executed for one of two reasons
        #
        #  1) it's being called from the nvsm-cli that set/valdidate everything
        #  2) it's being called because of a smash command from an "expert" user
        #
        # in both cases, we must assume that the image/username/password are good,
        # If they are wrong then no prompting as a secondary measure is
        # necessary -- the caller screwed up: simply indicate error and stop
        #
        # Note: Orignally this code use to prompt, don't do that any mode since
        #       code is smarter in cli, and we expect those who use the scripting
        #       smash interface 1) to be smart, and 2) not to want to prompt.

        # Ask API server to start FW download
        body = {'StartDump': True}
        post_response = self._context._queryBackend(self.__ref, method='POST', body=body)
        message=post_response.get('Message')
        jsop=json.loads(message)
        job_uri=jsop['taskUri']
        # Return handle to image fetch job
        time.sleep(30)
        return DockerFetchJob(target=self, statusref=job_uri)
