# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.


import json
import os
import re
import time
import sys

from nvsm.backend.context import Target, Property, dict_to_properties

import nvsm.log
import nvsm.prompt

from nvsm.backend.alerts import AlertsTarget
from nvsm.backend.policy import StoragePolicyTarget
from nvsm.color import colorize

bad_status = ['warning', 'critical', 'fail', 'predictive failure']
good_status = ['ok', 'normal']


def colorize_status(value):
    if value == None:
        return "none"
    if value.lower() in good_status:
        return 'green'
    if value.lower() in bad_status:
        return 'red'
    return 'none'


class StorageControllerTarget(Target):
    def __init__(self, controllerindex, controllerref, address, context):
        super(StorageControllerTarget, self).__init__(
            address=address, context=context)
        self.__controllerref = controllerref
        self.__controllerindex = controllerindex

        self.__properties = None

    def __getitem__(self, key):
        raise KeyError

    @property
    def targets(self):
        return []

    @property
    def properties(self):
        if self.__properties == None:
            # Enumerate controller properties upon first access
            controllerinfo = self._context._queryBackend(self.__controllerref)[
                'StorageControllers'][0]
            controllerinfo = controllerinfo.get(
                "Oem").get("Nvidia").get("Phys", [])
            controllerinfo = controllerinfo[int(self.__controllerindex)]
            # Squash hierarchies (SMASH does not support proprety hierarchies)

            def filter_properties(name, value):
                if 'odata' in name.lower():
                    return False
                if name.lower() in ['port', 'valid', 'deviceaddress',
                                    'linkspeed', 'phyid', 'invlddwrdcnt', 'lsdwrdsyncn',
                                    'phyrst_prblcn', 'phyno']:
                    return True
                return False
            self.__properties = list(dict_to_properties(
                target=self, d=controllerinfo,
                criteria=filter_properties))
            default_properties = ['linkspeed',
                                  'phyrst_prblcn',
                                  'lsdwrdsyncn',
                                  'phyno',
                                  'valid',
                                  'phyid',
                                  'deviceaddress',
                                  'port',
                                  'invlddwrdcnt', ]

            def compare_properties(a):
                try:
                    i = default_properties.index(a.name.lower())
                except ValueError:
                    i = len(default_properties)
                return i
            self.__properties.sort(key=compare_properties)
        return self.__properties


class StorageControllersTarget(Target):
    def __init__(self, storageref, address, context):
        super(StorageControllersTarget, self).__init__(
            address=address, context=context)
        assert(self.name == 'controllers')

        self.__storageref = storageref
        self.__targets = None

        self.__properties = None

    def __initTargets(self):
        # Enumerate storage controllers with NVSM backend
        self.__targets = {}
        storageinfo = {}
        try:
            storageinfo = self._context._queryBackend(self.__storageref)
        except:
            pass
        controllersinfo = storageinfo["StorageControllers"][0]

        members = controllersinfo.get("Oem").get("Nvidia").get("Phys")
        if members == None:
            members = []
        for controllerref in members:
            controllerref = controllerref.get('@odata.id', None)
            if not controllerref:
                continue
            index = controllerref.split('/')[-1]
            index_name = "phy" + index
            controller = StorageControllerTarget(
                controllerindex=index,
                controllerref=self.__storageref,
                address=(self.address + index_name),
                context=self._context)
            self.__targets[index_name] = controller

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            storageinfo = self._context._queryBackend(self.__storageref)
            controllersinfo = storageinfo["StorageControllers"][0]

            def filter_properties(name, value):
                if 'odata' in name.lower():
                    return False
                if name.lower() in ['memberid', 'name', 'manufacturer',
                                    'model', 'serialnumber', 'partnumber', 'speedgbps',
                                    'firmwareversion', 'supportedcontrollerprotocols',
                                    'supporteddeviceprotocols', 'status_healthrollup']:
                    return True
                return False

            def map_properties(name, value):
                color = 'none'
                if 'health' in name.strip().lower():
                    color = colorize_status(value)
                elif name.strip().lower() == 'supporteddeviceprotocols':
                    value = ', '.join(value)
                    value = '[ {0} ]'.format(value)
                return (name, value, color)
            self.__properties = list(dict_to_properties(
                target=self, d=controllersinfo,
                criteria=filter_properties,
                mapping=map_properties))
            default_properties = ['status_healthrollup',
                                  'name',
                                  'supporteddeviceprotocols',
                                  'serialnumber',
                                  'memberid',
                                  'supportedcontrollerprotocols',
                                  'model',
                                  'speedgbps',
                                  'partnumber',
                                  'firmwareversion',
                                  'Manufacturer', ]

            def compare_properties(a):
                try:
                    i = default_properties.index(a.name.lower())
                except ValueError:
                    i = len(default_properties)
                return i
            self.__properties.sort(key=compare_properties)
        return self.__properties


class StorageDriveEjectTarget(Target):
    def __init__(self, ejectref, driveref, context, address):
        super(StorageDriveEjectTarget, self).__init__(
            context=context, address=address)
        self.__ejectref = ejectref
        self.__driveref = driveref
        self.__properties = None
        self.__targets = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties

    def start(self, skip_prompt=False, wait=False, argv=None):

        # Give user opportunity to cancel
        if not skip_prompt:
            message = '''Once the Drive is Ejected, volume on the drive will be unmounted.'''
            if not nvsm.prompt.Warn(
                    question='Start Drive Eject?',
                    message=message):
                return None  # User canceled drive eject

        # Ask the API server to send the request
        response = self._context._queryBackend(
            self.__ejectref, method='POST')
        last_update = ""
        file = sys.stdout
        storageuri = self.__driveref.rsplit('/', 2)[0]
        ejectdrive = self.__driveref.split('/')[-1]
        with nvsm.progress.ProgressBar(file=file) as progress_bar:
            progress_bar.update(
                percent=0.0,
                message="In Progress",
                error=False)
            retry = 0
            while True:
                response = self._context._queryBackend(
                    path=storageuri, forceRefresh=True)
                drives = response.get('Drives', [])
                found = False
                for driveref in drives:
                    driveref = driveref.get('@odata.id', None)
                    if not driveref:
                        continue
                    index = driveref.split('/')[-1]
                    if index == ejectdrive:
                        found = True
                        retry += 1
                        time.sleep(10)
                        progress_bar.update(
                            percent=25.0 * retry,
                            message="In Progress",
                            error=False)
                        break
                if retry >= 4:
                    last_update = "Failed"
                    progress_bar.update(
                        percent=0,
                        message="Unable to confirm drive eject. Please use 'nvsm show drives' to confirm.",
                        error=True)
                    break
                elif found == False:
                    last_update = "Success"
                    progress_bar.update(
                        percent=100.0,
                        message="Success",
                        error=False)
                    break

        print('Status: ', end='', file=file)
        if last_update.lower() in ['done', 'finished', 'success']:
            color = 'green'
        else:
            color = 'red'
        with colorize(color):
            print(last_update, file=file)
        return


class StorageDriveTarget(Target):
    def __init__(self, driveref, address, context):
        super(StorageDriveTarget, self).__init__(
            address=address, context=context)
        self.__driveref = driveref
        self.__targets = None
        self.__ejectRef = None
        self.__properties = None

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    def __initTargets(self):
        self.__targets = {}
        self.__initEjectDrive()

    def __initEjectDrive(self):
        assert self.__ejectRef == None
        # Drive supporting Eject
        driveuri = self._context._queryBackend(self.__driveref, forceRefresh=True)
        try:
            self.__ejectRef = driveuri['Oem']['Actions']['#Eject']['target'][0]
        except:
            return
        eject_drive = StorageDriveEjectTarget(
            ejectref=self.__ejectRef,
            driveref=self.__driveref,
            context=self._context,
            address=self.address + 'eject')
        self.__targets['eject'] = eject_drive

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Enumerate drive properties upon first access
            driveinfo = self._context._queryBackend(self.__driveref)
            # Squash hierarchies (SMASH does not support proprety hierarchies)

            def filter_properties(name, value):
                if 'odata' in name.lower() and 'protocol' not in name.lower():
                    return False
                if name.lower() == 'schema':
                    return False
                if name.lower() in ['physicallocation_info', 'capacitybytes', 'blocksizebytes',
                                    'serialnumber', 'capablespeedgbps', 'partnumber',
                                    'model', 'revision', 'handle', 'manufacturer',
                                    'status_state', 'status_health', 'encryption_ability',
                                    'name', 'mediatype', 'indicatorled',
                                    'encryptionstatus', 'hotsparetype', 'protocol_@odata.id',
                                    'failurepredected', 'negotiatedspeedsgbs', 'id', 'oem_ledstatus_status',
                                    'oem_ledstatus_activity', 'oem_ledstatus_fail', 'oem_ledstatus_locate']:
                    return True
                return self._context.showall

            def map_properties(name, value):
                color = 'none'
                if 'status' in name.strip().lower():
                    color = colorize_status(value)
                if 'physicallocation_info' == name.strip().lower():
                    value = 'Slot' + str(value)
                if 'oem_ledstatus_status' == name.strip().lower() or 'oem_ledstatus_fail' == name.strip().lower():
                    name = 'LedStatus'
                if 'oem_ledstatus_activity' == name.strip().lower():
                    name = 'ActivityLed_Status'
                if 'oem_ledstatus_locate' == name.strip().lower():
                    name = 'LocateLed_Status'
                if 'protocol_@odata.id' == name.strip().lower():
                    name = 'Protocol'
                return (name, value, color)
            self.__properties = list(dict_to_properties(
                target=self, d=driveinfo,
                criteria=filter_properties,
                mapping=map_properties))
            default_properties = ['physicallocation_info',
                                  'blocksizebytes',
                                  'serialnumber',
                                  'partnumber',
                                  'model',
                                  'revision',
                                  'manufacturer',
                                  'status_state',
                                  'status_health',
                                  'name',
                                  'indicatorled',
                                  'mediatype',
                                  'encryptionstatus',
                                  'hotsparetype',
                                  'protocol',
                                  'capacitybytes',
                                  'negotiatedspeedsgbs',
                                  'id', ]

            def compare_properties(a):
                try:
                    i = default_properties.index(a.name.lower())
                except ValueError:
                    i = len(default_properties)
                return i
            self.__properties.sort(key=compare_properties)
        return self.__properties

    def set(self, nameValuePairs):
        self.set_array(nameValuePairs)

    def set_array(self, nameValuePairs):
        body_dict = {}
        body_dict['LEDType'] = None
        body_dict['StatusCode'] = None
        for pair in nameValuePairs:
            name = pair[0]
            value = pair[1]
            found = False
            if name == 'LEDType':
                found = True
                body_dict['LEDType'] = value
                nvsm.log.debug('found', value)
            elif (name == 'StatusCode'):
                found = True
                body_dict['StatusCode'] = int(value)
        if found is False:
            raise nvsm.exceptions.MissingProperty(
                'Target property "{0}" does not exist'.format(name))
        elif len(nameValuePairs) != 2:
            raise nvsm.exceptions.CommandError(
                'Incorrect parameters for set operation')
        ref = re.sub(r'redfish', 'nvsm', self.__driveref + '/LEDStatus')
        response = self._context._queryBackend(ref,
                                               method='PUT', body=body_dict)
        if hasattr(response, 'get') and response.get('error'):
            nvsm.log.error(response['error'])


class StorageDrivesTarget(Target):
    def __init__(self, storageref, address, context):
        super(StorageDrivesTarget, self).__init__(
            address=address, context=context)
        assert(self.name == 'drives')

        self.__storageref = storageref
        self.__targets = None

    def __initTargets(self):
        # Enumerate storage drives with NVSM backend
        self.__targets = {}
        try:
            storageinfo = self._context._queryBackend(self.__storageref)
        except:
            pass
        if not hasattr(storageinfo, 'get'):
            storageinfo = {}
        drives = storageinfo.get('Drives', [])
        if drives == None:
            drives = []
        for driveref in drives:
            driveref = driveref.get('@odata.id', None)
            if not driveref:
                continue
            index = driveref.split('/')[-1]
            drive = StorageDriveTarget(
                driveref=driveref,
                address=(self.address + index),
                context=self._context)
            self.__targets[index] = drive

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?


class StorageVolumeEncryptTarget(Target):
    def __init__(self, volumeid, volumeref, encryptref, address, context):
        super(StorageVolumeEncryptTarget, self).__init__(
            address=address, context=context)
        self.__volumeid = volumeid
        self.__volumeref = volumeref
        self.__encryptref = encryptref

        self.__properties = None
        self.__targets = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
            if self.__volumeid == None:
                encryptinfo = self._context._queryBackend(
                    '/nvsm/v1/Systems/1/Storage/1/Volumes/Actions/Encrypt', forceRefresh=True)
            else:
                encryptinfo = self._context._queryBackend(
                    '/nvsm/v1/Systems/1/Storage/1/Volumes/Actions/Encrypt/'+self.__volumeid, forceRefresh=True) 

            def filter_properties(name, value):
                if 'odata' in name.lower():
                    return False
                if name.lower() == 'schema':
                    return False
                if name.lower() in ['autolocking_drives',
                                    'locked_drives',
                                    'sed_capable',
                                    'tpm_enabled',
                                    'tpm_version']:
                    return True
                return False

            def map_properties(name, value):
                color = 'none'
                if 'status' in name.strip().lower():
                    color = colorize_status(value)
                    if not value:
                        value = '[ ]'

                if name.lower() in ['locked_drives', 'sed_capable', 'autolocking_drives']:
                    nvsm.log.info("name: {} value:{}".format(name, value))

                    def to_int(s):
                        try:
                            result = int(s)
                        except:
                            result = s
                        return result
                    if value is not None:
                        value.sort(key=to_int)

                return (name, value, color)
            self.__properties = list(dict_to_properties(
                target=self, d=encryptinfo,
                criteria=filter_properties,
                mapping=map_properties))

        return self.__properties
    def EncryptJob(self):
        status = "Unknown"
        if not self.__encryptref:
            time.sleep(0.5)
            return status
        time.sleep(5.0)
        while True:
            response = self._context._queryBackend(
                path=self.__encryptref['status'], forceRefresh=True)
            # Verify status in loop until Complete
            try:
                status = response['status']
            except KeyboardInterrupt:
                print('\nKeyboard interrupt. (job running in background)')
                return response

            except:
                return response

            if status != "InProgress":
                return response    
            time.sleep(1)
                
    def enable(self, skip_prompt=False, wait=False, argv=None):
        key = None
        file_name = None
        tpm_key = ""
        if self.__volumeid == None:
            nvsm.log.error("Operation not permitted")
            return
        encryptinfo = self._context._queryBackend(
            '/nvsm/v1/Systems/1/Storage/1/Volumes/Actions/Encrypt/'+self.__volumeid)

        # Only if TPM is enabled and Owned and if version is less than TPM2
        try:
            if encryptinfo['TPM_Version'] != "":
                tpm_version = float(encryptinfo['TPM_Version'])
            if encryptinfo['TPM_Enabled'] == True and tpm_version < 2.0:
                tpm_key = nvsm.prompt.String(message='TPM Enabled',
                                             prompt='Please enter TPM Owner password:',
                                             password=True)
                if tpm_key == "":
                    nvsm.log.error("Invalid key entered")
                    return
        except:
            pass

        if (os.path.exists("/etc/nv-disk-encrypt/.secrets.enc")):
            try:
                response = nvsm.prompt.Warn(
                    message='''This system has previously been initialized for encryption.
                                   Initializing again will re-key your vault and drives.''',
                    question='Do you wish to continue?')
            except EOFError:
                # Upon receiving EOF, cancel the job and leave quickly
                raise

            if not response:
                return
        file_name = nvsm.prompt.String(prompt='Please enter filename(absolute path):',
                                       message='''Filename must be a json formatted initializtion file
                           that should contain mapping of drive device path to keys.
                           If no file input is given random keys shall be used''',
                                       )
        if file_name is None:
            filestring = ""
        else:
            filestring = file_name
        while True:
            key = nvsm.prompt.String(prompt='Please enter Vault Encryption password:',
                                     message='''Vault Encryption password should be a combination of numeric
                               and alphanumeric characters e.g: a-z,$#%.''',
                                     password=True,
                                     re_enter=True)
            if key == "" or len(key)<8:
                nvsm.log.error("Keys must be at least 8 characters in length")
                continue
            else:
                break


        nvsm.log.info("vaultkey:{} filename:{}".format(key, filestring))
        body = {"VaultKey": key, "Filename": filestring, "TpmKey": tpm_key}

        response = self._context._queryBackend(self.__encryptref['enable'],
                                               method='POST',
                                               body=body)

        job_response = self.EncryptJob()
        if job_response['message'] == "":
            print("Success")
        else:
            nvsm.log.error(job_response['message'])
        return

    def lock(self, skip_prompt=False, wait=False, keyvalues=None):
        self.__oparg = None
        key = ""
        if self.__volumeid == None:
            nvsm.log.error("Operation not permitted")
            return

        lock_args = ['key']
        nvsm.log.info('keyvalues:{}'.format(keyvalues))
        try:
            for pair in keyvalues:
                name = pair[0]
                if name not in lock_args:
                    nvsm.log.error("Invalid Argument {}".format(name))
                    return
                else:
                    key = pair[1]
        except:
            pass
        body = {"VaultKey": key}
        response = self._context._queryBackend(
            self.__encryptref['lock'], method='POST', body=body)
        job_response = self.EncryptJob()
        if job_response['message'] == "Invalid Key":
            key = nvsm.prompt.String(prompt='Please enter Vault Encryption password:',
                                     message='''Invalid key''',
                                     password=True)
            response = self._context._queryBackend(
                self.__encryptref['lock'], method='POST', body={"VaultKey": key})
            job_response = self.EncryptJob()

        if job_response['message'] != "":
            nvsm.log.error(job_response['message'])
        else:
            print("Success")
        return

    def unlock(self, skip_prompt=False, wait=False, keyvalues=None):
        self.__oparg = None
        key = ""
        unlock_args = ['key']
        nvsm.log.info('keyvalues:{}'.format(keyvalues))
        try:
            for pair in keyvalues:
                name = pair[0]
                if name not in unlock_args:
                    nvsm.log.error("Invalid Argument {}".format(name))
                    return
                else:
                    key = pair[1]
        except:
            pass
        body = {"VaultKey": key}
        response = self._context._queryBackend(
            self.__encryptref['unlock'], method='POST', body=body)
        job_response = self.EncryptJob()
        try:
            if job_response['message'] == "Invalid Key":
                key = nvsm.prompt.String(prompt='Please enter Vault Encryption password:',
                                        message='''Invalid key''',
                                        password=True)
                response = self._context._queryBackend(
                    self.__encryptref['unlock'], method='POST', body={"VaultKey": key})
                job_response = self.EncryptJob()

            if job_response['message'] != "" and self.__volumeid is not None:
                nvsm.log.error(job_response['message'])
            else:
                print("Success")
        except:
            pass
        return

    def disable(self, skip_prompt=False, wait=False, keyvalues=None):
        self.__oparg = None
        key = ""
        disable_args = ['key']
        nvsm.log.info('keyvalues:{}'.format(keyvalues))
        try:
            for pair in keyvalues:
                name = pair[0]
                if name not in disable_args:
                    nvsm.log.error("Invalid Argument {}".format(name))
                    return
                else:
                    key = pair[1]
        except:
            pass
        response = self._context._queryBackend(
            self.__encryptref['disable'], method='POST', body={"VaultKey": key})
        job_response = self.EncryptJob()
        try:
            if job_response['message'] == "Invalid Key":
                key = nvsm.prompt.String(prompt='Please enter Vault Encryption password:',
                                     message='''Invalid key''',
                                     password=True)
                response = self._context._queryBackend(
                    self.__encryptref['disable'], method='POST', body={"VaultKey": key})
                job_response = self.EncryptJob()

            if job_response['message'] != "" and self.__volumeid is not None:
                nvsm.log.error(job_response['message'])
            else:
                print("Success")
        except:
            pass
        return
        #job = NvDriveEncryptDisableJob(target=self, arg=argv)
        # return job

    def factoryreset(self, skip_prompt=False, wait=False, argv=None):
        try:
            response = nvsm.prompt.Warn(
                message='''DATA_LOSS: This will erase all the data present 
                    on the drives and unlock the drives''',
                question='Do you wish to continue?',
                color='red')
        except EOFError:
            # Upon receiving EOF, cancel the job and leave quickly
            raise

        if not response:
            return

        file_name = None
        file_name = nvsm.prompt.String(prompt='Please enter filename(absolute path):',
                                       message='''File must be a JSON file containing mapping 
                of drive serial no. to PSID (Present on Drive Label). ''',
                                       )
        if file_name != "":
            filestring = file_name
        else:
            drive_id = nvsm.prompt.String(prompt='Please enter drive serial number:',
                                          message='''Drive serial number printed on the drive label''',)
            if drive_id is None:
                return

            psid = nvsm.prompt.String(prompt='Please enter PSID:',
                                      message='''PSID printed on the drive label''',)
            if psid is None:
                return
            file_name = drive_id + "@#" + psid
        body = {"Filename": file_name}
        response = self._context._queryBackend(
            self.__encryptref['factoryreset'], method='POST', body=body)
        job_response = self.EncryptJob()
        if job_response['message'] != "" and self.__volumeid is not None:
            nvsm.log.error(job_response['message'])
        else:
            print("Success")
        return

        #job = NvDriveEncryptRecoverJob(target=self, arg=argv)
        # return job

    def erase(self, skip_prompt=False, wait=False, keyvalues=None):

        key = ""
        erase_args = ['key']
        nvsm.log.info('keyvalues:{}'.format(keyvalues))
        try:
            for pair in keyvalues:
                name = pair[0]
                if name not in erase_args:
                    nvsm.log.error("Invalid Argument {}".format(name))
                    return
                else:
                    key = pair[1]
        except:
            pass
        try:
            response = nvsm.prompt.Warn(
                message='''This will erase all the data on the drives, 
                               recovery of data is not possible''',
                question='Do you wish to continue?')
        except EOFError:
            # Upon receiving EOF, cancel the job and leave quickly
            raise

        if not response:
            return

        body = {"VaultKey": key}
        response = self._context._queryBackend(
            self.__encryptref['erase'], method='POST', body=body)
        job_response = self.EncryptJob()
        try:
            if job_response['message'] == "Invalid Key":
                key = nvsm.prompt.String(prompt='Please enter Vault Encryption password:',
                                        message='''Invalid key''',
                                        password=True)
                response = self._context._queryBackend(
                    self.__encryptref['erase'], method='POST', body={"VaultKey": key})
                job_response = self.EncryptJob()

            if job_response['message'] != "" and self.__volumeid is not None:
                nvsm.log.error(job_response['message'])
            else:
                print("Success")
        except:
            pass
        return
        #job = NvDriveEncryptDeleteJob(target=self, arg=argv)
        # return job

    def export(self, skip_prompt=False, wait=False, argv=None):
        if self.__volumeid == None:
            nvsm.log.error("Operation not permitted")
            return
        file_name = ""
        try:
            file_name = argv[1]
        except:
            pass
        nvsm.log.info('Args:{}'.format(argv))
        key = nvsm.prompt.String(prompt='Please enter the password',
                                 message='''Use the password entered for Vault Encryption''',
                                 password=True)
        if key == "" :
            nvsm.log.error("Invalid Vault Key")
            return
    
        body = {"VaultKey": key}
        response = self._context._queryBackend(
            self.__encryptref['export'], method='POST', body=body)
        job_response = self.EncryptJob()
        if job_response['message'] != "" and self.__volumeid is not None:
                nvsm.log.error(job_response['message'] )
        else:
            print("Drive Keys exported to /tmp/secrets.out")
        return

    def rotate(self, skip_prompt=False, wait=False, argv=None):
        if self.__volumeid == None:
            nvsm.log.error("Operation not permitted")
            return
        key = None
        file_name = None

        file_name = nvsm.prompt.String(prompt='Please enter filename(absolute path):',
                                       message='''Filename must be a json formatted initializtion file
                           that should contain mapping of drive device path to keys.
                           If no file input is given random keys shall be used''',
                                       )
        if file_name is None:
            filestring = ""
        else:
            filestring = file_name

        key = nvsm.prompt.String(prompt='Please enter Vault Encryption password:',
                                 message='''Vault Encryption password should be a combination of numeric
                           and alphanumeric characters e.g: a-z,$#%.''',
                                 password=True)
        nvsm.log.info("VaultKey:{} Filename:{}".format(key, filestring))
        body = {"VaultKey": key, "Filename": filestring}
        response = self._context._queryBackend(self.__encryptref['rotate'],
                                               method='POST',
                                               body=body)
        job_response = self.EncryptJob()
        if job_response['message'] != "" and self.__volumeid is not None:
                nvsm.log.error(job_response['message'])
        else:
            print("Success")
        return

        return


class StorageVolumeEspRebuildTarget(Target):
    def __init__(self, esprebuildref, context, address, progressref):
        super(StorageVolumeEspRebuildTarget, self).__init__(
            context=context, address=address)
        self.__esprebuildref = esprebuildref
        self.__progressref = progressref
        self.__properties = None
        self.__volumetype = "EFI Partition"
        self.__targets = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    
    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties

    def start(self, skip_prompt=False, wait=False, argv=None):
        class VolumeRebuildJob(object):
            def __init__(self, target, volumetype, progressref):
                super(VolumeRebuildJob, self).__init__()
                self.__target = target
                self.__volumetype = volumetype
                self.__progressref = progressref

            @property
            def uninterruptible(self):
                # Storage rebuild jobs can run in background
                return False

            @property
            def status_stream(self):
                if not self.__progressref:
                    # Some volumes do not support rebuild progress
                    status = {
                        'State': 'Done',
                        'Message': 'Rebuilding {type}...'.format(
                            type=self.__volumetype),
                        'Progress': 100.0
                    }
                    yield json.dumps(status)
                    time.sleep(0.5)
                    return
                # Wait for 5 seconds to give server time to start rebuild
                status = {
                    'State': 'Initiating',
                    'Message': 'Initiating {type} rebuild...'.format(
                        type=self.__volumetype),
                    'Progress': 0.0
                }
                yield json.dumps(status)
                time.sleep(5.0)
                # Periodically poll the API server (using sleep and yield)
                success = False
                last_progress = 0.0
                while True:
                    response = self.__target._context._queryBackend(
                        path=self.__progressref, forceRefresh=True)
                    # Attempt to parse progress into numerical value
                    try:
                        progress = float(
                            response['rebuild_percent'].strip('%'))
                        last_progress = progress
                    except:
                        # API sever returned to "none" progress, indicating
                        # rebuild is complete, or has failed.
                        if 'last_status' in response:
                            status = {
                                'State': 'Warning',
                                'Message': response['last_status'],
                                'Progress': last_progress
                            }
                            yield json.dumps(status)
                            break
                        else:
                            # No error in status
                            success = True
                            break  # Done
                    if progress >= 100.0:
                        success = True
                        break  # Done
                    status = {
                        'State': 'Rebuilding',
                        'Message': 'Rebuilding {type}...'.format(
                            type=self.__volumetype),
                        'Progress': progress
                    }
                    yield json.dumps(status)
                    time.sleep(0.5)
                if success:
                    status = {
                        'State': 'Done',
                        'Message': 'Finished rebuilding {type}'.format(
                            type=self.__volumetype),
                        'Progress': 100.0
                    }
                    yield json.dumps(status)

        # Give user opportunity to cancel rebuild
        if not skip_prompt:
            message = '''Once the ESP rebuild process is started, the
                           process cannot be stopped.'''
            if not nvsm.prompt.Warn(
                    question='Start ESP rebuild?',
                    message=message):
                return None  # User canceled rebuild

        # Ask the API server to start volume rebuild
        self._context._queryBackend(
            self.__esprebuildref, method='POST', body={'RaidType': 'EFI Partition'})
        
        # Return handle representing volume rebuild job progress
        job = VolumeRebuildJob(
            target=self,
            volumetype=self.__volumetype,
            progressref=self.__progressref)
        return job


class StorageVolumeRebuildTarget(Target):
    def __init__(self, volumeid, volumeref, rebuildref, progressref, address, context):
        super(StorageVolumeRebuildTarget, self).__init__(
            address=address, context=context)
        self.__volumeid = volumeid
        self.__volumeref = volumeref
        self.__rebuildref = rebuildref
        self.__progressref = progressref
        self.__properties = None
        self.__targets = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        self.__initProgress()

    def __initProgress(self):
        progress = None
        if self.__progressref:
            # Query rebuild progress from API server
            try:
                response = self._context._queryBackend(self.__progressref)
                progress = response['rebuild_percent']
            except:
                pass
        if not progress:
            progress = 'N/A'
        p = Property(
            target=self,
            name='Progress',
            value=progress)
        self.__properties.append(p)

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties

    def start(self, skip_prompt=False, wait=False, argv=None):
        class VolumeRebuildJob(object):
            def __init__(self, target, volumeid, volumetype, progressref):
                super(VolumeRebuildJob, self).__init__()
                self.__target = target
                self.__volumeid = volumeid
                self.__volumetype = volumetype
                self.__progressref = progressref

            @property
            def uninterruptible(self):
                # Storage rebuild jobs can run in background
                return False

            @property
            def status_stream(self):
                if not self.__progressref:
                    # Some volumes do not support rebuild progress
                    status = {
                        'State': 'Done',
                        'Message': 'Rebuilding {type} on volume {id}...'.format(
                            type=self.__volumetype,
                            id=self.__volumeid),
                        'Progress': 100.0
                    }
                    yield json.dumps(status)
                    time.sleep(0.5)
                    return
                # Wait for 5 seconds to give server time to start rebuild
                status = {
                    'State': 'Initiating',
                    'Message': 'Initiating {type} rebuild on volume {id}...'.format(
                        type=self.__volumetype,
                        id=self.__volumeid),
                    'Progress': 0.0
                }
                yield json.dumps(status)
                time.sleep(5.0)
                # Periodically poll the API server (using sleep and yield)
                success = False
                last_progress = 0.0
                while True:
                    response = self.__target._context._queryBackend(
                        path=self.__progressref, forceRefresh=True)
                    # Attempt to parse progress into numerical value
                    try:
                        progress = float(
                            response['rebuild_percent'].strip('%'))
                        last_progress = progress
                    except:
                        # API sever returned to "none" progress, indicating
                        # rebuild is complete, or has failed.
                        if 'last_status' in response:
                            status = {
                                'State': 'Warning',
                                'Message': response['last_status'],
                                'Progress': last_progress
                            }
                            yield json.dumps(status)
                            break
                        else:
                            # No error in status
                            success = True
                            break  # Done
                    if progress >= 100.0:
                        success = True
                        break  # Done
                    status = {
                        'State': 'Rebuilding',
                        'Message': 'Rebuilding {type} on volume {id}...'.format(
                            type=self.__volumetype,
                            id=self.__volumeid),
                        'Progress': progress
                    }
                    yield json.dumps(status)
                    time.sleep(0.5)
                if success:
                    status = {
                        'State': 'Done',
                        'Message': 'Finished rebuilding {type} on volume {id}'.format(
                            type=self.__volumetype,
                            id=self.__volumeid),
                        'Progress': 100.0
                    }
                    yield json.dumps(status)

        # Determine volume type by querying the API
        try:
            if self.__volumeref == None:
                while True:
                    volumetype = ''
                    volumetype = nvsm.prompt.String(
                        prompt='Type of volume rebuild',
                        message='''In order to rebuild volume, volume type
                                    is required. Please specify the volume type to rebuild. (Options: raid-0 only)''')
                    if not volumetype:
                        return None
                    volumetype = volumetype.upper()
                    if volumetype == 'RAID-1' or volumetype == 'RAID-0':
                        self.__volumeid = 'raid'
                        break

                    if not nvsm.prompt.Warn(
                            question='Volume type invalid. Try again?',
                            message=''''''):
                        return None
            else:
                volumeinfo = self._context._queryBackend(self.__volumeref)
                volumetype = volumeinfo['VolumeType']
        except:
            volumetype = 'volume'

        # Rebuild of md0 requires spare drive, so we must prompt the user
        spare_drive = ''
        if volumetype == 'RAID-1':
            spare_drive = nvsm.prompt.String(
                prompt='Name of spare drive for {id} rebuild'.format(
                    id=self.__volumeid),
                message='''In order to rebuild this volume, a spare drive
                               is required. Please specify the spare drive to
                               use to rebuild {id}.'''.format(
                        id=self.__volumeid))
            if not spare_drive:
                return None  # User canceled rebuild
            # TODO: Check for validity of spare drive? How can we tell which drives are valid?

        # Give user opportunity to cancel rebuild
        if not skip_prompt:
            if volumetype == 'RAID-0':
                message = '''Once the RAID-0 rebuild process is started, all
                           data currently stored on {id} will be lost.'''.format(id=self.__volumeid)
            else:
                message = '''Once the volume rebuild process is started, the
                           process cannot be stopped.'''
            if not nvsm.prompt.Warn(
                    question='Start {type} rebuild on {id}?'.format(
                        type=volumetype,
                        id=self.__volumeid),
                    message=message):
                return None  # User canceled rebuild
        body = {'Drive': spare_drive, 'RaidType': volumetype}
        # Ask the API server to start volume rebuild
        self._context._queryBackend(
            self.__rebuildref, method='POST', body=body)

        # Return handle representing volume rebuild job progress
        job = VolumeRebuildJob(
            target=self,
            volumeid=self.__volumeid,
            volumetype=volumetype,
            progressref=self.__progressref)
        return job


class StorageVolumeTarget(Target):
    def __init__(self, volumeid, volumeref, rebuildref, esprebuildref, progressref, encryptref, address, context):
        super(StorageVolumeTarget, self).__init__(
            address=address, context=context)
        self.__volumeid = volumeid
        self.__volumeref = volumeref
        self.__rebuildref = rebuildref
        self.__progressref = progressref
        self.__encryptref = encryptref
        self.__esprebuildref = esprebuildref

        self.__properties = None
        self.__targets = None
        self.__volumeEncryptOpRefs = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        if self.__volumeref:
            self.__initVolumeProperties()
        else:
            # The volume is missing
            self.__initMissingVolumeProperties()

    def __initVolumeProperties(self):
        volumeinfo = self._context._queryBackend(self.__volumeref)
        # Squash hierarchies (SMASH does not support proprety hierarchies)

        def filter_properties(name, value):
            if name.lower() == '@odata.id':
                return False
            if 'actions' in name.lower():
                return False
            if name.lower() == 'schema':
                return False
            if name.lower() in ['status_state', 'status_health', 'name',
                                'encrypted', 'volumetype', 'links_oem_drives_@odata.id',
                                'capacitybytes', 'id']:
                return True
            return self._context.showall

        def map_properties(name, value):
            color = 'none'
            if 'status' in name.strip().lower():
                color = colorize_status(value)
            if name.strip().lower() == 'links_oem_drives_@odata.id':
                try:
                    drive_indices = value
                    drive_indices = [x.split('/')[-1] for x in drive_indices]

                    def to_int(s):
                        try:
                            result = int(s)
                        except:
                            result = s
                        return result
                    drive_indices.sort(key=to_int)
                    drive_indices = ', '.join(drive_indices)
                    drive_indices = '[ {0} ]'.format(drive_indices)
                    return ('Drives', drive_indices, color)
                except TypeError:
                    return (name, value, color)
            return (name, value, color)
        self.__properties.extend(list(dict_to_properties(
            target=self, d=volumeinfo,
            criteria=filter_properties,
            mapping=map_properties)))

    def __initMissingVolumeProperties(self):
        # API server has indicated that this volume is missing
        self.__properties.append(
            Property(
                target=self,
                name='Status_Health',
                value='Critical',
                color='red'))
        self.__properties.append(
            Property(
                target=self,
                name='Status_State',
                value='Missing',
                color='none'))

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}
        self.__initRebuildTarget()
        self.__initEspRebuildTarget()
        self.__initVolumeEncryptionTarget()

    def __initEspRebuildTarget(self):
        # Create CLI target representing the rebuild action for this volume
        if not self.__esprebuildref:
            return  # No encrypt action associated with this volume
        esp_rebuild = StorageVolumeEspRebuildTarget(
            esprebuildref=self.__esprebuildref,
            progressref=self.__progressref,
            context=self._context,
            address=self.address + 'rebuild')
        self.__targets['rebuild'] = esp_rebuild

    def __initRebuildTarget(self):
        if not self.__rebuildref:
            return  # No rebuild action associated with this volume
        # Create CLI target representing the rebuild action for this volume
        rebuildTarget = StorageVolumeRebuildTarget(
            volumeid=self.__volumeid,
            volumeref=self.__volumeref,
            rebuildref=self.__rebuildref,
            progressref=self.__progressref,
            address=self.address + 'rebuild',
            context=self._context)
        self.__targets['rebuild'] = rebuildTarget

    def __initVolumeEncryptionTarget(self):
        # Create CLI target representing the rebuild action for this volume
        if not self.__encryptref:
            return  # No encrypt action associated with this volume
        # Create CLI target representing the encrypt action for this volume
        encryptTarget = StorageVolumeEncryptTarget(
            volumeid=self.__volumeid,
            volumeref=self.__volumeref,
            encryptref=self.__encryptref,
            address=self.address + 'encryption',
            context=self._context)
        self.__targets['encryption'] = encryptTarget

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class StorageVolumesTarget(Target):
    def __init__(self, storageref, address, context):
        super(StorageVolumesTarget, self).__init__(
            address=address, context=context)
        assert(self.name == 'volumes')

        self.__storageref = storageref
        self.__targets = None
        self.__rebuildRefs = None
        self.__espRebuildRefs = None
        self.__progressRefs = None
        self.__encryptRefs = None
        self.__rebuildRef = None
        self.__rebuildProgressRef = None
        self.__encryptRef = None

    def __initTargets(self):
        assert self.__targets == None
        self.__targets = {}
        self.__initRebuildRefs()
        self.__initEspRebuildRefs()
        self.__initProgressRefs()
        self.__initEncryptRefs()
        self.__initRebuildTarget()
        self.__initEncryptTarget()
        self.__initVolumeTargets()

    def __initRebuildRefs(self):
        assert self.__rebuildRefs == None
        self.__rebuildRefs = {}
        # Enumerate all of the storage volumes that support rebuild
        volumesinfo = self._context._queryBackend(
            '{0}/Volumes'.format(self.__storageref))
        try:
            rebuild_targets = volumesinfo['Oem']['Actions']['#Volume.Rebuild']['target']
            for ref in rebuild_targets:
                # XXX: Fix for API path bug in nvsm-apis-18.10.4
                ref = re.sub(r'redfish', 'nvsm', ref)
                volumeid = ref.rsplit('/', 1)[-1]
                self.__rebuildRefs[volumeid] = ref
        except:
            pass

    def __initEspRebuildRefs(self):
        assert self.__espRebuildRefs == None
        self.__espRebuildRefs = {}
        volumesinfo = self._context._queryBackend(
            '{0}/Volumes'.format(self.__storageref))
        try:
            esp_rebuild_targets = volumesinfo['Oem']['Actions']['#Esp.Rebuild']['target']
            for ref in esp_rebuild_targets:
                ref = re.sub(r'redfish', 'nvsm', ref)
                volumeid = ref.rsplit('/', 1)[-1]
                self.__espRebuildRefs[volumeid] = ref
        except Exception as e:
            pass

    def __initEncryptRefs(self):
        assert self.__encryptRefs == None
        self.__encryptRefs = dict()
        encrypt_ops = dict()
        # Enumerate all of the storage volumes that support rebuild
        volumesinfo = self._context._queryBackend(
            '{0}/Volumes'.format(self.__storageref))
        try:
            def buildEncryptRefs(action, targets):
                for ref in targets:
                    volumeid = ref.rsplit('/', 1)[-1]
                    #encrypt_ops[action] = targets
                    if volumeid not in self.__encryptRefs:
                        self.__encryptRefs[volumeid] = {}
                    self.__encryptRefs[volumeid][action] = ref
                return
            buildEncryptRefs(
                'enable', volumesinfo['Oem']['Actions']['#Encrypt.Enable']['target'])
            buildEncryptRefs(
                'disable', volumesinfo['Oem']['Actions']['#Encrypt.Disable']['target'])
            buildEncryptRefs(
                'lock', volumesinfo['Oem']['Actions']['#Encrypt.Lock']['target'])
            buildEncryptRefs(
                'unlock', volumesinfo['Oem']['Actions']['#Encrypt.Unlock']['target'])
            buildEncryptRefs(
                'rotate', volumesinfo['Oem']['Actions']['#Encrypt.Rotate']['target'])
            buildEncryptRefs(
                'export', volumesinfo['Oem']['Actions']['#Encrypt.Export']['target'])
            buildEncryptRefs(
                'erase', volumesinfo['Oem']['Actions']['#Encrypt.Erase']['target'])
            buildEncryptRefs(
                'factoryreset', volumesinfo['Oem']['Actions']['#Encrypt.FactoryReset']['target'])
            buildEncryptRefs(
                'status', volumesinfo['Oem']['Actions']['#Encrypt.Status']['target'])

        except:
            pass

    def __initProgressRefs(self):
        assert self.__progressRefs == None
        self.__progressRefs = {}
        # Enumerate all of the storage volumes that support rebuild progress
        volumesinfo = self._context._queryBackend(
            '{0}/Volumes'.format(self.__storageref))
        try:
            progress_targets = volumesinfo['Oem']['Actions']['#Volume.Rebuild.Progress']['target']
            for ref in progress_targets:
                # XXX: Fix for API path bug in nvsm-apis-18.10.4
                ref = re.sub(r'redfish', 'nvsm', ref)
                volumeid = ref.rsplit('/', 1)[-1]
                self.__progressRefs[volumeid] = ref

            progress_targets = volumesinfo['Oem']['Actions']['#Esp.Rebuild.Progress']['target']
            for ref in progress_targets:
                # XXX: Fix for API path bug in nvsm-apis-18.10.4
                ref = re.sub(r'redfish', 'nvsm', ref)
                volumeid = ref.rsplit('/', 1)[-1]
                self.__progressRefs[volumeid] = ref
        except:
            pass

    def __initRebuildTarget(self):
        assert self.__rebuildRef == None
        assert self.__rebuildProgressRef == None
        # Enumerate all of the storage volumes that support rebuild progress
        volumesinfo = self._context._queryBackend(
            '{0}/Volumes'.format(self.__storageref))
        try:
            if volumesinfo.get('Oem', None) is not  None:
                rebuild_target = volumesinfo['Oem']['Actions']['#Rebuild']['target']
                for ref in rebuild_target:
                    # XXX: Fix for API path bug in nvsm-apis-18.10.4
                    ref = re.sub(r'redfish', 'nvsm', ref)
                    self.__rebuildRef = ref
                    break
                progress_target = volumesinfo['Oem']['Actions']['#Rebuild.Progress']['target']
                for ref in progress_target:
                    # XXX: Fix for API path bug in nvsm-apis-18.10.4
                    ref = re.sub(r'redfish', 'nvsm', ref)
                    self.__rebuildProgressRef = ref
                    break
        except Exception as e:
            print(e)
            pass

    def __initEncryptTarget(self):
        assert self.__encryptRef == None
        self.__encryptRef = {}
        # Enumerate all of the storage volumes that support rebuild progress
        volumesinfo = self._context._queryBackend(
                '{0}/Volumes'.format(self.__storageref))
        
        try:
            if volumesinfo.get('Oem', None) is not  None:
                disable_target = volumesinfo['Oem']['Actions']['#Encrypt.Disable']['target']
                for ref in disable_target:
                    ref = re.sub(r'redfish', 'nvsm', ref)
                    assert 'Encrypt.Disable' == ref.rsplit('/', 1)[-1]
                    self.__encryptRef['disable'] = ref

                unlock_target = volumesinfo['Oem']['Actions']['#Encrypt.Unlock']['target']
                for ref in unlock_target:
                    ref = re.sub(r'redfish', 'nvsm', ref)
                    self.__encryptRef['unlock'] = ref

                disable_target = volumesinfo['Oem']['Actions']['#Encrypt.FactoryReset']['target']
                for ref in disable_target:
                    ref = re.sub(r'redfish', 'nvsm', ref)
                    self.__encryptRef['factoryreset'] = ref

                status_target = volumesinfo['Oem']['Actions']['#Encrypt.Status']['target']
                for ref in status_target:
                    ref = re.sub(r'redfish', 'nvsm', ref)
                    self.__encryptRef['status'] = ref
            
        except Exception as e:
            print(e)
            pass

    def __initVolumeTargets(self):
        # Enumerate storage volumes with NVSM backend
        volumesinfo = self._context._queryBackend(
            '{0}/Volumes'.format(self.__storageref))
        if not hasattr(volumesinfo, 'get'):
            volumesinfo = {}
        members = volumesinfo.get("Members", [])
        if members == None:
            members = []
        for volumeref in members:
            volumeref = volumeref.get('@odata.id', None)
            if not volumeref:
                continue
            volumeid = volumeref.split('/')[-1]
            encryptref = self.__encryptRefs.get(volumeid)
            # Iterate over the keys in dictionary, access value & print line by line
            if encryptref is not None:
                for key in encryptref:
                    nvsm.log.info('Key:{}  Value:{}'.format(
                        key, encryptref[key]))
            volume = StorageVolumeTarget(
                volumeid=volumeid,
                volumeref=volumeref,
                rebuildref=self.__rebuildRefs.get(volumeid),
                esprebuildref=self.__espRebuildRefs.get(volumeid),
                progressref=self.__progressRefs.get(volumeid),
                encryptref=self.__encryptRefs.get(volumeid),
                address=(self.address + volumeid),
                context=self._context)
            self.__targets[volumeid] = volume
        # Create "empty" volume targets for missing volumes
        for volumeid in list(self.__rebuildRefs.keys()):
            if volumeid in self.__targets:
                continue
            volume = StorageVolumeTarget(
                volumeid=volumeid,
                volumeref=None,  # Missing volume
                rebuildref=self.__rebuildRefs.get(volumeid),
                esprebuildref=self.__espRebuildRefs.get(volumeid),
                progressref=self.__progressRefs.get(volumeid),
                encryptref=self.__encryptRefs.get(volumeid),
                address=(self.address + volumeid),
                context=self._context)
            self.__targets[volumeid] = volume

        rebuildTarget = StorageVolumeRebuildTarget(
            volumeid=None,
            volumeref=None,  # Missing volume
            rebuildref=self.__rebuildRef,
            progressref=self.__rebuildProgressRef,
            address=self.address + 'rebuild',
            context=self._context)
        self.__targets['rebuild'] = rebuildTarget

        if len(self.__encryptRef) > 0:
            encryptTarget = StorageVolumeEncryptTarget(
                volumeid=None,
                volumeref=None,  # Missing volume
                encryptref=self.__encryptRef,
                address=self.address + 'encryption',
                context=self._context)
            self.__targets['encryption'] = encryptTarget

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?


class StorageTarget(Target):
    def __init__(self, storageref, address, context):
        super(StorageTarget, self).__init__(
            address=address, context=context)
        self.__storageref = storageref
        self.__properties = None
        self.__targets = {}

        # Assume every system has storage alert targets (for DSHM alerts)
        alertsTarget = AlertsTarget(
            ref='/nvsm/v1/Alerts/Storage',  # XXX
            address=(self.address + 'alerts'),
            context=self._context)
        self.__targets['alerts'] = alertsTarget

        # Assume every system has storage policy targets (for configuring DSHM)
        policyTarget = StoragePolicyTarget(
            # FIXME: NVSM backend does not provide OData reference here
            policyref='/nvsm/v1/Policy/Storage',
            address=(self.address + 'policy'),
            context=self._context)
        self.__targets['policy'] = policyTarget

        # Provide target references to both drives and volumes
        drivesTarget = StorageDrivesTarget(
            storageref=self.__storageref,
            address=(self.address + 'drives'),
            context=self._context)
        self.__targets['drives'] = drivesTarget

        volumesTarget = StorageVolumesTarget(
            storageref=self.__storageref,
            address=(self.address + 'volumes'),
            context=self._context)
        self.__targets['volumes'] = volumesTarget

        # FIXME there may be better approach
        controllersinfo = self._context._queryBackend(self.__storageref)
        if controllersinfo['StorageControllers@odata.count']:
            controllersTarget = StorageControllersTarget(
                storageref=self.__storageref,
                address=(self.address + 'controllers'),
                context=self._context)
            self.__targets['controllers'] = controllersTarget

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        storageinfo = self._context._queryBackend(self.__storageref)
        if not hasattr(storageinfo, 'get'):
            storageinfo = {}
        # Count the drives to get "DriveCount" property
        drives = storageinfo.get('Drives', None)
        if not drives:
            driveCount = 0
        else:
            driveCount = len(drives)
        p = Property(
            target=self,
            name='DriveCount',
            value=driveCount)
        self.__properties.append(p)
        # Enumerate the storage volumes to get "Volumes" property
        volumeNames = []
        try:
            # XXX: In the future, do not construct OData object here
            import nvsm.odata
            storageinfo = nvsm.odata.from_value(
                storageinfo, context=self._context)
            volumes = storageinfo['Volumes']['Members']
            for v in volumes:
                volumeNames.append(v['Name'])
        except:
            nvsm.log.warning('Could not enumerate storage volumes')
        else:
            p = Property(
                target=self,
                name='Volumes',
                value='[ {} ]'.format(', '.join(volumeNames)))
            self.__properties.append(p)

    def __getitem__(self, key):
        return self.__targets[key]

    @property
    def targets(self):
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class StorageCollectionTarget(Target):
    def __init__(self, systemref, address, context):
        super(StorageCollectionTarget, self).__init__(
            address=address, context=context)
        self.__systemref = systemref
        self.__targets = None

    def __initTargets(self):
        # Enumerate storage subsystems with NVSM backend
        self.__targets = {}
        storageinfo = {}
        try:
            storageinfo = self._context._queryBackend(
                '{0}/Storage'.format(self.__systemref))
        except:
            pass
        members = storageinfo.get('Members', [])
        if members == None:
            members = []
        for storageref in members:
            storageref = storageref.get('@odata.id', None)
            if not storageref:
                continue
            index = storageref.split('/')[-1]
            storage = StorageTarget(
                storageref=storageref,
                address=(self.address + 'storage'),
                context=self._context)
            # self.__targets[index] = storage
            self.__targets['storage'] = storage

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?
