# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



from nvsm.backend.context import Target, dict_to_properties

from nvsm.backend.firmware import DockerTarget
from nvsm.backend.gpu import GPUCollectionTarget
from nvsm.backend.memory import MemoryCollectionTarget
from nvsm.backend.processor import ProcessorCollectionTarget
from nvsm.backend.storage import StorageCollectionTarget
from nvsm.backend.pcie import PCIeCollectionTarget
from nvsm.backend.networkinterfaces import NetworkInterfaceCollectionTarget
from nvsm.backend.nvswitch import NVSwitchCollectionTarget
from nvsm.backend.health import HealthTarget
from nvsm.backend.alerts import AlertsTarget
from nvsm.backend.rds import RdsDockerTarget


class SystemTarget(Target):
    def __init__(self, systemref, address, context):
        super(SystemTarget, self).__init__(
            address=address, context=context)
        self.__systemref = systemref
        self.__properties = None
        self.__targets = None

    def __initProperties(self):
        assert self.__properties == None
        self.__properties = []
        systeminfo = self._context._queryBackend(self.__systemref)
        properties = ['hostname', 'name', 'manufacturer', 'serialnumber',
                      'uuid', 'partnumber', 'assettag', 'biosversion',
                      'status_state', 'status_health']

        def filter_properties(name, value):
            if 'odata' in name.lower():
                return False
            if str(value).lower() == 'n/a':
                return self._context.showall
            if name.lower() in properties:
                return True
            return self._context.showall
        self.__properties = list(dict_to_properties(
                    target=self, d=systeminfo, criteria=filter_properties))
        def compare_properties(a):
            try:
                i = properties.index(a.name.lower())
            except ValueError:
                i = len(properties)
            return i
        self.__properties.sort(key=compare_properties)

    def __initTargets(self):
        self.__targets = {}
        alertsTarget = AlertsTarget(
            ref='/nvsm/v1/Alerts/Systems',  # XXX
            address=(self.address + 'alerts'),
            context=self._context)
        self.__targets['alerts'] = alertsTarget
        dcsHwaApiKeyTarget = DockerTarget(
            systemref=self.__systemref,
            address=(self.address + 'dcshwapikey'),
            context=self._context,
            ref='nvsm/v1/Systems/1/Dcshwapikey',
            showproperties=['expiry'])
        self.__targets['dcshwapikey'] = dcsHwaApiKeyTarget
        # Assume every system has storage, GPUs, memory, processors, firmware,
        # and pcie
        systeminfo  = {}
        try:
            systeminfo = self._context._queryBackend(self.__systemref)
        except:
            pass
        if systeminfo.get("NetworkInterfaces", None) is not None:
            if systeminfo.get("NetworkInterfaces", None).get("@odata.id" , None) is not None:
                networkTarget = NetworkInterfaceCollectionTarget(
                    nw_intf_collection_ref=self.__systemref + '/NetworkInterfaces',
                    address=(self.address + 'network_interfaces'),
                    context=self._context)
                self.__targets['network_interfaces'] = networkTarget

        if systeminfo.get("Storage", None) is not None:
            if systeminfo.get("Storage", None).get("@odata.id" , None) is not None:
                storageTarget = StorageCollectionTarget(
                    systemref=self.__systemref,
                    address=(self.address),
                    context=self._context)
                targets = []
                targets.extend(storageTarget.targets)
                for t in targets:
                    self.__targets[t.name] = t
        if systeminfo.get("Memory", None) is not None:
            if systeminfo.get("Memory", None).get("@odata.id" , None) is not None:
                memoryTarget = MemoryCollectionTarget(
                    systemref=self.__systemref,
                    address=(self.address + 'memory'),
                    context=self._context)
                self.__targets['memory'] = memoryTarget

        if systeminfo.get("Processors", None) is not None:
            if systeminfo.get("Processors", None).get("@odata.id" , None) is not None:
                processorsTarget = ProcessorCollectionTarget(
                    systemref=self.__systemref,
                    address=(self.address + 'processors'),
                    context=self._context)
                self.__targets['processors'] = processorsTarget

        try:
            self._context._queryBackend('/nvsm/v1/Systems/1/GPUs')
            gpusTarget = GPUCollectionTarget(
                systemref=self.__systemref,
                address=(self.address + 'gpus'),
                context=self._context)
            self.__targets['gpus'] = gpusTarget
        except:
            pass
        try:
            firmwareTarget = DockerTarget(
                systemref=self.__systemref,
                address=(self.address + 'firmware'),
                context=self._context)
            self.__targets['firmware'] = firmwareTarget
        except:
            pass
        try:
            rdsTarget = RdsDockerTarget(
                ref=self.__systemref,
                address=(self.address + 'collectkdblog'),
                context=self._context)
            self.__targets['collectkdblog'] = rdsTarget
        except:
            pass
        try:
            self._context._queryBackend('/nvsm/v1/Systems/1/PCIeDevices')
            pcieTarget = PCIeCollectionTarget(
                systemref=self.__systemref,
                address=(self.address + 'pcie'),
                context=self._context)
            self.__targets['pcie'] = pcieTarget
        except:
            pass
        try:
            self._context._queryBackend('/nvsm/v1/Systems/1/NVSwitches')
            nvswitchTarget = NVSwitchCollectionTarget(
                systemref=self.__systemref,
                address=(self.address + 'nvswitches'),
                context=self._context)
            self.__targets['nvswitches'] = nvswitchTarget
        except:
           pass
        try:
            self._context._queryBackend('/nvsm/v1/Systems/1/Health')
            healthTarget = HealthTarget(
                systemref=self.__systemref,
                address=(self.address + 'health'),
                context=self._context)
            self.__targets['health'] = healthTarget
        except:
            pass

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]


    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        if self.__properties == None:
            # Initialize properties on first access
            self.__initProperties()
        return self.__properties


class SystemCollectionTarget(Target):
    def __init__(self, address, context):
        super(SystemCollectionTarget, self).__init__(
            address=address, context=context)
        self.__targets = None

    def __initTargets(self):
        # Enumerate systems with
        self.__targets = {}

        systemsinfo = {}
        try:
            systemsinfo = self._context._queryBackend('/redfish/v1/Systems')
        except:
            pass
        members = systemsinfo.get('Members', [])
        if members == None:
            members = []
        for systemref in members:
            systemref = systemref.get('@odata.id', None)
            if not systemref:
                continue
            index = systemref.split('/')[-1]
            # import socket
            # system_host = socket.gethostname()
            if index == '1':
                system_host = 'localhost'
            else:
                system_host = index#'localhost'
            system = SystemTarget(
                systemref=systemref,
                # address=(self.address + index),
                address=(self.address + system_host),
                context=self._context)
            # self.__targets[index] = system
            self.__targets[system_host] = system

    def __getitem__(self, key):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return self.__targets[key]

    @property
    def targets(self):
        if self.__targets == None:
            # Initialize targets upon first access
            self.__initTargets()
        return list(self.__targets.values())

    @property
    def properties(self):
        return []  # TODO?
