# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.


import nvsm.exceptions


class CommaSeparatedList(list):
    def __init__(self, string):
        l = string.split(',')
        super(CommaSeparatedList, self).__init__(l)


class ColonSeparatedList(list):
    def __init__(self, string):
        l = string.split(':')
        super(ColonSeparatedList, self).__init__(l)


class Command(object):
    # Initialize the NVSM Command factory pattern
    __commands = []
    __commandAliases = []

    @classmethod
    def register_command(cls, subclass):
        """Register a subclass for the NVSM Command factory pattern.

        This class method allows classes that implement an NVSM command to
        register with the NVSM Command factory pattern. Once registered,
        subclasses can be constructed with the Command.from_command_node()
        factory method.

        Subclasses registered with the NVSM Command factory pattern must
        implement the "implements()" static method, which takes a command name
        as input and returns true if and only if the subclass implements the
        given command name.
        """
        assert(subclass not in cls.__commands)
        assert(subclass not in cls.__commandAliases)
        cls.__commands.append(subclass)

    @classmethod
    def register_command_alias(cls, subclass):
        """Register a command alias for the NVSM Command factory pattern.

        Command aliases take precedence over ordinary commands. Other than
        precedence, command aliases are functionally equivalent to ordinary
        commands.

        See the documentation for the Command.register_command() class method
        for a description of the NVSM Command factory pattern.
        """
        assert(subclass not in cls.__commands)
        assert(subclass not in cls.__commandAliases)
        cls.__commandAliases.append(subclass)

    @classmethod
    def list_aliases(cls):
        # XXX
        return [
            'show alerts',
            'show drives',
            'show versions',
            'show fans',
            'show firmware',
            'update firmware',
            'show gpus',
            'dump health',
            'show health',
            'show memory',
            'show networkadapters',
            'show networkdevicefunctions',
            'show networkinterfaces',
            'show networkports',
            'show nvswitches',
            'show policy',
            'show power',
            'show processors',
            'show storage',
            'show temperature',
            'show volumes',
            'show powermode',
            'show led',
            'enable dashboards',
            'enable exporters',
            'disable exporters'
        ]

    @classmethod
    def from_command_node(cls, command_node, working_target='/'):
        """Factory method to construct a Command object from CommandNode.

        Given a CommandNode object from the parse tree, the class implementing
        the command is not immediately known. This factory method searches
        through the list of known commands and constructs the appropriate
        Command object.

        If the command specified in the command node is not known, an
        nvsm.exceptions.UnknownCommand exception is thrown.
        """
        # Determine which subclass implements this command
        implementing_class = None
        if working_target == '/':  # XXX: Only evaluate command aliases at
            #      root of target hierarchy
            for subclass in cls.__commandAliases:
                if subclass.implements(command_node.argv):
                    # Programming error; only one command alias should exist
                    assert(not implementing_class)
                    implementing_class = subclass
        if not implementing_class:
            for subclass in cls.__commands:
                if subclass.implements(command_node.argv):
                    # Programming error; only one command implementation should exist
                    assert(not implementing_class)
                    implementing_class = subclass
        # Check for unknown commands
        if not implementing_class:
            raise nvsm.exceptions.UnknownCommand(
                'Unknown command "{0}"'.format(command_node.name))
        # Finally construct the command object
        command = implementing_class(argv=command_node.argv)
        return command

    def __init__(self, argv):
        self.__argv = argv

    @property
    def argv(self):
        return self.__argv

    @property
    def name(self):
        return self.__argv[0]

    def run(self, context=None, file=None):
        """ Abstract method to run the command.

        Implementing classes must implement the run() method.
        """
        raise NotImplementedError
