# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.




import argparse
import sys

import nvsm.command
import nvsm.log

class CdCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(CdCommand, self).__init__(argv=argv)

        self.__show_help = False
        self.__parse_arguments()

    def __parse_arguments(self):
        self.__argparser = argparse.ArgumentParser(
                prog='cd',
                description='''Change the working target address used by NVSM verbs''',
                add_help=False)
        self.__argparser.add_argument('-help', '-h', action='store_true', default=False,
                help='show this help message and exit')
        self.__argparser.add_argument('target', nargs='?', action='store', default='/',
                help='''The desired working target address. The given target
                address can be relative to the current working target, or it
                can be absolute.''')
        try:
            # TODO: Redirect the sys.stdout printed by the parser (for unit testing purposes)
            self.options = self.__argparser.parse_args(self.argv[1:])
        except SystemExit as e:
            if e.code != 0:
                # TODO: Somehow raise exception with the error message printed by argparse
                raise nvsm.exceptions.InvalidArguments(
                        'Invalid cd command arguments: {0}'.format(
                            ' '.join(self.argv[1:])))

    def run(self, context, file=sys.stdout):
        if self.options.help:
            self.__argparser.print_help(file=file)
            return
        nvsm.log.debug(
                'Running cd command...\n' \
                '    name: {name}\n' \
                '    argv: {argv}\n' \
                '    options: {options}',
                name=self.name, argv=self.argv, options=self.options)

        # Attempt to change the current working target address
        desired_target = nvsm.targetAddress.TargetAddress(
                self.options.target, context=context)
        desired_target = desired_target.canonicalize()
        try:
            context.workingTargetAddress = desired_target
        except nvsm.exceptions.MissingTarget as e:
            nvsm.log.error('{0}', str(e))
            return
        nvsm.log.debug('Successfully changed working target to "{0}"',
                    str(desired_target))

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an NVSM "cd" command.

        This method allows the NVSM Command factory pattern to determine if
        the ShowCommand class implements the given command name.

        This method returns True if argv[0] is "cd", case insensitive.
        Returns False otherwise.
        """
        if argv[0].lower() == 'cd':
            return True
        return False

# Register with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command(CdCommand)
