# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.




import os
import signal
import subprocess
import sys

import nvsm.command
import nvsm.log
import logging

import argparse

class DumpHealthCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(DumpHealthCommand, self).__init__(argv=argv)

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an alias for "dump health".

        This method allows the NVSM Command factory pattern to determine if
        the "DumpHealthCommand" class implements the given command.

        This method returns True if the first two arguments of argv look like
        "dump health" case insensitive. Returns False otherwise.
        """
        aliases = [
            [ 'dump', 'health' ],
            [ 'dump', 'nvsysinfo' ] ]
        return [x.lower() for x in argv[:2]] in aliases

    def check_if_platform_support_legacy_dump_only(self):
        platform_supporting_legacy_dump = ["DGX Station"] #FIXME: Is this also disabling dump health for red october
        cmdstr = "dmidecode -s system-product-name | grep -x '{}'"
        for platform in platform_supporting_legacy_dump:
            dmidecode = subprocess.Popen(cmdstr.format(platform), stdout=subprocess.PIPE,
                          stderr=subprocess.STDOUT, shell=True)
            (cmdout, cmderr) = dmidecode.communicate()
            if dmidecode.returncode:
                continue
            else:
                return True
        return False



    def run(self, context, file=sys.stdout):
        # Suppress deprecation warning messages from nvsysinfo
        import copy
        env = copy.copy(os.environ)
        env['CALLED_FROM_NVSM_CLI'] = 'true'

        # Pass remaining arguments unmodified to nvsysinfo
        args = [ 'nvsysinfo' ]
        args.extend(self.argv[2:])

        parser = argparse.ArgumentParser(
                description="NVIDIA System Information",
                usage='nvsm dump health [help] [-o OUTPUT] [-t, --tags "tag1,tag2"] [-tfp, --tar_file_path "/x/y/path"] [-tfn, --tar_file_name "name.tar.xz"]',
                #usage='nvsm dump health [help] [-o OUTPUT] [--log-level LEVEL] [-st, --sosreport_timeout TIMEOUT IN MINS] [-tt, --task_timeout TIMEOUT IN MINS]',
                add_help=False
                )
        parser.add_argument('help',
                default='*',
                nargs='?',
                help='show this help message and exit')
        parser.add_argument('-o', '--output', nargs='?',
                help='File path in which to write output.')
        parser.add_argument('-st', '--sosreport_timeout', nargs='?',
                help='sosreport timeout in mins. Default=12mins')
        parser.add_argument('-tt', '--task_timeout', nargs='?',
                help='Individual task timeout in mins. \
                    Default is as set with RunCommand, if unspecified it\'s 5mins.')
        parser.add_argument('--log-level', nargs='?',
                help='Set the output logging level.',
                choices=[ 'debug', 'info', 'warning', 'error', 'critical' ])
        parser.add_argument('-t', '--tags', nargs='?', 
                help='Select the operations to be executed through dump health using the comma seperated operation tags')
        parser.add_argument('-tfp', '--tar_file_path', nargs='?', 
                help='File path in which to write output without file name')
        parser.add_argument('-tfn', '--tar_file_name', nargs='?', 
                help='File name of tar file')
        parser.add_argument('-lgd','--legacy_dump',
                action='store_true',
                help='Run old nvsm dump health.')

        args, _ = parser.parse_known_args(self.argv[2:])
        
        if args.help != '*':
            parser.print_help()
            return

        tar_file_path = None
        tar_file_name = None
        if args.tar_file_path:
            tar_file_path = args.tar_file_path
        if args.tar_file_name:
            tar_file_name = args.tar_file_name

        out_file = None
        if args.output:
            out_file = args.output

        if args.legacy_dump and args.tags:
            logging.error("--legacy_dump and --tags option can not be used together")
            return
        if self.check_if_platform_support_legacy_dump_only() and args.tags:
            logging.error("--tags can not be used")
            return

        run_legacy_dump = False
        if args.legacy_dump or self.check_if_platform_support_legacy_dump_only():
            run_legacy_dump = True

        if run_legacy_dump:
            # If it is legacy dump health, then tags, tarball file path and file name are not supported
            if args.tags:
                logging.warning("-t/--tags option is ignored")
            if args.tar_file_path:
                logging.warning("-tfp/--tar_file_path option is ignored")
            if args.tar_file_name:
                logging.warning("-tfn/--tar_file_name option is ignored")
        else:
            # If it is new dump health, then sosreport timeout, loglevel and task timeout are not supported
            if args.sosreport_timeout:
                logging.warning("-st/--sosreport_timeout option is ignored")
            if args.log_level:
                logging.warning("--log-level option is ignored")
            if args.task_timeout:
                logging.warning("-tt/--task-timeout optoin is ignored")

        if run_legacy_dump:
            from nvsm.backend.health import run_dump_health
            run_dump_health(out_file, logging.getLogger().getEffectiveLevel(), sosreport_timeout=args.sosreport_timeout, task_timeout=args.task_timeout)
        else:
            #Handle new nvsm health code
            from nvsm.backend.health import run_new_dump_health
            run_new_dump_health(tar_file_name, tar_file_path, args.tags)




# Register as a command alias with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command_alias(DumpHealthCommand)
