# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.




import argparse
import datetime
import json
import re
import signal
import sys
import time
from contextlib import contextmanager

from nvsm.color import colorize
import nvsm.command
import nvsm.log


@contextmanager
def ignore_keyboard_interrupts(uninterruptible=True):
    if not uninterruptible:
        yield None
    else:
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        try:
            yield None
        finally:
            # Restore default signal handler
            signal.signal(signal.SIGINT, signal.SIG_DFL)



class LockCommand(nvsm.command.Command):

    __nameValuePattern = re.compile(r'^(?P<name>[^=]+)=(?P<value>.*)$')
    def __init__(self, argv):
        super(LockCommand, self).__init__(argv=argv)

        self.__parse_arguments()

    def __parse_arguments(self):
        self.__argparser = argparse.ArgumentParser(
                prog='lock',
                description='''Enable locking of SED''',
                add_help=False)
        self.__argparser.add_argument('-help', '-h', action='store_true', default=False,
                help='show this help message and exit')
        self.__argparser.add_argument('key', nargs='?', action='store', default='.',
                help=''' Key used to encrypt the drive(s).
                        To be used when encryption key is wrong/missing.
                        Example: lock key=xxxxxxxx''')        
        try:
            # TODO: Redirect the sys.stdout printed by the parser (for unit testing purposes)
            self.options, self.__remaining = self.__argparser.parse_known_args(
                self.argv[1:])
        except SystemExit as e:
            if e.code != 0:
                # TODO: Somehow raise exception with the error message printed by argparse
                raise nvsm.exceptions.InvalidArguments(
                        'Invalid lock command arguments: {0}'.format(
                            ' '.join(self.argv[1:])))
        if self.options.help:
            return  # User requested help; nothing to parse

        # Treat "target" with equals sign as propertyname=value pair
        if '=' in self.options.key:
            self.__remaining.insert(0, self.options.key)
            self.options.key = '.'

        # Parse remaining argparse options as propertyname=value pairs
        self.__nameValuePairs = []

        for arg in self.__remaining:
            m = LockCommand.__nameValuePattern.match(arg)
            if not m:
                raise nvsm.exceptions.InvalidArguments(
                    'Expected propertyname=value pair: {0}'.format(arg))
            pair = (m.group('name'), m.group('value'))
            self.__nameValuePairs.append(pair)

    def run(self, context, file=sys.stdout):
        if self.options.help:
            self.__argparser.print_help(file=file)
            return
        nvsm.log.debug(
                'Running lock command...\n' \
                '    name: {name}\n' \
                '    argv: {argv}\n' \
                '    options: {options}',
                name=self.name, argv=self.argv, options=self.options)

        # Traverse the target hierarchy to find matching target
        address = nvsm.targetAddress.TargetAddress(
                self.options.key, context=context)
        target = context.getTarget(address)
        # Make sure target supports "start" verb
        if not hasattr(target, 'lock'):
            raise nvsm.exceptions.CommandError(
                'Target "{target}" does not support "lock" verb'.format(
                target=target.address))
        # Start the job associated with this target
        job = target.lock(skip_prompt=False, wait=False, keyvalues=self.__nameValuePairs)
        if not job:
            return
        print(file=file)


    @staticmethod
    def implements(argv):
        """ Determine if the given command is an NVSM "lock" command.

        This method allows the NVSM Command factory pattern to determine if
        the SetCommand class implements the given command name.

        This method returns True if argv[0] is "lock", case insensitive.
        Returns False otherwise.
        """
        if argv[0].lower() == 'lock':
            return True
        return False

# Register with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command(LockCommand)


class EnableCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(EnableCommand, self).__init__(argv=argv)

        self.__parse_arguments()

    def __parse_arguments(self):
        self.__argparser = argparse.ArgumentParser(
                prog='enable',
                description='''Enable property of target''',
                add_help=False)
        self.__argparser.add_argument('-help', '-h', action='store_true', default=False,
                help='show this help message and exit')
        self.__argparser.add_argument('target', nargs='?', action='store', default='.',
                help='''The target address of the Managed Element . The
                        target address can be relative to the current working
                        target, or it can be absolute. Globbing of targets is
                        not supported for this command.''')
        try:
            # TODO: Redirect the sys.stdout printed by the parser (for unit testing purposes)
            self.options = self.__argparser.parse_args(self.argv[1:])
        except SystemExit as e:
            if e.code != 0:
                # TODO: Somehow raise exception with the error message printed by argparse
                raise nvsm.exceptions.InvalidArguments(
                        'Invalid enable command arguments: {0}'.format(
                            ' '.join(self.argv[1:])))

    def run(self, context, file=sys.stdout):
        if self.options.help:
            self.__argparser.print_help(file=file)
            return
        nvsm.log.debug(
                'Running enable command...\n' \
                '    name: {name}\n' \
                '    argv: {argv}\n' \
                '    options: {options}',
                name=self.name, argv=self.argv, options=self.options)

        # Traverse the target hierarchy to find matching target
        address = nvsm.targetAddress.TargetAddress(
                self.options.target, context=context)
        target = context.getTarget(address)
        # Make sure target supports "start" verb
        if not hasattr(target, 'enable'):
            raise nvsm.exceptions.CommandError(
                'Target "{target}" does not support "enable" verb'.format(
                target=target.address))
        # Start the job associated with this target
        job = target.enable(skip_prompt=False, wait=False, argv=self.argv)
        if not job:
            return
        print(file=file)


    @staticmethod
    def implements(argv):
        """ Determine if the given command is an NVSM "lock" command.

        This method allows the NVSM Command factory pattern to determine if
        the SetCommand class implements the given command name.

        This method returns True if argv[0] is "lock", case insensitive.
        Returns False otherwise.
        """
        if argv[0].lower() == 'enable':
            return True
        return False

# Register with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command(EnableCommand)

class UnlockCommand(nvsm.command.Command):
    __nameValuePattern = re.compile(r'^(?P<name>[^=]+)=(?P<value>.*)$')
    def __init__(self, argv):
        super(UnlockCommand, self).__init__(argv=argv)
        self.__parse_arguments()

    def __parse_arguments(self):
        self.__argparser = argparse.ArgumentParser(
                prog='unlock',
                description='''unlock sed drives''',
                add_help=False)
        self.__argparser.add_argument('-help', '-h', action='store_true', default=False,
                help='show this help message and exit')
        self.__argparser.add_argument('key', nargs='?', action='store', default='.',
                help=''' Key used to encrypt the drive(s).
                        To be used when encryption key is wrong/missing.
                        Example: unlock key=xxxxxxxx''')

        try:
            # TODO: Redirect the sys.stdout printed by the parser (for unit testing purposes)
            self.options, self.__remaining = self.__argparser.parse_known_args(
                self.argv[1:])
        except SystemExit as e:
            if e.code != 0:
                # TODO: Somehow raise exception with the error message printed by argparse
                raise nvsm.exceptions.InvalidArguments(
                        'Invalid unlock command arguments: {0}'.format(
                        ' '.join(self.argv[1:])))
        if self.options.help:
            return  # User requested help; nothing to parse

        # Treat "target" with equals sign as propertyname=value pair
        if '=' in self.options.key:
            self.__remaining.insert(0, self.options.key)
            self.options.key = '.'

        # Parse remaining argparse options as propertyname=value pairs
        self.__nameValuePairs = []
        for arg in self.__remaining:
            m = UnlockCommand.__nameValuePattern.match(arg)
            if not m:
                raise nvsm.exceptions.InvalidArguments(
                    'Expected propertyname=value pair: {0}'.format(arg))
            pair = (m.group('name'), m.group('value'))
            self.__nameValuePairs.append(pair)
                

    def run(self, context, file=sys.stdout):
        if self.options.help:
            self.__argparser.print_help(file=file)
            return
        nvsm.log.debug(
                'Running lock command...\n' \
                '    name: {name}\n' \
                '    argv: {argv}\n' \
                '    options: {options}',
                name=self.name, argv=self.argv, options=None)

        # Traverse the target hierarchy to find matching target
        address = nvsm.targetAddress.TargetAddress(
                self.options.key, context=context)
        target = context.getTarget(address)
        # Make sure target supports "start" verb
        if not hasattr(target, 'unlock'):
            raise nvsm.exceptions.CommandError(
                'Target "{target}" does not support "unlock" verb'.format(
                target=target.address))
        # Start the job associated with this target
        job = target.unlock(skip_prompt=False, wait=False, keyvalues=self.__nameValuePairs)
        if not job:
            return
        print(file=file)

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an NVSM "lock" command.

        This method allows the NVSM Command factory pattern to determine if
        the SetCommand class implements the given command name.

        This method returns True if argv[0] is "lock", case insensitive.
        Returns False otherwise.
        """
        if argv[0].lower() == 'unlock':
            return True
        return False

nvsm.command.Command.register_command(UnlockCommand)

class EraseCommand(nvsm.command.Command):
    __nameValuePattern = re.compile(r'^(?P<name>[^=]+)=(?P<value>.*)$')
    def __init__(self, argv):
        super(EraseCommand, self).__init__(argv=argv)
        self.__parse_arguments()

    def __parse_arguments(self):
        self.__argparser = argparse.ArgumentParser(
                prog='erase',
                description='''erase property''',
                add_help=False)
        self.__argparser.add_argument('-help', '-h', action='store_true', default=False,
                help='show this help message and exit')
        self.__argparser.add_argument('key', nargs='?', action='store', default='.',
                help=''' Key used to encrypt the drive(s).
                        To be used when encryption key is wrong/missing.
                        Example: erase key=xxxxxxxx''')
        try:
            # TODO: Redirect the sys.stdout printed by the parser (for unit testing purposes)
            self.options, self.__remaining = self.__argparser.parse_known_args(
                self.argv[1:])
        except SystemExit as e:
            if e.code != 0:
                # TODO: Somehow raise exception with the error message printed by argparse
                raise nvsm.exceptions.InvalidArguments(
                        'Invalid erase command arguments: {0}'.format(
                            ' '.join(self.argv[1:])))
        if self.options.help:
            return  # User requested help; nothing to parse

        # Treat "target" with equals sign as propertyname=value pair
        if '=' in self.options.key:
            self.__remaining.insert(0, self.options.key)
            self.options.key = '.'

        # Parse remaining argparse options as propertyname=value pairs
        self.__nameValuePairs = []

        for arg in self.__remaining:
            m = EraseCommand.__nameValuePattern.match(arg)
            if not m:
                raise nvsm.exceptions.InvalidArguments(
                    'Expected propertyname=value pair: {0}'.format(arg))
            pair = (m.group('name'), m.group('value'))
            self.__nameValuePairs.append(pair)

    def run(self, context, file=sys.stdout):
        if self.options.help:
            self.__argparser.print_help(file=file)
            return
        nvsm.log.debug(
                'Running erase command...\n' \
                '    name: {name}\n' \
                '    argv: {argv}\n' \
                '    options: {options}',
                name=self.name, argv=self.argv, options=None)

        # Traverse the target hierarchy to find matching target
        address = nvsm.targetAddress.TargetAddress(
                self.options.key, context=context)
        target = context.getTarget(address)
        # Make sure target supports "erase" verb
        if not hasattr(target, 'erase'):
            raise nvsm.exceptions.CommandError(
                'Target "{target}" does not support "erase" verb'.format(
                target=target.address))
        # Start the job associated with this target
        job = target.erase(skip_prompt=False, wait=False, keyvalues=self.__nameValuePairs)
        if not job:
            return
        print(file=file)

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an NVSM "lock" command.

        This method allows the NVSM Command factory pattern to determine if
        the SetCommand class implements the given command name.

        This method returns True if argv[0] is "lock", case insensitive.
        Returns False otherwise.
        """
        if argv[0].lower() == 'erase':
            return True
        return False

# Register with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command(EraseCommand)

class DisableCommand(nvsm.command.Command):
    __nameValuePattern = re.compile(r'^(?P<name>[^=]+)=(?P<value>.*)$')
    def __init__(self, argv):
        super(DisableCommand, self).__init__(argv=argv)
        self.__parse_arguments()

    def __parse_arguments(self):
        self.__argparser = argparse.ArgumentParser(
                prog='disable',
                description='''disable property''',
                add_help=False)
        self.__argparser.add_argument('-help', '-h', action='store_true', default=False,
                help='show this help message and exit')
        self.__argparser.add_argument('key', nargs='?', action='store', default='.',
                help=''' Key used to encrypt the drive(s).
                        To be used when encryption key is wrong/missing.
                        Example: disable key=xxxxxxxx''')
        try:
            # TODO: Redirect the sys.stdout printed by the parser (for unit testing purposes)
            self.options, self.__remaining = self.__argparser.parse_known_args(
                self.argv[1:])
        except SystemExit as e:
            if e.code != 0:
                # TODO: Somehow raise exception with the error message printed by argparse
                raise nvsm.exceptions.InvalidArguments(
                        'Invalid unlock command arguments: {0}'.format(
                            ' '.join(self.argv[1:])))
        if self.options.help:
            return  # User requested help; nothing to parse

        # Treat "target" with equals sign as propertyname=value pair
        if '=' in self.options.key:
            self.__remaining.insert(0, self.options.key)
            self.options.key = '.'

        # Parse remaining argparse options as propertyname=value pairs
        self.__nameValuePairs = []

        for arg in self.__remaining:
            m = DisableCommand.__nameValuePattern.match(arg)
            if not m:
                raise nvsm.exceptions.InvalidArguments(
                    'Expected propertyname=value pair: {0}'.format(arg))
            pair = (m.group('name'), m.group('value'))
            self.__nameValuePairs.append(pair)

    def run(self, context, file=sys.stdout):
        if self.options.help:
            self.__argparser.print_help(file=file)
            return
        nvsm.log.debug(
                'Running disable command...\n' \
                '    name: {name}\n' \
                '    argv: {argv}\n' \
                '    options: {options}',
                name=self.name, argv=self.argv, options=self.options, keyvalues=self.__nameValuePairs)

        # Traverse the target hierarchy to find matching target
        address = nvsm.targetAddress.TargetAddress(
                self.options.key, context=context)
        target = context.getTarget(address)
        # Make sure target supports "start" verb
        if not hasattr(target, 'disable'):
            raise nvsm.exceptions.CommandError(
                'Target "{target}" does not support "disable" verb'.format(
                target=target.address))
        # Start the job associated with this target
        job = target.disable(skip_prompt=False, wait=False, keyvalues=self.__nameValuePairs)
        if not job:
            return
        print(file=file)


    @staticmethod
    def implements(argv):
        """ Determine if the given command is an NVSM "disable" command.

        This method allows the NVSM Command factory pattern to determine if
        the SetCommand class implements the given command name.

        This method returns True if argv[0] is "lock", case insensitive.
        Returns False otherwise.
        """
        if argv[0].lower() == 'disable':
            return True
        return False

# Register with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command(DisableCommand)

class ExportCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(ExportCommand, self).__init__(argv=argv)

        self.__parse_arguments()

    def __parse_arguments(self):
        self.__argparser = argparse.ArgumentParser(
                prog='export',
                description='''export property''',
                add_help=False)
        self.__argparser.add_argument('-help', '-h', action='store_true', default=False,
                help='show this help message and exit')
        self.__argparser.add_argument('target', nargs='?', action='store', default='.',
                help='''The target address of the Managed Element . The
                        target address can be relative to the current working
                        target, or it can be absolute. Globbing of targets is
                        not supported for this command.''')
        try:
            # TODO: Redirect the sys.stdout printed by the parser (for unit testing purposes)
            self.options = self.__argparser.parse_args(self.argv[1:])
        except SystemExit as e:
            if e.code != 0:
                # TODO: Somehow raise exception with the error message printed by argparse
                raise nvsm.exceptions.InvalidArguments(
                        'Invalid unlock command arguments: {0}'.format(
                            ' '.join(self.argv[1:])))


    def run(self, context, file=sys.stdout):
        if self.options.help:
            self.__argparser.print_help(file=file)
            return
        nvsm.log.debug(
                'Running export command...\n' \
                '    name: {name}\n' \
                '    argv: {argv}\n' \
                '    options: {options}',
                name=self.name, argv=self.argv, options=self.options)

        # Traverse the target hierarchy to find matching target
        address = nvsm.targetAddress.TargetAddress(
                self.options.target, context=context)
        target = context.getTarget(address)
        # Make sure target supports "start" verb
        if not hasattr(target, 'export'):
            raise nvsm.exceptions.CommandError(
                'Target "{target}" does not support "export" verb'.format(
                target=target.address))
        # Start the job associated with this target
        job = target.export(skip_prompt=False, wait=False)
        if not job:
            return
        print(file=file)


    @staticmethod
    def implements(argv):
        """ Determine if the given command is an NVSM "lock" command.

        This method allows the NVSM Command factory pattern to determine if
        the SetCommand class implements the given command name.

        This method returns True if argv[0] is "lock", case insensitive.
        Returns False otherwise.
        """
        if argv[0].lower() == 'export':
            return True
        return False

# Register with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command(ExportCommand)

class FactoryResetCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(FactoryResetCommand, self).__init__(argv=argv)

        self.__parse_arguments()

    def __parse_arguments(self):
        self.__argparser = argparse.ArgumentParser(
                prog='factoryreset',
                description='''factoryreset setting''',
                add_help=False)
        self.__argparser.add_argument('-help', '-h', action='store_true', default=False,
                help='show this help message and exit')
        self.__argparser.add_argument('target', nargs='?', action='store', default='.',
                help='''The target address of the Managed Element . The
                        target address can be relative to the current working
                        target, or it can be absolute. Globbing of targets is
                        not supported for this command.''')

        try:
            # TODO: Redirect the sys.stdout printed by the parser (for unit testing purposes)
            self.options = self.__argparser.parse_args(self.argv[1:])
        except SystemExit as e:
            if e.code != 0:
                # TODO: Somehow raise exception with the error message printed by argparse
                raise nvsm.exceptions.InvalidArguments(
                        'Invalid factory-reset command arguments: {0}'.format(
                            ' '.join(self.argv[1:])))

    def run(self, context, file=sys.stdout):
        if self.options.help:
            self.__argparser.print_help(file=file)
            return
        nvsm.log.debug(
                'Running factoryreset command...\n' \
                '    name: {name}\n' \
                '    argv: {argv}\n' \
                '    options: {options}',
                name=self.name, argv=self.argv, options=None)

        # Traverse the target hierarchy to find matching target
        address = nvsm.targetAddress.TargetAddress(
                self.options.target, context=context)
        target = context.getTarget(address)
        # Make sure target supports "factoryReset" verb
        if not hasattr(target, 'factoryreset'):
            raise nvsm.exceptions.CommandError(
                'Target "{target}" does not support "factoryreset" verb'.format(
                target=target.address))
        # Start the job associated with this target
        job = target.factoryreset(skip_prompt=False, wait=False)
        if not job:
            return
        print(file=file)

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an NVSM "lock" command.

        This method allows the NVSM Command factory pattern to determine if
        the SetCommand class implements the given command name.

        This method returns True if argv[0] is "recover", case insensitive.
        Returns False otherwise.
        """
        if argv[0].lower() == 'factoryreset':
            return True
        return False
nvsm.command.Command.register_command(FactoryResetCommand)

class RotateCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(RotateCommand, self).__init__(argv=argv)

        self.__parse_arguments()

    def __parse_arguments(self):
        self.__argparser = argparse.ArgumentParser(
                prog='rotate',
                description='''rotate keys''',
                add_help=False)
        self.__argparser.add_argument('-help', '-h', action='store_true', default=False,
                help='show this help message and exit')
        self.__argparser.add_argument('target', nargs='?', action='store', default='.',
                help='''The target address of the Managed Element . The
                        target address can be relative to the current working
                        target, or it can be absolute. Globbing of targets is
                        not supported for this command.''')

        try:
            # TODO: Redirect the sys.stdout printed by the parser (for unit testing purposes)
            self.options = self.__argparser.parse_args(self.argv[1:])
        except SystemExit as e:
            if e.code != 0:
                # TODO: Somehow raise exception with the error message printed by argparse
                raise nvsm.exceptions.InvalidArguments(
                        'Invalid rotate command arguments: {0}'.format(
                            ' '.join(self.argv[1:])))

    def run(self, context, file=sys.stdout):
        if self.options.help:
            self.__argparser.print_help(file=file)
            return
        nvsm.log.debug(
                'Running rotate command...\n' \
                '    name: {name}\n' \
                '    argv: {argv}\n' \
                '    options: {options}',
                name=self.name, argv=self.argv, options=None)

        # Traverse the target hierarchy to find matching target
        address = nvsm.targetAddress.TargetAddress(
                self.options.target, context=context)
        target = context.getTarget(address)
        # Make sure target supports "start" verb
        if not hasattr(target, 'rotate'):
            raise nvsm.exceptions.CommandError(
                'Target "{target}" does not support "rotate" verb'.format(
                target=target.address))
        # Start the job associated with this target
        job = target.rotate(skip_prompt=False, wait=False)
        if not job:
            return
        print(file=file)

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an NVSM "rotate" command.

        This method allows the NVSM Command factory pattern to determine if
        the SetCommand class implements the given command name.

        This method returns True if argv[0] is "lock", case insensitive.
        Returns False otherwise.
        """
        if argv[0].lower() == 'rotate':
            return True
        return False

nvsm.command.Command.register_command(RotateCommand)
