# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import copy
import os
import signal
import subprocess
import sys
import re
import logging
from nvsm.color import colorize
import nvsm.command
import nvsm.interpreter
from nvsm.command.queryNGC import PickAndSelect1
import argparse

class EnableExportersCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(EnableExportersCommand, self).__init__(argv=argv)

        self.prompt = "Enabling NVSM Exporters will, \
        \n - Enable & Start the nvsm-exporter systemd service for metric collection and use port 9123. \
        \n \nDo you want to continue"

        self.logfile = "Logfile: %s" % "/var/log/nvsm/nvsm_exporters.log"

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an alias for "show health".

        This method allows the NVSM Command factory pattern to determine if
        the "EnableExportersCommand" class implements the given command.

        This method returns True if the first two arguments of argv look like
        "show health" case insensitive. Returns False otherwise.
        """
        aliases = [
            [ 'enable', 'exporters' ],
            [ 'enable', 'exporter' ],
            [ 'enable', 'metrics' ] ]
        return [x.lower() for x in argv[:2]] in aliases

    def printColor(self, text, color):
        with colorize(color):
            print(text)

    def run(self, context, file=sys.stdout):

        rc, ans = PickAndSelect1(["no", "yes"], self.prompt)

        if rc != 0 or ans != "yes":
            return


        from nvsm.backend.exporters import run_exporters
        exit_code = run_exporters(enable=True)
        if exit_code != 0:
            print("Error Observed: ", exit_code)
            print(self.logfile)
        else:
            self.printColor("Success", "green")
            self.printColor("NVSM-Exporter is running at <node-ip>:9123", "blue")


# Register as a command alias with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command_alias(EnableExportersCommand)

class DisableExportersCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(DisableExportersCommand, self).__init__(argv=argv)

        self.prompt = "Disabling NVSM Exporters will, \
        \n - Disable & Stop the nvsm-exporter systemd service for metric collection on port 9123. \
        \n \nDo you want to continue"

        self.logfile = "Logfile: %s" % "/var/log/nvsm/nvsm_exporters.log"

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an alias for "show health".

        This method allows the NVSM Command factory pattern to determine if
        the "DisableExportersCommand" class implements the given command.

        This method returns True if the first two arguments of argv look like
        "show health" case insensitive. Returns False otherwise.
        """
        aliases = [
            [ 'disable', 'exporters' ],
            [ 'disable', 'exporter' ],
            [ 'disable', 'metrics' ] ]
        return [x.lower() for x in argv[:2]] in aliases

    def printColor(self, text, color):
        with colorize(color):
            print(text)

    def run(self, context, file=sys.stdout):

        rc, ans = PickAndSelect1(["no", "yes"], self.prompt)

        if rc != 0 or ans != "yes":
            return


        from nvsm.backend.exporters import run_exporters
        exit_code = run_exporters(enable=False)
        if exit_code != 0:
            print("Error Observed: ", exit_code)
            print(self.logfile)
        else:
            self.printColor("Success", "green")
            self.printColor("NVSM-Exporter is stopped", "blue")


# Register as a command alias with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command_alias(DisableExportersCommand)
