# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.


import sys

import nvsm.command
from nvsm.command.set import SetCommand
from nvsm.command.show import ShowCommand


class SetPowermodeCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(SetPowermodeCommand, self).__init__(argv=argv)

        # Construct the "show" command that implements this command alias
        # NOTE: argv is of the form "set powermode=foo"
        powermode = argv[1].split('=', 1)[1]
        set_argv = ['set', '/chassis/localhost',
                    'PowerMode={}'.format(powermode)]
        self.__set_command = SetCommand(argv=set_argv)

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an alias for "set Powermode=*".

        This method allows the NVSM Command factory pattern to determine if
        the "SetPowermodeCoimmand" class implements the given command.

        This method returns True if argv looks like "set Powermode", case
        insensitive. Returns False otherwise.
        """
        if len(argv) != 2:
            return False
        if argv[0].lower() != 'set':
            return False
        if '=' not in argv[1]:
            return False
        if argv[1].split('=')[0].lower() != 'powermode':
            return False
        return True

    def run(self, context, file=sys.stdout):
        return self.__set_command.run(context=context, file=file)


# Register as a command alias with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command_alias(SetPowermodeCommand)


class ShowPowermodeCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(ShowPowermodeCommand, self).__init__(argv=argv)

        # Construct the "show" command that implements this command alias
        show_argv = ['show',
                     '-display', 'Properties=Power_Mode',
                     '/chassis/*']
        self.__show_powermode = ShowCommand(argv=show_argv)

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an alias for "show powermode".

        This method allows the NVSM Command factory pattern to determine if
        the "ShowPowermodeCommand" class implements the given command.

        This method returns True if argv looks like "show powermode", case
        insensitive. Returns False otherwise.
        """
        aliases = [
            ['show', 'powermode']]
        return list([x.lower() for x in argv]) in aliases

    def run(self, context, file=sys.stdout):
        return self.__show_powermode.run(context=context, file=file)


# Register as a command alias with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command_alias(ShowPowermodeCommand)
