#!/usr/bin/python3
#
# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.
#
# NGC API Explorer
#   https://docs.ngc.nvidia.com/api
# NGC User Guide - Batch
#   https://docs.google.com/document/d/1kDdYTrEfhmpvTFCAtfw_Ad-KPHSoTv34PaLqL21Tipc
# AuthN sevices
#   https://gitlab-master.nvidia.com/kaizen/auth/authn
# NGC Cli code
#   https://gitlab-master.nvidia.com/ngc/apps/ngc-cli

import requests
import os
import base64
import json
import pprint
import re
import sys
import copy
import traceback
import nvsm.log

###############################################################################

gdbug_flag = False
gapi_key   = None                             # save console entered api key
gpp4       = pprint.PrettyPrinter(indent=4)   # json pretty printer method -

def set_debug(v):
    global gdbug_flag
    gdbug_flag = v;
def get_debug():
    global gdbug_flag
    return(gdbug_flag)

def pp(msg):                                   # easy formatting of json
    global gpp4
    gpp4.pprint(msg)                           # prints output to console

def pf(msg):
    global gpp5
    return(gpp4.pformat(msg))

# prompt_gets
#
# non-command line history retaining escape-ignoring stdin gets type
# function.
#
# Normal input() and nvsm.prompt.String function add input to the command
# line history, and support the up/down arrow buttons to replay history
# entries -- This code is used for multi-level list prompting where keeping
# those history entries is anoying.
#
def prompt_gets(prompt):
    line = ""                                 # set a default, if ^C and have no line
    ansi_escape =re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')
    print(prompt, end = ' ?: ', flush=True)   # python3 feature
    for line in sys.stdin:                    # only get one line
        break

    v = ansi_escape.sub('', line.strip())  # strip escape and leading/trailing whitespace
    if type(v) == bytes:
        v = v.decode('utf-8')
    return v

###############################################################################
# check_for_prefix_flags()
#
def check_for_debug_flags(s):
    if " " in s:       # normal input (api keys) will not have spaces in them
        s = s.strip()  # removei any leading/trailing whitespace
        while s.startswith("--"):
            pos = s.find(' ')  # space
            flag = s[2:pos]
            s = s[(pos+1):].strip()  # skip over spaces till next valie
            if flag == "debug":
                set_debug(True)
            else:
                print("unknown flag '%s'" %(flag))
    return(s)         # original string with any leading flags removed

###############################################################################
# split_image_ref
#
# image_ref is '[nvcr.io/][org/]repo:tag' - break it up
#           or '[nvcr.io/][org/team/]repo:tag'
#
# returns root, org, repo, tag from above string
#
# NOTE: If there is a org/team, like 'nvsupport/dgx1' this code will not break
#       out the team -- the way we are using it here is always 'org/team'
#
def split_image_ref(image_ref):
    pieces = image_ref.split('/')

    # check for root - only valid one is nvcr.io?
    idx = 0
    last = len(pieces)-1

    root = None
    if pieces[0] == "nvcr.io":
        root  = pieces[0]
        idx   = 1                # used one piece

    org = None
    while idx < last:            # org or org/team?
        if org == None:
            org = pieces[idx]
        else:
            org += '/'           # This seperates the org from the team
            org += pieces[idx]
        idx += 1
                                 # get the tag
    parts = pieces[last].split(':')
    repo  = parts[0]

    tag = None
    if len(parts) > 1:           # have a ':' in string? -- it's the tag
       tag = parts[1]

    #print("split_image_ref: %s --> %s %s %s %s" %(image_ref, root, org, repo, tag))
    return root, org, repo, tag  # the 4 parts, many are optional
###############################################################################
# build_image_ref
#
# inverse of split -- given parts, make a whole - Given args do not have any
#                     trailing seperators like '/' on them
#
# Tries to handle case where either beginning or end of path is not given,
# making sure that the leading and trailing seperators make sense. But does
# not attempt to handle holes.
#
def build_image_ref(root, org, repo, tag):
    image_ref = ""

    if root != None and root != "":
        image_ref += root
    if org != None and org != "":
        if root != None and root != "":
            image_ref += '/'
        image_ref += org
    if repo != None and repo != "":
        if org != None and org != "":
            image_ref += '/'
        image_ref += repo
    if tag != None and tag != "":
        if repo != None and repo != "":
            image_ref += ':'
        image_ref += tag

    # print("%s %s %s %s --> %s" %(root, org, repo, tag, image_ref))
    return image_ref

###############################################################################
# Docker_SetupGetLocalImagesCmd
#
# As of 02/2020, the RedHat version of docker that we are using does not
# support the "docker images --format '{{json .}}" command. It works for
# Ubuntu however. This is a porting issue by RedHat as of 2/2020 and earlier.
# I don't know if/when it will be fixed.
#
# But we can fake it with "--format table"
#
# This function returns a string that is more or less equivelent to the
# original -format '{{json .}} form.
#
# When the entire command is called, this function returns a list of currently
# installed on the box. However Ubuntu returns the first line as the title
# of the fields "{"Containers":"CONTAINERS","CreatedAt":"CREATED AT",..}" while
# RedHat doesn't. This needs to be taken into account by caller
#
# The "Digest" field also seems to be returned differenty -- for the
# "--format '{{json .}}" it seems to always be returned as "<None>", but for
# the "--format table" it's returned as "sha256:3028017c6e5435df58ae0..."
#
# <image> is a specific docker image name. By default this is none, so the
# entire list is returned.
#
def Docker_SetupGetLocalImagesCmd(image=None):
    fields=["Containers", "CreatedAt", "CreatedSince", "Digest", "ID",
            "Repository", "SharedSize", "Size", "Tag", "UniqueSize",
            "VirtualSize" ]
    fields=["Tag", "ID", "Repository", "CreatedAt"]
    s=None
    for f in fields:
        if s == None:
            s = "{"                         # start
        else:
            s += ","                        # in between tokens
        s += "\\\"%s\\\":\\\"{{.%s}}\\\"" %(f, f)
    s += "}"                                # ending bracket
    cmd = "sudo docker images"
    if image != None:
        cmd += " %s" %(image)               # particular image
    cmd += " --format \"table %s\"" %(s)
    return cmd                                # output

###############################################################################
# Docker_GetListOfLocalImages
#
# Get list of docker images currently on the box -- I.E "docker images"
#
# returned  format simular to other lists, it contains full org/repo:tag
# followed date for each image. Intened to be used by picker code
#
def Docker_GetListOfLocalImages(image_ref=None):
    images = []
    line = ""


    #NOTE: This standard docker command which works correctly on Ubuntu version
    #      of docker, does not work in RedHat release as of Febuary 2020
    #
    # cmd = "docker images --format '{{json .}}'"   # get all fields with '.'
    #
    # which is why this was written to more-or-less emulate it using the '--format table'
    # command. This is used for both Ubuntu and Redhat so the coding changes are minimal.
    #
    cmd = Docker_SetupGetLocalImagesCmd(image_ref) # get command to get all fields of one or all local images
    try:
        if get_debug() == True:
            print(cmd)
        data = os.popen(cmd).read().strip()        # remove leading/trailing lines/spaces
        if(len(data)==0):
            return 1, "No local docker image found. Retry with remote option. Exiting...", ""

        # {"Tag":"20.03.1","ID":"c8a1816d6531","Repository":"nvcr.io/nvdgx1-dev/nvfw-dgx1-pbradstreet",i
        #  "CreatedAt":"2020-03-06 16:13:36 -0500 EST"}

        ary  = data.split('\n')               # list of '\n' seperated '{..lines...}'
        fmt  = "%12s %s"
        for line in ary:

            # Redhat docker version 1.13.1 as of 3/2020 will prepend the tags for the line
            # as a non-json string BEFORE the first line of data when using the
            # 'docker images --format 'table'
            #
            #   'TAG                 IMAGE ID            REPOSITORY          CREATED AT'
            #
            # The json decoder will have problems with this, so don't accept it
            # Note that Ubuntu docker version as of this date doesn't do this.

            if line.strip()[0] != '{':
                continue                      # not a json line

            jline      = json.loads(line)     # load each individual line into json format for easy parsing
            if get_debug() == True:
                pp(jline)

            orgrepo    = jline["Repository"]  # "Repository":"nvfw-dgx1-root" or
                                              # "Repository":"nvcr.io/nvdgx1-dev/nvfw-dgx1-pbradstreet"
            tag        = jline["Tag"]
            date       = jline["CreatedAt"]
            image_ref  = "%s:%s" %(orgrepo, tag)
            ID         = jline["ID"]

            # This is an area of needed improvement required because of the "format --table" form
            # of the command -- it may return the field Names in line one: as "Containers":"CONTAINERS"
            # Ignore that line if we see it -- See Docker_SetupGetLocalImagesCmd()

            if tag != "TAG":
                images.append((image_ref, fmt %(ID, date[0:10]) ))   # just the date, skip the time

        # return list of local docker images -- note that this list has not been
        # filtered at this point -- let the picker do it, same as done for  other ngc lists

        title = fmt %("Image ID", "Date")
        return 0, images, title               # success with d$responses output string
    except Exception as e:
        nvsm.log.error("Getting list of local docker images")
        print(e)
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_tb(exc_traceback, limit=3, file=sys.stdout)
        return 1, cmd, ""

##############################################################################
# Docker_ValidateLocalImageRef
#
# Used by nvsm cli after the download to verify that the image that was
# downloaded is what was expected - another parnoid check
#
def Docker_ValidateLocalImageRef(image_ref):
    root, org, repo, tag = split_image_ref(image_ref)   # break out the pieces

    # use the above "get the list" function to do the work -- but it only
    # should return the line matching the supplied "image_ref"

    rc, list_of_images, title = Docker_GetListOfLocalImages(image_ref)   # returns 1 item
    if rc != 0:
        return rc, list_of_images                  # list_of_images is error message string

    # now check what's actual got on first line -- it should match our image_ref

    if len(list_of_images) > 0:                    # assuming we got any thing
        val = list_of_images[0]
        dorgrepo = val[0]                          # val line is two fields: ["ref", "Id date"]
        droot, dorg, drepo, dtag = split_image_ref(dorgrepo)
        if get_debug() == True:
            print(list_of_images)
            print("'%s' --> '%s' '%s' '%s' '%s'" %(dorgrepo, droot, dorg, drepo, dtag))
        if root == droot and org == dorg and repo == drepo and tag == dtag:
            return 0, "success"

    return 1, "did not find '%s' in local machine docker repository" %(image_ref)

##############################################################################
# Docker_ValidateLocalRepo
#
# Same as above, different args
#
# Given the split root, org, repo, tag strings, verify that it's on the box
# within the local docker repos.
#
# Don't care if it's been pulled down remotely -- only checking for if it
# exists locally -- I.E just checks the full name -- and a remote name is
# kept within local docker repo
#
def Docker_ValidateLocalRepo(root, org, repo, tag):
    image_ref = build_image_ref(root, org, repo, tag)

    return Docker_ValidateLocalImageRef(image_ref)

##############################################################################
# NGC_ErrorToString
#
# Converts responses.response status to a string
#
# Note -- caller should check response.status before this call and call it if
#         the error code is non-success (200). This code is for the error path
#
def NGC_ErrorToString(response):
    lst = { 200:"Success",
            400:"Bad Request", 401:"Unauthorized",  403:"Forbidden",
            404:"Not Found",   409:"Object Exists", 500:"Internal Error"
          }
    code = response.status_code
    if code in lst:
        msg = "%3d:%s" %(code, lst[code] )
    else:
        msg = "%4d:%s" %(code, "Unknown error code")
    return msg

###############################################################################
# NGC_request
#
# Make the call to python request libaray
#
# Single location where call is made -- hook into nvsm job control here
#
# Returns:      0, dictionary      success, data allways converted to python dict
#              !0, msg             failure, returned error message is a string
#
def NGC_Request(cmd, url, headers, querystring):
    # TODO: put nvsm job timeout/check/... here


    if querystring != None and querystring != "":
        if get_debug():
            print("GET %s\n  headers=%s\n  params=%s" %(url, pf(headers), pf(querystring)))
        response = requests.request("GET", url, headers=headers, params=querystring)
    else:
        #if get_debug():
        #    print("GET %s\n  headers=%s" %(url, pf(headers)))
        response = requests.request("GET", url, headers=headers)

    # handle any error we see -- top level

    if response.status_code != 200:
        msg  = NGC_ErrorToString(response)     # decode the string
        msg += " for '%s'" %(url)
        return(response.status_code, msg)

    # convert text to json,
    rjson = json.loads(response.text)          # result json
    if get_debug():
        tjson = copy.deepcopy(rjson)           # make a copy that can be modifed
                                               # -- for debuging only, time not an issue
        if "token" in tjson:
            l = len(tjson["token"])            # token str can be > 10k
            if l > 100:                        # truncate it for debug output if it's really long
                tjson["token"] = "%s...truncated...len=%d" %(tjson["token"][0:50], l)  # string 10k or larger, trim it
        pp(tjson)                              # pretty print it

    # return the unmodified response
    return 0, rjson                            #  success with responses

###############################################################################
# NGC_GetAuthenticationToken
#
# Given a API key, returns a token. These tokens are only alive for about 10
# minutes.
#
# Note -- throughout this code, the 'org' being passed around will include
#         both the traditional 'org' name like "nvdgx1-dev" if the org doesn't
#         have teams (not followed by a "/"), or both the org and team for
#         something like 'nvsupport/dgx1'.  There are very few places here where
#         they can not be treated as one string -- except for here of course
#
# See:    https://gitlab-master.nvidia.com/kaizen/wiki/blob/master/AuthFlow.md
#
def NGC_GetAuthenticationToken(api_key, org):

    # For org/teams, need ot have the scope string for both the org, and the org/team
    # resulting params json structure that is being created will look like this
    #
    #  {   'scope':    [  {'group/ngc': 'nvsupport'},
    #                     {'group/ngc': 'nvsupport/dgx1'}
    #                  ],
    #       'service': 'ngc'
    #  }
    #
    # For reference, If this was passed on the cli command line, it would be of the form
    #
    #  service=ngc,scope=group/ngc:nvsupport&scope=group/ngc:nvsupport/dgx1
    #

    parts = org.split('/')
    p_org = parts[0]
    qval    = []                                # could be multiple scopes depending on org or org/team
    qval.append( {'group/ngc':'%s' %(p_org)} )
    if len(parts) == 2:                         # should only be org or org/parts...
        p_team = parts[1]                       # use the magic '&' operator to include org/team
        qval.append( {'group/ngc':'%s/%s' %(p_org, p_team) } )
    query={"service":"ngc", "scope":qval}

    url           = "https://authn.nvidia.com/token"
    api_header    = "$oauthtoken:{0}".format(api_key)

    headers = {
        "Authorization": "Basic {}".format(base64.b64encode(api_header.encode('utf-8')).decode('utf-8')),
        'Content-Type': "application/json",
        'User-Agent': "User/7.17.1",
        'Accept': "*/*",
        'Cache-Control': "no-cache",
        'Host': "authn.nvidia.com",
        'Accept-Encoding': "gzip, deflate",
        'Connection': "keep-alive",
        'cache-control': "no-cache"
        }
    rc, rtext = NGC_Request("GET", url, headers, query)
    if rc != 0:
        return rc, rtext         # rtext is error code in this case

    # convert from json back to real python dictionary

    token = rtext["token"]    # json form of token (valid time is avaliable here also)

    return(0, token)          # success, with token

###############################################################################
# NGC_Get
#
# Gets data from a url - simplest format
#
def NGC_Get(token, url):
    api_header = "Authorization:".format(token)
    headers = {
        'Authorization': "Bearer {}".format(token),
        'User-Agent': "User/7.17.1",
        'Accept': "*/*",
        'Cache-Control': "no-cache",
        'Host': "api.ngc.nvidia.com",
        'Accept-Encoding': "gzip, deflate",
        'Connection': "keep-alive",
        'cache-control': "no-cache"
        }
    # returns rc and errmsg if rc!=0  or python dict if rc=0
    rc, rtext = NGC_Request("GET", url, headers, None)
    return rc, rtext              # rtext is python_dict if rc=0, or errmsg if rc != 0

###############################################################################
# def Get_UsersAPIKey
#
# Returns the 64-ish byte long API key, provided by NGC
#
# Ether this comes from an 1)  API_KEY enviornment variable, 2) from the promptin
# the user or 3) from what has been gathered before
#
# If 'clear' is True, then any old value is cleared, forcing refresh from
# env variable or user prommpt
#
# Returns   -1      help text was
#
def Get_UsersAPIKey(platform_name, clear):
    global gapi_key

    if (clear == True):
        gapi_key = None                             # clear cached version if asked

    api_key = gapi_key                              # get saved copy from prior run

    if api_key == None:                             # try to get from env
        api_key = os.getenv("API_KEY")
    if api_key == None:                             # prompt the user
        try:
            prompt = "Enter your NGC API Key, or '?' for help"

            text   = prompt_gets(prompt)
            if text == "?":                       # bit of a hint
                msg = "\n"\
                       "The API Key that is provided will determine which images that you are authorized\n"\
                       "to pull from NCC. Please seen the NGC documentation and perhaps Nvidia Account\n"\
                       "Manager for details on how to obtain your key.\n"\
                       "\n"\
                       "You can also use the  enviornment variable API_KEY' to pass in your key, as in:\n"\
                       "\n"\
                       "    export API_KEY=<value>\n"\
                       "\n"\
                       "If this env variable is set, then you will not be prompted to provide any\n"\
                       "API Key here in this series of prompts.\n\n"\
                       "When using env variables as non-root user, be sure to use 'sudo -E nvsm ...' \n"\
                       "to preserve them into the sudo call to nvsm"

                return(-1, msg)                     # use '-1' return not to print this error message

            api_key = check_for_debug_flags(text)   # pull out any debug options
        except KeyboardInterrupt:                   # ^C
            print("")
            return 1, "KeyboardInterrupt"
        except Exception as e:
            nvsm.log.error("Getting users API key")
            print(e)
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_tb(exc_traceback, limit=3, file=sys.stdout)
            return 1, "no select"

    if api_key == None:
        msg = "No NGC API_KEY, unable to proceed"
        return(1, msg)
    gapi_key = api_key                             # save copy so don't need to ask again
    return 0, api_key

###############################################################################
# NGC_AuthenticationStart
#
# Given a user's NGC generated api_key, Verfies that
#   1) Can connect to NGC with that key, if 'org' is None, use a default org
#   2) Can get their authenticaion token
#
def NGC_AuthenticationStart(api_key, org="nvidian"):      # 'nvidian' is a default, low access rights  user

    # get token for api_key (which needs to be string, not bytes)
    # return rc=0 if good, or !0 if bad

    rc, token = NGC_GetAuthenticationToken(api_key, org)
    return(rc, token)           # already handled error message if needed

##############################################################################
# NGC_AuthenticateUser
#
# Verifies that the user can access ngc with the given token, and optional
# org. This is a cross-check after creating the token, and getting
#
# Returns:  rc        0 or error
#           info      string error msg of rc=1, or list of orgs user can access
#                     list is a tuple, in (name, optional_text)
#           title     description for optional text above
#           username  Name of key's owner
#
def NGC_AuthenticateUser(token, org=None):
    if org == None:        # no org, basic user access
        rc, rtext, title, username = NGC_GetUserInfo(token)
        if rc != 0:
            msg = "Have valid NGC token, but unable to get any NGC user info: %s" %(rtext)
            return rc, msg, ""
    else:                  # check to see if user has access to the given org
        rc, rtext, title, username = NGC_GetUserInfo(token, org)
        if rc != 0:
            msg = "Can't get user info for NGC org '%s': - do you have access? : %s" %(org, rtext)
            return rc, msg, ""
        if rtext == None:
            msg = "No orgs found"
            return 1, msg, ""

    return 0, rtext, title, username          # non-org info

###############################################################################
# NGC_GetUserInfo
#
# Returns a structure that gives user info -- if an Org name is given,
# provides additional information about that org if it exists
#
# Without the org, gives info on all the orgs and teams that the user has
# access to -- thus, this code also returns a list of those, simular
# to what NGC_GetListOfOrgsUserCanAccess.  Difference is that this list
# ALSO returns any teams the user can access, while th 'orgs' command only
# returns the orgs
#
# Returns  rc         0:ok, 1:err
#          orgs/msg   if rc=0, then list of orgs, if err then msg string
#          title      if rc=0, then title bar for list of orgs
#          username   name of key holder
#
# List is returned in form of below, with both org and org/team entries
#
#    [  "org1":"description", "org2":"description", ... ]
#
def NGC_GetUserInfo(token, org=None):
    url = "https://api.ngc.nvidia.com/v2/users/me"
    if org != None:
        url += "?org-name=" + org   # optional org name - like dgx1-dev

    rc, rtext = NGC_Get(token, url)
    if rc != 0:
        return rc, rtext, "", ""    # error string in 'rtext'

    #   {
    #     "requestStatus": {
    #       "statusCode": "SUCCESS",
    #       "requestId": "56527f75-37c8-4d64-bda5-e4dec9d9b139"
    #     },
    #     "user": {
    #       "isActive": false,
    #       "name": "Peter Bradstreet",
    #       "roles": [
    #         {
    #           "orgRoles": [
    #             "REGISTRY_USER_USER"
    #           ],
    #           "team": {
    #             "repoScanSettings": {
    #               "repoScanShowResults": false,
    #               "repoScanByDefault": false,
    #               "repoScanAllowOverride": false
    #             },
    #             "name": "nvidia",
    #             "description": "nvidia internal (for DCSHWAPIKEY)",
    #             "isDeleted": false,
    #             "id": 501
    #           },
    #           "org": {
    #             "repoScanSettings": {
    #               "repoScanShowResults": false,
    #               "repoScanByDefault": false,
    #               "repoScanAllowOverride": false,
    #               "repoScanEnabled": false,
    #               "repoScanEnableTeamOverride": false
    #             },
    #             "description": "Nvidia Internal org. For Drive Constellation dev team use, test here before moving to QA phase.",
    #             "country": "US",
    #             "industry": "Other",
    #             "type": "ENTERPRISE",
    #             "id": 52064,
    #             "name": "driveconst-dev"
    #           },
    #           "teamRoles": [
    #             "REGISTRY_USER_ADMIN"
    #           ]
    #         },
    #
    orgs = []
    username = rtext['user']['name']             # key holder's user name
    for i in rtext['user']['roles']:
        name        = i['org']['name']
        description = i['org']['description']

            # in this NGC 'users/me' query, it will also give us team info
            # a team is a single sub-directory to an org. Each one is listed
            # here in this command seperatly -- this is not a list.
            # 'roles' above is the list
            # Make the 'name', the '<org>/<team>' value

        if 'team' in i:
            name = name + '/' + i['team']['name']
            description = i['team']['description']
        orgs.append((name, description))
    title = "Description"
    return 0, orgs, title, username

###############################################################################
# NGC_GetListOfOrgsUserCanAccess
#
# returns a list of orgs user can access, and their descriptions
#
# list is returned in simple form, and array of
#
#    [  "org1":"description", "org2":"description", ... ]
#
# This code doesn't return the teams ones belongs to in the orgs -- See
# NGC_GetUserInfo for that.
#
def NGC_GetListOfOrgsUserCanAccess(token, org = ""):
    url = "https://api.ngc.nvidia.com/v2/orgs"
    rc, rtext = NGC_Get(token, url)
    if rc != 0:
        return rc, rtext, ""       # problems, return error code and error msg

    # Grab the name/description out of our array

    #   {
    #     "organizations": [
    #                          {   'country': 'US',
    #                              'description': 'NVIDIA owned. For RTX QA team use.',
    #                              'id': 56264,
    #                              'industry': 'Other',
    #                              'name': 'nvrtx-qa',
    #                              'repoScanSettings': {   'repoScanAllowOverride': False,
    #                                                      'repoScanByDefault': False,
    #                                                      'repoScanEnableTeamOverride': False,
    #                                                      'repoScanEnabled': False,
    #                                                      'repoScanShowResults': False},
    #                              'type': 'ENTERPRISE'},
    #                          { . . . }
    #                      ],
    #        'paginationInfo': {   'index': 0,
    #                              'size': 100,
    #                              'totalPages': 1,
    #                              'totalResults': 9},
    #        'requestStatus':  {   'requestId': '843bee08-6852-4aee-900c-8c60135b34cc',
    #                              'statusCode': 'SUCCESS'}
    #   }

    orgs = []
    for i in rtext["organizations"]:
        orgs.append((i["name"], i["description"]))
    title = "Description"
    return 0, orgs, title

###############################################################################
# NGC_GetListOfReposAtOrg
#
# returns a list of version tags for repo,
#
# 'org' can be a single org name, like 'nvdgx1_dev'
#       or it can be a org/team as in 'nvsupport/dgx1"
#
# While this function call is agnostic, to that, the actual NGC calls that we
# make must be different depending on if a team was specified or not
#
def NGC_GetListOfReposAtOrg(token, org = ""):

    parts=org.split('/')
    if len(parts) == 1:
        get_teams = "False"         # True or False
        get_public = "False"        # True or False
        url = "https://api.ngc.nvidia.com/v2/org/%s/repos?include-teams=%s&include-public=%s" \
                                  %(org, get_teams, get_public)
    else:
        get_org = "False"           # True or False
        get_public = "False"        # True or False
        p_org  = parts[0]
        p_team = parts[1]
        url  = "https://api.ngc.nvidia.com/v2/org/%s/team/%s/repos?include-org=%s&include-public=%s" \
                                  %(p_org, p_team, get_org, get_public)

    rc, data = NGC_Get(token, url)
    if rc != 0:
        return rc, data, ""       # problems, return error code and error msg

    #   {   'paginationInfo': {    'index': 0,
    #                              'size': 5,
    #                              'totalPages': 1,
    #                              'totalResults': 5},
    #       'repositories': [   {  'canGuestPull': False,
    #                              'description': 'Please add description',
    #                              'displayName': 'nvfw-dgx1',
    #                              'isPublic': False,
    #                              'isReadOnly': False,
    #                              'latestImageSize': 330096405,
    #                              'latestTag': 'latest',
    #                              'name': 'nvfw-dgx1',
    #                              'namespace': 'nvdgx1-dev',
    #                              'scanEnabled': True,
    #                              'sharedWithOrgs': ['nvdgx1-dev'],
    #                              'shortDescription': 'Please add description',
    #                              'tags': [   '19.10.4',
    #                                          '19.10.3',
    #                                          '19.10.2',
    #                                          '19.10.1',
    #                                          '19.09.5',
    #                                          '19.09.4',
    #                                          '19.09.99',
    #                                          '19.09.3',
    #                                          '19.06.2',
    #                                          'latest',
    #                                          '19.06.1'],
    #                               'updatedDate': '2019-10-14T23:16:05.674Z'
    #                             },
    #                             { ... }  # more
    #                         ],
    #       'requestStatus': {   'requestId': '80c4ecbb-7f47-48d3-9ab2-4595ff1af360',
    #                             'statusCode': 'SUCCESS'}
    # }
    repos = []
    for i in data["repositories"]:
        repos.append((i["name"], i["shortDescription"]))   # also a long description

    title = "Description"
    return rc, repos, title

###############################################################################
# NGC_GetListOfTagsAndImageInfoForRepo
#
# simular to above, less upper level info, more detail about images
#
def NGC_GetListOfTagsAndImageInfoForRepo(token, org, repo):
    parts=org.split('/')
    if len(parts) == 1:        # just a simple org
        url = "https://api.ngc.nvidia.com/v2/org/%s/repos/%s/images" %(org, repo)
    else:                      # org/team -- different call point
        p_org  = parts[0]
        p_team = parts[1]
        url  = "https://api.ngc.nvidia.com/v2/org/%s/team/%s/repos/%s/images" %(p_org, p_team, repo)

    rc, data = NGC_Get(token, url)
    if rc != 0:
        return rc, rtext, ""    # problems, return error code and error msg

    #   {
    #     "images": [
    #       {
    #         "scanStatus": "IN_PROGRESS",
    #         "scanEnabled": true,
    #         "updatedDate": "2019-10-14T23:16:05.734Z",
    #         "tag": "latest",
    #         "user": { # THIS STRUCT APPEARS TO BE EMPTY/FILLED FIELDS BASED ON ACCESS RIGHTS !
    #           "name": "svcdgxbuilds",
    #           "email": "svcdgxbuilds@nvidia.com",
    #           "id": 50569
    #         },
    #         "size": 330096405,
    #         "digest": "sha256:e034c9b0d552365198083276abc97a59d638ef1810e23cedd7ddbb40f97c834f"
    #       },
    #       { ... }
    #     ]
    #   }
    tags = []
    fmt  = "%8.8s %10s %10.10s %s"
    for i in data["images"]:
        digest = i["digest"].split(':')       #  "sha256:e034c9b0d...
        if len(digest) > 0:
            digest = digest[1]                # get the number, ignore the type
        else:
            digest = ""                       # a valid string -- don't know why didn't get value

        ## NVBug: 200583382
        ## Some Accounts don't have the user struct populated
        ## For those: hardcode email field
        if("email" not in i["user"]):
            i["user"]["email"] = "enterpriseservices@nvidia.com"
        text = fmt %( digest, i["size"],  i["updatedDate"], i["user"]["email"])
        tags.append((i["tag"], text))
    title = fmt %("Digest", "Size", "Updated", "Contact email")
    return 0, tags, title

###############################################################################
# NGC_ValidateRemoteRepo
#
# validate the given repo -- grabs blob and histroy information from the
# tagged repo -- if we can do this, know should be able to download it
#
def NGC_ValidateRemoteRepo(token, org, repo, tag):
    parts=org.split('/')
    if len(parts) == 1:        # just a simple org
        url = "https://api.ngc.nvidia.com/v2/org/%s/repos/%s/images/%s" %(org, repo, tag)
    else:                      # org/team -- different call point
        p_org  = parts[0]      # don't overwrite full 'org' -- need it later to verify
        p_team = parts[1]
        url = "https://api.ngc.nvidia.com/v2/org/%s/team/%s/repos/%s/images/%s" %(p_org, p_team, repo, tag)
    rc, data = NGC_Get(token, url)
    if rc != 0:
        return rc, data       # problems, return error code and error msg

    #   {
    #     "requestStatus": {
    #       "statusCode": "SUCCESS",
    #       "statusDescription": "ok",
    #       "requestId": "a9a986a8-589b-474d-9b80-3293cbec460f"
    #     },
    #     "signatures": [
    #       {
    #         "header": {
    #           "alg": "ES256",
    #           "jwk": {
    #             "y": "NlTmQ_3AUuT_uL_t3yTQRJPlP2o7Ug3N7MwzqB2sKiU",
    #             "x": "xZ-dpX_8lZAoXOLBThxMYbXU57_Won2VsdlZvYjESOA",
    #             "crv": "P-256",
    #             "kty": "EC",
    #             "kid": "2FD2:K2GJ:FW4X:B7XQ:74BZ:I4WZ:OX75:XJRJ:ALVE:UV5I:OPHY:7CM5"
    #           }
    #         },
    #         "signature": "i0HtTQmjUHeDae70JMiVyugJJARTMmfnjai3F_i-rv81z_Eg85sgBPc9Yaj7MC2ZEDW-BwnvZh3GTGbNfdGGeg"
    #       }
    #     ],
    #     "name": "nvdgx1-dev/nvfw-dgx1",
    #     "tag": "19.10.4",
    #     "architecture": "amd64",
    #     "fsLayers": [
    #       {
    #         "blobSum": "sha256:a3ed95caeb02ffe68cdd9fd84406680ae93d633cb16422d00e8a7c22955b46d4"
    #       },
    #       { ... }
    #     ],
    #     "schemaVersion": 1,
    #     "history": [
    #       {
    #         "v1Compatibility":  # ! potentially intersting things here
    #       },
    #       { ... }
    #     ]
    #   }

    expected_name = "%s/%s:%s" %(org, repo, tag)
    code = data["requestStatus"]["statusCode"]
    if code != "SUCCESS":
        desc = data["requestStatus"]["statusDescription"]
        return 1, "Failed validate on %s: %s %s" %(expected_name, code, desc)

    # painoid check -- the org/repo/tag name in the output here should agree with
    # the values we passed it -- this should always be true -- but...

    got_name = "%s:%s" %(data["name"], data["tag"])
    if got_name != expected_name:
        return 1, "Asked for '%s', got info back on '%s' - something is not working as expected" %(expected_name, got_name)

    # information we want to pass back to user?
    text = "%s" %(got_name)
    return(0, text)

###############################################################################
# PickAndSelect1
#
# Used like:   PickAndSelect1(["n", "y"], "Pull %s ?" %(repo_info))
#
# 'prompt' is a simple question, and 'picks' are a short list answers. These
# picks can be multiple characteters. They are case-insensitive checked against
# what the user enters, and any match is returned in it's entirely.
#
# Returns: 0, value        picked seleccted
#          1, msg          aborted, ^C, -- not picked
#
def PickAndSelect1(picks, prompt):

    # convert list to "n/y" for display
    cnt = 0
    select = ""
    for i in picks:
       if cnt > 0:
           select += "/"
       select += i    # the pick item
       cnt += 1

    while 1:
        try:
            text = prompt_gets("%s (%s)" %(prompt, select))
            v    = check_for_debug_flags(text)   # pull out any debug options
            for i in picks:
                if v.lower() == i.lower():
                    return 0, i   # return the pick value

        except ValueError as e:
            pass                  # will de-display prompt again
        except KeyboardInterrupt:            # ^C
            print("")
            return 1, "KeyboardInterrupt"
        except Exception as e:
            nvsm.log.error("Picking '%s'" %(prompt))
            print(e)
            return 1, "no select"

###############################################################################
# ShowListAndSelect1
#
# Simple list show utility -- gives user a list of items with and index and
# description, allows the user to select one by index, and returns with the
# name
#
# list consists of   [  ("name1", "description1"),
#                       ("name2", "description2"),
#                        ...
#                    ]
#
# name_filt:     regex expression of names allowed to be seen
# sort_dir:      is -1 for reverse, 0 for no sort, 1 for forward sort
# item_on_top:   if see this value in list, will always be placed on top
# sortfunc:      func to fixup value for sorting -- add leading zeros to numbers for example
#
# Returns   0, name        # for selected list
#           1, <text>      # for error
#
# The 'fieldname' and 'title' are used above to the outputed list to provide more
# info only
#
def ShowListAndSelect1(prompt, lst, sortdir = 1, fieldname="Value", title = None,
                       name_filt = None, item_on_top = None, sortfunc=None):

    # filter out any names -- full regex expressions like - "nvfw-*"

    if name_filt != None:
        newlst = []
        p = re.compile(name_filt)
        for val in lst:
            if p.match(val[0]):                 # match?
                newlst.append(val)
        lst = newlst                            # make a new list, can't remove in place

    # check that the list is not empty

    if len(lst) == 0:
        return 1, "'%s' filtered by '%s' is empty, nothing to download here" %(prompt, name_filt)

    # sort, with fancy option

    if sortdir == 1:
        lst.sort(reverse=False, key=sortfunc)                    # sort the list
    elif sortdir == -1:
        lst.sort(reverse=True, key=sortfunc)                     # sort the list

    # let caller move one item -- like 'latest' to start of list if it exists

    if item_on_top != None:
        for val in lst:
            if val[0] == item_on_top:      # looking for 'latest' or simular
                lst.remove(val)
                lst.insert(0, val)         # move item (slowly) to first item in list

    # get longest name in the list -- will do some formatting

    maxlen = 24                    # minimum space of first item
    for val in lst:                # for all the items in the list
        if len(val[0]) > maxlen:
            maxlen = len(val[0])
    maxlen += 2

    # show title/list, select and return, or re-display if invalid select
    while 1:
        try:
            # show the title, note that 'title' string is just for the optional data, not for
            # the index and the key value, so we need to create that here

            if title != None:
                print( " %3s  %s %s" %("Idx", fieldname.ljust(maxlen), title))  # line up with data

            # now the data in the list

            pos = 0
            for val in lst:              # sorted list
                print(" %3d  %s %s" %(pos, val[0].ljust(maxlen), val[1]))
                pos += 1

            text = prompt_gets("%s" %(prompt))          # FIXME: 'input' in python 3?
            v    = check_for_debug_flags(text)   # pull out any debug options

            # allow user to enter either the full tag name, or the index
            for val in lst:
                if v == val[0]:
                    return 0, val[0]          # found by name, return the name
            # ok, try it as an index
            idx = int(v)
            if idx >= 0 and idx < pos:
                return 0, lst[idx][0]         # good return: just want item, not description
            else:
                print("Please enter value between 0 and %d" %(pos-1))
        except ValueError as e:
            print("%s is not a value in the list or valid index number." %(e.args[0].split(": ")[1]))
            pass
        except KeyboardInterrupt:            # ^C
            print("")
            return 1, "KeyboardInterrupt"
        except Exception as e:
            nvsm.log.error("Selecting for '%s'" %(prompt))
            print(e)
            return 1, "no select"           # never let them leave until do right

###############################################################################
# ShowListOf Orgs/Repos/Tags AndSelect1
#
# Show and Select functions -- given one of the "list_of...", sorts it properly,
# displays it and allows user to select one of the items from it
#
# Returns -- 0, "item"   or 1, "error msg"
#
def ShowListOfOrgsAndSelect1(prompt, list_of_orgs, title, orgs_filter = None):
    return ShowListAndSelect1(prompt,list_of_orgs, 1, "Org", title, orgs_filter)

def ShowListOfReposAndSelect1(prompt, list_of_repos, title, repos_filter = None):
    return ShowListAndSelect1(prompt, list_of_repos, 1, "Repo", title, repos_filter)   # want only nvfw- prefixed containers

def tag_fix_version(elem):          # used with sort()
    pieces = elem[0].split('.') # want to put "1.1.18" below "19.1.18"
    output = ""
    cnt    = 0
    for i in pieces:
        try:
            k = int(i)          # successfull if python thinks string is a int
            j = "%04d" %(k)     # lead all number with same number of zeros
        except:
            j = i               # not a number, don't modify it
        if cnt > 0:
            output += '.'       # tag version seperator
        output += j             # 0 leading number, or unmodifed text
        cnt += 1
    return output               # all version numbers are 0 - prepended "0001.0001.0018"

def ShowListOfRepoTagsAndSelect1(prompt, list_of_version_tags, title, tags_filter = None):
    return ShowListAndSelect1(prompt, list_of_version_tags, -1, "Tag", title, tags_filter, "latest", tag_fix_version)

def tag_extract_fix_version(elem):        # elem is full or partial root/org/tag:version -- only want version
    pieces = elem[0].split(":")
    if len(pieces) > 0:                   # got the ':tag'
        v = []
        v.append(pieces[1])               # 1 item long list
        return tag_fix_version(v)         # consistent version formatting - see above
    return ""                             # no version tag, shouldn't really be here, but what the heck..

def ShowListOfDockerImagesAndSelect1(prompt, list_of_images, title, image_filter = None):
    return ShowListAndSelect1(prompt, list_of_images, -1, "Repository", title, image_filter, "latest", tag_extract_fix_version)

###############################################################################
# NGC_ValidateRepoAndAskIfPull
#
# Given the root/org/... pieces, validate that it exists and that the user
# has access rights to pull it down (same thing most of the time)
#
# If 'auto' is not true, will present full image_ref and prompt the user if
# they want to download it
#
# Returns the 0 and structure full of info if Success, or partial structure
# with error in results["status"] field and a return code of 1 if there
# was a problem.
#
#   {   'auto': False,
#       'backend': {   'DockerImageRef': u'nvcr.io/driveconst-dev/nvfw-dcc:19.09.09',
#                      'DockerPassword': 'ZG9...1ZjNi',
#                      'DockerUsername': '$oauthtoken',
#                      'DockerMode': '<value>'     # 'remote' or 'local,
#                      'Flags': 'show_version'
#                   },
#       'local': False,
#       'msg': '',
#       'org': u'driveconst-dev',
#       'repo': u'nvfw-dcc',
#       'status': 'Success',
#       'tag': u'19.09.09'
#   }
#
# Both local-from-docker and remote images go through here. The returned information
# is all that is necessary to pass onto the nvsm backend
#
def NGC_ValidateRepoAndAskIfPull(install_args, api_key, token, root, org, repo, tag, auto=False, local=False, username="unknown"):
    image_ref = build_image_ref(root, org, repo, tag)

    # prepare dictionary that we are going to return with info
    # two layers -- info, and fields we want to pass to backend

    results            = {}          # top level dictionary
    results["msg"]     = ""
    results["org"]     = org
    results["repo"]    = repo
    results["tag"]     = tag
    results["auto"]    = auto
    results["local"]   = local       # True for local, False for remote (via nvcr.io)
#   results["token"]   = token       # the 800'ish byte log version -- needed?

    results["backend"] = {}          # dictionary of keys to pass to nvsm backend - later
    results["backend"]["DockerImageRef"] = image_ref

    # strange escaped space tokenizing of flag args required by backend parser/comm system
    # these are the update container command line options, being passed to it by nvsm
    results["backend"]["Flags"] = ' '.join(install_args)

        # handle verification different if local vs remote image

    if local == True:                # local image -- in local docker repository
        results["backend"]["DockerMode"] = "local"          # value saved in backend
        # print("NGC_ValidateRepoAndAskIfPull: root:'%s', org:'%s', repo:'%s', tag:'%s'" %(root, org, repo, tag))
        rc, status = Docker_ValidateLocalRepo(root, org, repo, tag)

    else:                            # remote - backend information to be passed down
        results["backend"]["DockerMode"]     = "remote"
        results["backend"]["DockerPassword"] = api_key         # 60-ish byte long ngc key
        results["backend"]["DockerUsername"] = "$oauthtoken"
        rc, status = NGC_ValidateRemoteRepo(token, org, repo, tag)

    if rc != 0:
        results["status"] = "Unable to validate '%s': %s" %(image_ref, status)
        return rc, results
    repo_info = status       # info on what to download

    # have a valid image, download without asking?
    if auto == False:
        # ask user to verify what they asked us to do?
        if local == True:
            prompt = "Use local image %s" %(image_ref)
        else:
            prompt = "Pull %s as user '%s' and use" %(image_ref, username)
        rc, status = PickAndSelect1(["n", "y"], prompt)
        if rc != 0:
            return rc, status    # ^C or some other abort
        if status != "y":        # user answered the question?
            results["status"] = "Canceled"
            return -1, results

    # sweet smell of success
    results["status"] = "Success"
    return 0, results

###############################################################################
# NGC_AuthenicateAndQueryImageRef
#
# The main purpose of this function is to return a validated image reference
# that the user has authorization to download, of the form 'nvcr.io/org/repo:tag'
# It returns the name of that image if successful, or and error code and message
# if not.
#
# The code can be given complete defaults for all the component, or can query
# any of them via ngc api calls, allowing the user to select what's avaliable.
#
# Filters on each of these is also privided -- these are regex strings to pass
# through to the user select lists only those items names that match. For example,
# the default filter on the repo names is "nvfw-.*" to only allow in known
# firmware update repos names.  The 'driveconst-dev' org is loaded with a lot of
# other container (repo) names that are not firmware updates
#
# One important thing -- For authenticaion, you MUST get org-specific token (based
# on your API key) pull org-specific information down. Grabbing the tag numbers of
# a repo requires this. But you don't know the org if you haven't figured out which
# one you want yet, and STILL need an org name to get a token so you can pull down
# the lsit of orgs.
#
# When we don't have a org name, a more general "nvidian" org is used in this case.
# Meaning that we may need to pull down two authentication tokens -- the first to
# grab the list of org names that your account has access to, with "nvidian" org
# scope, and then, once you have a valid org name, create another token using that
# specific org name.
#
# root:    base path of download. as of 10/2019, for NGC this will always be
#          'nvcr.io', but is passed into this function to make it more genric.
#          Not used here in any of this code, but is pre-pended to the org/repo:tag
#          return string to make the result a full proper docker image reference
#
# Returns  status, msg
#   status:                0   success, msg = valid image ref
#                         !0   failure, msg = some error text
#
def NGC_AuthenicateAndQueryImageRef(install_args, platform_name, root, org=None, org_filter=None,
                                    repo=None, repo_filter="nvfw-.*",
                                    tag=None, tag_filter=None, auto=False):

    rc, source = PickAndSelect1(["local", "remote"], "Select image source")
    if source == "local":

        # select from docker images local on the machine  I.E: 'docker images'

        rc, list_of_images, title = Docker_GetListOfLocalImages()     # get list of local docker images
        if rc != 0:
            return rc, list_of_images   # list_of_images is error message string

        # docker image may be prefixed with nvcr.io or other prefix, ignore that when filtering
        repo_filter = ".*" + repo_filter
        rc, image_ref = ShowListOfDockerImagesAndSelect1("Local Docker Image", list_of_images, title, repo_filter)
        if rc != 0:
            return rc, image_ref        # image is error message string

        # these are fields passed to image verification function - note we have
        # not logged into NGC yet
        # image_ref is '[nvcr.io/][org/]repo:tag' - break it up

        api_key  = None
        token    = None
        root, org, repo, tag = split_image_ref(image_ref)   # break out the pieces
        local    = True
        username = "localdocker"   # not actually needed in fetch but variable must be defined
    else:

        local   = False       # use this flag later durring validate

        # query NGC for org/repo:tag for image to download

        rc, api_key = Get_UsersAPIKey(platform_name, False)   # do not clear NGC key if asked before
        if rc != 0:
            return rc, api_key        # api_key is error message

        # check basic authenticaion of user -- not yet in a specificed org scope
        rc, token = NGC_AuthenticationStart(api_key)
        if rc != 0:
            return rc, token          # token is error message

        # do we have an org? -- if not, do preliminary validation, then grab
        # lists of orgs
        if org == None:
            image_ref = build_image_ref(root, "<org>", None, None)

            rc, list_of_orgs, title, username = NGC_AuthenticateUser(token)
            if rc != 0:
                return rc, list_of_org    # list_of_orgs is a error string if rc != 0

            rc, org = ShowListOfOrgsAndSelect1(image_ref, list_of_orgs, title, org_filter)
            if rc != 0:
                return rc, org

        # Have an org, but first, verify that user has access to it
        rc, list_of_orgs, title, username = NGC_AuthenticateUser(token, org)
        if rc != 0:
            return rc, list_of_orgs       # list_of_orgs is an error string if rc != 0

        # know user can access org, get another org-specific rights token
        rc, token = NGC_AuthenticationStart(api_key, org)
        if rc != 0:
            return rc, token

        # if no repo, Get list of repos and allow user to select 1
        if repo == None:          # nvcr.io path/name of repo
            image_ref = build_image_ref(root, org, "<repo>", None)

            rc, list_of_repos, title = NGC_GetListOfReposAtOrg(token, org)
            if rc != 0:
                return rc, list_of_repos

            rc, repo = ShowListOfReposAndSelect1(image_ref, list_of_repos, title, repo_filter)
            if rc != 0:
                return rc, repo

        # if no version tag, get list and allow user to select one
        if tag == None:
            image_ref = build_image_ref(root, org, repo, "<tag>")

            rc, list_of_tags, title = NGC_GetListOfTagsAndImageInfoForRepo(token, org, repo)
            if rc != 0:
                return rc, list_of_tags
            rc, tag = ShowListOfRepoTagsAndSelect1(image_ref, list_of_tags, title, tag_filter)
            if rc != 0:
                return rc, tag

    # validate if can access repo, and if auto=False ask user to verify
    # If success, returns a dictionary containing all values to customer
    rc, results = NGC_ValidateRepoAndAskIfPull(install_args, api_key, token, root, org, repo, tag, auto, local, username)
    return rc, results    # results is a dictionary, results.msg is set if any error

###############################################################################
# NGC_AuthenicateAndVerifyImageRef
#
# Simular to above, but takes a full image reference vs walking down the tree
# Does full user and image autheniticataion is same way.
#
# Will need to pull 'org' out of this string in order to authenicate user
# nvcr.io/
#
# With 'auto' == False, user will be prompted to verify download
#
# <platform> is used in prompt to tell use what box running on
#
# Returns: same as above
#
def NGC_AuthenicateAndVerifyImageRef(install_args, platform_name, image_ref, auto=False, local=False):
    api_key              = None
    token                = None
    root, org, repo, tag = split_image_ref(image_ref)   # break out the pieces

    # see if this is a local image -- no nvcr.io root or org is given
    # NOTE: ref's will have root/org on them  if they were been previously
    #       downloaded. If no specifically told it's local, check NGC authenication
    #       In that case, image will still be 'fetched' from nvcr.io, to refresh it
    #       if it's not the latest. Requires valid authentication to do this.
    #
    # NGC authenticaion of local image is ONLY skipped if it's no root/org is given
    # or caller specificly declares that the image is "local"

    if root == None and org == None:
        api_key  = None
        token    = None
        username = "localdocker"
        local    = True

    # if remote, need to authenticate user to nvsm first
    if local == False:
       # query NGC for org/repo:tag for image to download

        rc, api_key = Get_UsersAPIKey(platform_name, True)   # fresh ask for NGC API key (password)
        if rc != 0:
            return rc, api_key        # api_key is error message

        # check basic authenticaion of user -- not yet in a specificed org scope
        rc, token = NGC_AuthenticationStart(api_key)
        if rc != 0:
            return rc, token          # token is error message

        # Have an org, but first, verify that user has access to it
        rc, list_of_orgs, title, username = NGC_AuthenticateUser(token, org)
        if rc != 0:
            return rc, list_of_orgs   # list of orgs is a error message if rc != 0

        # know user can access org, get another org-specific rights token
        rc, token = NGC_AuthenticationStart(api_key, org)
        if rc != 0:
            return rc, token

    # validate if can access repo, and if auto=False ask user to verify
    # If success, returns a dictionary containing all values to customer
    rc, results = NGC_ValidateRepoAndAskIfPull(install_args, api_key, token, root, org, repo, tag, auto, local, username)
    return rc, results    # results is a dictionary, results.msg is set if any error

###############################################################################
# test code, if module directly started
#
if __name__ == "__main__":
    install_args = ["query", "test", "args"]         # args are an array
    rc, msg = NGC_AuthenicateAndQueryImageRef(install_args, "all", root="nvcr.io", org=None, repo=None, tag=None, auto=False)
    print(rc, msg)
    if rc >= 0:
        if type(msg) == dict:
            pp(msg)     # json data
        else:
            print(msg)  # simple text
    sys.exit(rc)

