# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.


import argparse
import re
import sys

import nvsm.command
import nvsm.log


class SetCommand(nvsm.command.Command):
    __nameValuePattern = re.compile(r'^(?P<name>[^=]+)=(?P<value>.*)$')

    def __init__(self, argv):
        super(SetCommand, self).__init__(argv=argv)

        self.__parse_arguments()

    def __parse_arguments(self):
        self.__argparser = argparse.ArgumentParser(
            prog='set',
            description='''Change the value of NVSM target properties''',
            add_help=False)
        self.__argparser.add_argument('-help', '-h', action='store_true', default=False,
                                      help='show this help message and exit')
        self.__argparser.add_argument('target', nargs='?', action='store', default='.',
                                      help='''The target address of the Managed Element . The
                target address can be relative to the current working target,
                or it can be absolute. Globbing of targets is not supported
                for this command.''')
        try:
            # TODO: Redirect the sys.stdout printed by the parser (for unit testing purposes)
            self.options, self.__remaining = self.__argparser.parse_known_args(
                self.argv[1:])
        except SystemExit as e:
            if e.code != 0:
                # TODO: Somehow raise exception with the error message printed by argparse
                raise nvsm.exceptions.InvalidArguments(
                    'Invalid set command arguments: {0}'.format(
                        ' '.join(self.argv[1:])))

        if self.options.help:
            return  # User requested help; nothing to parse

        # Treat "target" with equals sign as propertyname=value pair
        if '=' in self.options.target:
            self.__remaining.insert(0, self.options.target)
            self.options.target = '.'

        # Parse remaining argparse options as propertyname=value pairs
        self.__nameValuePairs = []
        if not self.__remaining:
            raise nvsm.exceptions.InvalidArguments(
                'set command requires propertyname=value pairs')
        for arg in self.__remaining:
            m = SetCommand.__nameValuePattern.match(arg)
            if not m:
                raise nvsm.exceptions.InvalidArguments(
                    'Expected propertyname=value pair: {0}'.format(arg))
            pair = (m.group('name'), m.group('value'))
            self.__nameValuePairs.append(pair)

    def run(self, context, file=sys.stdout):
        if self.options.help:
            self.__argparser.print_help(file=file)
            return
        nvsm.log.debug(
            'Running set command...\n'
            '    name: {name}\n'
            '    argv: {argv}\n'
            '    options: {options}',
            name=self.name, argv=self.argv, options=self.options)

        # Traverse the target hierarchy to find matching targets
        address = nvsm.targetAddress.TargetAddress(
            self.options.target, context=context)
        target = context.getTarget(address)
        # Make sure target supports "set" verb
        if not hasattr(target, 'set'):
            raise nvsm.exceptions.CommandError(
                'Target "{target}" does not support "set" verb'.format(
                    target=target.address))

        # Push array of name,value pari to Target class.Class decides how to handle it.
        if hasattr(target, 'set_array'):
            target.set(self.__nameValuePairs)
        else:
            # Call "set" verb for each given propertyname=value pair
            for pair in self.__nameValuePairs:
                target.set(name=pair[0], value=pair[1])

        # Show current value of each propertyname=value pair
        # (as per SMASH CLP specification)
        for pair in self.__nameValuePairs:
            name = pair[0]
        # Print properties affected by the change
        for p in target.properties:
            if p.name.replace("_", "").lower() not in [x[0].lower() for x in self.__nameValuePairs]:
                continue
            # Skip to show the PowerState and IndicatorLed as they do not reflect immediately
            if p.name.replace("_", "").lower() in ['powerstate', 'indicatorled']:
                continue
            print('{name}={value}'.format(
                name=p.name, value=p.value), file=file)
        print('', file=file)

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an NVSM "set" command.

        This method allows the NVSM Command factory pattern to determine if
        the SetCommand class implements the given command name.

        This method returns True if argv[0] is "set", case insensitive.
        Returns False otherwise.
        """
        if argv[0].lower() == 'set':
            return True
        return False


# Register with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command(SetCommand)
