# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.


import argparse
import sys
import fnmatch
import re
import subprocess
import signal

from nvsm.color import colorize
import nvsm.command
import nvsm.log
from nvsm.targetAddress import TargetAddress
import time


class ShowCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(ShowCommand, self).__init__(argv=argv)

        self.__parse_arguments()

    def __parse_arguments(self):
        self.__argparser = argparse.ArgumentParser(
            prog='show',
            description='''Display information about devices and other entities managed by NVSM''',
            add_help=False)
        self.__argparser.add_argument('-help', '-h', action='store_true', default=False,
                                      help='show this help message and exit')
        self.__argparser.add_argument('-level', '-l', metavar='LEVEL',
                                      action='store', default=1,
                                      help='''Specify the target depth level to which the show
                command will traverse the target hierarchy. The default value
                for LEVEL is 1, which means "the current target only".''')
        self.__argparser.add_argument('-display', '-d', metavar='CATEGORIES',
                                      action='store', default='all',
                                      type=nvsm.command.ColonSeparatedList,
                                      help='''Select the categories of information displayed about
                the given target. Valid values for CATEGORIES are
                'associations', 'targets', 'properties', 'verbs', and 'all'.
                The default value for CATEGORY is 'all'. Multiple values can be
                specified by separating those values with colon. Sub-arguments
                for properties are supported which are separated by comma with paranthesis as optional.''')
        self.__argparser.add_argument('-all', '-a', action='store_true', default=False,
                                      help='''Show data that are normally hidden. This includes OEM
                properties and OEM targets unique to NVSM.''')
        self.__argparser.add_argument('target', nargs='?', action='store', default='.',
                                      help='''The target address of the Managed Element to show. The
                target address can be relative to the current working target,
                or it can be absolute. Simple globbing to select multiple
                Managed Elements is also possible.''')
        self.__argparser.add_argument('where', nargs='?', action='store', default='all',
                                      help='''Using this argument, targets can be filtered based on
                the value of their properties. This can be used to quickly find
                targets with interesting properties. Currently this supports
                '==' and '!=' operations, which mean 'equal' and 'not equal'
                respectively. UNIX-style wildcards using '*' are also
                supported.''')
        try:
            # TODO: Redirect the sys.stdout printed by the parser (for unit testing purposes)
            self.options = self.__argparser.parse_args(self.argv[1:])
        except SystemExit as e:
            if e.code != 0:
                # TODO: Somehow raise exception with the error message printed by argparse
                raise nvsm.exceptions.InvalidArguments(
                    'Invalid show command arguments: {0}'.format(
                        ' '.join(self.argv[1:])))

        where_pattern = re.compile(
            r'^(?P<name>[^=!]+)(!|=)=(?P<value>[^=!]+)$')
        # swap target and where if misplaced
        if where_pattern.match(self.options.target):
            self.options.target, self.options.where = self.options.where, self.options.target
            if self.options.target == 'all':
                self.options.target = '.'
            if self.options.where == '.':
                self.options.where = 'all'
        # check for illegal where statement and back to default
        if not where_pattern.match(self.options.where):
            self.options.where = 'all'
        # Interpret the where statement as conditional expression
        w = where_pattern.match(self.options.where.lower())
        if '==' in self.options.where:
            def condition(p):
                if not fnmatch.fnmatch(p.name.lower(), w.group('name')):
                    return False
                if not fnmatch.fnmatch(str(p.value).lower(), w.group('value')):
                    return False
                return True
            self.options.where = condition
        elif '!=' in self.options.where:
            def condition(p):
                if not fnmatch.fnmatch(p.name.lower(), w.group('name')):
                    return False
                if fnmatch.fnmatch(str(p.value).lower(), w.group('value')):
                    return False
                return True
            self.options.where = condition

        # Parse level as integer
        try:
            self.options.level = int(self.options.level)
        except ValueError:
            if self.options.level != 'all':
                raise nvsm.exceptions.InvalidArguments(
                    'Invalid -level value "{0}"'.format(self.options.level))
        else:
            if self.options.level <= 0:
                raise nvsm.exceptions.InvalidArguments(
                    'Invalid -level value "{0}"'.format(self.options.level))

        # Validate the categories passed to -display
        self.options.display = [x.lower() for x in self.options.display]

        self.display_args = {}
        subargs_pattern = re.compile(r'^(?P<name>[^=]+)=(?P<value>[^=]+)$')
        for c in self.options.display:
            subargs = subargs_pattern.match(c)
            if subargs:
                self.display_args[subargs.group('name')] = subargs.group(
                    'value').strip('()').split(',')
            else:
                self.display_args[c] = 'all'

        for c in self.display_args:
            if c not in [
                    'aliases',
                    'associations',
                    'targets',
                    'properties',
                    'verbs',
                    'all']:
                raise nvsm.exceptions.InvalidArguments(
                    'Unknown display category "{0}"'.format(c))
        self.options.display = set(self.options.display)

    def cancel_handler(self, signum, frame):
        sys.exit(0)

    def run(self, context, file=sys.stdout):

        # Bug 200643113
        # Trap broken pipe signal and exit gracefully
        signal.signal(signal.SIGPIPE, self.cancel_handler)

        if self.options.help:
            self.__argparser.print_help(file=file)
            return
        nvsm.log.debug(
            'Running show command...\n'
            '    name: {name}\n'
            '    argv: {argv}\n'
            '    options: {options}',
            name=self.name, argv=self.argv, options=self.options)

        # Traverse the target hierarchy to find matching targets
        address = nvsm.targetAddress.TargetAddress(
            self.options.target, context=context)
        targets = list(context.findTargets(
            search_address=address,
            level=self.options.level,
            showall=self.options.all))

        # Sort targets by address, with integers sorted numerically
        def get_targets(t):
            try:
                targets = t.targets
                return targets
            except nvsm.exceptions.BackendError as e:
                nvsm.log.error('{0} for target address {1}', e, t.address)
                return []

        def get_properties(t):
            try:
                properties = t.properties
                return properties
            except nvsm.exceptions.BackendError as e:
                nvsm.log.error('{0} for target address {1}', e, t.address)
                return []

        targets.sort(key=lambda a: a.address.canonicalPath)

        # Show information for the specified display categories
        def is_display(category):
            if 'all' in self.display_args:
                return True
            return category in self.display_args

        def has_desired_property(target, where):
            if where == 'all':
                return True
            return any(map(self.options.where, get_properties(properties)))

        def filter_properties(properties):
            if not is_display('properties'):
                return []
            if 'properties' not in self.display_args:
                return properties
            if self.display_args['properties'] == 'all':
                return properties

            def is_display_property(p):
                return any([fnmatch.fnmatch(p.name.lower(), name) for name in self.display_args['properties']])
            return list(filter(is_display_property, properties))

        # we should only retry at the top level of each path
        # i.e. /redfish/v1/chassis or /redfish/v1/managers
        # this corresponds to path_list ["chassis"] and ["managers"] respectively
        # do not retry when further down in path
        # i.e. /redfish/v1/chassis/policy -> ["chassis", "policy"]
        def should_retry(retry_list, path_list):
            if len(path_list) != 1:
                return False
            for path in retry_list:
                if path == path_list[0]:
                    return True
            return False

        max_timeout, delay_time = 100, 20
        retry_count = int(max_timeout / delay_time)
        retry_list = ["chassis", "managers"]

        # Convert text to int
        def string_to_int(text):
            return int(text) if text.isdigit() else text

        # Only valid in case of DRIVES
        # Sorting function for drive slots
        # If the target has a property 'PhysicalLocation_info',
        # We will sort the targets using the value of that property
        # This is valid only for drives where,
        # PhysicalLocation_Info = 'U.2_SlotX/M.2_SlotX' or '8:0'
        def sort_drive_slot(target):
            '''
            alist.sort(key=natural_keys) sorts in human order
            http://nedbatchelder.com/blog/200712/human_sorting.html
            (See Toothy's implementation in the comments)
            '''
            for p in filter_properties(get_properties(target)):
                if p.name == 'PhysicalLocation_Info':
                    return [ string_to_int(text) for text in re.split(r'(\d+)', p.value) ]
            return []

        def runshellcmd(cmdstr, verbose=False):
            if verbose:
                print("Command Run: %s\n" % (cmdstr))
            sp = subprocess.Popen(cmdstr, stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT, shell=True)
            (cmdout, cmderr) = sp.communicate()
            cmdret = sp.wait()
            cmdout = cmdout.decode('utf-8').rstrip('\n')
            if verbose:
                print("Command Return: %d\n" % (cmdret))
            return cmdret, cmdout

        ret_dmidecmd, model_dmidecmd = runshellcmd(
        "dmidecode -s system-product-name", False)

        # Sort the list of targets according to drive slots
        if targets is not None:
            targets.sort(key=sort_drive_slot)

        # iterate over targets and print out sections for properties, targets, verbs
        for i, target in enumerate(targets):
            if not has_desired_property(target, self.options.where):
                continue
            # i.e. /managers or /chassis
            path_list = target.address.canonicalPath
            print('{0}'.format(target.address), file=file)
            properties = filter_properties(get_properties(target))
            if properties:
                print('Properties:', file=file)
            for p in properties:
                print('    {0} = '.format(p.name), end='', file=file)
                with colorize(p.color):
                    if (p.value != None) and (p.value.__class__.__name__ == "str") and (p.value == ""):
                        print("<NOT_SET>", file=file)
                    else:
                        print(str(p.value), file=file)
            t1 = target
            for j in range(retry_count):
                if not is_display("targets"):
                    break
                # getTarget is a list of targets to print out
                # if it is empty, something went wrong and we should retry
                getTarget = get_targets(t1)
                if len(getTarget) == 0 and should_retry(retry_list, path_list):
                    target_str = target.address.canonicalAddress
                    nvsm.log.warning(
                        "Error fetching NVSM Target {0}. Retrying...".format(target_str))
                    context.refresh()
                    time.sleep(delay_time)
                    address = nvsm.targetAddress.TargetAddress(
                        self.options.target, context=context)
                    targs = list(context.findTargets(
                        search_address=address,
                        level=self.options.level,
                        showall=self.options.all))
                    t1 = targs[i]
                else:
                    print('Targets:', file=file)
                    # TODO CLEANUP HACK: NVBug 3153171
                    # Only one target for DGX Station
                    if ret_dmidecmd == 0 and "dgx station" == model_dmidecmd.lower():
                        print('    versions')
                        break
                    subtargets = sorted(
                        getTarget, key=lambda a: a.address.canonicalPath)
                    for t in subtargets:
                        print('    {0}'.format(t.name), file=file)
                    break

            if is_display('verbs'):
                print('Verbs:', file=file)
                verbs = sorted(['cd', 'show'] + target.verbs)
                for v in verbs:
                    print('    {0}'.format(v), file=file)
            if is_display('aliases') and target.address == TargetAddress('/'):
                print('Aliases:', file=file)
                # TODO CLEANUP HACK: NVBug 3153171
                # Only two aliases for DGX Station
                if ret_dmidecmd == 0 and "dgx station" == model_dmidecmd.lower():
                        print('    dump health')
                        print('    show health')
                        break
                aliases = nvsm.command.Command.list_aliases()
                for a in aliases:
                    print('    {0}'.format(a), file=file)
            print('', file=file)

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an NVSM "show" command.

        This method allows the NVSM Command factory pattern to determine if
        the ShowCommand class implements the given command name.

        This method returns True if argv[0] is "show", case insensitive.
        Returns False otherwise.
        """
        if argv[0].lower() == 'show':
            return True
        return False


# Register with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command(ShowCommand)

