# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.




import copy
import os
import signal
import subprocess
import sys
import re
import logging

import nvsm.command
import nvsm.interpreter

import argparse

class ShowHealthCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(ShowHealthCommand, self).__init__(argv=argv)

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an alias for "show health".

        This method allows the NVSM Command factory pattern to determine if
        the "ShowHealthCommand" class implements the given command.

        This method returns True if the first two arguments of argv look like
        "show health" case insensitive. Returns False otherwise.
        """
        aliases = [
            [ 'show', 'health' ],
            [ 'show', 'nvhealth' ] ]
        return [x.lower() for x in argv[:2]] in aliases

    def __get_callhome_policy(self, context):
        callhome_enabled = False
        callhome_email = ""
        try:
            response = context._queryBackend("/nvsm/v1/Policy/Global")
            callhome_policy = response.get('Policy')[0].get('CallHome')
            callhome_enabled = callhome_policy.get('enable')
            callhome_email = callhome_policy.get("email_contact", "")
            email_pattern = re.compile(r'^[^@]+@[^@]+\.[^@]+$')
            if callhome_enabled and not email_pattern.match(callhome_email):
                logging.warning(f"Callhome Invalid email address: {callhome_email}")
                callhome_email = ""
        except:
            logging.warning("Fetch callhome policy failed.")

        return callhome_enabled, callhome_email

    def __get_callhome_description(self, context, bom_check_collection_summary=None):
        callhome_description = "Callhome triggered by nvsm show health on unhealthy status."

        # get health status
        # instead of getting the api-gateway /nvsm/v1/BOMChecks which may not be
        # the latest, get the recent runtime BomCheckCollectionSummary which can
        # 1. come from standalone [uncached]
        # 2. come from client [cached]
        try:
            healthy       = bom_check_collection_summary.get("Healthy", 0)
            unhealthy     = bom_check_collection_summary.get("UnHealthy", 0)
            unknown       = bom_check_collection_summary.get("UnKnown", 0)
            informational = bom_check_collection_summary.get("Informational", 0)
            disabled      = bom_check_collection_summary.get("Disabled", 0)
            skipped       = bom_check_collection_summary.get("Skipped", 0)
            total_checks  = bom_check_collection_summary.get("TotalChecks", 0)
            callhome_description += f" {healthy} healthy, {unhealthy} unhealthy, {unknown} unknown, {informational} informational, {disabled} disabled and {skipped} skipped out of {total_checks} total health checks."
        except:
            logging.warning(f"Error getting health results: {bom_check_collection_summary}")

        return callhome_description

    def __trigger_callhome(self, context, bom_check_collection_summary=None):
        callhome_enabled, callhome_email = self.__get_callhome_policy(context)

        callhome_success_banner = '''Callhome successfully triggered.
Check the status with: nvsm show /callhome
'''
        callhome_fail_banner = '''Callhome trigger failed.'''
        callhome_running_banner = '''Callhome is in progress, wait for it to complete.
Check the status with: nvsm show /callhome
'''
        email_option = ""
        if callhome_email:
            email_option = f"email={callhome_email}"
        callhome_cmd = f'nvsm set /callhome trigger=true {email_option} description="{self.__get_callhome_description(context, bom_check_collection_summary)}" > /dev/null'

        if not callhome_enabled:
            logging.warning("Callhome policy not enabled, not triggering callhome.")
        else:
            callhome_state = "Idle"
            try:
                callhome_response = context._queryBackend("/nvsm/v1/CallHome")
                callhome_state = callhome_response.get('Op').get('State')
            except:
                pass

            if callhome_state.lower() == "running":
                logging.warning(callhome_running_banner)
                return

            rc = os.system(callhome_cmd)
            exit_code = rc >> 8
            if not exit_code:
                print(callhome_success_banner)
            else:
                logging.error(callhome_fail_banner)

    def check_if_platform_support_legacy_health_only(self):
        #FIXME: Add CGX-SM to list of platform supporting only legacy show health   
        platform_supporting_legacy_health = ["DGX Station"]
        cmdstr = "dmidecode -s system-product-name | grep -x '{}'"
        for platform in platform_supporting_legacy_health:
            dmidecode = subprocess.Popen(cmdstr.format(platform), stdout=subprocess.PIPE,
                          stderr=subprocess.STDOUT, shell=True)
            (cmdout, cmderr) = dmidecode.communicate()
            if dmidecode.returncode:
                continue
            else:
                return True
        return False

    def run(self, context, file=sys.stdout):
        # Suppress deprecation warning messages from nvhealth
        env = copy.copy(os.environ)
        env['CALLED_FROM_NVSM_CLI'] = 'true'

        parser = argparse.ArgumentParser(
                description="NVIDIA System Information",
                usage='nvsm show health [help] [-o OUTPUT] [--with_callhome]',
                add_help=False
                )
        parser.add_argument('help',
                default='*',
                nargs='?',
                help='show this help message and exit')
        parser.add_argument('-o', '--output', nargs='?',
                help='File path in which to write output.')
        parser.add_argument('-wch', '--with_callhome',
                action='store_true',
                help='Dial callhome on unhealthy status.')
        parser.add_argument('-lgh','--legacy_health',
                action='store_true',
                help='Run old nvsm health.')
        parser.add_argument('--cached',
                action='store_true',
                help='Run nvsm show health with cached result')

        args, _ = parser.parse_known_args(self.argv[2:])

        if args.help != '*':
            parser.print_help()
            return
        
        if args.legacy_health and args.cached:
            logging.error("--legacy_health and --cached option can not be used together")
            return
        if self.check_if_platform_support_legacy_health_only() and args.cached:
            logging.error("--cached can not be used")
            return

        out_file = None
        if args.output:
            out_file = args.output

        bom_check_collection_summary = None
        run_legacy_health = False
        if args.legacy_health or self.check_if_platform_support_legacy_health_only():
            run_legacy_health = True

        if run_legacy_health:
            from nvsm.backend.health import run_show_health
            exit_code = run_show_health(out_file)
        else:
            #Handle new nvsm health code
            from nvsm.backend.health import run_new_show_health
            exit_code, bom_check_collection_summary = run_new_show_health(out_file, args.cached)

        if args.with_callhome and exit_code:
            self.__trigger_callhome(context, bom_check_collection_summary)

# Register as a command alias with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command_alias(ShowHealthCommand)
