# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.




import sys

import nvsm.command
from nvsm.command.show import ShowCommand

class ShowStorageCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(ShowStorageCommand, self).__init__(argv=argv)

        # Construct the "show" command that implements this command alias
        interpreter = nvsm.interpreter.Interpreter()
        show_command = 'show -level all -display properties /systems/*/storage'
        self.__showCommand = interpreter.evaluate(show_command)

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an alias for "show storage".

        This method allows the NVSM Command factory pattern to determine if
        the "ShowStorageCommand" class implements the given command.

        This method returns True if argv looks like "show storage inventory",
        case insensitive. Returns False otherwise.
        """
        aliases = [
            [ 'show', 'storage' ] ]
        return [x.lower() for x in argv] in aliases

    def run(self, context, file=sys.stdout):
        return self.__showCommand.run(context=context, file=file)

# Register as a command alias with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command_alias(ShowStorageCommand)


class ShowStorageDrivesCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(ShowStorageDrivesCommand, self).__init__(argv=argv)

        # Construct the "show" command that implements this command alias
        show_argv = [ 'show',
                      '-level', 'all',
                      '-display', 'properties',
                      '/systems/*/storage/drives/*' ]
        self.__showCommand = ShowCommand(argv=show_argv)

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an alias for "show storage drives".

        This method allows the NVSM Command factory pattern to determine if
        the "ShowStorageDrivesCommand" class implements the given command.

        This method returns True if argv looks like "show storage drives" or
        "show drives", case insensitive. Returns False otherwise.
        """
        aliases = [
            [ 'show', 'storage', 'drives' ],
            [ 'show', 'drives' ] ]
        return [x.lower() for x in argv] in aliases

    def run(self, context, file=sys.stdout):
        return self.__showCommand.run(context=context, file=file)

# Register as a command alias with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command_alias(ShowStorageDrivesCommand)


class ShowStorageVolumesCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(ShowStorageVolumesCommand, self).__init__(argv=argv)

        # Construct the "show" command that implements this command alias
        show_argv = [ 'show',
                      '-level', 'all',
                      '-display', 'properties',
                      '/systems/*/storage/volumes/*' ]
        self.__showCommand = ShowCommand(argv=show_argv)

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an alias for "show storage volumes".

        This method allows the NVSM Command factory pattern to determine if
        the "ShowStorageVolumesCommand" class implements the given command.

        This method returns True if argv looks like "show storage volumes" or
        "show volumes", case insensitive. Returns False otherwise.
        """
        aliases = [
            [ 'show', 'storage', 'volumes' ],
            [ 'show', 'volumes' ] ]
        return [x.lower() for x in argv] in aliases

    def run(self, context, file=sys.stdout):
        return self.__showCommand.run(context=context, file=file)

# Register as a command alias with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command_alias(ShowStorageVolumesCommand)
