# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.




import argparse
import datetime
import json
import signal
import sys
import time
from contextlib import contextmanager

from nvsm.color import colorize
import nvsm.command
import nvsm.log

@contextmanager
def ignore_keyboard_interrupts(uninterruptible=True):
    if not uninterruptible:
        yield None
    else:
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        try:
            yield None
        finally:
            # Restore default signal handler
            signal.signal(signal.SIGINT, signal.SIG_DFL)

class StartCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(StartCommand, self).__init__(argv=argv)

        self.__parse_arguments()

    def __parse_arguments(self):
        self.__argparser = argparse.ArgumentParser(
                prog='start',
                description='''Start a job managed by NVSM''',
                add_help=False)
        self.__argparser.add_argument('-help', '-h', action='store_true', default=False,
                help='show this help message and exit')
        self.__argparser.add_argument('-noblock',
                action='store_true', default=False,
                help='''Do not wait for the running job to finish. By default, the
                        "start" verb will wait until the job finishes
                        (synchronous operation).''')
        self.__argparser.add_argument('-force', '-f',
                action='store_true', default=False,
                help='''Start this job without prompting with warning. By
                        default, the "start" verb will prompt the user before
                        starting certain jobs. With the "-force" argument
                        given, the job is started immediately without
                        prompting.''')
        self.__argparser.add_argument('-quiet',
                action='store_true', default=False,
                help='''Do not print job progress information. This can be
                        useful for use inside scripts or for slow
                        terminals.''')
        self.__argparser.add_argument('-timeout',
                action='store', default=None,
                help='''If no progress has been made, the time (in seconds) to wait for the
                        job before giving up. The default behavior is no timeout.''')
        self.__argparser.add_argument('target', nargs='?', action='store', default='.',
                help='''The target address of the Managed Element . The
                        target address can be relative to the current working
                        target, or it can be absolute. Globbing of targets is
                        not supported for this command.''')
        try:
            # TODO: Redirect the sys.stdout printed by the parser (for unit testing purposes)
            self.options = self.__argparser.parse_args(self.argv[1:])
        except SystemExit as e:
            if e.code != 0:
                # TODO: Somehow raise exception with the error message printed by argparse
                raise nvsm.exceptions.InvalidArguments(
                        'Invalid show command arguments: {0}'.format(
                            ' '.join(self.argv[1:])))

    def run(self, context, file=sys.stdout):
        if self.options.help:
            self.__argparser.print_help(file=file)
            return
        nvsm.log.debug(
                'Running start command...\n' \
                '    name: {name}\n' \
                '    argv: {argv}\n' \
                '    options: {options}',
                name=self.name, argv=self.argv, options=self.options)

        if self.options.timeout:
            try:
                self.options.timeout = int(self.options.timeout)
                if self.options.timeout < 0:
                    raise RuntimeError
            except:
                raise nvsm.exceptions.CommandError(
                    'Timeout in seconds must be a positive integer')

        # Traverse the target hierarchy to find matching target
        address = nvsm.targetAddress.TargetAddress(
                self.options.target, context=context)
        target = context.getTarget(address)
        # Make sure target supports "start" verb
        if not hasattr(target, 'start'):
            raise nvsm.exceptions.CommandError(
                'Target "{target}" does not support "start" verb'.format(
                target=target.address))
        # Start the job associated with this target
        job = target.start(skip_prompt=(self.options.force), wait=(not self.options.noblock))
        if not job:
            return
        # XXX: Get time info from the API server itself?
        print('{target} started at {now}'.format(
                target=target.address,
                now=datetime.datetime.now()), file=file)
        # If requested, wait for the job to finish
        if not self.options.noblock:
            self.__waitForJob(job=job, file=file, timeout=self.options.timeout)
        print(file=file)

    def __waitForJob(self, job, file=sys.stdout, timeout=None):
        last_state = 'Unknown'
        last_progress = 0.0
        last_update_time = time.time()
        with ignore_keyboard_interrupts(job.uninterruptible):
            with nvsm.progress.ProgressBar(file=file) as progress_bar:
                try:
                    # Monitor progress of job as JSON lines format
                    # NOTE: The streaming API will block as each line is read,
                    #       which should keep the progress bar update frequency
                    #       within reasonable limits.
                    for l in job.status_stream:
                        try:
                            blob = json.loads(l)
                        except:
                            nvsm.log.debug('Could not decode job status: {}', l.replace("\n", r'\n'))
                            continue
                        last_state=blob.get('State', 'Unknown')
                        error_state = blob.get('State', '').lower() in [ 'error', 'warning', 'Exception' ]
                        progress = blob.get('Progress', 0.0)
                        progress_bar.update(
                                percent=progress,
                                message=blob.get('Message', '...'),
                                error=error_state)
                        if progress != last_progress:
                            last_update_time = time.time()
                            last_progress = progress
                        if timeout and ((time.time() - last_update_time) > timeout):
                            progress_bar.update(
                                    percent=progress,
                                    message='Timeout reached.',
                                    error='timeout')
                            raise nvsm.exceptions.CommandError('Timeout reached.')
                except KeyboardInterrupt:
                    print('\nKeyboard interrupt. (job running in background)', file=file)
                    return
        print('Status: ', end='', file=file)
        if last_state.lower() in [ 'done', 'finished', 'success', 'completed' ]:
            color = 'green'
        else:
            color = 'red'
        with colorize(color):
            print(last_state, file=file)

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an NVSM "start" command.

        This method allows the NVSM Command factory pattern to determine if
        the SetCommand class implements the given command name.

        This method returns True if argv[0] is "start", case insensitive.
        Returns False otherwise.
        """
        if argv[0].lower() == 'start':
            return True
        return False

# Register with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command(StartCommand)
