# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
import nvsm.command
from time import sleep
from json import dumps
from ast import literal_eval
import datetime
import os
import sys
from nvsm.color import colorize
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()


class StressCommand(nvsm.command.Command):

    def __init__(self, argv):
        super(StressCommand, self).__init__(argv=argv)

        self.usage = "usage: sudo nvsm stress-test [--usage] [--force] [--no-prompt] [<test>...] [DURATION]\n"

        self.banner = "NVIDIA System Management Stress Testing\n"

        self.optional = {"--usage": "show this help message and exit",
                         "--no-prompt": "skip warning prompt before starting stress tests",
                         "--force": "run stress tests anyways even if system alerts have not yet been cleared",
                         "<test>...": "list of space separated tests; available tests:",
                         "DURATION": "provide duration for tests to run in seconds, defaults to 20 minutes"
                         }
        self.tests = ["gpu", "cpu", "memory", "storage", "nic (only on supported platforms)", "all (default)"]

        self.examples = ["sudo nvsm stress-test --no-prompt gpu cpu 90", "sudo nvsm stress-test storage memory", "sudo nvsm stress-test --usage",
                         "sudo nvsm stress-test 600", "sudo nvsm stress-test"]

        self.warning = ("WARNING:\n"
                        "These stress tests are potentially disruptive and can interfere with other jobs you might be running.\n"
                        "Make sure to be running as few jobs as possible before proceeding.\n"
                        "(Pass --no-prompt option in command next time to suppress this prompt)\n"
                        "If running GPU Stress Test, expect test to take longer than countdown duration.\n"
                        "The NIC (network interface card) Stress Test is only available on DGX-1, DGX-2, and Luna Systems.\n"
                        "If it is run on non-supported platforms, results for this component will be blank.\n"
                        )

        self.alertsMsg = ("There are currently system alerts that have not been cleared.\n"
                          "Stress Testing will not be done.\n"
                          "Internal failures may occur, but can run tests anyways by passing --force flag.\n"
                          )

        self.troubleshootMsg = ("Tests failed to start.\n"
                                "Please run 'nvidia-smi' and 'dcgmi diag -r diagnostic,pcie' from the shell.\n"
                                "If either of these fail to run or produce an error, stress-testing will fail to function properly.\n"
                                "Also ensure that all nvsm plugins are active: 'sudo systemctl | grep nvsm-plugin'.\n"
                                )

        self.migMsg = ("Stress tests will not be run if MIG is enabled on the system.\n"
                       "Run 'nvidia-smi --query-gpu=mig.mode.current --format=csv,noheader' to check MIG status.\n"
                      )
   
        self.kvmMsg = ("Stress tests will not be run if KVM is enabled on the system.\n"
                       "Run 'dmesg | grep KVM' to see if KVM has been enabled.\n"
                      )

        # to be populated later by PUT request response so we don't have to hardcode here in CLI
        self.additionalTime = 0
        self.defaultDuration = 0
        self.logName = ""

        # keep a data structure to buffer the exact text reported by CLI
        # when tests are done, write to the logs file in addition to current logged data
        self.completeReport = []

        # the exact stress-test command run from the shell
        self.stressCommand = ""

    # prints out help screen
    def helpScreen(self):
        print(self.usage)
        print(self.banner)
        print("optional arguments:")
        fmt = "\t{:<15}  {}"
        for opt in self.optional:
            print(fmt.format(opt, self.optional[opt]))
            if opt == "<test>...":
                for test in self.tests:
                    print(fmt.format("", test))
        print("\nadditional examples:")
        for ex in self.examples:
            print("\t" + ex)

    def byteDecode(self, x):
        return x if not isinstance(x, bytes) else x.decode("utf-8")

    # shows the user a countdown to approximate progress on stress test
    def countdown(self, t):
        self.completeReport.append(
            "\nTest Duration: {} Seconds\n\n".format(str(t)))
        try:
            print()
            for i in range(t, -1, -1):
                sys.stdout.write(
                    "\rApproximate Time Remaining: {}     ".format(i))
                sys.stdout.flush()
                sleep(1)
            sys.stdout.write("\nPreparing results ...\n\n")
        except (KeyboardInterrupt, SystemExit):
            return

    # helper method to recursively fix json responses whose keys and vals are unicode
    def unicodeFix(self, obj):
        if not isinstance(obj, dict):
            if isinstance(obj, list):
                return obj
            return str(obj)
        return dict((str(k), self.unicodeFix(v)) for k, v in obj.items())

    # helper method to check if any alerts are present prior to running test
    def alertsPresent(self, context):
        try:
            response = self.unicodeFix(context._queryBackend(
                path="/nvsm/v1/Alerts", method="GET"))
            return len(list(filter(lambda x: x["severity"] != "Cleared" and x["severity"] != "Warning", response["Alerts"]))) > 0
        except Exception as e:
            return True

    # helper method for colored prints to command line
    def printColor(self, text, color):
        with colorize(color):
            print(text)

    # before reporting system metrics from json, ensure that all numeric fields updated correctly
    # occasional issue: fields not populated yet and we try to read an empty field
    def fieldsFilled(self, response):
        numericFields = ["PreTest", "Min", "Max", "Avg"]
        for key in response:
            inner = response[key]
            if isinstance(inner, dict) and numericFields[0] in inner:
                for nKey in numericFields:
                    if inner[nKey] == "":
                        return False
        return True

    # return True for done, False for something went wrong
    def waitCompletion(self, context, response, leeway):
        while 1:
            if leeway <= 0:
                return False, None
            if response["TimeEnded"] != "" and self.fieldsFilled(response):
                return True, response
            response = self.unicodeFix(context._queryBackend(
                path="/nvsm/v1/Stress", method="GET"))
            leeway -= 1
            sleep(1)

    # print out in table format the results of stress test
    def reportStressTestsResults(self, response):
        fmt = "{:<15}{:<15}{:<15}"
        resultsHeader, columnLabels = "Stress Test Results Summary", fmt.format(
            "Test", "Status", "Result")
        self.printColor(resultsHeader, "blue")
        self.printColor(columnLabels, "blue")
        self.completeReport.append(resultsHeader + "\n")
        self.completeReport.append(columnLabels + "\n")
        testResults = []
        for key in response:
            if isinstance(response[key], dict) and "Errors" in response[key]:
                res = response[key]
                testResults.append(fmt.format(
                    key, res["Status"], res["Result"]))
        for t in sorted(testResults):
            self.completeReport.append(t + "\n")
            print(t)

    # print out in table format the system metrics recorded during stress test
    def reportSystemMetrics(self, response):
        fmt, fmtNumeric = "{:<15}" * 6,  "{:<15}" + "{:<15.2f}" * 4 + "{:<15}"
        fmtNonNumeric = "{:<15}" + "{:<15}" * 4 + "{:<15}"
        metricsHeader = "\nSystem Resource Metrics"
        columnLabels = fmt.format(
            "Component", "Pre-test", "Min", "Avg", "Max", "Unit")
        self.printColor(metricsHeader, "blue")
        self.printColor(columnLabels, "blue")
        self.completeReport.append(metricsHeader + "\n")
        self.completeReport.append(columnLabels + "\n")
        metrics = []
        for key in response:
            if isinstance(response[key], dict) and "Min" in response[key]:
                m = response[key]
                pre, minm, maxm, avg, unit = m["PreTest"], m["Min"], m["Max"], m["Avg"], m["Unit"]
                # check for NA as the current platform might not support this field being populated
                # any of the fields can be checked (does not matter)
                if pre == "NA":
                    metrics.append(fmtNonNumeric.format(
                        key, pre, minm, avg, maxm, unit))
                else:
                    metrics.append(fmtNumeric.format(key, float(
                        pre), float(minm), float(avg), float(maxm), unit))
        for each in sorted(metrics, key=lambda x: x if "_" in x else "z"+x[0]):
            self.completeReport.append(each + "\n")
            print(each)
        print()

    def writeCompleteReport(self):
        with open(self.logName, "a") as logFile:
            logFile.write("Command run: {}\n\n".format(self.stressCommand))
            for line in self.completeReport:
                logFile.write(line)


    # GPU/PCIe tests take variable lengths of time on different platforms
    # this method is to account for that
    def determineDuration(self, duration):
        return duration + self.additionalTime


    # checks for invalid arguments and returns the duration if any, else ""
    # empty string "" will tell APIs to use default duration in PDF
    def invalidArgsCheck(self, args):
        errStr = "\nInvalid Argument(s). Run 'sudo nvsm stress-test --usage' for help."
        valid = ["gpu", "cpu", "memory", "storage", "nic", "--usage",
                 "--no-prompt", "--force", "stress-test", "all"]
        # means command was "sudo nvsm stress-test"
        if len(args) == 1:
            return True, 0
        dur = 0
        # number of numeric arguments seen, should only be 1 max for the duration
        numCount = 0
        for i, arg in enumerate(args):
            if arg not in valid:
                if arg.isdigit():
                    if numCount >= 1:
                        self.printColor(errStr, "red")
                        return False, 0
                    dur = arg
                    numCount += 1
                else:
                    self.printColor(errStr, "red")
                    return False, 0
        return True, dur

    def run(self, context, file=sys.stdout):
        banner = "{} NVSM STRESS {}\n".format("#" * 30, "#" * 30)
        # command: sudo nvsm stress-test
        # args = ["stress-test"] (everything after "sudo nvsm" is an argument)
        args = [i.lower().strip() for i in self.argv]
        self.stressCommand = "sudo nvsm {}".format(" ".join(args))
        # this is where we do a check for invalid arguments
        valid, dur = self.invalidArgsCheck(args)
        if not valid:
            return
        if "--usage" in args:
            self.helpScreen()
            return
        if self.alertsPresent(context) and "--force" not in args:
            print(self.alertsMsg)
            return
        footer = "#" * len(banner)

        print(banner)
        if "--no-prompt" not in args:
            self.printColor(self.warning, "red")
            proceed = self.byteDecode(
                input("Are you sure you want to proceed [y/n]? ").lower().strip())
            if proceed != "y":
                return
            print()

        def migCheck():
            # MIG check: stress tests should not be run on systems with MIG enabled
            # if the output contains any "Enabled" lines, then MIG is enabled
            cmd = "nvidia-smi --query-gpu=mig.mode.current --format=csv,noheader"
            output = os.popen(cmd).read().split("\n")
            for line in output:
                if line.strip() == "Enabled":
                    print(self.migMsg)
                    return True
            return False


        def kvmCheck():
            # KVM check: stress tests should not be run on systems in KVM mode
            cmd = "dmesg | grep KVM"
            output = os.popen(cmd).read().strip()
            if len(output) > 0:
                print(self.kvmMsg)
                return True
            return False

        # see NVBug 200708253 
        # first stress-test run always results in "Not Found" error code
        # workaround is to POST mock run where nothing is stressed
        # in order to avoid this issue
        def mockRun():
            print("Preparing NVSM for Stress Test. Please Wait ...\n")
            reqBody = {"GPU": False, "CPU": False, "Memory": False, "Storage": False, "NIC": False, "Duration": 10}
            try:
                context._queryBackend(
                    path="/nvsm/v1/Systems/1/Stress", method="POST", body=reqBody)
            except Exception as e:
                print(e)
            sleep(10)

        # Now lets see what all tests user is interested in
        if "gpu" in args or "cpu" in args or "storage" in args or "memory" in args or "nic" in args or "all" in args:  # user wants selected tests only
            gpu, cpu, stor, mem, nic = False, False, False, False, False
            if "gpu" in args:
                gpu = True
            if "cpu" in args:
                cpu = True
            if "storage" in args:
                stor = True
            if "memory" in args:
                mem = True
            if "nic" in args:
                nic = True
            if "all" in args:
                gpu, cpu, stor, mem, nic = True, True, True, True, True
        else:  # user doesnt care, so select all tests to run
            gpu, cpu, stor, mem, nic = True, True, True, True, True

        mockRun()
        try:
            # POST Stress test request
            reqBody = {"GPU": gpu, "CPU": cpu, "Memory": mem,
                    "Storage": stor, "NIC" : nic, "Duration": int(dur)}
            postResponse = context._queryBackend(
                path="/nvsm/v1/Systems/1/Stress", method="POST", body=reqBody)

            if gpu:
                print("Stressing GPU : Multiplying matrices on GPUs")
            if cpu:
                print("Stressing CPU : Multiplying matrices on CPUs")
            if mem:
                print("Stressing Memory : Mapping pages into memory")
            if stor:
                print("Stressing Storage : Performing disk I/O operations")
            if nic:
                print("Stressing Network Cards : Performing loopback bandwidth stress")

            print("\n")

            prevTimeLeft = 0
            while True:
                # GET Stress test status
                getResponse = self.unicodeFix(context._queryBackend(
                    path="/nvsm/v1/Systems/1/Stress", method="GET", forceRefresh=True))
                self.logName = getResponse["LogName"]
                if getResponse['TimeEnded'] == "0001-01-01T00:00:00Z":
                    # print remaining time
                    # need to account for negative in CLI
                    timeLeft = getResponse['ApproxTimeLeft']
                    timeLeft = timeLeft if int(timeLeft) > 0 else str(0)
                    prevTimeLeft = int(timeLeft)
                    print("Approximate Time Remaining: {}     ".format(timeLeft), end="\r", flush=True)
                    # Avoid bombarding API gateway with calls causing intermittent crashes
                    # for longer running jobs (i.e. > 10 min) poll less frequently
                    # see NVBug 200679053 
                    sleep(10 if int(dur) > 600 else 5)
                    continue
                else:
                    # print results and make sure that counter stops at 0
                    if prevTimeLeft > 0:
                        print("Approximate Time Remaining: {}     ".format(0), end="\r", flush=True)
                    print("\n")
                    self.reportStressTestsResults(getResponse)
                    self.reportSystemMetrics(getResponse)
                    # report alerts and logs location
                    alerts = getResponse["Alerts"]
                    if int(alerts) > 0:
                        alertNoSuccessText = "{} new alerts were generated during the stress tests. Run 'sudo nvsm show alerts' to see them.\n".format(
                            alerts)
                        self.printColor(alertNoSuccessText, "red")
                        self.completeReport.append(alertNoSuccessText)
                    else:
                        alertSuccessText = "\nSystem stress test successful.\nNo alerts seen."
                        self.printColor(alertSuccessText, "green")
                        self.completeReport.append(alertSuccessText)
                    self.printColor("\nStress test log can be found at {} .\n".format(
                        getResponse['LogName']), "blue")
                    self.writeCompleteReport()
                    break
        except KeyboardInterrupt:
            # kill running stress tests when ctrl-c out of the cli
            self.printColor(
                "\nKeyboard Interrupt. Stress tests aborted.", "red")
            context._queryBackend(
                path="/nvsm/v1/Systems/1/Stress", method="POST", body={"Kill": True})
        except Exception as e:
            # API gateway will return error response after POST, triggering exception here if MIG or KVM enabled
            if not migCheck() and not kvmCheck():
                print("Failed to communicate with backend")
                print(e)
            return

        print(footer)

    @staticmethod
    def implements(argv):
        return argv[0].lower() == "stress-test"


nvsm.command.Command.register_command(StressCommand)
