# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.




import os
import re
import io
import sys
import time
import syslog
import traceback
import json
from nvsm.command.queryNGC import NGC_AuthenicateAndQueryImageRef
from nvsm.command.queryNGC import NGC_AuthenicateAndVerifyImageRef
from nvsm.command.queryNGC import Docker_ValidateLocalImageRef
import nvsm.command
import nvsm.exceptions
import nvsm.interpreter
import nvsm.prompt
from nvsm.command.show import ShowCommand

default_password  = '<NOT_SET>'   # see lib/apihandlers/api_firmware.go
default_image_ref = '<NOT_SET>'

update_aliases = [
  [ 'update', 'fw' ],
  [ 'update', 'firmware' ],
]

show_aliases = [
  [ 'show', 'fw' ],
  [ 'show', 'firmware' ],
]

generic_aliases = [
  [ 'fw' ],
  [ 'firmware' ],
]

platform_dict = {
        "DGXA100"         : { "platform": "dgxa100",        "prefix": "nv.*" },
        "DGX Station A100": { "platform": "dgxstationa100", "prefix": "nv.*" },
        "CGX"             : { "platform": "driveconst",     "prefix": "",    },
        "DCS"             : { "platform": "driveconst",     "prefix": "",    },
        "DGX-2"           : { "platform": "dgx2",           "prefix": "nv.*" },
        "DGX-1"           : { "platform": "dgx1",           "prefix": "nv.*" }
}

def get_fwuc_name_filter_prefix():

    data = ""
    with open("/sys/class/dmi/id/board_name") as data_file:
        data = data_file.read()

    for dict_key, dict_val in platform_dict.items():
        if dict_key in data:
            return dict_val

    return {}

def is_fwuc_running():

    cmd = "docker ps --filter \"label=fwc_build\" --format \"{{ .ID }}  {{.Image}}  {{ .Names }}\""

    handle = os.popen(cmd)

    data = handle.read().strip()

    if len(data) == 0:
        return False, []

    fwuc_instances = []
    for line in data.split('\n'):

        tmp = line.split(' ')
        try:
            instance_dict = { "Image-Id" : tmp[0],
                              "Image-Ref": tmp[2]
                            }
            fwuc_instances.append(instance_dict)
        except:
            pass

    return True, fwuc_instances

###############################################################################
# FIXME: find better interface
def NVSM_AuthenticationStart():
    name = "/etc/nvsm/nvsm-perpetual.jwt"
    try:
        with open(name) as f:
            token = f.read()
            return token
    except:
        Fail("could not open %s" %(name))
        return None

class FirmwareCommand(nvsm.command.show.ShowCommand):
    FW_TARGET = "/systems/localhost/firmware"

    def __init__(self):
        super(FirmwareCommand, self).__init__(argv=self.get_command())

    def get_command(self):
        """ Abstract method. Subclasses mus implement the get_command() method """
        raise NotImplementedError

class FirmwareCommandProperties(FirmwareCommand):
    def __init__(self):
        super(FirmwareCommandProperties, self).__init__()
        self.image_ref = None       # defaults in backend to '<NOT_SET>'
        self.password  = None       # defaults in backend to '<NOT_SET>'
        self.username  = None       # defaults in backend to '$oauthuser'
        self.mode      = None       # defaults in backend to 'mode'
        self.state     = None

    def get_command(self):
        command = "show -display Properties %s/install" %(FirmwareCommand.FW_TARGET)
        return command.split(' ')

    def run(self, context, file=sys.stdout):
        #
        # This overrides the ShowCommand run() method to fetch the firmware properties from
        # the backend. As NVSM backend is slow to come fully ready, we will retry here
        # until the backend targets come up.
        max_timeout = 180
        delay_time = 20
        retry_count = int(max_timeout / 20)

        for count in range(0, retry_count):
            try:
                address = nvsm.targetAddress.TargetAddress(self.options.target, context)
                targets = context.findTargets(address, self.options.level,self.options.all)
            except:
                nvsm.log.warning("Error connecting to NVSM backend. Retrying ...")

                # Repopulate the target hierachies
                context.refresh()
                time.sleep(delay_time)

        return super(FirmwareCommand, self).run(context, file)

    def query_properties_from_backend(self, context, argv):

        buff = io.StringIO()
        rc = self.run(context, file=buff)

        m = re.search(r'^\s*DockerImageRef\s*=\s*(?P<image_ref>\S+)$', buff.getvalue(), flags=re.MULTILINE)
        if m:
            self.image_ref = m.group('image_ref')

        m = re.search(r'^\s*DockerPassword\s*=\s*(?P<password>\S+)$',  buff.getvalue(), flags=re.MULTILINE)
        if m:
            self.password = m.group('password')

        m = re.search(r'^\s*DockerUsername\s*=\s*(?P<username>\S+)$',  buff.getvalue(), flags=re.MULTILINE)
        if m:
            self.username = m.group('username')

        m = re.search(r'^\s*DockerMode\s*=\s*(?P<mode>\S+)$',  buff.getvalue(), flags=re.MULTILINE)
        if m:
            self.mode = m.group('mode')

        m = re.search(r'^\s*State\s*=\s*(?P<state>\S+)', buff.getvalue(), flags=re.MULTILINE)
        if m:
            self.state = m.group('state').lower()

    def get_update_state(self):

        if self.state == 'installing':
           return True, 'Busy installing firmware image: "{}"'.format(self.image_ref)

        if self.state.lower() == 'downloading':
            return True, 'Busy downloading firmware image: "{}"'.format(self.image_ref)

        return False, ""

class FirmwareCommandSet(nvsm.command.set.SetCommand):
    def __init__(self, key, val):
        self.key = key
        self.val = val
        super(FirmwareCommandSet, self).__init__(argv=self.get_command())

    def get_command(self):
        command = "set %s/install" %(FirmwareCommand.FW_TARGET)
        cmd_ary = command.split(' ')
        cmd_ary.append("%s=%s" %(self.key, self.val))
        return cmd_ary

class FirmwareCommandFetch(nvsm.command.start.StartCommand):
    def __init__(self):
        super(FirmwareCommandFetch, self).__init__(argv=self.get_command())

    def get_command(self):
        command = "start %s/fetch" %(FirmwareCommand.FW_TARGET)
        return command.split(' ')

class FirmwareCommandInstall(nvsm.command.start.StartCommand):
    def __init__(self):
        super(FirmwareCommandInstall, self).__init__(argv=self.get_command())

    def get_command(self):
        command = "start %s/install" %(FirmwareCommand.FW_TARGET)
        return command.split(' ')

###############################################################################
# class UpdateFirmwareCommand
#
class UpdateFirmwareCommand(nvsm.command.Command):
    def __init__(self, argv):
        super(UpdateFirmwareCommand, self).__init__(argv=argv)

        # Construct the command that implements this command alias
        self.__cancel_existing_install = FirmwareCommandSet("State", "Cancel")
        self.__show_properties = FirmwareCommandProperties()
        self.__start_fetch = FirmwareCommandFetch()
        self.__start_install = FirmwareCommandInstall()

        # Look for supported aliases and set corresponding install flags
        argv_lower = [x.lower() for x in argv]
        install_args = []
        if any([x == argv_lower[:len(x)] for x in update_aliases]):
            if len(argv_lower) == 2:
                # Update all firmware by default
                install_args = [ "update_fw", "all" ]
            else:
                # Use flags passed by user
                install_args = [ "update_fw" ]
                install_args.extend(argv[2:])
        elif any([x == argv_lower[:len(x)] for x in show_aliases]):
            # Show version of firmware installed on system
            install_args = [ "show_version" ]
            install_args.extend(argv[2:])
        elif any([x == argv_lower[:len(x)] for x in generic_aliases]):
            # Pass user flags unmodified to firmware update container
            #
            # Some fwuc cmd options may conflict with nvsm options,
            # such as fwuc -i for inactive and nvsm -i interpreted as
            # interactive by argparser. So we 'll use sys.argv for
            # generic aliases instead. Other fw aliases do not have
            # options.
            install_args = sys.argv[2:]
        self.install_args = install_args   # passed to query code in 'run'

    @staticmethod
    def implements(argv):
        """ Determine if the given command is an alias for "update fw".

        This method allows the NVSM Command factory pattern to determine if
        the "UpdateFirmwareCommand" class implements the given command.

        This method returns True if argv looks like "update fw", case
        insensitive. Returns False otherwise.
        """
        argv_lower = [x.lower() for x in argv]
        aliases = []
        for x in [ show_aliases, update_aliases, generic_aliases ]:
            aliases.extend(x)
        return any([x == argv_lower[:len(x)] for x in aliases])

    ###########################################################################
    # get_boxtypefilter
    #
    # Return a regex filter string to select the ngc org names that carry repos
    # related to doing firmware updates on this particular box type
    #
    # This is a bridge to the container orgs in nvcr.io
    #
    #     dgx-1   nvdgx1
    #
    # returns rc, platform_name, nvname_prefix. nvfilt_prefix
    #         0    "dgx1"         "nvdgx1"       "nvdgx1.*"
    #
    def get_boxtypefilter(self):
        name_filter_prefix_dict = get_fwuc_name_filter_prefix()

        if len(name_filter_prefix_dict) == 0:
            nvsm.log.debug("unable to get platform name")
            return 1, "", "", ""

        preprefix = name_filter_prefix_dict["prefix"]  # nvdgx1, nvsupport/dgx1 etc..
        platform_name = name_filter_prefix_dict["platform"]

        nvname_prefix = "%s%s"   %(preprefix, platform_name)
        nvfilt_prefix = "%s%s.*" %(preprefix, platform_name)    # convert to regex expression  'nvdgx1.*'
        return 0, platform_name, nvname_prefix, nvfilt_prefix   # return platform, org_name and filter

    ###########################################################################
    # query_image_ref
    #
    # Authenicates user, and allows them to select an image to download off
    # of ngc, based on box type and user's authorization
    #
    # The 'install_args' is the firmware container command line. It's built here
    # in nvsm as a combonation of the nvsm command that is run, and user args.
    # Not actually used in the authentication/query below, but it can be modified
    # by that code if desired. It's passed back as part of the "backend" information
    # to be passed back to nvsm to actually peform the fetch and install
    #
    def query_image_ref(self, context, install_args):
        auto = False
        rc = -1                   # we don't have values yet

        # get platform name
        # no quick way through the env variable or previous run, so going to have to work for it.
        # get box type name 'nvdgx1' and filter 'nvdgx1.*' which is used for prefix of NGC org

        rc1, platform_name, nvname_prefix, nvfilt_prefix = self.get_boxtypefilter()
        if rc1 != 0:
            return rc1, platform_name      # error message in first text string returned by above call

        image_ref = self.__show_properties.image_ref
        username = self.__show_properties.username
        password = self.__show_properties.password

        # user override image ref with env variable -- take as is, validate before use

        if 'NVSM_FW_IMAGE_REF' in os.environ:
            image_ref = os.environ['NVSM_FW_IMAGE_REF']
            nvsm.log.debug("NVSM_FW_IMAGE_REF='%s'" %(image_ref))

            # this code is simular to the query below, except in this case we have
            # the entire path so don't need to query. Actually query makes this same
            # verify call when its determined the path.

            #
            # Find the image given in NVSM_FW_IMAGE_REF. If image_ref exists
            # locally, bypass local/remote selection prompt
            bRet = find_fwuc(image_ref)

            if not bRet:
                print ("Image %s does not exist" %(image_ref))
                rc, text = NGC_AuthenicateAndVerifyImageRef(install_args, platform_name, image_ref, auto)
            else:
                rc = 0
                backend_data = { "DockerImageRef": image_ref,
                                 "Flags"         : " ".join(install_args),
                                 "DockerMode"    : "local"
                               }
                text = {}
                text["backend"] = backend_data
                text["status"]  = "Success"
                text["local"]  = True

        if rc != 0:

            # is user pre-defining a org stage "dev", "qa", "stg" ??

            if 'NVSM_FW_IMAGE_STAGE' in os.environ:
                user_stage = os.environ['NVSM_FW_IMAGE_STAGE']
                nvsm.log.debug("NVSM_FW_IMAGE_STAGE='%s'" %(user_stage))
                org        = "%s-%s" %(nvname_prefix, user_stage)   # "nvdgx1-dev"
                org_filter = ""
            else:
                org        = None
                org_filter = nvfilt_prefix       # nvdgx1*"

            # This func can be pre-selected with org/repo:tag info, or filtered. See code
            # for more info
            # also authenticates the user -- only orgs/repos that the user is authorized
            # to see will be displayed.

            rc, text = NGC_AuthenicateAndQueryImageRef(install_args, platform_name, root="nvcr.io", org=org,
                                                       org_filter=org_filter, repo_filter="nvfw-.*",
                                                       auto=auto)

        return(rc, text)         # text is error msg or dictionary containing full
                                 # info on how to obtain and load the container


    ###########################################################################
    # run
    #
    # attempts to get the name of a container to download, and if gets then
    # does the download and install.
    def run(self, context, file=sys.stdout):
#       try:
#           with open(os.devnull, 'w') as devnull:
#               # To avoid confusing the user, cancel any install stuck at "prompt"
#               nvsm.log.debug("To avoid confusing the user, cancel any install stuck at 'prompt'")
#               self.__cancel_existing_install.run(context=context, file=devnull)
#               nvsm.log.debug("after self.__cancel_existing_install.run")
#               time.sleep(2)
#       except nvsm.exceptions.BackendError:
#           nvsm.log.debug("in nvsm.exceptions.BackendError: after __cancel_existing_install.run")
#           pass


        #
        # First, query the firmware properties from the backend. If container launched from
        # NVSM is still running, bail out.
        self.__show_properties.query_properties_from_backend(context, self.argv)
        update_state, msg = self.__show_properties.get_update_state()

        if update_state == True:
            raise nvsm.exceptions.BackendError(msg)

        bRet, fwuc_instances = is_fwuc_running()

        if bRet:
            msg = ""
            for each in fwuc_instances:
                msg += "Image Id:%s  image name:%s\n" \
                        %(each["Image-Id"], each["Image-Ref"])

            msg = "Following firmware update container already running:\n" + msg
            raise nvsm.exceptions.BackendError(msg)

        # get the container name to use, given platform name and various env flags

        try:
            # asks local/remote and returns a pre-validated sructure like this:
            #
            #   {   'auto': False,
            #       'backend': {   'DockerImageRef': u'nvcr.io/driveconst-qa/nvfw-dcs:19.07.4',
            #                      'DockerPassword': 'ZG9...Ni',
            #                      'DockerUsername': '$oauthtoken',
            #                      'DockerMode':     'remote' or 'local'
            #                      'Flags': 'update_firmware PSU -s 1 -f"},  # flags from user command line
            #       'local': False,
            #       'msg': '',
            #       'org': u'driveconst-qa',
            #       'repo': u'nvfw-dcs',
            #       'status': u'Success',
            #       'tag': u'19.07.4'
            #   }
            #
            # if 'local' above is true, then DockerImageRef points (validated) at a local
            # image on box, and the docker username/password will not be valid. Note that
            # for this local case, we did not authenticate the user.

            rc, data = self.query_image_ref(context, self.install_args)
            if rc != 0:
                # if error, may not have gotten around yet to completing the above struct
                # and simple string will be all that is returned. Check type to know
                if type(data) == dict:
                    status = data["status"]
                else:
                    status = data
                nvsm.log.error(status) # issues getting name - abort or authentication
                return 1
        except Exception as e:
            nvsm.log.error("Problem querying image ref '%s'" %(self.install_args))
            print(e)

            exception_msg = traceback.format_exc().splitlines()
            for line in exception_msg:
                nvsm.log.debug (line)

            return 1                    # error/logs printed in call above

        # verify that we were given a image_ref to use
        try:
            image_ref = data["backend"]["DockerImageRef"]   # pull this out for error messages -- will except if not provided
        except:
            nvsm.log.error("No image specified\n%s" %(data))
            return 1

        # send 'DockerUsername', 'DockerPassword', 'DockerImageRef', and any other data
        # values determined by earlier query call to backend -- these are all
        # 'set /systems/localhost/firmware/install key=value' type commands
        try:

            for key in data["backend"]:                     # multiple fields created by query to 'set' to backend
                value = data["backend"][key]                # key/value pairs, same names as backend expects
                fw_cmd = FirmwareCommandSet(key, value)

                with open(os.devnull, 'w') as devnull:
                    fw_cmd.run(context=context, file=devnull)

        except KeyboardInterrupt:
            nvsm.log.debug("received KeyboardInterrupt")
            print('Canceled', file=file)
            return 1
        except nvsm.exceptions.BackendError as e:
            print(e)
            raise e

        # This kicks off the fetch
        # If flag is 'local' then the fetch is not done -- DockerMode == 'local'
        # When 'remote' then container is pulled from NGC and placed in Local Docker space
        # In both cases, the image is install from Local Docker Space -- using the
        # same image_ref name. This the only real difference between 'local' and 'remote'
        # is that 'remote' does a fetch first

        try:
            if  data["local"] == False:        # must be remote, if not local
                action = "fetching"            # used in exception error msg below

                # start /systems/localhost/firmware/fetch'

                self.__start_fetch.run(context=context, file=file)

            # paronoid check to be sure image was downloaded

            action = "verifying"
            rc, msg = Docker_ValidateLocalImageRef(image_ref)
            if (rc != 0):
                print(msg)                 # msg is error message
                return rc                  # error, did not download image
            else:
                nvsm.log.debug("Verified that '%s' was downloaded successfully" %(image_ref))

            # And finally the install: 'start /systems/localhost/firmware/install'
            # NOTE: as of 03/03/2020 this backend run command does not verify that image exists

            action = "installing"
            self.__start_install.run(context=context, file=file)
            nvsm.log.debug("Successful install of '%s'" %(image_ref))
        except Exception as e:
            nvsm.log.error("Problem %s '%s'" %(action, image_ref))  #"Problem fetching" or "Problem installing"
            print(e)

            exception_msg = traceback.format_exc().splitlines()
            for line in exception_msg:
                nvsm.log.debug (line)

            return 1
        return 0

# Register as a command alias with the NVSM Command factory method.
# This allows NVSM to easily construct Commands of this type.
nvsm.command.Command.register_command_alias(UpdateFirmwareCommand)
