# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



import sys

import nvsm.exceptions
from nvsm.targetAddress import TargetAddress

import nvsm.log

import fnmatch
from functools import reduce

def get_targets(t):
    try:
        targets = t.targets
        return targets
    except nvsm.exceptions.BackendError as e:
        nvsm.log.error('{0} for target address {1}', e, t.address)
        return []

class SMASHContext(object):
    def __init__(self, workingTargetAddress=TargetAddress('/')):
        self.__workingTargetAddress = workingTargetAddress
        self.__showall = False

    def getTarget(self, address):
        raise NotImplementedError

    @property
    def workingTargetAddress(self):
        return self.__workingTargetAddress

    @workingTargetAddress.setter
    def workingTargetAddress(self, address):
        target = self.getTarget(address)
        if not target:
            raise nvsm.exceptions.MissingTarget(
                    'Target address "{0}" does not exist'.format(
                    str(address)))
        self.__workingTargetAddress = target.address.canonicalize()

    @property
    def showall(self):
        return self.__showall

    @showall.setter
    def showall(self, value):
        self.__showall = value

    def refresh(self):
        pass

    def __resolveGlob(self, search_address):
        next_working_targets = [
            self.getTarget(TargetAddress('/', context=self)) ]
        for name in search_address.canonicalPath:
            working_targets = next_working_targets
            next_working_targets = []
            # Expand glob at this level
            for t in working_targets:
                targets = []
                targets.extend(get_targets(t))
                for k in targets:
                    if fnmatch.fnmatch(k.name, name):
                        next_working_targets.append(k)
        return next_working_targets

    def findTargets(self, search_address, level=1, showall=False):
        address_str = search_address.canonicalAddress
        address_list = address_str.strip().split("/")
        if "1" in address_list and address_list[2] != "localhost":
            ix = address_list.index("1")
            address_list[ix] = "localhost"
            prev_context = search_address.context
            search_address = nvsm.targetAddress.TargetAddress("/".join(address_list), context=prev_context)
        nvsm.log.debug('Looking for targets under "{address}"',
                    address=search_address)
        self.showall = showall
        if level == 'all':
            level = sys.maxsize
        matching_targets = []
        # Start search with targets matching given search address
        top_targets = self.__resolveGlob(search_address)
        if not top_targets:
            raise nvsm.exceptions.MissingTarget(
                'Target address "{0}" does not exist'.format(
                search_address))
        matching_targets.extend(top_targets)
        # Traverse the target hierarchy looking for matching targets
        working_level = 1
        next_working_targets = reduce(lambda a, b: a + b, list(map(get_targets, top_targets)), [])
        while working_level < level:
            working_targets = next_working_targets
            matching_targets.extend(working_targets)
            next_working_targets = []
            for target in working_targets:
                # TODO: Apply some filter criterion to targets?
                next_working_targets.extend(get_targets(target))
            if not next_working_targets:
                break
            working_level += 1
        return matching_targets
