# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.




from . import ply
import re
import sys

import nvsm.grammar
import nvsm.lexicon
import nvsm.program

import nvsm.log

class Interpreter(object):
    ''' NVSM CLI interpreter encompassing both the NVSM lexer and parser.

    The same NVSM CLI interpreter is used for both the interactive and
    non-interactive modes of the NVSM CLI.

    This interpreter is also used for unit testing.
    '''
    def __init__(self):
        # Prepare the NVSM lexer and parser
        try:
            import nvsm.parsetab
        except ImportError:
            print('ERROR: Could not find parse tables; try running '
                  './generateParseTables.py', file=sys.stderr)
            sys.exit(1)
        self.__lexer = ply.lex.lex(
            module=nvsm.lexicon,
            reflags=re.UNICODE | re.VERBOSE)
        self.__parser = ply.yacc.yacc(
                debug=False,
                write_tables=False,
                tabmodule=nvsm.parsetab,
                module=nvsm.grammar)

    def evaluate(self, user_input, working_target='/'):
        # Reset the lexer and parser
        self.__lexer.input('')
        try:
            self.__parser.restart()
        except:
            pass
        # Tokenize and parse the user input
        parse_result = self.__parser.parse(user_input, lexer=self.__lexer)
        nvsm.log.debug('parse result: {0}', parse_result)
        if not parse_result:
            nvsm.log.error('Unable to parse user input')
            return None

        # Construct the program object from the parse tree
        try:
            program = nvsm.program.Program.from_program_node(
                parse_result, working_target=working_target)
        except (nvsm.exceptions.InvalidArguments,
                nvsm.exceptions.UnknownCommand) as e:
            nvsm.log.error('{0}', e)
            return None

        return program
