# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.


from functools import reduce

class OData(object):
    def __init__(self, context):
        super(OData, self).__init__()
        self.__context = context

    @staticmethod
    def from_value(value, context):
        if isinstance(value, OData):
            return value
        if type(value) is dict:
            if (len(value) == 1) and ('@odata.id' in value):
                return ODataReference(value, context=context)
            return ODataObject(value, context=context)
        if type(value) is list:
            return ODataArray(value, context=context)
        return value

def from_value(value, context):
    return OData.from_value(value=value, context=context)

class ODataReference(OData):
    def __init__(self, ref, context):
        super(ODataReference, self).__init__(context=context)
        assert type(ref) is dict
        assert '@odata.id' in ref
        self.__ref = ref
        self.__needsDereference = True
        self.__value = None
        self.__context = context

    def __dereference(self, forceRefresh=False):
        # Split reference into path and fragment identifier
        ref = self.__ref['@odata.id']
        if '#' in ref:
            hash_split = ref.split('#', maxsplit=1)
            path = hash_split[0]
            fragment_path = hash_split[1].strip('/').split('/')
        else:
            path = ref
            fragment_path = []
        path = path.strip('/')
        # Fetch the response from the backend
        response = self.__context._queryBackend(path, forceRefresh=forceRefresh)  # XXX: In the future, pass method='GET' here
        # Access the value specified by the fragment identifier
        # NOTE: This uses both dict and list semantics
        def to_int(k, v):
            if type(v) is not list:
                return k
            try:
                k = int(k)
            except:
                pass
            return k
        value = reduce(lambda v, k: v[to_int(k, v)], fragment_path, response)
        self.__value = OData.from_value(value, context=self.__context)
        self.__needsDereference = False

    def __getitem__(self, key):
        if self.__needsDereference:
            self.__dereference()
        return self.__value[key]

    def get(self, key, default=None, forceRefresh=False):
        if self.__needsDereference:
            self.__dereference(forceRefresh=forceRefresh)
        return self.__value.get(key, default)

    def items(self):
        if self.__needsDereference:
            self.__dereference()
        return iter(list(self.__value.items()))

    def __contains__(self, key):
        if self.__needsDereference:
            self.__dereference()
        return key in self.__value

    def __iter__(self):
        if self.__needsDereference:
            self.__dereference()
        for i in range(len(self.__value)):
            yield self[i]

class ODataObject(OData):
    def __init__(self, value, context):
        super(ODataObject, self).__init__(context=context)
        assert type(value) is dict
        self.__value = value
        self.__lower = { k.lower(): k for k in list(value.keys()) }
        self.__cache = {}
        self.__context = context

    def __getitem__(self, key):
        lower = key.lower()
        if lower not in self.__lower:
            raise KeyError
        if lower not in self.__cache:
            value = self.__value[self.__lower[lower]]
            odata_value = OData.from_value(value, context=self.__context)
            self.__cache[lower] = odata_value
        return self.__cache[lower]

    def get(self, key, default=None):
        try:
            result = self[key]
        except KeyError:
            result = default
        return result

    def items(self):
        for key in list(self.__value.keys()):
            yield key, self[key]

    def __contains__(self, key):
        try:
            self[key]
        except KeyError:
            return False
        return True

class ODataArray(OData):
    def __init__(self, value, context):
        super(ODataArray, self).__init__(context=context)
        assert type(value) is list
        self.__value = value
        self.__cache = {}
        self.__context = context

    def __getitem__(self, key):
        if key not in self.__cache:
            value = self.__value[key]
            odata_value = OData.from_value(value, context=self.__context)
            self.__cache[key] = odata_value
        return self.__cache[key]

    def __iter__(self):
        for i in range(len(self.__value)):
            yield self[i]

    def __len__(self):
        return len(self.__value)
