#!/usr/bin/env python3

# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.





import curses
import sys
from contextlib import contextmanager
from nvsm.color import colorize

class ProgressBar(object):
    MAX_BAR_LENGTH = 40
    CURSORS = [ '|', '/', '-', '\\' ]
    def __init__(self, file=sys.stdout):
        super(ProgressBar, self).__init__()
        self.__file = file
        self.__percent = 0.0
        self.__message = ''

        self.__barWidth = 40
        self.__cursor = '|'
        self.__cursorIndex = 0

        # Obtain terminal control characters from terminfo database
        curses.setupterm(fd=self.__file.fileno())
        self.__civis = curses.tigetstr('civis').decode('utf-8')
        self.__cnorm = curses.tigetstr('cnorm').decode('utf-8')
        self.__cud1 = curses.tigetstr('cud1').decode('utf-8')
        self.__cuu1 = curses.tigetstr('cuu1').decode('utf-8')
        self.__el = curses.tigetstr('el').decode('utf-8')
        self.__hpa = curses.tigetstr('hpa')

        self.__validControlCharacters = all([
            self.__civis,
            self.__cnorm,
            self.__cud1,
            self.__cuu1,
            self.__el,
            self.__hpa
        ])

    def __enter__(self):
        # Hide the cursor while progress bar is shown
        if self.__validControlCharacters:
            print(self.__civis, end='', file=self.__file)
        # Make room for two-line progress bar
        print(file=self.__file)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        print(file=self.__file)
        # Show the cursor upon leaving progress bar
        if self.__validControlCharacters:
            print(self.__cnorm, end='', file=self.__file)

    def __insertLine(self):
        print('\n', end='', file=self.__file)

    def __clearLine(self):
        if not self.__validControlCharacters:
            return
        print(curses.tparm(self.__hpa, 0).decode('utf-8'), end='', file=self.__file)  # move to beginning of line
        print(self.__el, end='', file=self.__file)  # delete the line
        print(curses.tparm(self.__hpa, 0).decode('utf-8'), end='', file=self.__file)  # move to beginning of line

    def clearLine(self):
        self.__clearLine()
        self.__file.flush()

    def __upLine(self):
        if not self.__validControlCharacters:
            return
        print(self.__cuu1, end='', file=self.__file)  # move up one line

    def __downLine(self):
        if not self.__validControlCharacters:
            return
        print(self.__cud1, end='', file=self.__file)  # move down one line

    def __upLine(self):
        if not self.__validControlCharacters:
            return
        print(self.__cuu1, end='', file=self.__file)  # move up one line

    def __downLine(self):
        if not self.__validControlCharacters:
            return
        print(self.__cud1, end='', file=self.__file)  # move down one line

    def __advanceCursor(self):
        self.__cursorIndex = (self.__cursorIndex + 1) % len(ProgressBar.CURSORS)
        self.__cursor = ProgressBar.CURSORS[self.__cursorIndex]

    def __drawProgressBar(self):
        bar_length = int(0.01 * self.__percent * ProgressBar.MAX_BAR_LENGTH)
        print('{percent:5.1f}% [{bar}{cursor}{space}]'.format(
                    percent=self.__percent,
                    bar='=' * bar_length,
                    cursor=self.__cursor if self.__percent < 100.0 else '=',
                    space=' ' * (ProgressBar.MAX_BAR_LENGTH - bar_length)),
                end='',
                file=self.__file)

    def __drawMessage(self):
        print('{message}'.format(message=self.__message),
                end='', file=self.__file)

    def update(self, percent, message, error=False):
        if error and (not message.endswith('\n')):
            # Keep error messages on the screen
            message = message + '\n'
        self.__percent = percent
        self.__message = message
        # Count newlines in message and insert lines accordingly
        line_count = self.__message.count('\n')
        lines = curses.tigetnum('lines')

        #
        # Bug 200584575
        # cursor up will fail if line_count is greater than terminal line count
        if line_count > lines:
            line_count = lines
        for i in range(line_count):
            self.__insertLine()
        # Clear lines to make room for message
        for i in range(line_count + 1):
            self.__upLine()
            self.__clearLine()
        if error:
            with colorize('red'):
                self.__drawMessage()
        else:
            self.__drawMessage()
        # Clear line and draw current progress bar
        self.__downLine()
        self.__clearLine()
        self.__advanceCursor()
        self.__drawProgressBar()
        self.__file.flush()

    def updateProgress(self, percent=None):
        if percent is not None:
            self.__percent = percent
        self.__clearLine()
        self.__advanceCursor()
        self.__drawProgressBar()
        print(curses.tparm(self.__hpa, 0).decode('utf-8'), end='', file=self.__file)  # move to beginning of line
        self.__file.flush()
