# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.




from .pyrepl import readline  # MIT-licensed readline clone
import sys
import traceback

from nvsm.color import colorize
from nvsm.backend.context import NVSMContext
from nvsm.backend.handle import NVSMHandle
import nvsm.interpreter
from nvsm.targetAddress import TargetAddress

from nvsm.completer import init_history, make_completer
from .pyrepl.readline import _ReadlineWrapper

PROMPT = 'nvsm{working_target}-> '

def main_loop():
    ''' The read-evaluate-print loop for NVSM.

    This uses the pyrepl library for command line editing, history, and
    completion.
    '''

    # Initialize the NVSM interpreter (persists across loop iterations)
    interpreter = nvsm.interpreter.Interpreter()

    # Handle representing persistent NVSM backend connection
    backend = NVSMHandle()
    backend.connect()
    
    init_history()
    # Keep track of working target address for duration of REPL
    workingTargetAddress = TargetAddress('/')

    # read, evaluate, print, loop
    while True:
        # NVSM API context persists for the duration of a single REPL invocation
        with NVSMContext(
                backendHandle=backend,
                workingTargetAddress=workingTargetAddress) as context:
            try:
                # Format NVSM prompt depending on context
                wt = str(context.workingTargetAddress)
                if wt == '/':
                    wt = ''
                else:
                    wt = '({0})'.format(wt)
                prompt = PROMPT.format(working_target=wt)

                make_completer(context=context)
                # Read user input
                #user_input = (input(prompt))
                user_input = readline._wrapper.raw_input(prompt).decode('utf-8')

            except EOFError:
                nvsm.log.debug('Received EOF; exiting NVSM shell')
                sys.exit(0)
            except KeyboardInterrupt:
                # Keyboard interrupt; user wants to clear line editor input
                nvsm.log.debug('Keyboard interrupt')
                continue

            # Ignore empty input strings
            if user_input.strip() == '':
                continue

            # Evaluate user input string
            program = interpreter.evaluate(
                    user_input,
                    working_target=workingTargetAddress.canonicalAddress)

            if not program:
                # Some error in constructing the program
                # remove unknown program from history
                if readline.get_current_history_length() > 0:
                    readline.remove_history_item(readline.get_current_history_length()-1)
                continue

            try:
                # Run command(s) and print output
                program.run(context=context)
            except EOFError:
                nvsm.log.info('Received EOF; aborting command')
                context.invalidate()
                continue
            except KeyboardInterrupt:
                nvsm.log.info('Keyboard interrupt; aborting command')
                context.invalidate()
                continue
            except (nvsm.exceptions.BackendError,
                    nvsm.exceptions.CommandError,
                    nvsm.exceptions.MissingProperty,
                    nvsm.exceptions.MissingTarget,
                    nvsm.exceptions.SchemaError) as e:
                nvsm.log.error('{0}', e)
                context.invalidate()
                continue
            except Exception as e:
                # For robustness, log unhandled exceptions and continue
                nvsm.log.error('Unhandled exception: {}'.format(e))
                nvsm.log.error('Run nvsm again with "--log-level=debug" for details')
                nvsm.log.debug(traceback.format_exc())
                continue

            # Keep track of working target address
            workingTargetAddress = context.workingTargetAddress
