# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.



import nvsm.exceptions

class TargetAddress(object):
    def __init__(self, string, context=None):
        # Optional argument: NVSM context in which address string was given
        self.__context = context

        # Required argument: string representing target address
        self.__string = str(string)

    def __str__(self):
        return self.__string 

    def __unicode__(self):
        return self.__string

    def __add__(self, other):
        new_string = '{0}/{1}'.format(
                self.__string.rstrip('/'),
                str(other).lstrip('/'))
        new_target_address = TargetAddress(
                string=new_string, context=self.context)
        return new_target_address

    def __eq__(self, other):
        if self.context != other.context:
            return False
        return self.canonicalAddress == other.canonicalAddress

    @property
    def context(self):
        return self.__context

    @property
    def canonicalAddress(self):
        return '/{0}'.format('/'.join(self.canonicalPath))

    @property
    def canonicalPath(self):
        relative_path = self.__string.split('/')
        working_path = []
        if self.isRelative():
            # Address is relative to the current target address of the context
            if not self.__context:
                raise nvsm.exceptions.MissingTarget(
                        'Tried to resolve relative target address "{0}" without context'.format(self.__string))
            working_path.extend(
                    self.__context.workingTargetAddress.canonicalPath)
        # Resolve relative paths in working path
        while len(relative_path):
            next_target = relative_path.pop(0)
            if not next_target:
                # Ignore empty path
                continue
            if next_target == '.':
                # Relative to the current working path
                continue
            if next_target == '..':
                # Move up in the target hierarchy
                # NOTE: In UNIX style paths, the path "/../../.." is equivalent to "/"
                if len(working_path) > 0:
                    working_path.pop()
                continue
            # Append to the working path (existence of target is not important)
            working_path.append(next_target)
        return working_path

    def canonicalize(self):
        return TargetAddress(self.canonicalAddress, context=self.__context)

    @property
    def basename(self):
        path = self.canonicalPath
        if len(path) == 0:
            return ''
        return path[-1]

    @property
    def string(self):
        return self.__string

    def isRelative(self):
        # XXX: What about zero-length strings? Implicit targets?
        if self.__string[0] != '/':
            return True
        return False
