#!/usr/bin/python
# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
dcshwapikey monitoring module

"""

from nvsm_dshm.thea import Topic
from nvsm_dshm.thea import Endpoint
from nvsm_dshm.thea import Client
from nvsm_dshm.policy import policy
from nvsm_dshm.policy import IPolicy
from nvsm_dshm.infra import log
from nvsm_dshm.infra import ErrCode as Err
from nvsm_dshm.infra import config_settings
from nvsm_dshm.equipment import Equipment
from nvsm_dshm.equipment import catalog
from nvsm_dshm.ds import AlertsDataStore
import errno
import time
import logging
import sys
import syslog
import select

import time
import nvsm_sdk
from nvsm_sdk.rest import ApiException
from datetime import datetime, date
import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
urllib3.disable_warnings(urllib3.exceptions.SubjectAltNameWarning)


NVSM_CONN_RETRY_TIME = 5  # in seconds

logger = logging.getLogger(__name__)


def init_configuration(func):
    def inner():
        try:
            config_settings(config_file_name='/etc/nvsm/dshm/nvsm_dshm.cfg')
        except Exception as e:
            syslog.syslog(syslog.LOG_CRIT,
                          'dcs_hw_api_key: Unable to read the config file : {}'.format(e))
            sys.exit(1)
        func()
    return inner


def init_logger(func):
    """
        Initializes the log file.
    """
    def inner():
        # Init logger
        if not log.initialize(
                config_settings().get('logging', 'config-file'),
                config_settings().get('logging', 'log-dir'),
                'dcshwapikey'):
            syslog.syslog(syslog.LOG_CRIT,
                          'dcs_hw_api_key: Logger init failed')
            sys.exit(1)

        func()
    return inner


def push_defaults(func):
    # TODO move this to other module
    """
        Push defaults.
    """
    def inner():
        logger.debug('begin')

        endpoint = None
        nvsm_api_cfg_file = config_settings().get('nvsm-api', 'config-file')
        nvsm_api_jwt_file = config_settings().get('nvsm-api', 'jwt-file')

        (ca_cert_file,
         _,
         _,
         _,
         port,
         address,
         _) = config_settings.nvsm_api_config_helper(nvsm_api_cfg_file)

        # TODO change Endpoint arguments to kargs
        if port is not None:
            endpoint = Endpoint(port=str(port), jwt_file=nvsm_api_jwt_file,
                                ca_cert_file=ca_cert_file, ip_addr=address)
        else:
            endpoint = Endpoint()

        connected = False
        attempts = 0
        while True:
            if not endpoint.accessible():
                logger.warning('{} attempt, Retrying after {} sec'.format(
                    attempts, NVSM_CONN_RETRY_TIME))
                time.sleep(NVSM_CONN_RETRY_TIME)
                attempts = attempts + 1
            else:
                connected = True
                break

        if not connected:
            logger.critical('Connection to thea is failed')
            sys.exit(1)


        catalog.init('dcs_hw_api_key')
        policy.push_defaults(['DCSHWAPIKey'])

        logger.debug('end')

        func()
    return inner


def check_key():
    """
        process the event
    """
    data_store = AlertsDataStore.factory('thea_db', 'dcshwapikey')

    # Configure API key authorization: bearerAuth
    configuration = nvsm_sdk.Configuration()
    configuration.ssl_ca_cert = '/etc/nvsm/nvsm-ca.crt'
    with open('/etc/nvsm/nvsm-perpetual.jwt', 'r') as jwt_file:
        configuration.api_key['Authorization'] = "Bearer " + jwt_file.read()

    api_instance = nvsm_sdk.NVSMApi(nvsm_sdk.ApiClient(configuration))
    # str | The value of the Id property of the ComputerSystem resource
    computer_system_id = '1'
    numberOfDaysLeftToExpireLicense = 0
    try:
        api_response = api_instance.nvsm_v1_systems_computer_system_id_dcshwapikey_get(
            computer_system_id)
        date_time_obj = datetime.strptime(api_response.expiry, '%d-%b-%Y')
        now = date.today()
        numberOfDaysLeftToExpireLicense = (date_time_obj.date() - now).days
        print('Date: ', date_time_obj.date(), ' numberOfDaysLeftToExpireLicense : ', numberOfDaysLeftToExpireLicense)
    except ApiException as e:
        print("Exception when calling NVSMApi->nvsm_v1_systems_computer_system_id_dcshwapikey_get: %s\n" % e)
        exit(1)
    except Exception as e:
        print(e)
        exit(0)

    policy = IPolicy.factory('DCSHWAPIKey')
    logger.debug('Policy used {}'.format(policy))

    json_obj = {
        'numberOfDaysLeftToExpireLicense': numberOfDaysLeftToExpireLicense}

    device = Equipment.factory('DCSHWAPIKey',
                               json_obj,
                               Endpoint().hostname(),
                               Endpoint().serial(),
                               uri='nvsm/v1/Systems/1/Dcshwapikey')

    if device is not None and policy.monitor(device.get_name()):
        logger.debug('Checking device {}:  {}'.format(
            device.get_name(), device))

        ret = data_store.fetch_alerts(device.get_name())
        if ret['errored']:
            logger.error('Unable to fetch alert data from dcshwapikey from thea')
            logger.error(
                'Thea is in-accessible')
            return
        alerts = ret['alerts']
        device.apply_rules(device, alerts)
        data_store.store_alerts(device.get_name(), alerts)


@init_configuration
@init_logger
@push_defaults
def dcs_hw_api_key_main():
    """
    Main
    """
    logger.debug('Entering')

    check_key()

    logger.info('Exiting')
