# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Data store interface
"""
from abc import ABCMeta, abstractmethod
import shelve
import logging
from nvsm_dshm.thea import Client

from nvsm_dshm.equipment import Alert

logger = logging.getLogger(__name__)


class AlertsDataStore(object, metaclass=ABCMeta):

    """
    Alert Class.

    """

    @staticmethod
    def factory(ds_type, ds_input=None):
        """
            Factory method.
            Creates an instance that implements fetch and store operations for
            shelve db or the thea db.
            ds_input: Is filename for shelve db.
            ds_input: Is the device name for Nvsm db.
        """
        if ds_type == "shelve":
            logger.debug("Initializing Shelve DB for {}".format(ds_input))
            return AlertsShelveDS(ds_input)
        elif ds_type == 'thea_db':
            logger.debug(
                "Initializing Alerts class to access Nvsm DB {}".format(
                    ds_input))
            return AlertsNvsmDB(ds_input)
        return None

    @abstractmethod
    def fetch_alerts(self, device_name):
        """
        Method used to fetch all active alerts for the given device name
        """
        pass

    @abstractmethod
    def store_alerts(self, device_name, alerts):
        """
        Method to store all alerts for the given device name.

        This method only stores those alerts which were modified by the device
        monitor, in other words the alerts that were marked as dirty.
        """
        pass


class AlertsShelveDS(AlertsDataStore):

    """
        The implementation which performs fetch and store on shelve db.
    """

    def __init__(self, file_name):
        self.__sh_db = shelve.open(
            file_name, flag='c', protocol=2, writeback=True)

    def __del__(self):
        self.__sh_db.close()

    def has_device(self, device_name):
        """
        Checks if the shelve db has an entry for this device.
        """
        if device_name in self.__sh_db:
            return True
        return False

    def store_alerts(self, device_name, alerts):
        """
        Stores the alerts to the db.
        """
        device_name = device_name.encode("utf-8")
        if not self.has_device(device_name):
            logger.debug(
                "Entry for device {} not in db, creating it".format(
                    device_name))
            self.__sh_db[device_name] = {"history": [], "active": {}}

        active_alerts = self.__sh_db[device_name]["active"]
        alert_history = self.__sh_db[device_name]["history"]

        for alert_id in list(alerts.keys()):
            if alerts[alert_id].dirty:
                logger.debug("Alert {} for device {} is dirty".format(
                    alerts[alert_id], device_name))
                if not alerts[alert_id].asserted:
                    logger.debug("Alert is de-asserted, appending to history")
                    alert_history.append(alerts[alert_id].to_json())
                    logger.debug(
                        "Alert is de-asserted, deleting from active alerts")
                    del active_alerts[alert_id]
                else:
                    logger.debug(
                        "Alert is asserted, updating the entry in active \
                                alerts")
                    active_alerts[alert_id] = alerts[alert_id].to_json()

        logger.debug(
            "Writing back the history and active alerts for device \
                    {} to db".format(device_name))
        self.__sh_db[device_name]["active"] = active_alerts
        self.__sh_db[device_name]["history"] = alert_history

    def fetch_alerts(self, device_name):
        """
        Fetches alerts for the given device
        """
        alerts = {}
        device_name = device_name.encode("utf-8")

        if self.has_device(device_name):
            active_alerts = self.__sh_db[device_name]["active"]
            for key in list(active_alerts.keys()):
                alerts[key] = Alert.create(active_alerts[key])
                logger.debug("Fetched: {}".format(alerts[key]))

        return {'errored': False, 'alerts': alerts}


# FIXME: If thea is inaccessible at the time that alert is being stored back
# to the DB, then the system behaviour will be affected. Fix these issues
# post Phase 1.
# If thea is inaccessible at the time that the alerts are read, then the
# error is propogated to the monitoring subsystem. The monitoring
# subsystem will go back to connect() job where it re-establishes the
# connection to thea.
class AlertsNvsmDB(AlertsDataStore):

    """
        The implementation which performs fetch and store on Nvsm's Mongo db.
    """
    __URI_MAP = {
        'Fan': '/nvsm/v1/Alerts/Thermal',
        'Pdb': '/nvsm/v1/Alerts/Thermal',
        'PowerSupplyUnit': '/nvsm/v1/Alerts/Power',
        'PowerState': '/nvsm/v1/Alerts/Power',
        'Cpu': '/nvsm/v1/Alerts/Processors',
        'Gpu': '/nvsm/v1/Alerts/GPUs',
        'Dimm': '/nvsm/v1/Alerts/Memory',
        'Volume': '/nvsm/v1/Alerts/Storage',
        'Drive': '/nvsm/v1/Alerts/Storage',
        'StorageController': '/nvsm/v1/Alerts/Storage',
        'PCIe': '/nvsm/v1/Alerts/PCIeDevices',
        'CMM': '/nvsm/v1/Alerts/CMM',
        'Blade': '/nvsm/v1/Alerts/Blade',
        'Network': '/nvsm/v1/Alerts/Network',
        'NetworkPort': '/nvsm/v1/Alerts/NetworkPort',
        'NetworkDevFunc': '/nvsm/v1/Alerts/NetworkDevFunc',
        'NetworkAdapter': '/nvsm/v1/Alerts/Network',
        'NVLink': '/nvsm/v1/Alerts/GPUs',
        'NVSwitch': '/nvsm/v1/Alerts/NVSwitches',
        'dcshwapikey': '/nvsm/v1/Alerts/Systems'
    }

    def __init__(self, device_type):
        self.device_type = device_type

    def has_device(self, device_name):
        """
        Checks if the shelve db has an entry for this device.
        """
        pass

    #Return supported device types
    @staticmethod
    def get_supported_devices():
        return AlertsNvsmDB.__URI_MAP.keys()


    def get_all_active_alerts(self):
        """
        Get All Active alerts from the Nvsm DB 
        """
        records = None
        client = Client(
            {Client.URI: self.__URI_MAP[self.device_type], Client.ALL: True})
        records = client.get_objects()

        # Failed to fetch alert records from thea.
        if records is None:
            logger.error("Unable to fetch alert records from DB")
            return {'thea_errored': True, 'active_alerts': []}

        _required_json_attribs = set(('MemberCount', 'Alerts'))
        if not _required_json_attribs.issubset(list(records.keys())):
            logger.error(
                "Alert records from DB does not have valid data {}".format(records))
            return {'thea_errored': True, 'active_alerts': []}

        if records['MemberCount'] == 0:
            logger.debug("No Alerts found in DB ")
            return {'thea_errored': False, 'active_alerts': []}

        alert_objs = records['Alerts']
        active_alerts = []

        for alert_obj in alert_objs:
            _required_json_attribs = set(('component_id', 'severity'))
            if not _required_json_attribs.issubset(list(alert_obj.keys())):
                logger.error("Alert data is not valid {}".format(alert_obj))
                continue

            if alert_obj['severity'] != 'Cleared':
                logger.debug("Active alerts: {}".format(alert_obj))
                active_alerts.append(alert_obj)

        return {'thea_errored': False, 'active_alerts': active_alerts}

    def get_alerts(self, device_name, serial):
        """
        Get Active alerts from the Nvsm DB for this device
        """
        records = None
        client = Client(
            {Client.URI: self.__URI_MAP[self.device_type], Client.ALL: True})
        records = client.get_objects()

        # Failed to fetch alert records from thea.
        if records is None:
            logger.error("Unable to fetch alert records from {}".format(
                self.__URI_MAP[self.device_type]))
            return {'thea_errored': True, 'active_alerts': []}

        _required_json_attribs = set(('MemberCount', 'Alerts'))
        if not _required_json_attribs.issubset(list(records.keys())):
            logger.error(
                "Alert records from {} does not have valid data {}".format(
                    self.__URI_MAP[self.device_type],
                    records))
            return {'thea_errored': True, 'active_alerts': []}

        if records['MemberCount'] == 0:
            logger.debug("No Alert found at {}".format(
                self.__URI_MAP[self.device_type]))
            return {'thea_errored': False, 'active_alerts': []}

        alert_objs = records['Alerts']
        active_alerts = []
        for alert_obj in alert_objs:
            _required_json_attribs = set(('component_id', 'severity'))
            if not _required_json_attribs.issubset(list(alert_obj.keys())):
                logger.error("Alert data is not valid {}".format(alert_obj))
                continue

            if alert_obj['component_id'] == device_name \
                    and alert_obj['severity'] != 'Cleared':
                # Check if Serial number is set for the data store, filter
                # based on serial too.
                if serial and serial != alert_obj['system_serial']:
                    continue
                active_alerts.append(alert_obj)

        return {'thea_errored': False, 'active_alerts': active_alerts}

    def put_alert(self, alert):
        """
        Updates an existing alert in Nvsm DB
        """
        if not Client.put(self.__URI_MAP[self.device_type], alert.to_json(True)):
            logger.critical("PUT alert {} failed".format(alert))

    def post_alert(self, alert):
        """
        Creates a new  alert in Nvsm DB
        """
        if not Client.post(self.__URI_MAP[self.device_type], alert.to_json()):
            logger.critical("POST alert {} failed".format(alert))

    def store_alerts(self, device_name, alerts):
        """
        Stores the alerts to the db.
        """
        for alert_id in list(alerts.keys()):
            if alerts[alert_id].dirty:
                logger.debug("Alert {} for device {} is dirty".format(
                    alerts[alert_id], device_name))
                if not alerts[alert_id].asserted:
                    logger.debug("Alert is de-asserted, appending to history")
                    self.put_alert(alerts[alert_id])
                else:
                    logger.debug(
                        "Alert is asserted, updating the entry in active \
                                alerts")
                    if alerts[alert_id].new is True:
                        self.post_alert(alerts[alert_id])
                    else:
                        logger.critical(
                            "Alert is asserted and is dirty. \
                                    !!!! Shouldn't be here !!!")

        logger.debug(
            "Writing back the alerts for device {} to db".format(device_name))

    def fetch_alerts(self, device_name, serial=None):
        """
        Fetches alerts for the given device

        NOTE: If thea becomes in-accessible while fetching the alerts,
        then this function returns None.
        """
        ret = self.get_alerts(device_name, serial)
        if ret['thea_errored']:
            return {'errored': True, 'alerts': {}}

        alert_objs = ret['active_alerts']
        if alert_objs is None:
            return {'errored': False, 'alerts': {}}

        alerts_map = {}
        for alert_obj in alert_objs:
            alert = Alert.create(alert_obj)
            alerts_map[alert.alert_id()] = alert

        return {'errored': False, 'alerts': alerts_map}


    def fetch_all_active_alerts(self):
        """
        Fetches all alerts in the DB

        NOTE: If thea becomes in-accessible while fetching the alerts,
        then this function returns None.
        """
        ret = self.get_all_active_alerts()
        if ret['thea_errored']:
            return {'errored': True, 'alerts': {}}

        alert_objs = ret['active_alerts']
        if alert_objs is None:
            return {'errored': False, 'alerts': {}}

        return {'errored': False, 'alerts': alert_objs}

