# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""

Template for alert class

"""

import logging
import socket
import datetime

from nvsm_dshm.thea import Endpoint
from . import catalog
import sys
import os
sys.path.append(os.path.abspath(os.path.join(__file__, '..')))
from . import catalog


logger = logging.getLogger(__name__)


class Alert(object):

    """
    Alert Class.

    """

    def __init__(self, identifier, device_inst,
                 hostname, serial, create_new=True, dynamic_attrs=None):
        """
            Creates an alert object.

            If create_new is 'True':
            1. Alerts creation time is set to 'now'
            2. Sets alert to dirty, indicating that the alert has been
                created/modified and should be put back to datastore.

            If create_new is 'False':
            1. Indicates that the alert is already in the DB,
                and alert object is created for DSHMs use.

            Dynamic alerts are:
               1. Alerts that do not have pre-defined rules.
               2. Alerts that are not statically defined in the alert json files. 
               3. Example: DCGM errors for GPU are dynamically mapped to NVSM alerts. 

            For Dynamic alerts, dyanmic_attrs is not None. 
            dynamic_attrs is a dictionary with keys 'severity', 'message', 
            'message_details', 'recommended_action' and 'description'
        """
        self.id = identifier
        self.device_inst = device_inst
        self.__asserted = True
        self.dirty = False
        self.new = False
        self.hostname = hostname
        self.serial = serial
        self.msg_param = None
        self.chassis_id = None
        self.dynamic_attrs = dynamic_attrs
        self.case_id = None
        self.notifier = None
        self.notifier_update = False

        if create_new:
            self.origin_time = \
                datetime.datetime.now().astimezone().isoformat()
            self.dirty = True
            self.new = True
            self.clear_time = '-'

    @property
    def asserted(self):
        """
            Accessor for __asserted property.
        """
        return self.__asserted

    @asserted.setter
    def asserted(self, val):
        """
            Sets __asserted property and marks the alert object dirty,
            for writeback to the DB.
        """
        if self.__asserted != val:
            self.__asserted = val
            self.dirty = True

    @classmethod
    def create(cls, json_obj):
        """
        This method creats the alert object from the alert json in the DB.
        """
        alert = None
        required_json_attribs = set(
            ('alert_id',
             'event_time',
             'component_id',
             'system_name',
             'system_serial'))

        if required_json_attribs.issubset(list(json_obj.keys())):
            alert = cls(
                json_obj['alert_id'],
                json_obj['component_id'],
                json_obj['system_name'],
                json_obj['system_serial'],
                False)
            alert.origin_time = json_obj['event_time']
            if 'clear_time' in list(json_obj.keys()):
                alert.clear_time = json_obj['clear_time']
            else:
                alert.clear_time = '-'
            
            if 'case_id' in json_obj:
                alert.case_id = json_obj['case_id']

            if 'notifier' in json_obj:    
                alert.notifier = json_obj['notifier']

            if 'dynamic_attrs' in json_obj:
                alert.dynamic_attrs = json_obj['dynamic_attrs']


        return alert

    def alert_id(self):
        return self.id

    def notifiable(self):
        event_info = catalog.get(self.id)
        if event_info is not None and event_info['notify_user'] == 'No':
            return False
        else:
            return True

    def personalise_email(self, event, model):
        try:
            if 'personalise' in event \
                    and model in event['personalise']:
                '''
                Personalise alert message for the platform.
                placeholder is for the updated value to continue working with msg_param.
                '''
                for key in event['personalise'][model]:
                    location_info = Endpoint().location(self.chassis_id)
                    event[key] = event['personalise'][model][key].format(
                        placeholder={}, ID=self.chassis_id, **location_info)
        except Exception as e:
            logger.debug(e, exc_info=True)

        return event

    def get_property(self, prop, default):
        if self.dynamic_attrs is not None:
            if prop in self.dynamic_attrs.keys():
                return self.dynamic_attrs[prop]
            else:
                return None
        else:
            if default is not None:
                return default[prop]
            else: 
                return None


    def to_json(self, brief=False):
        """
        This method returns a json object.
        It combins the data in Event object and that in Event catalog
        maintained by nvsm_dshm.
        """
        event_info = catalog.get(self.id)
        if event_info is None:
            logger.info("Could not find static informationf or alert {}".format(self.id))
        else:
            event_info = self.personalise_email(event_info, Endpoint().model())

        json_obj = {}
        json_obj['alert_id'] = self.id
        json_obj['component_id'] = self.device_inst
        if not self.asserted and not self.notifier_update:
            json_obj['severity'] = 'Cleared'
            json_obj['clear_time'] = \
                datetime.datetime.now().astimezone().isoformat() + "."
        else:
            json_obj['severity'] = self.get_property('severity', event_info)
            json_obj['clear_time'] = '-'

        json_obj['system_serial'] = self.serial

        if self.case_id:
            json_obj['case_id'] = self.case_id

        if self.notifier:
            json_obj['notifier'] = self.notifier

        # If only brief output is desired then only return the minimum required data.
        # This is used when clearing an alert.
        if brief:
            return json_obj

        json_obj['description'] = self.get_property('description', event_info)
        json_obj['event_time'] = self.origin_time
        json_obj['system_name'] = self.hostname


        # FIXME +id: when we rework primary key for the alerts db, clean this
        # up
        msg_device_inst = self.device_inst
        if '+' in msg_device_inst:
            msg_device_inst = msg_device_inst.split('+')[-1]
        json_obj['message'] = self.get_property('message', event_info).format(msg_device_inst)

        msg = self.get_property('message_details', event_info)
        if self.msg_param is not None:
            msg = msg.format(*self.msg_param)
        json_obj['message_details'] = msg
        json_obj['recommended_action'] = self.get_property('recommended_action', event_info)


        return json_obj

    def __repr__(self):
        """
            Repr
        """
        return '{}(id:{},\
                device_inst:{},\
                serial:{},\
                asserted:{},\
                dirty:{},\
                origin_time:{},\
                clear_time:{})'.format(
            self.__class__.__name__,
            self.id,
            self.device_inst,
            self.serial,
            self.asserted,
            self.dirty,
            self.origin_time,
            self.clear_time
        )
