# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Container module

Contains the definition for Equipment Container class and its derivations.
"""

# import json
from abc import ABCMeta, abstractmethod
import logging
from nvsm_dshm.thea import Client
# from .rule import IContainerRule
# from nvsm_dshm.infra import pa
from enum import IntEnum


class LedStatus(IntEnum):
    NORMAL = 0,
    LOCATE = 1,
    FAIL = 2,
    REBUILD = 3,
    PREDFAIL = 4,
    HOTSPARE = 5


logger = logging.getLogger(__name__)


class EquipmentContainer(object, metaclass=ABCMeta):

    """
    Equipment Container:
        Represents a list of homogenous devices.

        Could be used to represent redfish containers like
        drives, volumes, dimms.
    """

    def __init__(self, device_pres_map):
        """
        Constructor
        device_type: Indicates the device type contained by this object.
        device_pres_map: It is a map that indicates if a device is present
        or absent.
        """
        self.device_type = 'EquipmentContainer'
        if device_pres_map is None:
            self.device_pres_map = {}
        else:
            self.device_pres_map = device_pres_map

    @staticmethod
    def factory(equipment_type, json_obj):
        """
        Factory for various types of equipment supported by DSHM
        """
#        if equipment_type == "DimmContainer":
#           return DimmContainer.create(json_obj)
#        if equipment_type == "DriveContainer":
#            return DriveContainer.create(json_obj)
        if equipment_type == "VolumeContainer":
            return VolumeContainer.create(json_obj)
        raise AssertionError(
            "Container equiment type undefined: " + equipment_type)

    @abstractmethod
    def create(self, json_obj):
        """
        Creates the equipment container class from the json object.
        """
        pass

    @abstractmethod
    def get_device_presence_map(self):
        """
        Returns map of device and its presence information.
        """
        return self.device_pres_map

    @classmethod
    def get_rules(cls):
        """
        Returns the rules defined for this equipment
        """
        pass

    @classmethod
    def apply_rules(cls, device, is_present, hostname, serial, alerts):
        """
        Applies the rules defined for the equipment.
        """
        rules = cls.get_rules()
        for rule in rules:
            rule.apply(device, is_present, hostname, serial, alerts)

    def __repr__(self):
        missing_devices = None

        for device in list(self.device_pres_map.keys()):
            if not self.device_pres_map[device]:
                missing_devices += device + ':'

        return 'device_type:{},missing:{}'.format(
            self.device_type, missing_devices)


# Even though DIMM container executes, it does not raise any alerts. This has been the behavior since the changes for RTX were made. Commenting out this code, delete it later.
# Rules for Dimm Container defined here.
# class DimmMissing(IContainerRule):
#
#    """
#    Dimm is missing.
#    """
#
#    def __init__(self):
#        self.alert_id = 'NV-DIMM-03'
#
#
# class DimmContainer(EquipmentContainer):
#
#    """
#    Dimm Container:
#        Container of Dimms
#    """
#
#    __rules = []
#
#    def __init__(self, device_type, device_pres_map):
#        """
#        Constructor
#        device_type: Indicates the device type contained by this object.
#        device_pres_map: It is a map that indicates if a device is present
#        or absent.
#        """
#        self.device_type = device_type
#        self.device_pres_map = device_pres_map
#
#    @classmethod
#    def create(cls, json_obj):
#        """
#        Creates the equipment container class from the json object.
#        """
#        dimm_uri = 'Systems/Memory/DIMM/List'
#        device_pres_map = dict((key, False)
#                               for key in pa().enumerate_obj(dimm_uri))
#
#        if 'Members' not in list(json_obj.keys()):
#            logger.error(
#                "DimmContainer object creation failed, \
#                        json object incomplete %s", json.dumps(json_obj))
#            return None
#
#        for member in json_obj['Members']:
#            if '@odata.id' not in list(member.keys()):
#                logger.error("@odata.id key missing from member object")
#                return None
#
#            dimm = member['@odata.id'].split('/')[-1]
#            if dimm not in list(device_pres_map.keys()):
#                logger.error(
#                    "{} dimm is not present in the platform dimm list".format(
#                        dimm)
#                )
#                continue
#
#            device_pres_map[dimm] = True
#
#        return cls('Dimm', device_pres_map)
#
#    def get_device_presence_map(self):
#        """
#        Returns map of device and its presence information.
#        """
#        return super(DimmContainer, self).get_device_presence_map()
#
#    @classmethod
#    def get_rules(cls):
#        """
#        Returns the rules defined for this equipment
#        """
#        return cls.__rules
#
#    def __repr__(self):
#        return '{}({})'.format(
#            self.__class__.__name__,
#            super(DimmContainer, self).__repr__())
#


# Rules for Drive Container defined here.
# class DriveMissing(IContainerRule):

    """
    Drive is missing.
    """

#    def __init__(self):
#        self.alert_id = 'NV-DRIVE-01'


# class DriveContainer(EquipmentContainer):

    """
    Drive Container:
        Container of Drives
    """
#    __rules = [DriveMissing()]
#    slot_map = None
#    led_map = None

#    def __init__(self, device_type, device_pres_map):
    """
        Constructor
        device_type: Indicates the device type contained by this object.
        device_pres_map: It is a map that indicates if a device is present
        or absent.
    """
#        self.device_type = device_type
#        self.device_pres_map = device_pres_map

#    @classmethod
#    def create(cls, json_obj):
    """
        Creates the equipment container class from the json object.
        # TODO: Later get the initial map from a platform catalog file, do not
        # hardcode it.
    """
#        device_pres_map = \
#            dict((key, False) for key in
#                 pa().enumerate_obj('Systems/Storage/Drive/SlotListAll'))

#        for obj in json_obj:
#            if obj is None:
#                continue

#            if 'PhysicalLocation' not in list(obj.keys()):
#                logger.error("PhysicalLoc key missing from member object")
#                return None

#            physical_slot = "Slot" + obj['PhysicalLocation']['Info']
#            if physical_slot not in list(device_pres_map.keys()):
#                logger.error("Slot is not present {}".format(
#                    physical_slot))
#                return None
#            device_pres_map[physical_slot] = True
#        return cls('Drive', device_pres_map)

#    def get_device_presence_map(self):
    """
        Returns map of device and its presence information.
    """
#        return super(DriveContainer, self).get_device_presence_map()

#    @classmethod
#    def get_rules(cls):
    """
        Returns the rules defined for this equipment
    """
#        return cls.__rules

#    def __repr__(self):
#        return '{}({})'.format(
#            self.__class__.__name__,
#            super(DriveContainer, self).__repr__())


class VolumeContainer(EquipmentContainer):

    """
    Volume Container:
        Container of Volumes
    """
    monitored_volumes = list()

    def __init__(self, device_type, device_pres_map):
        """
        Constructor
        device_type: Indicates the device type contained by this object.
        device_pres_map: It is a map that indicates if a device is present
        or absent.
        """
        self.device_type = device_type
        self.device_pres_map = device_pres_map

    @classmethod
    def create(cls, json_obj):
        """
        Creates the equipment container class from the json object.
        """
        # Expected Volume info read from policies
        # VolumeList = pa().enumerate_obj('Systems/Storage/Volume/VolumeList')

        # Monitored volumes
        VolumeList = VolumeContainer.monitored_volumes
        device_pres_map = dict()
        if VolumeList is not None:
            for vol in list(VolumeList):
                device_pres_map[vol['Name']] = False

        if 'Members' not in list(json_obj.keys()):
            logger.error("Volume container creation failed")
            return None

        for member in json_obj['Members']:
            if '@odata.id' not in list(member.keys()):
                logger.error("@odata.id key missing from member object")
                return None

            # If the volume name is not any pre-existing
            volume_name = member['@odata.id'].rsplit('/', 1)[1]
            if volume_name not in list(device_pres_map.keys()):
                logger.debug("{} new volume found".format(volume_name))
                # Fetch Volume Data
                try:
                    client = Client({Client.URI:member['@odata.id']})
                    volume_data = client.get_objects()
                    vol_obj = {
                               'Name': volume_name,
                               'DriveCount': len(volume_data['Links']['Oem']['Drives']['@odata.id']),
                               'VolumeType': volume_data['VolumeType']
                              }
                    VolumeContainer.monitored_volumes.append(vol_obj)
                except:
                    logger.error("Unable to fetch volume data for {}".format(volume_name))
                    continue
            device_pres_map[volume_name] = True
        return cls('Volume', device_pres_map)

    def get_device_presence_map(self):
        """
        Returns map of device and its presence information.
        """
        return super(VolumeContainer, self).get_device_presence_map()

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            super(VolumeContainer, self).__repr__())
