# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Equipment module

Contains the definition for Equipment class and its derivations.
"""

from nvsm_dshm.infra import pa
from nvsm_dshm.thea import Endpoint
from .sliding_window import Window
from .rule import IRule
from .sensor import Sensor
from .alerts import Alert
import json
import sys
from abc import ABCMeta, abstractmethod
import logging
import os.path
import datetime
import re

sys.path.append(os.path.abspath(os.path.join(__file__, '..')))

logger = logging.getLogger(__name__)


class Equipment(object, metaclass=ABCMeta):

    """
    Equipment class:
        Encapsulates the following
        - Device identity (entityId, name)

    TODO:
        Device identity currently is tracked using entity id and name.
        This should evolve later.

        Device health might have to be abstracted some more later on.
    """

    def __init__(self, entity_id, name, hostname, serial):
        """
        Constructor
        """
        self.entity_id = entity_id
        self.name = name
        self.hostname = hostname
        self.serial = serial

    def __repr__(self):
        return 'entity_id:{},name:{}'.format(self.entity_id, self.name)

    @staticmethod
    def factory(equipment_type, json_obj, hostname, serial, **kwargs):
        """
        Factory for various types of equipment supported by DSHM
        """
        if equipment_type == "PowerSupplyUnit":
            return PowerSupplyUnit.create(json_obj, hostname, serial, **kwargs)
        if equipment_type == "PowerState":
            return PowerState.create(json_obj, hostname, serial)
        if equipment_type == "Fan":
            return Fan.create(json_obj, hostname, serial)
        if equipment_type == "Pdb":
            return Pdb.create(json_obj, hostname, serial)
        if equipment_type == "Dimm":
            return Dimm.create(json_obj, hostname, serial)
        if equipment_type == "MemoryCollection":
            return MemoryCollection.create(json_obj, hostname, serial)
        if equipment_type == "Cpu":
            return Cpu.create(json_obj, hostname, serial)
        if equipment_type == "Gpu":
            return Gpu.create(json_obj, hostname, serial)
        if equipment_type == "GpuLimits":
            return GpuLimits.create(json_obj, hostname, serial, **kwargs)
        if equipment_type == "PCIe":
            return PCIe.create(json_obj, hostname, serial, **kwargs)
        if equipment_type == "CMM":
            return CMM.create(json_obj, hostname, serial, **kwargs)
        if equipment_type == "Blade":
            return Blade.create(json_obj, hostname, serial, **kwargs)
        if equipment_type == "NetworkPort":
            return NetworkPort.create(json_obj, hostname, serial, **kwargs)
        if equipment_type == "NetworkDevFunc":
            return NetworkDevFunc.create(json_obj, hostname, serial, **kwargs)
        if equipment_type == "NetworkAdapter":
            return NetworkAdapter.create(json_obj, hostname, serial, **kwargs)
        if equipment_type == "Network":
            return Network.create(json_obj, hostname, serial, **kwargs)
        if equipment_type == "NVSwitch":
            return NVSwitch.create(json_obj, hostname, serial)
        if equipment_type == "NVLink":
            return NVLink.create(json_obj, hostname, serial)
        if equipment_type == "DCSHWAPIKey":
            return DCSHWAPIKey.create(json_obj, hostname, serial)
        raise AssertionError("Equiment type undefined: " + equipment_type)

    @abstractmethod
    def create(self, json_obj, hostname, serial, **kwargs):
        """
        Creates the equipment class from the json object.
        """
        pass

    @abstractmethod
    def get_name(self):
        """
        Returns the device name.
        """
        return self.name

    @abstractmethod
    def config(self, config):
        """
        Additional configuration applied to the Equipment object,
        can be used to change the behaviour of the rules applied to the object

        Like setting thresholds
        """
        pass

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns the rules defined for this equipment
        """
        pass

    @classmethod
    def apply_rules(cls, equipment, alerts):
        """
        Applies the rules defined for the equipment.
        """
        rules = cls.get_rules(equipment)
        for rule in rules:
            rule.apply(equipment, alerts)

    @classmethod
    def apply_boottime_rules(cls, equipment, alerts):
        """
        Applies the rules defined for the equipment
        at boottime.
        """
        rules = cls.get_boottime_rules()
        for rule in rules:
            rule.apply(equipment, alerts)


# Rules for Fan defined here.
class FanSpeedNoReading(IRule):

    """
    Fan speed sensor reports 'NoReading'.
    """

    def __init__(self):
        self.alert_id = 'NV-FAN-02'

    @classmethod
    def chassis_id(cls, psu):
        """
           Returns the chassis id
        """
        return psu.chassis_id

    @classmethod
    def condition(cls, fan):
        if fan.get_assertion() == 'No Reading':
            return True


class FanSpeedLowerThresholdCritical(IRule):

    """
    Fan speed sensor reports 'LowerThresholdCritical'.
    """

    def __init__(self):
        self.alert_id = 'NV-FAN-01'

    @classmethod
    def condition(cls, fan):
        if fan.get_assertion() == 'LowerThresholdCritical':
            return True

    @classmethod
    def msg(cls, fan):
        """
           Returns the details for msg
        """
        return [fan.get_reading(), fan.get_threshold()]

    @classmethod
    def chassis_id(cls, psu):
        """
           Returns the chassis id
        """
        return psu.chassis_id

# Fan class defined here.


class Fan(Equipment):

    """
    Fan: Extends Equipment Class.
    """
    _required_json_attribs = set(('Reading',
                                  'LowerThresholdCritical',
                                  'MemberId',
                                  'Name'))

    __rules = [FanSpeedNoReading(), FanSpeedLowerThresholdCritical()]

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def __init__(self, entity_id, name, hostname,
                 serial, fan_speed_sensor, chassis_id):
        super(Fan, self).__init__(entity_id, name, hostname, serial)
        self._speed = fan_speed_sensor
        self.chassis_id = chassis_id

    def get_assertion(self):
        """
            Gets the fan speed assertion error.
        """
        return self._speed.get_assertion()

    def get_threshold(self):
        """
            Gets the fan speed threshold value.
        """
        return self._speed.get_threshold()

    def get_reading(self):
        """
            Gets the fan speed threshold value.
        """
        return self._speed.get_reading()

    def get_name(self):
        return super(Fan, self).get_name()

    def config(self, config):
        return super(Fan, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial):
        ret = None

        if cls._required_json_attribs.issubset(list(json_obj.keys())):
            sensor = Sensor('Performance', 'Unknown',
                            json_obj['ReadingUnits'],
                            LowerThresholdCritical=json_obj
                            ['LowerThresholdCritical'])
            sensor.set_reading(json_obj['Reading'])

            # TODO: remove redundant hostname and serial parameters
            host_id = json_obj['@odata.id'].split('/')[4]
            hostname = Endpoint().hostname(host_id)
            serial = Endpoint().serial(host_id)

            ret = cls(json_obj['MemberId'], json_obj['Name'],
                      hostname, serial, sensor, host_id)
        else:
            logger.error(
                "Fan object creation failed, json object incomplete %s",
                json.dumps(json_obj))

        return ret

    def __repr__(self):
        return '{}({},_speed:{})'.format(
            self.__class__.__name__,
            super(Fan, self).__repr__(),
            self._speed
        )


class PDBThermalUpperThresholdCritical(IRule):

    """
    PDB Thermal sensor reports 'UpperThresholdCritical'.
    """

    def __init__(self):
        self.alert_id = 'NV-PDB-01'

    @classmethod
    def condition(cls, pdb):
        if pdb.get_thermal_assertion() == 'UpperThresholdCritical':
            return True


class Pdb(Equipment):

    """
    Pdb: Extends Equipment Class
    """

    _required_json_attribs = set(
        ('Name',
         'MemberId',
         'UpperThresholdCritical',
         'UpperThresholdNonCritical',
         'ReadingCelsius'))

    __rules = [PDBThermalUpperThresholdCritical()]

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def get_name(self):
        return super(Pdb, self).get_name()

    def config(self, config):
        return super(Pdb, self).config(config)

    def __init__(self, entity_id, name, thermal_sensor, hostname, serial):
        super(Pdb, self).__init__(entity_id, name, hostname, serial)
        self._thermal = thermal_sensor

    def get_thermal_assertion(self):
        """
            Gets the pdb thermal assertion error.
        """
        return self._thermal.get_assertion()

    @classmethod
    def create(cls, json_obj, hostname, serial):
        ret = None

        if cls._required_json_attribs.issubset(list(json_obj.keys())):
            sensor = Sensor('Thermal', 'Unknown', 'Celsius',
                            UpperThresholdCritical=json_obj
                            ['UpperThresholdCritical'],
                            UpperThresholdNonCritical=json_obj
                            ['UpperThresholdNonCritical'])
            sensor.set_reading(json_obj['ReadingCelsius'])
            ret = cls(json_obj['MemberId'], json_obj['Name'],
                      sensor, hostname, serial)
        else:
            logger.error(
                "PDB object creation failed, json object incomplete %s",
                json.dumps(json_obj))

        return ret

    def __repr__(self):
        return '{}({},_thermal:{})'.format(
            self.__class__.__name__,
            super(Pdb, self).__repr__(),
            self._thermal
        )


class PsuAbsent(IRule):

    """
    PSU is absent.
    """

    def __init__(self):
        self.alert_id = 'NV-PSU-05'

    @classmethod
    def msg(cls, psu):
        """
           Returns the details for msg
        """
        return [psu.msg()]

    @classmethod
    def chassis_id(cls, psu):
        """
           Returns the chassis id
        """
        return psu.chassis_id

    @classmethod
    def condition(cls, psu):
        if psu.presence == 'Absent':
            return True


class PsuFailed(IRule):

    """
    PSU Status sensor reports that failure is detected in one of the
    subcomponents of the PSU.
    """

    def __init__(self):
        self.alert_id = 'NV-PSU-01'

    @classmethod
    def msg(cls, psu):
        """
           Returns the details for msg
        """
        return [psu.msg()]

    @classmethod
    def chassis_id(cls, psu):
        """
           Returns the chassis id
        """
        return psu.chassis_id

    @classmethod
    def condition(cls, psu):
        if 'Presence detected' in psu.errors and\
           'Power supply Failure detected' in psu.errors:
            return True


class PsuPredictiveFailureDetected(IRule):

    """
    PSU Status sensor reports that predictive failure is detected in one of
    the subcomponents of the PSU.
    """

    def __init__(self):
        self.alert_id = 'NV-PSU-02'

    @classmethod
    def msg(cls, psu):
        """
           Returns the details for msg
        """
        return [psu.msg()]

    @classmethod
    def chassis_id(cls, psu):
        """
           Returns the chassis id
        """
        return psu.chassis_id

    @classmethod
    def condition(cls, psu):
        if 'Presence detected' in psu.errors and\
           'Power supply Predictive Failure' in psu.errors:
            return True


class PsuInputLost(IRule):

    """
    PSU Status sensor reports that input supply is missing.
    """

    def __init__(self):
        self.alert_id = 'NV-PSU-03'

    @classmethod
    def msg(cls, psu):
        """
           Returns the details for msg
        """
        return [psu.msg()]

    @classmethod
    def chassis_id(cls, psu):
        """
           Returns the chassis id
        """
        return psu.chassis_id

    @classmethod
    def condition(cls, psu):
        if 'Presence detected' in psu.errors and\
           'Power Supply AC Lost' in psu.errors:
            return True


class PsuInputLostOutOfRange(IRule):

    """
    PSU Status sensor reports that input power(voltage?) values are out of
    range.
    """

    def __init__(self):
        self.alert_id = 'NV-PSU-04'

    @classmethod
    def msg(cls, psu):
        """
           Returns the details for msg
        """
        return [psu.msg()]

    @classmethod
    def chassis_id(cls, psu):
        """
           Returns the chassis id
        """
        return psu.chassis_id

    @classmethod
    def condition(cls, psu):
        if 'Presence detected' in psu.errors and\
           'Power Supply AC Lost' not in psu.errors and\
           'AC Lost or out-of-range' in psu.errors:
            return True


class PowerSupplyUnit(Equipment):

    """
    PowerSupplyUnit: Extends Equipment Class.
    """

    _required_json_attribs = set(('Name', 'MemberId', 'Oem', 'Status'))

    __rules = [PsuAbsent(), PsuFailed(), PsuInputLost(),
               PsuInputLostOutOfRange(), PsuPredictiveFailureDetected()]

    def __init__(self, entity_id, name, hostname,
                 serial, presence, error, power_state, power_mode, chassis_id):
        super(PowerSupplyUnit, self).__init__(
            entity_id, name, hostname, serial)
        self.presence = presence
        self.errors = list(map(str.strip, error.split('|')))
        self.power_state = power_state
        self.power_mode = power_mode
        self.chassis_id = chassis_id

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def get_name(self):
        return super(PowerSupplyUnit, self).get_name()

    def config(self, config):
        return super(PowerSupplyUnit, self).config(config)

    def msg(self):
        if self.power_state == "Full Power" or \
                (self.power_state == "Reduced Power" and self.power_mode == 'MAXQ'):
            return " System is operating in full power state. "

        elif self.power_state == "Reduced Power" and self.power_mode == 'MAXP':
            return " System is operating in reduced power state. "

        elif self.power_state == "Power Fail":
            return " System is in failed power state. "
        else:
            return ""

    @classmethod
    def create(cls, json_obj, hostname, serial, **kwargs):
        ret = None

        if 'pwr_state' not in list(kwargs.keys()):
            logger.error(
                "Required arguments are not passed to the create method")
            return ret

        presence = None
        if cls._required_json_attribs.issubset(list(json_obj.keys())) and \
           'State' in list(json_obj['Status'].keys()) and \
           'Error' in list(json_obj['Oem']['PSU'].keys()):

            presence = json_obj['Status']['State']
            error = json_obj['Oem']['PSU']['Error']
            power_state = kwargs['pwr_state']
            power_mode = kwargs['pwr_mode']
            # TODO: remove redundant hostname and serial parameters
            host_id = json_obj['@odata.id'].split('/')[4]
            hostname = Endpoint().hostname(host_id)
            serial = Endpoint().serial(host_id)

            ret = cls(json_obj['MemberId'], json_obj['Name'],
                      hostname, serial, presence, error, power_state, power_mode, host_id)

        else:
            logger.error(
                "PSU object creation failed, json object incomplete %s",
                json.dumps(json_obj))
        return ret

    def __repr__(self):
        return '{}({}, presence:{}, errors:{}, power_state:{})'.format(
            self.__class__.__name__, super(
                PowerSupplyUnit, self
            ).__repr__(), self.presence, self.errors, self.power_state)

# Rules for CMM defined here


class CMMFailed(IRule):

    """
    CMM Status reports that the CMM is in failed condition
    """

    def __init__(self):
        self.alert_id = 'NV-CMM-01'

    def msg(self, cmm):
        return [""]

    @classmethod
    def condition(cls, cmm):
        if cmm.state and ('Disabled'.lower() == cmm.state.lower()):
            return True


class CMMSwitchFaulty(IRule):

    """
    CMM Switch is in faulty condition
    """

    def __init__(self):
        self.alert_id = 'NV-CMM-02'

    def msg(self, cmm):
        return [""]

    @classmethod
    def condition(cls, cmm):
        if cmm.health and ('Critical'.lower() == cmm.health.lower()):
            return True


class CMM(Equipment):

    """
    CMM: Extends Equipment Class.
    """

    _required_json_attribs = set(('Name', 'Id', 'Status'))

    __rules = [CMMFailed(), CMMSwitchFaulty()]

    def __init__(self, entity_id, name, hostname,
                 serial, state, health):
        super(CMM, self).__init__(
            entity_id, name, hostname, serial)
        self.state = state
        self.health = health

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def get_name(self):
        return super(CMM, self).get_name()

    def config(self, config):
        return super(CMM, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial, **kwargs):
        ret = None
        if cls._required_json_attribs.issubset(list(json_obj.keys())):
            try:
                state = json_obj['Status']['State']
            except KeyError:
                logger.warning('State key not found')
                state = None
            try:
                health = json_obj['Status']['Health']
            except KeyError:
                logger.warning('Health key not found')
                health = None
            # TODO: remove redundant hostname and serial parameters
            # CMM id in the form Enclosure0_CMM
            chassis_id = json_obj['@odata.id'].split('/')[4].split('_')[0]
            hostname = Endpoint().hostname(chassis_id, "Chassis")
            serial = Endpoint().serial(chassis_id, "Chassis")
            if not serial:
                logger.error(
                    "CMM object creation failed, serial number missing %s",
                    json.dumps(json_obj))
                return None

            ret = cls(json_obj['Id'], json_obj['Id'],
                      hostname, serial, state, health)

        else:
            logger.error(
                "CMM object creation failed, json object incomplete %s",
                json.dumps(json_obj))
        return ret

    def __repr__(self):
        return '{}({}, state:{}, health:{})'.format(
            self.__class__.__name__, super(
                CMM, self
            ).__repr__(), self.state, self.health)

# Rules for Blade defined here


class BladeError(IRule):

    """
    Blade is in failed condition
    """

    def __init__(self):
        self.alert_id = 'NV-BLADE-01'

    def msg(self, blade):
        return [""]

    @classmethod
    def condition(cls, blade):
        if blade.health and ('Critical'.lower() == blade.health.lower()):
            return True


class BladeMissing(IRule):

    """
    Blade is missing
    """

    def __init__(self):
        self.alert_id = 'NV-BLADE-02'

    def msg(self, blade):
        return [""]

    @classmethod
    def condition(cls, blade):
        if blade.state and ('Absent'.lower() == blade.state.lower()):
            return True


class Blade(Equipment):

    """
    Blade: Extends Equipment Class.
    """

    _required_json_attribs = set(('Name', 'Id', 'Status'))

    __rules = [BladeError(), BladeMissing()]

    def __init__(self, entity_id, name, hostname,
                 serial, state, health):
        super(Blade, self).__init__(
            entity_id, name, hostname, serial)
        self.state = state
        self.health = health

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def get_name(self):
        return super(Blade, self).get_name()

    def config(self, config):
        return super(Blade, self).config(config)

    def msg(self):
        return ""

    @classmethod
    def create(cls, json_obj, hostname, serial, **kwargs):
        ret = None
        if cls._required_json_attribs.issubset(list(json_obj.keys())):
            try:
                state = json_obj['Status']['State']
            except KeyError:
                logger.warning('State key not found')
                state = None
            try:
                health = json_obj['Status']['Health']
            except KeyError:
                logger.warning('Health key not found')
                health = None
            # TODO: remove redundant hostname and serial parameters
            chassis_id = json_obj['@odata.id'].split('/')[4].split('_')[0]
            hostname = Endpoint().hostname(chassis_id)
            serial = Endpoint().serial(chassis_id)
            if not serial:
                logger.error(
                    "Blade object creation failed, serial number missing %s",
                    json.dumps(json_obj))
                return None

            ret = cls(json_obj['Id'], json_obj['Id'],
                      hostname, serial, state, health)

        else:
            logger.error(
                "Blade object creation failed, json object incomplete %s",
                json.dumps(json_obj))
        return ret

    def __repr__(self):
        return '{}({}, state:{}, health:{})'.format(
            self.__class__.__name__, super(
                Blade, self
            ).__repr__(), self.state, self.health)

# Rules for Network defined here


class SwitchFailure(IRule):

    """
    Switch module is in failed condition
    """

    def __init__(self):
        self.alert_id = 'NV-ETHSWITCH-01'

    def msg(self, network):
        return [""]

    @classmethod
    def chassis_id(cls, psu):
        """
           Returns the chassis id
        """
        return psu.chassis_id

    @classmethod
    def condition(cls, network):
        if network.state and ('Disabled'.lower() == network.state.lower()):
            return True


class Network(Equipment):

    """
    Network: Extends Equipment Class.
    """

    _required_json_attribs = set(('Name', 'Id', 'Status'))

    __rules = [SwitchFailure()]

    def __init__(self, entity_id, name, hostname,
                 serial, state, chassis_id):
        super(Network, self).__init__(
            entity_id, name, hostname, serial)
        self.state = state
        self.chassis_id = chassis_id

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def get_name(self):
        return super(Network, self).get_name()

    def config(self, config):
        return super(Network, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial, **kwargs):
        ret = None
        if cls._required_json_attribs.issubset(list(json_obj.keys())):
            try:
                state = json_obj['Status']['State']
            except KeyError:
                logger.warning('State key not found')
                state = None
            # TODO: remove redundant hostname and serial parameters
            chassis_id = json_obj['@odata.id'].split('/')[4]
            hostname = Endpoint().hostname(chassis_id)
            serial = Endpoint().serial(chassis_id)
            if not serial:
                logger.error(
                    "Network object creation failed, serial number missing %s",
                    json.dumps(json_obj))
                return None

            ret = cls(json_obj['Id'], json_obj['Name'],
                      hostname, serial, state, chassis_id)
        else:
            logger.error(
                "Network object creation failed, json object incomplete %s",
                json.dumps(json_obj))
        return ret

    def __repr__(self):
        return '{}({}, state:{})'.format(
            self.__class__.__name__, super(
                Network, self
            ).__repr__(), self.state)


class SwitchPortFailed(IRule):

    """
    NetworkPort Status reports that the network port is in failed condition
    """

    def __init__(self):
        self.alert_id = 'NV-ETHSWITCH-02'

    def msg(self, network_port):
        return ["NetworkAdapter {} ".format(network_port.adapter)]

    @classmethod
    def chassis_id(cls, psu):
        """
           Returns the chassis id
        """
        return psu.chassis_id

    @classmethod
    def condition(cls, network_port):
        if (Endpoint().model() == Endpoint().MODEL_MANAGEMENT_SERVER) and \
                network_port.state and ('Disabled'.lower() == network_port.state.lower()):
            return True


class PortFailed(IRule):

    """
    NetworkPort Status reports that the network port is in failed condition
    """

    def __init__(self):
        self.alert_id = 'NV-NET-05'

    def msg(self, network_port):
        return ["NetworkAdapter {} ".format(network_port.adapter)]

    @classmethod
    def chassis_id(cls, psu):
        """
           Returns the chassis id
        """
        return psu.chassis_id

    @classmethod
    def condition(cls, network_port):
        if (Endpoint().model() != Endpoint().MODEL_MANAGEMENT_SERVER) and \
                network_port.state and ('disabled' == network_port.state.lower()):
            return True


class LinkDown(IRule):

    """
    NetworkPort Status reports that the network port is in failed condition
    """

    def __init__(self):
        self.alert_id = 'NV-NET-01'

    def msg(self, network_port):
        return ["NetworkAdapter {} ".format(network_port.adapter)]

    @classmethod
    def condition(cls, network_port):
        if network_port.state and \
            ('enabled' == network_port.state.lower()) and \
                network_port.linkstate and ('Down'.lower() == network_port.linkstate.lower()):
            return True
        return False


class EthNVRAMFailure(IRule):

    """
    NetworkPort Status reports that the network port is in failed condition
    """

    def __init__(self):
        self.alert_id = 'NV-ETH-01'

    def msg(self, network_port):
        return ["NetworkAdapter {} ".format(network_port.adapter)]

    @classmethod
    def condition(cls, network_port):
        if network_port.state and \
            ('enabled' == network_port.state.lower()) and \
            network_port.oem_data and \
            'online_self_test' in network_port.oem_data and \
            'NVRAM' in network_port.oem_data['online_self_test'] and \
                (network_port.oem_data['online_self_test']['NVRAM'].lower() == 'fail'):
            return True
        return False


class IBCounterFailure(IRule):

    """
    NetworkPort Status reports that the network port is in failed condition
    """

    def __init__(self):
        self.alert_id = 'NV-IB-02'

    def msg(self, network_port):
        return ["NetworkAdapter {} ".format(network_port.adapter)]

    @classmethod
    def condition(cls, network_port):
        if network_port.state and \
            ('enabled' == network_port.state.lower()) and \
            network_port.oem_data and \
            'hca_self_test' in network_port.oem_data and \
            'CA_ERROR_COUNTER_CHECK' in network_port.oem_data['hca_self_test'] and \
                (network_port.oem_data['hca_self_test']['CA_ERROR_COUNTER_CHECK'].lower() == 'fail'):
            return True
        return False


class NetworkPort(Equipment):

    """
    NetworkPort: Extends Equipment Class.
    """

    _required_json_attribs = set(('Name', 'Id'))

    __rules = [PortFailed(
    ),
        SwitchPortFailed(),
        LinkDown()]

    def __init__(self, entity_id, name, hostname,
                 serial, state, linkstate, oem_data, adapter, chassis_id):
        super(NetworkPort, self).__init__(
            entity_id, name, hostname, serial)
        self.state = state
        self.linkstate = linkstate
        self.oem_data = oem_data
        self.adapter = adapter
        self.chassis_id = chassis_id

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        rules = cls.__rules.copy()
        try:
            check = equipment.oem_data['hca_self_test']['CA_ERROR_COUNTER_CHECK']
            if check.lower() != 'na':
                rules.append(IBCounterFailure())
        except:
            # key is missing
            pass
        try:
            check = equipment.oem_data['online_self_test']['NVRAM']
            if check.lower() != 'na':
                rules.append(EthNVRAMFailure())
        except:
            # key is missing
            pass

        return rules

    def get_name(self):
        return super(NetworkPort, self).get_name()

    def config(self, config):
        return super(NetworkPort, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial, **kwargs):
        ret = None
        if cls._required_json_attribs.issubset(list(json_obj.keys())):
            try:
                state = json_obj['Status']['State']
            except KeyError:
                logger.warning('State key not found')
                state = None
            try:
                linkstate = json_obj['LinkStatus']
            except KeyError:
                logger.warning('LinkStatus key not found')
                linkstate = None
            try:
                oem_data = json_obj['Oem']
            except KeyError:
                logger.warning('Oem key not found')
                oem_data = None
            # TODO: remove redundant hostname and serial parameters
            chassis_id = json_obj['@odata.id'].split('/')[4]
            hostname = Endpoint().hostname(chassis_id)
            serial = Endpoint().serial(chassis_id)
            if not serial:
                logger.error(
                    "NetworkPort object creation failed, serial number missing %s",
                    json.dumps(json_obj))
                return None
            adapter = json_obj['@odata.id'].split('/')[6]

            # FIXME +id: when we rework primary key for the alerts db, clean up
            # the name prep
            ret = cls(json_obj['Id'],
                      '{}+{}'.format(adapter, json_obj['Id']),
                      hostname, serial, state, linkstate, oem_data, adapter, chassis_id)

        else:
            logger.error(
                "NetworkPort object creation failed, json object incomplete %s",
                json.dumps(json_obj))
        return ret

    def __repr__(self):
        return '{}({}, state:{}, linkstate{}, oem:{})'.format(
            self.__class__.__name__, super(
                NetworkPort, self
            ).__repr__(), self.state, self.linkstate, self.oem_data)


class NVSwitchFatalError(IRule):

    """
    NVSwitch or the NVSwitch subsystem has reported a fatal error.
    """

    def __init__(self):
        self.alert_id = 'NV-NVSWITCH-01'

    @classmethod
    def condition(cls, nvswitch):
        if nvswitch.error == 'nvswitchfatalerror' and \
                nvswitch.sno != "OK":
            return True

        return False

    @classmethod
    def msg(cls, nvswitch):
        """
           Returns the details for msg
        """
        return [nvswitch.nvswitch_id, nvswitch.sno]


class NVSwitchNonFatalError(IRule):

    """
    NVSwitch or the NVSwitch subsystem has reported a non fatal error.
    """

    def __init__(self):
        self.alert_id = 'NV-NVSWITCH-02'

    @classmethod
    def condition(cls, nvswitch):
        if nvswitch.error == 'nvswitchnonfatalerror' and \
                nvswitch.sno != "OK":
            return True

        return False

    @classmethod
    def msg(cls, nvswitch):
        """
           Returns the details for msg
        """
        return [nvswitch.nvswitch_id, nvswitch.sno]


class NVSwitchNVLinkStateError(IRule):

    """
    NVSwitch or the NVSwitch subsystem has reported a fatal error.
    """

    def __init__(self):
        self.alert_id = 'NV-NVLINK-01'

    @classmethod
    def condition(cls, nvswitch):
        if "nvlink" in nvswitch.error and \
                nvswitch.sno != "UP":
            return True

        return False

    @classmethod
    def msg(cls, nvswitch):
        """
           Returns the details for msg
        """
        return [nvswitch.nvswitch_id, nvswitch.error, nvswitch.sno, "UP"]

# NVSwitch class defined here.


class NVSwitch(Equipment):
    """
    NVSwitch: Extends Equipment Class.
    """
    __rules = [NVSwitchFatalError(),
               NVSwitchNonFatalError(),
               NVSwitchNVLinkStateError()]
    _required_json_attribs = set(('Error', 'NVSwitchId', 'SNo'))

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def __init__(self, entity_id, name, hostname, serial, error, nvswitch_id, sno):
        super(NVSwitch, self).__init__(entity_id, name, hostname, serial)
        self.error = error
        self.nvswitch_id = nvswitch_id
        self.sno = sno

    def get_name(self):
        return super(NVSwitch, self).get_name()

    def config(self, config):
        return super(NVSwitch, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial):
        ret = None

        if cls._required_json_attribs.issubset(list(json_obj.keys())):
            ret = cls(
                json_obj['Id'],
                json_obj['Name'],
                hostname,
                serial,
                json_obj['Error'],
                json_obj['NVSwitchId'],
                json_obj['SNo'])
        else:
            logger.error(
                "NVSwitch Health object creation failed, json object incomplete %s",
                json.dumps(json_obj))

        return ret

    def __repr__(self):
        return '{}({},NVSwitchId:{},Error:{},SNo:{})'.format(
            self.__class__.__name__,
            super(NVSwitch, self).__repr__(),
            self.nvswitch_id,
            self.error,
            self.sno
        )


class NVLinkCRCFlitError(IRule):

    """
    NVSwitch NVLink has reported a CRC FLIT errors.
    """

    def __init__(self):
        self.alert_id = 'NV-NVLINK-02'

    @classmethod
    def condition(cls, nvlink):
        if "nvlinkcrcfliterrorcount" in nvlink.error and \
                nvlink.error_count != "OK":
            return True

        return False

    @classmethod
    def msg(cls, nvlink):
        """
           Returns the details for msg
        """
        return [nvlink.error_count, nvlink.gpu_id, nvlink.lane_id]


class NVLinkCRCDataError(IRule):

    """
    GPU NVLink has reported a CRC Data errors.
    """

    def __init__(self):
        self.alert_id = 'NV-NVLINK-03'

    @classmethod
    def condition(cls, nvlink):
        if "nvlinkcrcdataerrorcount" in nvlink.error and \
                nvlink.error_count != "OK":
            return True

        return False

    @classmethod
    def msg(cls, nvlink):
        """
           Returns the details for msg
        """
        return [nvlink.error_count, nvlink.gpu_id, nvlink.lane_id]


class NVLinkReplayError(IRule):

    """
    GPU NVLink has reported a Replay errors.
    """

    def __init__(self):
        self.alert_id = 'NV-NVLINK-04'

    @classmethod
    def condition(cls, nvlink):
        if "nvlinkreplayerrorcount" in nvlink.error and \
                nvlink.error_count != "OK":
            return True

        return False

    @classmethod
    def msg(cls, nvlink):
        """
           Returns the details for msg
        """
        return [nvlink.error_count, nvlink.gpu_id, nvlink.lane_id]


class NVLinkRecoveryError(IRule):

    """
    GPU NVLink has reported a Recovery errors.
    """

    def __init__(self):
        self.alert_id = 'NV-NVLINK-05'

    @classmethod
    def condition(cls, nvlink):
        if "nvlinkrecoveryerrorcount" in nvlink.error and \
                nvlink.error_count != "OK":
            return True

        return False

    @classmethod
    def msg(cls, nvlink):
        """
           Returns the details for msg
        """
        return [nvlink.error_count, nvlink.gpu_id, nvlink.lane_id]


# NVLink class defined here.
class NVLink(Equipment):
    """
    NVLink: Extends Equipment Class.
    """
    __rules = [NVLinkCRCFlitError(),
               NVLinkCRCDataError(),
               NVLinkReplayError(),
               NVLinkRecoveryError()]
    _required_json_attribs = set(('Error', 'GpuId', 'Count', 'LaneId'))

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def __init__(self, entity_id, name, hostname, serial, error, gpu_id=None,
                 error_count=None, lane_id=None):
        super(NVLink, self).__init__(entity_id, name, hostname, serial)
        self.error = error
        self.gpu_id = gpu_id
        self.error_count = error_count
        self.lane_id = lane_id

    def get_name(self):
        return super(NVLink, self).get_name()

    def config(self, config):
        return super(NVLink, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial):
        ret = None

        if cls._required_json_attribs.issubset(list(json_obj.keys())):
            ret = cls(
                json_obj['Id'],
                json_obj['Name'],
                hostname,
                serial,
                json_obj['Error'],
                json_obj['GpuId'],
                json_obj['Count'],
                json_obj['LaneId'])
        else:
            logger.error(
                "NVLink Health object creation failed, json object incomplete %s",
                json.dumps(json_obj))

        return ret

    def __repr__(self):
        return '{}({},Error:{},ErrorCount:{},GPUId:{},LaneId:{})'.format(
            self.__class__.__name__,
            super(NVLink, self).__repr__(),
            self.error,
            self.error_count,
            self.gpu_id,
            self.lane_id)


class MACMissing(IRule):

    """
    NetworkDevFunc Status reports that mac is missing
    """

    def __init__(self):
        self.alert_id = 'NV-ETH-02'

    def msg(self, dev):
        return ["NetworkAdapter {} ".format(dev.adapter)]

    @classmethod
    def condition(cls, dev):
        try:
            eth = dev.oem_data["nvidia"]["Ethernet"]
            if 'MACAddress' not in eth or \
                    not eth["MACAddress"]:
                return True
        except:
            # mostly key not found
            pass
        return False


class GUIDMissing(IRule):

    """
    NetworkDevFunc Status reports that GUID is missing
    """

    def __init__(self):
        self.alert_id = 'NV-IB-03'

    def msg(self, dev):
        return ["NetworkAdapter {} ".format(dev.adapter)]

    @classmethod
    def condition(cls, dev):
        try:
            ib = dev.oem_data["nvidia"]["Infiniband"]
            if 'GUIDAddress' not in ib or \
                    not ib["GUIDAddress"]:
                return True
        except:
            # mostly key not found
            pass
        return False


class RxCollisionRate(IRule):

    """
    NetworkDevFunc Status reports that rx collision rate crossed threshold
    """

    def __init__(self):
        self.alert_id = 'NV-NET-02'

    def msg(self, dev):
        return [f"{dev.oem_data['Counters']['rx_collision_rate']}", f"{dev.rx_collision_threshold}", f"NetworkAdapter {dev.adapter} "]

    @classmethod
    def condition(cls, dev):
        if dev.oem_data and \
            'Counters' in dev.oem_data and \
            'rx_collision_rate' in dev.oem_data['Counters'] and \
            dev.rx_collision_threshold and \
                (dev.oem_data['Counters']['rx_collision_rate'] >= dev.rx_collision_threshold):
            return True
        return False


class TxCollisionRate(IRule):

    """
    NetworkDevFunc Status reports that tx collision rate crossed threshold
    """

    def __init__(self):
        self.alert_id = 'NV-NET-03'

    def msg(self, dev):
        return [f"{dev.oem_data['Counters']['tx_collision_rate']}",
                f"{dev.tx_collision_threshold}",
                f"NetworkAdapter {dev.adapter} "]

    @classmethod
    def condition(cls, dev):
        if dev.oem_data and \
            'Counters' in dev.oem_data and \
            'tx_collision_rate' in dev.oem_data['Counters'] and \
            dev.tx_collision_threshold and \
                (dev.oem_data['Counters']['tx_collision_rate'] >= dev.tx_collision_threshold):
            return True
        return False


class RxCRCRate(IRule):

    """
    NetworkDevFunc Status reports that rx crc rate crossed threshold
    """

    def __init__(self):
        self.alert_id = 'NV-NET-04'

    def msg(self, dev):
        return [f"{dev.oem_data['Counters']['crc_error_rate']}",
                f"{dev.rx_crc_threshold}",
                f"NetworkAdapter {dev.adapter} "]

    @classmethod
    def condition(cls, dev):
        if dev.oem_data and \
            'Counters' in dev.oem_data and \
            'crc_error_rate' in dev.oem_data['Counters'] and \
            dev.rx_crc_threshold and \
                (dev.oem_data['Counters']['crc_error_rate'] >= dev.rx_crc_threshold):
            return True
        return False


class NetworkDevFunc(Equipment):

    """
    NetworkDevFunc: Extends Equipment Class.
    """

    _required_json_attribs = set(('Name', 'Id'))

    __rules = [MACMissing(), RxCollisionRate(),
               TxCollisionRate(), RxCRCRate(), GUIDMissing()]

    def __init__(self, entity_id, name, hostname,
                 serial, oem_data, adapter, chassis_id):
        super(NetworkDevFunc, self).__init__(
            entity_id, name, hostname, serial)
        self.oem_data = oem_data
        self.adapter = adapter
        self.chassis_id = chassis_id
        self.rx_collision_threshold = None
        self.rx_crc_threshold = None
        self.tx_collision_threshold = None

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def get_name(self):
        return super(NetworkDevFunc, self).get_name()

    def config(self, config):
        try:
            self.rx_collision_threshold = config.get_val(
                'rx_collision_threshold')
            self.rx_crc_threshold = config.get_val('rx_crc_threshold')
            self.tx_collision_threshold = config.get_val(
                'tx_collision_threshold')
        except Exception as _:
            pass

    @classmethod
    def create(cls, json_obj, hostname, serial, **kwargs):
        ret = None
        if cls._required_json_attribs.issubset(list(json_obj.keys())):
            try:
                oem_data = json_obj['Oem']
            except KeyError:
                logger.warning('Oem key not found')
                oem_data = None
            # TODO: remove redundant hostname and serial parameters
            chassis_id = json_obj['@odata.id'].split('/')[4]
            hostname = Endpoint().hostname(chassis_id)
            serial = Endpoint().serial(chassis_id)
            if not serial:
                logger.error(
                    "NetworkDevFunc object creation failed, serial number missing %s",
                    json.dumps(json_obj))
                return None
            adapter = json_obj['@odata.id'].split('/')[6]

            # FIXME +id: when we rework primary key for the alerts db, clean up
            # the name prep
            ret = cls(json_obj['Id'],
                      '{}+{}'.format(adapter, json_obj['Id']),
                      hostname, serial, oem_data, adapter, chassis_id)

        else:
            logger.error(
                "NetworkDevFunc object creation failed, json object incomplete %s",
                json.dumps(json_obj))
        return ret

    def __repr__(self):
        return '{}({}, oem:{})'.format(
            self.__class__.__name__, super(
                NetworkDevFunc, self
            ).__repr__(), self.oem_data)


class IBDriverInitFailure(IRule):

    """
    NetworkAdapter Status reports that driver initialization failed
    """

    def __init__(self):
        self.alert_id = 'NV-IB-01'

    @classmethod
    def condition(cls, dev):
        if dev.oem_data and \
                'hca_self_test' in dev.oem_data:
            keys = ["PCI_DEV_CHECK", "DRIVER_RPM_CHECK",
                    "DRIVER_INIT", "KERNEL_SYSLOG_CHECK"]
            for k in keys:
                if k in dev.oem_data['hca_self_test']:
                    if dev.oem_data['hca_self_test'][k].lower() == "fail":
                        return True
        return False


class NetworkAdapter(Equipment):

    """
    NetworkAdapter: Extends Equipment Class.
    """

    _required_json_attribs = set(('Name',))

    __rules = [IBDriverInitFailure()]

    def __init__(self, name, hostname,
                 serial, oem_data, chassis_id):
        super(NetworkAdapter, self).__init__(
            name, name, hostname, serial)
        self.oem_data = oem_data
        self.chassis_id = chassis_id

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def get_name(self):
        return super(NetworkAdapter, self).get_name()

    def config(self, config):
        return super(NetworkAdapter, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial, **kwargs):
        ret = None
        if cls._required_json_attribs.issubset(list(json_obj.keys())):
            try:
                oem_data = json_obj['Oem']
            except KeyError:
                logger.warning('Oem key not found')
                oem_data = None
            # TODO: remove redundant hostname and serial parameters
            chassis_id = json_obj['@odata.id'].split('/')[4]
            hostname = Endpoint().hostname(chassis_id)
            serial = Endpoint().serial(chassis_id)
            if not serial:
                logger.error(
                    "NetworkAdapter object creation failed, serial number missing %s",
                    json.dumps(json_obj))
                return None

            # FIXME +id: when we rework primary key for the alerts db, clean up
            # the name prep
            ret = cls(json_obj['Name'],
                      hostname, serial, oem_data, chassis_id)

        else:
            logger.error(
                "NetworkAdapter object creation failed, json object incomplete %s",
                json.dumps(json_obj))
        return ret

    def __repr__(self):
        return '{}({}, oem:{})'.format(
            self.__class__.__name__, super(
                NetworkAdapter, self
            ).__repr__(), self.oem_data)


# Rules for Dimm defined here.
class DimmUncorrectableError(IRule):

    """
    Dimm reports 'uncorrectable' error.
    """

    def __init__(self):
        """
        Init
        """
        self.alert_id = 'NV-DIMM-01'

    @classmethod
    def condition(cls, dimm):
        if dimm.uncorr_count != 0:
            return True


class DimmCorrectableError(IRule):

    """
    Dimm reports 'correctable' error.
    """

    def __init__(self):
        """
        Init
        """
        self.alert_id = 'NV-DIMM-02'

    @classmethod
    def condition(cls, dimm):
        if dimm.corr_count != 0:
            return True

        return False


class DimmError(IRule):

    """
    Dimm reports 'correctable' error.
    """

    def __init__(self):
        """
        Init
        """
        self.alert_id = 'NV-DIMM-03'

    @classmethod
    def condition(cls, dimm):
        # This alert is deprecated, raises Memory misconfiguration alert instead.
        # The rule is left as is, so that any existing alerts NV-DIMM-03 alerts in the DB are automatically cleared.
        # if dimm.state == 'Present_disabled':
        #    return True

        return False


class DimmThrottleError(IRule):

    """
    Dimm reports 'Throttle/Overheat' error.
    """

    def __init__(self):
        """
        Init
        """
        self.alert_id = 'NV-DIMM-06'

    @classmethod
    def condition(cls, dimm):
        if dimm.throttle_count != 0:
            return True

        return False


class DimmUnExpected(IRule):

    """
    Dimm reports 'correctable' error.
    """

    def __init__(self):
        """
        Init
        """
        self.alert_id = 'NV-DIMM-04'

    @classmethod
    def condition(cls, dimm):
        # This alert is deprecated, raises Memory misconfiguration alert instead.
        # The rule is left as is, so that any existing alerts NV-DIMM-04 alerts in the DB are automatically cleared.
        # if dimm.state == 'NotPresent_enabled':
        #    return True

        return False

# DIMM class defined here.


class Dimm(Equipment):

    """
    Dimm: Extends Equipment Class.
    """
    _required_json_attribs = set(('CorrectableCount',
                                  'CorrectableTimestamp',
                                  'UncorrectableCount',
                                  'DimmThrottleCount'))

    __rules = [DimmCorrectableError(), DimmUncorrectableError(),
               DimmError(), DimmUnExpected(), DimmThrottleError()]

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def __init__(self, entity_id, name, hostname, serial,
                 corr_cnt, corr_times, uncorr_cnt, throttle_cnt):
        """
        Init
        """
        super(Dimm, self).__init__(entity_id, name, hostname, serial)
        self.corr_count = corr_cnt
        self.corr_time_stamps = corr_times
        self.uncorr_count = uncorr_cnt
        self.throttle_count = throttle_cnt

        if self.corr_time_stamps is not None:
            for (index, timestamp) in enumerate(self.corr_time_stamps):
                # TODO: Ignoring timezone for now.
                self.corr_time_stamps[index] = \
                    datetime.datetime.strptime(
                        timestamp[0:19], '%Y-%m-%dT%H:%M:%S')

            self.corr_time_stamps.sort()

        # Default threshold is 60 minutes, and count is 1000 errors
        self.threshold = {'duration': 3600, 'count': 10}

    def get_name(self):
        return super(Dimm, self).get_name()

    def config(self, config):
        required_attribs = set(('duration', 'count'))
        if required_attribs.issubset(config.keys()):
            self.threshold['duration'] = config['duration']
            self.threshold['count'] = config['count']

    @classmethod
    def create(cls, json_obj, hostname, serial):
        ret = None

        if 'Oem' in json_obj.keys() and \
           'Error' in json_obj['Oem'].keys() and \
           cls._required_json_attribs.issubset(
            json_obj['Oem']['Error'].keys()
        ):
            error_obj = json_obj['Oem']['Error']
            ret = cls(json_obj['Id'],
                      json_obj['Id'],
                      hostname,
                      serial,
                      error_obj['CorrectableCount'],
                      error_obj['CorrectableTimestamp'],
                      error_obj['UncorrectableCount'],
                      error_obj['DimmThrottleCount'])
        else:
            logger.error(
                "Dimm object creation failed, json object incomplete %s",
                json.dumps(json_obj))
        return ret

    def __repr__(self):
        return '{}({},corr_cnt:{},uncorr_cnt:{},threshold:{})'.format(
            self.__class__.__name__,
            super(Dimm, self).__repr__(),
            self.corr_count,
            self.uncorr_count,
            self.threshold)


class MemoryTopologyUnsupported(IRule):

    """
    MemoryCollection reports Unsupported configuration error.
    """

    def __init__(self):
        """
        Init
        """
        self.alert_id = 'NV-DIMM-05'

    @classmethod
    def condition(cls, mem_collection):
        for topology in mem_collection.supported_topologies:
            if set(topology) == set(mem_collection.actual_topology):
                return False

        return True

    @classmethod
    def msg(cls, mem_collection):
        """
           Returns the details for msg
        """
        fmt_supported_topology = ''
        for topology in mem_collection.supported_topologies:
            fs = "\n\t\t\t{}".format(sorted(topology))
            fmt_supported_topology += fs

        return [len(mem_collection.actual_topology), sorted(mem_collection.actual_topology), fmt_supported_topology]


# Memory Collection class defined here.
class MemoryCollection(Equipment):

    """
    MemoryCollection: Extends Equipment Class.
    """
    __rules = [MemoryTopologyUnsupported()]

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def __init__(self, entity_id, name, hostname, serial,
                 actual_topology, supported_topologies):
        """
        Init
        """
        super(MemoryCollection, self).__init__(
            entity_id, name, hostname, serial)
        self.actual_topology = actual_topology
        self.supported_topologies = supported_topologies

    def get_name(self):
        return super(MemoryCollection, self).get_name()

    def config(self, config):
        return super(MemoryCollection, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial):
        ret = None

        supported_topologies = pa().enumerate_obj(
            'SystemCollection/1/Memory/SupportedConfiguration/Topologies')

        if supported_topologies and 'Members' in json_obj.keys():
            members = json_obj['Members']
            actual_topology = []
            for member in members:
                if '@odata.id' not in list(member.keys()):
                    logger.error("@odata.id key missing from member object")
                    return ret

                dimm_slot = member['@odata.id'].split('/')[-1]
                actual_topology.append(dimm_slot)

            ret = cls('MemorySubsystem',
                      'MemorySubsystem',
                      hostname,
                      serial,
                      actual_topology,
                      supported_topologies)
        else:
            logger.error(
                "MemoryCollection object creation failed, json object incomplete %s",
                json.dumps(json_obj))
        return ret

    def __repr__(self):
        return '{}({},Actual:{},Supported:{})'.format(
            self.__class__.__name__,
            super(MemoryCollection, self).__repr__(),
            self.actual_toplogy,
            self.supported_topologies
        )


# Rules for CPU defined here.
class CpuIErrRule(IRule):

    """
    CPU reports IErr.
    """

    def __init__(self):
        """
        Init
        """
        self.alert_id = 'NV-CPU-01'

    @classmethod
    def condition(cls, cpu):
        if cpu.ierror_count != 0:
            return True


class CpuThermtripRule(IRule):

    """
    CPU reports Thermtrip.
    """

    def __init__(self):
        """
        Init
        """
        self.alert_id = 'NV-CPU-02'

    @classmethod
    def condition(cls, cpu):
        if cpu.thermtrip_count != 0:
            return True


class CpuProcHotRule(IRule):

    """
    CPU reports Proc Hot.
    """

    def __init__(self):
        """
        Init
        """
        self.alert_id = 'NV-CPU-03'

    @classmethod
    def condition(cls, cpu):
        if cpu.cpu0_hot != 0:
            return True


class CpuSysThrottleRule(IRule):

    """
    CPU reports Sys Throttle.
    """

    def __init__(self):
        """
        Init
        """
        self.alert_id = 'NV-CPU-04'

    @classmethod
    def condition(cls, cpu):
        if cpu.sys_throttle != 0:
            return True


class CpuPchThermTripRule(IRule):

    """
    CPU reports Pch Therm Trip.
    """

    def __init__(self):
        """
        Init
        """
        self.alert_id = 'NV-PCH-01'

    @classmethod
    def condition(cls, cpu):
        if cpu.pch_therm_trip != 0:
            return True

# CPU class defined here.


class Cpu(Equipment):

    """
    Cpu: Extends Equipment Class.
    """
    _required_json_attribs = set(('IerrCount', 'ThermTripCount'))

    __rules = [CpuIErrRule(
    ),
        CpuThermtripRule(),
        CpuProcHotRule(),
        CpuSysThrottleRule(),
        CpuPchThermTripRule()]

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def __init__(self, entity_id, name, hostname,
                 serial, error_obj):
        super(Cpu, self).__init__(entity_id, name, hostname, serial)
        self.ierror_count = error_obj['IerrCount']
        self.thermtrip_count = error_obj['ThermTripCount']
        self.cpu0_hot = error_obj.get('Cpu0Hot', 0)
        self.sys_throttle = error_obj.get('SysThrottle', 0)
        self.pch_therm_trip = error_obj.get('PchThermTrip', 0)

    def get_name(self):
        return super(Cpu, self).get_name()

    def config(self, config):
        return super(Cpu, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial):
        ret = None

        if 'Oem' in list(json_obj.keys()) and \
           'Error' in list(json_obj['Oem'].keys()) and \
           cls._required_json_attribs.issubset(
            list(json_obj['Oem']['Error'].keys())
        ):
            error_obj = json_obj['Oem']['Error']
            ret = cls(json_obj['Id'],
                      json_obj['Id'],
                      hostname,
                      serial,
                      error_obj)
        else:
            logger.error(
                "Cpu object creation failed, json object incomplete %s",
                json.dumps(json_obj))

        return ret

    def __repr__(self):
        return '{}({},ierror_cnt:{},thermtrip_cnt:{})'.format(
            self.__class__.__name__,
            super(Cpu, self).__repr__(),
            self.ierror_count,
            self.thermtrip_count
        )


class GpuCriticalError(IRule):

    """
    GPU or the GPU subsystem has reported a critical error.
    """

    def __init__(self):
        self.alert_id = 'NV-GPU-01'

    @classmethod
    def condition(cls, gpu):
        if cls.generic_alert_supported(gpu) and gpu.health == 'Critical':
            return True

        return False

    @staticmethod
    def generic_alert_supported(gpu):
        """
        Check if generic alerting is supported for this gpu instance.
        """
        if gpu.incidents is None:
            return True

        for incident in gpu.incidents:
            if 'Errors' in incident.keys() and incident['Errors'] is not None:
                for error in incident['Errors']:
                    if 'Code' in error.keys() and error['Code'] is not None:
                        logger.debug(
                            "Detailed error information available, NV-GPU-01 alert is not supported")
                        return False
        return True


# GPU class defined here.
class Gpu(Equipment):

    """
    Gpu: Extends Equipment Class.
    """
    __rules = [GpuCriticalError()]
    _required_json_attribs = set(('Name', 'Health', 'Incidents'))

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def __init__(self, entity_id, name, hostname, serial, health, incidents):
        super(Gpu, self).__init__(entity_id, name, hostname, serial)
        self.health = health
        self.incidents = incidents

    def get_name(self):
        return super(Gpu, self).get_name()

    def config(self, config):
        return super(Gpu, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial):
        ret = None

        if cls._required_json_attribs.issubset(list(json_obj.keys())):
            ret = cls(
                json_obj['Name'],
                json_obj['Name'],
                hostname,
                serial, json_obj['Health'], json_obj['Incidents'])
        else:
            logger.error(
                "GPU Health object creation failed, json object incomplete %s",
                json.dumps(json_obj))

        return ret

    def __repr__(self):
        return '{}({},health:{},incidents:{})'.format(
            self.__class__.__name__,
            super(Gpu, self).__repr__(),
            self.health, self.incidents
        )


class ReducedPowerState(IRule):

    """
    System is operating in reduced power state.
    """

    def __init__(self):
        self.alert_id = 'NV-PSU-06'

    @classmethod
    def condition(cls, power):
        # System is in reduced power state only when in MaxP.
        if power.state == 'Reduced Power' and power.mode == 'MAXP':
            return True
        else:
            return False


class FailedPowerState(IRule):

    """
    System is in failed power state.
    """

    def __init__(self):
        self.alert_id = 'NV-PSU-07'

    @classmethod
    def condition(cls, power):
        if power.state == 'Power Fail':
            return True
        else:
            return False


class PowerState(Equipment):

    """
    PowerState: Extends Equipment Class.
    """
    __rules = [ReducedPowerState(), FailedPowerState()]

    def __init__(self, entity_id, name,
                 hostname, serial, power_state, power_mode):
        super(PowerState, self).__init__(
            entity_id, name, hostname, serial)
        self.state = power_state
        self.mode = power_mode

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        # If power state is not set properly, it indicates that Thea is unable to fetch the PowerState.
        # Do not run any rules, just leave the alerts as they are.
        if equipment.state in ["Unknown", "", "Failed"]:
            return []
        else:
            return cls.__rules

    def get_name(self):
        return super(PowerState, self).get_name()

    def config(self, config):
        return super(PowerState, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial):
        ret = None

        try:
            power_mode = json_obj['Oem']['Power']['Mode']
            power_state = json_obj['Oem']['Power']['State']
        except Exception:
            logger.exception(
                "PowerState object creation failed, json object incomplete %s", json.dumps(json_obj))
            return ret

        # TODO: remove redundant hostname and serial parameters
        host_id = json_obj['@odata.id'].split('/')[4]
        hostname = Endpoint().hostname(host_id)
        serial = Endpoint().serial(host_id)

        ret = cls('PSU Subsystem',
                  'PSU Subsystem',
                  hostname,
                  serial,
                  power_state,
                  power_mode)

        if ret:
            logger.debug("Powerstate object created {}".format(ret))

        return ret

    def __repr__(self):
        return '{}({}, power_state:{}, power_mode:{})'.format(
            self.__class__.__name__, super(
                PowerState, self
            ).__repr__(), self.state, self.mode)


class GpuLimitMisconfigError(IRule):

    """
    GPU limits are misconfigured.
    """

    def __init__(self):
        self.alert_id = 'NV-GPU-02'
        self.pwr_mode = None
        self.pwr_limit = None

    @classmethod
    def condition(cls, gpu):

        limits_uri = 'SystemCollection/1/GPU/PowerLimits'
        limits = pa().enumerate_obj(limits_uri)
        gpu_count_uri = 'SystemCollection/1/GPU/Count'
        rest_of_system_uri = 'SystemCollection/1/GPU/RestofSystemPower'

        # PowerMode is unsupported when:
        # 1. DGX1 Platform.
        # 2. KVM Mode enabled.
        # 3. NVIDIA Driver is not loaded.
        # When PowerMode is unsupported, DSHM should not raise GPU MisConfig alert.
        # This check below will ensure that any GPU MisConfig alerts raised are
        # cleared.
        if gpu.pwr_mode == 'Not Supported':
            return False

        # Get the limits from the PA file and use that to generate the alert
        if limits:

            # In case of LUNA, powermode is in the form: MaxP(6500W), Custom(5000W)
            # Extract the PowerMode i.e. MaxP, Custom
            cls.pwr_mode = re.sub(r'\([^)]*\)', '', gpu.pwr_mode).upper()
            cls.pwr_limit = limits[cls.pwr_mode][gpu.pwr_state]['Power']

            # Check if PowerMode and PowerState are present in the Platform Definition file
            if cls.pwr_mode in list(limits.keys()) and gpu.pwr_state in list(limits[cls.pwr_mode].keys()):

                # Check if PowerMode is set to Custom
                # Calculate custom perGPU Power for custom PowerMode
                if cls.pwr_mode == 'CUSTOM':
                    try:

                        # Powermode is in the form: Custom(5000W)
                        # Extract power: 5000
                        pwr_limit = re.findall(
                            '\((.*?)\)', gpu.pwr_mode)[0].rstrip("W")

                        # Get RestofSystemPower from platform definition file
                        pwr_rest_system = int(
                            pa().enumerate_obj(rest_of_system_uri))

                        # Get GPU count from the platform definition file
                        gpu_count = int(pa().enumerate_obj(gpu_count_uri))

                        # Get perGPU Power using the custom power
                        # Formula used in case of LUNA:
                        #   GPU_TGP = (ChassisPowerLimit – RestOfSystemPower) / NumGPUs
                        cls.pwr_limit = (
                            int(pwr_limit) - pwr_rest_system) * 1000 / int(gpu_count)
                    except Exception as e:
                        logger.error(
                            "ERROR: POWERMODE: {} POWER_LIMIT: {} EXCEPTION:{}".format(cls.pwr_mode, cls.pwr_limit, str(e)))
                        pass

                if gpu.pwr_limit != cls.pwr_limit or gpu.clk_limit != limits[cls.pwr_mode][gpu.pwr_state]['Clock']:
                    return True

        return False

    @classmethod
    def msg(cls, gpu):
        """
           Returns the details for msg
        """
        limits_uri = 'SystemCollection/1/GPU/PowerLimits'
        limits = pa().enumerate_obj(limits_uri)

        return [cls.pwr_limit, limits[cls.pwr_mode][gpu.pwr_state]['Clock'], gpu.pwr_limit, gpu.clk_limit]


# GPU Limits class defined here.
class GpuLimits(Equipment):

    """
    Gpu: Extends Equipment Class.
    """
    __rules = [GpuLimitMisconfigError()]
    _required_json_attribs = set(
        ('Name',
         'EnforcedPowerLimit',
         'MaxCustomerBoostClock'))

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        # Even if pwr_state is unknown, if pwr_mode is
        # unsupported go ahead and clear the alerts.
        if equipment.pwr_mode == "Not Supported":
            return cls.__rules

        # If power state or power mode is not set properly, it indicates that Thea is unable to fetch the data.
        # Do not run any rules, just leave the alerts as they are.
        if (equipment.pwr_state in ["Unknown", "", "Failed"]) or \
                (equipment.pwr_mode in ["Unknown", "", "Failed", "Inconsistent"]):
            return []

        return cls.__rules

    def __init__(self, entity_id, name, hostname,
                 serial, pwr_limit, clk_limit, pwr_state, pwr_mode):
        super(GpuLimits, self).__init__(entity_id, name, hostname, serial)
        self.pwr_limit = pwr_limit
        self.clk_limit = clk_limit
        self.pwr_state = pwr_state
        self.pwr_mode = pwr_mode

    def get_name(self):
        return super(GpuLimits, self).get_name()

    def config(self, config):
        return super(GpuLimits, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial, **kwargs):
        ret = None

        kwargs_rqrd_set = set(('power_state', 'power_mode'))
        if not kwargs_rqrd_set.issubset(list(kwargs.keys())):
            logger.error(
                "Required arguments are not passed to the create method")
            return ret

        if cls._required_json_attribs.issubset(list(json_obj.keys())):
            ret = cls(json_obj['Name'],
                      json_obj['Name'],
                      hostname,
                      serial,
                      json_obj['EnforcedPowerLimit'],
                      json_obj['MaxCustomerBoostClock'],
                      kwargs['power_state'],
                      kwargs['power_mode'])
        else:
            logger.error(
                "GPU Limits object creation failed, json object incomplete %s",
                json.dumps(json_obj))

        return ret

    def __repr__(self):
        return '{}({},pwr_limit:{}, clk_limit:{}, pwr_state:{}, pwr_mode:{})'.format(
            self.__class__.__name__,
            super(GpuLimits, self).__repr__(),
            self.pwr_limit, self.clk_limit, self.pwr_state, self.pwr_mode)


class PCIeLinkSpeedNotMatchingError(IRule):

    """
    PCIe speed mismatch
    """
    pcie_speed = {
        1: '2.5GT/s',
        2: '5GT/s',
        3: '8GT/s',
        4: '16GT/s',
    }

    def __init__(self):
        self.alert_id = 'NV-PCI-01'

    @classmethod
    def condition(cls, pcie):
        try:
            health = pcie.json_obj['Status']['Health']
            state = pcie.json_obj['Status']['State']
            if state == 'Enabled' and health == 'Critical':
                current = pcie.json_obj['PCIeInterface']['CurrentSpeed']
                expected = pcie.json_obj['PCIeInterface']['ExpectedSpeed']
                if current != expected:
                    device_type = pcie.json_obj['PCIeInterface']['DeviceType']
                    if device_type == 'Endpoint' or \
                            device_type == 'UpstreamPort':
                        return True
        except Exception:
            return False
        return False

    @classmethod
    def msg(cls, pcie):
        """
           Returns the details for msg
        """
        bdf_1 = pcie.json_obj['BDF']
        bdf_2 = pcie.json_obj['PCIeInterface']['ParentChild']['BDF']
        msg_list = [bdf_1 + ' ' + bdf_2,
                    cls.pcie_speed[
                        pcie.json_obj['PCIeInterface']['ExpectedSpeed']],
                    cls.pcie_speed[
                        pcie.json_obj['PCIeInterface']['CurrentSpeed']]
                    ]
        return msg_list


class PCIeLinkWidthNotMatchingError(IRule):

    """
    PCIe width mismatch
    """

    def __init__(self):
        self.alert_id = 'NV-PCI-02'

    @classmethod
    def condition(cls, pcie):
        try:
            health = pcie.json_obj['Status']['Health']
            state = pcie.json_obj['Status']['State']
            if state == 'Enabled' and health == 'Critical':
                current = pcie.json_obj['PCIeInterface']['CurrentWidth']
                expected = pcie.json_obj['PCIeInterface']['ExpectedLanes']
                if current != expected:
                    device_type = pcie.json_obj['PCIeInterface']['DeviceType']
                    if device_type == 'Endpoint' or\
                            device_type == 'UpstreamPort':
                        return True
        except Exception:
            return False
        return False

    @classmethod
    def msg(cls, pcie):
        """
           Returns the details for msg
        """
        bdf_1 = pcie.json_obj['BDF']
        bdf_2 = pcie.json_obj['PCIeInterface']['ParentChild']['BDF']
        msg_list = [bdf_1 + ' ' + bdf_2,
                    pcie.json_obj['PCIeInterface']['ExpectedLanes'],
                    pcie.json_obj['PCIeInterface']['CurrentWidth']]
        return msg_list


class PCIeCorrectableError(IRule):

    """
    PCIe correctable error
    """

    def __init__(self):
        self.alert_id = 'NV-PCI-03'

    @classmethod
    def condition(cls, pcie):
        try:
            health = pcie.json_obj['Status']['Health']
            state = pcie.json_obj['Status']['State']
            if state == 'Enabled' and health == 'Warning':
                if pcie.json_obj['PCIeDeviceErrorCount']['CorrectableCount']:
                    return True
        except Exception:
            return False
        return False

    @classmethod
    def msg(cls, pcie):
        """
           Returns the details for msg
        """
        msg_list = [pcie.json_obj['BDF']]
        return msg_list


class PCIeUnCorrectableError(IRule):

    """
    PCIe un-correctable error
    """

    def __init__(self):
        self.alert_id = 'NV-PCI-04'

    @classmethod
    def condition(cls, pcie):
        try:
            health = pcie.json_obj['Status']['Health']
            state = pcie.json_obj['Status']['State']
            if state == 'Enabled' and health == 'Warning':
                if pcie.json_obj['PCIeDeviceErrorCount']['UncorrectableCount']:
                    return True
        except Exception:
            return False
        return False

    @classmethod
    def msg(cls, pcie):
        """
           Returns the details for msg
        """
        msg_list = [pcie.json_obj['BDF']]
        return msg_list


class PCIeDeviceMissing(IRule):

    """
    PCIe drive missing
    """

    def __init__(self):
        self.alert_id = 'NV-PCI-05'

    @classmethod
    def condition(cls, pcie):
        try:
            health = pcie.json_obj['Status']['Health']
            state = pcie.json_obj['Status']['State']
            device_type = pcie.json_obj['PCIeInterface']['DeviceType']
            if state == 'Absent' and \
                    health == '' and \
                    (device_type == 'Endpoint' or device_type == 'PCIeLegacy'):
                return True
        except Exception:
            return False
        return False

    @classmethod
    def msg(cls, pcie):
        """
           Returns the details for msg
        """
        msg_list = [pcie.json_obj['BDF']]
        return msg_list


class PCIeDeviceDisbled(IRule):

    """
    PCIe device is Disabled
    """

    def __init__(self):
        self.alert_id = 'NV-PCI-07'

    @classmethod
    def condition(cls, pcie):
        try:
            health = pcie.json_obj['Status']['Health']
            state = pcie.json_obj['Status']['State']
            device_type = pcie.json_obj['PCIeInterface']['DeviceType']
            if state == 'Disabled' and \
                    health == '' and \
                    (device_type == 'Endpoint' or device_type == 'PCIeLegacy'):
                return True
        except Exception:
            return False
        return False

    @classmethod
    def msg(cls, pcie):
        """
           Returns the details for msg
        """
        msg_list = [pcie.json_obj['BDF']]
        return msg_list


class PCIeAERSupport(IRule):

    """
    PCIe AER support check
    """

    def __init__(self):
        self.alert_id = 'NV-PCI-06'

    @classmethod
    def condition(cls, pcie):
        try:
            # Set watcher for error reporting, ignoring
            # UnsupportedRequestError.
            health = pcie.json_obj['Status']['Health']
            state = pcie.json_obj['Status']['State']
            if state == 'Absent' and \
                    health == '':
                return False
            watchers = ['CorrectableError', 'NonfatalError', 'FatalError']
            source_aer_dict = {}
            for watcher in watchers:
                source_aer_dict[
                    watcher] = pcie.json_obj[
                        'PCIeDeviceErrorReporting'][
                            watcher]

            aer_dict = {
                k: v for k, v in list(source_aer_dict.items()) if not v}

            if bool(aer_dict):
                return True
        except Exception:
            return False
        return False

    @classmethod
    def msg(cls, pcie):
        """
           Returns the details for msg
        """
        msg_list = [pcie.json_obj['BDF']]

        # Set watcher for error reporting, ignoring UnsupportedRequestError.
        watchers = ['CorrectableError', 'NonfatalError', 'FatalError']
        source_aer_dict = {}
        for watcher in watchers:
            source_aer_dict[
                watcher] = pcie.json_obj[
                    'PCIeDeviceErrorReporting'][
                        watcher]

        aer_dict = {
            k: v for k, v in list(source_aer_dict.items()) if not v}
        aer_str = ",".join(list(aer_dict.keys()))
        msg_list.append(aer_str)
        return msg_list


# PCIe class defined here.
class PCIe(Equipment):

    """
    PCIe: Extends Equipment Class.
    """

    # TODO test it.
    __boottime_rules = [PCIeLinkSpeedNotMatchingError(),
                        PCIeLinkWidthNotMatchingError(),
                        PCIeCorrectableError(),
                        PCIeUnCorrectableError(),
                        PCIeDeviceMissing(),
                        PCIeAERSupport(),
                        PCIeDeviceDisbled()]

    __rules = [PCIeLinkSpeedNotMatchingError(),
               PCIeLinkWidthNotMatchingError(),
               PCIeCorrectableError(),
               PCIeUnCorrectableError(),
               PCIeDeviceMissing(),
               PCIeAERSupport(),
               PCIeDeviceDisbled()]

    __CGX_boottime_rules = [PCIeLinkSpeedNotMatchingError(),
                            PCIeLinkWidthNotMatchingError(),
                            PCIeCorrectableError(),
                            PCIeUnCorrectableError(),
                            PCIeDeviceMissing(),
                            PCIeDeviceDisbled()]

    __CGX_rules = [PCIeLinkSpeedNotMatchingError(),
                   PCIeLinkWidthNotMatchingError(),
                   PCIeCorrectableError(),
                   PCIeUnCorrectableError(),
                   PCIeDeviceMissing(),
                   PCIeDeviceDisbled()]

    __RTX_boottime_rules = [PCIeLinkSpeedNotMatchingError(),
                            PCIeLinkWidthNotMatchingError(),
                            # PCIeCorrectableError(),
                            # PCIeUnCorrectableError(),
                            PCIeDeviceMissing(),
                            PCIeAERSupport(),
                            PCIeDeviceDisbled()]

    __RTX_rules = [PCIeLinkSpeedNotMatchingError(),
                   PCIeLinkWidthNotMatchingError(),
                   # PCIeCorrectableError(),
                   # PCIeUnCorrectableError(),
                   PCIeDeviceMissing(),
                   PCIeAERSupport(),
                   PCIeDeviceDisbled()]

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        if Endpoint().model() == "CGX":
            return cls.__CGX_rules
        if Endpoint().model() == "RTX":
            return cls.__RTX_rules

        return cls.__rules

    @classmethod
    def get_boottime_rules(cls):
        """
        Returns a list of boottime rules supported for this device type
        """
        if Endpoint().model() == "CGX":
            return cls.__CGX_boottime_rules
        if Endpoint().model() == "RTX":
            return cls.__RTX_boottime_rules

        return cls.__boottime_rules

    def __init__(self, entity_id, name, hostname, serial, json_obj):
        super(PCIe, self).__init__(entity_id, name, hostname, serial)
        self.json_obj = json_obj

    def get_name(self):
        return super(PCIe, self).get_name()

    def config(self, config):
        return super(PCIe, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial, **kwargs):
        ret = None

        if 'uri' not in list(kwargs.keys()):
            logger.error(
                "Required arguments are not passed to the create method")
            return ret

        pcie_name = json_obj['ID']
        try:
            ret = cls(
                pcie_name,
                pcie_name,
                hostname,
                serial,
                json_obj)
        except Exception:
            logger.exception(
                "PCIe object creation failed, \
                        json object incomplete %s",
                json.dumps(json_obj))

        return ret

    def __repr__(self):
        return '{}({},health:{})'.format(
            self.__class__.__name__,
            super(PCIe, self).__repr__(),
            self.json_obj['Status']['Health']
        )


class DCSHWAPIKeyExpiring(IRule):

    """
    DCSHW API Key expiring
    """
    def __init__(self):
        self.alert_id = 'NV-DCS-01'

    @classmethod
    def condition(cls, dcshwapikey):
        days = dcshwapikey.json_obj['numberOfDaysLeftToExpireLicense']
        if days <= 30 and days > 0:
            return True
        return False

    @classmethod
    def msg(cls, dcshwapikey):
        """
           Returns the details for msg
        """
        return [dcshwapikey.json_obj['numberOfDaysLeftToExpireLicense']]



class DCSHWAPIKeyExpired(IRule):

    """
    DCSHW API Key expired
    """
    def __init__(self):
        self.alert_id = 'NV-DCS-02'

    @classmethod
    def condition(cls, dcshwapikey):
        if dcshwapikey.json_obj['numberOfDaysLeftToExpireLicense'] <= 0:
            return True
        return False

    @classmethod
    def msg(cls, dcshwapikey):
        """
           Returns the details for msg
        """
        return []


class DCSHWAPIKey(Equipment):
    """
    DCSHWAPIKey: Extends Equipment Class.
    """
    __rules = [DCSHWAPIKeyExpiring(),
               DCSHWAPIKeyExpired()]

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def __init__(self, entity_id, name, hostname, serial, json_obj):
        super(DCSHWAPIKey, self).__init__(entity_id, name, hostname, serial)
        self.json_obj = json_obj

    def get_name(self):
        return super(DCSHWAPIKey, self).get_name()

    def config(self, config):
        return super(DCSHWAPIKey, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial, **kwargs):
        ret = None

        key_name = 'DCSHWAPIKEY'
        try:
            ret = cls(
                key_name,
                key_name,
                hostname,
                serial,
                json_obj)
        except Exception:
            logger.exception(
                "Dcshwapikey object creation failed, \
                        json object incomplete %s",
                json.dumps(json_obj))

        return ret

    def __repr__(self):
        return '{}({},health:{})'.format(
            self.__class__.__name__,
            super(DCSHWAPIKey, self).__repr__(),
            self.json_obj['numberOfDaysLeftToExpireLicense']
        )
