# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Equipment module

Contains the definition for Equipment class and its derivations.
"""

from abc import ABCMeta, abstractmethod
import logging

import os, sys
sys.path.append(os.path.abspath(os.path.join(__file__, '..')))
from .alerts import Alert


logger = logging.getLogger(__name__)


class IRule(object, metaclass=ABCMeta):

    """
        Rule class.
        Defines a rule to be applied on the equipment.

        When the rule is applied if the condition is met, it raises an alert.
        If the condition is not met, it clears the alert.
    """

    def __init__(self):
        self.alert_id = None

    @classmethod
    def condition(cls, equipment):
        """
            Tests for the condition that is evaluted for this rule
        """
        pass

    @classmethod
    def msg(cls, equipment):
        """
           Returns the custom message string for the alert, that is used to extend the message details.
        """
        return None

    @classmethod
    def chassis_id(cls, equipment):
        """
           Returns the chassis id for the alert, that is used to add in the message details.
        """
        return ''

    def exists(self, alerts):
        """
            Checks if the alert raised for this rule already exits.
        """
        if self.alert_id in list(alerts.keys()) and alerts[self.alert_id].asserted:
            return True

    def apply(self, device, alerts):
        """
            Applies the rule.

            If condition is true:
                Check if alert for the condition exists.
                    If the alert already exists, then do nothing.
                    else raise an alert.

            If condition is false:
                Check if alert for the condition exists.
                    If the alert exists, then clear the alert.
                    else do nothing.
        """
        if self.condition(device):
            logger.debug('{}: Result: True'.format(self))
            if not self.exists(alerts):
                alerts[self.alert_id] = Alert(self.alert_id,
                                              device.get_name(),
                                              device.hostname,
                                              device.serial)
                alerts[self.alert_id].msg_param = self.msg(device)
                alerts[self.alert_id].chassis_id = self.chassis_id(device)
                logger.alert('Alert asserted {}'.format(alerts[self.alert_id]))
        else:
            logger.debug('{}: Result: False'.format(self))
            if self.exists(alerts):
                alerts[self.alert_id].asserted = False
                logger.alert(
                    'Alert de-asserted {}'.format(alerts[self.alert_id]))

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            self.alert_id
        )


class IContainerRule(object, metaclass=ABCMeta):

    """
        Container Rule class.
        Defines a rule to be applied on the equipment container.

        Currently it checks if the container has all the equipment or of some
        are missing. It raises alerts for missing devices.
    """

    def __init__(self, device_list):
        self.alert_id = None

    def exists(self, alerts):
        """
            Checks if the alert raised for this rule already exits.
        """
        if self.alert_id in list(alerts.keys()) and alerts[self.alert_id].asserted:
            return True

    def apply(self, device, is_present, hostname, serial, alerts):
        """
            Applies the rule.

            If condition is true:
                Check if alert for the condition exists.
                    If the alert already exists, then do nothing.
                    else raise an alert.

            If condition is false:
                Check if alert for the condition exists.
                    If the alert exists, then clear the alert.
                    else do nothing.
        """
        op = getattr(self, "condition", None)
        if callable(op):
            if self.condition(device):
                logger.debug('{}: Result: True'.format(self))
                if not is_present:
                    if not self.exists(alerts):
                        alerts[self.alert_id] = Alert(self.alert_id,
                                                      device,
                                                      hostname,
                                                      serial)
                        logger.alert(
                            'Alert asserted {}'.format(alerts[self.alert_id]))
                else:
                    logger.debug('{}: Device present'.format(self))
                    if self.exists(alerts):
                        alerts[self.alert_id].asserted = False
                        logger.alert(
                            'Alert de-asserted {}'.format(alerts[self.alert_id]))

        else:
            if not is_present:
                logger.debug('{}: Device missing'.format(self))
                if not self.exists(alerts):
                    alerts[self.alert_id] = Alert(self.alert_id,
                                                  device,
                                                  hostname,
                                                  serial)
                    logger.alert(
                        'Alert asserted {}'.format(alerts[self.alert_id]))
            else:
                logger.debug('{}: Device present'.format(self))
                if self.exists(alerts):
                    alerts[self.alert_id].asserted = False
                    logger.alert(
                        'Alert de-asserted {}'.format(alerts[self.alert_id]))

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            self.alert_id
        )
