# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Sensor module

Encapsulates the sensor reading and its various threashold data.
"""
import logging

logger = logging.getLogger(__name__)


class Sensor(object):

    """
    Sensor Class
    Description:
    Represents the sensor and its data

    Attributes:
        units    :  Units of measurement
        reading  :  Sensor Reading
        assertion:  Indicates sensor assertion
        name     :  Name of the sensor
        kind     :  Type of the sensor
                    Ex: 'Performance', 'Thermal', 'Voltage', 'Current'

    FIXME: should kind attribute be a enum?
"""

    __allowed = ("UpperThresholdNonCritical", "UpperThresholdCritical",
                 "UpperThresholdFatal", "LowerThresholdNonCritical",
                 "LowerThresholdCritical", "LowerThresholdFatal")

    def __init__(self, kind, name, units, **kwargs):
        self.units = units
        self._reading = None
        self.kind = kind
        self.name = name
        self._assertion = None

        # FIXME: Check if the string is float before converting it.
        for attr, value in list(kwargs.items()):
            assert attr in self.__allowed
            setattr(self, attr, float(value))

    def set_reading(self, reading):
        """
        Records sensor reading and checks if the sensor has crossed any of the
        defined thresholds.
        """
        # TODO: Thea should ideally not send No Reading, remove this check
        # later.
        if reading == 'No Reading':
            self._reading = None
        else:
            self._reading = reading
        self._set_assertion()

    # FIXME: Can this be implemented better?
    def _set_assertion(self):
        """
        This method checks if the sensor reading has crossed any of the
        thresholds and sets '_assertion' to the asserted threshold

        Current implementation will NOT alert the user if sensor is asserted.
        It is assumed that when the sensor is asserted the device health will
        be set to 'Warning'/'Critical'.

        When device health is set to 'Warning'/'Critical', DSHM will check the
        assertion(/s) on the device and include details of all assertions in
        the alert.
        """
        if self._reading is None:
            self._assertion = "No Reading"

        elif hasattr(self, 'UpperThresholdFatal') and \
                self._reading >= getattr(self, 'UpperThresholdFatal'):
            self._assertion = "UpperThresholdFatal"
            # TODO: Is there a way to format this string better?
            logger.error('{} --> UpperThresholdFatal is: {}'.format(
                self,
                getattr(self, 'UpperThresholdFatal')))

        elif hasattr(self, 'UpperThresholdCritical') and \
                self._reading >= getattr(self, 'UpperThresholdCritical'):
            self._assertion = "UpperThresholdCritical"
            logger.error('{} --> UpperThresholdCritical is: {}'.format(
                self,
                getattr(self, 'UpperThresholdCritical')))

        elif hasattr(self, 'UpperThresholdNonCritical') and \
                self._reading >= getattr(self, 'UpperThresholdNonCritical'):
            self._assertion = "UpperThresholdNonCritical"
            logger.error('{} --> UpperThresholdNonCritical is: {}'.format(
                self,
                getattr(self, 'UpperThresholdNonCritical')))

        elif hasattr(self, 'LowerThresholdFatal') and \
                self._reading <= getattr(self, 'LowerThresholdFatal'):
            self._assertion = "LowerThresholdFatal"
            logger.error('{} --> LowerThresholdFatal is: {}'.format(
                self,
                getattr(self, 'LowerThresholdFatal')))

        elif hasattr(self, 'LowerThresholdCritical') and \
                self._reading <= getattr(self, 'LowerThresholdCritical'):
            self._assertion = "LowerThresholdCritical"
            logger.error('{} --> LowerThresholdCritical is: {}'.format(
                self,
                getattr(self, 'LowerThresholdCritical')))

        elif hasattr(self, 'LowerThresholdNonCritical') and \
                self._reading <= getattr(self, 'LowerThresholdNonCritical'):
            self._assertion = "LowerThresholdNonCritical"
            logger.error('{} --> LowerThresholdNonCritical is: {}'.format(
                self,
                getattr(self, 'LowerThresholdNonCritical')))

        else:
            self._assertion = None

    def get_reading(self):
        """
        Returns sensor reading.
        """
        return self._reading

    def get_assertion(self):
        """
        Returns sensor assertion.
        """
        return self._assertion

    def get_threshold(self):
        """
        Returns sensor threshold value that was exceeded.
        """
        if hasattr(self, self._assertion):
            return getattr(self, self._assertion)
        else:
            return "Unknown"

    def __repr__(self):
        return '{}(name:{},reading:{},units:{},assertion:{}'.format(
            self.__class__.__name__,
            self.name,
            self._reading,
            self.units,
            self._assertion
        )
