# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Sliding Window module

"""

import logging

logger = logging.getLogger(__name__)

#
# Sample usage of this window
#
#   window = Window(time_stamps_list, duration)
#   while window.roll() is True:
#   if window.errors() >= threshold_count:
#       logger.debug("Threshold has been crossed")
#       return True
#


class Window(object):

    """
        Window class.

        Contains the implementation for the sliding window algorithm
        implementation to process time series data with variable sampling
        frequency.

        This is used to compute thresholds for correctable errors for DIMMs
        and HDDs.
    """

    def __init__(self, timestamps, duration, errs_per_entry=1):
        """
        Initializer

        'timestamps': Array of timestamps.
        'duration': The duration monitored by the sliding window (in seconds).
        'errs_per_entry': Number of errors tracked by each log entry.
                          For DIMMs, each sel entry corresponds to 1
                            correctable error.
                          This implementation assumes that the number of
                            errors reported per log entry is fixed.
        """
        self.timestamps = timestamps
        self.duration = duration
        self.start = -1
        self.end = 0
        self.errs_per_entry = errs_per_entry

    def roll(self):
        """
        Rolls start index of the window by 1.

        This method returns False if it has reached the end of the timestamps
        data.
        """
        rolled = False
        self.start += 1

        if self.end < self.start:
            self.end = self.start

        index = self.end
        while index < len(self.timestamps):
            delta = self.timestamps[index] - self.timestamps[self.start]
            if delta.total_seconds() <= self.duration:
                self.end = index
                rolled = True
                index += 1
            else:
                break

        return rolled

    def errors(self):
        """
            This method errors seen for the current window
        """
        return (self.end - self.start + 1) * self.errs_per_entry

    def __repr__(self):
        return '{}(duration:{},\
                start_time:{},\
                end_time:{},\
                errs_per_entry:{})'.format(
            self.__class__.__name__,
            self.duration,
            self.timestamps[self.start],
            self.timestamps[self.end],
            self.errs_per_entry
        )
