#!/usr/bin/python
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import logging
import socket
import sys
import json

from nvsm_dshm.thea import Client
from nvsm_dshm.thea import Endpoint
from nvsm_dshm.equipment import Equipment
from nvsm_dshm.ds import AlertsDataStore
from nvsm_dshm.policy import IPolicy
from nvsm_dshm.infra import ErrCode as Err
from nvsm_dshm.infra import config_settings
from nvsm_dshm.infra import file_not_found
from nvsm_dshm.infra import invalid_input
from nvsm_dshm.infra import pa

logger = logging.getLogger(__name__)


def get_ip_addr():
    """
    Returns the IP Address of the machine.
    """
    ip_addr = "Unknown"
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect(("8.8.8.8", 80))
        ip_addr = sock.getsockname()[0]
    except Exception:
        logger.exception("Getting the Host IP Address failed")

    return ip_addr


def process(obj, device_type, filter_device,
            filter_func=None, policy=None, **kwargs):
    """
        Job to fetch the device_type objects and execute the rules
        on the  objects
    """
    logger.debug(
        "Alert processing for device_type : " +
        device_type +
        " Enter")

    # Flag to indicate if a new alert was generated
    new_alert = False

    if obj is None:
        logger.error(
            'JSON OBJ received from the event is None for {} '.format(
                device_type))
        return {"err": Err.ERR_EMPTY_JSON_OBJS, "new_alert": new_alert}

    # If Policy is not provided then fetch the policy
    if policy is None:
        policy = IPolicy.factory(device_type)
    logger.debug("Policy used {}".format(policy))

    # Fetch the device type objects
    json_objs = Client.get_prop_value(obj, filter_device)
    if json_objs is None:
        logger.error(
            'Invalid payload received,device_type{} is empty.'.format(
                device_type))
        return {"err": Err.ERR_INVALID_JSON_OBJ, "new_alert": new_alert}

    # If the input object is not a list, convert it into a list of single
    # element.
    if not isinstance(json_objs, list):
        lst = [json_objs]
        json_objs = lst

    # If filter_func is supplied, filter the device objects
    if filter_func is not None:
        try:
            filtered_objs = list(filter(filter_func, json_objs))
            json_objs = filtered_objs
        except Exception:
            logger.exception(
                "Filtering of {} objects failed".format(device_type))
            return {"err": Err.ERR_INVALID_JSON_OBJ, "new_alert": new_alert}

        if json_objs is None:
            logger.error(
                "Filtered objects of device type: {} are not available".format(device_type))
            return {"err": Err.ERR_INVALID_JSON_OBJ, "new_alert": new_alert}

    # Fetch the alerts in the data store
    data_store = AlertsDataStore.factory('thea_db', device_type)

    for obj in json_objs:
        device = Equipment.factory(device_type, obj,
                                   Endpoint().hostname(),
                                   Endpoint().serial(), **kwargs)

        if device is not None and policy.monitor(device.get_name(), device.serial):
            logger.debug("Checking device {}".format(device.get_name()))

            ret = data_store.fetch_alerts(device.get_name(), device.serial)
            if ret['errored']:
                logger.error(
                    "Unable to fetch alert data from {} for thea".format(
                        device_type))
                logger.error(
                    'Thea is in-accessible, waiting till connection \
                        is re-established')
                return {"err": Err.ERR_CONN_REST_SRV, "new_alert": new_alert}

            alerts = ret['alerts']

            device.config(policy)
            device.apply_rules(device, alerts)

            if not new_alert:
                new_alert = any(
                    alerts[alert_id].new for alert_id in list(alerts.keys()))

            data_store.store_alerts(device.get_name(), alerts)

    logger.debug("Alert processing for device_type : " + device_type + " Exit")

    return {"err": Err.SUCCESS, "new_alert": new_alert}


def init_pa(monitor):
    """
    Initializes pa.
    If pa is unable to initialize then there is no point in continuing monitor
    so exit the monitor
    """
    # try:
    #    prof = config_settings().get(
    #        'pa', Endpoint().model() + '_profile_file')
    #    pa(file_name=prof)
    # except file_not_found as e:
    #    logger.fatal('PA: {0}'.format(e))
    #    sys.exit(1)
    # except Exception:
    #    logger.exception("PA Init failed")
    #    sys.exit(1)

    client = Client({Client.URI: '/nvsm/v1/PlatformProfile', Client.ALL: True})
    profile = json.dumps(client.get_objects())

    try:
        prof = pa(obj=profile)
    except invalid_input_data as e:
        logger.fatal("PA: {0}".format(e))
        sys.exit(1)
    except Exception:
        logger.exception("PA Init failed")
        sys.exit(1)

    if profile is None:
        logger.fatal('Fetching platform profile through REST API failed')
        sys.exit(1)

    services = pa().enumerate_obj('Capabilities/Services')
    if not services or monitor not in services:
        logger.info("{} not required for this Platform [{}], gracefully exiting.".format(
            monitor, Endpoint().model()))
        sys.exit(0)


def handle_chassis_init(client, msg, subscribe_topics, unsubscribe_topics):
    import json
    msg = json.loads(msg.payload)
    logger.debug("Chassis init message [{}] received ".format(msg))

    if msg['state'] == "INIT_DONE":
        subscribe_topics(client)
    elif msg['state'] == "INIT_WAIT":
        unsubscribe_topics(client)

def get_selected_severity(alert_severity):
     
    selected_alerts = []  
    alert_severity_table = {0:'Warning',
                            1:'Critical'
                           }

    alert_severity_map = {'warning':0,
                          'critical':1
                         }
    sev_level = alert_severity_map[alert_severity.lower()]
    #By default all Cleared alerts are to be forwarded/selected. 
    selected_alerts.append('Cleared')  

    for key in alert_severity_table.keys():
        if sev_level <= key:
            selected_alerts.append( alert_severity_table[key])

    return selected_alerts  
