#!/usr/bin/python
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.


"""
Message Queue Broker Client.

"""

import paho.mqtt.client as mqtt
from abc import ABCMeta, abstractmethod

import logging
import os.path
import sys
import datetime
import ssl
from nvsm_dshm.infra import config_settings

from nvsm_dshm.infra import ErrCode as Err

# DONT REMOVE,USED FOR REFERENCE ##################
# MQTT_BROKER = "127.0.0.1"
# MQTT_PORT = 159
# CA_CERTS="/etc/nvsm/nvsm-ca.crt"
# SERVER_CERT_FILE="/etc/nvsm/nvsm-server.crt"
# SERVER_KEY_FILE="/etc/nvsm/nvsm-server.key"

logger = logging.getLogger(__name__)

# This module adds a lot of noise without providing any valuable debug data.
# Force it to log-level error
logger.setLevel(logging.ERROR)


class MQBrokerClient(object, metaclass=ABCMeta):

    """
    MqBrokerClient class:


    """

    MQTT_KEEPALIVE = 30
    QOS_LEVEL = 1
    RETAIN_FLAG = 1
    MQTT_CLIENT = "mqtt_client"

    MQTT_BROKER_IP = None
    MQTT_PORT = None
    CA_CERTS_FILE = None
    SERVER_CERT_FILE = None
    SERVER_KEY_FILE = None

    def __init__(self, mq_type, name, mq_broker_ip, mq_broker_port,
                 ca_cert_file, srv_cert_file, srv_cert_key_file,
                 on_connect, on_message, on_disconnect, mq_ttl, client):

        logger.debug("Enter MQBrokerClient Constructor")
        """
        Constructor
        """
        self.mq_type = mq_type
        self.name = name
        self.mq_broker_ip = mq_broker_ip
        self.mq_broker_port = mq_broker_port
        self.ca_cert_file = ca_cert_file
        self.srv_cert_file = srv_cert_file
        self.srv_cert_key_file = srv_cert_key_file
        self.on_connect = on_connect
        self.on_message = on_message
        self.on_disconnect = on_disconnect
        self.mq_ttl = mq_ttl
        self.client = client
        logger.debug("Exit MQBrokerClient Constructor")

    """
    Factory for various types of message qeue broker

    mq_type: Type of MQ Broker Client.
    mq_broker_ip_addr:MQ Broker IP Address.
    mq_broker_port: Port on which MQ Broker listens.
    ca_cert_file_name: ca certficate MQ Broker has to trust.
    srv_cert_file_name: Server certificate.
    srv_cert_key_file_name: Server private key.
    on_connect: call back to be called on a succesful connect.
    on_message: call back to be called on, when a message received.
    on_disconnect: call back to be called when disconnect happens.
    mq_ttl: minimum time in seconds to ping the MQ Broker.

    """
    @staticmethod
    def factory(mq_type, name,
                on_connect, on_message=None, on_disconnect=None,
                mq_broker_ip_addr=None, mq_broker_port=None,
                ca_cert_file_name=None, srv_cert_file_name=None,
                srv_cert_key_file_name=None,
                mq_ttl=MQTT_KEEPALIVE):
        logger.debug("Enter MQBrokerClient:factory %s", mq_type)

        if mq_type == MQBrokerClient.MQTT_CLIENT:

            if mq_broker_ip_addr is None or mq_broker_port is None or \
                ca_cert_file_name is None or srv_cert_file_name is None or \
                    srv_cert_key_file_name is None:

                logger.debug(
                    "MQBrokerClient:factory read it from Thea config file")

                ret = MQBrokerClient.get_mqtt_config()
                if ret is not Err.SUCCESS:
                    return None

                mq_broker_ip_addr = MQBrokerClient.MQTT_BROKER_IP
                mq_broker_port = int(MQBrokerClient.MQTT_PORT)
                ca_cert_file_name = MQBrokerClient.CA_CERTS_FILE
                srv_cert_file_name = MQBrokerClient.SERVER_CERT_FILE
                srv_cert_key_file_name = MQBrokerClient.SERVER_KEY_FILE

            logger.debug('Mqtt Config: ' + 'ca_cert:' + ca_cert_file_name +
                         ' server_cert:' + srv_cert_file_name + ' server_key:' + srv_cert_key_file_name +
                         ' mqtt_broker ip:' + mq_broker_ip_addr + ' mqtt_port:' + str(mq_broker_port))

            return MQTTClient.create(
                mq_type, name, mq_broker_ip_addr, mq_broker_port,
                              ca_cert_file_name, srv_cert_file_name,
                              srv_cert_key_file_name,
                              on_connect, on_message, on_disconnect, mq_ttl)

        logger.debug("Undefined MQ Broker Type:" + mq_type)

        return None

    """
    Gets the MQTT Broker IP and PORT, CA certificate, SA certificate and
    Server Key file name from the Thea Config File.
    """
    @classmethod
    def get_mqtt_config(cls):

        logger.debug("Enter")

        nvsm_api_config_file = config_settings().get('nvsm-api', 'config-file')

        (cls.CA_CERTS_FILE,
         cls.SERVER_CERT_FILE,
         cls.SERVER_KEY_FILE,
         mqtt_address,
         _,
         _,
         _) = config_settings.nvsm_api_config_helper(
            nvsm_api_config_file)

        if cls.CA_CERTS_FILE is None:
            logger.error('init: ca certificate file is not set')
            return Err.ERR_INVALID_NVSM_API_CONFIG

        if cls.SERVER_CERT_FILE is None:
            logger.error('init: server certificate file is not set')
            return Err.ERR_INVALID_NVSM_API_CONFIG

        if cls.SERVER_KEY_FILE is None:
            logger.error('init: server key file is not set')
            return Err.ERR_INVALID_NVSM_API_CONFIG

        if mqtt_address is None:
            logger.error('init: mqtt broker address is NULL')
            return Err.ERR_INVALID_NVSM_API_CONFIG

        l_st = mqtt_address.split(':')

        if len(l_st) != 3:
            logger.error('init: Invalid mqtt broker address')
            return Err.ERR_INVALID_MQTT_ADDRESS

        cls.MQTT_BROKER_IP = l_st[1][2:]
        cls.MQTT_PORT = l_st[2]

        logger.debug("Exit")

        return Err.SUCCESS

    @staticmethod
    def subscribe(client, topic, qos=QOS_LEVEL, call_back=None):
        """
        Subscribes to given topic.
        """
        logger.debug("Enter MQBrokerClient:subscribe")

        ret_val = Err.ERROR

        if client is None:
            Err.ERR_INVLAID_MQ_CLIENT

        if type(client).__name__ == type(mqtt.Client()).__name__:
            ret_val = MQTTClient.subscribe_topic(client, topic, qos)

        logger.debug("Exit MQBrokerClient:subscribe")

        return ret_val

    @staticmethod
    def register_callback(client, topic, qos=QOS_LEVEL):
        """
        Subscribes to given topic.
        """
        logger.debug("Enter MQBrokerClient:subscribe")

        ret_val = Err.ERROR

        if client is None:
            Err.ERR_INVLAID_MQ_CLIENT

        if type(client).__name__ == type(mqtt.Client()).__name__:
            ret_val = MQTTClient.subscribe_topic(client, topic, qos)

        logger.debug("Exit MQBrokerClient:subscribe")

        return ret_val

    @staticmethod
    def publish(client, topic, obj, qos=QOS_LEVEL, retain=RETAIN_FLAG):
        """
        Publishes the given payload to the topic.
        """
        logger.debug("Enter MQBrokerClient:publish")

        ret_val = Err.ERROR

        if client is None:
            Err.ERR_INVLAID_MQ_CLIENT

        if type(client).__name__ == type(mqtt.Client()).__name__:
            ret_val = MQTTClient.publish_topic(client, topic, obj, qos, retain)
            if ret_val[0] == 0:
                ret_val = Err.SUCCESS

        logger.debug("Exit MQBrokerClient:publish")

        return ret_val

    @abstractmethod
    def create(cls, mq_type, name, mq_broker_ip_addr, mq_broker_port,
               ca_cert_file_name, srv_cert_file_name,
               srv_cert_file_key,
               on_connect, on_message, on_disconnect, mq_ttl):
        """
        Creates the MQ Client based on the provided type.
        """
        pass

    @abstractmethod
    def subscribe_topic(cls, client, topic, qos=QOS_LEVEL):
        """
        Implement Broker specific subscribe method.
        """
        pass

    @abstractmethod
    def publish_topic(
            cls, client, topic, obj, qos=QOS_LEVEL, retain=RETAIN_FLAG):
        """
        Implement Broker specific publish method.
        """
        pass

    @abstractmethod
    def listen_for_ever(self):
        """
        Implement Broker specific forever loop method.
        """
        pass

    @abstractmethod
    def loop_start(self):
        """
        Implement Broker specific loop start method.
        """
        pass

    @abstractmethod
    def loop_stop(self):
        """
        Implement Broker specific loop stop method.
        """
        pass

    @abstractmethod
    def disconnect(self):
        """
        Implement Broker specific disconnect method.
        """
        pass

    def __repr__(self):
        return 'name:{}'.format(self.name)


class MQTTClient(MQBrokerClient):

    def __init__(self, mq_type, name, mq_broker_ip, mq_broker_port,
                 ca_cert_file, srv_cert_file, srv_cert_key_file_name,
                 on_connect, on_message, on_disconnect, mq_ttl, client):

        logger.debug("Enter MQTTClient Constructor")

        super(
            MQTTClient, self).__init__(mq_type, name, mq_broker_ip, mq_broker_port,
                                       ca_cert_file, srv_cert_file,
                                       srv_cert_key_file_name,
                                       on_connect, on_message, on_disconnect,
                                       mq_ttl, client)

        logger.debug("Exit MQTTClient Constructor")

    @classmethod
    def create(
        cls, mq_type, name, mq_broker_ip_addr, mq_broker_port, ca_cert_file_name,
                            srv_cert_file_name, srv_cert_file_key,
                            on_connect, on_message, on_disconnect, mq_ttl):

        logger.debug("Enter MQTTClient Create")

        client = None
        mqtt_broker = None
        client = mqtt.Client(name)

        if on_connect is not None:
            client.on_connect = on_connect

        if on_message is not None:
            client.on_message = on_message

        if on_disconnect is not None:
            client.on_disconnect = on_disconnect

        try:
            client.tls_set(
                ca_cert_file_name,
                srv_cert_file_name,
                srv_cert_file_key)
            client.tls_insecure_set(True)
        except:
            logger.error(
                "Incorrect ca_certificate/server_certificate/server_key")
            return Err.ERR_INVLAID_MQ_SECURITY_CERT
        try:
            client.connect(mq_broker_ip_addr, mq_broker_port, mq_ttl)
        except:
            logger.error("Incorrect MqBroker IP-Address or Port Number")
            return Err.ERR_INVLAID_MQ_ADDR_PORT

        mqtt_broker = cls(mq_type, name, mq_broker_ip_addr, mq_broker_port,
                          ca_cert_file_name, srv_cert_file_name, srv_cert_file_key,
                          on_connect, on_message, on_disconnect,
                          mq_ttl, client)

        logger.debug("Exit MQTTClient Create")

        return mqtt_broker

    def disconnect(self):

        logger.debug("Enter MQTTClient Disconnect")

        if self.client is None:
            logger.error("Client is invalid ")
            return Err.ERR_INVLAID_MQ_CLIENT

        self.client.disconnect()

        logger.debug("Exit MQTTClient Disconnect")

    def listen_for_ever(self):

        logger.debug("Enter MQTTClient Run")

        if self.client is None:
            logger.error("Client is invalid ")
            return Err.ERR_INVLAID_MQ_CLIENT

        self.client.loop_forever()

        logger.debug("Exit MQTTClient Run")

    def loop_start(self):
        logger.debug("Enter MQTTClient start loop")

        if self.client is None:
            logger.error("Client is invalid ")
            return Err.ERR_INVLAID_MQ_CLIENT

        self.client.loop_start()

        logger.debug("Exit MQTTClient start loop")

    def loop_stop(self):
        logger.debug("Enter MQTTClient stop loop")

        if self.client is None:
            logger.error("Client is invalid ")
            return Err.ERR_INVLAID_MQ_CLIENT

        self.client.loop_stop()

        logger.debug("Exit MQTTClient stop loop")

    @classmethod
    def subscribe_topic(cls, client, topic, qos=MQBrokerClient.QOS_LEVEL):

        logger.debug("Enter MQTTClient: Subscribe Topic: ")

        ret_val = Err.SUCCESS

        if topic is None:
            logger.error("Topic is invalid ")
            return Err.ERR_INVLAID_MQ_TOPIC

        if client is None:
            logger.error("Client is invalid ")
            return Err.ERR_INVLAID_MQ_CLIENT

        client.subscribe(topic, qos)

        logger.debug("Exit MQTTClient: Subscribe")

        return ret_val

    @classmethod
    def publish_topic(cls, client, topic, obj,
                      qos=MQBrokerClient.QOS_LEVEL,
                      retain=MQBrokerClient.RETAIN_FLAG):

        logger.debug("Enter MQTTClient: Publish Topic: " + topic)

        ret_val = Err.ERROR

        if topic is None:
            logger.error("Topic is invalid ")
            return Err.ERR_INVLAID_MQ_TOPIC

        if obj is None:
            logger.error("MQ Payload is invalid ")
            return Err.ERR_INVLAID_MQ_PAYLOAD

        if client is None:
            logger.error("Client is invalid ")
            return Err.ERR_INVLAID_MQ_CLIENT

        ret_val = client.publish(topic, obj, qos, retain)

        logger.debug("Exit MQTTClient: Publish Topic")

        return ret_val

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            super(MQTTClient, self).__repr__()
        )
