import os
import json
import logging

from nvsm_dshm.notifier.state_machine import StateMachine, state, retry
from nvsm_dshm.policy import IPolicy
from nvsm_dshm.notifier.call_home import webserver
from nvsm_dshm.notifier.call_home import log
from nvsm_dshm.notifier.call_home.offlinecallhome import OfflineCallHomeDestination
from nvsm_dshm.mqclient import MQBrokerClient as MQBroker
from nvsm_dshm.infra import ErrCode as Err
from nvsm_dshm.thea import Topic
from nvsm_dshm.infra import config_settings
from nvsm_dshm.ds import AlertsDataStore
from nvsm_dshm.equipment import Alert
from nvsm_dshm.ds import AlertsNvsmDB
from nvsm_dshm.equipment import catalog

logger = logging.getLogger(__name__)

#__DEVICE_TYPE_MAP maps the device type defined in the array DataStoreTypes in lib/apihandlers/api_util.go,
#with the map __URI_MAP defined in ds.py.
#This mapping is required to store the alerts under proper URI.
__DEVICE_TYPE_MAP = {
        #'Volume': '/nvsm/v1/Alerts/Storage',
        #'Drive': '/nvsm/v1/Alerts/Storage',
        #'StorageController': '/nvsm/v1/Alerts/Storage',
        'Storage':'Volume',

        #'Cpu': '/nvsm/v1/Alerts/Processors',
        'Processors':'Cpu',

        #'Gpu': '/nvsm/v1/Alerts/GPUs',
        #'NVLink': '/nvsm/v1/Alerts/GPUs',
        'GPUs':'Gpu',

        #'Fan': '/nvsm/v1/Alerts/Thermal',
        #'Pdb': '/nvsm/v1/Alerts/Thermal',
        'Thermal':'Fan',

        #'PowerSupplyUnit': '/nvsm/v1/Alerts/Power',
        #'PowerState': '/nvsm/v1/Alerts/Power',
        'Power': 'PowerSupplyUnit',

        #'Dimm': '/nvsm/v1/Alerts/Memory',
        'Memory': 'Dimm',

        #'PCIe': '/nvsm/v1/Alerts/PCIeDevices',
        'PCIeDevices':'PCIe',

        #'NVSwitch': '/nvsm/v1/Alerts/NVSwitches',
        'NVSwitches': 'NVSwitch',

        #'dcshwapikey': '/nvsm/v1/Alerts/Systems'
        'Systems': 'dcshwapikey'
    }


# TODO: Refactor code, move everything related to on_demand call home to a different file and simply call the
# methods from this file.


class CallHome(StateMachine):
    name = 'CallHome'
    states = {}
    start_state = ''

    def __init__(self, server_url, stage, version, batch, on_demand_op=None, is_diag_test=False):
                 #jwt_filename
        super(CallHome, self).__init__()
        self.e_context['notifications'] = batch
        self.e_context['server_url'] = '/'.join([server_url, stage, version])
        #self.e_context['jwt_filename'] = jwt_filename
        self.e_context['on_demand_op'] = on_demand_op
        self.e_context['mq_client'] = None
        self.e_context['diag_test'] = is_diag_test
        self.e_context['failure_text'] = None

        if on_demand_op:
            # FIXME: Notifier main thread creates an MQTT Client instance, which is used to subscribe to alert topics.
            # Use the same instance to publish On Demand call home status, DO
            # NOT create anyother client instance here.
            mq_client = MQBroker.factory(MQBroker.MQTT_CLIENT,
                                         'nvsm_on_demand_client',
                                         None)
            if mq_client == Err.ERR_INVLAID_MQ_SECURITY_CERT or mq_client == Err.ERR_INVLAID_MQ_ADDR_PORT:
                # TODO: Add appropriate handling here when state machine init
                # fails.
                logger.critical('Failed to create MQBroker Client')
            else:
                self.e_context['mq_client'] = mq_client
                mq_client.loop_start()

    def terminate(self):
        super(CallHome, self).terminate()
        error_message = 'Call Home operation {}'.format(self.result)

        if 'log_location' in list(self.e_context.keys()) and self.e_context['log_location'] and \
                os.path.exists(self.e_context['log_location']) and \
                'upload_dump_file' not in list(self.e_context.keys()):
            os.remove(self.e_context['log_location'])
            self.e_context['log_location'] = None

        if self.result == 'Failed' or ('upload_dump_file' in list(self.e_context.keys()) and \
                self.e_context['upload_dump_file'] == "failed"):
            error_message = " "
            if 'upload_dump_file' in list(self.e_context.keys()) and \
                    'log_location' in list(self.e_context.keys()) and self.e_context['log_location']:
                error_message = 'Call Home operation failed in upload log to NVES portal.File name {}'.format(self.e_context['log_location'].decode())
            else:
                error_message = 'Call Home operation {}'.format(self.result+".\n"+self.e_context['failure_text'])
                
            logger.debug(error_message)    

        if self.e_context['mq_client']:
            mqtt_notify(
                self.e_context,
                self.result,
                error_message)
            self.e_context['mq_client'].loop_stop()
            self.e_context['mq_client'].disconnect()


def mqtt_notify(e_context, state, message, case_id = "None"):
    if not e_context['on_demand_op']:
        return True

    e_context['on_demand_op']['State'] = state
    e_context['on_demand_op']['Message'] = message
    e_context['on_demand_op']['CaseId'] = case_id
    err = MQBroker.publish(e_context['mq_client'].client,
                           Topic.TOPIC_CALLHOME_STATUS,
                           json.dumps(e_context['on_demand_op']), 1, 0)
    if err == Err.SUCCESS:
        return True

    return False


def has_new(notifications):
    first_n = notifications[0]
    if first_n['severity'] != 'Cleared':
        return True
    return False


def check_callhome_policy(e_context):
    e_context['offline_call_home_enabled'] = False
    callhome_policy = IPolicy.factory('CallHome')
    if callhome_policy and (callhome_policy.get()['enable'] or e_context['on_demand_op']):
        e_context['policy'] = callhome_policy.get()
        return True
    else:
        offlinecallhome_policy = IPolicy.factory('OfflineCallHome')
        if offlinecallhome_policy and offlinecallhome_policy.get()['enable']:
            e_context['policy'] = offlinecallhome_policy.get()
            e_context['offline_call_home_enabled'] = True
            logger.debug('e_context[offline_call_home_enabled]: {}'.format(e_context['offline_call_home_enabled']))
            return True
        else:
            return False
    return False


@CallHome.add(start_state=True)
@state(on_success='authorize_server', on_failure='terminate')
@retry()  # Fix this better later, if retry is not required don't mention the decorator, but how do we handle cancel_event argument?
def get_config(e_context):
    e_context['failure_text'] = "\t\t\tFailed in stage: get_config.\n"

    msg_str = 'Started execution of Call Home operation(attempt:{})'.format(e_context['retry_attempt'])

    try:
        mqtt_notify(e_context,'Running', msg_str)
        if (check_callhome_policy(e_context) or e_context['on_demand_op']) \
        and log.dump_supported():
            return True
        else:
            return False
    except Exception as excep:
        logger.error("Failed in get_config stage, {}".format(str(excep)))
        return False

@CallHome.add
@state(on_success='collect_log', on_failure='terminate')
@retry(retries=5, timeout=40, exponential=True)
def authorize_server(e_context):

    if e_context['offline_call_home_enabled'] and not e_context['on_demand_op']:
        return True
    e_context['failure_text'] = "\t\t\tFailed in stage: authorize_server.\n" \
                                "\t\t\tCheck the following settings:\n" \
                                "\t\t\t 1. Server/Asset registered in NVEX portal.\n" \
                                "\t\t\t 2. Port 443 is enabled.\n" \

    msg_str = 'Getting server authorized(attempt:{})'.format(e_context['retry_attempt'])

    error_settings_msg = "\t\t\tCheck the following settings:\n" \
                         "\t\t\t 1. Server/Asset registered in NVEX portal.\n" \
                         "\t\t\t 2. Port 443 is enabled.\n" \

    error_msg_str =  msg_str +"\n" + error_settings_msg

    if e_context['retry_attempt'] > 1:
        mqtt_notify(e_context, 'Running', error_msg_str)
    else:
        mqtt_notify(e_context, 'Running', msg_str)

    notification = e_context['notifications'][0]

    ret = None

    try:
        ret = webserver.validate_server(
                notification,
                e_context['server_url'])
                #e_context['jwt_filename'],
    except Exception as excep:
        logger.error("Failed in authorize_server stage, {}".format(str(excep)))
        return False

    if ret['error']:
        logger.error(
             "authorize_server:Failed in authorizing the server {}".format(notification))
        return False

    if not 'authorized_server' in list(ret.keys()):
        logger.error("authorize_server:Key not present {}".format(notification))
        return False 
    if ret['authorized_server'] is False:
        logger.error("authorize_server:Server not registered in NVES portal {}".format(notification))
        return False 

    return True


#@CallHome.add
#@state(on_success='collect_log', on_failure='do_registration')
#@retry()
#def check_token_expiry(e_context):
#    mqtt_notify(e_context, 'Running', 'Validating credentials')
#
#    ret = webserver.check_token(e_context['jwt_filename'])
#
#    if ret['error']:
#        logger.debug("Failed in check_token_expiry")
#        return False
#
#    logger.debug("JWT has not expired.")
#    return True


#@CallHome.add
#@state(on_success='collect_log', on_failure='terminate')
#@retry(retries=6, timeout=60, exponential=True)
#def do_registration(e_context):
#    mqtt_notify(e_context, 'Running', 'Registration in progress')
#    notification = e_context['notifications'][0]
#
#    if(not 'system_serial' in notification):
#        logger.error("System serial not present")
#        return False
#
#    ret = webserver.register(
#        notification,
#        e_context['server_url'],
#        e_context['jwt_filename'],
#        e_context['cert_filename'])
#    if ret['error']:
#        logger.debug("Failed in registration process")
#        return False
#
#    return True


@CallHome.add
@state(on_success='call_home_first', on_failure='terminate')
@retry(retries=6, timeout=40, exponential=True)
def collect_log(e_context):
    e_context['failure_text'] = "\t\t\tFailed in stage: collect_log.\n"

    if e_context['diag_test']:
        return True

    msg_str = 'Collecting logs(attempt:{})'.format(e_context['retry_attempt'])
    mqtt_notify(e_context, 'Running', msg_str)

    # If there are no new notifications, then do not collect the logs
    if not has_new(e_context['notifications']):
        logger.debug(
            "Skip log file collection, there are no new notifications")
        return True

    log_file = None

    try:
        log_file = log.collect()
        e_context['log_location'] = log_file
        logger.debug('Location {}'.format(e_context['log_location']))
    except Exception as excep:
        logger.error("Failed in collect_log stage, {}".format(str(excep)))
        return False

    if log_file:
        return True
    else:
        return False


@CallHome.add
@state(on_success='upload_log', on_failure='terminate')
@retry(retries=10, timeout=30)
def call_home_first(e_context):
    e_context['failure_text'] = "\t\t\tFailed in stage: create case.\n"

    if e_context['diag_test']:
        return True

    if (e_context['offline_call_home_enabled'] and not e_context['on_demand_op']):
        return True

    msg_str = 'Creating Call Home(attempt:{})'.format(e_context['retry_attempt'])
    mqtt_notify(e_context, 'Running', msg_str)

    notification = e_context['notifications'][0]

    ret = None
    try:
        ret = webserver.notify(
               notification,
               e_context['server_url'],
               #e_context['jwt_filename'],
               e_context['policy']['email_contact'])
    except Exception as excep:
        logger.error("Failed in create case stage, {}".format(str(excep)))
        return False

    e_context['notifications'][0]['raised_ticket'] = False

    if ret['error']:
        logger.error(
            "Failed to callhome for notification {}".format(notification))
        return False

    e_context['log_url'] = ret['log_url']
    e_context['log_filename'] = ret['log_filename']
    e_context['secure_keys'] = ret['secure_keys']
    e_context['notifications'][0]['raised_ticket'] = True
    return True


@CallHome.add
@state(on_success='call_home_rest', on_failure='handle_log_upload_failure')
@retry(retries=5, timeout=60, exponential=True)
def upload_log(e_context):
    e_context['failure_text'] = "\t\t\tFailed in stage: upload log to NVES portal.\n"

    if e_context['diag_test']:
        return True

    msg_str = 'Uploading log file(attempt:{})'.format(e_context['retry_attempt'])
    mqtt_notify(e_context, 'Running', msg_str)

    try:
        # Upload the logs only if they were collected
        if 'log_location' in list(e_context.keys()) and e_context['log_location']:
            if e_context['offline_call_home_enabled'] and not e_context['on_demand_op']:
                log_destination_client = OfflineCallHomeDestination.get_client(e_context)
                log_destination_client.dump_metadata_file()
                log_destination_client.connect()
                log_destination_client.transfer_dump()
                log_destination_client.disconnect()
                return True
            else:
                ret = webserver.upload_log(e_context['log_url'],
                                   e_context['secure_keys'],
                                   e_context['log_location'])
                return ret
    except Exception as excep:
        logger.error("Failed in upload log stage, {}".format(str(excep)))
        return False
    
    return True

@CallHome.add
@state(on_success='call_home_rest', on_failure='terminate')
@retry(retries=5, timeout=10)
def handle_log_upload_failure(e_context):
    e_context['failure_text'] = "\t\t\tFailed in stage: handle log upload failure.\n"

    if e_context['diag_test']:
        return True

    if e_context['offline_call_home_enabled'] and not e_context['on_demand_op']:
        return True

    msg_str = 'Creating Call Home(attempt:{})'.format(e_context['retry_attempt'])
    mqtt_notify(e_context, 'Running', msg_str)

    for index, notification in enumerate(e_context['notifications']):
         notification['raised_ticket'] = False
         notification['log_file_upload'] = "failed"
         if 'log_location' in e_context and  e_context['log_location']:
              notification['log_file_location'] = e_context['log_location'].decode()

    e_context['upload_dump_file'] = "failed"

    return True

@CallHome.add
@state(on_success='query_tickets', on_failure='terminate')
#@state(on_success='terminate', on_failure='terminate')
@retry(retries=10, timeout=30)
def call_home_rest(e_context):
    e_context['failure_text'] = "\t\t\tFailed in stage: create case for batched alerts.\n"

    if e_context['diag_test']:
        return True

    if e_context['offline_call_home_enabled'] and not e_context['on_demand_op']:
        return True

    for index, notification in enumerate(e_context['notifications']):
        #if not notification:
        if 'raised_ticket' in list(notification.keys()) and notification['raised_ticket'] is True:   
            continue

        msg_str = 'Creating Call Home for batched entries(attempt:{})'.format(e_context['retry_attempt'])
        mqtt_notify(e_context, 'Running', msg_str)

        ret = None
        try:
            ret = webserver.notify(
                    notification,
                    e_context['server_url'],
                    #e_context['jwt_filename'],
                    e_context['policy']['email_contact'],
                    e_context['log_filename'])
        except Exception as excep:
            logger.error("Failed in create case for batched alerts stage, {}".format(str(excep)))
            return False


        e_context['notifications'][index]['raised_ticket'] = False

        if ret['error']:
            logger.error(
                "Failed to callhome for notification {}".format(notification))
            return False

        e_context['notifications'][index]['raised_ticket'] = True
        #e_context['notifications'][index] = None

    return True

@CallHome.add
@state(on_success='terminate', on_failure='terminate')
@retry(retries=5, timeout=30)
def query_tickets(e_context):
    e_context['failure_text'] = "\t\t\tFailed in stage: query case/ticket number.\n"

    if e_context['diag_test']:
        return True

    if e_context['offline_call_home_enabled'] and not e_context['on_demand_op']:
        return True

    msg_str = 'Verifying ticket creation(attempt:{})'.format(e_context['retry_attempt'])
    mqtt_notify(e_context, 'Running', msg_str)

    device_types = None
    try:
        device_types = AlertsNvsmDB.get_supported_devices()
    except Exception as excep:
        device_types = None
        logger.error("Failed in retriving supported NVSM devices, {}".format(str(excep)))

    for index, notification in enumerate(e_context['notifications']):
        
        if not 'raised_ticket' in list(notification.keys()) or \
                       notification['raised_ticket'] is False:   
            continue


        if  'severity' in e_context['notifications'][index] and \
                     e_context['notifications'][index]['severity'] == 'Cleared':
            logger.debug("Query Ticket: Alert skipped as its in Cleared state")
            continue

        ret = None
        try:
            ret = webserver.query_tickets(
                        notification,
                        e_context['server_url'])
                        #e_context['jwt_filename'],
        except Exception as excep:
            logger.error("Failed in query case number stage, {}".format(str(excep)))
            return False

        if ret['error']:
            logger.error(
                "Failed while querying ticket/case id {}".format(notification))
            return False

        if not 'case_id' in list(ret.keys()):
            logger.error(
                    "query_tickets: case_id not present {}".format(notification))
            return False

        e_context['notifications'][index]['case_id'] = ret['case_id']

        if(e_context['on_demand_op']):
            mqtt_notify(e_context, 'Running', 'Ticket created.',ret['case_id'])
            return True

        if ( (not 'severity' in e_context['notifications'][index]) or 
             (not 'type' in e_context['notifications'][index]) or
             (not 'component_id' in e_context['notifications'][index]) ):
            logger.error("Query Ticket: Alert not in proper format")
            continue

        if e_context['notifications'][index]['severity'] != 'Cleared':

            if 'notifier' in e_context['notifications'][index]:
                e_context['notifications'][index]['notifier'] = 'call_home' + '|' + e_context['notifications'][index]['notifier']
            else:
                e_context['notifications'][index]['notifier'] = 'call_home' + '|'
                
            alerts_map = {}
            device_type = e_context['notifications'][index]['type']

            if device_types and device_type not in device_types:
                if device_type in __DEVICE_TYPE_MAP:
                    device_type = __DEVICE_TYPE_MAP[device_type]
                else:
                    logger.error("Failed to find the device type {}".format(device_type))
                    continue
            try:
                data_store = AlertsDataStore.factory('thea_db', device_type)

                #Check whether the Alert is a dynamic alert, if so fill the required details manually.  
                event_id = catalog.get(e_context['notifications'][index]['alert_id'])

                if event_id is None:
                    dynamic_attr = {}
                    dynamic_attr['severity'] = e_context['notifications'][index]['severity']
                    dynamic_attr['description'] = e_context['notifications'][index]['description']
                    dynamic_attr['message'] = e_context['notifications'][index]['message']
                    dynamic_attr['message_details'] = e_context['notifications'][index]['message_details']
                    dynamic_attr['recommended_action'] = e_context['notifications'][index]['recommended_action']
                    e_context['notifications'][index]['dynamic_attrs'] = dynamic_attr

                alert = Alert.create(e_context['notifications'][index])
                alert.asserted = False
                alert.notifier_update = True
                alerts_map[alert.alert_id()] = alert
                data_store.store_alerts(e_context['notifications'][index]['component_id'],alerts_map)
            except Exception as excep:
                logger.error("Failed in updating the case details to DB, {}".format(str(excep)))

    return True
