import os
import logging

import json
import requests
import jwt
import datetime

logger = logging.getLogger(__name__)
TIME_BEFORE_EXPIRY = 15  # in number of Days


def to_json(notification, email_contact, log_filename, log_format):
    obj = {}

    obj['system_serial'] = notification['system_serial']
    obj['alert_id'] = notification['alert_id']
    obj['component_id'] = notification['component_id']
    obj['event_time'] = notification['event_time']
    obj['severity'] = notification['severity']
    obj['clear_time'] = notification['clear_time']
    obj['description'] = notification['description']
    obj['message_details'] = notification['message_details']
    obj['message'] = notification['message']
    obj['system_name'] = notification['system_name']

    obj['email_contact'] = email_contact
    if 'email' in notification:
        obj['email_contact'] = notification['email']

    obj['log_filename'] = None
    obj['batch_id'] = notification['batch_id']
    if notification['severity'] != 'Cleared':
        obj['log_filename'] = log_filename

    if 'log_file_upload' in notification and notification['log_file_upload']:
        obj['log_file_upload'] = notification['log_file_upload']

    if 'log_file_location' in notification and  notification['log_file_location']:
        obj['log_file_location'] = notification['log_file_location']


    # obj['log_file_format'] = log_format
    return json.dumps([obj])


def check_token(jwt_filename):
    if((jwt_filename is None) or
            (os.path.exists(jwt_filename) == False)):
        logger.error("call home jwt token file is None or not present")
        return {'error': True}

    jwt_token = None
    val = 0
    payload = None

    try:
        with open(jwt_filename, "rb") as f:
            jwt_token = f.read()
    except Exception as excep:
        logger.error(
            "validate_credentials:Exception in reading jwt from file, error {}".format(excep))
        return {'error': True}

    try:
        payload = jwt.decode(jwt_token, verify=False)
    except Exception as excep:
        logger.error("Failed to decode jwt, error {}".format(excep))
        return {'error': True}
    try:
        exp = datetime.datetime.fromtimestamp(payload['exp'])
        cur = datetime.datetime.utcnow()
        val = exp - cur
        logger.debug(
            "JWT-TOKEN: {}, expiry in {} days".format(payload, val.days))
    except Exception as excep:
        logger.error(
            "Failed in extracting expiry date from JWT, error {}".format(excep))
        return {'error': True}

    if(val.days < TIME_BEFORE_EXPIRY):
        logger.error(
            "JWT expires soon,in days {}.Get registration done.".format(val.days))
        return {'error': True}

    return {'error': False}


def notify(notification,
           server_url,
           #jwt_filename,
           email_contact=None,
           log_filename=None,
           time_out=60,
           log_format='tar.xz'):

    #if((jwt_filename is None) or (os.path.exists(jwt_filename) == False)):
    #    logger.error("CallHome jwt file is None or not present")
    #    return {'error': True, 'log_url': None, 'log_filename': None, 'secure_keys': None}

    json_data = to_json(notification, email_contact, log_filename, log_format)
    server_url = server_url + '/call-home'
    ret = None

    if((notification is None) or (not 'system_serial' in notification)):
        logger.error("Notify:System serial not present")
        return {'error': True, 'log_url': None, 'log_filename': None, 'secure_keys': None}

    serial_number = notification['system_serial']

    #jwt_token = None
    #try:
    #    with open(jwt_filename, "rb") as f:
    #        jwt_token = f.read()
    #except Exception as excep:
    #    logger.error(
    #        "Exception occurred while reading jwt token from file {}".format(excep))
    #    return {'error': True, 'log_url': None, 'log_filename': None, 'secure_keys': None}

    if(serial_number is None):
        logger.error("Notify:CallHome invalid system serial number")
        return {'error': True, 'log_url': None, 'log_filename': None, 'secure_keys': None}

    #if(jwt_token is None):
    #    logger.error("Notify:CallHome invalid jwt token")
    #    return {'error': True, 'log_url': None, 'log_filename': None, 'secure_keys': None}

    headers = {"Content-Type": "application/json",
               "system_serial": serial_number,
               "nvsm_app"     : "True" 
               #"authorization": jwt_token
               }

    try:
        logger.debug(
            "Json data sent to call home {}".format(json.dumps(json_data)))
        ret = requests.put(server_url, json=json_data, headers=headers,timeout=time_out)
    except requests.exceptions.RequestException as excep:
        logger.exception(
            'HTTP PUT of data ({}) for url {} failed : '.format(json_data, server_url))

    if ret:
        if ret.status_code == 200:
            response = json.loads(ret.text)
            if 'status_code' in list(response.keys()) and response['status_code'] == 200:
                if log_filename or notification['severity'] == 'Cleared':
                    logger.debug(
                        "Call home succeeded, response is {}".format(response))
                    return {'error': False, 'log_url': None, 'log_filename': log_filename, 'secure_keys': None}
                if log_filename is None and all(k in list(response.keys()) for k in ("url", "log_filename", "fields")):
                    logger.debug(
                        "Call home succeeded, response is {}".format(response))
                    return {'error': False, 'log_url': response['url'], 'log_filename': response['log_filename'], 'secure_keys': response['fields']}
                else:
                    logger.error(
                        "CallHome response did not contain the expected fields: {}".format(response))
            else:
                logger.error(
                    "Call home failed, response is {}".format(response))
        else:
            logger.error(
                "HTTPS PUT of {} failed, returned {}".format(
                    server_url,
                    ret.status_code))

    return {'error': True, 'log_url': None, 'log_filename': None, 'secure_keys': None}


def register(notification, server_url,
             jwt_filename, cert_filename, time_out=240):

    ret = None

    if((jwt_filename is None)):
        logger.error("CallHome jwt file is None")
        return {'error': True}

    if((cert_filename is None) or (os.path.exists(cert_filename) == False)):
        logger.error(
            "CallHome middleware server certificate is None or not present")
        return {'error': True}

    if((notification is None) or (not 'system_serial' in notification)):
        logger.error("System serial not present")
        return {'error': True}

    serial_number = notification['system_serial']

    if(serial_number is None):
        logger.error("CallHome invalid system serial  number")
        return {'error': True}

    server_url = server_url + '/call-home'
    ret = None
    headers = {"Content-Type": "application/json",
               "system_serial": serial_number,
               "nvsm_app"     : "True", 
               "register": "True"
               }

    try:
        json_data = to_json(notification, None, None, None)
        logger.debug(
            "Register: Json data sent to call home {}".format(json.dumps(json_data)))
        ret = requests.put(server_url, json=json_data, headers=headers,
                           verify=cert_filename, timeout=time_out)
    except requests.exceptions.RequestException as excep:
        logger.exception(
            'Register: HTTP PUT of data ({}) for url {} failed : '.format(json_data, server_url))

    if ret:
        if ret.status_code == 200:
            response = json.loads(ret.text)
            if 'status_code' in list(response.keys()) and response['status_code'] != 200:
                if 'message' in list(response.keys()):
                    logger.debug(
                        "Call home registration failed, response is {}".format(response['message']))
                return {'error': True}

            if not 'jwt_token' in list(response.keys()):
                logger.debug("Call home registration failed, jwt not found.")
                return {'error': True}

            jwt_token = response['jwt_token']

            try:
                with open(jwt_filename, "w") as f:
                    f.write(jwt_token)
            except Exception as excep:
                logger.error(
                    "Exception occurred while updating jwt token to file {}".format(excep))
                return {'error': True}

            return {'error': False}

    return {'error': True}


def upload_log(log_url, secure_keys, log_file, time_out=240):

    ret = None

    try:
        log_data = open(log_file.decode(), 'rb')

        ret = requests.post(log_url,
                            data=secure_keys,
                            files={'file': log_data},
                            timeout=time_out)

    except Exception as excep:
        logger.exception(
            "Exception occurred while uploading the log {}, error {}".format(log_file, excep))
    finally:
        log_data.close()

    if ret and ret.status_code == 200 or ret.status_code == 201 or ret.status_code == 204:
        logger.debug("Uploading the log file {} succeeded".format(log_file))
        return True

    logger.debug("Uploading the log file {} failed".format(log_file))
    return False

def query_tickets(notification,
           server_url,
           #jwt_filename,
           time_out=240):

    #if((jwt_filename is None) or (os.path.exists(jwt_filename) == False)):
    #    logger.error("CallHome:query_tickets jwt file is None or not present")
    #    return {'error': True, 'case_id': None}


    json_data = to_json(notification, None, None, None)
    server_url = server_url + '/call-home'
    ret = None

    if((notification is None) or (not 'system_serial' in notification)):
        logger.error("query_tickets:System serial not present")
        return {'error': True, 'case_id': None}

    serial_number = notification['system_serial']

    #jwt_token = None
    #try:
    #    with open(jwt_filename, "rb") as f:
    #        jwt_token = f.read()
    #except Exception as excep:
    #    logger.error(
    #            "query_tickets:Exception occurred while reading jwt token from file {}".format(excep))
    #    return {'error': True, 'case_id': None}

    if(serial_number is None):
        logger.error("query_tickets:CallHome invalid system serial number")
        return {'error': True, 'case_id': None}

    #if(jwt_token is None):
    #    logger.error("query_tickets:CallHome invalid jwt token")
    #    return {'error': True, 'case_id': None}

    headers = {"Content-Type": "application/json",
               "system_serial": serial_number,
               #"authorization": jwt_token,
               "query_tickets": "True",
               "nvsm_app"     : "True" 
               }

    try:
        logger.debug(
                "query_tickets:Json data sent to call home {}".format(json.dumps(json_data)))

        #TODO: Change it to get Method.
        ret = requests.put(server_url, json=json_data, headers=headers, timeout=time_out)
    except requests.exceptions.RequestException as excep:
        logger.exception(
                'query_tickets::HTTP PUT of data ({}) for url {} failed : '.format(json_data, server_url))

    if ret:
        if ret.status_code == 200:
            response = json.loads(ret.text)
            if 'status_code' in list(response.keys()) and response['status_code'] == 200:

                if 'case_id' in list(response.keys()):
                    logger.debug(
                        "Call home succeeded, response is {}".format(response))

                    return {'error': False, 'case_id':response['case_id']}
                else:
                    logger.error(
                     "query_tickets:CallHome response did not contain the expected fields: {}".format(response))
            else:
                logger.error(
                    "query_tickets:Call home failed, response is {}".format(response))
        else:
            logger.error(
                "query_tickets: HTTPS PUT of {} failed, returned {}".format(
                    server_url,
                    ret.status_code))

    return {'error': True, 'case_id': None}


def validate_server(notification,
           server_url,
           #jwt_filename,
           time_out=240):

    #if((jwt_filename is None) or (os.path.exists(jwt_filename) == False)):
    #    logger.error("CallHome:validate_server jwt file is None or not present")
    #    return {'error': True, 'authorized_server': False}

    json_data = to_json(notification, None, None, None)
    server_url = server_url + '/call-home'
    ret = None

    if((notification is None) or (not 'system_serial' in notification)):
        logger.error("validate_server:System serial not present")
        return {'error': True, 'authorized_server': False}

    serial_number = notification['system_serial']

    #jwt_token = None
    #try:
    #    with open(jwt_filename, "rb") as f:
    #        jwt_token = f.read()
    #except Exception as excep:
    #    logger.error(
    #            "validate_server:Exception occurred while reading jwt token from file {}".format(excep))
    #    return {'error': True, 'authorized_server': False}

    if(serial_number is None):
        logger.error("validate_server:CallHome invalid system serial number")
        return {'error': True, 'authorized_server': False}

    #if(jwt_token is None):
    #    logger.error("validate_server:CallHome invalid jwt token")
    #    return {'error': True, 'authorized_server': False}

    headers = {"Content-Type": "application/json",
               "system_serial": serial_number,
               #"authorization": jwt_token,
               "authorize_server": "True",
               "nvsm_app"     : "True" 
               }

    try:
        logger.debug(
                "validate_server:Json data sent to call home {}".format(json.dumps(json_data)))

        #TODO: Change it to get Method.
        ret = requests.put(server_url, json=json_data, headers=headers, timeout=time_out)
    except requests.exceptions.RequestException as excep:
        logger.exception(
                'validate_server:HTTP PUT of data ({}) for url {} failed : '.format(json_data, server_url))

    if ret:
        if ret.status_code == 200:
            response = json.loads(ret.text)
            if 'status_code' in list(response.keys()) and response['status_code'] == 200:

                if 'authorized_server' in list(response.keys()):
                    logger.debug(
                        "validate_server:Call home response is {}".format(response))

                    return {'error': False, 'authorized_server':response['authorized_server']}
                else:
                    logger.error(
                     "validate_server:CallHome response did not contain the expected fields: {}".format(response))
            else:
                logger.error(
                    "validate_server:Call home failed, response is {}".format(response))
        else:
            logger.error(
                "validate_server: HTTPS PUT of {} failed, returned {}".format(
                    server_url,
                    ret.status_code))

    return {'error': True, 'authorized_server': False}
