#!/usr/bin/python
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Notifier module

This module sends E-Mail and CallHome Notifications.
"""

import time
import os
import sys
import signal
import logging
import json
import syslog
from functools import partial


from nvsm_dshm.thea import Endpoint
from nvsm_dshm.equipment import catalog
from nvsm_dshm.policy import policy
from nvsm_dshm.infra import config_settings
from nvsm_dshm.infra import log
from nvsm_dshm.mqclient import MQBrokerClient as MQBroker
from nvsm_dshm.thea import Topic
from nvsm_dshm.infra import ErrCode as Err

from nvsm_dshm.notifier.call_home import OnDemandCHTask
from nvsm_dshm.notifier.call_home import CallHomeTask
from nvsm_dshm.notifier.pager_duty import OnDemandPDTask
from nvsm_dshm.notifier.soak import SoakTask

NVSM_CONN_RETRY_TIME = 5  # in seconds
NVSM_RETRY_ATTEMPT_ON_MESSAGE = 5
NVSM_CLIENT_NAME = "nvsm_notifier_client"

logger = logging.getLogger(__name__)

SOAK_TASK = None
ONDEMAND_NOTIFIER = None
ONDEMAND_PD_NOTIFIER = None


def init():
    """
        Initializer for the notifier entry module.
        1) Initializes the log file.
        2) Copies the static alerts data to memory.
    """
    # read the config filename from system
    try:
        config_settings(config_file_name=os.environ['DSHM_CONFIG_FILE'])
    except Exception as e:
        syslog.syslog(syslog.LOG_CRIT,
                        '__name__ + __file__ + : Unable to read the config file')
        return Err.ERROR
    # Init logger
    ret = log.initialize(config_settings().get('logging', 'config-file'),
                         config_settings().get('logging', 'log-dir'),
                         'notifier')
    if ret is False:
        return Err.ERROR

    logger.info('Initializing Notifier Service')

    catalog.init('environmental', 'storage', 'pcie', 'system', 'network', 'dcs_hw_api_key')

    ret_val = Err.ERROR
    while ret_val is not Err.SUCCESS:
        ret_val = thea_connect()
        logger.debug("Thea_Connect ret:" + str(ret_val))
        if ret_val is not Err.SUCCESS:
            time.sleep(NVSM_CONN_RETRY_TIME)

    logger.debug('Exit')
    return Err.SUCCESS


def thea_connect():
    """
        Checks connectivity to Thea is established.
    """
    logger.debug("Thea_connect Enter")

    endpoint = None
    nvsm_api_config_file = config_settings().get('nvsm-api', 'config-file')
    nvsm_api_jwt_file = config_settings().get('nvsm-api', 'jwt-file')

    (ca_cert_file,
     _,
     _,
     _,
     port,
     address,
     _) = config_settings.nvsm_api_config_helper(nvsm_api_config_file)

    # TODO change Endpoint arguments to kargs
    if port is not None:
        endpoint = Endpoint(port=str(port), jwt_file=nvsm_api_jwt_file,
                            ca_cert_file=ca_cert_file, ip_addr=address)
    else:
        endpoint = Endpoint()

    if not endpoint.accessible():
        logger.error('Endpoint is in-accessible')
        return Err.ERR_CONN_REST_SRV

    logger.debug('Endpoint is accessible')

    policy.push_defaults(['EmailConfig', 'CallHome', 'PagerDuty','AlertFilter', 'OfflineCallHome'])

    logger.debug("Thea_connect Exit")

    return Err.SUCCESS


def on_connect(client, userdata, flags, rc):
    """
        On Connect callback that is registered with MQTT client.
    """

    logger.debug("MQ ON Connect Ret-Code:" + str(rc))

    ret_val = Err.ERROR

    # useful when Thea API Server goes down,helps in reconnection.
    while ret_val is not Err.SUCCESS:
        ret_val = thea_connect()
        logger.debug("Thea_Connect ret:" + str(ret_val))
        if ret_val is not Err.SUCCESS:
            time.sleep(NVSM_CONN_RETRY_TIME)

    if rc == 0:
        # useful when mqtt broker goes down,
        # in reconnection we will still subscribe to all topics.
        logger.debug("MQ Client Connect Success, result code: " + str(rc))
        for topic in Topic.ALERT_TOPICS:
            MQBroker.subscribe(client, topic)
            logger.debug("Registered with MQTT for topic {}".format(topic))

        MQBroker.subscribe(client, Topic.TOPIC_CALLHOME)
        MQBroker.subscribe(client, Topic.TOPIC_PAGERDUTY)
        MQBroker.subscribe(client,Topic.TOPIC_CALLHOME_ENABLED)
        logger.debug("Registered with MQTT for topic {}".format(topic))

    else:
        logger.error("MQ Client Connect failure, with result code: " + str(rc))

    logger.debug("MQ ON Connect Handled")


def on_message(client, userdata, msg):
    """
        On Message callback registered with MQTT Client.
    """
    global ONDEMAND_NOTIFIER
    global ONDEMAND_PD_NOTIFIER
    global SOAK_TASK

    logger.debug("MQ Message Received for Topic: " + msg.topic)
    logger.debug("Payload is: " + msg.payload.decode('utf-8'))

    if msg is None or msg.payload is None:
        return

    payload = json.loads(msg.payload)

    if msg.topic == Topic.TOPIC_CALLHOME:
        ONDEMAND_NOTIFIER.put(payload)
        logger.debug("MQ Message for Topic: " + msg.topic + " Handled")
        return

    if msg.topic == Topic.TOPIC_PAGERDUTY:
        ONDEMAND_PD_NOTIFIER.put(payload)
        logger.debug("MQ Message for Topic: " + msg.topic + " Handled")
        return

    if msg.topic == Topic.TOPIC_CALLHOME_ENABLED:
        CallHomeTask.push_existing_alerts(
                 config_settings().get('notifier', 'webserver'),
                 config_settings().get('notifier', 'stage'),
                 config_settings().get('notifier', 'version'))
        logger.debug("MQ Message for Topic: " + msg.topic + " Handled")
        logger.debug("Call Home pushed all active alerts in the system")
        return

    # Send the notification object (alert) to the soak task
    SOAK_TASK.put(msg)

    logger.debug("MQ Message for Topic: " + msg.topic + " Handled")


def graceful_exit(mq_client, signum, frame):
    """
    Handler to gracefully terminate all threads.
    """
    logger.info("Executing graceful exit")
    mq_client.disconnect()


def notifier_main():
    """
    Main
    """
    global ONDEMAND_NOTIFIER
    global ONDEMAND_PD_NOTIFIER
    global SOAK_TASK

    # Initialize the System monitoring subsystem.
    if init() != Err.SUCCESS:
        logger.critical('Failed to launch Notifier  monitor')
        sys.exit(1)

    SOAK_TASK = SoakTask()
    ONDEMAND_NOTIFIER = OnDemandCHTask(
        config_settings().get('notifier', 'webserver'),
        config_settings().get('notifier', 'stage'),
        config_settings().get('notifier', 'version'))
        #config_settings().get('notifier', 'jwt-file'),

    ONDEMAND_PD_NOTIFIER = OnDemandPDTask(
        config_settings().get('notifier', 'webserver'),
        config_settings().get('notifier', 'stage'),
        config_settings().get('notifier', 'version'))
        #config_settings().get('notifier', 'jwt-file'),



    ONDEMAND_NOTIFIER.start()
    ONDEMAND_PD_NOTIFIER.start()
    SOAK_TASK.start()

    # Initialize MQ Borker
    mq_client = MQBroker.factory(MQBroker.MQTT_CLIENT, NVSM_CLIENT_NAME,
                                 on_connect, on_message)
    if mq_client is None or \
            mq_client is Err.ERR_INVLAID_MQ_SECURITY_CERT or \
            mq_client is Err.ERR_INVLAID_MQ_ADDR_PORT:
        logger.critical('Failed in creating the MQBroker Client')
        sys.exit(1)

    # Register an handler for graceful shutdown.
    signal.signal(signal.SIGTERM, partial(graceful_exit, mq_client))

    logger.debug("Run the MQ Client and listen to MQ Broker")

    # Start listening to MQTT
    mq_client.listen_for_ever()

    logger.info("Exiting the main thread")
