import time
from datetime import datetime
import logging
from threading import Lock


from nvsm_dshm.notifier.pager_duty.state_machine import PagerDuty as PagerDutySM
from nvsm_dshm.notifier.notifier_thread import NotifierTask
from nvsm_dshm.thea import Endpoint

logger = logging.getLogger(__name__)
DIAG_TEST_MESSAGE = 'This is a diagnostic test for the server with system-name: '

def mock_notification(comment, diagTest=False):
    notification = {}

    notification['system_serial'] = Endpoint().serial()
    notification['alert_id'] = 'OnDemandPagerDuty'
    notification['component_id'] = 'Server'
    notification['event_time'] = datetime.now().astimezone().isoformat()
    notification['severity'] = 'Critical'
    notification['clear_time'] = ''
    notification['system_name'] = Endpoint().hostname()


    if diagTest:
        comment = DIAG_TEST_MESSAGE
        notification['severity'] = 'Warning'
        if notification['system_name']:
            comment = DIAG_TEST_MESSAGE +  notification['system_name']

    notification['description'] = comment #'User triggered pagerduty operation'

    notification['message_details'] = notification['description']
    notification['message'] = notification['message_details']
    notification['batch_id'] = None
    notification['batch_id'] = notification['event_time']
        
    return ([notification])

def timestamp(op):
    if not op:
        return 0

    op_start_time = datetime.strptime(op['StartTime'], "%Y-%m-%dT%H:%M:%SZ")
    epoch_time = datetime(1970, 1, 1, 0, 0, 0)
    return (op_start_time - epoch_time).total_seconds()


class OnDemandPDTask(NotifierTask):
    queue_size = 100
    name = 'OnDemandPDTask'

    def __init__(self, webserver, stage, version):
        super(OnDemandPDTask, self).__init__()
        self.webserver = webserver
        self.stage = stage
        self.version = version

        self.current_op = None
        self.state_machine = None

        self.lock = Lock()

    def put(self, op):
        if op['State'] != 'Cancelled' and op['State'] != 'Pending':
            logger.debug(
                "On Demand pagerduty operation {} is not supported".format(op['State']))
            return

        if op['State'] == 'Cancelled':
            with self.lock:
                if self.current_op and self.state_machine and self.current_op['StartTime'] == op['StartTime']:
                    self.state_machine.stop()
                    return

        if not self.queue.full():
            self.queue.put(op)
        else:
            logger.debug(
                "{} queue is full, unable to process the notification".format(self.name))

    def run(self):
        while True:
            op = self.queue.get()

            ops = {}
            while True:
                key = op['StartTime']
                if key in list(ops.keys()):
                    if op['State'] == 'Cancelled':
                        logger.debug("Op {} was dropped because it was cancelled before it could be processed.".format(
                            op['StartTime']))
                        ops[key] = None
                else:
                    if op['State'] == 'Pending':
                        ops[key] = op
                self.queue.task_done()

                if not self.queue.empty():
                    op = self.queue.get()
                else:
                    ops_list = [o for key, o in list(ops.items()) if o is not None]
                    op = None
                    if ops_list:
                        op = max(ops_list, key=timestamp)
                    break

            if op:
                with self.lock:
                    self.current_op = op
                    batch = mock_notification(op['Description'], op['DiagTest'])
                    self.state_machine = PagerDutySM(
                        self.webserver,
                        self.stage,
                        self.version,
                        batch,
                        op,
                        op['DiagTest'])

                self.state_machine.run()

                with self.lock:
                    self.op_timestamp = None
                    self.state_machine = None
