import os
import json
import logging

from nvsm_dshm.notifier.state_machine import StateMachine, state, retry
from nvsm_dshm.policy import IPolicy
from nvsm_dshm.notifier.pager_duty import webserver
from nvsm_dshm.mqclient import MQBrokerClient as MQBroker
from nvsm_dshm.infra import ErrCode as Err
from nvsm_dshm.thea import Topic
from nvsm_dshm.infra import config_settings

logger = logging.getLogger(__name__)


# TODO: Refactor code, move everything related to on_demand pagerduty to a different file and simply call the
# methods from this file.


class PagerDuty(StateMachine):
    name = 'PagerDuty'
    states = {}
    start_state = ''

    def __init__(self, server_url, stage, version, batch, on_demand_op=None, is_diag_test=False):
        super(PagerDuty, self).__init__()
        self.e_context['notifications'] = batch
        self.e_context['server_url'] = '/'.join([server_url, stage, version])
        self.e_context['on_demand_op'] = on_demand_op
        self.e_context['mq_client'] = None
        self.e_context['diag_test'] = is_diag_test
        self.e_context['failure_text'] = None

        if on_demand_op:
            # FIXME: Notifier main thread creates an MQTT Client instance, which is used to subscribe to alert topics.
            # Use the same instance to publish On Demand pagerduty status, DO
            # NOT create anyother client instance here.
            mq_client = MQBroker.factory(MQBroker.MQTT_CLIENT,
                                         'nvsm_pd_on_demand_client',
                                         None)
            if mq_client == Err.ERR_INVLAID_MQ_SECURITY_CERT or mq_client == Err.ERR_INVLAID_MQ_ADDR_PORT:
                # TODO: Add appropriate handling here when state machine init
                # fails.
                logger.critical('Failed to create MQBroker Client')
            else:
                self.e_context['mq_client'] = mq_client
                mq_client.loop_start()

    def terminate(self):
        super(PagerDuty, self).terminate()
        error_message = 'Pager Duty operation {}'.format(self.result)

        if 'log_location' in list(self.e_context.keys()) and self.e_context['log_location'] and os.path.exists(self.e_context['log_location']):
            os.remove(self.e_context['log_location'])
            self.e_context['log_location'] = None

        if self.result == 'Failed':
            error_message = None
            error_message = 'Pager Duty operation {}'.format(self.result+".\n"+self.e_context['failure_text'])

        if self.e_context['mq_client']:
            mqtt_notify(
                self.e_context,
                self.result,
                error_message)
            self.e_context['mq_client'].loop_stop()
            self.e_context['mq_client'].disconnect()


def mqtt_notify(e_context, state, message, case_id = "None"):
    if not e_context['on_demand_op']:
        return True

    e_context['on_demand_op']['State'] = state
    e_context['on_demand_op']['Message'] = message
    err = MQBroker.publish(e_context['mq_client'].client,
                           Topic.TOPIC_PAGERDUTY_STATUS,
                           json.dumps(e_context['on_demand_op']), 1, 0)
    if err == Err.SUCCESS:
        return True

    return False


def has_new(notifications):
    first_n = notifications[0]
    if first_n['severity'] != 'Cleared':
        return True

    return False


@PagerDuty.add(start_state=True)
@state(on_success='diagnostic_test', on_failure='terminate')
@retry()  # Fix this better later, if retry is not required don't mention the decorator, but how do we handle cancel_event argument?
def get_config(e_context):
    e_context['failure_text'] = "\t\t\tFailed in stage: get_config.\n"

    msg_str = 'Started execution of Pager Duty operation(attempt:{})'.format(e_context['retry_attempt'])

    try:
        mqtt_notify(e_context,
                'Running',
                msg_str)
        notification = e_context['notifications'][0]
        pagerduty_policy = IPolicy.factory('PagerDuty')
        e_context['policy'] = pagerduty_policy.get()

        if pagerduty_policy.get()['enable'] or e_context['on_demand_op']:
            if e_context['on_demand_op'] and 'ServiceKey' in e_context['on_demand_op'] and e_context['on_demand_op']['ServiceKey']:
                e_context['service_key'] = e_context['on_demand_op']['ServiceKey']
            elif 'service_key' in e_context['policy'] and e_context['policy'] ['service_key']:
                e_context['service_key'] = e_context['policy']['service_key']
            else:
                return False
        else:
            return False

    except Exception as excep:
        logger.error("Failed in get_config stage, {}".format(str(excep)))
        return False

    return True

@PagerDuty.add
@state(on_success='create_page', on_failure='terminate')
@retry(retries=6, timeout=40, exponential=True)
def diagnostic_test(e_context):

    if not e_context['diag_test']:
        logger.debug("Skip diagnostic_test,create_page issued")
        return True

    ROUTING_KEY = e_context['service_key']

    e_context['failure_text'] = "\t\t\tFailed in stage: diagnostic_test.\n" \
                                "\t\t\tCheck the following settings:\n" \
                                "\t\t\t 1. Check the Integration/Service key set in PagerDuty policy.\n" \
                                "\t\t\t 2. Port 443 is enabled.\n"

    error_settings_msg = "\t\t\tCheck the following settings:\n" \
                         "\t\t\t 1. Check the Integration/Service key set in PagerDuty policy.\n" \
                         "\t\t\t 2. Port 443 is enabled.\n"

    msg_str = 'diagnostic_test (attempt:{})'.format(e_context['retry_attempt'])

    error_msg_str =  msg_str +"\n" + error_settings_msg

    try:
        if e_context['retry_attempt'] > 1:
            mqtt_notify(e_context, 'Running', error_msg_str)
        else:
            mqtt_notify(e_context, 'Running', msg_str)

        notification = e_context['notifications'][0]

        ret = webserver.create_page(
          notification,
          ROUTING_KEY,
          e_context['diag_test'])

        if ret['error']:

            if "service_key" in notification:
                del notification["service_key"]

            logger.error(
            "Failed in diagnostic_test stage {}".format(notification))
            return False
    except Exception as excep:
        logger.error("Failed in diagnostic_test stage, {}".format(str(excep)))
        return False

    return True

@PagerDuty.add
@state(on_success='terminate', on_failure='terminate')
@retry(retries=6, timeout=40, exponential=True)
def create_page(e_context):

    if e_context['diag_test']:
        logger.debug("Diagnostic Test issued,skip create_page stage")
        return True

    e_context['failure_text'] = "\t\t\tFailed in stage: Create Page.\n" \
                                "\t\t\tCheck the following settings:\n" \
                                "\t\t\t 1. Check the Integration/Service key set in PagerDuty policy.\n" \
                                "\t\t\t 2. Port 443 is enabled.\n"

    error_settings_msg = "\t\t\tCheck the following settings:\n" \
                         "\t\t\t 1. Check the Integration/Service key set in PagerDuty policy.\n" \
                         "\t\t\t 2. Port 443 is enabled.\n"

    msg_str = 'Create Page (attempt:{})'.format(e_context['retry_attempt'])

    error_msg_str =  msg_str +"\n" + error_settings_msg

    ROUTING_KEY = e_context['service_key']

    try:
        for index, notification in enumerate(e_context['notifications']):
            e_context['failure_text'] = "\t\t\tFailed in stage: create_page.\n"
            if e_context['retry_attempt'] > 1:
                mqtt_notify(e_context, 'Running', error_msg_str)
            else:
                mqtt_notify(e_context, 'Running', msg_str)
            
            ret = webserver.create_page(notification,ROUTING_KEY)
            if ret['error']:
                if "service_key" in notification:
                    del notification["service_key"]
                logger.error("Failed to create page for notification {}".format(notification))
                return False
    except Exception as excep:
        logger.error("Failed in  create page stage, {}".format(str(excep)))
        return False

    return True
