import datetime
import logging

from nvsm_dshm.thea import Client
from nvsm_dshm.equipment import catalog
from nvsm_dshm.policy.policy import filter_policy

logger = logging.getLogger(__name__)


def collate_mute_lst(policy):
    if not policy or 'policy_id' not in list(policy.keys()):
        return []

    classification = {'PCIeDevices': ['PCIe'],
                      'thermal': ['Fan', 'Pdb'],
                      'power': ['PowerSupplyUnit'],
                      'memory': ['Dimm'],
                      'processor': ['Cpu'],
                      'storage': ['Volume', 'Drive', 'StorageController'],
                      'gpu': ["Gpu"],
                      
                      'network': ["Network"],
                      'networkport': ["NetworkPort"],
                      'networkdevfunc': ["NetworkDevFunc"],
                      'networkadapter': ["NetworkAdapter"],
                      'nvlink': ["NVLink"],
                      'nvswitch': ["NVSwitch"]}

    if policy['policy_id'] not in list(classification.keys()):
        return []

    devices = classification[policy['policy_id']]

    lst = []
    for device in devices:
        if device in list(policy.keys()):
            lst += policy[device]['mute_notification']

    return lst


def fetch_mute_list(topic):
    uri = '/'.join(topic.split('/')[:-2]).replace("Alerts", "Policy")

    # Fetch User Policy
    usr_policy = {}
    client = Client({Client.URI: uri, Client.JSON_NAME: "Policy"})
    objs = client.get_objects()
    if objs is not None:
        usr_policy = objs[0]

    # Get and return the mute list
    return collate_mute_lst(usr_policy)


def alert_notifiable(alert_id):
    # Check if alert is notifiable
    alert_info = catalog.get(alert_id)
    if alert_info is not None and alert_info['notify_user'] == 'No':
        logger.info(
            "Alert {} is not notifiable, dropping this alert notification ".format(alert_id))
        return False

    return True


def get_soak_interval(alert_id):
    alert_info = catalog.get(alert_id)
    if alert_info is not None and 'notification_soak' in list(alert_info.keys()):
        return alert_info['notification_soak']

    return None


def device_notifiable(topic, device_id, system_serial=None):
    # Check notification is muted for device
    mute_list = fetch_mute_list(topic)
    mute_list = filter_policy(mute_list, system_serial)
    if device_id in mute_list:
        logger.info(
            "Device {} notification is muted, dropping this alert notificaton ".format(device_id))
        return False

    return True


def str_to_utc(date_time_str, precision=6):
    if precision >=10:
        raise Exception("Incorrect Timestamp")
    try:
        dt = datetime.datetime.strptime(date_time_str[:-precision],
                                    "%Y-%m-%dT%H:%M:%S.%f")
        if date_time_str[-6] == "+":
            dt -= datetime.timedelta( hours=int(date_time_str[-5:-3]), minutes=int(date_time_str[-2:]))
        elif date_time_str[-6] == "-":
            dt += datetime.timedelta(hours=int(date_time_str[-5:-3]), minutes=int(date_time_str[-2:]))
        return dt
    except ValueError:
        return str_to_utc(date_time_str, precision+1)


def get_firing_time(date_time_str, soak_interval):
    if not date_time_str:
        return 0

    l_seconds = 0
    l_minutes = 0

    if soak_interval[-1] == 'm' and soak_interval[:-1].isdigit():
        l_minutes = int(soak_interval[:-1])
    if soak_interval[-1] == 's' and soak_interval[:-1].isdigit():
        l_seconds = int(soak_interval[:-1])

    alert_origin_time = str_to_utc(date_time_str)
    firing_time = alert_origin_time + \
        datetime.timedelta(seconds=l_seconds, minutes=l_minutes)
    return firing_time
