import time
import json
import datetime
from threading import Lock
import logging


from nvsm_dshm.notifier.notifier_thread import NotifierTask
from nvsm_dshm.notifier.e_mail import EmailTask
from nvsm_dshm.notifier.call_home import CallHomeTask
from nvsm_dshm.notifier.pager_duty import PagerDutyTask
from nvsm_dshm.infra import config_settings
from nvsm_dshm.notifier.soak.helper import alert_notifiable, device_notifiable, get_soak_interval, get_firing_time

logger = logging.getLogger(__name__)


EXECUTION_INTERVAL = 30  # CallHomeTask's batch processing is triggered at most once every 10 minutes.


class SoakTask(NotifierTask):
    queue_size = 0  # Infinite queue size
    name = 'SoakTask'

    def __init__(self):
        super(SoakTask, self).__init__()
        self.soaking = {}
        self.notifiers = []

        # Create notifier threads
        self.notifiers.append(EmailTask())
        self.notifiers.append(
            CallHomeTask(config_settings().get('notifier', 'webserver'),
                         config_settings().get('notifier', 'stage'),
                         config_settings().get('notifier', 'version')))
                         #config_settings().get('notifier', 'jwt-file'),
        self.notifiers.append(
            PagerDutyTask(config_settings().get('notifier', 'webserver'),
                         config_settings().get('notifier', 'stage'),
                         config_settings().get('notifier', 'version')))
                         #config_settings().get('notifier', 'jwt-file'),


        # Launch notifier threads
        for n in self.notifiers:
            n.start()

    def put(self, msg):
        topic = msg.topic
        notification = json.loads(msg.payload)

        try:
            if not alert_notifiable(notification['alert_id']):
                return

            if not device_notifiable(topic, notification['component_id'], notification['system_serial']):
                return

            soak_interval = get_soak_interval(notification['alert_id'])
            if not soak_interval:
                for n in self.notifiers:
                    n.put(notification)
                logger.debug(
                    "Soak is not required, notification object is sent to the notifiers")
                return

            super(SoakTask, self).put(notification)
        except Exception as e:
            logger.debug(e, exc_info=True)

    def run(self):
        while True:
            # Process all the messages in the input queue
            while not self.queue.empty():
                item = self.queue.get()
                key = '|'.join([item['alert_id'],
                                item['component_id'],
                                item['event_time']])

                firing_time = None

                if item['severity'] == 'Cleared':
                    if key in list(self.soaking.keys()):
                        del self.soaking[key]
                        logger.info(
                            "Notification was cleared before soak timeout: {}".format(key))
                        continue
                    else:
                        logger.debug(
                            "Queueing clear notification for immediate processing: {}".format(key))
                        firing_time = datetime.datetime.utcnow()

                else:
                    soak_interval = get_soak_interval(item['alert_id'])
                    firing_time = get_firing_time(item['event_time'],
                                                  soak_interval)
                    logger.debug(
                        "Soaking notification {} for {} duration, it will be processed at {}".format(key, soak_interval, firing_time))

                self.soaking[key] = {'firing_time': firing_time,
                                     'notification': item}

            # Iterate through all the items in soaking dict and check if any of
            # them has to be fired
            empty_keys = []
            current_time = datetime.datetime.utcnow()
            logger.debug("Current time is: {}".format(current_time))
            for key in list(self.soaking.keys()):
                if self.soaking[key]['firing_time'] <= current_time:
                    logger.info("Firing notification: {}".format(key))
                    for n in self.notifiers:
                        n.put(self.soaking[key]['notification'])
                    self.soaking[key] = None
                    empty_keys.append(key)

            # Discard all the keys in soaking dict that are empty
            for key in empty_keys:
                logger.debug("Discarding soak entry for {}".format(key))
                del self.soaking[key]

            time.sleep(EXECUTION_INTERVAL)
