from functools import wraps
import functools
import time
import threading
import logging

logger = logging.getLogger(__name__)


# NOTE: If we want to re-write this, consider pytransitions or
# python-statemachine packages.

class State(object):

    def __init__(self, func, on_success, on_failure, tags=None):
        functools.update_wrapper(self, func)
        self.name = func.__name__
        self.on_failure = on_failure
        self.on_success = on_success
        self.handler = func
        # List of tags for the state
        self.tags = tags

    def __call__(self, *args, **kwargs):
        return self.handler(*args, **kwargs)


class StateMachine(object):
    name = 'Base'
    states = {}
    start_state = ''

    def __init__(self):
        self.e_context = {}
        self.e_context['retry_attempt'] = 0
        self.e_context['nvsm_state_machine_context'] = 1
        self.states['terminate'] = None
        self.skip = []
        self.result = 'Succeeded'
        self.cancel_event = threading.Event()

    @classmethod
    def add(cls, st=None, start_state=False):
        def dec_add(st):
            cls.states[st.name] = st
            if start_state:
                cls.start_state = st.name
            return st
        if st is None:
            return dec_add
        else:
            return dec_add(st)

    def exists_state(self, st_name):
        if st_name == 'terminate':
            return True

        if st_name not in list(self.states.keys()):
            logger.error(
                "State {} does not exist in StateMachine {}".format(st_name, self.name))
            raise Exception(
                "State {} does not exist in StateMachine {}".format(st_name, self.name))

        return True

    def skip_states(self, tags):
        self.skip = tags

    def stop(self):
        logger.info("Cancelling the state machine")
        self.cancel_event.set()

    def run(self):
        logger.debug("Start execution of StateMachine {}".format(self.name))
        if self.exists_state(self.start_state):
            curr_state = self.states[self.start_state]

        while curr_state:
            # Check if the statemachine execution was cancelled.
            if self.cancel_event.isSet():
                self.result = 'Cancelled'
                break

            logger.debug(
                "Executing state {}".format(
                    self.name +
                    '_' +
                    curr_state.name))

            if curr_state(self.e_context, cancel_event=self.cancel_event):
                logger.debug(
                    "Execution of state {} succeeded".format(self.name + '_' + curr_state.name))
                if self.exists_state(curr_state.on_success):
                    curr_state = self.states[curr_state.on_success]
            else:
                logger.debug(
                    "Execution of state {} failed".format(self.name + '_' + curr_state.name))

                # StateMachine's overall result is marked as failed only if a
                # state has failed and has entered into the terminate state.
                if curr_state.on_failure == 'terminate':
                    self.result = 'Failed'

                if self.exists_state(curr_state.on_failure):
                    curr_state = self.states[curr_state.on_failure]

            # Skip the state and execute the next state in the state-machine if
            # it is skipped.
            # This is not tested yet.
            # TODO check it later
            # if curr_state:
            #     skip_tags = set(curr_state.tags) & set(self.skip)
            #     if len(skip_tags):
            #         curr_state = self.states[curr_state.on_success]

        self.terminate()

    def terminate(self):
        if self.result == 'Succeeded':
            logger.debug(
                "Executing terminate state for the workflow {}, work flow execution was successfull".format(self.name))
        else:
            logger.debug(
                "Executing terminate state for the workflow {}, work flow execution was unsuccessfull".format(self.name))


def retry(retries=1, timeout=10, exponential=False):
    def dec_retry(func):
        @wraps(func)
        def retry_wrap(*args, **kwargs):

            cancel_event = None
            if 'cancel_event' in list(kwargs.keys()):
                cancel_event = kwargs['cancel_event']
                kwargs.pop('cancel_event', None)

            # If the statemachine is cancelled, then don't execute the blocking
            # state handler.
            if cancel_event and cancel_event.isSet():
                logger.info(
                    "Cancel event was set, aborting state handler execution")
                return False

            for arg in args:
                if isinstance(arg, dict):
                    if 'nvsm_state_machine_context' in arg.keys():
                        arg['retry_attempt'] = 1

            value = func(*args, **kwargs)
            if value:
                return value

            _try = 1
            while _try < retries:
                logger.debug(
                    "Retrying {} -> try ->{}".format(func.__name__, _try))

                # If the statemachine is cancelled, then don't execute the
                # blocking sleep call.
                if cancel_event and cancel_event.isSet():
                    logger.info("Cancel event was set, aborting retries")
                    return False

                e = 2 if exponential else 1
                interval = pow(e, _try - 1) * timeout
                time.sleep(interval)

                # If the statemachine is cancelled, then don't execute the
                # blocking state handler.
                if cancel_event and cancel_event.isSet():
                    logger.info("Cancel event was set, aborting retries")
                    return False

                for arg in args:
                    if isinstance(arg, dict):
                        if 'nvsm_state_machine_context' in arg.keys():
                            arg['retry_attempt'] = _try+1

                value = func(*args, **kwargs)
                if value:
                    return value
                _try += 1

            return value

        return retry_wrap

    return dec_retry


def state(on_success='', on_failure=''):
    def decorate_state(func):
        return State(func, on_success, on_failure)
    return decorate_state
