#!/usr/bin/python
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
pcie monitoring module

"""

import errno
import time
import sched
import json
import logging
import os
import signal
import sys
import syslog
import select
import traceback

from functools import partial
from nvsm_dshm.ds import AlertsDataStore
from nvsm_dshm.equipment import catalog
from nvsm_dshm.equipment import Equipment
from nvsm_dshm.infra import config_settings
from nvsm_dshm.infra import ErrCode as Err
from nvsm_dshm.infra import log
from nvsm_dshm.mqclient import MQBrokerClient as MQBroker
from nvsm_dshm.policy import IPolicy
from nvsm_dshm.policy import policy
from nvsm_dshm.thea import Client
from nvsm_dshm.thea import Endpoint
from nvsm_dshm.thea import Topic
from nvsm_dshm.infra.helper import init_pa

NVSM_CONN_RETRY_TIME = 5  # in seconds
NVSM_CLIENT_NAME = 'nvsm_pcie_client'

logger = logging.getLogger(__name__)


def sigterm_handler(mq_client, signum, frame):
    """
    Sig term handler.
    """
    logger.info("Executing graceful exit")
    mq_client.disconnect()


def init_configuration(func):
    def inner():
        try:
            config_settings(config_file_name=os.environ['DSHM_CONFIG_FILE'])
        except Exception as e:
            syslog.syslog(syslog.LOG_CRIT,
                          'pcie_mon: Unable to read the config file')
            sys.exit(1)
        func()
    return inner


def init_logger(func):
    """
        Initializes the log file.
    """
    def inner():
        # Init logger
        if not log.initialize(
                config_settings().get('logging', 'config-file'),
                config_settings().get('logging', 'log-dir'),
                'pcie'):
            syslog.syslog(syslog.LOG_CRIT, 'pcie_mon: Logger init failed')
            sys.exit(1)

        func()
    return inner


def push_defaults(func):
    # TODO move this to other module
    """
        Push defaults.
    """
    def inner():
        logger.debug('begin')

        endpoint = None
        nvsm_api_cfg_file = config_settings().get('nvsm-api', 'config-file')
        nvsm_api_jwt_file = config_settings().get('nvsm-api', 'jwt-file')

        (ca_cert_file,
         _,
         _,
         _,
         port,
         address,
         _) = config_settings.nvsm_api_config_helper(nvsm_api_cfg_file)

        # TODO change Endpoint arguments to kargs
        if port is not None:
            endpoint = Endpoint(port=str(port), jwt_file=nvsm_api_jwt_file,
                                ca_cert_file=ca_cert_file, ip_addr=address)
        else:
            endpoint = Endpoint()

        connected = False
        attempts = 0
        while True:
            if not endpoint.accessible():
                logger.warning('{} attempt, Retrying after {} sec'.format(
                    attempts, NVSM_CONN_RETRY_TIME))
                time.sleep(NVSM_CONN_RETRY_TIME)
                attempts = attempts + 1
            else:
                connected = True
                break

        if not connected:
            logger.critical('Connection to thea is failed')
            sys.exit(1)

        init_pa(__name__)

        catalog.init('pcie')
        policy.push_defaults(['PCIe'])

        logger.debug('end')

        func()
    return inner


def on_connect(client, userdata, flags, rc):
    logger.debug('MQ ON Connect Ret-Code:' + str(rc))

    if not rc:
        logger.debug('MQ Client Connect Success, result code: ' + str(rc))
        MQBroker.subscribe(client, Topic.TOPIC_PCIE)
    else:
        logger.error(
            'MQ Client Connect failure, with result code: ' + str(rc))


def on_message(client, userdata, msg):
    logger.debug('MQ Message Received for Topic:' + msg.topic)

    if msg is None or msg.payload is None:
        return
    process_topic(msg)

    logger.debug('MQ Message for Topic: ' + msg.topic + ' Handled')


def process_topic(msg):
    """
        process the event
    """
    logger.debug('process event for pcies')
    policy = IPolicy.factory('PCIe')
    logger.debug('Policy used {}'.format(policy))

    data_store = AlertsDataStore.factory('thea_db', 'PCIe')
    # data_store = AlertsDataStore.factory(u'shelve', './sys_data_store')

    topic = msg.topic
    payload = str(msg.payload.decode('utf-8', 'ignore'))
    json_obj = json.loads(payload)  # decode json data

    # FIXME hack
    # does mqtt provides suppress topics machanism?
    logger.info('topic {}'.format(topic))
    if topic == '/nvsm/v1/Systems/1/PCIeDevices':
        logger.info('ignoring topic')
        return

    if json_obj is None:
        logger.error('empty dict')
        return

    device = Equipment.factory('PCIe',
                               json_obj,
                               Endpoint().hostname(),
                               Endpoint().serial(),
                               uri=topic)

    # FIXME Hack
    if device.get_name() == 'bmcvga':
        logger.info('No monitoring for bmcvga')
        return

    if device is not None and policy.monitor(device.get_name()):
        logger.debug('Checking device {}:  {}'.format(
            device.get_name(), device))

        ret = data_store.fetch_alerts(device.get_name())
        if ret['errored']:
            logger.error('Unable to fetch alert data from PCIe from thea')
            logger.error(
                'Thea is in-accessible')
            return
        alerts = ret['alerts']

        if msg.retain:
            logger.info('Boottime topic handling')
            device.apply_boottime_rules(device, alerts)
        else:
            device.apply_rules(device, alerts)
        data_store.store_alerts(device.get_name(), alerts)


@init_configuration
@init_logger
@push_defaults
def pcie_main():
    """
    Main
    """
    mq_client = MQBroker.factory(MQBroker.MQTT_CLIENT, NVSM_CLIENT_NAME,
                                 on_connect, on_message)

    if mq_client is None or \
            mq_client is Err.ERR_INVLAID_MQ_SECURITY_CERT or \
            mq_client is Err.ERR_INVLAID_MQ_ADDR_PORT:
        logger.critical('Failed to create mqtt client')
        sys.exit(1)

    signal.signal(signal.SIGTERM, partial(sigterm_handler, mq_client))

    logger.debug('mqtt broker listen')

    try:
        mq_client.listen_for_ever()
    except (OSError, select.error) as e:
        logger.error('Select Error: {}'.format(e))
    except Exception as e:
        logger.error('Unhandled Exception: {}'.format(e))
        logger.debug(traceback.format_exc())

    logger.info('Exiting the main thread')
