# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Helper module

Contains the helper methods used by policy package.
"""
import os
import json
import jsonschema
import logging

from nvsm_dshm.infra import config_settings

logger = logging.getLogger(__name__)

DEFAULT_DATA_DIR = '/usr/share/dshm/policy/'
SCHEMA_FILE = 'policy_schema.json'


def copy_default(default_dict, stored_dict):
    """
    Copies the default device policy dictionary from the DSHM defaults to the
    DB. It recursively checks if any  DSHM defines any additional properties
    and adds them to the DB.
    """
    copied = False
    for key in list(default_dict.keys()):
        if key in list(stored_dict.keys()):
            if isinstance(default_dict[key], dict):
                copied = copied or copy_default(
                    default_dict[key], stored_dict[key])
        else:
            copied = True
            stored_dict[key] = default_dict[key]

    return copied


def read_json_file(file_name):
    """
    Read the json file and return the object.
    """
    contents = None
    data_dir = config_settings().get('data', 'data-dir')
    if data_dir is None:
        data_dir = DEFAULT_DATA_DIR
    abs_path_file_name = data_dir + file_name

    if os.path.isfile(abs_path_file_name):
        with open(abs_path_file_name, 'r') as fhandle:
            contents = fhandle.read()

    if contents is None:
        return None
    else:
        return json.loads(contents)


def schema_valid(in_policy):
    """
    Checks if the json schema of the policy is valid.
    """
    # Open Schema file
    # Assumes that schema file read does not fail,
    # this is a file used by DSHM and not user input file.
    schema = read_json_file(SCHEMA_FILE)

    # Validate input policy
    try:
        jsonschema.validate(in_policy, schema)
    except jsonschema.exceptions.ValidationError:
        logger.exception('JSON schema validation for policy failed')
        return False

    return True
