# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Policy File module

Contains the implementation for the PolicyFile class which reads the policy from a file.
"""
import json
import logging

from .policy import IPolicy
from nvsm_dshm.policy.policy import IPolicy
from nvsm_dshm.policy.helper import schema_valid
from nvsm_dshm.policy.helper import read_json_file
from nvsm_dshm.infra import config_settings

logger = logging.getLogger(__name__)

DEFAULT_DATA_DIR = '/usr/share/dshm/policy/'


# NOTE: This is not tested
class PolicyFile(IPolicy):

    """
    Fetches Policy from a file
    """

    def __init__(self, device_type):
        super(PolicyFile, self).__init__(device_type)

    @classmethod
    def source_type(cls):
        """
        Returns type
        """
        return "file"

    @classmethod
    def read_policy(cls, source, policy_type, default):
        """
        Read the file and return the object.
        """
        usr_policy = read_json_file(source)
        if usr_policy is None:
            logger.info("Unable to read the policy file {}".format(source))
            return {'usr_policy': default, 'usr_policy_whole': None, 'exists': False}

        if not schema_valid(usr_policy):
            logger.info("Policy json schema is invalid, returning default")
            return {'usr_policy': default, 'usr_policy_whole': None, 'exists': False}

        if policy_type in list(usr_policy.keys()):
            logger.debug("Valid policy found")
            return {'usr_policy': usr_policy[policy_type], 'usr_policy_whole': usr_policy, 'exists': True}
        else:
            logger.info(
                "Cannot find the specified policy type in user configured policy")
            return {'usr_policy': default, 'usr_policy_whole': usr_policy, 'exists': True}

    @classmethod
    def write_policy(self, destination, data, modify=False):
        """
        Write policy to the json file and return True if write is successful.
        """
        logger.debug("Writing {} to {}".format(data, destination))
        try:
            data_dir = config_settings().get('data', 'data-dir')
            if data_dir is None:
                data_dir = DEFAULT_DATA_DIR
            abs_path_file_name = data_dir + destination

            with open(abs_path_file_name, 'w') as fhandle:
                fhandle.write(json.dumps(data))
        except (OSError, IOError) as exc:
            logger.exception(
                'Writing json data ({}) to file {} failed: '.format(
                    data, destination))
            return False

        return True

    @classmethod
    def push_default(cls, device_type):
        super(PolicyFile, cls).push_default(device_type)
