# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Policy Uri module

Contains the implementation for the PolicyUri class which reads the policy from Thea.
"""
import logging

from .policy import IPolicy
from nvsm_dshm.policy.helper import schema_valid
from nvsm_dshm.thea import Client

logger = logging.getLogger(__name__)


class PolicyUri(IPolicy):

    """
    Fetches Policy from thea URI
    """

    def __init__(self, policy_type):
        super(PolicyUri, self).__init__(policy_type)

    @classmethod
    def read_policy(cls, source, policy_type, default):
        """
        Read the file and return the object.
        """
        usr_policy = None
        client = Client({Client.URI: source, Client.JSON_NAME: "Policy"})
        objs = client.get_objects()
        if objs is not None:
            usr_policy = objs[0]

        # FIXME: read_policy will return policy exists as 'False under two conditions:
        # 1) Thea is inaccessible
        # 2) First time install, policy is not created in the DB.
        #
        # Fix this better later.
        if usr_policy is None:
            logger.info("Unable to read the policy from {}".format(source))
            return {'usr_policy': default, 'usr_policy_whole': None, 'exists': False}

        # If schema is invalid, then assume that the policy is corrupted and
        # re-write the policy (exist is set to False)
        if not schema_valid(usr_policy):
            logger.info("Policy json schema is invalid, returning default")
            return {'usr_policy': default, 'usr_policy_whole': None, 'exists': False}

        if policy_type in list(usr_policy.keys()):
            logger.debug("Valid policy found")
            return {'usr_policy': usr_policy[policy_type], 'usr_policy_whole': usr_policy, 'exists': True}
        else:
            logger.info(
                "Cannot find the specified policy type in user configured policy")
            return {'usr_policy': default, 'usr_policy_whole': usr_policy, 'exists': True}

    @classmethod
    def write_policy(self, destination, data, modify=False):
        """
        Write policy to the json file and return True if write is successful.
        """
        logger.debug("Writing {} to {}".format(data, destination))
        if modify:
            return Client.put(destination, data)
        else:
            return Client.post(destination, data)

    @classmethod
    def source_type(cls):
        """
        Returns type
        """
        return "uri"

    @classmethod
    def push_default(cls, policy_type):
        super(PolicyUri, cls).push_default(policy_type)
