# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Policy module

Contains the definition for Policy class and its derivations.
"""
from abc import ABCMeta
import logging

import os, sys
sys.path.append(os.path.abspath(os.path.join(__file__, '..')))
from nvsm_dshm.policy.helper import copy_default
from nvsm_dshm.policy.helper import read_json_file

logger = logging.getLogger(__name__)

POLICY_INFO = 'policy_info.json'
DEFAULT_POLICY = 'policy_defaults.json'


def push_defaults(policy_types, cls_name='PolicyUri'):
    """
    It is invoked to store the default policy to the DB.
        a) If the policy is absent in the DB, it will create it.

        b) If the policy exists in the DB, it will only update the missing
            properties in the DB. This will be needed if DSHM defines
            additional properties in the next version .
    """
    if policy_types is None:
        return False

    for cls_type in IPolicy.__subclasses__():
        if cls_type.__name__ == cls_name:
            for policy_type in policy_types:
                cls_type.push_default(policy_type)
            return True

    raise AssertionError("Class undefined: " + cls_name)

def filter_policy(policy, chassis_serial):
    if not chassis_serial:
        return policy
    # filter policy based on chassis serial
    filtered_policy = []
    for p in policy:
        if '+' in p:
            if (p.split('+')[0].lower() != chassis_serial.lower()):
                continue
            filtered_policy.append('+'.join(p.split('+')[1:]))
        else:
            filtered_policy.append(p)
    return filtered_policy

class IPolicy(object, metaclass=ABCMeta):

    """
    Policy class:
        Defines an interface for the policy classes to be implemented by all
        subsystems.
    """

    def __init__(self, policy_type):
        """
        Constructor
        """
        self.usr_policy = {}
        self.type = policy_type

        source = read_json_file(POLICY_INFO)[self.type][self.source_type()]
        default_policy = read_json_file(DEFAULT_POLICY)[self.type]

        # Get user configured policy file
        ret = self.read_policy(source, policy_type, default_policy)
        self.usr_policy = ret['usr_policy']

    def __repr__(self):
        return '{}(usr_policy:{}, policy_type:{})'.format(
            self.__class__.__name__,
            self.usr_policy,
            self.type)

    @staticmethod
    def factory(policy_type, cls_name='PolicyUri'):
        """
        Factory for various types of policies
        """
        for cls_type in IPolicy.__subclasses__():
            if cls_type.__name__ == cls_name:
                return cls_type(policy_type)

        raise AssertionError("Class undefined: " + cls_name)

    @classmethod
    def push_default(cls, policy_type):
        """
        Method used to create/update the policy to the DB
        This is a class method that must be defined by the derived classes

        1) Writes the default policy to data store(file/uri)  if it isn't
            already present.
        2) If the policy is present in the data store(file/uri), it updates
            the missing properties.
        """
        src_type = cls.source_type()
        source = read_json_file(POLICY_INFO)[policy_type][src_type]
        policy_id = read_json_file(POLICY_INFO)[policy_type]["policy_id"]
        default_policy = read_json_file(DEFAULT_POLICY)[policy_type]

        # Get usr configured policy
        policy_exists = True
        ret = cls.read_policy(source, policy_type, None)
        usr_policy = {} if not ret['usr_policy'] else ret['usr_policy']
        usr_policy_whole = {} if not ret[
            'usr_policy_whole'] else ret['usr_policy_whole']
        policy_exists = ret['exists']

        # Copy the properties that are in the default policy but are missing in
        # the input policy
        if copy_default(default_policy, usr_policy):
            wr_policy = usr_policy_whole
            wr_policy[policy_type] = usr_policy
            if src_type == "uri":
                wr_policy['policy_id'] = policy_id

            ret = cls.write_policy(source, wr_policy, policy_exists)
            if not ret:
                logger.error(
                    'Failed to write back to file ({})'.format(wr_policy))
            else:
                logger.info('({}) was written to the file {}'.format(
                    wr_policy, source))

    def get(self):
        """
        Returns the user configured policy
        """
        return self.usr_policy

    def get_val(self, prop):
        """
        Returns the specified property in the policy
        """
        if prop in list(self.usr_policy.keys()):
            return self.usr_policy[prop]

        return None

    def monitor(self, device_name, chassis_serial=None):
        """
        Returns True if the device_name is not present in the mute_monitoring
        list for device policies.

        If called for global policy it simply returns True.
        """
        if 'mute_monitoring' in list(self.usr_policy.keys()) and \
           self.usr_policy['mute_monitoring'] is not None:
            policy = [d.lower() for d in self.usr_policy['mute_monitoring']]
            policy = filter_policy(policy, chassis_serial)
            if device_name.lower() in policy:
                logger.error("Monitoring muted for device {}".format(device_name))
                return False

        return True

    def notify(self, device_name, chassis_serial=None):
        """
        Returns True if the device_name is not present in the mute_notification
        list for device policies.

        If called for global policy it simply returns True.
        """
        if 'mute_notification' in list(self.usr_policy.keys()) and \
           self.usr_policy['mute_notification'] is not None:
            policy = [d.lower() for d in self.usr_policy['mute_notification']]
            policy = filter_policy(policy, chassis_serial)
            if device_name.lower() in policy:
                logger.debug(
                    "Notification muted for device {}".format(device_name))
                return False
        return True

    @classmethod
    def read_policy(cls, source, policy_type, default):
        """
        Subclasses must implement this as staticmethod

        Used to read the policy
        If read successful: The read policy is returned.
        If read fails: The default policy is returned.
        """
        pass

    @classmethod
    def write_policy(cls, destination, data, modify=False):
        """
        Subclasses must implement this as staticmethod

        Used to write the policy
        Returns False if write fails.
        Else returns True.
        """
        pass

    @classmethod
    def source_type(cls):
        """
        Subclasses should implement this
        """
        return 'Base'
