# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Storage module

Contains the definition for Equipment class and its derivations.
"""

import json
from abc import ABCMeta, abstractmethod
import logging
import time
from nvsm_dshm.equipment.rule import IRule
# from nvsm_dshm.infra import pa
from nvsm_dshm.equipment.container import VolumeContainer
# from nvsm_dshm.equipment.container import DriveContainer
from nvsm_dshm.equipment.container import LedStatus
from nvsm_dshm.thea import Endpoint
from nvsm_dshm.thea import Client
from nvsm_dshm.ds import AlertsDataStore
from nvsm_dshm.infra.helper import pa

logger = logging.getLogger(__name__)

HEALTH_OK = "OK"
HEALTH_WARN = "Warning"
HEALTH_CRIT = "Critical"
HEALTH_UNKN = "H_Unknown"

STATE_ENABLED = "Enabled"
STATE_UPDATING = "Updating"
STATE_STANDBYOFFLINE = "StandByOffline"
STATE_UNKNOWN = "S_Unknown"

alert_led = {
    'NV-DRIVE-01': "IGNORE",
    'NV-DRIVE-02': "PREDFAIL",
    'NV-DRIVE-03': "PREDFAIL",
    'NV-DRIVE-04': "FAIL",
    'NV-DRIVE-05': "PREDFAIL",
    'NV-DRIVE-06': "PREDFAIL",
    'NV-DRIVE-07': "IGNORE",
    'NV-DRIVE-08': "FAIL"
}


class Storage(object, metaclass=ABCMeta):
    """
    Storage class:
        Encapsulates the following
        - Storage element identity (entityId, name)

    TODO:
        storage element identity currently is tracked using entity id and name.
        This should evolve later.

        Storage health might have to be abstracted some more later on.
    """

    def __init__(self, id, name, hostname, serial, status, present=True):
        """
        Constructor
        """
        self.entity_id = id
        self.name = name
        self.hostname = hostname
        self.serial = serial
        self.status = status
        self.prescence = present

    def __repr__(self):
        return 'entity_id:{},name:{}'.format(self.entity_id, self.name)

    @staticmethod
    def factory(storage_element,
                json_obj,
                hostname,
                serial,
                supported_config=None):
        """
        Factory for various types of equipment supported by DSHM
        """
        if storage_element == "Volume":
            logger.debug('Factory create volume')
            return Volume.create(json_obj, hostname, serial)

        if storage_element == 'Drive':
            return Drive.create(json_obj, hostname, serial, supported_config)

        if storage_element == 'DriveCollection':
            return DriveCollection.create(json_obj, hostname, serial,
                                          supported_config)

        if storage_element == 'StorageController':
            return StorageController.create(json_obj, hostname, serial)

        raise AssertionError('Storage element undefined: ' + storage_element)

    @abstractmethod
    def create(self, json_obj, hostname):
        """
        Creates the storage element class from the json object.
        """
        pass

    @abstractmethod
    def get_name(self):
        """
        Returns the device name.
        """
        return self.name

    @abstractmethod
    def config(self, config):
        """
        Additional configuration applied to the Storage object,
        can be used to change the behaviour of the rules applied to the object

        Like setting thresholds
        """
        pass

    @classmethod
    def get_rules(cls):
        """
        Returns the rules defined for this storage element
        """
        pass

    @classmethod
    def apply_rules(cls, storage_element, alerts):
        """
        Applies the rules defined for the storage element.
        """
        rules = cls.get_rules()
        for rule in rules:
            rule.apply(storage_element, alerts)


# Rules for Volume defined here.
class VolumeCorrupted(IRule):
    """
    1) Check Volume(if RAID-1) Health status
    2) Check if the expected drive count is matching
    """

    def __init__(self):
        self.alert_id = 'NV-VOL-01'

    @classmethod
    def condition(cls, volume):
        # define condtition here
        if volume.prescence is not True:
            logger.info("Volume not found {}".format(volume.name))
            return True

        if volume.volume_type is None:
            return False

        if volume.status != HEALTH_OK:
            return True
        # Drive count validation
        if int(volume.drive_count) < int(volume.exp_drive_count):
            logger.info('Drive count Expected {} Current {}'.format(
                volume.exp_drive_count, volume.drive_count))
            return True
        else:
            return False
        return False

    @classmethod
    def msg(cls, volume):
        """
            Returns the details for msg
        """
        msg_list = [volume.volume_type]
        return msg_list

# Rules for Volume defined here.


class VolumeSEDConfig(IRule):
    """
    1) Check Volume(if RAID-1) has all the U.2 drive available
    """

    def __init__(self):
        self.alert_id = 'NV-VOL-02'

    @classmethod
    def condition(cls, volume):
        # define condtition here
        if volume.prescence is not True:
            return False
        else:
            if volume.encrypted is True:
                if volume.data_drive_count != volume.drive_count:
                    return True
        return False

    @classmethod
    def msg(cls, volume):
        """
            Returns the details for msg
        """
        msg_list = [volume.name]
        return msg_list


# Volume Class defined here
class Volume(Storage):
    """
    Volume: Extends Storage Class.
    """

    _required_json_attribs = set(('Id', 'Name', 'Status'))

    __rules = [VolumeCorrupted(), VolumeSEDConfig()]

    def __init__(self, entity_id, name, hostname, serial,
                 status, drv_count, vol_type, exp_drv_count,
                 u2_drive_count=0, present=True, encrypted=False):
        super(Volume, self).__init__(entity_id, name, hostname, serial,
                                     status, present)
        self.drive_count = drv_count
        self.exp_drive_count = exp_drv_count
        self.volume_type = vol_type
        self.encrypted = encrypted
        self.data_drive_count = u2_drive_count

    @classmethod
    def get_rules(cls):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def get_name(self):
        return super(Volume, self).get_name()

    def get_id(self):
        return super(Volume, self).get_id()

    def get_status(self):
        return super(Volume, self).get_status()

    def get_volume_type(self):
        return self._volume_type

    def get_drive_count(self):
        return self._drive_count

    def config(self, config):
        return super(Volume, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial):
        ret = None
        status = 'Ok'
        if json_obj["IsPresent"] == 'No':
            logger.debug('Volume missing {}'.format(json_obj['Name']))
            ret = cls(json_obj['Name'], json_obj['Name'], hostname,
                      serial, None, None, json_obj['VolumeType'], False)
            # Remove the volume from monitoring until it comes back
            VolumeContainer.monitored_volumes[:] = \
                [d for d in list(VolumeContainer.monitored_volumes)
                 if d.get('Name') != json_obj['Name']]
            return ret

        if cls._required_json_attribs.issubset(list(json_obj.keys())) and \
           'Health' in list(json_obj['Status'].keys()):
            logger.debug('Creating Volume: %s', json.dumps(json_obj))
            status = json_obj['Status']['Health']
            drive_count = len(json_obj['Links']['Oem']['Drives']['@odata.id'])
            volume_type = json_obj['VolumeType']
            expct_drv_count = 0
            found = False
            if VolumeContainer.monitored_volumes is not None:
                for volume in list(VolumeContainer.monitored_volumes):
                    if volume['Name'] == json_obj['Name']:
                        expct_drv_count = int(volume['DriveCount'])
                        found = True

            if found is False:
                new_vol = {'Name': json_obj['Name'],
                           'DriveCount': drive_count,
                           'VolumeType': volume_type}
                # Get from predefined list
                expct_drv_count = drive_count
                if VolumeContainer.monitored_volumes is not None:
                    VolumeContainer.monitored_volumes.append(new_vol)

            logger.debug('VolID:{} VolName:{} status:{} DrvCount:{} '
                         'VolType:{} ExpDrv:{}'.format(
                             json_obj['Id'], json_obj['Name'], status,
                             drive_count, volume_type, expct_drv_count))
            # Fetch all the U.2 drives
            try:
                storage_uri = Endpoint.storage_uri
                client = Client({Client.URI: storage_uri})
                storage_data = client.get_objects()
                u2_drive_count = storage_data['Drives@odata.count'] - 2
            except:
                u2_drive_count = 0
            ret = cls(json_obj['Id'], json_obj['Name'], hostname, serial,
                      status, drive_count, volume_type, expct_drv_count,
                      u2_drive_count, encrypted=json_obj['Encrypted'])
        else:
            logger.error('Volume object creation failed, '
                         'json object incomplete %s',
                         json.dumps(json_obj))
        return ret

    def __repr__(self):
        return '{}({}, health:{})'.format(self.__class__.__name__,
                                          super(Volume, self).__repr__(),
                                          self.health)


def u2drive_led(led_cmd, drive_id, slot_name, alert_id, led_support):
    # Ignore for M.2 slot
    if slot_name == "SlotM.2_1" or slot_name == "SlotM.2_0":
        return
    if led_support is False:
        return

    if slot_name in list(DriveSlot.get_led_map().keys()):
        led_status = DriveSlot.get_led_status(slot_name)
    else:
        return
    # No change in status
    if led_status == led_cmd:
        return

    logger.debug("LED Status: {}".format(led_status))
    data_store = AlertsDataStore.factory('thea_db', 'Drive')
    ret = data_store.fetch_alerts(slot_name)
    alerts = ret['alerts']
    led_change = True
    # FAIL gets maximum priority
    if led_cmd == LedStatus.FAIL:
        led_change = True
    # If predictive failure and current status is greater
    if led_cmd == LedStatus.PREDFAIL:
        if led_status.value == 4:
            led_change = False
        else:
            led_change = True
    # If drive LED to be turned off
    if led_cmd == LedStatus.NORMAL:
        # Verify if other alerts present
        count = 0
        led_next = "IGNORE"
        for alert in alerts:
            # Take the next highest priority
            if alert != alert_id:
                count = count+1
                alert_level = alert_led[alert]
                if alert_level == "FAIL":
                    led_next = "FAIL"
                    break
                else:
                    if alert_level == "PREDFAIL" and led_next != "FAIL":
                        led_next = "PREDFAIL"

        # No outstanding alerts
        if count == 0 or led_next == "IGNORE":
            led_change = True
            led_cmd = LedStatus.NORMAL
        # Alerts present of FAIL
        elif led_next == "FAIL":
            if led_status == LedStatus.FAIL:
                led_change = False
            led_cmd = LedStatus.FAIL
        # Alerts present of PREDFAIL
        elif led_next == "PREDFAIL":
            if led_status == LedStatus.PREDFAIL:
                led_change = False
            led_cmd = LedStatus.PREDFAIL

    if led_change is False:
        return

    drive_led_uri = Endpoint.storage_drive_led_uri.format(drive_id)
    client = Client({Client.URI: drive_led_uri})

    put_data = '{{\"StatusCode\":\"{}\"}}'.format(led_cmd.value)
    json_data = json.loads(put_data)
    retcode = client.put(drive_led_uri, json_data)
    if retcode is not True:
        logger.info("LED Drive status change failed")
    else:
        logger.info('LED Drive {} status changed to '
                    '{}'.format(drive_id, led_cmd.value))
        DriveSlot.set_led_status(slot_name, led_cmd)
        led_status = led_cmd
    return


class MediaErrors(IRule):
    def __init__(self):
        self.alert_id = 'NV-DRIVE-02'

    @classmethod
    def condition(cls, drive):
        if drive.drive_model == 'KCM5DRUG3T84':
            return False

        if drive.media_errors:
            u2drive_led(LedStatus.PREDFAIL, drive.drive_name,
                        drive.name, 'NV-DRIVE-02', drive.led_support)
            return True
        u2drive_led(LedStatus.NORMAL, drive.drive_name,
                    drive.name, 'NV-DRIVE-02', drive.led_support)
        return False


class IOErrors(IRule):
    def __init__(self):
        self.alert_id = 'NV-DRIVE-03'

    @classmethod
    def condition(cls, drive):
        if drive.io_errors:
            u2drive_led(LedStatus.PREDFAIL, drive.drive_name,
                        drive.name, 'NV-DRIVE-03', drive.led_support)
            return True
        u2drive_led(LedStatus.NORMAL, drive.drive_name,
                    drive.name, 'NV-DRIVE-03', drive.led_support)
        return False


class ControllerFailure(IRule):
    def __init__(self):
        self.alert_id = 'NV-DRIVE-04'

    @classmethod
    def condition(cls, drive):
        if drive.ctrl_failure:
            u2drive_led(LedStatus.FAIL, drive.drive_name,
                        drive.name, 'NV-DRIVE-04', drive.led_support)
            return True
        u2drive_led(LedStatus.NORMAL, drive.drive_name,
                    drive.name, 'NV-DRIVE-04', drive.led_support)
        return False


class AvailableCapacity(IRule):
    def __init__(self):
        self.alert_id = 'NV-DRIVE-05'

    @classmethod
    def condition(cls, drive):
        if drive.avlbl_cpcty < 10:
            u2drive_led(LedStatus.PREDFAIL, drive.drive_name,
                        drive.name, 'NV-DRIVE-05', drive.led_support)
            return True
        u2drive_led(LedStatus.NORMAL, drive.drive_name,
                    drive.name, 'NV-DRIVE-05', drive.led_support)
        return False


class DriveNotOperational(IRule):
    def __init__(self):
        self.alert_id = 'NV-DRIVE-08'

    @classmethod
    def condition(cls, drive):
        if drive.state != 'Enabled':
            u2drive_led(LedStatus.FAIL, drive.drive_name,
                        drive.name, 'NV-DRIVE-08', drive.led_support)
            return True
        u2drive_led(LedStatus.NORMAL, drive.drive_name,
                    drive.name, 'NV-DRIVE-08', drive.led_support)
        return False

    @classmethod
    def msg(cls, drive):
        """
            Returns the details for msg
        """
        msg_list = [drive.state]
        return msg_list


class DriveUsage(IRule):
    def __init__(self):
        self.alert_id = 'NV-DRIVE-06'

    @classmethod
    def condition(cls, drive):
        if drive.percent_used > 90:
            u2drive_led(LedStatus.PREDFAIL, drive.drive_name,
                        drive.name, 'NV-DRIVE-06', drive.led_support)
            return True
        u2drive_led(LedStatus.NORMAL, drive.drive_name,
                    drive.name, 'NV-DRIVE-06', drive.led_support)
        return False


class CriticalWarning(IRule):
    def __init__(self):
        self.alert_id = 'NV-DRIVE-12'

    @classmethod
    def condition(cls, drive):
        if drive.critical_wrng:
            u2drive_led(LedStatus.PREDFAIL, drive.drive_name,
                        drive.name, 'NV-DRIVE-12', drive.led_support)
            return True
        u2drive_led(LedStatus.NORMAL, drive.drive_name,
                    drive.name, 'NV-DRIVE-12', drive.led_support)
        return False


class UnsupportedDrive(IRule):
    def __init__(self):
        self.alert_id = 'NV-DRIVE-07'

    @classmethod
    def condition(cls, drive):
        logger.debug('Supported Drive List {}'.format(drive.supported_drives))
        if drive.supported_drives is None:
            return False
        if drive.drive_model not in drive.supported_drives:
            return True
        return False


# DriveSlot map
class DriveSlot:
    __led_map = dict()

    def __init__(self, led_map=None):
        self.__led_map = led_map

    @classmethod
    def set_led_map(cls, led_map):
        cls.__led_map = led_map

    @classmethod
    def get_led_map(cls):
        return cls.__led_map

    @classmethod
    def set_led_status(cls, slot_name, status):
        cls.__led_map[slot_name] = status

    @classmethod
    def get_led_status(cls, slot_name):
        if cls.__led_map is None:
            return None
        if slot_name in cls.__led_map.keys():
            return cls.__led_map[slot_name]
        else:
            return None


# Drive Class defined here
class Drive(Storage):
    """
    Drive: Extends Storage Class.
    """

    _required_json_attribs = set(('Errors', 'Metrics'))
    __rules = [
        MediaErrors(),
        IOErrors(),
        ControllerFailure(),
        AvailableCapacity(),
        DriveUsage(),
        UnsupportedDrive(),
        DriveNotOperational(),
        CriticalWarning()
    ]

    def __init__(self, entity_id, name, drv_name, hostname, serial, state,
                 health, model, hm_errors, metrics, supported_drives):
        super(Drive, self).__init__(entity_id, name, hostname, serial, health)
        self.media_errors = hm_errors['Media']['Count']
        self.io_errors = hm_errors['IO']['Count']
        self.ctrl_failure = hm_errors['IO']['ControllerFailure']
        self.critical_wrng = hm_errors['CriticalWarning']['Count']
        self.state = state
        self.avlbl_cpcty = 100
        self.percent_used = 0
        self.supported_drives = supported_drives
        if 'AvailableSpare' in list(metrics.keys()):
            if metrics['AvailableSpare'] != 0:
                self.avlbl_cpcty = metrics['AvailableSpare']
        if 'WearLevelingCount' in list(metrics.keys()):
            if metrics['WearLevelingCount'] != "":
                self.avlbl_cpcty = int(metrics['WearLevelingCount'])
        if 'MediaWearOutIndicator' in list(metrics.keys()):
            if metrics['MediaWearOutIndicator'] != "":
                self.avlbl_cpcty = int(metrics['MediaWearOutIndicator'])
        if 'PercentUsed' in list(metrics.keys()):
            self.percent_used = metrics['PercentUsed']
        self.drive_model = model
        self.drive_name = drv_name
        self.led_support = False
        # TBD Check if platform definition file supports LED status
        # Till then we keep this
        platform_flags = pa().enumerate_obj(
            'SystemCollection/1/StorageCollection/1/PlatformFlags')
        if 'LedStatus' not in list(platform_flags.keys()):
            return
        if platform_flags['LedStatus'] == True:
            if name == "SlotM.2_1" or name == "SlotM.2_0":
                return
            self.led_support = True
            drive_led_uri = Endpoint.storage_drive_led_uri.format(drv_name)
            client = Client({Client.URI: drive_led_uri,
                             Client.ALL: True,
                             Client.IS_REFERENCE: False})
            time.sleep(2)
            resp = client.get_objects()
            if resp is not None:
                DriveSlot.set_led_status(name, LedStatus(int(resp['Status'])))
            logger.debug('Drive LED Map {}'.format(DriveSlot.get_led_map()))

    @classmethod
    def get_rules(cls):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def get_name(self):
        return super(Drive, self).get_name()

    def config(self, config):
        return super(Drive, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial, supported_config):
        ret = None
        state = 'Ok'
        if 'Model' in list(json_obj.keys()) and \
           'Oem' in list(json_obj.keys()) and \
           'Health' in list(json_obj['Status'].keys()) and \
           'Nvidia_HM' in list(json_obj['Oem'].keys()) and \
           'Id' in list(json_obj.keys()) and \
           'PhysicalLocation' in list(json_obj.keys()) and \
            cls._required_json_attribs.issubset(
                list(json_obj['Oem']['Nvidia_HM'].keys())):

            logger.debug('Creating Drives: %s', json.dumps(json_obj))
            health = json_obj['Status']['Health']
            state = json_obj['Status']['State']
            model = json_obj['Model']
            hm_errors = json_obj['Oem']['Nvidia_HM']['Errors']
            hm_metric = json_obj['Oem']['Nvidia_HM']['Metrics']
            name = 'Slot-' + json_obj['PhysicalLocation']['Info']
            drive_name = json_obj['Name']
            if supported_config is not None and \
                    'SupportedDrives' in supported_config.keys():
                supported_drives = supported_config['SupportedDrives']
            else:
                supported_drives = None
            ret = cls(json_obj['Id'], name, drive_name, hostname, serial,
                      state, health, model, hm_errors, hm_metric,
                      supported_drives)
        else:
            logger.error('Drive object creation failed, '
                         'json object incomplete %s', json.dumps(json_obj))
        return ret

    def __repr__(self):
        return '{}({}, health:{})'.format(self.__class__.__name__,
                                          super(Drive, self).__repr__(),
                                          self.health)


class DriveMisconfiguration(IRule):
    """
    Checks if drives are present in expected slots
    """

    def __init__(self):
        self.alert_id = 'NV-DRIVE-01'

    @classmethod
    def condition(cls, driveslot_collection):
        if driveslot_collection.supported_topologies is None:
            return False
        for topology in driveslot_collection.supported_topologies:
            if set(topology) == set(driveslot_collection.actual_topology):
                return False

        return True

    @classmethod
    def msg(cls, driveslot_collection):
        """
           Returns the details for msg
        """
        fmt_supported_topology = ''
        for topology in driveslot_collection.supported_topologies:
            fs = "\n\t\t\t{}".format(sorted(topology))
            fmt_supported_topology += fs

        return [sorted(driveslot_collection.actual_topology), fmt_supported_topology]


class DriveSEDMisconfig(IRule):
    def __init__(self):
        self.alert_id = 'NV-DRIVE-09'

    @classmethod
    def condition(cls, sed_info):
        if len(sed_info.sed_capable) > 0:
            if(all(x in sed_info.sed_capable for x in sed_info.data_drives)):
                return False
            else:
                return True
        return False


class DriveEncryptError(IRule):
    def __init__(self):
        self.alert_id = 'NV-DRIVE-10'

    @classmethod
    def condition(cls, sed_info):
        if len(sed_info.encrypted_drives) > 0:
            if(all(x in sed_info.encrypted_drives for x in sed_info.data_drives)):
                return False
            else:
                return True
        return False


class DriveM2FwVersionCheck(IRule):
    def __init__(self):
        self.alert_id = 'NV-DRIVE-11'

    @classmethod
    def condition(cls, drivecollection):
        fw_version = drivecollection.m2_drive_fw_version
        if len(fw_version) > 0:
            def check_dict_value_equal(x):
                if len(x) == 1:
                    return True
                idx = iter(x.values())
                val = next(idx)
                try:
                    next(j for j in idx if j != val)
                    return False
                except StopIteration:
                    return True
            if check_dict_value_equal(fw_version):
                return False
            else:
                return True
        return False

    @classmethod
    def msg(cls, drivecollection):
        """
           Returns the details for msg
        """
        msg_string = []
        fw_version = drivecollection.m2_drive_fw_version
        for value in fw_version.values():
            msg_string.append(value)
        logger.debug('{}'.format(msg_string))
        return [msg_string]


class DriveCollection(Storage):
    """
    DriveCollection: Extends Storage Class.
    """
    __rules = [DriveMisconfiguration(),
               DriveSEDMisconfig(),
               DriveEncryptError(),
               DriveM2FwVersionCheck()]

    @classmethod
    def get_rules(cls, equipment=None):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def __init__(self, entity_id, name, hostname, serial,
                 actual_topology, supported_topologies,
                 sed_capable, encrypted_drives, data_drives,
                 m2_drive_fw_version):
        """
        Init
        """
        super(DriveCollection, self).__init__(entity_id, name,
                                              hostname, serial,
                                              actual_topology,
                                              supported_topologies)

        self.actual_topology = actual_topology
        self.supported_topologies = supported_topologies
        self.sed_capable = sed_capable
        self.encrypted_drives = encrypted_drives
        self.data_drives = data_drives
        logger.debug("M2 Drive FW version:{}".format(m2_drive_fw_version))
        self.m2_drive_fw_version = m2_drive_fw_version

    def get_name(self):
        return super(DriveCollection, self).get_name()

    def config(self, config):
        return super(DriveCollection, self).config(config)

    @classmethod
    def create(cls, json_objs, hostname, serial, supported_config):
        ret = None
        if supported_config is not None and \
           'ExpectedSlots' in supported_config.keys():
            supported_topologies = supported_config['ExpectedSlots']
            actual_topology = []
            sed_capable = []
            encrypted_drives = []
            data_drives = []
            m2_drive_fw = dict()
            for json_obj in json_objs:
                if 'Info' not in json_obj['PhysicalLocation'].keys():
                    continue
                drive_slot = json_obj['PhysicalLocation']['Info']
                actual_topology.append(drive_slot)
                # Encryption capability
                if 'EncryptionAbility' in json_obj.keys():
                    if json_obj['EncryptionAbility'] == "SelfEncryptingDrive":
                        sed_capable.append(drive_slot)
                    if json_obj['EncryptionStatus'] == "Locked" or \
                            json_obj['EncryptionStatus'] == "Foreign":
                        encrypted_drives.append(drive_slot)
                # Exclude M.2 Drives
                if (drive_slot.find("M.2") == -1):
                    data_drives.append(drive_slot)
                else:
                    # M.2 Drive fw versions
                    if 'Revision' in json_obj.keys():
                        m2_drive_fw[drive_slot] = json_obj['Revision']
                    else:
                        m2_drive_fw[drive_slot] = 'Unknown'
            ret = cls('StorageDrive',
                      'StorageDrive',
                      hostname,
                      serial,
                      actual_topology,
                      supported_topologies,
                      sed_capable,
                      encrypted_drives,
                      data_drives,
                      m2_drive_fw)
        else:
            logger.error('DriveCollection object creation failed, '
                         ' json object incomplete %s', json.dumps(json_obj))
        return ret


class ControllerError(IRule):
    """
    Checks for controller errors
    """

    def __init__(self):
        self.alert_id = 'NV-CONTROLLER-01'

    @classmethod
    def condition(cls, controller):
        if controller.status != HEALTH_OK:
            return True
        return False


class ControllerConnectError(IRule):
    """
    Checks for controller errors
    """

    def __init__(self):
        self.alert_id = 'NV-CONTROLLER-02'

    @classmethod
    def condition(cls, controller):
        if controller.ports is None:
            return False
        for port in controller.ports:
            if port == "Fail":
                return True
        return False


class ControllerLinkConfigError(IRule):
    """
    Checks for controller errors
    """

    def __init__(self):
        self.alert_id = 'NV-CONTROLLER-03'

    @classmethod
    def condition(cls, controller):
        if controller.link_speed is None:
            return False
        return controller.link_speed


class StorageController(Storage):
    """
    Drive: Extends Storage Class.
    """

    __rules = [
        ControllerError(),
        ControllerConnectError(),
        ControllerLinkConfigError()
    ]

    def __init__(self, entity_id, name, hostname, serial, status,
                 ports, link_speed_config_error):
        super(StorageController, self).__init__(entity_id, name, hostname,
                                                serial, status)
        self.ports = ports
        self.link_speed = link_speed_config_error

    @classmethod
    def get_rules(cls):
        """
        Returns a list of rules supported for this device type
        """
        return cls.__rules

    def get_name(self):
        return super(StorageController, self).get_name()

    def config(self, config):
        return super(StorageController, self).config(config)

    @classmethod
    def create(cls, json_obj, hostname, serial):
        status = HEALTH_OK
        if 'Status' in list(json_obj.keys()):
            if 'Health' in json_obj['Status']:
                status = json_obj['Status']['Health']
        name = "c" + json_obj['MemberId']
        link_speed_config_err = False
        ports = ['Enabled', 'Enabled']
        if 'Oem' in list(json_obj.keys()):
            if 'Nvidia' in list(json_obj['Oem'].keys()):
                if 'Phys' in list(json_obj['Oem']['Nvidia'].keys()):
                    phys = json_obj['Oem']['Nvidia']['Phys']
                    for phy in phys:
                        if phy['InvldDwrdCnt'] > 0 or \
                           phy['DisPrtErrCnt'] > 0 or \
                           phy['LsDwrdSynCn'] > 0 or \
                           phy['PhyRst_PrblCn'] > 0:
                            ports[phy['Port']] = "Fail"
                        if "no limit" not in phy['LinkSpeed'].lower():
                            link_speed_config_err = True

        ret = cls(json_obj['MemberId'], name, hostname, serial, status,
                  ports, link_speed_config_err)
        return ret

    def __repr__(self):
        return '{}({}, health:{})'.format(self.__class__.__name__,
                                          super(StorageController,
                                                self).__repr__(),
                                          self.health)
