#!/usr/bin/python
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Storage monitoring module

This module monitors Storage - Volumes and drives.
"""
import time
import sys
import signal
import logging
import logging.config
import traceback
import select
import syslog
import os
import re

from functools import partial
from nvsm_dshm.thea import Client
from nvsm_dshm.thea import Endpoint
from nvsm_dshm.equipment.alerts import catalog
from nvsm_dshm.storage import Storage
from nvsm_dshm.ds import AlertsDataStore
from nvsm_dshm.policy import IPolicy
from nvsm_dshm.policy import policy
from nvsm_dshm.infra import config_settings
from nvsm_dshm.equipment.container import EquipmentContainer
from nvsm_dshm.equipment.container import VolumeContainer
from nvsm_dshm.infra.helper import init_pa
from nvsm_dshm.infra.helper import pa

from nvsm_dshm.infra import log

from nvsm_dshm.mqclient import MQBrokerClient as MQBroker
from nvsm_dshm.thea import Topic
from nvsm_dshm.infra import ErrCode as Err

NVSM_CONN_RETRY_TIME = 5  # in seconds
NVSM_RETRY_ATTEMPT_ON_MESSAGE = 5
NVSM_CLIENT_NAME = "nvsm_storage_client"


STORAGE_POLL_PRIORITY = 2
CONNECT_PRIORITY = 1
STORAGE_INIT_RETRY_TIME = 30

logger = logging.getLogger(__name__)


def init_configuration(func):
    def inner():
        try:
            config_settings(config_file_name=os.environ['DSHM_CONFIG_FILE'])
        except Exception as e:
            syslog.syslog(syslog.LOG_CRIT,'storage_mon: Unable to read the config file')
            sys.exit(1)
        func()
    return inner


# Initialize Logging
def init_logger(func):
    """
        Initializes the log file.
    """
    def inner():
        # Init logger
        if not log.initialize(
                config_settings().get('logging', 'config-file'),
                config_settings().get('logging', 'log-dir'), 'storage'):
            syslog.syslog(syslog.LOG_CRIT, 'Logger init failed')
            sys.exit(1)

        func()
    return inner


def push_defaults(func):
    """
        Push defaults.
    """
    def inner():
        logger.debug('begin')

        endpoint = None
        nvsm_api_cfg_file = config_settings().get('nvsm-api', 'config-file')
        nvsm_api_jwt_file = config_settings().get('nvsm-api', 'jwt-file')

        (ca_cert_file,
         _,
         _,
         _,
         port,
         address,
         _) = config_settings.nvsm_api_config_helper(nvsm_api_cfg_file)

        # TODO change Endpoint arguments to kargs
        if port is not None:
            endpoint = Endpoint(port=str(port), jwt_file=nvsm_api_jwt_file,
                                ca_cert_file=ca_cert_file, ip_addr=address)
        else:
            endpoint = Endpoint()

        connected = False
        attempts = 0
        while True:
            if not endpoint.accessible():
                logger.warning('{} attempt, Retrying after {} sec'.format(
                    attempts, NVSM_CONN_RETRY_TIME))
                time.sleep(NVSM_CONN_RETRY_TIME)
                attempts = attempts + 1
            else:
                connected = True
                break

        if not connected:
            logger.critical('Connection to thea is failed')
            sys.exit(1)

        catalog.init('storage')
        policy.push_defaults(['Volume', 'Drive'])
        # Check if storage controllers are present
        client = Client({Client.URI: Endpoint.storage_uri})
        json_obj = client.get_objects()
        if json_obj is not None:
            try:
                if len(json_obj['StorageControllers']) > 0:
                    policy.push_defaults(['StorageController'])
            except Exception as e:
                logger.error('Unable to fetch StorageController uri')
        logger.debug('end')
        func()
    return inner


# MQTT event registration and handling
def on_connect(client, userdata, flags, rc):
    logger.debug('MQ ON Connect Ret-Code:' + str(rc))

    if not rc:
        logger.debug('MQ Client Connect Success, result code: ' + str(rc))
        MQBroker.subscribe(client, Topic.TOPIC_STORAGE)
        MQBroker.subscribe(client, Topic.TOPIC_STORAGE_VOLUME_COLLECTION)
        MQBroker.subscribe(client, Topic.TOPIC_STORAGE_VOLUME)
        MQBroker.subscribe(client, Topic.TOPIC_STORAGE_DRIVE_COLLECTION)
    else:
        logger.error('MQ Client Connect failure, with result code: ' + str(rc))


def on_message(client, userdata, msg):
    logger.debug('MQ Message Received for Topic:' + msg.topic)

    if msg is None or msg.payload is None:
        return
    process_topic(msg)
    logger.debug('MQ Message for Topic: ' + msg.topic + ' Handled')


def process_topic(msg):
    logger.debug('process event for storage')
    ret_val = Err.ERR_CONN_REST_SRV
    attempt = 0

    if msg is None or msg.payload is None:
        return

    if msg.topic == Topic.TOPIC_STORAGE:
        # we will retry only incase of REST communication failure.
        while ret_val is Err.ERR_CONN_REST_SRV and \
              attempt < NVSM_RETRY_ATTEMPT_ON_MESSAGE:
            ret_val = process_storage(msg.payload)
            logger.debug('Storage Event: return value {}'.format(ret_val))
            if ret_val is Err.ERR_CONN_REST_SRV:
                attempt = attempt + 1
                continue

    elif Topic.TOPIC_STORAGE_DRIVE in msg.topic:
        while ret_val is Err.ERR_CONN_REST_SRV and \
                attempt < NVSM_RETRY_ATTEMPT_ON_MESSAGE:
            ret_val = process_drive(msg.payload)
            logger.debug('Drive Event: return value {}'.format(ret_val))
            if ret_val is Err.ERR_CONN_REST_SRV:
                attempt = attempt + 1
                continue

    elif re.compile(Topic.TOPIC_STORAGE_VOLUME_REGEX).match(msg.topic):
        while ret_val is Err.ERR_CONN_REST_SRV and \
                attempt < NVSM_RETRY_ATTEMPT_ON_MESSAGE:
            ret_val = process_volume(msg.payload)
            logger.debug('Volume Event: return value {}'.format(ret_val))
            if ret_val is Err.ERR_CONN_REST_SRV:
                attempt = attempt + 1
                continue

    elif Topic.TOPIC_STORAGE_VOLUME_COLLECTION in msg.topic:
        while ret_val is Err.ERR_CONN_REST_SRV and \
                attempt < NVSM_RETRY_ATTEMPT_ON_MESSAGE:
            ret_val = process_volume_collection(msg.payload)
            logger.debug(
                'Volume Collection Event: return value {}'.format(ret_val))
            if ret_val is Err.ERR_CONN_REST_SRV:
                attempt = attempt + 1
                continue

    elif Topic.TOPIC_STORAGE_CONTROLLER in msg.topic:
        while ret_val is Err.ERR_CONN_REST_SRV and \
                attempt < NVSM_RETRY_ATTEMPT_ON_MESSAGE:
            ret_val = process_controller(msg.payload)
            logger.debug('Controller Event: return value {}'.format(ret_val))
            if ret_val is Err.ERR_CONN_REST_SRV:
                attempt = attempt + 1
                continue
    else:
        logger.debug('Invalid TOPIC received, Topic: ' + msg.topic)

    if attempt is NVSM_RETRY_ATTEMPT_ON_MESSAGE:
        logger.error('messgage dropped, Thea server is down')

    logger.info('storage_dshm: MQ Message for Topic: '
                + msg.topic + ' Handled')
    return


# Check for drive missing
def process_storage(payload):
    policy = IPolicy.factory('Drive')
    logger.debug('Policy used {}'.format(policy))

    json_obj = Client.get_prop_value(payload)

    if json_obj is None:
        logger.error('Empty storage payload')
        return Err.ERROR
    try:
        supported_configuration = pa().enumerate_obj(
            'SystemCollection/1/StorageCollection/1/SupportedConfiguration')
    except Exception as e:
        supported_configuration = None
        logger.error('Unable to fetch supported configuration')

    # fetch drive list
    client = Client({Client.URI: Endpoint.storage_uri,
                     Client.JSON_NAME: "Drives",
                     Client.IS_REFERENCE: True})
    json_objs = client.get_objects()
    if json_objs is None:
        logger.error('Empty drive list')
        return Err.ERROR

    storage_element = Storage.factory('DriveCollection',
                                      json_objs,
                                      Endpoint().hostname(),
                                      Endpoint().serial(),
                                      supported_configuration
                                      )
    logger.debug("Created drive collection factory element")
    data_store = AlertsDataStore.factory('thea_db', 'Drive')
    if storage_element is not None and  \
       policy.monitor(storage_element.get_name()):
        logger.debug("Checking device {}".format(storage_element.get_name()))
        ret = data_store.fetch_alerts(storage_element.get_name())
        if ret['errored']:
            logger.error('Unable to fetch alert data from thea for volume')
            logger.error('Thea is in-accessible, waiting till connection'
                         'is re-established')
            return Err.ERR_CONN_REST_SRV
        alerts = ret['alerts']
        storage_element.apply_rules(storage_element, alerts)
        data_store.store_alerts(storage_element.get_name(), alerts)
    # process storage controller
    retVal = Err.SUCCESS
    if json_obj['StorageControllers@odata.count'] > 0:
        storage_controller_json = json_obj['StorageControllers'][0]
        retVal =  process_controller(storage_controller_json)
    return retVal


# Process volume collection for missing volumes
def process_volume_collection(payload):
    logger.debug('Process Volume Collection')
    if payload is None:
        logger.error('JSON OBJ received from the event'
                     'is None for VolumeCollection.')
        return Err.ERR_EMPTY_JSON_OBJS

    policy = IPolicy.factory('Volume')
    logger.debug('Policy used {}'.format(policy))
    json_obj = Client.get_prop_value(payload)
    if json_obj is None:
        logger.error('Invalid payload received, '
                     'not in VolumeCollection format.')
        return Err.ERR_INVALID_JSON_OBJ

    container = EquipmentContainer.factory('VolumeContainer', json_obj)
    if container is None:
        return Err.ERROR

    device_map = container.get_device_presence_map()
    logger.debug("Volume Device Map: {}".format(device_map))
    data_store = AlertsDataStore.factory('thea_db', 'Volume')

    for device in device_map.keys():
        if device_map[device] is True:
            continue
        else:
            vol_obj = {'Name': device,
                       'VolumeType': 'Unknown',
                       'IsPresent': 'No'}
            # Check if the volume data is available in the list
            volume = filter(lambda x: x['Name']==device, VolumeContainer.monitored_volumes)
            vol_list = list(volume)
            if len(vol_list) > 0:
                vol_obj['VolumeType'] = vol_list[0]['VolumeType']
            data_store = AlertsDataStore.factory('thea_db', 'Volume')
            # Create the volume and run the rules
            storage_element = Storage.factory('Volume',
                                              vol_obj,
                                              Endpoint().hostname(),
                                              Endpoint().serial())
            logger.debug('Created factory element')
            if storage_element is not None and \
                policy.monitor(storage_element.get_name()):
                logger.debug('Checking device '
                             '{}'.format(storage_element.get_name()))

                ret = data_store.fetch_alerts(storage_element.get_name())
                if ret['errored']:
                    logger.error('Unable to fetch alert data for Volume')
                    logger.error('Thea is in-accessible, wait till connection'
                                 'is re-established')
                    return Err.ERR_CONN_REST_SRV
                alerts = ret['alerts']

                storage_element.apply_rules(storage_element, alerts)
                data_store.store_alerts(storage_element.get_name(), alerts)

    return Err.SUCCESS


# Process volume payload for volume corruption
def process_volume(payload):
    logger.debug('Process Volume')
    if payload is None:
        logger.error('JSON OBJ received from the event is None for Volume.')
        return Err.ERR_EMPTY_JSON_OBJS

    policy = IPolicy.factory('Volume')
    logger.debug('Policy used {}'.format(policy))
    json_obj = Client.get_prop_value(payload)
    if json_obj is None:
        logger.error('Invalid payload received, not in Volume format.')
        return Err.ERR_INVALID_JSON_OBJ

    data_store = AlertsDataStore.factory('thea_db', 'Volume')
    # Create the volume and run the rules
    json_obj['IsPresent'] = 'Yes'
    storage_element = Storage.factory('Volume',
                                      json_obj,
                                      Endpoint().hostname(),
                                      Endpoint().serial())
    logger.debug('Created factory element')
    if storage_element is not None and \
       policy.monitor(storage_element.get_name()):
        logger.debug('Checking device {}'.format(storage_element.get_name()))

        ret = data_store.fetch_alerts(storage_element.get_name())
        if ret['errored']:
            logger.error('Unable to fetch alert data from thea for Drive')
            logger.error('Thea is in-accessible, waiting till connection'
                         'is re-established')
            return Err.ERR_CONN_REST_SRV
        alerts = ret['alerts']

        storage_element.apply_rules(storage_element, alerts)
        data_store.store_alerts(storage_element.get_name(), alerts)

    return Err.SUCCESS


# Process drive payload for drive parameters
def process_drive(payload):
    logger.debug('Process Drive')
    if payload is None:
        logger.error('Payload received from the event is None for Drives.')
        return Err.ERR_EMPTY_JSON_OBJS

    policy = IPolicy.factory('Drive')
    logger.debug('Policy used {}'.format(policy))
    json_obj = Client.get_prop_value(payload)
    if json_obj is None:
        logger.error('Invalid payload received, not in Drive format.')
        return Err.ERR_INVALID_JSON_OBJ
    try:
        supported_configuration = pa().enumerate_obj(
            'SystemCollection/1/StorageCollection/1/SupportedConfiguration')
    except Exception as e:
        supported_configuration = None
        logger.error('Unable to fetch supported configuration')

    data_store = AlertsDataStore.factory('thea_db', 'Drive')
    # Create the drive and run the rules
    storage_element = Storage.factory('Drive',
                                      json_obj,
                                      Endpoint().hostname(),
                                      Endpoint().serial(),
                                      supported_configuration
                                      )
    logger.debug('Created factory element')
    if storage_element is not None and \
       policy.monitor(storage_element.get_name()):
        logger.debug('Checking device {}'.format(storage_element.get_name()))

        ret = data_store.fetch_alerts(storage_element.get_name())
        if ret['errored']:
            logger.error('Unable to fetch alert data from thea for Drive')
            logger.error('Thea is in-accessible, waiting till connection'
                         'is re-established')
            return Err.ERR_CONN_REST_SRV
        alerts = ret['alerts']

        storage_element.apply_rules(storage_element, alerts)
        data_store.store_alerts(storage_element.get_name(), alerts)
    return Err.SUCCESS


# Storage Controller
def process_controller(json_obj):
    logger.debug('Process StorageController')
    policy = IPolicy.factory('StorageController')
    logger.debug("Policy used {}".format(policy))

    #json_obj = Client.get_prop_value(payload)
    if json_obj is None:
        logger.error('Invalid payload received, not in '
                     'StorageController format.')
        return Err.ERR_INVALID_JSON_OBJ

    data_store = AlertsDataStore.factory('thea_db', 'StorageController')
    storage_element = Storage.factory('StorageController',
                                      json_obj,
                                      Endpoint().hostname(),
                                      Endpoint().serial())

    logger.debug("Created factory element")
    if storage_element is not None and \
       policy.monitor(storage_element.get_name()):
        logger.debug('Checking device {}'.format(storage_element.get_name()))

        ret = data_store.fetch_alerts(storage_element.get_name())
        if ret['errored']:
            logger.error('Unable to fetch alert data from thea for '
                         'StorageController')
            logger.error('Thea is in-accessible, waiting till connection '
                         'is re-established')
            return Err.ERR_CONN_REST_SRV

        alerts = ret['alerts']

        storage_element.apply_rules(storage_element, alerts)

        data_store.store_alerts(storage_element.get_name(), alerts)
    return Err.SUCCESS


def sigterm_handler(mq_client, signum, frame):
    """
    Sig term handler.
    """
    logger.info('Executing graceful exit')
    mq_client.disconnect()


def init_storage():
    return Err.SUCCESS

@init_configuration
@init_logger
@push_defaults
def storage_main():
    """
    Main
    """
    ret_val = Err.ERROR
#    while ret_val != Err.SUCCESS:
#        time.sleep(STORAGE_INIT_RETRY_TIME)
#        ret_val = init_storage()
    init_pa(__name__)

    mq_client = MQBroker.factory(MQBroker.MQTT_CLIENT, NVSM_CLIENT_NAME,
                                 on_connect, on_message)
    if mq_client is None or \
       mq_client is Err.ERR_INVLAID_MQ_SECURITY_CERT or \
       mq_client is Err.ERR_INVLAID_MQ_ADDR_PORT:
        logger.critical('Failed in creating the MQBroker Client')
        sys.exit(1)
        # Register an handler for graceful shutdown.
    signal.signal(signal.SIGTERM, partial(sigterm_handler, mq_client))
    logger.debug('Run the MQ Client and listen to MQ Broker')
    # Start listening to MQTT
    try:
        mq_client.listen_for_ever()
    except (OSError, select.error) as e:
        logger.error('Select Error: {}'.format(e))
    except Exception as e:
        logger.error('Unhandled Exception: {}'.format(e))
        logger.debug(traceback.format_exc())

    logger.info('Exiting the main thread')
