#!/usr/bin/python
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Sys monitoring module

This module monitors CPUs, DIMMs and GPUs.
"""

import time
import os
import sys
import signal
import logging
import re
import subprocess
import select
import traceback
import syslog
from functools import partial


from nvsm_dshm.infra import pa
from nvsm_dshm.infra.helper import init_pa
from nvsm_dshm.thea import Endpoint
from nvsm_dshm.equipment import catalog
from nvsm_dshm.policy import policy
from nvsm_dshm.infra import config_settings
from nvsm_dshm.infra import log
from nvsm_dshm.mqclient import MQBrokerClient as MQBroker
from nvsm_dshm.thea import Topic
from nvsm_dshm.infra import ErrCode as Err
from nvsm_dshm.system import handler

NVSM_CONN_RETRY_TIME = 5  # in seconds
NVSM_RETRY_ATTEMPT_ON_MESSAGE = 5
NVSM_CLIENT_NAME = "nvsm_sys_client"

logger = logging.getLogger(__name__)


def init():
    """
        Initializer for the System mon entry module.
        1) Initializes the log file.
        2) Copies the alerts data to memory.
    """

    # read the config filename from system
    try:
        config_settings(config_file_name=os.environ['DSHM_CONFIG_FILE'])
    except Exception as e:
        syslog.syslog(syslog.LOG_CRIT,
                        'pcie_mon: Unable to read the config file')
        return Err.ERROR
    # Init logger
    ret = log.initialize(config_settings().get('logging', 'config-file'),
                         config_settings().get('logging', 'log-dir'),
                         'sys')
    if ret is False:
        return Err.ERROR

    logger.info('Initializing System monitoring')
    catalog.init('system')

    ret_val = Err.ERROR

    while ret_val is not Err.SUCCESS:
        ret_val = thea_connect()
        logger.debug("Thea_Connect ret:" + str(ret_val))
        if ret_val is not Err.SUCCESS:
            time.sleep(NVSM_CONN_RETRY_TIME)

    logger.debug('Exit')
    return Err.SUCCESS


def thea_connect():
    """
        Checks connectivity to Thea is established.
    """
    logger.debug("Thea_connect Enter")

    endpoint = None
    nvsm_api_config_file = config_settings().get('nvsm-api', 'config-file')
    nvsm_api_jwt_file = config_settings().get('nvsm-api', 'jwt-file')

    (ca_cert_file,
     _,
     _,
     _,
     port,
     address,
     _) = config_settings.nvsm_api_config_helper(nvsm_api_config_file)

    # TODO change Endpoint arguments to kargs
    if port is not None:
        endpoint = Endpoint(port=str(port), jwt_file=nvsm_api_jwt_file,
                            ca_cert_file=ca_cert_file, ip_addr=address)
    else:
        endpoint = Endpoint()

    if not endpoint.accessible():
        logger.error('Endpoint is in-accessible')
        return Err.ERR_CONN_REST_SRV

    logger.debug('Endpoint is accessible')

    init_pa(__name__)

    policy.push_defaults([ 'NVSwitch'])

    logger.debug("Thea_connect Exit")

    return Err.SUCCESS



def on_connect(client, userdata, flags, rc):
    """
        On Connect callback that is registered with MQTT client.
    """

    logger.debug("MQ ON Connect Ret-Code:" + str(rc))

    ret_val = Err.ERROR

    # useful when Thea API Server goes down,helps in reconnection.
    while ret_val is not Err.SUCCESS:
        ret_val = thea_connect()
        logger.debug("Thea_Connect ret:" + str(ret_val))
        if ret_val is not Err.SUCCESS:
            time.sleep(NVSM_CONN_RETRY_TIME)

    if rc == 0:
        # useful when mqtt broker goes down,
        # in reconnection we will still subscribe to all topics.
        logger.debug("MQ Client Connect Success, result code: " + str(rc))

        MQBroker.subscribe(client, Topic.TOPIC_NVSWITCH_HEALTH)
        MQBroker.subscribe(client, Topic.TOPIC_GPU_NVLINK_HEALTH)

    else:
        logger.error("MQ Client Connect failure, with result code: " + str(rc))

    logger.debug("MQ ON Connect Handled")


def call_handler(func):
    """
        Wrapper for the handler that processes the event.
    """

    # we set this as default,so that we callthe job() atleast once.
    ret_val = Err.ERR_CONN_REST_SRV
    attempt = 0

    if not func:
        logger.error("Funciton passed is None")
        return

    # we will retry only incase of REST communication failure.
    while ret_val is Err.ERR_CONN_REST_SRV and \
            attempt < NVSM_RETRY_ATTEMPT_ON_MESSAGE:
        ret_val = func()
        attempt = attempt + 1

    if attempt == NVSM_RETRY_ATTEMPT_ON_MESSAGE and ret_val is Err.ERR_CONN_REST_SRV:
        logger.error("Thea server is down, dropping the message")
    elif ret_val == Err.ERR_INVLAID_MQ_PAYLOAD:
        logger.error("Received invalid payload, dropping the message")
    elif ret_val != Err.SUCCESS:
        logger.error("Error occurred {}, dropping the message".format(ret_val))
    else:
        logger.debug("Successfully handled the message")


def on_message(client, userdata, msg):
    """
        On Message callback registered with MQTT Client.
    """

    logger.debug("MQ Message Received for Topic:" + msg.topic)

    if msg is None or msg.payload is None:
        return

    elif msg.topic == Topic.TOPIC_GPU_NVLINK_HEALTH:
        call_handler(partial(handler.nvlink_health, msg.payload))

    elif msg.topic == Topic.TOPIC_NVSWITCH_HEALTH:
        call_handler(partial(handler.nvswitch_health, msg.payload))

    else:
        # Message is not handled, just set the handler_ret to True.
        logger.debug(
            "Invalid TOPIC received, it was not handled, Topic: " +
            msg.topic)

    logger.debug("MQ Message for Topic: " + msg.topic + " Handled")


def graceful_exit(mq_client, signum, frame):
    """
    Handler to gracefully terminate all threads.
    """
    logger.info("Executing graceful exit")
    mq_client.disconnect()
    handler.cancel_timer()


def sys_main():
    """
    Main
    """
    # Initialize the System monitoring subsystem.
    if init() != Err.SUCCESS:
        logger.critical('Failed to launch System monitor')
        sys.exit(1)

    # Initialize MQ Borker
    mq_client = MQBroker.factory(MQBroker.MQTT_CLIENT, NVSM_CLIENT_NAME,
                                 on_connect, on_message)
    if mq_client is None or \
            mq_client is Err.ERR_INVLAID_MQ_SECURITY_CERT or \
            mq_client is Err.ERR_INVLAID_MQ_ADDR_PORT:
        logger.critical('Failed in creating the MQBroker Client')
        sys.exit(1)

    # Register an handler for graceful shutdown.
    signal.signal(signal.SIGTERM, partial(graceful_exit, mq_client))

    logger.debug("Run the MQ Client and listen to MQ Broker")

    # Start listening to MQTT
    try:
        mq_client.listen_for_ever()
    except (OSError, select.error) as e:
        logger.error('Select Error: {}'.format(e))
    except Exception as e:
        logger.error('Unhandled Exception: {}'.format(e))
        logger.debug(traceback.format_exc())

    logger.info("Exiting the main thread")
