# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""

Thea Parser Module

Implements the functionality to fetch the data/ JSON objects returned by Thea.
"""

import json
import logging

from nvsm_dshm.thea.endpoint import Endpoint


logger = logging.getLogger(__name__)


class Client(object):

    """
    Thea Client Class.

    Main functionality:
        Fetches the list of objects from the given URI
    """
    URI = "uri"
    JSON_NAME = "json_name"
    FILTER_FUNC = "filter_func"
    IS_REFERENCE = "is_reference"
    ALL = "all"

    def __init__(self, parse_info):
        self.uri = parse_info[Client.URI]

        self.json_name = None
        if Client.JSON_NAME in list(parse_info.keys()):
            self.json_name = parse_info[Client.JSON_NAME]

        self.is_reference = False
        if Client.IS_REFERENCE in list(parse_info.keys()):
            self.is_reference = parse_info[Client.IS_REFERENCE]

        self.all = False
        if Client.ALL in list(parse_info.keys()):
            self.all = parse_info[Client.ALL]

        self.filter_func = None
        if Client.FILTER_FUNC in list(parse_info.keys()):
            self.filter_func = parse_info[Client.FILTER_FUNC]

    def get_objects(self):
        """
        This method uses the URI, JSON_NAME and FILTER_FUNC that were
        initialized through the constructor.

        Gets the objects with the given json attribute name, and returns
        the list of objects.

        If the filter function is passed, it filters the objects returned.
        """
        ret = Endpoint().issue(self.uri, "GET")
        if ret is None:
            return None

        if self.all is True:
            devices = Client.get_prop_value(ret.text)
        else:
            devices = Client.get_prop_value(ret.text, self.json_name)

        if devices is None:
            return None

        if self.is_reference is True:
            if isinstance(devices, list):
                for index, device in enumerate(devices):
                    devices[index] = Client.get_referenced(device)
            else:
                    devices = Client.get_referenced(devices)

        if self.filter_func is not None:
            try:
                device_list = []
                for device in devices:
                    if self.filter_func(device) is True:
                        device_list.append(device)
                devices = device_list
            except Exception:
                logger.error("Filter func failed: {}".format(self))

        if len(devices) != 0:
            return devices

        return None

    @staticmethod
    def get_referenced(json_obj):
        """
        Returns the object referenced using '@odata.id' attribute in the
        json_obj.
        """
        obj = None

        if "@odata.id" in list(json_obj.keys()):
            ret = Endpoint().issue(json_obj["@odata.id"], "GET")

            if ret is not None:
                obj = json.loads(ret.text)
            else:
                logger.error("Could not fetch the json object at {}".format(
                    json_obj["@odata.id"]))

        return obj

    @staticmethod
    def get_prop_value(rawdata, prop_name=None, obj_name=None, obj_index=None):
        """
        From the JSON blob <rawdata>, it fetches the property value.
        i.e; rawdata[obj_name][obj_index][prop_name] is returned.

        obj_name and obj_index are optional.
        """
        try:
            obj = json.loads(rawdata)
            if prop_name is None:
                value = obj
            elif obj_name is None:
                value = obj[prop_name]
            elif obj_index is None:
                value = obj[obj_name][prop_name]
            else:
                value = obj[obj_name][obj_index][prop_name]
        except ValueError as err:
            logger.error(
                "Json decoding error, value error for key: %s" % (err))
            return None
        except KeyError as err:
            logger.error("Json decoding error, key absent: %s" % (err))
            return None

        return value

    # TODO: See if these can be enhanced same as get_objects.
    @classmethod
    def put(cls, uri, data):
        ret = Endpoint().issue(uri, 'PUT', data)
        if ret is None:
            logger.error("Failed to put ({}) to uri {}".format(data, uri))
            return False

        return True

    # TODO: See if these can be enhanced same as get_objects.
    @classmethod
    def post(cls, uri, data):
        ret = Endpoint().issue(uri, 'POST', data)
        if ret is None:
            logger.error("Failed to post ({}) to uri {}".format(data, uri))
            return False

        return True

    def __repr__(self):
        return '{}(uri:{!r},\
                json_name:{!r},\
                filter_func:{!r},\
                is_reference:{!r},\
                all:{!r})'.format(
            self.__class__.__name__,
            self.uri,
            self.json_name,
            self.filter_func.__name__,
            self.is_reference,
            self.all
        )
