# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""
Thea endpoint Module
"""

import os
import json
import logging
import requests
import subprocess
import urllib3
import time
from nvsm_dshm.infra import singleton
from nvsm_dshm.infra import pa

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
urllib3.disable_warnings(urllib3.exceptions.SecurityWarning)
logger = logging.getLogger(__name__)

_IP = '127.0.0.1'
_PORT = '273'
_SERVICE_ROOT = '/redfish/v1'
_NVSM_ROOT = '/nvsm/v1'
DEFAULT_JWT_FILE = "/etc/nvsm/nvsm-perpetual.jwt"
_TIMEOUT = 10
DEFAULT_CERT_FILE = '/etc/nvsm/nvsm-ca.crt'


class Endpoint(object, metaclass=singleton):

    """
        Endpoint class represents the Thea endpoint that DSHM connects to.

        The end point class connects to the default ip/port (127.0.0.1/273).

        Users of this class must treat it as a singleton and use 'instance'
        class method to invoke it.

        This is a crude implementation to support current needs. Might have
        to modify to support:
            1) Allowing the user to configure the Thea end point details.
            2) Allowing DSHM to connect to more than one Thea end point.
    """

    # TODO:
    # Later implement code to fetch these URIs dynamically from the endpoint
    # after initializing the connection.
    #
    # Use @property instead of the accessor methods bearer(), method(),
    # hostname(), serial().
    system_uri = _SERVICE_ROOT + '/Systems/1'
    thermal_uri = _SERVICE_ROOT + '/Chassis/1/Thermal'
    power_uri = _SERVICE_ROOT + '/Chassis/1/Power'
    power_alerts_uri = '/nvsm/v1/Chassis/1/Power/Alerts'
    gpu_uri = _NVSM_ROOT + '/Systems/1/GPUs'
    nvswitch_uri = _NVSM_ROOT + '/Systems/1/NVSwitches'
    memory_uri = system_uri + '/Memory'
    processor_uri = system_uri + '/Processors'
    storage_drive_uri = system_uri + '/Storage/1'
    storage_uri = system_uri + '/Storage/1'
    storage_drive_led_uri = _NVSM_ROOT + '/Systems/1/Storage/1/Drives/{}/LEDStatus'
    storage_volume_uri = system_uri + '/Storage/1/Volumes'
    storage_controller_uri = system_uri + '/Storage/1'
    storage_pcie_uri = _NVSM_ROOT + '/Systems/1/PCIeDevices'
    pci_uri = _NVSM_ROOT + '/Systems/1/PCIeDevices'
    #config_uri = _NVSM_ROOT + '/NvsmConfig'
    chassis_uri = _SERVICE_ROOT + '/Chassis/1'
    

    MODEL_MANAGEMENT_SERVER = 'ManagementServer'

    def __init__(self, ip_addr=_IP, port=_PORT,
                 root=_SERVICE_ROOT, jwt_file=DEFAULT_JWT_FILE,
                 ca_cert_file=DEFAULT_CERT_FILE):
        """
            Initialize instance
        """
        self.ip_addr = ip_addr
        self.port = port
        self.root = root
        self._bearer = None
        self._host_name = {}
        self._serial = {}
        self._location = {}
        self._model = None
        self.jwt_file = jwt_file
        self.ca_cert_file = ca_cert_file

    def bearer(self):
        """
            Return the bearer key for this endpoint
        """
        return self._bearer

    def _update_host_info(self, _id, _parent):
        ret = self.issue("{}/{}/{}".format(_SERVICE_ROOT,
                                           _parent,
                                           _id), "GET")
        if ret:
            obj = json.loads(ret.text)
            try:
                self._host_name[_id] = obj['Name']
            except:
                pass
            try:
                self._serial[_id] = obj["SerialNumber"]
            except:
                pass
            try:
                self._location[_id] = obj["Location"]
            except:
                pass

    def hostname(self, _id=None, _parent="Chassis"):
        """
            Return hostname
        """
        if not self._host_name and not _id:
            return None

        if not _id:
            _id = list(self._host_name)[0]

        if _id not in self._host_name:
            # try to get hostname from parent uri
            self._update_host_info(_id, _parent)

        # check again if hostname is fetched
        if _id not in self._host_name:
            return None

        return self._host_name[_id]

    def serial(self, _id=None, _parent="Chassis"):
        """
            Return serial
        """
        if not self._serial and not _id:
            return None

        if not _id:
            _id = list(self._serial)[0]

        if _id not in self._serial:
            # try to get srial from chassis uri
            self._update_host_info(_id, _parent)

        # check again if serial is fetched
        if _id not in self._serial:
            return None

        return self._serial[_id]

    def location(self, _id=None, _parent="Chassis"):
        """
            Return location
        """
        default_location = {"Placement": {
            "Row": "", "Rack": "", "RackOffset": ""}}

        if not self._location and not _id:
            return default_location

        if not _id:
            _id = list(self._location)[0]

        if _id not in self._location:
            # try to get hostname from parent uri
            self._update_host_info(_id, _parent)

        # check again if hostname is fetched
        if _id not in self._location:
            return default_location

        location_info = self._location[_id]
        try:
            # add spaces
            location_info["Placement"]["Row"] = str(
                location_info["Placement"]["Row"]).strip() + ' '
            location_info["Placement"]["Rack"] = str(
                location_info["Placement"]["Rack"]).strip() + ' '
            location_info["Placement"]["RackOffset"] = str(
                location_info["Placement"]["RackOffset"]).strip() + ' '
        except Exception as e:
            logger.debug(e, exc_info=True)
        return location_info

    def model(self):
        """
            Return model
        """
        if not self._model and pa().enumerate_obj('Capabilities/Monitoring/ManagementServerModel'):
            self._model = self.MODEL_MANAGEMENT_SERVER
            logger.info('Turning on ManagementServerModel')
        return self._model

    def accessible(self):
        """
        This method is used to establish connection with thea.

        This can be executed by the client, whenever Thea goes in-accessible to
            1) Re-initialize connection with Thea.
            2) Check if Thea is accessible.

        If Thea was uninstalled and re-installed, then the bearer key will
        change. This method fetches the bearer key afresh during
        re-initialization.

        It also gets the Thea endpoint's identity (hostname, serial num)
        through the redfish API.
        """
        # Fetch the bearer key.
        self._bearer = None
        if os.path.isfile(self.jwt_file):
            with open(self.jwt_file, 'r') as fhandle:
                self._bearer = fhandle.read()

        # Return false, if bearer key could not be read.
        if self._bearer is None:
            logger.critical('Unable to get the bearer key')
            return False

        endpoint_accesible = False
        # Check for accessibility through system_uri or chassis uri or redfish root uri
        uris_for_accessibility_test = [Endpoint.system_uri, _SERVICE_ROOT, Endpoint.chassis_uri]

        for uri in uris_for_accessibility_test:
            uri_ret = self.issue(uri, "GET")
            if uri_ret.status_code == 200:
                endpoint_accesible=True
                break

        if not endpoint_accesible:
            logger.critical('Unable to get the endpoint connectivity')
            return False


        # Issue get on system uri and read hostname/serial number
        ret = self.issue(Endpoint.system_uri, "GET")
        if ret is not None:
            json_obj = json.loads(ret.text)
            self._host_name['1'] = json_obj['HostName']
            self._serial['1'] = json_obj['SerialNumber']
            self._model = json_obj['Model']
        else:
            logger.debug(
                'Unable to get the endpoint identity on system_uri')
            cmd = "test -f /usr/sbin/nvsm_rtx_integration"
            result = subprocess.Popen(cmd.split(),
                                      stdout=subprocess.PIPE,
                                      shell=False)
            _, _= result.communicate()
            # returncode = 0 - bladerunner pod platfrom
            if result.returncode:
                return False
        return True

    def issue(self, url_path, access_type, data=None):
        """
        Issues http request and returns the response
        """
        if not url_path:
            return None

        if access_type == "GET":
            try:
                url_prefix = 'https://' + self.ip_addr + ':' + self.port
                ret = requests.get(url_prefix + url_path,
                                   verify=self.ca_cert_file,
                                   headers={'Authorization': self.bearer()},
                                   timeout=_TIMEOUT)
                if ret.status_code == 200:
                    return ret
                else:
                    logger.error("HTTP GET %s returned %d" %
                                 (url_prefix + url_path, ret.status_code))
                    return None
            except requests.exceptions.RequestException as excep:
                logger.exception('HTTP GET failed: ')
            return None

        elif access_type == "POST":
            try:
                url_prefix = 'https://' + self.ip_addr + ':' + self.port
                ret = requests.post(url_prefix + url_path,
                                    verify=self.ca_cert_file,
                                    headers={'Authorization': self.bearer()},
                                    json=data,
                                    timeout=_TIMEOUT)
                if ret.status_code == 200 or ret.status_code == 201:
                    return ret
                else:
                    logger.error("HTTP POST %s returned %d" %
                                 (url_prefix + url_path, ret.status_code))
                    return None
            except requests.exceptions.RequestException as excep:
                logger.exception(
                    'HTTP POST of data ({}) failed : '.format(data))
            return None

        elif access_type == "PUT":
            try:
                url_prefix = 'https://' + self.ip_addr + ':' + self.port
                ret = requests.put(url_prefix + url_path,
                                   verify=self.ca_cert_file,
                                   headers={'Authorization': self.bearer()},
                                   json=data,
                                   timeout=_TIMEOUT)
                if ret.status_code == 200:
                    return ret
                else:
                    logger.error("HTTP PUT %s returned %d" %
                                 (url_prefix + url_path, ret.status_code))
                    return None
            except requests.exceptions.RequestException as excep:
                logger.exception(
                    'HTTP PUT of data ({}) failed : '.format(data))
            return None

        else:
            return None
