#!/usr/bin/env python3

import argparse
import io
import logging
import os
import signal
import sys
import traceback

def getArgs():
    parser = argparse.ArgumentParser(prog = 'nvsm-health',
            description = 'NVIDIA System Information')
    parser.add_argument('-o', '--output', '--output-file', default=None,
            help='''File path in which to write output.''')
    parser.add_argument('--version', action='store_true',
            help='''Print the NVSMHealth version string and exit.''')
    parser.add_argument('--discover', action='store_true',
            help='''Print a summary of what NVSMHealth will do and exit.''')
    parser.add_argument('--dump', action='store_true',
            help='''Create health dump of system and exit.''')
    parser.add_argument('--graph', action='store_true',
            help='''Print graphviz representation of health checks and exit.''')
    parser.add_argument('--order', action='store_true',
            help='''Print graphviz representation of health checks in execution order and exit.''')
    parser.add_argument('--show', action='store_true',
            help='''Show system health checks result''')
    parser.add_argument('--log-level',
        choices=[ 'debug', 'info', 'warning', 'error', 'critical' ],
        metavar='LEVEL', default='warning',
        help='''Set the output logging level. Possible values for LEVEL are
        "debug", "info", "warning", "error", and "critical". The default value
        is "warning".''')
    parser.add_argument('--with-progress', action='store_true',
        help='''Enable progress widget.''')
    return parser.parse_args()


def initialize_logger(log_level, outpath=None, is_dump=False):
    logger = logging.getLogger()
    logger.setLevel(logging.NOTSET)

    # Set log level for stderr as requested
    stderr_log_handler = logging.StreamHandler(
            stream=sys.stderr)
    if outpath and not is_dump:
        stderr_log_handler = logging.FileHandler(outpath)

    stderr_log_handler.setFormatter(
            logging.Formatter(
                fmt='{levelname}: {message}', style='{'))
    stderr_log_handler.setLevel(log_level)
    logger.addHandler(stderr_log_handler)
    # Log everything to the NVSMHealth log file
    logging.nvsmhealth_log_file = io.TextIOWrapper(buffer=io.BytesIO())
    logging.nvsmhealth_log_handler = logging.StreamHandler(
            stream=logging.nvsmhealth_log_file)
    logging.nvsmhealth_log_handler.setFormatter(
            logging.Formatter(
                fmt='{asctime} {levelname}: {message}',
                style='{'))
    logging.nvsmhealth_log_handler.setLevel(logging.NOTSET)
    logger.addHandler(logging.nvsmhealth_log_handler)


def __main(args):
    try:
        import nvsmhealth.version
        version = nvsmhealth.version.NVSMHEALTH_VERSION
    except:
        version = 'version unknown'

    # (optional) Print the version string
    if args.version:
        print('NVSMHealth {0}'.format(version),
            file=sys.stdout)
        sys.exit(0)

    args.log_level = {
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warning': logging.WARNING,
        'error': logging.ERROR,
        'critical': logging.CRITICAL
    }.get(args.log_level, logging.WARNING)

    initialize_logger(args.log_level, args.output, args.dump)

    # Always log the NVSMHealth version string
    logging.info('NVSMHealth {0}'.format(version))

    if args.graph:
        # Print graphviz representation of health checks and exit
        from nvsmhealth.discovery import graph_tasks
        graph_tasks()
        sys.exit(0)

    if args.order:
        # Print graphviz representation of health checks in sorted order and exit
        from nvsmhealth.discovery import tasks_order
        tasks_order()
        sys.exit(0)

    progress_bar = None
    if args.with_progress:
        from nvsm.progress import ProgressBar
        try:
            progress_bar = ProgressBar()
        except:
            pass

    # Ensure superuser privileges
    if os.geteuid() != 0:
        print('ERROR: nvsm-health must be run with root privileges', file=sys.stderr)
        sys.exit(1)

    # Ignore all keyboard interrupts
    def ignore_keyboard_interrupt(signum, stackframe):
        logging.info('Keyboard interrupt received; ignoring')
        stack_trace = io.StringIO()
        traceback.print_stack(f=stackframe, file=stack_trace)
        logging.debug('Stack trace at keyboard interrupt:\n{0}'.format(
                    stack_trace.getvalue()))
    signal.signal(signal.SIGINT, ignore_keyboard_interrupt)

    # Ignore SIG_PIPE and don't throw exceptions on it - BUG200510749
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)

    if args.discover:
        # Print summary of test content and exit (without running any commands)
        from nvsmhealth.discovery import discover_tasks
        discover_tasks()
        sys.exit(0)

    if args.dump:
        # Create nvsm health dump of system
        rc = 0
        try:
            # Collect nvsmhealth (as well as sosreport) files
            from nvsmhealth.handler import dump
            rc = dump(output_path=args.output, progress_bar=progress_bar)
        except Exception as e:
            print('ERROR: Could not create NVSMHealth log: {0}'.format(e),
                    file=sys.stderr)
            traceback.print_exc()
            sys.exit(1)
        sys.exit(rc)

    # For no args execute all health checks and pretty print the results
    from nvsmhealth.handler import health
    rc = 0
    try:
        rc = health(output_path=args.output, progress_bar=progress_bar)
    except Exception as e:
        print('ERROR: {0}'.format(e),
                file=sys.stderr)
        traceback.print_exc()
        sys.exit(1)
    sys.exit(rc)

def main():
    args = getArgs()
    __main(args=args)

if __name__ == '__main__':
    main()

