# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from subprocess import TimeoutExpired
import datetime
import fnmatch
import glob
import io
import json
import logging
import os
import re
import shutil
import signal
import socket
import stat
import subprocess
import sys
import tarfile
import tempfile
import time
import traceback

class CollectionError(RuntimeError):
    pass

class ConfigurationError(RuntimeError):
    pass

class CollectionContext:
    def __init__(self, output_path=None, stream_collector=None):
        self.__output_path = output_path
        self.__stream_collector = stream_collector
        self.__currentTask = None
        self.__taskCollectedFiles = {}

    def __enter__(self):
        # Create a temporary directory, shared among data collectors
        self.__tempdir = tempfile.mkdtemp(prefix='nvsm-health-temp-')
        logging.debug('Created temporary working directory {0}'.format(
                    self.__tempdir))
        # Create the output tarball for final result
        if not self.__output_path:
            # Output path not specified; create output in /tmp
            filename = 'nvsm-health-{hostname}-{time}.tar.xz'.format(
                    hostname=socket.gethostname(),
                    time=datetime.datetime.now().strftime('%Y%m%d%H%M%S'))
            self.__output_path = os.path.join(tempfile.gettempdir(), filename)
        else:
            if not self.__output_path.endswith('.tar.xz'):
                self.__output_path = self.__output_path + '.tar.xz'
        # Open the output file with restrictive permissions
        self.__output_file = io.FileIO(os.open(
                self.__output_path,
                os.O_CREAT | os.O_EXCL | os.O_WRONLY,
                0o600), 'w')
        self.__tarball = tarfile.open(
                mode='w|xz',
                fileobj=self.__output_file, format=tarfile.GNU_FORMAT)
        logging.info('Writing output tarball to {0}'.format(
                    os.path.realpath(self.__output_path)))
        # Note: Downstream parsers, for e.g., Call home in DSHM is dependent on
        # the format of this print output.
        dump_msg = f'''
        This command will collect system configuration and diagnostic information to help diagnose the system.
        The output of this command will be written to: {os.path.realpath(self.__output_path)}
        The output may contain data considered sensitive and should be reviewed before sending to any third party.
        '''
        if self.__stream_collector:
            self.__stream_collector.stream(dump_msg)

        # Limit access to prefix directory of tarball by specifying mode 0700
        # NOTE: This is merely a convenience, to reduce the likelihood of a
        # user accidentally exposing the contents of the tarball to
        # unprivileged users.
        info = tarfile.TarInfo()
        info.name = self.prefix
        info.type = tarfile.DIRTYPE
        info.mode = 0o700
        info.uid = 0
        info.gid = 0
        info.uname = 'root'
        info.gname = 'root'
        info.mtime = time.time()
        self.__tarball.addfile(tarinfo=info)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__writeReport()
        self.__writeLog()
        self.__tarball.close()
        self.__output_file.close()
        shutil.rmtree(self.__tempdir, ignore_errors=True)

    def _setCurrentTask(self, task):
        self.__currentTask = task
    def useTask(self, task):
        collection_context = self
        old_task = self.__currentTask
        class CurrentTaskContext(object):
            def __init__(self, task):
                self.__task = task
            def __enter__(self):
                collection_context._setCurrentTask(self.__task)
                return self
            def __exit__(self, type, value, traceback):
                collection_context._setCurrentTask(old_task)
        return CurrentTaskContext(task)

    @property
    def prefix(self):
        return os.path.basename(self.__output_path).split('.')[0]

    @property
    def tempdir(self):
        return self.__tempdir

    def add_file(self, fileobj, tarinfo=None, path=None):
        # Use name of file object if path is not given
        if not path:
            path=fileobj.name
        logging.info('Adding file to collection at path "{0}"'.format(path))
        temp = None
        try:
            s = os.fstat(fileobj.fileno())
            logging.debug('File status for "{0}": {1}'.format(path, str(s)))
        except:
            pass  # fileobj does not have OS file descriptor
        else:
            # Check for virtual files (such as in /proc) with zero length
            if s[stat.ST_SIZE] == 0:
                logging.debug('Found special zero-length file "{0}"'.format(path))
                # Copy entire file to /tmp before adding to archive
                temp = tempfile.NamedTemporaryFile(
                        mode='w+b', dir=self.tempdir)
                shutil.copyfileobj(fileobj, temp)
                temp.seek(0)
                fileobj = temp
        # Add the file to the tar archive
        if not tarinfo:
            tarinfo = self.__tarball.gettarinfo(fileobj=fileobj)
        tarinfo.name = '{0}/{1}'.format(self.prefix, path.lstrip('/'))
        self.__tarball.addfile(tarinfo, fileobj=fileobj)
        # Close (and delete) temporary file if used
        if temp:
            temp.close()
        # Record task to file path association
        if self.__currentTask:
            task_name = self.__currentTask.getName()
            if task_name not in self.__taskCollectedFiles:
                self.__taskCollectedFiles[task_name] = []
            self.__taskCollectedFiles[task_name].append(path)

    def __applyReportTemplate(self, original_path):
        relative_path = original_path.replace('/usr/share/doc/nvsm-health/html/', '')

        with open(original_path, mode='rb') as orig:
            data = orig.read()

        # Apply template for all collected files
        for task_name, paths in self.__taskCollectedFiles.items():
            marker = '<!--NVSM_TASK_COLLECTED_FILE_LIST_{}-->'.format(task_name)
            file_list = ''.join(map(lambda x: '<li><a href="../{}" target="_blank">{}</a></li>'.format(x.lstrip('/'), x), paths))
            replacement = '''
                <div class="section" id="{anchor}">
                    <h4>
                        Files in Report
                        <a class="headerlink" href="#{anchor}" title="Permalink to this headline">¶</a>
                    </h4>
                    <ul class="simple">{file_list}</ul>
                </div>
            '''.format(file_list=file_list,
                       anchor='{}__files_in_report'.format(task_name))
            data = data.replace(marker.encode('utf-8'), replacement.encode('utf-8'))

        with tempfile.NamedTemporaryFile(mode='r+b', dir=self.tempdir) as temp:
            # Write result to temporary file
            temp.write(data)
            temp.seek(0)
            # Add resulting file to archive
            tarinfo = self.__tarball.gettarinfo(fileobj=temp)
            tarinfo.name = '{0}/nvsmhealth_reports/{1}'.format(self.prefix, relative_path)
            logging.info('Adding file to NVSM Health report from template at "{0}"'.format(original_path))
            self.__tarball.addfile(fileobj=temp, tarinfo=tarinfo)

    def __writeReport(self):
        for root, dirs, files in os.walk('/usr/share/doc/nvsm-health/html'):
            for f in files:
                path = os.path.join(root, f)
                self.__applyReportTemplate(original_path=path)

    def __writeLog(self):
        logging.info("Adding NVSMHealth log to collection (last log entry)")
        logging.nvsmhealth_log_handler.flush()
        logger = logging.getLogger()
        logger.removeHandler(logging.nvsmhealth_log_handler)
        info = tarfile.TarInfo()
        info.name = 'nvsmhealth_log.txt'
        info.type = tarfile.REGTYPE
        info.mode = 0o644
        info.uid = 0
        info.gid = 0
        info.uname = 'root'
        info.gname = 'root'
        info.mtime = time.time()
        nvsmhealth_log_buff = logging.nvsmhealth_log_file.detach()
        nvsmhealth_log_buff.seek(0, io.SEEK_END)
        info.size = nvsmhealth_log_buff.tell()
        nvsmhealth_log_buff.seek(0)
        self.add_file(
                fileobj=nvsmhealth_log_buff,
                tarinfo=info,
                path='/nvsmhealth_log.txt')
