# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import importlib
import logging
import os
import pkgutil
import sys
import json
import io
from nvsmhealth.graph import TaskDependencyGraph
from nvsmhealth import tasks, config, modules
from nvsmhealth.tasks.stream import HealthSummaryStream

# JSON encoder for python additional objects
class CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (set, range)):
            return list(obj)
        if isinstance(obj, io.TextIOWrapper):
            return str(obj)
        if isinstance(obj, bytes):
            return obj.decode("utf-8")
        return json.JSONEncoder.default(self, obj)

def discover_tasks():
    import nvsmhealth.modules

    # Generator function to discover and iterate over all tasks
    def task_iterable(module_names=nvsmhealth.modules.__all__):
        for module_name in module_names:
            module = importlib.import_module('nvsmhealth.modules.{}'.format(module_name))
            logging.info('Found NVSMHealth module: {}'.format(module_name))
            for member_name in module.__all__:
                if member_name not in module.__dict__:
                    logging.warning(
                        'Missing member "{member}" in NVSMHealth module "{module}"'.format(
                        member=member_name,
                        module=module_name))
                    continue
                member = module.__dict__[member_name]
                if not isinstance(member, nvsmhealth.tasks.ITask):
                    continue
                yield (member, member_name, module_name, module)

    # Assign names and modules to each task based on Python object name
    for task, task_name, module_name, module in task_iterable():
        task.name(task_name)
        task.path('nvsmhealth.modules.{module_name}.{task_name}'.format(
                module_name=module_name, task_name=task_name))
        task.setModule(module)

    # Create implicit dependency for tasks that declare parameters
    for task, task_name, module_name, module in task_iterable():
        if task.getParameters():
            task.depends(modules.parameter.parameters)

    # Store parent -> child edge information (now that inter-module
    # dependencies have been resolved)
    for task, task_name, module_name, module in task_iterable():
        task._registerWithParents()

    # Create the dependency graph for all tasks
    g = TaskDependencyGraph()

    # Find all tasks and root tasks
    tasks = set()
    root_tasks = set()
    for task, task_name, module_name, module in task_iterable():
        # Add dependency task to adjacency list of current task in the graph
        for parent_task in task.parents:
            g.addTaskDependency(parent_task, task)

        tasks.add(task)
        if task.isRoot():
            root_tasks.add(task)

    # Check for cycles in task dependency graph
    visited = set()
    for task in root_tasks:
        task.checkCycles(visited)
    if tasks > visited:
        raise nvsmhealth.exceptions.CyclicTaskDependency(
                'Cycle detected in task dependency graph')

    # Perform topological sort for all tasks in the dependency graph
    g.taskTopologicalSort(tasks)

    return root_tasks, g

def graph_tasks(file=sys.stdout):
    # Print Graphviz representation of task dependency graph
    # The Graphviz representation can later be converted to PNG using the "dot"
    # command
    root_tasks, _ = discover_tasks()
    print(
        'digraph G {\n' \
        '  rankdir=RL;',
        file=file)
    for task in root_tasks:
        task.printGraphviz(file=file)
    print('}', file=file)

def tasks_order(file=sys.stdout):
    # Print Graphviz representation of tasks execution order from dependency graph
    # The Graphviz representation can later be converted to PNG using the "dot"
    # command
    _, g = discover_tasks()
    g.printSortedGraphviz()

def run_with_timeout(func):
    def handler(signum, frame):
        raise TimeoutError()
    import signal
    signal.signal(signal.SIGALRM, handler)
    # 3 mins timeout for every icheck and itask
    signal.alarm(180)
    try:
        func()
    except TimeoutError:
        logging.warning(f"{func} timeout")
    signal.alarm(0)

def execute_tasks(publish_cb=None, dump=False, progress_bar=None, output_file=None):
    import traceback
    import time
    import tempfile

    # Get the sorted order of tasks from the graph
    # Execute each task from the sorted order and collect the results
    if not execute_tasks.__order:
        _, g = discover_tasks()
        execute_tasks.__order = g.getOrder()

    # Create tempfile for storing nvsm health checks result
    temp_file = '{}/nvsm-health-{}.json'.format(tempfile.gettempdir(), int(time.time()))
    temp_data = {}
    temp_result = ''

    # Collect all health checks
    health_result = {'healthy':0, 'unhealthy':0, 'unknown':0, 'informational':0}
    health_summary_stream = HealthSummaryStream(publish_cb)
    health_summary_stream.stream_obj.totalChecks = len(execute_tasks.__order)
    health_summary_stream.stream(config.get_nvsm_health_header())

    temp_result += '{}\n'.format(health_summary_stream.stream_obj.message)

    # Set the current working directory to current path
    # Needed to identify executable scripts system path
    os.chdir(os.path.dirname(__file__))

    plat_not_found = False
    task_num = 0
    for task in execute_tasks.__order:
        if not dump:
            # Skip tasks that are for "dump health"
            if task.getModule() in [ modules.dump, modules.sosreport ]:
                continue

        task_name = task.getName()
        temp_data[task_name] = {}

        if not task.checkPredicate():
            # Some tasks need to be skipped under certain conditions
            # (e.g. platform support, unavailable tools)
            logging.info('Skipping task "{name}" due to predicate'.format(name=task_name))
            continue
        logging.info('Starting task: {name}'.format(name=task_name))
        try:
            # Install publish callback
            task.setCallback(publish_cb)
            if isinstance(task, tasks.command.RunCommand):
                task.run()
                # Store run comand task info into tempfile
                temp_data[task_name]['type'] = 'command'
                temp_data[task_name]['command'] = task.getArgs()
                temp_data[task_name]['timeout'] = task.getTimeout()
                temp_data[task_name]['output'] = task.getOutput()
                temp_data[task_name]['return_code'] = task.getReturnCode()
                temp_data[task_name]['error'] = task.getError()
            elif isinstance(task, tasks.ICheck):
                run_with_timeout(task.run)
                result = task.getResult()
                for key in health_result.keys():
                    health_result[key] += result[key]
                # Store check task info into tempfile
                temp_data[task_name]['type'] = 'check'
                temp_data[task_name]['result'] = task.getResult()
            elif isinstance(task, tasks.ILogs):
                pass
            elif isinstance(task, tasks.ITask):
                run_with_timeout(task.run)
                # Store parse task info into tempfile
                temp_data[task_name]['type'] = 'parse'
                if hasattr(task, 'getResult'):
                    temp_data[task_name]['result'] = task.getResult()
            else:
                pass
            task.sendComplete()
            if task.getMessages():
                temp_result += task.getMessages()
        except modules.platform.PlatformNotFound:
            plat_not_found = True
            break
        except:
            logging.warning('Unhandled exception from task {name}'.format(name=task_name))
            logging.warning('Run with --log-level=debug for exception details.')
            logging.debug('{err}'.format(err=traceback.format_exc()))
            pass
        task_num += 1
        percent = (task_num/health_summary_stream.stream_obj.totalChecks)*100
        if progress_bar:
            progress_bar.updateProgress(percent)

    if not plat_not_found:
        # we can use module tasks here as the run is already completed
        health_summary_stream.stream(modules.system.system_summary.getResult())
        temp_result += modules.system.system_summary.getResult()
    health_summary_stream.stream("\nHealth Summary\n--------------")
    temp_result += '\n{}\n'.format(health_summary_stream.stream_obj.message)

    total_checks = sum(health_result.values())
    for key in health_result.keys():
        health_summary_stream.stream((config.health_check_result_str).format(health_result[key], total_checks, key))
        temp_result += '\n{}'.format(health_summary_stream.stream_obj.message)
    if plat_not_found:
        health_summary_stream.stream_obj.result = 'unhealthy'
        t = tasks.RunCommand(['bash -c "cat /sys/devices/virtual/dmi/id/sys_vendor"'], shell=True)
        t.run()
        system_vendor = t.getOutput().strip()
        error_msg = f'''
Unknown system vendor "{system_vendor}"
nvsm health could not determine system SKU

Please ensure that NVSM is running on a supported NVIDIA system.
'''
        health_summary_stream.stream_obj.message = error_msg
    elif ((health_result['unhealthy'] > 0) or (health_result['unknown'] > 0)):
        health_summary_stream.stream_obj.result = 'unhealthy'
        health_summary_stream.stream_obj.message = (config.health_status_str).format('unhealthy')
    else:
        health_summary_stream.stream_obj.result = 'healthy'
        health_summary_stream.stream_obj.message = (config.health_status_str).format('healthy')

    health_summary_stream.stream_obj.totalChecks       = total_checks
    health_summary_stream.stream_obj.healthyChecks     = health_result['healthy']
    health_summary_stream.stream_obj.unhealthyChecks   = health_result['unhealthy']
    health_summary_stream.stream_obj.unknownChecks     = health_result['unknown']
    health_summary_stream.stream_obj.informationChecks = health_result['informational']

    temp_result += '\n{}'.format(health_summary_stream.stream_obj.message)
    temp_data['results'] = temp_result
    # For unhealthy system, write nvsm health checks result into tempfile
    if health_result['unhealthy'] or health_result['unknown'] or plat_not_found:
        try:
            with open(temp_file, 'w') as outfile:
                json.dump(temp_data, outfile, cls=CustomEncoder)
            health_summary_stream.stream_obj.message += (config.nvsm_health_unhealthy_message).format(temp_file)
            temp_result += '\n{}'.format((config.nvsm_health_unhealthy_message).format(temp_file))
        except:
            logging.warning('Error while writing nvsm health checks result to JSON file')
    # Write the results output to the file
    if output_file:
        try:
            with open(output_file, 'w') as outfile:
                outfile.write(temp_result)
        except Exception as e:
            logging.warning('Error while writing nvsm health checks output: {}'.format(e))

    health_summary_stream.stream()

    return health_summary_stream.stream_obj.result

execute_tasks.__order = None
