# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import os
import io
import logging
import tarfile
import time
import traceback
from threading import Thread

from nvsmhealth.collect import CollectionContext
from nvsmhealth.discovery import discover_tasks

class __SosReportTaskProgress(object):
    def __init__(self, current_progress, max_progress, timeout, progress_bar):
        self.__current_progress = current_progress
        self.__max_progress = max_progress
        self.__timeout = timeout
        self.__sleep_time = 1 # progress ticker - 1secs
        self.__finished = False
        self.__equated_progress = 0
        self.__progress_bar = progress_bar

    def __enter__(self):
        if not self.__progress_bar:
            return self
        # For SOSReport progress calculation divide current progress to max progress into
        # equal fragments spread over the SOSReport task timeout
        self.__equated_progress = (self.__max_progress - self.__current_progress) / (self.__timeout / self.__sleep_time)
        def loop(s=self):
            progress = s.__current_progress
            while not s.__finished:
                s.__progress_bar.updateProgress(progress)
                progress += s.__equated_progress
                if progress > self.__max_progress:
                    progress = self.__max_progress
                time.sleep(s.__sleep_time)

            s.__progress_bar.updateProgress(s.__max_progress)
        # Update progress status in a thread, to keep progress bar responsive
        self.__thread = Thread(target=loop)
        self.__thread.start()
        return self
    def __exit__(self, type, value, traceback):
        self.__finished = True
        return True

class __TaskProgress(object):
    def __init__(self, progress, progress_bar):
        self.__progress = progress
        self.__progress_bar = progress_bar
        self.__sleep_time = 1
        self.__finished = False

    def __enter__(self):
        if not self.__progress_bar:
            return self
        # Wait for a moment, till earlier task progress is over
        time.sleep(self.__sleep_time)
        def loop(s=self):
            # For generic tasks, update the progress calculated on tasks order execution
            while not s.__finished:
                s.__progress_bar.updateProgress(s.__progress)
                time.sleep(s.__sleep_time)
        # Update status stream in a thread, to keep progress bar responsive
        self.__thread = Thread(target=loop)
        self.__thread.start()
        return self

    def __exit__(self, type, value, traceback):
        self.__finished = True
        return True

def dump_health(output_path=None, stream_collector=None, progress_bar=None):
    # SOSReport individual task consumes nearly 50% of overall health dump
    # Stream progress for SOSReport task accordingly
    sosreport_progress_weight = 50

    _, task_graph = discover_tasks()
    # Health dump collects output from ICommand and ILogs tasks
    # which defines 'collect' method for log or command output collection.
    # Only run tasks with collect method defined and not having predicate clause
    tasks = [task for task in task_graph.getOrder() if hasattr(task, 'collect') and task.checkPredicate()]
    # Avoid zero division issue
    tasks_len = len(tasks) if len(tasks) else 1
    # Calculate progress contributed by individual tasks and SOSReport task explicitly
    calibrated_progress = (100 - sosreport_progress_weight) / tasks_len
    progress = 0

    # The collection context provides the temporary directory and output
    # tarball used for data collection
    with CollectionContext(
            output_path=output_path,
            stream_collector=stream_collector) as context:
        if progress_bar:
            progress_bar.updateProgress(0.0)
        # Run the entire task graph and dump all output to collection context
        for task in tasks:
            with context.useTask(task):
                try:
                    if task.getName() == "sosreport":
                        with __SosReportTaskProgress(
                            current_progress=progress,
                            max_progress=(progress + sosreport_progress_weight),
                            timeout=task.getTimeout(),
                            progress_bar=progress_bar):

                            task.run()
                            task.collect(context=context)
                        progress += sosreport_progress_weight
                    else:
                        with __TaskProgress(
                            progress=progress,
                            progress_bar=progress_bar):

                            try:
                                timeout = int(os.getenv('TASK_TIMEOUT')) * 60
                            except:
                                timeout = None
                            if timeout:
                                task.run(timeout=timeout)
                            else:
                                task.run()
                            task.collect(context=context)
                        progress += calibrated_progress

                except:
                    logging.warning('Unhandled exception from task {name}'.format(name=task.getName()))
                    logging.warning('Run with --log-level=debug for exception details.')
                    logging.debug('{err}'.format(err=traceback.format_exc()))
