#!/usr/bin/env python3
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import time
import os
import sys
import signal
import logging
import json
import io
from functools import partial

from nvsmhealth.tasks.stream import HealthSummaryStream, HealthTaskStream

logger = logging.getLogger(__name__)

def on_connect(client, userdata, flags, rc):
    """
        On Connect callback registered with MQTT Client
        This callback is called each time the client connects/reconnects to the broker
    """
    if rc == 0:
        logger.debug("MQTT client connected successfully")
    else:
        logger.debug("MQTT client connection failed, with ret-code:{}".format(str(rc)))

def get_mqtt_client():
    from nvsm_dshm.mqclient.mqclient import MQBrokerClient as MQBroker
    from nvsm_dshm.infra.err_code import ErrCode as Err
    from nvsm_dshm.infra import config_settings

    try:
        config_settings(config_file_name='/etc/nvsm/dshm/nvsm_dshm.cfg')
    except:
        logging.debug("Failed to load config file")

    # disable logs from mqtt lib as this is optional for nvsmhealth
    logging.getLogger("nvsm_dshm.mqclient.mqclient").setLevel(logging.CRITICAL)
    mq_client = MQBroker.factory(MQBroker.MQTT_CLIENT, str(os.getpid()), on_connect, None)
    if mq_client is None or \
             mq_client is Err.ERR_INVLAID_MQ_SECURITY_CERT or \
             mq_client is Err.ERR_INVLAID_MQ_ADDR_PORT:
         logging.debug('Failed in creating the MQBroker Client')
         return None
    return mq_client

prev_msg = None
def publish_cb(output_path, progress_bar, topic, obj):
    json_obj = json.dumps(obj, default=lambda o: o.__dict__, sort_keys=True)
    try:
        mq_client = get_mqtt_client()
        if json_obj != None and mq_client:
            mq_client.client.publish(topic, json_obj)
    except Exception as e:
        logging.debug(e)
    if obj.message:
        if output_path:
            # Do nothing - Results are pre-collected in execute tasks.
            # Dump the results to the output file in single write.
            pass
        else:
            global prev_msg
            if prev_msg != obj.message:
                # log only new messages on screen
                if progress_bar:
                    progress_bar.clearLine()
                print(obj.message)
                prev_msg = obj.message
    if progress_bar:
        progress_bar.updateProgress()
    try:
        mq_client.disconnect()
    except:
        pass


# return 0 - if success
def result_exit_code(result):
    if any(a == result.lower() for a in ["success", "healthy"]):
        return 0
    return 1

def progress_complete(status, progress_bar, output_path=None):
    if progress_bar:
        progress_bar.updateProgress(100.0)
        # For output path don't print status
        if output_path:
            print('\n')
            return
        print('\nStatus: ', end='')
        import nvsm.color as colorizer
        colorizer.init('always')
        if not result_exit_code(status):
            color = 'green'
        else:
            color = 'red'
        with colorizer.colorize(color):
            print(status)

def health(output_path=None, progress_bar=None):
    # log start time
    import nvsmhealth.config as config
    import time
    config.timestamp = f'{time.strftime("%c")}'

    if output_path:
        output_path = os.path.realpath(output_path)
        print(f"Writing health output to {output_path}")

    # Execute all health checks and pretty print the results
    from nvsmhealth.discovery import execute_tasks
    result = execute_tasks(publish_cb=partial(publish_cb, output_path, progress_bar),
                           progress_bar=progress_bar, output_file=output_path)
    progress_complete(result, progress_bar, output_path)
    return result_exit_code(result)

def dump(output_path=None, progress_bar=None):

    # log start time
    import nvsmhealth.config as config
    import time
    config.timestamp = f'{time.strftime("%c")}'

    # Stream health dump progress
    health_dump_stream = HealthSummaryStream(partial(publish_cb, None, progress_bar))
    health_dump_stream.stream_obj.totalChecks = 1
    health_dump_stream.stream()
    health_dump_stream.stream('Health dump started')

    # Dump health while streaming progress
    from nvsmhealth.dump import dump_health
    try:
        dump_health(output_path=output_path,
                    stream_collector=health_dump_stream,
                    progress_bar=progress_bar)
        result = "success"
    except Exception as e:
        result = str(e)

    health_dump_stream.stream_obj.result = result
    health_dump_stream.stream('Health dump completed')

    progress_complete(result, progress_bar)
    return 0
