#!/usr/bin/python

# Copyright (c) 2017-2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from __future__ import unicode_literals

import os
import re
import sys

from nvsmhealth.lib.deepdiff import DeepDiff

class DictionarySuperset():
    def __init__(self, missing_message=None, type_message=None, changed_message=None):
        super(DictionarySuperset, self).__init__()

        self.missing_message = missing_message
        self.type_message = type_message
        self.changed_message = changed_message

        # Optional field: missing_message
        if self.missing_message == None:
            self.missing_message = '"{}" is missing'

        # Optional field: type_message
        if self.type_message == None:
            self.type_message = '"{}" is the wrong type'

        # Optional field: changed_message
        if self.changed_message == None:
            self.changed_message = '"{}" has value "{}" when "{}" was expected'


    def compare(self, ref_dict, expected_dict, additional_message_key=None):
        """Verify whether the reference dictionary is a proper superset
        of the input dictionary. Uses the third party DeepDiff module
        (https://github.com/seperman/deepdiff) - MIT license.

        Since the number of comparisons may be large, only differences
        are included in the return message. Differences may be items
        missing or corresponding keys with different values.

        Note that order is important if the input dictionary contains
        lists. In that case, the reference dictionary must contain
        placeholders for each list item. The DeepDiff 'ignore_order'
        parameter changes the behavior in an unacceptable way so it
        cannot be used.
        """

        ddiff = DeepDiff(ref_dict, expected_dict)

        if any(key in ddiff for key in ['dictionary_item_added',
                                        'type_changes', 'values_changed']):

            message = ''

            if 'dictionary_item_added' in ddiff:
                for item in ddiff['dictionary_item_added']:
                    key = re.findall('\[\'(.*?)\'\]', item)
                    message += '\n'
                    message += self.missing_message.format(' -> '.join(key))

            if 'type_changes' in ddiff:
                for key, value in ddiff['type_changes'].items():
                    key = re.findall('\[\'(.*?)\'\]', key)
                    message += '\n'
                    message += self.type_message.format(' -> '.join(key))

            if 'values_changed' in ddiff:
                for key, value in ddiff['values_changed'].items():
                    key = re.findall('\[\'(.*?)\'\]', key)
                    message += '\n'
                    # Best effort to add additional_message_key information
                    if additional_message_key:
                        try:
                            message += 'For {} with '.format(ref_dict[key[0]][additional_message_key])
                        except:
                            pass
                    message += self.changed_message.format(
                            ' -> '.join(key),
                            value['old_value'],
                            value['new_value'])

            return message.strip()

        else:
            return None

