# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from nvsmhealth import modules, tasks, config
from nvsmhealth.lib.xmltodict import xmltodict

import json

__all__ = [
    'check_blacklist_recommendations',
    'check_instant_blacklist_recommendations'
]

class CheckBlacklistRecommendations(tasks.ICheck):
    def __init__(self):
        super(CheckBlacklistRecommendations, self).__init__()
        # Declare dependency on parameters
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.depends(modules.kvm.kvm_mode_on)

    def run(self):

        own_params = self.getParameters()
        params = self.__parameters_task.getResult()

        check_blacklist_recommendations = False
        for param, req in own_params.items():
            if param in params.keys():
                check_blacklist_recommendations = params[param]

        # Return if parameter is not defined
        if not check_blacklist_recommendations:
            return

        # check if kvm mode is on
        if modules.kvm.kvm_mode_on.getResult() == True:
            self.addCheckMessage('KVM mode is on, skipping blacklist recommendations check.')
            self.addInformational()
            return

        # Run the blacklist_recommendations task
        args = ['./modules/blacklist_recommendations/_gpu_blacklist_recommendations', \
            '--detect', '--watches']
        args.extend(check_blacklist_recommendations)
        collect_task = tasks.RunCommand(args=args, timeout=1000)
        collect_task.run()

        # For DCGM failures, GPU blacklist recommendations can
        # exit with returncode 1, handle it gracefully
        # Return, for no respone or exitcode > 1
        if (collect_task.getReturnCode() > 1) or not collect_task.getOutput():
            self.addCheckMessage('No response or error while running GPU blacklist recommendations: {}'.format(
                                    collect_task.getError()))
            self.addUnknown()
            return

        healthy = True

        try:
            result = json.loads(collect_task.getOutput())
            blacklist = result.get('blacklistedGpus', {})
            # Check for GPU/NVSwitch blacklist recommendations
            if len(blacklist) > 0:
                healthy = False
                self.addCheckMessage('Found {count} device(s) recommended for blacklist:'.format(
                    count=len(blacklist)))
            else:
                self.addCheckMessage('No devices found recommended for blacklist.')

            for entity_id in sorted(blacklist.keys()):
                details = blacklist[entity_id]
                device_uuid = details.get('UUID')
                device_bdf = details.get('BDF')
                failure_explanation = details.get('Failure Explanation')
                self.addCheckMessage('\t"GPU{entity_id}":\n' \
                                     '\t"BDF": "{device_bdf}"\n' \
                                     '\t"UUID": "{device_uuid}"\n' \
                                     '\t"Failure Explanation": {failure_explanation}'.format(
                                     entity_id=entity_id,
                                     device_bdf=device_bdf,
                                     device_uuid=device_uuid,
                                     failure_explanation=failure_explanation))
            # Check for other errors in blacklist recommendation script
            error_list = result.get('errors', [])
            if error_list:
                nv_hostengine_running = True
                self.addCheckMessage('Errors encountered:')
                for e in error_list:
                    if 'host engine is not valid any longer' in e:
                        nv_hostengine_running = False
                    self.addCheckMessage('\t{}'.format(e))

                # If nv-hostengine is not running return as unknown
                if not nv_hostengine_running:
                    self.addUnknown()
                    return

                healthy = False

        except Exception as e:
            self.addCheckMessage('Error while parsing GPU blacklist recommendations: {}'.format(e.message))
            self.addUnknown()
            return

        # make sure SBE page pending retirements are caught as informational,
        # as the blacklist_recommendations script ignores them as warnings
        if healthy:
            nvidia_smi_res = modules.nvidia_smi.parse_nvidia_smi.getResult()
            if nvidia_smi_res:
                for gpu, info in nvidia_smi_res.items():
                    gpu_dict = xmltodict.parse(info)
                    check_cls = modules.nvidia_smi.GpuCheckRetiredPagesPending(gpu, gpu_dict)
                    check_cls.setCallback(self.getCallback())
                    check_cls.run()
                    bad_count = check_cls.getResult()['unhealthy'] + check_cls.getResult()['unknown']
                    if bad_count:
                        healthy = False
                        self.addCheckMessage(check_cls.getTitle() )
                if not healthy:
                    self.addCheckMessage(config.gpu_total_retired_pages_pending_error)
                    self.addInformational()
                    return

        if healthy:
            self.addHealthy()
        else:
            self.addUnHealthy()


check_blacklist_recommendations = CheckBlacklistRecommendations() \
    .title(config.blacklist_recommendations_info_str) \
    .parameters('dcgm_watch_type') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(modules.nvidia_smi.parse_nvidia_smi) \
    .predicate(lambda: modules.parameter.parameters.getResult().get('dcgm_watch_type'))

check_instant_blacklist_recommendations = CheckBlacklistRecommendations() \
    .title(config.blacklist_recommendations_instant_info_str) \
    .parameters('dcgm_instant_watch_type') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(modules.nvidia_smi.parse_nvidia_smi) \
    .predicate(lambda: modules.parameter.parameters.getResult().get('dcgm_instant_watch_type'))

