# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""Checks for hardware bill-of-materials, also known as BOM.  For a given
platform configuration, the expected bill-of-materials is checked against the
observed output of the lspci command. The platform configuration is usually
determined based on product name observed in the DMI table or IPMI FRU.
"""

import sys
from nvsmhealth import modules, tasks, config
from nvsmhealth.lib import DictionarySuperset
from nvsmhealth.lib.deepdiff import DeepDiff
import json
import re
from copy import deepcopy

__all__ = [
    'check_bom_gpus',
    'check_bom_vgas',
    'check_bom_dimms',
    'check_bom_ib_controllers',
    'check_bom_pcie_switches',
    'check_bom_nvswitch',
    'check_bom_ethernet_controllers',
    'check_bom_disk_controllers',
    'check_ib_controller_link_info',
    'check_gpu_link_info',
    'check_nvme_link_info',
    'check_ethernet_controller_info'
]

def generic_bom_check(task, parameters_task, parse_task, \
    info_str, command_str, missing_str=None, changed_str=None):
    # Get parameters from own task
    own_params = task.getParameters()

    # Get parameters from parameter task
    params = parameters_task.getResult()

    bom_config = None
    for param, req in own_params.items():
        if param in params.keys():
           bom_config = params[param]
    # If parameter is not found in platform parameters - Do Nothing
    if bom_config == None:
        return
    # Print/Stream Task info and command messages
    task.title(info_str)
    task.addCheckMessage(command_str)

    if parse_task.getResult() != None:
        # Compare task output with expected config
        res = parse_task.getResult()
        if type(res) is dict:
            out_dict = res
        else:
            out_dict = json.loads(res)
        dictionary_superset = DictionarySuperset.DictionarySuperset( \
                missing_message=missing_str, \
                changed_message=changed_str)
        result = dictionary_superset.compare(out_dict, bom_config, additional_message_key='device')
        if result == None:
            # Healthy check - No diffs found in bill-of-materials
            task.addHealthy()
        else:
            # UnHealthy check - Print/Stream diffs found in bill-of-materials
            task.addUnHealthy()
            task.addCheckMessage(result)
    return

class CheckBomGpus(tasks.ICheck):
    def __init__(self, parse_lspci_task, parameters_task):
        super(CheckBomGpus, self).__init__()

        # Declare dependency on parse lspci task using the ITask interface
        self.__parse_lspci_task = parse_lspci_task
        self.__parameters_task = parameters_task
        self.depends(parse_lspci_task)
        self.depends(parameters_task)

    def run(self):
        generic_bom_check(self, self.__parameters_task, self.__parse_lspci_task, \
            config.gpus_info_str, config.gpus_command_str, \
            config.gpus_pci_device_missing_str, config.pci_device_changed_str)

check_bom_gpus = CheckBomGpus(modules.lspci.parse_lspci, modules.parameter.parameters) \
    .title('''Check GPUs PCIe devices information for consistency''') \
    .parameters('gpu_bom')


class CheckBomVgas(tasks.ICheck):
    def __init__(self, parse_lspci_task, parameters_task):
        super(CheckBomVgas, self).__init__()

        # Declare dependency on parse lspci task using the ITask interface
        self.__parse_lspci_task = parse_lspci_task
        self.__parameters_task = parameters_task
        self.depends(parse_lspci_task)
        self.depends(parameters_task)

    def run(self):
        generic_bom_check(self, self.__parameters_task, self.__parse_lspci_task, \
            config.vgas_info_str, config.vgas_command_str, \
            config.vgas_pci_device_missing_str, config.pci_device_changed_str)

check_bom_vgas = CheckBomVgas(modules.lspci.parse_lspci, modules.parameter.parameters) \
    .title('''Check VGA Controller PCIe devices information for consistency''') \
    .parameters('vga_bom')

class CheckBomDimms(tasks.ICheck):
    def __init__(self, parse_dmidecode_task, parameters_task):
        super(CheckBomDimms, self).__init__()

        # Declare dependency on parse dmidecode task using the ITask interface
        self.__parse_dmidecode_task = parse_dmidecode_task
        self.__parameters_task = parameters_task
        self.depends(parse_dmidecode_task)
        self.depends(parameters_task)

    def run(self):
        generic_bom_check(self, self.__parameters_task, self.__parse_dmidecode_task, \
            config.dimms_info_str, config.dimms_command_str)

check_bom_dimms = CheckBomDimms(modules.dmidecode.parse_dmidecode, modules.parameter.parameters) \
    .title('''Check Memory DIMMs devices information for consistency''') \
    .parameters('dimm_bom')

class CheckBomIbControllers(tasks.ICheck):
    def __init__(self, parse_lspci_task, parameters_task):
        super(CheckBomIbControllers, self).__init__()

        # Declare dependency on parse lspci task using the ITask interface
        self.__parse_lspci_task = parse_lspci_task
        self.__parameters_task = parameters_task
        self.depends(parse_lspci_task)
        self.depends(parameters_task)

    def run(self):
        generic_bom_check(self, self.__parameters_task, self.__parse_lspci_task, \
            config.ib_controllers_info_str, config.ib_controllers_command_str, \
            config.ib_controllers_pci_device_missing_str, config.pci_device_changed_str)

check_bom_ib_controllers = CheckBomIbControllers(modules.lspci.parse_lspci, modules.parameter.parameters) \
    .title('''Check Infiband controllers PCIe devices information for consistency''') \
    .parameters('ib_controllers_bom')

class CheckBomPcieSwitches(tasks.ICheck):
    def __init__(self, parse_lspci_task, parameters_task):
        super(CheckBomPcieSwitches, self).__init__()

        # Declare dependency on parse lspci task using the ITask interface
        self.__parse_lspci_task = parse_lspci_task
        self.__parameters_task = parameters_task
        self.depends(parse_lspci_task)
        self.depends(parameters_task)

    def run(self):
        generic_bom_check(self, self.__parameters_task, self.__parse_lspci_task, \
            config.pcie_switches_info_str, config.pcie_switches_command_str, \
            config.pcie_switches_pci_device_missing_str, config.pci_device_changed_str)

check_bom_pcie_switches = CheckBomPcieSwitches(modules.lspci.parse_lspci_n, modules.parameter.parameters) \
    .title('''Check PCIe Switches PCIe devices information for consistency''') \
    .parameters('pcie_switches_bom')

class CheckBomNvswitch(tasks.ICheck):
    def __init__(self, parse_lspci_task, parameters_task):
        super(CheckBomNvswitch, self).__init__()

        # Declare dependency on parse lspci task using the ITask interface
        self.__parse_lspci_task = parse_lspci_task
        self.__parameters_task = parameters_task
        self.depends(parse_lspci_task)
        self.depends(parameters_task)

    def run(self):
        generic_bom_check(self, self.__parameters_task, self.__parse_lspci_task, \
            config.nvswitch_info_str, config.nvswitch_command_str, \
            config.nvswitch_pci_device_missing_str, config.pci_device_changed_str)

check_bom_nvswitch = CheckBomNvswitch(modules.lspci.parse_lspci, modules.parameter.parameters) \
    .title('''Check NVSwitch controller PCIe devices information for consistency''') \
    .parameters('nvswitch_bom')

class CheckBomEthernetControllers(tasks.ICheck):
    def __init__(self, parse_lspci_task, parameters_task):
        super(CheckBomEthernetControllers, self).__init__()

        # Declare dependency on parse lspci task using the ITask interface
        self.__parse_lspci_task = parse_lspci_task
        self.__parameters_task = parameters_task
        self.depends(parse_lspci_task)
        self.depends(parameters_task)

    def run(self):
        # Get parameters from own task
        own_params = self.getParameters()

        # Get parameters from parameter task
        params = self.__parameters_task.getResult()

        bom_config = None
        for param, req in own_params.items():
            if param in params.keys():
               bom_config = params[param]
        # If parameter is not found in platform parameters - Do Nothing
        if bom_config == None:
            return
        # Print/Stream Task info and command messages
        self.title(config.ethernet_controllers_info_str)
        self.addCheckMessage(config.ethernet_controllers_command_str)

        if self.__parse_lspci_task.getResult() != None:
            # Compare task output with expected config
            res = self.__parse_lspci_task.getResult()
            if type(res) is dict:
                out_dict = res
            else:
                out_dict = json.loads(res)

            # dictionary compare
            ddiff = DeepDiff(out_dict, bom_config)

            message = ''
            result = 'Healthy'
            if any(key in ddiff for key in ['dictionary_item_added',
                                            'values_changed']):
                if 'dictionary_item_added' in ddiff:
                    for item in ddiff['dictionary_item_added']:
                        key = re.findall('\[\'(.*?)\'\]', item)
                        message += '\n'
                        message += config.ethernet_controllers_pci_device_missing_str.format(' -> '.join(key))
                    result = 'UnHealthy'

                if 'values_changed' in ddiff:
                    for key, value in ddiff['values_changed'].items():
                        key = re.findall('\[\'(.*?)\'\]', key)
                        message += '\n'
                        # Best effort to add additional_message_key information
                        try:
                            message += 'For {} with '.format(ref_dict[key[0]]['device'])
                        except:
                            pass
                        message += config.pci_device_changed_str.format(
                                ' -> '.join(key),
                                value['old_value'],
                                value['new_value'])
                    if result == 'Healthy':
                        result = 'Informational'
            self.addCheckMessage(message)

            if result == 'Healthy':
                # Healthy check - No diffs found in bill-of-materials
                self.addHealthy()
            elif result == 'UnHealthy':
                # UnHealthy check - Print/Stream diffs found in bill-of-materials
                self.addUnHealthy()
            elif result == 'Informational':
                # Informational status - print change in config
                self.addInformational()
            else:
                self.addUnknown()

check_bom_ethernet_controllers = CheckBomEthernetControllers(modules.lspci.parse_lspci, modules.parameter.parameters) \
    .title('''Check Ethernet Controllers PCIe devices information for consistency''') \
    .parameters('ethernet_controllers_bom')

class CheckBomDiskControllers(tasks.ICheck):
    def __init__(self, parse_lspci_task, parameters_task):
        super(CheckBomDiskControllers, self).__init__()

        # Declare dependency on parse lspci task using the ITask interface
        self.__parse_lspci_task = parse_lspci_task
        self.__parameters_task = parameters_task
        self.depends(parse_lspci_task)
        self.depends(parameters_task)

    def run(self):
        generic_bom_check(self, self.__parameters_task, self.__parse_lspci_task, \
            config.disk_controllers_info_str, config.disk_controllers_command_str, \
            config.disk_controllers_pci_device_missing_str, config.pci_device_changed_str)

check_bom_disk_controllers = CheckBomDiskControllers(modules.lspci.parse_lspci, modules.parameter.parameters) \
    .title('''Check Disk Controllers PCIe devices information for consistency''') \
    .parameters('disk_controllers_bom')

class CheckLink(tasks.ICheck):
    def __init__(self, device_id, config, link_type, parse_lspci_task, parameter=""):
        super().__init__()
        self.__parse_lspci_task = parse_lspci_task
        self.__config = config
        self.__link_type = link_type
        self.__device_id = device_id
        self.__parameter = parameter
        super().name("CheckLink")

    def parameters(self, param, required=False):
        assert False, "Do not add parameter dependency on this class"

    def depends(self, task):
        assert False, "Do not add task dependency on this class"

    def run(self):
        if not self.__parse_lspci_task.getResult():
            return

        res = self.__parse_lspci_task.getResult()
        if type(res) is dict:
            out_dict = res
        else:
            out_dict = json.loads(res)
        # Compare task output with expected config
        if self.__link_type == 'width':
            pattern = re.compile(r"^\tCapabilities:.*\n(\t\t.*\n)*\t\tLnkSta:.*Width x(?P<link_width>\d+)", flags=re.MULTILINE)
        elif self.__link_type == 'speed':
            pattern = re.compile(r"^\tCapabilities:.*\n(\t\t.*\n)*\t\tLnkSta:.*Speed (?P<link_speed>[^(?:,|\s)]+)", flags=re.MULTILINE)
        else:
            return
        if self.__device_id not in out_dict:
            nvsm_config = config.read_nvsm_config_file()
            if self.__parameter ==  "nvme_link_info" and nvsm_config != None and not nvsm_config["use_standard_config_storage"]:
                self.addCheckMessage('"use_standard_config_storage" is set to false in "/etc/nvsm/nvsm.config" file, skipping check.')
                self.addInformational()
            else:
                self.addCheckMessage("Device not present in 'lspci' output")
                self.addUnknown()
            # Device not present; nothing to check
            return
        if 'body' not in out_dict[self.__device_id]:
            # Could not parse device entry in lspci
            self.addUnknown()
            return
        match = pattern.search(out_dict[self.__device_id]['body'])
        if not match:
            # Could not parse device entry in lspci
            self.addUnknown()
            return
        res = match.groupdict()['link_{}'.format(self.__link_type)]
        if str(self.__config) == res:
            # Healthy check - No diffs found in bill-of-materials
            self.addHealthy()
        else:
            # UnHealthy check - Print/Stream diffs found in bill-of-materials
            self.addUnHealthy()

class CheckLinkInfo(tasks.ICheck):
    def __init__(self, list_parameter, title_str, is_gpu_check=False):
        super().__init__()

        self.__parameter = list_parameter
        self.parameters(self.__parameter)
        self.__title_str = title_str
        self.__is_gpu_check = is_gpu_check
        # Declare dependency on parse lspci task using the ITask interface
        self.__parse_lspci_task = modules.lspci.parse_lspci_verbose
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parse_lspci_task)
        self.depends(self.__parameters_task)
        self.__bdf_pstate = modules.nvidia_smi.gpu_bdf_pstate
        self.depends(self.__bdf_pstate)
        self.depends(modules.kvm.kvm_mode_on)

    def run(self):
        try:
            devices = self.__parameters_task.getResult()[self.__parameter]
        except:
            return
        for device in deepcopy(devices):
            pstate_not_found = False
            if self.__is_gpu_check:
                bdf_pstate = self.__bdf_pstate.getResult()
                try:
                    pstate = bdf_pstate[device['bdf']]
                    device['speed'] = device['speed'][pstate]
                    device['width'] = device['width'][pstate]
                except:
                    # not able to find the pstate for this gpu bdf
                    pstate_not_found = True
            kvm_mode_disabled = True
            if self.__parameter== "gpu_link_info"  and modules.kvm.kvm_mode_on.getResult() == True:
                kvm_mode_disabled = False
            for check_type in ['speed', 'width']:

                check_cls = CheckLink(device['bdf'], device[check_type], check_type, self.__parse_lspci_task, self.__parameter)
                if 'name' not in device:
                    device["name"] = ""
                check_cls.setCallback(self.getCallback())
                if kvm_mode_disabled:
                    if pstate_not_found:
                        device['speed'] = "None"
                        device['width'] = "None"
                        check_cls.addCheckMessage("unknown pstate for the GPU[{}]".format(device['bdf']))
                        check_cls.addUnknown()
                    else:
                        check_cls.run()
                else:
                    device['speed'] = "None"
                    device['width'] = "None"
                    check_cls.addCheckMessage('KVM mode is on, skipping check.')
                    check_cls.addInformational()
                    self.addInformational()

                check_cls.title(self.__title_str[check_type].format(**device).strip())
                if kvm_mode_disabled:
                    self.addHealthy(count=check_cls.getResult()['healthy'])
                    self.addUnHealthy(count=check_cls.getResult()['unhealthy'])
                    self.addUnknown(count=check_cls.getResult()['unknown'])
                    self.addInformational(count=check_cls.getResult()['informational'])
                check_cls.sendComplete()
                super().addMessages(check_cls.getMessages())

        # clear message as this task doesnt print anything
        self.title('')

check_gpu_link_info = CheckLinkInfo('gpu_link_info', \
                        config.gpu_link_info_str, \
                        True)
check_nvme_link_info = CheckLinkInfo('nvme_link_info', \
                        config.nvme_link_info_str)
check_ethernet_controller_info = CheckLinkInfo('ethernet_controller_info', \
                        config.ethernet_controller_link_info_str)

check_ib_controller_link_info = CheckLinkInfo('ib_controller_link_info', \
                        config.ib_controller_link_info_str)

