import re
import json
from nvsmhealth import modules, tasks, config

__all__ = [
    'run_cec_version',
    'run_cec_boot_status',
    'run_bmc_boot_slot_task',
    'parse_cec_info'
]

run_cec_version = tasks.RunCommand(
    args=['ipmitool', 'raw', '0x3C', '0xF', '0x9']) \
    .title('Run ipmitool raw 0x3C 0xF 0x9') \
    .describe('''
                Get CEC version.
                This command requires superuser privileges.''') \
    .depends(modules.common.check_superuser_privileges)

run_cec_boot_status = tasks.RunCommand(
    args=['ipmitool', 'raw', '0x3C', '0x68', '0x00']) \
    .title('Run ipmitool raw 0x3C 0x68 0x00') \
    .describe('''
                Get boot status.
                This command requires superuser privileges.''') \
    .depends(modules.common.check_superuser_privileges)

run_bmc_boot_slot_task = tasks.RunCommand(
    args=['ipmitool', 'raw', '0x3C', '0x3', '0x0']) \
    .title('Run ipmitool raw 0x3C 0x3 0x0') \
    .describe('''
                Get bmc boot slot.
                This command requires superuser privileges.''') \
    .depends(modules.common.check_superuser_privileges)

class ParseCECInfo(tasks.ITask):
    def __init__(self, cec_version_task, cec_boot_status_task, bmc_boot_slot_task):
        super(ParseCECInfo, self).__init__()
        self.__cec_version_task = cec_version_task
        self.__cec_boot_status_task = cec_boot_status_task
        self.__bmc_boot_slot_task = bmc_boot_slot_task
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        if self.__cec_version_task.getReturnCode():
            # command failed
            return
        res = {}
        output = self.__cec_version_task.getOutput().strip()
        output_bytes = output.split()
        if len(output_bytes) != 3:
            # Wrong output
            res['cec_version'] = "Not available"
        else:
            res['cec_version'] = "{}.{}".format(int(output_bytes[1].strip(), 16), int(output_bytes[2].strip(), 16))

        if self.__cec_boot_status_task.getReturnCode():
            # command failed
            res['EC_FW_TAG0'] = 'Not Authenticated'
            res['EC_FW_TAG1'] = 'Not Authenticated'
            res['bmc_fw_auth_state'] = 'Not Authenticated'
        else:
            output = self.__cec_boot_status_task.getOutput().strip()
            res['EC_FW_TAG0'] =  'Not Authenticated' if int(output[3])&0x01  == 0 else "Authenticated"
            res['EC_FW_TAG1'] =  'Not Authenticated' if int(output[3])&0x02  == 0 else "Authenticated"
            res['bmc_fw_auth_state'] =  'Not Authenticated' if int(output[1])&0x40  == 0 else "Authenticated"
        if self.__bmc_boot_slot_task.getReturnCode():
            # command failed
            res['bmc_boot_slot'] = "Not Available"
        else:
            output = self.__bmc_boot_slot_task.getOutput().strip()
            res['bmc_boot_slot']  =  'Slot 0' if output[0] == "00" else "Slot 1"
        self.__result = res


parse_cec_info = ParseCECInfo(run_cec_version, run_cec_boot_status, run_bmc_boot_slot_task) \
    .title('Parse output from CEC') \
    .depends(run_cec_version)
