# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import logging
from nvsmhealth import modules, tasks, config
from nvsmhealth.modules.dcs_modules.dcc_health_api import run_dcc_health_api


__all__ = [
    'check_application_config_health',
    'check_dcc_display_configuration',
    'check_dcc_display_synchronization',
    'check_dcc_can_reachability',
    'check_dcc_usb_reachability',
    'check_dcc_network_reachability',
    'check_dcc_serializer_configuration',
]
class CheckApplicationConfigHealth(tasks.ITask):
    def __init__(self, run_dcc_health_api_task):
        super().__init__()
        self.__run_dcc_health_api_task = run_dcc_health_api_task
        self.__output = None

    def getOutput(self):
        return self.__output

    def run(self):
        json_output = self.__run_dcc_health_api_task.getOutput()
        if json_output != None:
            try:
                self.__output = json_output['application_config_health']
                #print(self.__output)
            except:
                self.__output = None
                #self.addUnknown()
                return
        #self.addHealthy()

check_application_config_health = CheckApplicationConfigHealth(run_dcc_health_api) \
    .title('''Drive Constellation: Check DCC Application config health''') \
    .depends(run_dcc_health_api) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class CheckDccDisplayConfiguration(tasks.ICheck):
     def __init__(self, check_application_config_health):
        super().__init__()
        self.__run_check_application_config_health = check_application_config_health
        self.__output = ""

     def getOutput(self):
        return self.__output

     def run(self):
        #self.addCheckMessage("Checking Application Config health")
        #import pdb
        #pdb.set_trace()
        json_output = self.__run_check_application_config_health.getOutput()
        if json_output != None:
            try:
                dcc_display_configuration = json_output['display_configuration']
                if dcc_display_configuration['test_result']  == "Healthy":
                    self.addHealthy()
                else:
                    self.addUnHealthy()
                    self.addCheckMessage(dcc_display_configuration['test_information'])
                    #self.__output(dcc_display_configuration['test_information'])

            except Exception as e:
                logging.debug("Error fetching DCC Display health info: {}".format(e))
                self.addUnknown()
                return
        self.addHealthy()

check_dcc_display_configuration = CheckDccDisplayConfiguration(check_application_config_health) \
    .title('''Drive Constellation: Check DCC Display Configuration Health''') \
    .depends(check_application_config_health) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))


class CheckDccDisplaySynchronization(tasks.ICheck):
     def __init__(self, check_application_config_health):
        super().__init__()
        self.__run_check_application_config_health = check_application_config_health
        self.__output = ""

     def getOutput(self):
        return self.__output

     def run(self):
        json_output = self.__run_check_application_config_health.getOutput()
        if json_output != None:
            try:
                dcc_display_synchronization = json_output['display_synchronization']
                if dcc_display_synchronization['test_result']  == "Healthy":
                    self.addHealthy()
                else:
                    self.addUnHealthy()
                    self.addCheckMessage(dcc_display_synchronization['test_information'])
                    #self.__output(dcc_display_configuration['test_information'])

            except Exception as e:
                print("Error fetching ethernet health info: {}".format(e))
                self.addUnknown()
                return
        self.addHealthy()

check_dcc_display_synchronization = CheckDccDisplaySynchronization(check_application_config_health) \
    .title('''Drive Constellation: Check DCC Display Synchronization Health''') \
    .depends(check_application_config_health) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class CheckDccCanReachability(tasks.ICheck):
     def __init__(self, check_application_config_health):
        super().__init__()
        self.__run_check_application_config_health = check_application_config_health
        self.__output = ""

     def getOutput(self):
        return self.__output

     def run(self):

        json_output = self.__run_check_application_config_health.getOutput()
        if json_output != None:
            try:
                dcc_can_reachability = json_output['can_reachability']
                if dcc_can_reachability['test_result']  == "Healthy":
                    self.addHealthy()
                else:
                    self.addUnHealthy()
                    self.addCheckMessage(dcc_can_reachability['test_information'])
                    #self.__output(dcc_display_configuration['test_information'])

            except Exception as e:
                logging.debug("Error fetching DCC CAN health info: {}".format(e))
                self.addUnknown()
                return
        self.addHealthy()

check_dcc_can_reachability = CheckDccCanReachability(check_application_config_health) \
    .title('''Drive Constellation: Check DCC CAN reachability Health''') \
    .depends(check_application_config_health) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class CheckDccUsbReachability(tasks.ICheck):
     def __init__(self, check_application_config_health):
        super().__init__()
        self.__run_check_application_config_health = check_application_config_health
        self.__output = ""

     def getOutput(self):
        return self.__output

     def run(self):

        json_output = self.__run_check_application_config_health.getOutput()
        if json_output != None:
            try:
                dcc_usb_reachability = json_output['usb_reachability']
                if dcc_usb_reachability['test_result']  == "Healthy":
                    self.addHealthy()
                else:
                    self.addUnHealthy()
                    self.addCheckMessage(dcc_usb_reachability['test_information'])
                    #self.__output(dcc_display_configuration['test_information'])

            except Exception as e:
                logging.debug("Error fetching DCC USB reachability health info: {}".format(e))
                self.addUnknown()
                return
        self.addHealthy()

check_dcc_usb_reachability = CheckDccUsbReachability(check_application_config_health) \
    .title('''Drive Constellation: Check DCC USB reachability Health''') \
    .depends(check_application_config_health) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class CheckDccSerializerConfiguration(tasks.ICheck):
     def __init__(self, check_application_config_health):
        super().__init__()
        self.__run_check_application_config_health = check_application_config_health
        self.__output = ""

     def getOutput(self):
        return self.__output

     def run(self):
        json_output = self.__run_check_application_config_health.getOutput()
        if json_output != None:
            try:
                dcc_serializer_configuration = json_output['serializer_configuration']
                if dcc_serializer_configuration['test_result']  == "Healthy":
                    self.addHealthy()
                else:
                    self.addUnHealthy()
                    self.addCheckMessage(dcc_serializer_configuration['test_information'])
                    #self.__output(dcc_display_configuration['test_information'])

            except Exception as e:
                logging.debug("Error fetching serializer_configuration info: {}".format(e))
                self.addUnknown()
                return
        self.addHealthy()

check_dcc_serializer_configuration = CheckDccSerializerConfiguration(check_application_config_health) \
    .title('''Drive Constellation: Check DCC Serializer Configuration Health''') \
    .depends(check_application_config_health) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class CheckDccNetworkReachability(tasks.ICheck):
     def __init__(self, check_application_config_health):
        super().__init__()
        self.__run_check_application_config_health = check_application_config_health
        self.__output = ""

     def getOutput(self):
        return self.__output

     def run(self):
        json_output = self.__run_check_application_config_health.getOutput()
        if json_output != None:
            try:
                dcc_network_reachability = json_output['network_reachability']
                if dcc_network_reachability['test_result']  == "Healthy":
                    self.addHealthy()
                else:
                    self.addUnHealthy()
                    self.addCheckMessage(dcc_network_reachability['test_information'])
                    #self.__output(dcc_display_configuration['test_information'])

            except Exception as e:
                logging.debug("Error fetching network reachability info: {}".format(e))
                self.addUnknown()
                return
        self.addHealthy()

check_dcc_network_reachability = CheckDccNetworkReachability(check_application_config_health) \
    .title('''Drive Constellation: Check DCC Network Reachability''') \
    .depends(check_application_config_health) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))
