import re
import json
from nvsmhealth import modules, tasks, config, platforms

__all__ = [
    'dcc_passgen',
    'dcv_bmc_run_ipmi_info',
    'dcv_bmc_parse_ipmi_info',
    'dcv_bmc_firmware_revision_info',
    'dcv_run_ipmi_sensor',
    'dcv_parse_ipmi_sensor',
    'dcv_check_ipmi_sensor_thresholds',
    'dcv_run_ipmi_sdr_elist',
    'dcv_parse_ipmi_sdr_elist',
    'dcv_sdr_device_bom',
    'dcv_check_fan_bom',
    'dcv_check_psu_bom',
    'dcv_run_ipmi_getenables',
    'dcv_parse_ipmi_getenables',
    'dcv_run_ipmi_fru',
    'dcv_parse_ipmi_fru',
    'dcv_check_fru_consistency'
]

dcc_passgen = tasks.RunCommand(
    args=['dcc_passgen']) \
    .title('Run dcc_passgen tool') \
    .describe('''
            Run dcc_passgen for DCC BMC.
            This command requires superuser privileges.''') \
    .depends(modules.common.check_superuser_privileges) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class DcvBmcRunIpmiInfo(tasks.ITask):
    def __init__(self, dcc_passgen_task):
        super().__init__()
        self.__dcc_passgen_task = dcc_passgen_task
        self.__result = None
        self.__ret = 1

    def getOutput(self):
        return self.__result

    def getReturnCode(self):
        return self.__ret

    def run(self):
        try:
            if(self.__dcc_passgen_task.getReturnCode()==0):
                psswd = self.__dcc_passgen_task.getOutput().strip()
                cmd = 'ipmitool -I lanplus -H {} -U {} -P {} mc info'.format(platforms.cgx.dcv_bmc_config['ip'], platforms.cgx.dcv_bmc_config['user'], psswd)
                cmd_task = tasks.RunCommand(shell=True, args=[cmd])
                cmd_task.run()
                self.__ret = cmd_task.getReturnCode()
                self.__result = cmd_task.getOutput()
        except:
            return

dcv_bmc_run_ipmi_info = DcvBmcRunIpmiInfo(dcc_passgen) \
    .title('Run ipmitool mc info command') \
    .describe('''
            Check DCC BMC status with ipmitool.
            This command requires superuser privileges.''') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(dcc_passgen) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class DcvParseIpmiInfo(tasks.ITask):
    def __init__(self, ipmi_info_task):
        super(DcvParseIpmiInfo, self).__init__()
        self.__ipmi_info_task = ipmi_info_task
        self.__result = None
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('dcs_dcv_bmc')

    def getResult(self):
        return self.__result

    def run(self):
        try:
            dcs_dcv_bmc = self.__parameters_task.getResult()['dcs_dcv_bmc']
            if not dcs_dcv_bmc:
                return
        except:
            return

        if self.__ipmi_info_task.getReturnCode():
            # command failed
            return
        res = {}
        output = self.__ipmi_info_task.getOutput()

        ipmi_info_pattern = re.compile(r"^Device ID\s*:\s*(?P<device_id>.*)\nDevice Revision\s*:\s*(?P<device_revision>.*)\nFirmware Revision\s*:\s*(?P<firmware_revision>.*)\nIPMI Version\s*:\s*(?P<ipmi_version>.*)\nManufacturer ID\s*:\s*(?P<manufacturer_id>.*)\nManufacturer Name\s*:\s*(?P<manufacturer_name>.*)\nProduct ID\s*:\s*(?P<product_id>.*)\nProduct Name\s*:\s*(?P<product_name>.*)\nDevice Available\s*:\s*(?P<device_available>.*)\nProvides Device SDRs\s*:\s*(?P<provides_device_sdrs>.*)\nAdditional Device Support\s*:\s*\n(?P<additional_device_support>(.*\n)*)Aux Firmware Rev Info\s*:\s*\n(?P<aux_firmware_rev_info>(.*\n)*)")

        match = ipmi_info_pattern.match(output)
        if match:
            res = match.groupdict()

            res['additional_device_support'] = res['additional_device_support'].split('\n')
            res['additional_device_support'] = [a.strip() for a in res['additional_device_support'] if a]

            # BUG-2547367: Use Byte 13 or aux[0] (Most significant byte of Auxiliary firmware revision information)
            # to update firmware revision number
            rev_pattern = re.compile(r"^\s*0x(?P<hex1>.*)\n\s*0x(?P<hex2>.*)\n\s*0x(?P<hex3>.*)\n\s*0x(?P<hex4>.*)")
            rev_match = rev_pattern.match(res['aux_firmware_rev_info'])
            if rev_match:
                rev = rev_match.groupdict()
                res['firmware_revision'] = '{}.{}'.format(res['firmware_revision'], rev['hex1'])

            res['aux_firmware_rev_info'] = res['aux_firmware_rev_info'].split('\n')
            res['aux_firmware_rev_info'] = [a.strip() for a in res['aux_firmware_rev_info'] if a]

        self.__result = res

dcv_bmc_parse_ipmi_info = DcvParseIpmiInfo(dcv_bmc_run_ipmi_info) \
    .title('Parse output from ipmitool mc info of DCC BMC') \
    .depends(dcv_bmc_run_ipmi_info) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class DcvFirmwareRevisionInfo(tasks.ITask):
    def __init__(self, dcv_bmc_parse_ipmi_info_task):
        super(DcvFirmwareRevisionInfo, self).__init__()
        self.__dcv_bmc_parse_ipmi_info_task = dcv_bmc_parse_ipmi_info_task
        self.__result = None
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('dcs_dcv_bmc')

    def getResult(self):
        return self.__result

    def run(self):
        try:
            dcs_dcv_bmc = self.__parameters_task.getResult()['dcs_dcv_bmc']
            if not dcs_dcv_bmc:
                return
        except:
            return

        firmware_revision = "UNKNOWN"
        try:
            firmware_revision = self.__dcv_bmc_parse_ipmi_info_task.getResult()['firmware_revision']
        except:
            pass

        #TODO: Bug[200383435] - fix the defect after platform abstraction work
        self.send(msg=firmware_revision)
        self.__result = firmware_revision

dcv_bmc_firmware_revision_info = DcvFirmwareRevisionInfo(dcv_bmc_parse_ipmi_info) \
    .title('Drive Constellation: DCC BMC Firmware Revision') \
    .depends(dcv_bmc_parse_ipmi_info) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class DcvRunIpmiSensor(tasks.ITask):
    def __init__(self, dcc_passgen_task):
        super().__init__()
        self.__dcc_passgen_task = dcc_passgen_task
        self.__result = None
        self.__ret = 1

    def getOutput(self):
        return self.__result

    def getReturnCode(self):
        return self.__ret

    def run(self):
        try:
            if(self.__dcc_passgen_task.getReturnCode()==0):
                psswd = self.__dcc_passgen_task.getOutput().strip()
                cmd = 'ipmitool -I lanplus -H {} -U {} -P {} sensor'.format(platforms.cgx.dcv_bmc_config['ip'], platforms.cgx.dcv_bmc_config['user'], psswd)
                cmd_task = tasks.RunCommand(shell=True, args=[cmd])
                cmd_task.run()
                self.__ret = cmd_task.getReturnCode()
                self.__result = cmd_task.getOutput()
        except:
            return

dcv_run_ipmi_sensor = DcvRunIpmiSensor(dcc_passgen) \
    .title('Run ipmitool mc info command') \
    .describe('''
            Check DCC BMC status with ipmitool.
            This command requires superuser privileges.''') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(dcc_passgen) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class DcvParseIpmiSensor(tasks.ITask):
    def __init__(self, ipmi_sensor_task):
        super(DcvParseIpmiSensor, self).__init__()
        self.__ipmi_sensor_task = ipmi_sensor_task
        self.__result = None
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('dcs_dcv_bmc')

    def getResult(self):
        return self.__result

    def run(self):
        try:
            dcs_dcv_bmc = self.__parameters_task.getResult()['dcs_dcv_bmc']
            if not dcs_dcv_bmc:
                return
        except:
            return

        if self.__ipmi_sensor_task.getReturnCode():
            # command failed
            return
        res = {}
        output = self.__ipmi_sensor_task.getOutput().strip()

        sensor_pattern = re.compile(
                r'^(?P<name>[^|\n]+)\|' \
                r'(?P<current_reading>[^|\n]+)\|' \
                r'(?P<type>[^|\n]+)\|' \
                r'(?P<status>[^|\n]+)\|' \
                r'(?P<lower_non_recoverable>[^|\n]+)\|' \
                r'(?P<lower_critical>[^|\n]+)\|' \
                r'(?P<lower_non_critical>[^|\n]+)\|' \
                r'(?P<upper_non_critical>[^|\n]+)\|' \
                r'(?P<upper_critical>[^|\n]+)\|' \
                r'(?P<upper_non_recoverable>[^|\n]+)' ,
                flags=re.MULTILINE)

        # Attempt to parse the table of sensor values
        sensors = []
        try:
            for m in sensor_pattern.finditer(output):
                s = {
                    'name': m.group('name').strip(),
                    'status': m.group('status').strip(),
                    'type': m.group('type').strip(),
                    'current_reading': m.group('current_reading').strip(),
                    'lower_non_recoverable': m.group('lower_non_recoverable').strip(),
                    'lower_critical': m.group('lower_critical').strip(),
                    'lower_non_critical': m.group('lower_non_critical').strip(),
                    'upper_non_recoverable': m.group('upper_non_recoverable').strip(),
                    'upper_critical': m.group('upper_critical').strip(),
                    'upper_non_critical': m.group('upper_non_critical').strip()
                }
                sensors.append(s)
        except:
            pass

        self.__result = sensors

dcv_parse_ipmi_sensor = DcvParseIpmiSensor(dcv_run_ipmi_sensor) \
    .title('Parse output from ipmitool sensor for DCC BMC') \
    .depends(dcv_run_ipmi_sensor) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class DcvCheckSensorThresholds(tasks.ICheck):
    def __init__(self, dcv_parse_ipmi_sensor_task):
        super(DcvCheckSensorThresholds, self).__init__()

        self.__dcv_parse_ipmi_sensor_task = dcv_parse_ipmi_sensor_task
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('dcs_dcv_sensor_threshold')

    def run(self):

        try:
            dcs_dcv_sensor_threshold = self.__parameters_task.getResult()['dcs_dcv_sensor_threshold']
            if not dcs_dcv_sensor_threshold:
                return
        except:
            return

        threshold_check_dispatch = {
            'lower_non_recoverable': lambda observed, threshold: threshold < observed,
            'lower_critical': lambda observed, threshold: threshold < observed,
            'upper_non_recoverable': lambda observed, threshold: observed < threshold,
            'upper_critical': lambda observed, threshold: observed < threshold
        }

        threshold_display = {
            'lower_non_recoverable': '{name}: Observed value "{observed}" ({units}) below non-recoverable lower threshold "{threshold}"',
            'lower_critical': '{name}: Observed value "{observed}" ({units}) below critical lower threshold "{threshold}"',
            'upper_non_recoverable': '{name}: Observed value "{observed}" ({units}) above non-recoverable upper threshold "{threshold}"',
            'upper_critical': '{name}: Observed value "{observed}" ({units}) above critical upper threshold "{threshold}"'
        }

        # Look for any sensor values that fall outside of critical thresholds
        healthy = True
        try:
            sensors = self.__dcv_parse_ipmi_sensor_task.getResult()
            for sensor in sensors:
                name = sensor['name']
                observed = sensor['current_reading']
                try:
                    observed = float(observed)
                except:
                    continue
                units = sensor['type']
                if units.lower() == 'discrete':
                    continue
                for field in [
                        'lower_non_recoverable',
                        'lower_critical',
                        'upper_non_recoverable',
                        'upper_critical' ]:
                    threshold = sensor.get(field)
                    try:
                        threshold = float(threshold)
                    except:
                        continue
                    check = threshold_check_dispatch[field]
                    if check(observed, threshold):
                        continue  # Observed value is within threshold
                    healthy = False
                    display = threshold_display[field]
                    self.addCheckMessage(display.format(
                            name=name,
                            observed=observed,
                            units=units,
                            threshold=threshold))
            self.addCheckMessage('Checked {count} sensor values against DCC BMC thresholds.'.format(
                 count=len(sensors)))
        except:
            self.addUnknown()

        if healthy:
            self.addHealthy()
        else:
            self.addUnHealthy()

dcv_check_ipmi_sensor_thresholds = DcvCheckSensorThresholds(dcv_parse_ipmi_sensor) \
    .title('''Drive Constellation: Check DCC BMC sensor thresholds''') \
    .depends(dcv_parse_ipmi_sensor) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class DcvRunIpmiSdrElist(tasks.ITask):
    def __init__(self, dcc_passgen_task):
        super().__init__()
        self.__dcc_passgen_task = dcc_passgen_task
        self.__result = None
        self.__ret = 1

    def getOutput(self):
        return self.__result

    def getReturnCode(self):
        return self.__ret

    def run(self):
        try:
            if(self.__dcc_passgen_task.getReturnCode()==0):
                psswd = self.__dcc_passgen_task.getOutput().strip()
                cmd = 'ipmitool -I lanplus -H {} -U {} -P {} sdr elist'.format(platforms.cgx.dcv_bmc_config['ip'], platforms.cgx.dcv_bmc_config['user'], psswd)
                cmd_task = tasks.RunCommand(shell=True, args=[cmd])
                cmd_task.run()
                self.__ret = cmd_task.getReturnCode()
                self.__result = cmd_task.getOutput()
        except:
            return

dcv_run_ipmi_sdr_elist = DcvRunIpmiSdrElist(dcc_passgen) \
    .title('Run ipmitool mc info command') \
    .describe('''
        Check DCC BMC status with ipmitool.
        This command requires superuser privileges.''') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(dcc_passgen) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class DcvParseIpmiSdrElist(tasks.ITask):
    def __init__(self, ipmi_sdr_elist_task):
        super(DcvParseIpmiSdrElist, self).__init__()
        self.__ipmi_sdr_elist_task = ipmi_sdr_elist_task
        self.__result = None
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('dcs_dcv_bmc')

    def getResult(self):
        return self.__result

    def run(self):
        try:
            dcs_dcv_bmc = self.__parameters_task.getResult()['dcs_dcv_bmc']
            if not dcs_dcv_bmc:
                return
        except:
            return

        if self.__ipmi_sdr_elist_task.getReturnCode():
            return

        res = {}
        output = self.__ipmi_sdr_elist_task.getOutput().strip()

        sensor_pattern = re.compile(
                r'^(?P<name>[^|\n]+)\|' \
                r'(?P<id>[^|\n]+)\|' \
                r'(?P<status>[^|\n]+)\|' \
                r'(?P<entity_id>[^|\n]+)\|' \
                r'(?P<reading>[^|\n]+)' ,
                flags=re.MULTILINE)


        try:
            sensors = []
            for m in sensor_pattern.finditer(output):
                s = {
                    'name': m.group('name').strip(),
                    'id': m.group('id').strip(),
                    'status': m.group('status').strip(),
                    'entity_id': m.group('entity_id').strip(),
                    'reading': m.group('reading').strip()
                }
                sensors.append(s)
        except:
            pass

        self.__result = sensors

dcv_parse_ipmi_sdr_elist = DcvParseIpmiSdrElist(dcv_run_ipmi_sdr_elist) \
    .title('Parse output from ipmitool sdr elist for DCC BMC') \
    .depends(dcv_run_ipmi_sdr_elist) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class DcvSdrDeviceBom(tasks.ITask):
    def __init__(self, dcv_parse_ipmi_sdr_elist_task):
        super().__init__()
        self.__dcv_parse_ipmi_sdr_elist_task = dcv_parse_ipmi_sdr_elist_task
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        # Create dictionary containing devices that appear to be present
        try:
            bom = {}
            ns_pattern = re.compile(r'^[nN]/?[sS]$')
            for s in self.__dcv_parse_ipmi_sdr_elist_task.getResult():
                name = s['name']
                status = s['status']
                if not ns_pattern.match(status):
                    # Sensor status is not ns, so device appears to be present
                    bom[name] = True
        except:
            pass

        self.__result = bom

dcv_sdr_device_bom = DcvSdrDeviceBom(dcv_parse_ipmi_sdr_elist) \
    .title('Prepare Device Bom from ipmitool sdr elist') \
    .depends(dcv_parse_ipmi_sdr_elist) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class DcvCheckFanBom(tasks.ICheck):
    def __init__(self, dcv_sdr_device_bom_task):
        super().__init__()
        self.__dcv_sdr_device_bom_task = dcv_sdr_device_bom_task
        self.__result = None
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('dcv_fan_bom')

    def run(self):
        from nvsmhealth.lib import DictionarySuperset
        try:
            dcv_fan_bom = self.__parameters_task.getResult()['dcv_fan_bom']
        except:
            return

        output = self.__dcv_sdr_device_bom_task.getResult()
        dictionary_superset = DictionarySuperset.DictionarySuperset(
                    missing_message="Could not detect presence of DCC chassis fan: {}")
        result = dictionary_superset.compare(output, dcv_fan_bom)
        self.addCheckMessage("Checking output of 'ipmitool sdr elist' for expected chassis fans on DCC")
        if result:
            self.addCheckMessage(result)
            self.addUnHealthy()
        else:
            self.addHealthy()

dcv_check_fan_bom = DcvCheckFanBom(dcv_sdr_device_bom) \
    .title('Drive Constellation: Verify chassis fan presence for DCC') \
    .depends(dcv_sdr_device_bom) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class DcvCheckPsuBom(tasks.ICheck):
    def __init__(self, dcv_sdr_device_bom_task, dcv_parse_ipmi_sdr_elist_task):
        super().__init__()
        self.__dcv_sdr_device_bom_task = dcv_sdr_device_bom_task
        self.__dcv_parse_ipmi_sdr_elist_task = dcv_parse_ipmi_sdr_elist_task
        self.__result = None
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('dcv_psu_bom')

    def run(self):
        from nvsmhealth.lib import DictionarySuperset
        try:
            dcv_psu_bom = self.__parameters_task.getResult()['dcv_psu_bom']
        except:
            return

        output = self.__dcv_sdr_device_bom_task.getResult()
        dictionary_superset = DictionarySuperset.DictionarySuperset(
                    missing_message="Could not detect presence of DCC chassis power supply: {}")
        result = dictionary_superset.compare(output, dcv_psu_bom)
        self.addCheckMessage("Checking output of 'ipmitool sdr elist' for expected chassis PSUs on DCC")
        if result:
            self.addCheckMessage(result)
            self.addUnHealthy()
        else:
            healthy = True
            # NVBUG-200554527: Check for Power supply lost for DCC
            # Even if PSU status is ok, readings might have power supply ac lost message
            psu_res = self.__dcv_parse_ipmi_sdr_elist_task.getResult()
            # Filter PSU status keys for readings
            psu_status_keys = [k for k in dcv_psu_bom.keys() if 'status' in k.lower()]
            for s in psu_res:
                if s['name'] in psu_status_keys:
                    reading = s['reading']
                    if 'power supply ac lost' in reading.lower():
                        self.addCheckMessage("AC input is lost, {} has reading:\n{}".format(s['name'], s['reading']))
                        healthy = False

            if healthy:
                self.addHealthy()
            else:
                self.addUnHealthy()

dcv_check_psu_bom = DcvCheckPsuBom(dcv_sdr_device_bom, dcv_parse_ipmi_sdr_elist) \
    .title('Drive Constellation: Verify chassis power supply presence on DCC') \
    .depends(dcv_sdr_device_bom) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

class DcvRunIpmiGetEnables(tasks.ITask):
    def __init__(self, dcc_passgen_task):
        super().__init__()
        self.__dcc_passgen_task = dcc_passgen_task
        self.__result = None
        self.__ret = 1

    def getOutput(self):
        return self.__result

    def getReturnCode(self):
        return self.__ret

    def run(self):
        try:
            if(self.__dcc_passgen_task.getReturnCode()==0):
                psswd = self.__dcc_passgen_task.getOutput().strip()
                cmd = 'ipmitool -I lanplus -H {} -U {} -P {} mc getenables'.format(platforms.cgx.dcv_bmc_config['ip'], platforms.cgx.dcv_bmc_config['user'], psswd)
                cmd_task = tasks.RunCommand(shell=True, args=[cmd])
                cmd_task.run()
                self.__ret = cmd_task.getReturnCode()
                self.__result = cmd_task.getOutput()
        except:
            return

dcv_run_ipmi_getenables = DcvRunIpmiGetEnables(dcc_passgen) \
    .title('Run ipmitool mc info command') \
    .describe('''
            Check DCC BMC status with ipmitool.
            This command requires superuser privileges.''') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(dcc_passgen) \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

class DcvParseIpmiGetEnables(tasks.ITask):
    def __init__(self, dcv_run_ipmi_getenables_task):
        super().__init__()
        self.__dcv_run_ipmi_getenables_task = dcv_run_ipmi_getenables_task
        self.__result = None
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('dcs_dcv_bmc')

    def getResult(self):
        return self.__result

    def run(self):
        try:
            dcs_dcv_bmc = self.__parameters_task.getResult()['dcs_dcv_bmc']
            if not dcs_dcv_bmc:
                return
        except:
            return

        if self.__dcv_run_ipmi_getenables_task.getReturnCode():
            return

        res = {}
        output = self.__dcv_run_ipmi_getenables_task.getOutput().strip()

        try:
            ipmi_getenables_pattern = re.compile(r"^Receive Message Queue Interrupt\s*:\s*(?P<receive_message_queue_interrupt>.*)\nEvent Message Buffer Full Interrupt\s*:\s*(?P<event_message_buffer_full_interrupt>.*)\nEvent Message Buffer\s*:\s*(?P<event_message_buffer>.*)\nSystem Event Logging\s*:\s*(?P<system_event_logging>.*)\nOEM 0\s*:\s*(?P<oem_0>.*)\nOEM 1\s*:\s*(?P<oem_1>.*)\nOEM 2\s*:\s*(?P<oem_2>.*)")

            match = ipmi_getenables_pattern.match(output)
            if match:
                res = match.groupdict()
        except Exception as e:
            print("EXCEPTION:", e)
            pass

        self.__result = res

dcv_parse_ipmi_getenables = DcvParseIpmiGetEnables(dcv_run_ipmi_getenables) \
    .title('Parse output from ipmitool get enables for DCC BMC') \
    .depends(dcv_run_ipmi_getenables) \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

class DcvRunIpmiFru(tasks.ITask):
    def __init__(self, dcc_passgen_task):
        super().__init__()
        self.__dcc_passgen_task = dcc_passgen_task
        self.__result = None
        self.__ret = 1

    def getOutput(self):
        return self.__result

    def getReturnCode(self):
        return self.__ret

    def run(self):
        try:
            if(self.__dcc_passgen_task.getReturnCode()==0):
                psswd = self.__dcc_passgen_task.getOutput().strip()
                cmd = 'ipmitool -I lanplus -H {} -U {} -P {} fru print'.format(platforms.cgx.dcv_bmc_config['ip'], platforms.cgx.dcv_bmc_config['user'], psswd)
                cmd_task = tasks.RunCommand(shell=True, args=[cmd])
                cmd_task.run()
                self.__ret = cmd_task.getReturnCode()
                self.__result = cmd_task.getOutput()
        except:
            return

dcv_run_ipmi_fru = DcvRunIpmiFru(dcc_passgen) \
    .title('Run ipmitool mc info command') \
    .describe('''
            Check DCC BMC status with ipmitool.
            This command requires superuser privileges.''') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(dcc_passgen) \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

class DcvParseIpmiFru(tasks.ITask):
    def __init__(self, dcv_run_ipmi_fru_task):
        super().__init__()
        self.__dcv_run_ipmi_fru_task = dcv_run_ipmi_fru_task
        self.__result = None
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('dcs_dcv_bmc')

    def getResult(self):
        return self.__result

    def run(self):
        try:
            dcs_dcv_bmc = self.__parameters_task.getResult()['dcs_dcv_bmc']
            if not dcs_dcv_bmc:
                return
        except:
            return

        if self.__dcv_run_ipmi_fru_task.getReturnCode():
            return

        output = self.__dcv_run_ipmi_fru_task.getOutput().strip()

        fru_pattern = re.compile(
                r'^FRU Device Description\s*:\s*(?P<fru_name>[^\(]*)\((?P<fru_id>[^\)\n]*)\)?\n(?P<fru_body>(?:[^\n]*\S+[^\n]*\n?)+)',
                flags=re.MULTILINE)

        # Attempt to parse FRU values
        try:
            result = {}
            for m in fru_pattern.finditer(output):
                fru = {}
                fru_name = m.group('fru_name').strip()
                result[fru_name] = fru
                for line in m.group('fru_body').split('\n'):
                    line = line.strip()
                    if not line:
                        continue
                    field = line.split(':', 1)
                    if len(field) != 2:
                        continue
                    field_name = field[0].strip().lower().replace(' ', '_')
                    field_value = field[1].strip()
                    fru[field_name] = field_value
        except Exception as e:
            print("EXCEPTION:", e)
            pass

        self.__result = result

dcv_parse_ipmi_fru = DcvParseIpmiFru(dcv_run_ipmi_fru) \
    .title('Parse output from ipmitool fru print for DCC BMC') \
    .depends(dcv_run_ipmi_fru) \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

class DcvCheckFRUConsistency(tasks.ICheck):
    def __init__(self, fru_task):
        super(DcvCheckFRUConsistency, self).__init__()

        # Declare dependency on fru using the ITask interface
        self.__fru_task = fru_task
        self.depends(fru_task)

    def run(self):
        # TODO:[Kenzen-499] Check if FRU in self.__fru_task is consistent and report health using ICheck interface
        # TODO:Similar approach needs to be used by DCC BMC
        pass

dcv_check_fru_consistency = DcvCheckFRUConsistency(dcv_parse_ipmi_fru) \
    .title('''Drive Constellation: Check FRU information for consistency''') \
    .describe('''
            The FRU (field replaceable unit) information recorded in the BMC
            (baseboard management controller) includes serial numbers for
            various FRUs on the system. For any given system, these serial
            numbers should be consistent among all FRUs. However, it is
            possible for these serial numbers to become inconsistent as the
            result of normal maintenance (such as FRU replacement). This check
            makes sure serial numbers are consistent for all FRUs recorded in
            the DCC BMC.''') \
   .parameters('enable_dcv_checks') \
   .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))
