# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from nvsmhealth import modules, tasks, config, platforms
#from nvsmhealth.modules.dcs_modules import dcs_usb
from nvsmhealth.modules.dcs_modules.dcs_dcv_bmc import dcc_passgen
#from nvsmhealth.modules.dcs_modules.dcs_peg import dcs_peg_gen_up
import re

__all__ = [
    'run_psu0_vendor',
    'run_psu0_model',
    'run_psu0_serial_number',
    'run_psu0_fw_version',
    'parse_psu0',
    'run_psu1_vendor',
    'run_psu1_model',
    'run_psu1_serial_number',
    'run_psu1_fw_version',
    'parse_psu1',
    'check_dcs_psu_info',
    'show_psu0',
    'show_psu1',
    'show_dcs_psu_info'
]

#PSU Raw Output Run/Parse/Show Classes
class DcsRunPSUOutput(tasks.ITask):
    def __init__(self, dcc_passgen_task, dcc_psu_vendor_task, psu_cmd_type):
        super().__init__()
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('dcs_psu_info_commands')
        self.__dcc_passgen_task = dcc_passgen_task
        self.__dcc_psu_vendor_task = dcc_psu_vendor_task
        self.__psu_cmd_type = psu_cmd_type
        self.__result = None
        self.__ret = 1

    def getOutput(self, start_position):
        ret_value = None
        if self.__result and isinstance(self.__result, str):
            ret_value = self.__result.replace("\n","").strip()[start_position:]
        else:
            ret_value = self.__result

        return ret_value

    def getOutputAscii(self, start_position):
        ret_ascii = None
        try:
            ret_value = self.getOutput(start_position)
            ret_ascii =  bytes.fromhex(ret_value).decode('ascii')
        except:
            ret_ascii = None

        return ret_ascii

    def getReturnCode(self):
        return self.__ret

    def getPsuCmd(self):
        psu_cmd = None

        # Get the platform specific directory of PSU commands
        try:
            dcs_psu_info_commands = self.__parameters_task.getResult()['dcs_psu_info_commands']
        except:
            return None

        # Pull PSU command for the specified PSU Info from the platform specific directory
        try:
            impi_raw_command = dcs_psu_info_commands.get(self.__psu_cmd_type)
        except:
            return None


        if isinstance(impi_raw_command, str):
            # Single raw command passed in as a string, so just use it
            psu_cmd = impi_raw_command
        elif self.__dcc_psu_vendor_task:
            # Get vendor info to look up the vendor specific command
            psu_vendor = self.__dcc_psu_vendor_task.getOutputAscii(6)
            try:
                psu_cmd = impi_raw_command.get(psu_vendor, None)
            except:
                return None

        return psu_cmd

    def run(self):
        # Get the raw command to use
        psu_raw_cmd = self.getPsuCmd()
        if not psu_raw_cmd:
            self.__ret = 1
            self.__result = None
            return

        try:
            if(self.__dcc_passgen_task.getReturnCode()==0):
                psswd = self.__dcc_passgen_task.getOutput().strip()
                cmd = 'ipmitool -I lanplus -H {} -U {} -P {} raw {}'.format(platforms.cgx.dcv_bmc_config['ip'], platforms.cgx.dcv_bmc_config['user'], psswd, psu_raw_cmd)
                cmd_task = tasks.RunCommand(shell=True, args=[cmd])
                cmd_task.run()
                self.__ret = cmd_task.getReturnCode()
                self.__result = cmd_task.getOutput()
        except:
            self.__ret = 1
            self.__result = None
            return

class DcsParsePSUOutput(tasks.ITask):
    def __init__(self, run_psu_model_task, run_psu_vendor_task, run_psu_serial_number_task, run_psu_fw_ver_task):
        super().__init__()
        self.__run_psu_model_task = run_psu_model_task
        self.__run_psu_vendor_task = run_psu_vendor_task
        self.__run_psu_serial_number_task = run_psu_serial_number_task
        self.__run_psu_fw_ver_task = run_psu_fw_ver_task
        self.__output = None

    def getOutput(self):
        return self.__output

    def parse_task_output(self, run_task, start_position):
        if run_task.getReturnCode():
            # command failed
            return 'UNKNOWN'

        run_task_output = run_task.getOutput(start_position)

        if not run_task_output:
            # command did not return any output
            return 'UNKNOWN'

        return run_task_output

    def parse_task_output_ascii(self, run_task, start_position):
        if run_task.getReturnCode():
            # command failed
            return 'UNKNOWN'

        # Handle corner case where return code is 0 but no data is returned.
        run_task_output = run_task.getOutputAscii(start_position)
        if not run_task_output:
            # command did not return any output
            return 'UNKNOWN'

        return run_task_output

    def run(self):
        psu_info = {}
        psu_info['Model'] = self.parse_task_output_ascii(self.__run_psu_model_task, 6)
        psu_info['Vendor'] = self.parse_task_output_ascii(self.__run_psu_vendor_task, 6)
        psu_info['SN'] = self.parse_task_output_ascii(self.__run_psu_serial_number_task, 6)
        if psu_info.get('Vendor', '') == 'DELTA':
            fw_task_output = self.parse_task_output(self.__run_psu_fw_ver_task, 3)
            try:
                fw_ver = fw_task_output.replace(" ", ".")
            except:
                fw_ver = fw_task_output

            psu_info['FW Ver'] = fw_ver
        else:
            psu_info['FW Ver'] = self.parse_task_output_ascii(self.__run_psu_fw_ver_task, 3)

        self.__output = psu_info

class DcsShowPSUOuput(tasks.ITask):
    def __init__(self, parse_psu_task):
        super().__init__()
        self.__parse_psu_task = parse_psu_task
        self.__output = None

    def run(self):
        psu_info = self.__parse_psu_task.getOutput()

        org_title=self.getTitle()
        self.title(org_title+" {}".format("Model"))
        self.send(msg=psu_info.get('Model', 'UNKNOWN'))

        self.title(org_title+" {}".format("Vendor"))
        self.send(msg=psu_info.get('Vendor', 'UNKNOWN'))

        self.title(org_title+" {}".format("Serial Number"))
        self.send(msg=psu_info.get('SN', 'UNKNOWN'))

        self.title(org_title+" {}".format("FW Version"))
        self.send(msg=psu_info.get('FW Ver', 'UNKNOWN'))

        # Restore original title
        self.title(org_title)

        self.__output = psu_info

# PSU-0 Run Commands
run_psu0_vendor = DcsRunPSUOutput(dcc_passgen, None, 'psu0_vendor') \
    .title('Run ipmitool PSU-0 Vendor command') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(dcc_passgen) \
    .parameters('dcs_psu_info_commands') \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

run_psu0_model = DcsRunPSUOutput(dcc_passgen, None, 'psu0_model') \
    .title('Run ipmitool PSU-0 Model command') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(dcc_passgen) \
    .parameters('dcs_psu_info_commands') \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

run_psu0_serial_number = DcsRunPSUOutput(dcc_passgen, None, 'psu0_serial_number') \
    .title('Run ipmitool PSU-0 Serial Number command') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(dcc_passgen) \
    .parameters('dcs_psu_info_commands') \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

run_psu0_fw_version = DcsRunPSUOutput(dcc_passgen, run_psu0_vendor, 'psu0_fw_version') \
    .title('Run ipmitool PSU-0 Firmware Version command') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(dcc_passgen) \
    .depends(run_psu0_vendor) \
    .parameters('dcs_psu_info_commands') \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

# PSU=0 Parse command
parse_psu0 = DcsParsePSUOutput(run_psu0_model, run_psu0_vendor, run_psu0_serial_number, run_psu0_fw_version) \
    .title('''Parse PSU-0 output''') \
    .depends(run_psu0_vendor) \
    .depends(run_psu0_model) \
    .depends(run_psu0_serial_number) \
    .depends(run_psu0_fw_version) \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

# PSU-1 Run Commands
run_psu1_vendor = DcsRunPSUOutput(dcc_passgen, None, 'psu1_vendor') \
    .title('Run ipmitool PSU-1 Vendor command') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(dcc_passgen) \
    .parameters('dcs_psu_info_commands') \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

run_psu1_model = DcsRunPSUOutput(dcc_passgen, None, 'psu1_model') \
    .title('Run ipmitool PSU-1 Model command') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(dcc_passgen) \
    .parameters('dcs_psu_info_commands') \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

run_psu1_serial_number = DcsRunPSUOutput(dcc_passgen, None, 'psu1_serial_number') \
    .title('Run ipmitool PSU-1 Serial Number command') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(dcc_passgen) \
    .parameters('dcs_psu_info_commands') \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

run_psu1_fw_version = DcsRunPSUOutput(dcc_passgen, run_psu1_vendor, 'psu1_fw_version') \
    .title('Run ipmitool PSU-1 Firmware Version command') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(dcc_passgen) \
    .depends(run_psu1_vendor) \
    .parameters('dcs_psu_info_commands') \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

# PSU-1 parse command
parse_psu1 = DcsParsePSUOutput(run_psu1_model, run_psu1_vendor, run_psu1_serial_number, run_psu1_fw_version) \
    .title('''Parse PSU-1 output''') \
    .depends(run_psu1_vendor) \
    .depends(run_psu1_model) \
    .depends(run_psu1_serial_number) \
    .depends(run_psu1_fw_version) \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

# Check PSUs
class DcsCheckPSUInfo(tasks.ICheck):
    def __init__(self, parse_psu0_task, parse_psu1_task):
        super().__init__()
        self.__parse_psu0_task = parse_psu0_task
        self.__parse_psu1_task = parse_psu1_task
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('dcs_psu_attrib_values')

    def check_psu_attrib(self, attrib, attrib_values, results):
        actual_psu_attrib_value = None
        # Make sure all PSUs have the same same attribute value
        for psu, psu_info in results.items():
            psu_attrib_value = psu_info.get(attrib, 'UNKNOWN')
            if actual_psu_attrib_value == None:
                actual_psu_attrib_value = psu_attrib_value
            elif actual_psu_attrib_value != psu_attrib_value:
                return 'More than one PSU {} present: {}, {}'.format(attrib, actual_psu_attrib_value, psu_attrib_value)

        # Check if PSU attribute is in the supported attribute list
        if attrib_values != None:
            valid_attrib_values = attrib_values.get(attrib, None)
            if valid_attrib_values != None and actual_psu_attrib_value not in valid_attrib_values:
                return 'Unsupported PSU {}: {}'.format(attrib, actual_psu_attrib_value)

        return ''

    def run(self):
        Health = True
        try:
            dcs_psu_attrib_values = self.__parameters_task.getResult()['dcs_psu_attrib_values']
        except:
            # Could not get list of valid values
            self.addCheckMessage("Could not get list of valid PSU values")
            self.addUnknown()
            return

        # Get the parsed results
        dcs_psu_results = {}
        dcs_psu_results['PSU-0'] = self.__parse_psu0_task.getOutput()
        dcs_psu_results['PSU-1'] = self.__parse_psu1_task.getOutput()

        # Check PSU Vendor
        Msg = self.check_psu_attrib('Vendor', dcs_psu_attrib_values, dcs_psu_results)
        if Msg != '':
            self.addCheckMessage(Msg)
            Health = False

        # Check PSU Model
        Msg = self.check_psu_attrib('Model', dcs_psu_attrib_values, dcs_psu_results)
        if Msg != '':
            self.addCheckMessage(Msg)
            Health = False

        if Health == True:
            self.addHealthy()
        elif Health == False:
            self.addUnHealthy()
        else:
            self.addUnknown()

check_dcs_psu_info = DcsCheckPSUInfo(parse_psu0, parse_psu1) \
    .title('''Drive Constellation: Check DCC PSU Info''') \
    .depends(parse_psu0) \
    .depends(parse_psu1) \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

# Show the info for PSU's after the PSU info check has been run
# PSU-0 Show command
show_psu0 = DcsShowPSUOuput(parse_psu0) \
    .title('''Drive Constellation: DCC PSU-0''') \
    .depends(check_dcs_psu_info) \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

# PSU-1 show command
show_psu1 = DcsShowPSUOuput(parse_psu1) \
    .title('''Drive Constellation: DCC PSU-1''') \
    .depends(show_psu0) \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

class ShowDcsPSUInfo(tasks.ITask):
    def __init__(self):
        super().__init__()
        self.dcs_sensor_monitor = "0x30 0x9a"
        self.dcs_switch_psu_channel = "0x06 0x52 0x03 0xea 0x00"
        self.dcs_pmbus_cmd = "0x06 0x52 0x03 0xb0"
        self.disable_sensor_cmd = "0x00"
        self.enable_sensor_cmd = "0x01"
        self.dcs_psu_mfr_id = "0x0d 0x99" # CMD_MFR_ID
        self.dcs_psu_mfr_model = "0x0c 0x9a" # CMD_MFR_MODEL
        self.dcs_psu_mfr_serial = "0x19 0x9e" # CMD_MFR_SERIAL_ACBEL
        self.dcs_psu_fw_version = "0x08 0xd5" # CMD_FW_VERSION
        self.psus = ["0x01", "0x02", "0x04", "0x08"]

    def get_data(self, cmd_out):
        out = cmd_out.replace("\n","").strip().split(" ")
        # Get the length from first byte
        l = int(out[0], 16)
        # Get data from the length
        data = out[1:(l+1)]
        # Convert hex into ascii string
        return bytes.fromhex(" ".join(data)).decode('ascii')

    def get_manufacturer(self):
        try:
            cmd = 'ipmitool raw {}'.format(" ".join([self.dcs_pmbus_cmd, self.dcs_psu_mfr_id]))
            cmd_task = tasks.RunCommand(shell=True, args=[cmd])
            cmd_task.run()
            if not cmd_task.getReturnCode():
                return self.get_data(cmd_task.getOutput())
        except:
            return 'Unknown'

    def get_model(self):
        try:
            cmd = 'ipmitool raw {}'.format(" ".join([self.dcs_pmbus_cmd, self.dcs_psu_mfr_model]))
            cmd_task = tasks.RunCommand(shell=True, args=[cmd])
            cmd_task.run()
            if not cmd_task.getReturnCode():
                return self.get_data(cmd_task.getOutput())
        except:
            return 'Unknown'

    def get_serial(self):
        try:
            cmd = 'ipmitool raw {}'.format(" ".join([self.dcs_pmbus_cmd, self.dcs_psu_mfr_serial]))
            cmd_task = tasks.RunCommand(shell=True, args=[cmd])
            cmd_task.run()
            if not cmd_task.getReturnCode():
                return self.get_data(cmd_task.getOutput())
        except:
            return 'Unknown'

    def get_fw_version(self):
        try:
            cmd = 'ipmitool raw {}'.format(" ".join([self.dcs_pmbus_cmd, self.dcs_psu_fw_version]))
            cmd_task = tasks.RunCommand(shell=True, args=[cmd])
            cmd_task.run()
            if not cmd_task.getReturnCode():
                return self.get_data(cmd_task.getOutput())
        except:
            return 'Unknown'

    def disable_sensor(self):
        try:
            cmd = 'ipmitool raw {}'.format(" ".join([self.dcs_sensor_monitor, self.disable_sensor_cmd]))
            cmd_task = tasks.RunCommand(shell=True, args=[cmd])
            cmd_task.run()
        except:
            pass

    def enable_sensor(self):
        try:
            cmd = 'ipmitool raw {}'.format(" ".join([self.dcs_sensor_monitor, self.enable_sensor_cmd]))
            cmd_task = tasks.RunCommand(shell=True, args=[cmd])
            cmd_task.run()
        except:
            pass

    def switch_psu_channel(self, psu):
        try:
            cmd = 'ipmitool raw {}'.format(" ".join([self.dcs_switch_psu_channel, psu]))
            cmd_task = tasks.RunCommand(shell=True, args=[cmd])
            cmd_task.run()
        except:
            pass

    def run(self):
        # Iterate over all PSUs
        title = "DCS PSU-{}"
        try:
            for i, psu in enumerate(self.psus):
                self.disable_sensor()
                self.switch_psu_channel(psu)
                mfr = self.get_manufacturer()
                self.title(title.format(i) + " {}".format("Vendor"))
                self.send(msg=mfr)
                model = self.get_model()
                self.title(title.format(i) + " {}".format("Model"))
                self.send(msg=model)
                serial = self.get_serial()
                self.title(title.format(i) + " {}".format("Serial Number"))
                self.send(msg=serial)
                fw_version = self.get_fw_version()
                self.title(title.format(i) + " {}".format("FW Version"))
                self.send(msg=fw_version)
                self.enable_sensor()
        except:
            self.enable_sensor()
            pass

show_dcs_psu_info = ShowDcsPSUInfo() \
    .title('Run ipmitool for DCS PSU information') \
    .depends(modules.common.check_superuser_privileges) \
    .parameters('enable_dcv_checks') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))
