# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from nvsmhealth import modules, tasks
from nvsmhealth.lib.xmltodict import xmltodict
import logging

__all__ = [
    'check_nvidia_grid_license',
]

class CheckNvidiaGridLicense(tasks.ICheck):
    def __init__(self, parse_nvidia_smi_task):
        super().__init__()
        self.__parse_nvidia_smi_task = parse_nvidia_smi_task
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('dcs_grid_license')

    def run(self):
        try:
            dcs_grid = self.__parameters_task.getResult()['dcs_grid_license']
            if not dcs_grid:
                return
        except:
            return
        
        nvidia_smi_res = self.__parse_nvidia_smi_task.getResult()
        if nvidia_smi_res == None:
            return

        healthy = False

        try:
            for gpu, info in nvidia_smi_res.items():
                gpu_dict = xmltodict.parse(info)
                # GRID License check on GRID Products only
                # NVBug-2795033: GPU cards which can be licensed will have the `License Status` field and
                # those which cannot will not have this field.
                if gpu_dict['nvidia_smi_log']['gpu'].get('grid_licensed_product', None) == None:
                    continue
                elif gpu_dict['nvidia_smi_log']['gpu']['grid_licensed_product'].get('license_status', None) == None:
                    continue
                # NVBug:3145085 - If the product name is QVDCW and status is licensed then only show healthy
                elif gpu_dict['nvidia_smi_log']['gpu']['grid_licensed_product'].get('licensed_product_name', None) == None:
                    continue

                product_name = gpu_dict['nvidia_smi_log']['gpu']['grid_licensed_product']['licensed_product_name']
                product_name = product_name.strip()
                
                if product_name == 'Quadro Virtual Data Center Workstation':
                    if not gpu_dict['nvidia_smi_log']['gpu']['grid_licensed_product']['license_status'] == 'Licensed':
                        healthy = False
                        break
                    else:
                        healthy = True


            if healthy == True:
                self.addHealthy()
            else:
                self.addCheckMessage("Check GRID License: Contact Nvidia")
                self.addUnHealthy()

        except:
            self.addCheckMessage("Error while performing GRID License Check")
            self.addUnknown()

check_nvidia_grid_license = CheckNvidiaGridLicense(modules.nvidia_smi.parse_nvidia_smi) \
    .title('''Drive Constellation: GRID License Status''') \
    .depends(modules.nvidia_smi.parse_nvidia_smi) \
    .parameters('enable_dcv_checks') \
    .parameters('enable_cgx_sm_checks') \
    .predicate(lambda: (modules.parameter.parameters.getResult().get('enable_dcv_checks') or modules.parameter.parameters.getResult().get('enable_cgx_sm_checks')))

