# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import re
import json

from nvsmhealth import tasks, modules

__all__ = [
    'check_dimm_vendors',
    'check_dimm_part_number'
]

class CheckDimmVendors(tasks.ICheck):
    def __init__(self):
        super().__init__()
        self.__get_dimm_vendors_task = modules.dmidecode.get_dimm_vendors
        self.depends(self.__get_dimm_vendors_task)
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('dmidecode_dimm_vendors')

    def run(self):

        try:
            dimm_vendors = self.__parameters_task.getResult()['dmidecode_dimm_vendors']
        except:
            return

        out = self.__get_dimm_vendors_task.getResult()
        if not out:
            self.addCheckMessage('ERROR: Could not parse dmidecode output')
            self.addUnknown()
            return

        healthy = True
        for dimm in out:
            if dimm not in dimm_vendors:
                self.addCheckMessage('Unknown DIMM vendor "{value}"'.format(value=dimm))
                healthy = False
            else:
                # Found the expected dimm vendor
                pass

        if healthy:
            self.addHealthy()
        else:
            self.addUnHealthy()


check_dimm_vendors = CheckDimmVendors() \
        .title('Verify DIMM vendors')

class CheckDimmPartNumber(tasks.ICheck):
    def __init__(self):
        super().__init__()
        self.__dimm_task = modules.dmidecode.parse_dmidecode
        self.depends(self.__dimm_task)
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)

    def run(self):
        # Return if parameter is not defined
        try:
            dimm_boms = self.__parameters_task.getResult()['dimm_bom']
            part_number = self.__parameters_task.getResult()['dimm_part_number']
        except:
            return

        # Unknown check for no result from dmidecode
        res = self.__dimm_task.getResult()
        if not res:
            self.addCheckMessage('No result from parse dmidecode output')
            self.addUnknown()
            return

        healthy = True
        for dimm in dimm_boms:
            if dimm in res.keys():
                if 'part_number' in res[dimm].keys():
                    if res[dimm]['part_number'].strip() not in part_number:
                        self.addCheckMessage('Mismatch in DIMM "{}" part number, expected is "{}" found is "{}"'.format(dimm, " or ".join(part_number), res[dimm]['part_number']))
                        healthy = False
                else:
                    self.addCheckMessage('DIMM "{}" part number not found'.format(dimm))
                    healthy = False
            else:
                # Must be caught on checking DIMMs
                pass

        if healthy:
            self.addHealthy()
        else:
            self.addUnHealthy()

check_dimm_part_number = CheckDimmPartNumber() \
    .title('Verify DIMM part number') \
    .parameters('dimm_part_number') \
    .parameters('dimm_bom')
