# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import sys
import re

from nvsmhealth import modules, tasks, config
from nvsmhealth.lib import DictionarySuperset

__all__ = [
    'check_smartctl_megaraid_disk_count',
    'check_smartctl_disk_count',
    'check_smartctl_ssd_brick'
]

class CheckDiskCount(tasks.ICheck):
    def __init__(self, is_raid=True):
        super().__init__()
        self.__is_raid = is_raid
        self.__get_disk_count_task = modules.smartctl.parse_smartctl_device_info
        self.depends(self.__get_disk_count_task)
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)

    def run(self):
        try:
            exp_disk_count = self.__parameters_task.getResult()[list(self.getParameters())[0]]
        except:
            return

        self.addCheckMessage("Checking output of 'smartctl' for expected disks")
        if self.__is_raid:
            disk_count = self.__get_disk_count_task.getResult()
        else:
            disk_count = self.__get_disk_count_task.getDisksBySize()
        if not disk_count:
            self.addCheckMessage("No disk(s) found")
            self.addUnknown()
            return

        dictionary_superset = DictionarySuperset.DictionarySuperset(
                missing_message="No disks of capacity '{}' were found",
                changed_message="Disks of capacity '{}' were found '{}' when '{}' disk(s) were expected")
        result = dictionary_superset.compare(disk_count, exp_disk_count)
        if result == None:
            # Healthy check - No diffs found
            self.addHealthy()
        else:
            # UnHealthy check - Print/Stream diffs found
            self.addUnHealthy()
            self.addCheckMessage(result)

check_smartctl_megaraid_disk_count = CheckDiskCount() \
        .title('Verify installed MegaRAID disks') \
        .describe('''Count the disks attached to the MegaRAID controller using the smartctl command''') \
        .parameters('smartctl_megaraid_disk_count')

check_smartctl_disk_count = CheckDiskCount(is_raid=False) \
        .title('Verify installed disks') \
        .describe('''Verify that all of the expected disks are installed''') \
        .parameters('smartctl_disk_count')

class CheckSSDBrick(tasks.ICheck):
    def __init__(self):
        super().__init__()
        self.__smartctl_info_task = modules.smartctl.parse_smartctl_device_info
        self.depends(self.__smartctl_info_task)
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('smartctl_check_ssd_brick')

    def run(self):
        try:
            check = self.__parameters_task.getResult()['smartctl_check_ssd_brick']
            if not check:
                return
        except:
            return

        errormod_pattern = re.compile(r'ERRORMOD', flags=re.IGNORECASE)
        # Look for disks with bricked firmware conditions
        brick = False
        try:
            disk_bom = self.__smartctl_info_task.getBomList()

            if disk_bom is None:
                return

            for key, value in disk_bom.items():
                # Check for "ERRORMOD" in firmware version
                if not 'firmware_version' in value:
                    continue
                firmware_version = value['firmware_version']
                m = errormod_pattern.match(firmware_version)
                if not m:
                    continue
                # NOTE: It is likely that the disk capacity is also incorrect, but
                # we do not check for this here. Disk capacity is checked elsewhere
                # as part of the "smartctl-disk-count" check.
                brick = key
                break
        except Exception as e:
            self.addCheckMessage(str(e))
            self.addUnknown()
            return

        # Print details message
        if brick:
            self.addUnHealthy()
            self.addCheckMessage('Possible firmware bug on disk "{0}"'.format(brick))
        else:
            self.addHealthy()
            self.addCheckMessage('No disks with firmware bug found')

check_smartctl_ssd_brick = CheckSSDBrick() \
        .title('Check for SSD health')
