# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import re
import json

from nvsmhealth import tasks
from nvsmhealth.modules.common import VariantRef, check_superuser_privileges

__all__ = [
    'run_dmidecode',
    'run_dmidecode_memory',
    'parse_memory_dimms',
    'parse_dmidecode',
    'get_dimm_vendors',
]

run_dmidecode_memory = tasks.RunCommand(
        args=['dmidecode', '--type', 'memory']) \
    .title('''Run the dmidecode command''') \
    .describe('''
            Run the "dmidecode" command to get memory DMI type information.
            Some flags are added to output in a machine-readable format.
            This command does not require superuser privileges.''') \
    .depends(check_superuser_privileges)

run_dmidecode = tasks.RunCommand(
        args=['dmidecode']) \
    .title('''Run the dmidecode command''') \
    .describe('''
            Verify system as described by SMBIOS/DMI using the dmidecode tool''') \
    .depends(check_superuser_privileges)

class ParseMemoryDimms(tasks.ITask):
    def __init__(self, run_dmidecode_memory):
        super(ParseMemoryDimms, self).__init__()

        # Declare dependency on nvme list using the ITask interface
        self.__dimm_task = run_dmidecode_memory
        self.depends(run_dmidecode_memory)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        memory_pattern = re.compile(r'Memory\s*Device(.*\n)*?\s*Size:\s*(?P<size>.*)', flags=re.MULTILINE)
        output = self.__dimm_task.getOutput()
        if not output:
            return
        memory = []
        for m in memory_pattern.finditer(output):
            size = m.group('size').strip()
            memory.append(size)

        # Add the parsed result as JSON object
        self.__result = json.dumps(memory, ensure_ascii=True)

parse_memory_dimms = ParseMemoryDimms(run_dmidecode_memory) \
    .title('''Parse memory DMI type information gathered from dmidecode tool''')

class ParseDmidecode(tasks.ITask):
    def __init__(self):
        super().__init__()

        self.__dmi_command_task = run_dmidecode
        self.depends(self.__dmi_command_task)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        if self.__dmi_command_task.getReturnCode():
            return

        def to_underscore_name(name):
            return '_'.join(map(lambda w: w.lower(), name.split(':')[0].split(' ')))

        sub_block_pattern = re.compile(r'^(?P<block_name>\S.*)\n(?P<block_body>(\t.*\n?)*)', flags=re.MULTILINE)
        attribute_pattern = re.compile(r'^(?P<attribute_name>[^:]+):\s*(?P<attribute_value>.*\S)\s*$', flags=re.MULTILINE)
        def parse_block(block, blockValue):
            dictionary = {}
            array = []
            # Iterate over the sub-blocks contained in this block
            match_count = 0
            empty_count = 0
            for m in re.finditer(sub_block_pattern, block):
                match_count += 1
                block_name = to_underscore_name(m.group('block_name'))
                # Remove leading tabs from the sub-block body
                block_body = re.sub(r'^\t', '', m.group('block_body'), flags=re.MULTILINE)
                sub_block_value = VariantRef()
                parse_block(block_body, sub_block_value)
                if sub_block_value.value() == None:
                    # Check for an attribute
                    m2 = re.search(attribute_pattern, m.group(0))
                    if m2:
                        sub_block_value.attribute = m2.group('attribute_value')
                    else:
                        array.append(m.group(0).rstrip())
                        empty_count += 1
                # Check for this sub-block in our dictionary
                if block_name not in dictionary:
                    dictionary[block_name] = sub_block_value.value()
                else:
                    if not isinstance(dictionary[block_name], list):
                        # Convert the existing element into an array
                        old = dictionary[block_name]
                        dictionary[block_name] = []
                        dictionary[block_name].append(old)
                    dictionary[block_name].append(sub_block_value.value())
            # Determine the return value for this block
            if match_count == 0:
                pass  # The blockValue is None
            else:
                if empty_count == match_count:
                    # All of the sub-blocks are empty; return an array of strings
                    blockValue.array = array
                else:
                    # At least some of the sub-blocks are not empty; return a dictionary
                    blockValue.dict = dictionary

        def parse_record(record, recordValue):
            pass

        header_pattern = re.compile(r'^# dmidecode\s+(?P<dmidecode_version>.+)\n(.*\n)*SMBIOS\s+(?P<smbios_version>.+)\s+present.\n(.*\n)*Table at\s+(?P<dmi_table_address>.+)\.$', flags=re.MULTILINE)
        def parse_header(header, headerValue):
            dictionary = {}
            # Grab each field from the header
            m = re.search(header_pattern, header)
            if not m:
                self.send(info="Could not parse dmidecode header")
                return
            for key in ['dmidecode_version',
                      'smbios_version',
                      'dmi_table_address']:
                dictionary[key] = m.group(key)
            headerValue.dict = dictionary

        dmidecode_pattern = re.compile(r'^(?P<dmidecode_header># dmidecode.*\n(\S.*\n)*)\n(?P<dmidecode_body>Handle.*\n(.*\n)+)', flags=re.MULTILINE)
        record_pattern = re.compile(r'^Handle (?P<record_handle>0x[0-9A-F]{4}), DMI type (?P<record_dmi_type>\d+), (?P<record_bytes>\d+) bytes?\n(?P<record_name>.*)\n(?P<record_body>(\t.*\n)+)', flags=re.MULTILINE)
        def parse_dmidecode(output, dmidecodeValue):
            dictionary = {}
            # Get the header and body
            m1 = re.search(dmidecode_pattern, output)
            if not m1:
                self.send(info="Could not find dmidecode header")
                return
            # Parse the header
            header_value = VariantRef()
            parse_header(m1.group('dmidecode_header'), header_value)
            dictionary['header'] = header_value.value()
            # Iterate over each record in the body
            for m2 in re.finditer(record_pattern, m1.group('dmidecode_body')):
                # Remove leading tabs from the record body
                record_body = re.sub(r'^\t', '', m2.group('record_body'), flags=re.MULTILINE)
                # Parse the record body
                record_value = VariantRef()
                parse_block(record_body, record_value)
                # Add the record value to the output dictionary
                if m2.group('record_name') == 'Memory Device':
                    record_name = '{0} ({1})'.format(
                            m2.group('record_name'),
                            record_value.value().get('locator', m2.group('record_handle')))
                else:
                    record_name = '{0} ({1})'.format(m2.group('record_name'), m2.group('record_handle'))
                dictionary[record_name] = record_value.value()
            dmidecodeValue.dict = dictionary

        # Parse the dmidecode output into a Python dictionary
        dmidecode_value = VariantRef()
        parse_dmidecode(self.__dmi_command_task.getOutput().strip(), dmidecode_value)

        self.__result = dmidecode_value.value()

parse_dmidecode = ParseDmidecode() \
    .title('''Parse dmidecode tool output''')

class GetDimmVendors(tasks.ITask):
    def __init__(self):
        super().__init__()
        self.__parse_dmidecode_task = parse_dmidecode
        self.depends(self.__parse_dmidecode_task)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        out = self.__parse_dmidecode_task.getResult()
        if not out:
            return

        self.__result = set()

        try:

            def checkIfNoDimm(vendor):
                dimm_no_presence=['no dimm','not specified']
                if any(vendor.lower() in s for s in dimm_no_presence):
                    return True
                else:
                    return False

            for key, body in out.items():
                if 'Memory Device' in key \
                    and 'manufacturer' in body \
                    and not checkIfNoDimm(body['manufacturer']):
                    self.__result.add(body['manufacturer'])
        except:
            return

get_dimm_vendors = GetDimmVendors() \
        .title("Verify system as described by SMBIOS/DMI using the dmidecode tool")
