#!/usr/bin/env python3

# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.


import http.client
import json
import logging
import requests
import sys
import time

class BackendError(Exception):
    pass

class NVSMHandle(object):
    def __init__(self, host='127.0.0.1', port='273',
            token='/etc/nvsm/nvsm-perpetual.jwt'):
        super(NVSMHandle, self).__init__()
        self.__tokenfile = token
        self.__host = host
        self.__port = port

        self.__prepareToken()

    def __prepareToken(self):
        self.__tokenstring = None
        logging.debug('Attempting to read NVSM API token file: {0}',
                self.__tokenfile)
        try:
            with open(self.__tokenfile, 'r') as jwt_file:
                self.__tokenstring = jwt_file.read()
        except IOError as e:
            raise BackendError(
                    'Could not read NVSM API token: {0}'.format(e))
        logging.debug('Successfully read NVSM API token')

    def connect(self):
        pass  # TODO: Check if we are able to connect to NVSM backend

    def query(self, api):
        url = 'https://{host}:{port}/{path}'.format(
                host=self.__host, port=self.__port, path=api.lstrip('/'))
        logging.debug('Query NVSM backend: GET {0}', url)

        import urllib3
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)  # FIXME
        try:
            r = requests.get(url,
                    verify=False,  # XXX: Don't check the SSL certificate
                    headers={
                        'Authorization': 'Bearer {0}'.format(self.__tokenstring)
                    })
        except requests.exceptions.ConnectionError as e:
            raise BackendError(
                    'Error connecting to NVSM backend: {0}'.format(e))

        logging.debug('Response from NVSM backend:\n{0}', r.text)

        if r.status_code != requests.codes.ok:
            raise BackendError(
                    'Bad HTTP status code "{0}" from NVSM backend: {1}'.format(
                        r.status_code,
                        http.client.responses.get(r.status_code, 'Unknown Status Code')))

        try:
            response = r.json()
        except ValueError as e:
            raise BackendError(
                    'Error decoding JSON response from NVSM backend: {0}'.format(e))

        return response

def queryMembers(data, handle, responses):
    if type(data) is dict:
        for k, v in data.items():
            if k == '@odata.id':
                if type(v) is list:
                    for q in v:
                        queryEndpoint(endpoint=q, handle=handle, responses=responses)
                else:
                    queryEndpoint(endpoint=v, handle=handle, responses=responses)
            else:
                queryMembers(data=v, handle=handle, responses=responses)
    if type(data) is list:
        for v in data:
            queryMembers(data=v, handle=handle, responses=responses)

def queryEndpoint(endpoint, handle, responses):

    endpoint = endpoint.split('#')[0]  # Ignore local references

    response = {
        'method': 'GET',
        'endpoint': endpoint
    }

    if any(map(lambda x: x.get('endpoint') == endpoint, responses)):
        return  # Already queried this endpoint

    try:
        data = handle.query(endpoint)
    except BackendError as e:
        response['BackendError'] = str(e)
        responses.append(response)
        return

    response['response'] = data
    responses.append(response)

    queryMembers(data, handle=handle, responses=responses)

api_endpoints = [
        "/redfish/v1",
        "/redfish/v1/Chassis/1/Power",
        "/redfish/v1/Chassis/1/Thermal",
        "/redfish/v1/Chassis/2/Power",
        "/redfish/v1/Chassis/2/Thermal",
        "/redfish/v1/Systems",
        "/redfish/v1/Systems/1",
        "/redfish/v1/Systems/1/Memory",
        "/redfish/v1/Systems/1/Processors",
        "/redfish/v1/Systems/1/Storage",
        "/redfish/v1/Systems/1/Storage/1",
        "/redfish/v1/Systems/1/Storage/1/Volumes",
        "/nvsm/v1/Chassis/1/Alerts",
        "/nvsm/v1/Chassis/1/Policy",
        "/nvsm/v1/Chassis/1/Power/Alerts",
        "/nvsm/v1/Chassis/1/Power/Policy",
        "/nvsm/v1/Chassis/1/Thermal/Alerts",
        "/nvsm/v1/Chassis/1/Thermal/Policy",
        "/nvsm/v1/Chassis/2/Alerts",
        "/nvsm/v1/Chassis/2/Policy",
        "/nvsm/v1/Chassis/2/Power/Alerts",
        "/nvsm/v1/Chassis/2/Power/Policy",
        "/nvsm/v1/Chassis/2/Thermal/Alerts",
        "/nvsm/v1/Chassis/2/Thermal/Policy",
        "/nvsm/v1/Policy",
        "/nvsm/v1/Schemas/Systems/GPU",
        "/nvsm/v1/Schemas/Systems/Memory/Errors",
        "/nvsm/v1/Schemas/Systems/Pcie/Errors",
        "/nvsm/v1/Systems/1/Memory/Alerts",
        "/nvsm/v1/Systems/1/Memory/Policy",
        "/nvsm/v1/Systems/1/Processors/Alerts",
        "/nvsm/v1/Systems/1/Processors/Policy",
        "/nvsm/v1/Systems/1/Alerts",
        "/nvsm/v1/Systems/1/GPUs",
        "/nvsm/v1/Systems/1/Lshw",
        "/nvsm/v1/Systems/1/Memory/Errors",
        "/nvsm/v1/Systems/1/Pcie/Errors",
        "/nvsm/v1/Systems/1/Policy",
        "/nvsm/v1/Systems/1/Storage/1/Alerts",
        "/nvsm/v1/Systems/1/Storage/1/Policy",
]

def main():
    handle = NVSMHandle()
    handle.connect()

    responses = []
    for endpoint in api_endpoints:
        queryEndpoint(endpoint, handle=handle, responses=responses)
        time.sleep(1)

    json.dump(responses, fp=sys.stdout, indent=4)

if __name__ == '__main__':
    main()

