# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from nvsmhealth import modules, tasks
from nvsmhealth.modules.dcs_modules.dcs_peg import dcs_peg_gen_up

import os

# These are added to modules/dump/__init__.py wheh file is located in the in the modules/dump folder.
# Only needed here if the file is located in the modules/dcs_modules folder
#
#__all__ = [
#    "pegasus_syslog",
#    "pegasus_dbglog"
#]

# Copy specified files and/or log output for specified commands from Pegasus
class PegasusLogs(tasks.ILogs):
    def __init__(self, log_list):
        super(PegasusLogs, self).__init__()
# Platform data is not set for dump health tasks, so need to hard code
#        self.__parameters_task = modules.parameter.parameters
#        self.depends(self.__parameters_task)
#        self.parameters('net_ping')
        self.__log_list = log_list

    def check_login(self, addr, uln, ulnp, as_sudo_user):
        # First build the command prefix that can be used with with either ssh or scp
        cmd_prefix = []
        # If a password was specified then add in 'sshpass' to the prefix
        if isinstance(ulnp, str) and ulnp != '':
            cmd_prefix.append('sshpass')
            cmd_prefix.append('-p')
            cmd_prefix.append(ulnp)

        # If requested to run as normal user (i.e. the $SUDO_USER) then add that into the prefix
        if as_sudo_user == True:
            cmd_prefix.append('sudo')
            cmd_prefix.append('-u')
            cmd_prefix.append('$SUDO_USER')

        # Build up the ssh command to use
        ssh_cmd =[]
        ssh_cmd.append('ssh')
        ssh_cmd.append('-o')
        ssh_cmd.append('ConnectTimeout=10')
        ssh_cmd.append('{}@{}'.format(uln, addr))
        ssh_cmd.append('exit')

        # Finally create the full run command from the prefix and the ssh command
        run_cmd = cmd_prefix + ssh_cmd
        cmd_task = tasks.RunCommand(shell=True, args=[' '.join(run_cmd)])
        cmd_task.run()

        if cmd_task.getReturnCode() == 0:
            # Login was successful, return the prefix to use for login
            # Note, it could be that the prefix is blank (i.e. == []) depending on input parameters
            return cmd_prefix

        # Return a None type if login failed
        return None

    def make_tmp_user_path(self, tmp_path):
        mkdir_cmd = []
        mkdir_cmd.append('sudo')
        mkdir_cmd.append('-u')
        mkdir_cmd.append('$SUDO_USER')
        mkdir_cmd.append('mkdir')
        mkdir_cmd.append('-p')

        mkdir_task = tasks.RunCommand(shell=True, args=[' '.join(mkdir_cmd+[tmp_path])])
        mkdir_task.run()

        # Return True upon success, False otherwise
        return (mkdir_task.getReturnCode() == 0)

    def remove_tmp_user_path(self, tmp_path):
        rmdir_cmd = []
        rmdir_cmd.append('sudo')
        rmdir_cmd.append('-u')
        rmdir_cmd.append('$SUDO_USER')
        rmdir_cmd.append('rm -rf')

        rmdir_task = tasks.RunCommand(shell=True, args=[' '.join(rmdir_cmd+[tmp_path])])
        rmdir_task.run()

        return (rmdir_task.getReturnCode() == 0)

    def run(self):
        pass

    def collect(self, context):
# Platform data is not set for dump health tasks, so need to hard code data here
#        try:
#            dcs_net_ping = self.__parameters_task.getResult()['net_ping']
#            dcs_pegasus_creds = self.__parameters_task.getResult()['dcs_pegasus_creds']
#        except:
#            return
        # import global dcs_net_ping structure from cgx.py platform file since the platform data is not defined for --dump tasks
        # Note: These are here rather than at top of file to make it clearer that they are not needed if nvsm-health changes such that
        #       they can be retrieved from the platform data as in the commented out code a bove
        from nvsmhealth.platforms.cgx import dcs_net_ping, dcs_pegasus_creds

        # Set temporary base path to use the 'normal' $SUDO_USER is used to do the 'scp'
        # Start with assumption that the 'root' user will be used
        tmp_user_base_used = False
        tmp_user_base_path = context.tempdir+'_pegasus'

        # Collect logs from both Tegra_A and Tegra_B
        for dev_name in dcs_pegasus_creds:
            # Try to get the address for this device, skip this device if address was not retrieved successfually
            dev_addr = dcs_net_ping.get(dev_name, None)
            if not dev_addr:
                continue

            # Verify credentials are available for this device.  If not then skip this device
            dev_creds = dcs_pegasus_creds.get(dev_name, None)
            if not dev_creds or not dev_creds.get('1', None) or not dev_creds.get('2', None):
                continue

            # Decode credentials
            dev_uln = dcs_peg_gen_up(dev_creds['1'])
            dev_ulnp = dcs_peg_gen_up(dev_creds['2'])

            # Create the final destination base path for the collected files for this Tegra
            dst_base_path = os.path.join(context.tempdir, 'pegasus', dev_name)
            os.makedirs(dst_base_path, exist_ok=True)

            # Since this could be running on any platform, do a quick login check, skip log collection if it fails.
            # First try using the root user since this is how nvsm-health is run
            use_sudo_user = False
            ssh_prefix = self.check_login(dev_addr, dev_uln, dev_ulnp, use_sudo_user)
            if ssh_prefix == None:
                # If login as root user failed, try as 'normal' user that ran nvsm-health ($SUDO_USER)
                use_sudo_user = True
                ssh_prefix = self.check_login(dev_addr, dev_uln, dev_ulnp, use_sudo_user)

            # If all login check attempts failed, log failure and move to the next Pegasus Tegra node.
            if ssh_prefix == None:
                dst_file = os.path.join(dst_base_path, dev_name+'.log')
                output = "Could not ssh to Pegasus {}\r\n".format(dev_name)
                output = "No log files collected from Pegasus {}\r\n".format(dev_name)
                with open(dst_file, "w") as f:
                    f.write(output)

                collect_logs = tasks.CollectLogFiles(path=dst_file)
                collect_logs.collect(context=context)
                continue

            # Login succeeded, so go forward with collecting logs from Pegasus
            for log_args in self.__log_list:
                # Get next file(s) to copy or command to run from the input list
                log_cmd = log_args.get('cmd', '')
                log_path = log_args.get('path','')
                log_file = log_args.get('file','')
                log_exclude = log_args.get('exclude','')

                src_path = log_path
                dst_path = dst_base_path+log_path
                os.makedirs(dst_path, exist_ok=True)

                # Build the run command starting with 'sudo -u $SUDO_USER' in order to run the command as normal user
                # this is done because doing ssh or scp to Pegasus as the root user is problematic.
                # SSH Keys are not always present in root user 'known_hosts' file
                run_cmd = []

                # Logging output from a command? or just copy log file?
                if use_sudo_user == False:
                    # Using the root user, so use the final location
                    dst_path_to_use = dst_path
                else:
                    # Using 'normal' $SUDO_USER so use the temporary location
                    tmp_user_base_used = True
                    dst_path_to_use = os.path.join(tmp_user_base_path, dev_name)+log_path
                    self.make_tmp_user_path(dst_path_to_use)

                if log_cmd and log_cmd != '':
                    # Generate the destination file path/name to send output to.
                    if log_file and log_file != '':
                        dst_file = os.path.join(dst_path_to_use, log_file)
                    else:
                        log_file = log_cmd+'.log'
                        dst_file = os.path.join(dst_path_to_use, log_file)

                    # Build command to run command on pegasus and log the output to the destination file
                    run_cmd.append('ssh')
                    run_cmd.append('{}@{}'.format(dev_uln, dev_addr))
                    run_cmd.append(log_cmd)
                    run_cmd.append('>')
                    run_cmd.append(dst_file)
                else:
                    # Build command to scp copy the specified files (or recursive directory copy) from Pegasus to DCS temp dir.
                    run_cmd.append('scp')
                    if log_file and log_file != '':
                        src_path = os.path.join(src_path, log_file)
                    else:
                        run_cmd.append('-r')

                    run_cmd.append('{}@{}:{}'.format(dev_uln, dev_addr, src_path))
                    run_cmd.append(dst_path_to_use)

                # Combine the login prefix with the ssh/scp command to be executed
                full_run_cmd = ssh_prefix + run_cmd

                # Run the command
                cmd_task = tasks.RunCommand(shell=True, args=[' '.join(full_run_cmd)])
                cmd_task.run()

                # Collect the files into the dump health logs if command succeeded.
                if cmd_task.getReturnCode() == 0:
                    # Now collect the local pegasus log files
                    log_path_files = os.path.join(dst_path, log_file)

                    # If using the 'normal' $SUDO_USER, need to first move the files from the
                    # temporary user directory to the final directory before collecting them in the dump tar ball
                    if use_sudo_user == True:
                        import shutil
                        files = os.listdir(dst_path_to_use)
                        for f in files:
                            src_f = os.path.join(dst_path_to_use, f)
                            shutil.move(src_f, dst_path)

                    # Collect the files into the dump file
                    collect_logs = tasks.CollectLogFiles(path=log_path_files, exclude=log_exclude)
                    collect_logs.collect(context=context)

        # If the temporary 'normal' $SUDO_USER directory was created/used then delete it now
        if tmp_user_base_used == True:
            self.remove_tmp_user_path(tmp_user_base_path)

pegasus_syslog = PegasusLogs(
    log_list=[
        {'cmd':'dmesg', 'path':'/var/log', 'file':'dmesg.log', 'exclude':r'\d\.gz\Z'},
        {'cmd':None, 'path':'/var/log', 'file':'messages*', 'exclude':r'\d\.gz\Z'},
        {'cmd':None, 'path':'/var/log', 'file':'syslog*', 'exclude':r'\d\.gz\Z'},
        {'cmd':None, 'path':'/var/log', 'file':'kern.log*', 'exclude':r'\d\.gz\Z'},
        {'cmd':None, 'path':'/var/log', 'file':'kernel.log*', 'exclude':r'\d\.gz\Z'},
    ])
# Predicate checks aren't supported anymore for dump health tasks
#    .parameters('enable_dcv_checks')
#    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

# Log any outstanding data on the Pegasus USB ports.
class PegasusDebugLog(tasks.ILogs):
    def __init__(self):
        super(PegasusDebugLog, self).__init__()
# Platform data is not set for dump health tasks, so need to hard code
#        self.__parameters_task = modules.parameter.parameters
#        self.depends(self.__parameters_task)
#        self.parameters('dcs_usb_serial_cfg')

    def run(self):
        pass

    def collect(self, context):
        import serial, time

# Platform data is not set for dump health tasks, so need to hard code
#        try:
#            dcs_usb_serial_cfg = self.__parameters_task.getResult()['dcs_usb_serial_cfg']
#        except:
#            return
        # import global dcs_usb_serial_cfg structure from cgx.py platform file since the platform data is not defined for --dump tasks
        # Note: These are here rather than at top of file to make it clearer that they are not needed if nvsm-health changes such that
        #       they can be retrieved from the platform data as in the commented out code a bove
        from nvsmhealth.platforms.cgx import dcs_usb_serial_cfg

        dst_path = os.path.join(context.tempdir, 'pegasus')
        os.makedirs(dst_path, exist_ok=True)
        for interface in dcs_usb_serial_cfg:
            dev_parms = dcs_usb_serial_cfg[interface]
            dev_port = dev_parms['port']
            baudrate = dev_parms['baudrate']

            if not os.path.exists(dev_port):
                continue

            log_filename = "{}_debug.log".format(interface)
            dst_file = os.path.join(dst_path, log_filename)

            header = "Interface: {}\r\n".format(interface)
            header += "Port: {}\r\n".format(dev_port)
            header += "Baudrate: {}\r\n".format(baudrate)
            header += "\r\n"
            header += "Log:\r\n"
            header += "\r\n"

            # Track if connection is successful. If not, assume it's because this is not a CGX platform and don't log anything.
            output = ""
            with serial.Serial(dev_port, baudrate, timeout=1) as ser:
                # First read anything currently waiting in the buffer up to 500 lines (avoid infinite loop)
                max_line_cnt = 500
                while ser.inWaiting() > 0 and max_line_cnt > 0:
                    max_line_cnt -= 1
                    output += str(ser.readline(), 'utf-8')

                # Send a CR/LF and wait a second for any data to come back
                ser.write(b'\r\n')
                time.sleep(1)

                # Read back anything wating in the buffer up to 500 lines (avoid infinite loop)
                max_line_cnt = 500
                while ser.inWaiting() > 0 and max_line_cnt > 0:
                    max_line_cnt -= 1
                    output += str(ser.readline(), 'utf-8')

            if not output:
                output = "No data logged"

            with open(dst_file, "w") as f:
                f.write(header)
                f.write(output)

            collect_logs = tasks.CollectLogFiles(path=dst_file)
            collect_logs.collect(context=context)

pegasus_dbglog = PegasusDebugLog()
# Predicate checks aren't supported anymore for dump health tasks
#    .parameters('enable_dcv_checks')
#    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_dcv_checks'))

