# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import re
import json
from nvsmhealth import modules, tasks, config
__all__ = [
    'collect_fru',
    'parse_fru',
]

collect_fru = tasks.RunCommand(
        args=['ipmitool', 'fru', 'print']) \
        .title('Run ipmitool fru print command') \
        .describe('''
                This runs the "ipmitool fru" command to obtain FRU (field
                replaceable unit) information from the BMC (baseboard management
                controller). FRU information is important for keeping inventory of
                the components installed on the system and their serial
                numbers.''') \
        .depends(modules.common.check_superuser_privileges)

class ParseIpmiFru(tasks.ITask):
    def __init__(self, run_ipmi_fru_task):
        super().__init__()
        self.__run_ipmi_fru_task = run_ipmi_fru_task
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):

        if self.__run_ipmi_fru_task.getReturnCode():
            return


        output = self.__run_ipmi_fru_task.getOutput().strip()

        fru_pattern = re.compile(
                r'^FRU Device Description\s*:\s*(?P<fru_name>[^\(]*)\((?P<fru_id>[^\)\n]*)\)?\n(?P<fru_body>(?:[^\n]*\S+[^\n]*\n?)+)',
                flags=re.MULTILINE)

        # Attempt to parse FRU values
        try:
            result = {}
            for m in fru_pattern.finditer(output):
                fru = {}
                fru_name = m.group('fru_name').strip()
                result[fru_name] = fru
                for line in m.group('fru_body').split('\n'):
                    line = line.strip()
                    if not line:
                        continue
                    field = line.split(':', 1)
                    if len(field) != 2:
                        continue
                    field_name = field[0].strip().lower().replace(' ', '_')
                    field_value = field[1].strip()
                    fru[field_name] = field_value
        except Exception as e:
            print("EXCEPTION:", e)
            pass

        self.__result = result

parse_fru = ParseIpmiFru(collect_fru) \
        .title('Parse output from ipmitool fru print') \
        .depends(collect_fru)
