import re
import json
import logging
from nvsmhealth import modules, tasks, config

__all__ = [
    'parse_firmware_revision',
    'firmware_revision_info',
    'run_ipmi_sensor',
    'parse_ipmi_sensor',
    'check_ipmi_sensor_thresholds',
    'run_ipmi_sdr_elist',
    'parse_ipmi_sdr_elist',
    'sdr_device_bom',
    'check_fan_bom',
    'check_psu_bom',
    'run_ipmi_getenables',
    'parse_ipmi_getenables',
    'run_ipmi_fru',
    'parse_ipmi_fru',
    'check_fru_consistency'
]

class ParseFirmwareRevision(tasks.ITask):
    def __init__(self, parse_ipmi_info_task):
        super(ParseFirmwareRevision, self).__init__()
        self.__parse_ipmi_info_task = parse_ipmi_info_task
        self.__parameters_task = modules.parameter.parameters
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        version = "UNKNOWN"
        try:
            version = self.__parse_ipmi_info_task.getResult()['firmware_revision']
            is_type = self.__parameters_task.getResult()['aux_firmware_version_type']
            if is_type == 'hex':
                # Best effort - try to convert hex to decimal
                try:
                    ver = version.split(".")
                    dec = int(ver[2], 16)
                    # Only format the version when greater than '09'
                    if dec > 9:
                        ver[2] = "{}".format(dec)
                        ver = ".".join(ver)
                        version = ver
                except Exception as ex:
                    logging.debug('Error while converting hex to decimal: {}'.format(ex))
                    pass
        except:
            # Skip conversion
            pass

        self.__result =  version

parse_firmware_revision = ParseFirmwareRevision(modules.ipmitool_cmd.parse_ipmi_info) \
        .title('Parse BMC Firmware Revision') \
        .depends(modules.ipmitool_cmd.parse_ipmi_info) \
        .depends(modules.parameter.parameters) \
        .parameters('aux_firmware_version_type')

class FirmwareRevisionInfo(tasks.ITask):
    def __init__(self, parse_firmware_revision_task):
        super(FirmwareRevisionInfo, self).__init__()
        self.__parse_firmware_revision_task = parse_firmware_revision_task
        self.__parameters_task = modules.parameter.parameters
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        try:
            if self.__parameters_task.getResult()['skip_bmc_revision_check']:
                return
        except:
            pass

        firmware_revision = "UNKNOWN"
        try:
            firmware_revision = self.__parse_firmware_revision_task.getResult()
        except:
            pass

        #TODO: Bug[200383435] - fix the defect after platform abstraction work
        self.send(msg=firmware_revision)
        self.__result = firmware_revision

firmware_revision_info = FirmwareRevisionInfo(parse_firmware_revision) \
        .title('BMC Firmware Revision') \
        .depends(parse_firmware_revision) \
        .depends(modules.parameter.parameters) \
        .parameters('skip_bmc_revision_check')

run_ipmi_sensor = tasks.RunCommand(
        args=['ipmitool', 'sensor']) \
        .title('Run ipmitool sensor command') \
        .describe('''
                Check BMC sensor status with ipmitool.
                This command requires superuser privileges.''') \
        .depends(modules.ipmitool_cmd.parse_ipmi_info) \
        .depends(modules.common.check_superuser_privileges) \
        .depends(modules.parameter.parameters)

class ParseIpmiSensor(tasks.ITask):
    def __init__(self, ipmi_sensor_task):
        super(ParseIpmiSensor, self).__init__()
        self.__ipmi_sensor_task = ipmi_sensor_task
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        if self.__ipmi_sensor_task.getReturnCode():
            # command failed
            return
        res = {}
        output = self.__ipmi_sensor_task.getOutput().strip()

        sensor_pattern = re.compile(
                r'^(?P<name>[^|\n]+)\|' \
                r'(?P<current_reading>[^|\n]+)\|' \
                r'(?P<type>[^|\n]+)\|' \
                r'(?P<status>[^|\n]+)\|' \
                r'(?P<lower_non_recoverable>[^|\n]+)\|' \
                r'(?P<lower_critical>[^|\n]+)\|' \
                r'(?P<lower_non_critical>[^|\n]+)\|' \
                r'(?P<upper_non_critical>[^|\n]+)\|' \
                r'(?P<upper_critical>[^|\n]+)\|' \
                r'(?P<upper_non_recoverable>[^|\n]+)' ,
                flags=re.MULTILINE)

        # Attempt to parse the table of sensor values
        sensors = []
        try:
            for m in sensor_pattern.finditer(output):
                s = {
                    'name': m.group('name').strip(),
                    'status': m.group('status').strip(),
                    'type': m.group('type').strip(),
                    'current_reading': m.group('current_reading').strip(),
                    'lower_non_recoverable': m.group('lower_non_recoverable').strip(),
                    'lower_critical': m.group('lower_critical').strip(),
                    'lower_non_critical': m.group('lower_non_critical').strip(),
                    'upper_non_recoverable': m.group('upper_non_recoverable').strip(),
                    'upper_critical': m.group('upper_critical').strip(),
                    'upper_non_critical': m.group('upper_non_critical').strip()
                }
                sensors.append(s)
        except:
            pass

        self.__result = sensors

parse_ipmi_sensor = ParseIpmiSensor(run_ipmi_sensor) \
        .title('Parse output from ipmitool sensor') \
        .depends(run_ipmi_sensor)

class CheckSensorThresholds(tasks.ICheck):
    def __init__(self, parse_ipmi_sensor_task):
        super(CheckSensorThresholds, self).__init__()

        self.__parse_ipmi_sensor_task = parse_ipmi_sensor_task

    def run(self):
        threshold_check_dispatch = {
            'lower_non_recoverable': lambda observed, threshold: threshold < observed,
            'lower_critical': lambda observed, threshold: threshold < observed,
            'upper_non_recoverable': lambda observed, threshold: observed < threshold,
            'upper_critical': lambda observed, threshold: observed < threshold
        }

        threshold_display = {
            'lower_non_recoverable': '{name}: Observed value "{observed}" ({units}) below non-recoverable lower threshold "{threshold}"',
            'lower_critical': '{name}: Observed value "{observed}" ({units}) below critical lower threshold "{threshold}"',
            'upper_non_recoverable': '{name}: Observed value "{observed}" ({units}) above non-recoverable upper threshold "{threshold}"',
            'upper_critical': '{name}: Observed value "{observed}" ({units}) above critical upper threshold "{threshold}"'
        }

        # Look for any sensor values that fall outside of critical thresholds
        try:
            healthy = True
            sensors = self.__parse_ipmi_sensor_task.getResult()
            if not sensors:
                return
            for sensor in sensors:
                name = sensor['name']
                observed = sensor['current_reading']
                try:
                    observed = float(observed)
                except:
                    continue
                units = sensor['type']
                if units.lower() == 'discrete':
                    continue
                for field in [
                        'lower_non_recoverable',
                        'lower_critical',
                        'upper_non_recoverable',
                        'upper_critical' ]:
                    threshold = sensor.get(field)
                    try:
                        threshold = float(threshold)
                    except:
                        continue
                    check = threshold_check_dispatch[field]
                    if check(observed, threshold):
                        continue  # Observed value is within threshold
                    healthy = False
                    display = threshold_display[field]
                    self.addCheckMessage(display.format(
                            name=name,
                            observed=observed,
                            units=units,
                            threshold=threshold))
            self.addCheckMessage('Checked {count} sensor values against BMC thresholds.'.format(
                 count=len(sensors)))

            if healthy:
                self.addHealthy()
            else:
                self.addUnHealthy()

        except:
            self.addCheckMessage('No sensors found in "ipmitool sensor"')
            self.addUnknown()
            return

check_ipmi_sensor_thresholds = CheckSensorThresholds(parse_ipmi_sensor) \
    .title('''Check BMC sensor thresholds''') \
    .depends(parse_ipmi_sensor)

run_ipmi_sdr_elist = tasks.RunCommand(
        args=['ipmitool', 'sdr', 'elist']) \
        .title('Run ipmitool sdr elist command') \
        .describe('''
                Check BMC bom devices with ipmitool.
                This command requires superuser privileges.''') \
        .depends(modules.common.check_superuser_privileges)

class ParseIpmiSdrElist(tasks.ITask):
    def __init__(self, ipmi_sdr_elist_task):
        super(ParseIpmiSdrElist, self).__init__()
        self.__ipmi_sdr_elist_task = ipmi_sdr_elist_task
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        if self.__ipmi_sdr_elist_task.getReturnCode():
            return

        res = {}
        output = self.__ipmi_sdr_elist_task.getOutput().strip()

        sensor_pattern = re.compile(
                r'^(?P<name>[^|\n]+)\|' \
                r'(?P<id>[^|\n]+)\|' \
                r'(?P<status>[^|\n]+)\|' \
                r'(?P<entity_id>[^|\n]+)\|' \
                r'(?P<reading>[^|\n]+)' ,
                flags=re.MULTILINE)


        try:
            sensors = []
            for m in sensor_pattern.finditer(output):
                s = {
                    'name': m.group('name').strip(),
                    'id': m.group('id').strip(),
                    'status': m.group('status').strip(),
                    'entity_id': m.group('entity_id').strip(),
                    'reading': m.group('reading').strip()
                }
                sensors.append(s)
            self.__result = sensors
        except:
            pass


parse_ipmi_sdr_elist = ParseIpmiSdrElist(run_ipmi_sdr_elist) \
        .title('Parse output from ipmitool sdr elist') \
        .depends(modules.ipmitool_cmd.parse_ipmi_info) \
        .depends(run_ipmi_sdr_elist)

class SdrDeviceBom(tasks.ITask):
    def __init__(self, parse_ipmi_sdr_elist_task):
        super().__init__()
        self.__parse_ipmi_sdr_elist_task = parse_ipmi_sdr_elist_task
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        # Create dictionary containing devices that appear to be present
        try:
            bom = {}
            ns_pattern = re.compile(r'^[nN]/?[sS]$')
            for s in self.__parse_ipmi_sdr_elist_task.getResult():
                name = s['name']
                status = s['status']
                reading = s.get('reading')
                if not ns_pattern.match(status) and reading:
                    # Sensor status is not ns and reading is not empty, so device appears to be present
                    bom[name] = True
        except:
            pass

        self.__result = bom

sdr_device_bom = SdrDeviceBom(parse_ipmi_sdr_elist) \
        .title('Prepare Device Bom from ipmitool sdr elist') \
        .depends(parse_ipmi_sdr_elist)

class CheckFanBom(tasks.ICheck):
    def __init__(self, sdr_device_bom_task):
        super().__init__()
        self.__sdr_device_bom_task = sdr_device_bom_task
        self.__result = None
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('fan_bom')

    def run(self):
        from nvsmhealth.lib import DictionarySuperset
        try:
            fan_bom = self.__parameters_task.getResult()['fan_bom']
        except:
            return

        output = self.__sdr_device_bom_task.getResult()
        dictionary_superset = DictionarySuperset.DictionarySuperset(
                    missing_message="Could not detect presence of chassis fan {}")
        result = dictionary_superset.compare(output, fan_bom)
        self.addCheckMessage("Checking output of 'ipmitool sdr elist' for expected chassis fans")
        if result:
            self.addCheckMessage(result)
            self.addUnHealthy()
        else:
            self.addHealthy()

check_fan_bom = CheckFanBom(sdr_device_bom) \
        .title('Verify chassis fan presence') \
        .depends(sdr_device_bom)

class CheckPsuBom(tasks.ICheck):
    def __init__(self, sdr_device_bom_task, parse_ipmi_sdr_elist_task):
        super().__init__()
        self.__sdr_device_bom_task = sdr_device_bom_task
        self.__parse_ipmi_sdr_elist_task = parse_ipmi_sdr_elist_task
        self.__result = None
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('psu_bom')

    def run(self):
        from nvsmhealth.lib import DictionarySuperset
        try:
            psu_bom = self.__parameters_task.getResult()['psu_bom']
        except:
            return

        output = self.__sdr_device_bom_task.getResult()
        dictionary_superset = DictionarySuperset.DictionarySuperset(
                    missing_message="Could not detect presence of chassis power supply {}")
        result = dictionary_superset.compare(output, psu_bom)
        self.addCheckMessage("Checking output of 'ipmitool sdr elist' for expected chassis PSUs")
        if result:
            self.addCheckMessage(result)
            self.addUnHealthy()
        else:
            healthy = True
            # NVBUG-200528273: Check for Power supply lost
            # Even if PSU status is ok, readings might have power supply ac lost message
            psu_res = self.__parse_ipmi_sdr_elist_task.getResult()
            # Filter PSU status keys for readings
            psu_status_keys = [k for k in psu_bom.keys() if 'status' in k.lower()]
            for s in psu_res:
                if s['name'] in psu_status_keys:
                    reading = s['reading']
                    if 'power supply ac lost' in reading.lower():
                        self.addCheckMessage("AC input is lost, {} has reading:\n{}".format(s['name'], s['reading']))
                        healthy = False

            if healthy:
                self.addHealthy()
            else:
                self.addUnHealthy()

check_psu_bom = CheckPsuBom(sdr_device_bom, parse_ipmi_sdr_elist) \
        .title('Verify chassis power supply presence') \
        .depends(parse_ipmi_sdr_elist) \
        .depends(sdr_device_bom)

run_ipmi_getenables = tasks.RunCommand(
        args=['ipmitool', 'mc', 'getenables']) \
        .title('Run ipmitool mc getenables command') \
        .describe('''
                Check BMC status with ipmitool.
                This command requires superuser privileges.''') \
        .depends(modules.ipmitool_cmd.parse_ipmi_info) \

class ParseIpmiGetEnables(tasks.ITask):
    def __init__(self, run_ipmi_getenables_task):
        super().__init__()
        self.__run_ipmi_getenables_task = run_ipmi_getenables_task
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        if self.__run_ipmi_getenables_task.getReturnCode():
            return

        res = {}
        output = self.__run_ipmi_getenables_task.getOutput().strip()

        try:
            ipmi_getenables_pattern = re.compile(r"^Receive Message Queue Interrupt\s*:\s*(?P<receive_message_queue_interrupt>.*)\nEvent Message Buffer Full Interrupt\s*:\s*(?P<event_message_buffer_full_interrupt>.*)\nEvent Message Buffer\s*:\s*(?P<event_message_buffer>.*)\nSystem Event Logging\s*:\s*(?P<system_event_logging>.*)\nOEM 0\s*:\s*(?P<oem_0>.*)\nOEM 1\s*:\s*(?P<oem_1>.*)\nOEM 2\s*:\s*(?P<oem_2>.*)")

            match = ipmi_getenables_pattern.match(output)
            if match:
                res = match.groupdict()
        except Exception as e:
            print("EXCEPTION:", e)
            pass

        self.__result = res

parse_ipmi_getenables = ParseIpmiGetEnables(run_ipmi_getenables) \
        .title('Parse output from ipmitool get enables') \
        .depends(run_ipmi_getenables)

run_ipmi_fru = tasks.RunCommand(
        args=['ipmitool', 'fru', 'print']) \
        .title('Run ipmitool fru print command') \
        .describe('''
                This runs the "ipmitool fru" command to obtain FRU (field
                replaceable unit) information from the BMC (baseboard management
                controller). FRU information is important for keeping inventory of
                the components installed on the system and their serial
                numbers.''') \
        .depends(modules.ipmitool_cmd.parse_ipmi_info)

class ParseIpmiFru(tasks.ITask):
    def __init__(self, run_ipmi_fru_task):
        super().__init__()
        self.__run_ipmi_fru_task = run_ipmi_fru_task
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        if self.__run_ipmi_fru_task.getReturnCode():
            return

        output = self.__run_ipmi_fru_task.getOutput().strip()

        fru_pattern = re.compile(
                r'^FRU Device Description\s*:\s*(?P<fru_name>[^\(]*)\((?P<fru_id>[^\)\n]*)\)?\n(?P<fru_body>(?:[^\n]*\S+[^\n]*\n?)+)',
                flags=re.MULTILINE)

        # Attempt to parse FRU values
        try:
            result = {}
            for m in fru_pattern.finditer(output):
                fru = {}
                fru_name = m.group('fru_name').strip()
                result[fru_name] = fru
                for line in m.group('fru_body').split('\n'):
                    line = line.strip()
                    if not line:
                        continue
                    field = line.split(':', 1)
                    if len(field) != 2:
                        continue
                    field_name = field[0].strip().lower().replace(' ', '_')
                    field_value = field[1].strip()
                    fru[field_name] = field_value
        except Exception as e:
            print("EXCEPTION:", e)
            pass

        self.__result = result

parse_ipmi_fru = ParseIpmiFru(run_ipmi_fru) \
        .title('Parse output from ipmitool fru print') \
        .depends(run_ipmi_fru)

class CheckFRUConsistency(tasks.ICheck):
    def __init__(self, fru_task, dmidecode_task):
        super(CheckFRUConsistency, self).__init__()

        # Declare dependency on fru using the ITask interface
        self.__fru_task = fru_task
        self.__dmidecode_task = dmidecode_task
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__fru_task)
        self.depends(self.__dmidecode_task)
        self.depends(self.__parameters_task)

    def run(self):
        own_params = self.getParameters()
        params = self.__parameters_task.getResult()

        fru_devices = None
        for param, req in own_params.items():
            if param in params.keys():
                fru_devices = params[param]

        if fru_devices == None:
            return

        # Unknown check for parse tasks failure
        if not (self.__fru_task.getResult() and self.__dmidecode_task.getResult()):
            self.addCheckMessage("No results from 'ipmitool fru print' or 'dmidecode' commands")
            self.addUnknown()
            return

        result = "healthy"
        self.addCheckMessage(config.fru_command_str)

        try:
            fru_res = self.__fru_task.getResult()
            # Check for FRU devices
            devices_not_found = [device for device in fru_devices if device not in fru_res.keys()]
            if devices_not_found:
                self.addCheckMessage("FRU devices not found '{}'".format(", ".join(devices_not_found)))
                if len(devices_not_found) == len(fru_devices):
                    self.addUnHealthy()
                    return
                result = "unhealthy"

            devices_found = [device for device in fru_devices if device not in devices_not_found]
            # Check for FRU devices chassis serial number
            chassis_serial_not_found = [device for device in devices_found if 'chassis_serial' not in fru_res[device].keys()]
            if chassis_serial_not_found:
                self.addCheckMessage("Chassis serial number not found for FRU devices '{}'".format(", ".join(chassis_serial_not_found)))
                result = "unhealthy"

            chassis_serial_found = [device for device in devices_found if device not in chassis_serial_not_found]

            # Get expected serial number
            dmidecode_res = self.__dmidecode_task.getResult()
            chassis_info = [v['serial_number'] for k, v in dmidecode_res.items() if 'chassis information' in k.lower() and
                            'serial_number' in v.keys()]
            if chassis_info:
                expected_serial_number = chassis_info[0]
            else:
                self.addCheckMessage("Failed while fetching serial number from chassis information")
                self.addUnknown()
                return

            # Check and print the FRU devices having inconsistent chassis serial numbers
            diff = [device for device in chassis_serial_found if fru_res[device]['chassis_serial'] != expected_serial_number]
            for device in diff:
                self.addCheckMessage("FRU device '{}' got chassis serial '{}' whereas expected is '{}'"
                                        .format(device, fru_res[device]['chassis_serial'], expected_serial_number))
                # For change in FRU chassis serial print informational status
                result = "info"

            if result == "unhealthy":
                self.addUnHealthy()
            elif result == "info":
                self.addInformational()
            else:
                self.addHealthy()

        except:
            self.addCheckMessage("Failed while checking FRU serial number consistency")
            self.addUnknown()

check_fru_consistency = CheckFRUConsistency(parse_ipmi_fru, modules.dmidecode.parse_dmidecode) \
    .title('''Check FRU information for consistency''') \
    .describe('''
            The FRU (field replaceable unit) information recorded in the BMC
            (baseboard management controller) includes serial numbers for
            various FRUs on the system. For any given system, these serial
            numbers should be consistent among all FRUs. However, it is
            possible for these serial numbers to become inconsistent as the
            result of normal maintenance (such as FRU replacement). This check
            makes sure serial numbers are consistent for all FRUs recorded in
            the BMC.''') \
    .parameters('fru_devices')
