import re
import json
from nvsmhealth import modules, tasks, config

__all__ = [
    'run_ipmi_info',
    'parse_ipmi_info'
]

run_ipmi_info = tasks.RunCommand(
    args=['ipmitool', 'mc', 'info']) \
    .title('Run ipmitool mc info command') \
    .describe('''
                Check BMC status with ipmitool.
                This command requires superuser privileges.''') \
    .depends(modules.common.check_superuser_privileges)


class ParseIpmiInfo(tasks.ITask):
    def __init__(self, ipmi_info_task):
        super(ParseIpmiInfo, self).__init__()
        self.__ipmi_info_task = ipmi_info_task
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        if self.__ipmi_info_task.getReturnCode():
            # command failed
            return
        res = {}
        output = self.__ipmi_info_task.getOutput()

        ipmi_info_pattern = re.compile(r"^Device ID\s*:\s*(?P<device_id>.*)\nDevice Revision\s*:\s*(?P<device_revision>.*)\nFirmware Revision\s*:\s*(?P<firmware_revision>.*)\nIPMI Version\s*:\s*(?P<ipmi_version>.*)\nManufacturer ID\s*:\s*(?P<manufacturer_id>.*)\nManufacturer Name\s*:\s*(?P<manufacturer_name>.*)\nProduct ID\s*:\s*(?P<product_id>.*)\nProduct Name\s*:\s*(?P<product_name>.*)\nDevice Available\s*:\s*(?P<device_available>.*)\nProvides Device SDRs\s*:\s*(?P<provides_device_sdrs>.*)\nAdditional Device Support\s*:\s*\n(?P<additional_device_support>(.*\n)*)Aux Firmware Rev Info\s*:\s*\n(?P<aux_firmware_rev_info>(.*\n)*)")

        match = ipmi_info_pattern.match(output)
        if match:
            res = match.groupdict()

            res['additional_device_support'] = res['additional_device_support'].split(
                '\n')
            res['additional_device_support'] = [a.strip()
                                                for a in res['additional_device_support'] if a]

            # BUG-2547367: Use Byte 13 or aux[0] (Most significant byte of Auxiliary firmware revision information)
            # to update firmware revision number
            if res['manufacturer_id'] in [config.quanta_manufacturer_id, config.nvidia_manufacturer_id]:
                rev_pattern = re.compile(
                    r"^\s*0x(?P<hex1>.*)\n\s*0x(?P<hex2>.*)\n\s*0x(?P<hex3>.*)\n\s*0x(?P<hex4>.*)")
                rev_match = rev_pattern.match(res['aux_firmware_rev_info'])
                if rev_match:
                    rev = rev_match.groupdict()
                    res['firmware_revision'] = '{}.{}'.format(
                        res['firmware_revision'], rev['hex1'])

            res['aux_firmware_rev_info'] = res['aux_firmware_rev_info'].split(
                '\n')
            res['aux_firmware_rev_info'] = [a.strip()
                                            for a in res['aux_firmware_rev_info'] if a]

        self.__result = res


parse_ipmi_info = ParseIpmiInfo(run_ipmi_info) \
    .title('Parse output from ipmitool mc info') \
    .depends(run_ipmi_info)
