# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import json
import os
import re
import sys
import logging

from nvsmhealth import modules, tasks

__all__ = [
    'run_lsblk_scsi_device_info',
    'parse_lsblk_scsi_device_info'
]

run_lsblk_scsi_device_info = tasks.RunCommand(
    # -S Output info about SCSI devices only,
    # -P Produce output in the form of key="value" pairs,
    # -o Specify which output columns to print
        args=['lsblk', '-S', '-P', '-o', 'NAME,HCTL,TYPE,VENDOR,MODEL,REV,TRAN']) \
    .title('''Run the lsblk utility''') \
    .describe('''
            Run the "lsblk" utility to get info for scsi block devices.
            Get the output in json format.''') \

class ParseLsblkSCSIDeviceInfo(tasks.ITask):
    def __init__(self, lsblk_task):
        super(ParseLsblkSCSIDeviceInfo, self).__init__()

        # Declare dependency on lsblk using the ITask interface
        self.__lsblk_task = lsblk_task
        self.depends(run_lsblk_scsi_device_info)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        # Parse the result as a JSON object from parent task
        if self.__lsblk_task.getReturnCode():
            return

        if not self.__lsblk_task.getOutput():
            return

        try:
            output = self.__lsblk_task.getOutput()
            # Filter physical scsi devices
            devices = []
            info_regex = re.compile(r'^(NAME=\"(?P<name>[^\"]*)).*(HCTL=\"(?P<hctl>[^\"]*)).*(TYPE=\"(?P<type>[^\"]*)).*(VENDOR=\"(?P<vendor>[^\"]*)).*(MODEL=\"(?P<model>[^\"]*)).*(REV=\"(?P<rev>[^\"]*)).*(TRAN=\"(?P<tran>[^\"]*))?$')
            # Split each line
            output_split = output.splitlines()
            # Run regex pattern on each line
            for device in output_split:
                m = info_regex.match(device)
                if m:
                    if m.group() is not None and 'vdisk' in m.group('model').strip().lower():
                        continue
                    device_data = { k:v.strip() if v is not None else "" for k, v in m.groupdict().items()}
                    devices.append(device_data)
            self.__result = devices

        except Exception as e:
            logging.error("ERROR: Failed to parse 'lsblk -S -P -o NAME,HCTL,TYPE,VENDOR,MODEL,REV,TRAN' information: {}".format(str(e)))

parse_lsblk_scsi_device_info = ParseLsblkSCSIDeviceInfo(run_lsblk_scsi_device_info) \
    .title('''Parse lsblk scsi block devices information gathered from lsblk tool''')
