import re
import json
from nvsmhealth import modules, tasks, config

__all__ = [
    'run_lscpu',
    'parse_lscpu',
    'check_logical_core_count'
]

run_lscpu = tasks.RunCommand(
        args=['lscpu']) \
        .title('Run lscpu command') \
        .describe('''
                Verify hyperthreading and NUMA are enabled''') \

class ParseLsCpu(tasks.ITask):
    def __init__(self, lscpu_cmd_task):
        super().__init__()
        self.__lscpu_cmd_task = lscpu_cmd_task
        self.depends(lscpu_cmd_task)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        res = {}
        output = self.__lscpu_cmd_task.getOutput().strip()

        keys = [('hyperthread', "Thread\(s\) per core:\s+(?P<{}>\d+)"),
                ('NUMA', "NUMA node\(s\):\s+(?P<{}>\d+)"),
                ('CPU', "CPU\(s\):\s+(?P<{}>\d+)"),
                ('core_per_socket', "Core\(s\) per socket:\s+(?P<{}>\d+)"),
                ('cpu_flags', "Flags:\s*(?P<{}>.*)")]

        for key, pattern in keys:
            match_string = pattern.format(key)
            p = re.compile(r"{}".format(match_string))
            for m in p.finditer(output):
                if key in res.keys():
                    # consider first match only
                    continue
                res[key] = m.group(key).strip()
                try:
                    res[key] = eval(res[key])
                except:
                    pass
        self.__result = res

parse_lscpu = ParseLsCpu(run_lscpu) \
        .title('Parse output from lscpu') \

class CheckLogicalCoreCount(tasks.ICheck):
    def __init__(self, parse_lscpu_task):
        super().__init__()
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.__parse_lscpu_task = parse_lscpu_task
        self.depends(parse_lscpu_task)
        self.parameters('lscpu_number_of_cores')

    def run(self):
        output = self.__parse_lscpu_task.getResult()

        observed_core_count = output['CPU']
        self.title(self.getTitle().format(observed_core_count))

        try:
            expected_core_count = self.__parameters_task.getResult()['lscpu_number_of_cores']
        except:
            return

        if observed_core_count == expected_core_count:
            self.addCheckMessage('Observed {0} logical CPU cores, matching expectations'.format(
                        observed_core_count))
            self.addHealthy()
            return

        self.addCheckMessage('Observed {0} logical CPU cores when {1} cores were expected'.format(
                    observed_core_count, expected_core_count))
        self.addUnHealthy()

        if observed_core_count * 2 == expected_core_count:
            # When only half of the expected logical cores are observed, we
            # suspect hyperthreading might be disabled

            # Look for the hyperthreading flag
            hyperthreading_enabled = output['hyperthread'] == 2

            if not hyperthreading_enabled:
                self.addCheckMessage('It appears that Hyper-Threading is disabled.' \
                      ' Some customers choose to disable Hyper-Threading in' \
                  ' order to improve the performance of certain' \
                  ' workloads. If Hyper-Threading was intentionally' \
                  ' disabled, please ignore this message.')

check_logical_core_count = CheckLogicalCoreCount(parse_lscpu) \
        .title('Number of logical CPU cores [{0}]')
