# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import re
import json
from nvsmhealth import tasks
from nvsmhealth.modules.common import VariantRef, check_superuser_privileges

__all__ = [
    'run_lspci',
    'run_lspci_n',
    'run_lspci_verbose',
    'parse_lspci',
    'parse_lspci_n',
    'parse_lspci_verbose'
]

run_lspci = tasks.RunCommand(
        args=['lspci', '-vmm', '-nn']) \
    .title('''Run the lspci command''') \
    .describe('''
            Run the "lspci" command to list PCI devices. Some flags are added
            such that lspci output is printed in a machine-readable format.
            This command does not require superuser privileges.''')

run_lspci_n = tasks.RunCommand(
        args=['lspci', '-vmm', '-n']) \
    .title('''Run the lspci command''') \
    .describe('''
            Run the "lspci" command to list PCI devices. Some flags are added
            such that lspci output is printed in a machine-readable format.
            This command does not require superuser privileges.''')

run_lspci_verbose = tasks.RunCommand(
        args=['lspci', '-vvv', '-nn', '-D']) \
    .title('''Run the lspci command with verbose flags''') \
    .describe('''
            Run the "lspci" command with verbose flags to show detailed
            information about PCI devices. This command requires superuser
            privileges in order to read privileged PCI device registers. Much
            of the verbose output from lspci is not necessarily in a
            machine-readable format.''')
#    .depends(check_superuser_privileges)

class ParseLspci(tasks.ITask):
    def __init__(self, lspci_task):
        super(ParseLspci, self).__init__()
        self.__lspci_task = lspci_task
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        string_id_pattern = re.compile(r'^\s*(?P<field>[^:]*):\s*(?P<value>(?P<string>.*)\s*\[(?P<id>[0-9a-fA-F]{4})\])\s*$')
        no_id_pattern = re.compile(r'^\s*(?P<field>[^:]*):\s*(?P<value>.*)\s*$')

        output = self.__lspci_task.getOutput()
        devices = output.split('\n\n')

        res = {}
        for device in devices:
            if device == '':
                continue

            fields = device.split('\n')

            # The first field is guaranteed to be the slot
            m = no_id_pattern.match(fields.pop(0))
            if not m:
                raise RuntimeError("Unable to parse the output of lspci")
            slot = m.group('value').strip()

            res[slot] = {}

            for field in fields:
                # First look for the ID
                m = string_id_pattern.match(field)
                if m:
                    field = m.group('field').lower()
                    res[slot][field] = m.group('value').strip()
                    res[slot]['{0}_string'.format(field)] = m.group('string').strip()
                    res[slot]['{0}_id'.format(field)] = m.group('id')
                    continue
                # If the ID cannot be found, use the entire field value
                m = no_id_pattern.match(field)
                if m:
                    field = m.group('field').lower()
                    res[slot][field] = m.group('value').strip()
                    continue
                raise RuntimeError("Unable to parse the output of lspci")

        # Add the parsed result as JSON object
        self.__result = json.dumps(res, ensure_ascii=True)
        return res

parse_lspci = ParseLspci(run_lspci) \
    .title('''Parse output from the lspci command''') \
    .depends(run_lspci)

parse_lspci_n = ParseLspci(run_lspci_n) \
    .title('''Parse output from the lspci command''') \
    .depends(run_lspci_n)

class ParseLspciVerbose(tasks.ITask):
    def __init__(self, lspci_task):
        super(ParseLspciVerbose, self).__init__()
        self.__lspci_task = lspci_task
        self.depends(lspci_task)
        self.__result = None

    def getResult(self):
        return self.__result

    def parse_lspci(self, output, lspciValue):
        record_pattern = re.compile(r'^(?P<record_heading>(?P<pci_domain>[0-9a-fA-F]{4}):(?P<pci_bus>[0-9a-fA-F]{2}):(?P<pci_slot>[0-9a-fA-F]{2})\.(?P<pci_function>\d)\s+.*)\n(?P<record_body>(\t.*\n)+)', flags=re.MULTILINE)
        dictionary = {}
        # Iterate over each record
        for m in re.finditer(record_pattern, output):
            record_body = m.group('record_body')
            # Add the record value to the output dictionary
            record_name = '{0}:{1}:{2}.{3}'.format(
                    m.group('pci_domain'),
                    m.group('pci_bus'),
                    m.group('pci_slot'),
                    m.group('pci_function'))
            dictionary[record_name] = {
              "heading": m.group('record_heading'),
              "body": record_body
            }
        lspciValue.dict = dictionary

    def run(self):
        # If lspci verbose got failure run got a failure, skip parsing the output
        if self.__lspci_task.getReturnCode() != 0:
            return

        out = self.__lspci_task.getOutput()
        # Parse the lspci output
        lspci_value = VariantRef()
        self.parse_lspci(out, lspci_value)

        self.__result = lspci_value.value()

parse_lspci_verbose = ParseLspciVerbose(run_lspci_verbose) \
    .title('''Parse verbose output from the lspci command''')
