# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.
import os
import re

# Avoid locale differences in numbers printed
import locale
locale.setlocale(locale.LC_ALL, "POSIX")

from nvsmhealth import modules, tasks, config

__all__ = [
    'run_mdadm_examine',
    'mdadm_parse_examine',
    'run_mdadm_detail',
    'mdadm_parse_detail',
    'check_mdadm_volumes',
    'check_mdadm_disks'
]

class RunMdadmExamine(tasks.ITask):
    def __init__(self):
        super().__init__()
        self.depends(modules.common.check_superuser_privileges)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        import glob
        volumes = glob.glob("/sys/block/md*")
        disks = sum([glob.glob(v+"/md/dev-*") for v in volumes], [])

        # capture disk info with mdadm examine command
        disk_info = {}
        for disk in disks:
            uevent_path = disk+'/block/uevent'
            if not os.path.exists(uevent_path):
                continue
            dev_name = ''
            with open(uevent_path) as f:
                lines = f.readlines()
                for line in lines:
                    pattern = re.compile(r'DEVNAME=(?P<dev_name>.*)')
                    match = pattern.match(line)
                    if match:
                        dev_name = match.group('dev_name')
                        break
            if dev_name == '':
                continue

            # capture mdadm info for each device
            args = "mdadm --examine /dev/"+dev_name
            cmd_task = tasks.RunCommand(args.split())
            cmd_task.depends(modules.common.check_superuser_privileges)
            cmd_task.run()
            if cmd_task.getReturnCode():
                # could not run the info command
                continue
            out = cmd_task.getOutput().strip()
            disk_info[dev_name] = out
        self.__result = disk_info

run_mdadm_examine = RunMdadmExamine() \
        .title("Capture mdadm examine output for all the disks")

class MdadmParseExamine(tasks.ITask):
    def __init__(self, run_mdadm_examine):
        super().__init__()
        self.__run_mdadm_examine_task = run_mdadm_examine
        self.depends(self.__run_mdadm_examine_task)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        disk_info_out = self.__run_mdadm_examine_task.getResult()
        if not disk_info_out:
            return

        # Parse disk information printed by mdadm
        disk_info = {}
        for dev_name, out in disk_info_out.items():
            field_pattern = re.compile(
                    r'^(?P<name>[^:]*):(?P<value>.*)$',
                    flags=re.MULTILINE)

            # Parse disk details into JSON fields
            disk_fields = {}
            for field in field_pattern.finditer(out):
                field_name = field.group('name')
                field_name = field_name.strip().lower().replace(' ', '_')
                field_value = field.group('value').strip()
                disk_fields[field_name] = field_value
            disk_info[dev_name] = disk_fields

        self.__result = disk_info

mdadm_parse_examine = MdadmParseExamine(run_mdadm_examine) \
        .title('Parse mdadm examine')

class RunMdadmDetail(tasks.ITask):
    def __init__(self):
        super().__init__()
        self.depends(modules.common.check_superuser_privileges)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        import glob
        volumes = glob.glob("/dev/md*")

        # Parse disk information printed by mdadm
        volume_info = {}
        for volume in volumes:
            # capture mdadm info for each device
            args = "mdadm --detail "+volume
            cmd_task = tasks.RunCommand(args.split())
            cmd_task.depends(modules.common.check_superuser_privileges)
            cmd_task.run()
            if cmd_task.getReturnCode():
                # could not run the info command
                continue
            out = cmd_task.getOutput().strip()
            volume_info[volume] = out

        self.__result = volume_info

run_mdadm_detail = RunMdadmDetail() \
        .title("Capture mdadm detail output for all the disks")

class MdadmParseDetail(tasks.ITask):
    def __init__(self, run_mdadm_detail):
        super().__init__()
        self.__run_mdadm_detail = run_mdadm_detail
        self.depends(self.__run_mdadm_detail)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        volume_detail_out = self.__run_mdadm_detail.getResult()
        if not volume_detail_out:
            return
        # Parse disk information printed by mdadm
        volume_info = {}
        for volume, out in volume_detail_out.items():
            volume_details_pattern = re.compile(
                    r'^(?P<volume_name>[^:]*):\n(?P<volume_body>(?:.*\n?)*)',
                    flags=re.MULTILINE)
            field_pattern = re.compile(
                    r'^(?P<name>[^:]*):(?P<value>.*)$',
                    flags=re.MULTILINE)

            # Use regex to separate mdadm output into volume name and body
            volume_match = volume_details_pattern.match(out)
            if not volume_match:
                continue  # Could not parse mdadm output for this volume
            volume_name = volume_match.group('volume_name').strip()
            volume_body = volume_match.group('volume_body')
            # Parse volume details into JSON fields
            volume_fields = {}
            for field in field_pattern.finditer(volume_body):
                field_name = field.group('name')
                field_name = field_name.strip().lower().replace(' ', '_')
                field_value = field.group('value').strip()
                volume_fields[field_name] = field_value
            volume_info[volume] = volume_fields

        self.__result = volume_info

mdadm_parse_detail = MdadmParseDetail(run_mdadm_detail) \
        .title('Parse mdadm detail')

class CheckMdadmVolumes(tasks.ICheck):
    def __init__(self, mdadm_parse_detail):
        super().__init__()
        self.__mdadm_parse_details = mdadm_parse_detail
        self.depends(self.__mdadm_parse_details)
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('mdadm_volume_status')

    def run(self):
        try:
            check = self.__parameters_task.getResult()['mdadm_volume_status']
            if not check:
                return
        except:
            return

        volume_info = self.__mdadm_parse_details.getResult()
        if not volume_info:
            self.addCheckMessage("No result from parse 'mdadm --detail' for software RAID volume")
            self.addUnknown()
            return

        self.addCheckMessage("Checking output of 'mdadm --detail' for each software RAID volume")

        good_volume_states = [ 'clean', 'active', 'write-pending', 'active-idle' ]

        healthy = True
        for name, volume in volume_info.items():
            # Check the volume state
            if 'state' not in volume:
                self.addCheckMessage('State not known for RAID volume "{0}"'.format(name))
                healthy = False
            elif 'recovering' in volume['state'].lower():
                self.addCheckMessage('It appears that the RAID volume "{0}" is currently' \
                      ' recovering. This is normal. However, volume performance' \
                      ' might be reduced while the volume is recovering. The' \
                      ' recovery process should complete soon, but if it does' \
                      ' not please contact NVIDIA support.'.format(name))
            elif 'resync' in volume['state'].lower():
                self.addCheckMessage('It appears that the RAID volume "{0}" is currently' \
                      ' resyncing. This is normal. However, volume performance' \
                      ' might be reduced while the volume is resyncing. The' \
                      ' resync process should complete soon, but if it does' \
                      ' not please contact NVIDIA support.'.format(name))
            elif volume['state'].lower() not in good_volume_states:
                self.addCheckMessage('Observed unhealthy state "{0}" for RAID volume "{1}"'.format(
                            volume['state'], name))
                healthy = False
            # Check for failed devices in the volume
            if 'failed_devices' not in volume:
                pass  # Hmm...
            elif int(volume['failed_devices']) > 0:
                self.addCheckMessage('Observed {0} failed device(s) in RAID volume "{1}"'.format(
                            volume['failed_devices'], name))
                healthy = False

        # Return healthy/unhealthy status
        if healthy:
            self.addHealthy()
        else:
            self.addUnHealthy()

check_mdadm_volumes = CheckMdadmVolumes(mdadm_parse_detail) \
        .title('Status of software RAID volumes')

class CheckMdadmDisks(tasks.ICheck):
    def __init__(self, mdadm_parse_examine):
        super().__init__()
        self.__mdadm_parse_examine = mdadm_parse_examine
        self.depends(self.__mdadm_parse_examine)
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('mdadm_disk_status')

    def run(self):
        try:
            check = self.__parameters_task.getResult()['mdadm_disk_status']
            if not check:
                return
        except:
            return

        disk_info = self.__mdadm_parse_examine.getResult()
        if not disk_info:
            self.addCheckMessage("No result from parse 'mdadm --examine' for software RAID superblock")
            self.addUnknown()
            return

        self.addCheckMessage("Checking output of 'mdadm --examine' for each software RAID superblock")

        # Check the checksum of each RAID disk superblock managed by mdadm
        healthy = True
        for name, disk in disk_info.items():
            if 'checksum' not in disk:
                self.addCheckMessage('Checksum not known for RAID disk "{0}"'.format(name))
                healthy = False
                continue
            checksum = disk['checksum']
            if 'correct' not in checksum:
                self.addCheckMessage('Observed failed checksum "{0}" on RAID disk "{1}"'.format(
                            checksum, name))
                healthy = False

         # Return healthy/unhealthy status
        if healthy:
            self.addHealthy()
        else:
            self.addUnHealthy()

check_mdadm_disks = CheckMdadmDisks(mdadm_parse_examine) \
        .title('Status of software RAID disk superblocks')
