# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.
import re
from nvsmhealth import tasks, modules

__all__ = ['collect_mem_info',
           'check_meminfo_mem_size']

class CollectMemInfo(tasks.ITask):
    def __init__(self):
        super().__init__()
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        import os
        meminfo_path = os.path.join(os.sep, 'proc', 'meminfo')

        with open(meminfo_path) as f:
            meminfo_data = f.read()

        pattern = re.compile(r'MemTotal:\s+(?P<MemTotal>\d+)')
        match = pattern.match(meminfo_data)
        if not match:
            return

        # convert and factor
        mem_total = match.group('MemTotal').strip()

        try:
            mem_total = float(mem_total) * 9.537e-7
            self.__result = mem_total
        except:
            return

collect_mem_info = CollectMemInfo() \
        .title('Verify memory information using the content of /proc/meminfo')

class CheckMemInfoMemSize(tasks.ICheck):
    def __init__(self):
        super().__init__()
        self.__collect_meminfo_task = collect_mem_info
        self.depends(self.__collect_meminfo_task)
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('meminfo_memory_size')

    def run(self):
        try:
            threshold = self.__parameters_task.getResult()['meminfo_memory_size']
        except:
            return

        threshold = threshold * 9.537e-7
        self.title(self.getTitle().format(threshold))

        actual_mem = self.__collect_meminfo_task.getResult()
        if not actual_mem:
            self.addUnknown()
            return

        # check for actual mem w.r.t threshold with tolerance of 1GB
        tolerance = 1
        if modules.common.almost_equal(actual_mem, threshold, tolerance):
            self.addHealthy()
        else:
            self.addCheckMessage('Amount of memory is {0:.2f} GB'.format(actual_mem))
            self.addUnHealthy()

check_meminfo_mem_size = CheckMemInfoMemSize() \
        .title('Installed memory capacity [{0:.2f}GB]')
