# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from nvsmhealth import modules, tasks, config
from nvsmhealth.lib.xmltodict import xmltodict

__all__ = [
    'run_mlxfwmanager',
    'check_mlx_fw_version'
]

run_mlxfwmanager = tasks.RunCommand(
        args=['mlxfwmanager', '--query-format', 'xml']) \
    .title('''Collect details of mellanox devices firmware version using Mellanox Firmware Manager''')

class CheckMlxFwVersion(tasks.ICheck):
    def __init__(self, run_task, parameters_task):
        super(CheckMlxFwVersion, self).__init__()
        self.__run_task = run_task
        self.__parameters_task = parameters_task
        self.depends(run_task)
        self.depends(parameters_task)

    def run(self):

        own_params = self.getParameters()
        params = self.__parameters_task.getResult()

        firmware_versions = 0
        for param, req in own_params.items():
            if param in params.keys():
                firmware_versions = params[param]

        if firmware_versions == 0:
            return

        # If mlxfwmanager run got a failure, skip parsing the output
        mlnx_res = self.__run_task.getOutput()
        if self.__run_task.getReturnCode() or not mlnx_res:
            self.addCheckMessage('No result from mellanox firmware manager')
            self.addUnknown()
            return

        self.addCheckMessage(config.mlx_fw_ver_cmd_str)
        inconsistent_devices = {}

        try:
            res_dict = xmltodict.parse(mlnx_res)
            for res in res_dict['Devices']['Device']:
                pci = res['@pciName']
                fw_ver = res['Versions']['FW']['@current']
                if fw_ver not in inconsistent_devices:
                   inconsistent_devices[fw_ver] = []
                inconsistent_devices[fw_ver].append(pci)

            # Unhealthy check for multiple versions
            if len(inconsistent_devices) > firmware_versions:
                res = ""
                for k, v in inconsistent_devices.items():
                    res += "PCI device: '{}' has firmware version '{}'\n".format(", ".join(v), k)
                self.addCheckMessage(f"Different firmware version found on Mellanox devices\n{res}")
                self.addUnHealthy()
            else:
                self.addHealthy()
        # Unknown check
        except:
            self.addCheckMessage('Error while checking Mellanox devices firmware version consistency')
            self.addUnknown()

check_mlx_fw_version = CheckMlxFwVersion(run_mlxfwmanager, modules.parameter.parameters) \
    .title('''Verify Mellanox devices firmware version consistency''') \
    .parameters('mlnx_firmware_check')

