# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from nvsmhealth import modules, tasks, config

__all__ = [
    'run_net_ifconfig',     # Runs ifconfig
    'parse_net_ifconfig',   # Parses ifconfig, return dictionary
    'check_net_link',       # Check interface link using ethtool
    'check_net_ping'        # Check IP reachability using ping
]

run_net_ifconfig = tasks.RunCommand(
        args=['ifconfig', '-a']) \
        .title('Run ifconfig command to show all network interfaces') \
        .describe('''See all network interfaces''')

class ParseNetIfconfig(tasks.ITask):
    def __init__(self, run_net_ifconfig_task):
        super().__init__()
        self.__run_net_ifconfig_task = run_net_ifconfig_task 
        self.__output = {}

    def getOutput(self):
        return self.__output

    def getInterfaceNames(self, output):
        from re import findall
        regex = r'([A-Za-z0-9_.-]*):+ flags'
        devices = findall(regex, output)
        self.__output['interface_names'] = devices

    def run(self):
        if(self.__run_net_ifconfig_task.getReturnCode()!=0):
            return
        output = self.__run_net_ifconfig_task.getOutput()
        self.getInterfaceNames(output)

        # TODO: Add Code to get IP Address

parse_net_ifconfig = ParseNetIfconfig(run_net_ifconfig) \
    .title('''Parse network interfaces''') \
    .depends(run_net_ifconfig)

class CheckNetLink(tasks.ICheck):
    def __init__(self, parse_net_ifconfig_task):
        super().__init__()
        self.__parse_net_ifconfig_task = parse_net_ifconfig_task
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('net_link')
        self.__output = {}

    def getOutput(self):
        return self.__output

    def run(self):
        try:
            net_link = self.__parameters_task.getResult()['net_link']
            if not net_link:
                return
        except:
            return

        from re import findall
        from nvsmhealth.lib import DictionarySuperset

        output = self.__parse_net_ifconfig_task.getOutput()['interface_names']
        if output == None:
            return

        #~ TODO: Not all interfaces are currently up
        #~ Using the ones passed by net_link
        interfaces = list(net_link.keys())
        health_list = [True for x in range(len(interfaces))]

        j = {}
        for i, device in enumerate(interfaces):
            j[device] = {}
            ethtool = tasks.RunCommand(args=['ethtool', '{}'.format(device)])
            ethtool.run()
            if(ethtool.getReturnCode()!=0):
                self.addCheckMessage("Check network interface setup: {}".format(device))
                health_list[i] = False
                continue
            regex = r'Link detected:+ ([A-Za-z0-9_.-]*)'
            x = findall(regex, ethtool.getOutput())
            try:
                if(x[0] == 'yes'):
                    j[device]['link'] = True
                else:
                    j[device]['link'] = False
            except:
                j[device]['link'] = False
        
        dictionary_superset = DictionarySuperset.DictionarySuperset(
                    missing_message="Could not detect {}")
        result = dictionary_superset.compare(j, net_link)

        if (all(health_list) and result == None):
            self.addHealthy()
        else: 
            self.addCheckMessage(result)
            self.addUnHealthy()

check_net_link = CheckNetLink(parse_net_ifconfig) \
    .title('''Verify Network Interfaces Link''') \
    .depends(parse_net_ifconfig)


class CheckNetPing(tasks.ICheck):
    def __init__(self, ):
        super().__init__()
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('net_ping')
        self.__output = {}

    def getOutput(self):
        return self.__output

    def run(self):
        from re import findall
        try:
            net_ping = self.__parameters_task.getResult()['net_ping']
            if not net_ping:
                return
        except:
            return
        
        health_list = [True for x in range(len(net_ping))]

        for i, (interface, ip) in enumerate(net_ping.items()):
            run_ping_task = tasks.RunCommand(
                    args=['ping', '-c', str(config.check_net_ping_count), '-W', str(config.check_net_ping_timeout), ip]) \
                    .title('Run ping') \
                    .describe('''Check IP Reachability via Ping''')
            try:
                run_ping_task.run()
                if(run_ping_task.getReturnCode()!=0):
                    raise Exception("Unable to ping {} at {}".format(ip, interface))

                output = run_ping_task.getOutput()
            except:
                self.addCheckMessage("Unable to ping {} at {}".format(ip, interface))
                health_list[i] = False
                continue

            regex = r'[0-9 ]*[a-z].+?(?=,).+?(?=,).+(.+?(?=%))'
            packet_loss = findall(regex, output)
            
            self.addCheckMessage("Checking Packet Loss on {} at {}: {}%".format(interface, ip, packet_loss[0]))

            if(packet_loss[0]!='0'):
                health_list[i] = False
            elif(packet_loss[0]=='0'):
                health_list[i] = True

        if all(health_list):
            self.addHealthy()
        else: 
            self.addUnHealthy()

            

check_net_ping = CheckNetPing() \
    .title('''Verify Network IP Reachability''')
