# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from nvsmhealth import modules, tasks, config
from nvsmhealth.tasks.stream import HealthTaskStream
from nvsmhealth.lib import DictionarySuperset
from nvsmhealth.lib.xmltodict import xmltodict
import logging
import json
import re
import sys


__all__ = [
    'run_nvidia_smi_topology',
    'parse_nvidia_smi_topology',
    'gpu_mig_status',
    'check_gpu_direct_topology',
    'run_nvidia_smi_p2p_topology',
    'parse_nvidia_smi_p2p_topology',
    'check_gpu_p2p_topology',
    'run_nvidia_smi_gpu_bus_id',
    'parse_nvidia_smi_gpu_bus_id',
    'check_nvidia_smi_gpu_bus_id',
    'parse_nvidia_smi',
    'check_gpu_vbios_version_consistency',
    'check_gpus',
    'nvidia_driver_version',
    'check_nvidia_smi_nvlink_status',
    'nvidia_vbios_version',
    'gpu_bdf_pstate'
]

run_nvidia_smi_topology = tasks.RunCommand(
    args=['nvidia-smi', 'topo', '--matrix']) \
    .title('''Collect GPUDirect topology using the nvidia-smi tool''')


def genericParseNvidiaSmi(matrix_string, is_p2p=False):
    # Parse the result as a JSON object from parent task
    if is_p2p:
        matrix_header_pattern = re.compile(
            r'^\s*(?:\033\[4m)?(?P<column_header>(?:\S+\s+)+)\s*(?:\033\[0m)?')
    else:
        matrix_header_pattern = re.compile(
            r'^\s*(?:\033\[4m)?(?P<column_header>(?:\S+\s+)+)CPU Affinity(?:\033\[0m)?')
    matrix = {}
    matrix_string = matrix_string.strip()
    matrix_lines = matrix_string.splitlines()
    matrix_header = matrix_lines[0]

    # Collect header from the first line
    m = matrix_header_pattern.match(matrix_header)
    if not m:
        raise RuntimeError("Could not match topology matrix header")
    matrix_columns = m.group('column_header').split()

    # Iterate over and collect each row
    for line in matrix_lines[1:]:
        line = line.strip()
        if not line:
            # Empty line signifies the end of the matrix
            break
        row_cells = line.split()
        row_header = row_cells[0]
        if row_header not in matrix:
            matrix[row_header] = {}

        # Iterate over the cells in this row
        column_index = 0
        for cell in row_cells[1:]:
            if column_index >= len(matrix_columns):
                # Ignore the CPU Affinity cell
                break
            column = matrix_columns[column_index]
            matrix[row_header][column] = cell
            column_index += 1

    return matrix


class ParseNvidiaSmiTopology(tasks.ITask):
    def __init__(self, nvidia_smi_task):
        super(ParseNvidiaSmiTopology, self).__init__()

        # Declare dependency on nvidia_smi using the ITask interface
        self.__nvidia_smi_task = nvidia_smi_task
        self.depends(nvidia_smi_task)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        # If nvidia-smi run got a failure, skip parsing the output
        if self.__nvidia_smi_task.getReturnCode() != 0:
            return

        matrix_string = self.__nvidia_smi_task.getOutput()
        matrix = genericParseNvidiaSmi(matrix_string)
        # Add the parsed result as JSON object
        self.__result = json.dumps(matrix, ensure_ascii=True)


parse_nvidia_smi_topology = ParseNvidiaSmiTopology(run_nvidia_smi_topology) \
    .title('''Parse GPUDirect topology information gathered from nvidia-smi tool''') \



def genericGpuTopologyCheck(topology, expected_topology):
    healthy = True
    message = ''

    topology_gpus = topology.keys()
    expected_topology_gpus = expected_topology.keys()
    # Check missing GPUs
    count = 0
    for gpu in expected_topology_gpus:
        if gpu not in topology_gpus:
            count += 1

    if count:
        message += '{} of {} GPUs missing'.format(
            count, len(expected_topology_gpus))
        healthy = False

    # Check gpu link status mismatch:
    for gpu, gpus in expected_topology.items():
        # For missing GPU, mark link to all GPUs as missing
        if gpu not in topology_gpus:
            for g in expected_topology_gpus:
                message += '\n"{} -> {}" is missing'.format(gpu, g)

        else:
            for k, v in gpus.items():
                if k in topology[gpu].keys() and v != topology[gpu][k]:
                    message += '\n"{} -> {}" has value "{}" when "{}" was expected'.format(
                        gpu, k, topology[gpu][k], v)
                    healthy = False

    return healthy, message


class ParseNvidiaSmi(tasks.ITask):
    def __init__(self):
        super(ParseNvidiaSmi, self).__init__()
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        # If nvidia-smi run got a failure, skip parsing the output
        nvidia_smi = tasks.RunCommand(args=['nvidia-smi', '--list-gpus'])
        nvidia_smi.run()
        if nvidia_smi.getReturnCode() != 0:
            return
        nvidia_smi_output = nvidia_smi.getOutput().strip()
        gpus = nvidia_smi_output.split('\n')
        gpus_count = len(gpus)
        gpus_out = {}
        for gpu in range(gpus_count):
            nvidia_smi_query = tasks.RunCommand(
                args=['nvidia-smi', '-i', '{}'.format(gpu), '--query', '--xml-format'])
            nvidia_smi_query.run()
            if nvidia_smi_query.getReturnCode() == 0:
                nvidia_smi_query_out = nvidia_smi_query.getOutput()
                gpus_out[gpu] = nvidia_smi_query_out

        self.__result = gpus_out


parse_nvidia_smi = ParseNvidiaSmi() \
    .title('''Parse nvidia-smi tool output for GPUs consistency checks''')


class GpuMIGStatus(tasks.ITask):
    def __init__(self, parse_task, parameters_task):
        super(GpuMIGStatus, self).__init__()
        self.depends(parse_task)
        self.depends(parameters_task)
        self.__parse_task = parse_task
        self.__parameters_task = parameters_task
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        # If nvidia-smi run got a failure, skip parsing the output
        nvidia_smi_res = self.__parse_task.getResult()
        if not nvidia_smi_res:
            logging.debug("No result from nvidia-smi tool")
            return

        gpus_mig_enabled = {}
        try:
            for gpu, info in nvidia_smi_res.items():
                gpu_dict = xmltodict.parse(info)
                mig_mode = gpu_dict['nvidia_smi_log']['gpu']['mig_mode']['current_mig']
                gpus_mig_enabled[gpu] = 1 if mig_mode == 'Enabled' else 0
        except:
            return

        self.__result = gpus_mig_enabled


gpu_mig_status = GpuMIGStatus(parse_nvidia_smi, modules.parameter.parameters) \
    .title('''GPUs MIG Status Task''') \
    .parameters('enable_gpu_mig') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_gpu_mig'))

class CheckGpuDirectTopology(tasks.ICheck):
    def __init__(self, nvidia_smi_task, gpu_mig_status, parameters_task):
        super(CheckGpuDirectTopology, self).__init__()

        # Declare dependency on parse nvidia-smi task using the ITask interface
        self.__parse_task = nvidia_smi_task
        self.__parameters_task = parameters_task
        self.__gpu_mig_state = gpu_mig_status
        self.depends(nvidia_smi_task)
        self.depends(parameters_task)
        self.depends(gpu_mig_status)
        self.depends(modules.kvm.kvm_mode_on)

    def run(self):
        # Get parameters from own task
        own_params = self.getParameters()

        # Get parameters from parameter task
        params = self.__parameters_task.getResult()

        # Get GPU MIG State
        gpu_mig_state = self.__gpu_mig_state.getResult()

        # Check if kvm mode is on then skip
        if modules.kvm.kvm_mode_on.getResult() == True:
            self.addCheckMessage('KVM mode is on, skipping check.')
            self.addInformational()
            return

        # If GPU State Enabled
        if gpu_mig_state != None and any(gpu_mig_state.values()):
            logging.info("MIG State Detected: Modifying to MIG Topology")
            _gpu_direct_topology = params['gpu_direct_topology']
            num_gpus = len(gpu_mig_state.values())
            # gpu_mig_state is a dict
            gpus = range(num_gpus)
            mig_enabled = [x for x, y in gpu_mig_state.items() if y == 1]
            mig_enabled_gpus = ['GPU{}'.format(x) for x in mig_enabled]
            pxb_enabled_gpus = ['GPU{}'.format(
                y) for y in [x-1 if x % 2 != 0 else x+1 for x in mig_enabled]]
            non_mig_gpus = ['GPU{}'.format(x)
                            for x in gpus if x not in mig_enabled]

            # Go over all the MIG enabled GPUs first
            for index, mig_gpu in enumerate(mig_enabled_gpus):
                for k, v in _gpu_direct_topology[mig_gpu].items():
                    if k == mig_gpu:
                        continue  # Already marked as X
                    elif k == pxb_enabled_gpus[index]:
                        _gpu_direct_topology[mig_gpu][k] = 'PXB'
                    else:
                        _gpu_direct_topology[mig_gpu][k] = 'SYS'

            #Go over non-mig gpus next
            for index, gpu in enumerate(non_mig_gpus):
                for k, v in _gpu_direct_topology[gpu].items():
                    if k == gpu:
                        continue
                    elif k in mig_enabled_gpus and gpu in pxb_enabled_gpus:
                        if mig_enabled_gpus.index(k) == pxb_enabled_gpus.index(gpu):
                            _gpu_direct_topology[gpu][k] = 'PXB'
                        else:
                            _gpu_direct_topology[gpu][k] = 'SYS'
                    elif k in mig_enabled_gpus:
                        _gpu_direct_topology[gpu][k] = 'SYS'

            params['gpu_direct_topology'] = _gpu_direct_topology

        expected_topology = None
        for param, req in own_params.items():
            if param in params.keys():
               expected_topology = params[param]
        # If parameter is not found in platform parameters - Do Nothing
        if expected_topology == None:
            return

        # Print/Stream Task info and command messages
        self.addCheckMessage(config.gpu_direct_topology_command_str)

        # Unknown check for no result from nvidia-smi topology parse task
        if not self.__parse_task.getResult():
            self.addCheckMessage(
                'No result for GPUDirect topology information gathered from nvidia-smi tool')
            self.addUnknown()
            return

        try:
            # Compare task output with expected config
            topology = json.loads(self.__parse_task.getResult())
            healthy, message = genericGpuTopologyCheck(
                topology, expected_topology)
            if not healthy:
                self.addCheckMessage(message)
                self.addUnHealthy()
            else:
                self.addHealthy()
        except:
            self.addCheckMessage('Error while checking gpu direct topology')
            self.addUnknown()  # Unknown check


check_gpu_direct_topology = CheckGpuDirectTopology(parse_nvidia_smi_topology, gpu_mig_status, modules.parameter.parameters) \
    .title('''Check GPUDirect Topology information for consistency''') \
    .parameters('gpu_direct_topology')


run_nvidia_smi_p2p_topology = tasks.RunCommand(
    args=['nvidia-smi', 'topo', '-p2p', 'rw']) \
    .title('''Collect GPUs p2p topology using the nvidia-smi tool''')


class ParseNvidiaSmiP2PTopology(tasks.ITask):
    def __init__(self, nvidia_smi_task):
        super(ParseNvidiaSmiP2PTopology, self).__init__()

        # Declare dependency on nvidia_smi using the ITask interface
        self.__nvidia_smi_task = nvidia_smi_task
        self.depends(nvidia_smi_task)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        # If nvidia-smi run got a failure, skip parsing the output
        if self.__nvidia_smi_task.getReturnCode() != 0:
            return

        matrix_string = self.__nvidia_smi_task.getOutput()
        matrix = genericParseNvidiaSmi(matrix_string, True)
        # Add the parsed result as JSON object
        self.__result = json.dumps(matrix, ensure_ascii=True)


parse_nvidia_smi_p2p_topology = ParseNvidiaSmiP2PTopology(run_nvidia_smi_p2p_topology) \
    .title('''Parse GPUs p2p topology information gathered from nvidia-smi tool''')


class CheckGpuP2PTopology(tasks.ICheck):
    def __init__(self, nvidia_smi_task, parameters_task):
        super(CheckGpuP2PTopology, self).__init__()

        # Declare dependency on parse nvidia-smi task using the ITask interface
        self.__parse_task = nvidia_smi_task
        self.__parameters_task = parameters_task
        self.depends(nvidia_smi_task)
        self.depends(parameters_task)
        self.depends(modules.kvm.kvm_mode_on)

    def run(self):
        # Get parameters from own task
        own_params = self.getParameters()

        # Get parameters from parameter task
        params = self.__parameters_task.getResult()

        expected_topology = None
        for param, req in own_params.items():
            if param in params.keys():
               expected_topology = params[param]
        # If parameter is not found in platform parameters - Do Nothing
        if expected_topology == None:
            return
        if modules.kvm.kvm_mode_on.getResult() == True:
            self.addCheckMessage('KVM mode is on, skipping check.')
            self.addInformational()
            return
        # Print/Stream Task info and command messages
        self.addCheckMessage(config.gpu_p2p_topology_command_str)

        # Unknown check for no result from nvidia-smi topology parse task
        if not self.__parse_task.getResult():
            self.addCheckMessage(
                'No result for GPUs p2p topology information gathered from nvidia-smi tool')
            self.addUnknown()
            return

        try:
            # Compare task output with expected config
            topology = json.loads(self.__parse_task.getResult())
            healthy, message = genericGpuTopologyCheck(
                topology, expected_topology)
            if not healthy:
                self.addCheckMessage(message)
                self.addUnHealthy()
            else:
                self.addHealthy()
        except:
            self.addCheckMessage('Error while checking gpu direct topology')
            self.addUnknown()  # Unknown check


check_gpu_p2p_topology = CheckGpuP2PTopology(parse_nvidia_smi_p2p_topology, modules.parameter.parameters) \
    .title('''Check GPUs p2p Topology information for consistency''') \
    .parameters('gpu_p2p_topology')


run_nvidia_smi_gpu_bus_id = tasks.RunCommand(
    args=['nvidia-smi', '--query-gpu=gpu_bus_id', '--format=csv,noheader']) \
    .title('''Collect GPU's identified with the NVIDIA System Management Interface (nvidia-smi) tool''')


class ParseNvidiaSmiGpuBusId(tasks.ITask):
    def __init__(self, nvidia_smi_task):
        super(ParseNvidiaSmiGpuBusId, self).__init__()

        # Declare dependency on nvidia_smi using the ITask interface
        self.__nvidia_smi_task = nvidia_smi_task
        self.depends(nvidia_smi_task)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):

        # If nvidia-smi run got a failure, skip parsing the output
        if self.__nvidia_smi_task.getReturnCode() or not self.__nvidia_smi_task.getOutput():
            return

        res = self.__nvidia_smi_task.getOutput()

        try:
            res_bdfs = res.strip().split('\n')
            bdfs = []
            pattern = re.compile(
                "^(?P<domain>.*):(?P<bus>.*):(?P<device>.*)\.(?P<function>.*)")
            for bdf in res_bdfs:
                match = pattern.search(bdf)
                if match:
                    match_dict = match.groupdict()
                    bdfs.append('{}:{}.{}'.format(match_dict['bus'].lower(
                    ), match_dict['device'], match_dict['function']))

            self.__result = bdfs

        except:
            pass


parse_nvidia_smi_gpu_bus_id = ParseNvidiaSmiGpuBusId(run_nvidia_smi_gpu_bus_id) \
    .title('''Parse GPU's identified with the NVIDIA System Management Interface (nvidia-smi) tool''')


class CheckNvidiaSmiGpuBusId(tasks.ICheck):
    def __init__(self, nvidia_smi_task, parameters_task):
        super(CheckNvidiaSmiGpuBusId, self).__init__()

        # Declare dependency on parse nvidia-smi task using the ITask interface
        self.__parse_task = nvidia_smi_task
        self.__parameters_task = parameters_task
        self.depends(nvidia_smi_task)
        self.depends(parameters_task)

    def run(self):
        # Get parameters from own task
        own_params = self.getParameters()

        # Get parameters from parameter task
        params = self.__parameters_task.getResult()

        expected_bdfs = None
        for param, req in own_params.items():
            if param in params.keys():
               expected_bdfs = params[param]
        # If parameter is not found in platform parameters - Do Nothing
        if expected_bdfs == None:
            return
        if modules.kvm.kvm_mode_on.getResult() == True:
            self.addCheckMessage('KVM mode is on, skipping check.')
            self.addInformational()
            return
        # Unknown check for no result from nvidia-smi gpu_bus_id parse task
        if not self.__parse_task.getResult():
            self.addCheckMessage(
                'No result for gpu_bus_id information gathered from nvidia-smi tool')
            self.addUnknown()
            return

        # Print/Stream Task info and command messages
        self.addCheckMessage(
            'Checking output of "nvidia-smi --query-gpu=gpu_bus_id --format=csv,noheader" for expected GPUs')

        message = ''
        healthy = True

        try:
            bdfs = self.__parse_task.getResult()
            for gpu in expected_bdfs.keys():
                if gpu not in bdfs:
                    message += '\nGPU not identified at PCI address "{}"'.format(
                        gpu)
                    healthy = False

            if not healthy:
                self.addCheckMessage(message)
                self.addUnHealthy()
            else:
                self.addHealthy()

        except:
            self.addCheckMessage('Error while identifying GPUs bus_id')
            self.addUnknown()


check_nvidia_smi_gpu_bus_id = CheckNvidiaSmiGpuBusId(parse_nvidia_smi_gpu_bus_id, modules.parameter.parameters) \
    .title('''Verify GPU's identified using nvidia-smi''') \
    .parameters('gpu_bom')


class CheckGpuVbiosVersionConsistency(tasks.ICheck):
    def __init__(self, parse_task):
        super(CheckGpuVbiosVersionConsistency, self).__init__()
        self.depends(parse_task)
        self.depends(modules.kvm.kvm_mode_on)
        self.__parse_task = parse_task

    def run(self):
        if modules.kvm.kvm_mode_on.getResult() == True:
            self.addCheckMessage('KVM mode is on, skipping check.')
            self.addInformational()
            return
        # If nvidia-smi run got a failure, skip parsing the output

        nvidia_smi_res = self.__parse_task.getResult()
        if not nvidia_smi_res:
            self.addCheckMessage('No result from nvidia-smi tool')
            self.addUnknown()
            return

        inconsistent_gpus = {}
        self.addCheckMessage(config.vbios_command_str)
        '''
        product name: vbios_version: gpu_name
        '''
        try:
            for gpu, info in nvidia_smi_res.items():
                gpu_dict = xmltodict.parse(info)
                vbios_version = gpu_dict['nvidia_smi_log']['gpu']['vbios_version']
                product_name = gpu_dict['nvidia_smi_log']['gpu']['pci']['pci_device_id']
                if product_name not in inconsistent_gpus:
                    inconsistent_gpus[product_name] = {}
                if vbios_version not in inconsistent_gpus[product_name]:
                    inconsistent_gpus[product_name][vbios_version] = []
                inconsistent_gpus[product_name][vbios_version].append('GPU{}'.format(gpu))
            # Unhealthy check for multiple versions
            res = ""
            for product_name, vbios_version in inconsistent_gpus.items():
                if len(vbios_version) > 1:
                    for k, v in vbios_version.items():
                        res += "GPUs: {} has VBIOS version '{}'\n".format(
                            ", ".join(v), k)
            if res != "":
                self.addCheckMessage(
                    f"Different VBIOS version found on GPUs\n{res}")
                self.addUnHealthy()
            else:
                self.addHealthy()
        # Unknown check
        except:
            self.addCheckMessage(
                'Error while checking GPUs VBIOS version consistency')
            self.addUnknown()


check_gpu_vbios_version_consistency = CheckGpuVbiosVersionConsistency(parse_nvidia_smi) \
    .title('''Verify GPUs VBIOS version consistency''') \
    .parameters('enable_vbios_version_check') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_vbios_version_check'))


class GpuCheckRetiredPagesCount(tasks.ICheck):
    def __init__(self, gpu, gpu_dict, threshold):
        super(GpuCheckRetiredPagesCount, self).__init__()
        self.__gpu = gpu
        self.__gpu_dict = gpu_dict
        self.__threshold = threshold
        super().name('GpuCheckRetiredPagesCount')

    def run(self):
        # Attempt to sum the retired page counts
        pages = 'Unknown'
        try:
            retired_pages = self.__gpu_dict['nvidia_smi_log']['gpu']['retired_pages']
            dbe_retired_count = int(
                retired_pages['double_bit_retirement']['retired_count'])
            sbe_retired_count = int(
                retired_pages['multiple_single_bit_retirement']['retired_count'])
            ret_pages = dbe_retired_count + sbe_retired_count
            pages = ret_pages
            if (ret_pages >= self.__threshold):
                self.addCheckMessage((config.gpu_total_retired_pages_count_error).
                                     format(ret_pages, sbe_retired_count, dbe_retired_count, gpu, self.__threshold))
                self.addUnHealthy()
            else:
                self.addHealthy()
        except ValueError:
            # ECC probably not supported
            retired_pages = self.__gpu_dict['nvidia_smi_log']['gpu']['retired_pages']
            dbe_retired_pages = (
                retired_pages['double_bit_retirement']['retired_count']).lower().strip()
            sbe_retired_pages = (
                retired_pages['multiple_single_bit_retirement']['retired_count']).lower().strip()
            if sbe_retired_pages == 'n/a' and dbe_retired_pages == 'n/a':
                # Skip the health check
                return
            else:
                self.addCheckMessage('Unknown value for retired pages ({} SBE; {} DBE) on GPU {}'.format(
                    sbe_ret_pages, dbe_retired_pages, self.__gpu))
                self.addUnknown()
        except:
            self.addCheckMessage('Error while getting retired pages count')
            self.addUnknown()

        self.title(config.gpu_total_retired_pages_count_str.format(
            gpu_index=self.__gpu, pages=pages))


class GpuCheckRetiredPagesPending(tasks.ICheck):
    def __init__(self, gpu, gpu_dict):
        super(GpuCheckRetiredPagesPending, self).__init__()
        self.__gpu = gpu
        self.__gpu_dict = gpu_dict
        super().name('GpuCheckRetiredPagesPending')

    def run(self):
        pages = 'No'
        try:
            # Bug-2587712: From NVIDIA Driver version 418.40.04 the "pending_retirement" key
            # has been renamed to "pending_blacklist" in nvidia-smi xml output
            try:
                ret = self.__gpu_dict['nvidia_smi_log']['gpu']['retired_pages']['pending_retirement']
            except KeyError:
                ret = self.__gpu_dict['nvidia_smi_log']['gpu']['retired_pages']['pending_blacklist']
            match_pattern = re.compile("^[nN][oO]$")
            m = match_pattern.match(ret)
            if m:
                self.addHealthy()
            else:
                if ret.lower() == "n/a":
                    # Skip the health check
                    return
                else:
                    self.addCheckMessage(
                        config.gpu_total_retired_pages_pending_error)
                    self.addUnHealthy()
                    pages = ret

        except:
            self.addCheckMessage('Error while getting pending retired pages')
            self.addUnknown()

        self.title(config.gpu_total_retired_pages_pending_str.format(
            gpu_index=self.__gpu, pages=pages))


class GpuCheckVbiosVersion(tasks.ITask):
    def __init__(self):
        super(GpuCheckVbiosVersion, self).__init__()
        self.name('GpuCheckVbiosVersion')
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        pass


class GpuCheckInforomStorageVersion(tasks.ITask):
    def __init__(self):
        super(GpuCheckInforomStorageVersion, self).__init__()
        self.name('GpuCheckInforomStorageVersion')
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        pass


class CheckGpus(tasks.ICheck):
    def __init__(self, parse_nvidia_smi_task, parameters_task):
        super(CheckGpus, self).__init__()
        self.depends(parse_nvidia_smi_task)
        self.depends(parameters_task)
        self.__parse_nvidia_smi_task = parse_nvidia_smi_task
        self.__parameters_task = parameters_task
        self.__gpu_result = None

    def getGPUResult(self):
        return self.__gpu_result

    def run(self):
        own_params = self.getParameters()
        params = self.__parameters_task.getResult()

        gpu_total_retired_page_count = None
        for param, req in own_params.items():
            if param in params.keys():
                gpu_total_retired_page_count = params[param]

        # If nvidia-smi run got a failure, skip parsing the output
        nvidia_smi_res = self.__parse_nvidia_smi_task.getResult()
        if nvidia_smi_res == None:
            return

        gpus_info = {}

        for gpu, info in nvidia_smi_res.items():
            gpus_info[gpu] = {}
            try:
                gpu_dict = xmltodict.parse(info)
                check_cls = GpuCheckVbiosVersion()
                vbios_version = gpu_dict['nvidia_smi_log']['gpu']['vbios_version']
                gpus_info[gpu]['vbios_version'] = vbios_version
                product_name = gpu_dict['nvidia_smi_log']['gpu']['product_name']
                gpus_info[gpu]['product_name'] = product_name
                bdf = gpu_dict['nvidia_smi_log']['gpu']['pci']['pci_bus_id']
                msg_str = config.gpu_vbios_version_str.format(
                    gpu_index=gpu, bdf=bdf)
                check_cls.title(msg_str)
                check_cls.setCallback(self.getCallback())
                check_cls.send(msg=vbios_version)
                super().addMessages(check_cls.getMessages())

                check_cls = GpuCheckInforomStorageVersion()
                inforom_version = gpu_dict['nvidia_smi_log']['gpu']['inforom_version']['img_version']
                msg_str = config.gpu_inforom_version_str.format(
                    gpu_index=gpu, bdf=bdf)
                check_cls.title(msg_str)
                check_cls.setCallback(self.getCallback())
                check_cls.send(msg=inforom_version)
                super().addMessages(check_cls.getMessages())

                # NVBug-2691112: Disable GPU total retired page count and retired pages pending health checks
                # These checks are already covered in blacklist_recommendations health check
                '''if gpu_total_retired_page_count != None:
                    check_cls = GpuCheckRetiredPagesCount(gpu, gpu_dict, gpu_total_retired_page_count)
                    check_cls.setCallback(self.getCallback())
                    check_cls.run()
                    self.addHealthy(count=check_cls.getResult()['healthy'])
                    self.addUnHealthy(count=check_cls.getResult()['unhealthy'])
                    self.addUnknown(count=check_cls.getResult()['unknown'])
                    check_cls.sendComplete()
                    super().addMessages(check_cls.getMessages())

                    check_cls = GpuCheckRetiredPagesPending(gpu, gpu_dict)
                    check_cls.setCallback(self.getCallback())
                    check_cls.run()
                    self.addHealthy(count=check_cls.getResult()['healthy'])
                    self.addUnHealthy(count=check_cls.getResult()['unhealthy'])
                    self.addUnknown(count=check_cls.getResult()['unknown'])
                    check_cls.sendComplete()
                    super().addMessages(check_cls.getMessages())'''
            except:
                logging.error(
                    "ERROR: Failed to perform GPU {} health check".format(gpu))
                pass
        self.title('')

        self.__gpu_result = gpus_info


check_gpus = CheckGpus(parse_nvidia_smi, modules.parameter.parameters) \
    .title('''Check GPU health retired page count, retired pages pending, inforom storage version and vbios version''') \
    .parameters('gpu_total_retired_page_count')


class NvidiaDriverVersion(tasks.ITask):
    def __init__(self):
        super(NvidiaDriverVersion, self).__init__()
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        # If nvidia-smi run got a failure, skip parsing the output
        nvidia_smi = tasks.RunCommand(
            args=['nvidia-smi', '--query-gpu=driver_version', '--format=csv,noheader'])
        nvidia_smi.run()
        if nvidia_smi.getReturnCode() != 0:
            return
        nvidia_smi_output = nvidia_smi.getOutput()
        nvidia_driver_version = nvidia_smi_output.split('\n')[0]
        self.title(config.nvidia_driver_version_info_str)
        self.send(msg=nvidia_driver_version)
        self.__result = nvidia_driver_version


nvidia_driver_version = NvidiaDriverVersion() \
    .title('''Checking nvidia driver version''') \
    .depends(modules.parameter.parameters)


class NvlinkSpeed(tasks.ICheck):
    def __init__(self, req_speed, active_count, nvlink):
        super(NvlinkSpeed, self).__init__()
        self.__req_speed = req_speed
        self.__active_count = active_count
        self.__nvlink = nvlink
        super().name("NvlinkSpeed")

    def get_activeCount(self):
        return self.__active_count

    def run(self):
        if self.__nvlink['status'] == self.__req_speed:
            self.addHealthy()
        else:
            if self.__active_count == None or self.__active_count <= 0:
                self.addUnHealthy()
            else:
                self.__active_count -= 1
                self.addHealthy()


class CheckNvidiaSmiNvlinkStatus(tasks.ICheck):
    def __init__(self, driver_version_task, parse_nvidia_smi, parameters_task):
        super(CheckNvidiaSmiNvlinkStatus, self).__init__()
        self.depends(driver_version_task)
        self.depends(parse_nvidia_smi)
        self.depends(parameters_task)
        self.__parameters_task = parameters_task
        self.__driver_version = driver_version_task
        self.__parse_nvidia_smi_task = parse_nvidia_smi

    def is_valid_version(self, ver):
        req_ver = config.nvidia_driver_version
        ver_tuple = tuple(map(int, (ver.split('.'))))
        req_ver_tuple = tuple(map(int, (req_ver.split('.'))))
        return (ver_tuple >= req_ver_tuple)

    def run(self):

        own_params = self.getParameters()
        params = self.__parameters_task.getResult()

        nvlink_speed = None
        nvlink_active_count = None
        for param, req in own_params.items():
            if param in params.keys():
                if 'nvlink_speed' in param:
                    nvlink_speed = params[param]
                else:
                    nvlink_active_count = params[param]

        if nvlink_speed == None:
            return

        # Test for older versions of nvidia-smi (via NVIDIA driver version)
        # NVLink speed query requires NVIDIA driver version 384.98 or later
        # Log error and return for version lesser than 384.98
        nvidia_driver_version = self.__driver_version.getResult()
        if nvidia_driver_version == None:
            return

        if not self.is_valid_version(nvidia_driver_version):
            logging.error(
                "ERROR: This version {} of nvidia-smi does not query NVLink speed".format(nvidia_driver_version))
            logging.error(
                "ERROR: NVLink speed query requires NVIDIA driver version 384.98 or later")
            return

        # Get GPU BDF information
        nvidia_smi_res = self.__parse_nvidia_smi_task.getResult()
        gpus_info = {}
        try:
            for gpu, info in nvidia_smi_res.items():
                gpus_info[gpu] = xmltodict.parse(info)
        except:
            pass

        # Get the nvlink status
        nvlink_nvidia_smi = tasks.RunCommand(
            args=['nvidia-smi', 'nvlink', '-s'])
        nvlink_nvidia_smi.run()
        if nvlink_nvidia_smi.getReturnCode() != 0:
            return

        nvlink_nvidia_smi_output = nvlink_nvidia_smi.getOutput()

        if len(nvlink_nvidia_smi_output) == 0:
            logging.debug(
                "nvidia-smi nvlink -s return code 0 but returned empty. Check if MIG enabled.")
            return

        link_regex = re.compile(
            "\\s*Link (?P<link_index>\\d+):\\s*(?P<link_status>.*)$")
        gpu_regex = re.compile("GPU (?P<gpu_index>\\d+)")

        nvlink_lines = nvlink_nvidia_smi_output.split('\n')
        nvlinks = {}
        for line in nvlink_lines:
            gpu_match = gpu_regex.match(line)
            if gpu_match:
                gpu = gpu_match.group('gpu_index').strip()
                nvlinks[gpu] = []
            link_match = link_regex.match(line)
            if link_match:
                index = link_match.group('link_index').strip()
                status = link_match.group('link_status').strip()
                nvlinks[gpu].append({'link': index, 'status': status})

        for gpu, nvlink in nvlinks.items():
            try:
                gpu_dict = gpus_info[int(gpu)]
                bdf = gpu_dict['nvidia_smi_log']['gpu']['pci']['pci_bus_id']
                active_count = nvlink_active_count
            except:
                bdf = ""
                pass
            for link in nvlink:
                check_cls = NvlinkSpeed(nvlink_speed, active_count, link)
                check_cls.title(config.nvlink_speed_info_str.format(
                    gpu, bdf, link['link'], link['status']))
                check_cls.setCallback(self.getCallback())
                check_cls.run()
                if nvlink_active_count != None:
                    active_count = check_cls.get_activeCount()
                self.addHealthy(count=check_cls.getResult()['healthy'])
                self.addUnHealthy(count=check_cls.getResult()['unhealthy'])
                self.addUnknown(count=check_cls.getResult()['unknown'])
                check_cls.sendComplete()
                super().addMessages(check_cls.getMessages())

        self.title('')


check_nvidia_smi_nvlink_status = CheckNvidiaSmiNvlinkStatus(nvidia_driver_version,
                                                            parse_nvidia_smi, modules.parameter.parameters) \
    .title('''Parse nvlink status with NVIDIA System Management Interface (nvidia-smi)''') \
    .parameters('gpu_nvlink_speed') \
    .parameters('gpu_nvlink_active_count')


class NvidiaVbiosVersion(tasks.ITask):
    def __init__(self):
        super(NvidiaVbiosVersion, self).__init__()
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        # If nvidia-smi run got a failure, skip parsing the output
        nvidia_smi = tasks.RunCommand(
            args=['nvidia-smi', '-i', '0', '--query-gpu=vbios_version', '--format=csv,noheader'])
        nvidia_smi.run()
        if nvidia_smi.getReturnCode() != 0:
            return
        nvidia_vbios_version = nvidia_smi.getOutput()
        self.__result = nvidia_vbios_version


nvidia_vbios_version = NvidiaVbiosVersion() \
    .title('''Get the nvidia vbios version for health summary''') \
    .depends(modules.parameter.parameters)


class GpuBdfPstate(tasks.ITask):
    def __init__(self):
        super().__init__()
        self.__result = None
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.__parse_nvidia_smi = parse_nvidia_smi
        self.depends(self.__parse_nvidia_smi)

    def getResult(self):
        return self.__result

    def run(self):
        try:
            gpus = self.__parameters_task.getResult()['gpu_link_info']
        except:
            return

        # If parameter is not found in platform parameters - Do Nothing
        if gpus == None:
            return

        nvidia_smi = self.__parse_nvidia_smi.getResult()
        if not nvidia_smi:
            return

        bdf_pstate = {}
        for _, info in nvidia_smi.items():
            gpu_dict = xmltodict.parse(info)
            bdf = None
            for gpu in gpus:
                try:
                    if gpu['bdf'].lower() in gpu_dict['nvidia_smi_log']['gpu']['@id'].lower():
                        bdf = gpu['bdf']
                        break
                except:
                    pass
            if not bdf:
                continue
            try:
                bdf_pstate[bdf] = gpu_dict['nvidia_smi_log']['gpu']['power_readings']['power_state']
            except:
                pass
        self.__result = bdf_pstate


gpu_bdf_pstate = GpuBdfPstate() \
    .title('''Get gpu pstate mapped to bdfs''') \
    .parameters('gpu_link_info')
