# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import sys
import re
import os
import logging
from nvsmhealth import modules, tasks, config
from nvsmhealth.tasks.stream import HealthTaskStream

# Avoid locale differences in numbers printed
import locale
locale.setlocale(locale.LC_ALL, "POSIX")

__all__ = [
    'nvme_list',
    'parse_nvme_devices',
    'check_nvme_devices',
    'collect_nvme_smart_log',
    'parse_nvme_smart_log',
    'check_nvme_smart_log'
]

nvme_list = tasks.RunCommand(
    args=['nvme', 'list', '--output-format=json']) \
    .title('''Collect list of NVMe devices using the nvme-cli tool''') \
    .depends(modules.common.check_superuser_privileges)


class ParseNvmeDevices(tasks.ITask):
    def __init__(self, nvme_list_task):
        super(ParseNvmeDevices, self).__init__()

        # Declare dependency on nvme list using the ITask interface
        self.__nvme_list = nvme_list_task
        self.depends(nvme_list)
        self.__result = None

    def getResult(self):
        return self.__result

    def quantize_bytes(self, num_bytes):
        # Quantize bytes into 100GB increments, expressed in TB
        num_terabytes = float(num_bytes) / float(1e12)
        return '{0:.1f}TB'.format(num_terabytes)

    def run(self):
        import json
        # Return if nvme run throws error or no results
        if self.__nvme_list.getReturnCode() or not self.__nvme_list.getOutput():
            return

        # Parse the result as a JSON object from parent task
        nvme_list_in = self.__nvme_list.getOutput()
        nvme_stream = json.loads(nvme_list_in)
        count = {}
        try:
            # Attempt to quantize NVMe device counts in 100GB increments
            devices = nvme_stream['Devices']
            for d in devices:
                num_bytes = d.get("PhysicalSize", 0)
                quantized = self.quantize_bytes(num_bytes)
                if quantized not in count:
                    count[quantized] = 0
                count[quantized] += 1
        except:
            # unable to parse or no devices in the nvme output
            pass
        # Add the parsed result as JSON object
        self.__result = json.dumps(count, ensure_ascii=True)


parse_nvme_devices = ParseNvmeDevices(nvme_list) \
    .title('''Parse nvme list information gathered from nvme-cli tool''') \



class CheckNvmeDevices(tasks.ICheck):
    def __init__(self, parse_nvme_devices_task, parameters_task):
        super(CheckNvmeDevices, self).__init__()

        # Declare dependency on nvme devices parse using the ITask interface
        self.__parse_nvme_devices = parse_nvme_devices_task
        self.__parameters_task = parameters_task
        self.depends(parse_nvme_devices)
        self.depends(parameters_task)

    def run(self):
        import json

        own_params = self.getParameters()
        params = self.__parameters_task.getResult()

        nvme_config = None
        for param, req in own_params.items():
            if param in params.keys():
                nvme_config = params[param]

        if not nvme_config:
            return

        nvsm_config = config.read_nvsm_config_file()

        if nvsm_config != None:
            if not nvsm_config["use_standard_config_storage"]:
                return

        if self.__parse_nvme_devices.getResult():
            self.addCheckMessage(config.nvme_command_str)
            devices = json.loads(self.__parse_nvme_devices.getResult())
            if not [conf for conf in nvme_config if devices.items() == conf.items()]:
                count = 0
                res = 'Supported NVMe device(s) configuration:\n'
                for conf in nvme_config:
                    for size, count in conf.items():
                        res += '"{}" NVMe device(s) with capacity "{}"\n'.format(
                            count, size)
                    if count < len(nvme_config):
                        res += 'or \n'
                        count += 1

                res += 'Found NVMe device(s) configuration:'
                for size, count in devices.items():
                    res += '\n"{}" NVMe device(s) with capacity "{}"'.format(
                        count, size)
                self.addCheckMessage(res)
                self.addUnHealthy()

            else:
                self.addHealthy()
        else:
            self.addCheckMessage("No results from parse nvme devices")
            self.addUnknown()


check_nvme_devices = CheckNvmeDevices(parse_nvme_devices, modules.parameter.parameters) \
    .title(config.nvme_info_str) \
    .parameters('nvme')


class CollectNvmeSmartLog(tasks.ITask):
    def __init__(self):
        super().__init__()
        self.depends(modules.common.check_superuser_privileges)
        self.depends(nvme_list)
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('nvme_check_smart_log')
        self.__nvme_list = nvme_list
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        try:
            check = self.__parameters_task.getResult()['nvme_check_smart_log']
            if not check:
                return
        except:
            return

        # Return when no output from nvme list
        if self.__nvme_list.getReturnCode() or not self.__nvme_list.getOutput():
            return
        import json
        nvme_stream = json.loads(self.__nvme_list.getOutput())
        nvme_devices = []
        try:
            devices = nvme_stream['Devices']
            for device in devices:
                nvme_devices.append(device['DevicePath'])
        except Exception as e:
            logging.debug("Error while parsing nvsm devices:{}".format(e))
            pass

        nvme_smart_log = {}
        for device in nvme_devices:
            args = "nvme smart-log " + device
            cmd_task = tasks.RunCommand(args.split()).depends(
                modules.common.check_superuser_privileges)
            cmd_task.run()
            if cmd_task.getReturnCode():
                # could not run the info command
                continue
            out = cmd_task.getOutput().strip()
            nvme_smart_log[device] = out
        self.__result = nvme_smart_log


collect_nvme_smart_log = CollectNvmeSmartLog() \
    .title("Collect nvme smartlogs")


class ParseNvmeSmartLog(tasks.ITask):
    def __init__(self):
        super().__init__()
        self.__collect_nvme_smart_log = collect_nvme_smart_log
        self.depends(self.__collect_nvme_smart_log)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        nvme_smartlog_out = self.__collect_nvme_smart_log.getResult()
        if not nvme_smartlog_out:
            return

        nvme_smart_log = {}
        percent_pattern = re.compile(r'^(?P<percent_value>\d+(?:\.\d+)?)\s*%$')

        def percent_to_float(percent):
            m = percent_pattern.match(percent)
            if not m:
                return percent
            try:
                percent_value = float(m.group('percent_value'))
            except:
                return percent
            return percent_value * 1e-2

        for device, out in nvme_smartlog_out.items():
            field_pattern = re.compile(
                r'^(?P<name>[^:]*):(?P<value>.*)$', flags=re.MULTILINE)
            smartlog = {}
            # Iterate over fields of each device
            for f in field_pattern.finditer(out):
                # Normalize the field name and store the field value
                name = f.group('name').strip().lower().replace(' ', '_')
                value = f.group('value').strip()
                # Convert percentage values to floats
                if percent_pattern.match(value):
                    value = percent_to_float(value)
                smartlog[name] = value
            nvme_smart_log[device.split(os.sep)[-1]] = smartlog

        self.__result = nvme_smart_log


parse_nvme_smart_log = ParseNvmeSmartLog() \
    .title('Parse and create NVME smart log objects')


class CheckNvmeSmartLog(tasks.ICheck):
    def __init__(self):
        super().__init__()
        self.__nvme_smart_log_task = parse_nvme_smart_log
        self.depends(self.__nvme_smart_log_task)
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.__nvme_list = nvme_list
        self.depends(nvme_list)
        self.parameters('nvme_check_smart_log')
        self.parameters('skip_nvme_drive_model')

    def run(self):
        try:
            check = self.__parameters_task.getResult()['nvme_check_smart_log']
            if not check:
                return
        except:
            return

        # NVBUG2794792: Toshiba/Kioxia CM5: medium error observed in SMART log when I/O sent to the locked drive
        skipDriveModel = None
        try:
            skipDriveModel = self.__parameters_task.getResult()[
                'skip_nvme_drive_model']
        except:
            pass

        import json
        # Return if nvme run throws error or no results
        nvme_drive_map = {}
        if not self.__nvme_list.getReturnCode() and self.__nvme_list.getOutput():
            try:
                nvme_stream = json.loads(nvme_list.getOutput())
                devices = nvme_stream['Devices']
                for device in devices:
                    d = device['DevicePath']
                    deviceName = d.split(os.sep)[-1]
                    nvme_drive_map[deviceName] = device['ModelNumber']
            except Exception as e:
                logging.debug("Error while parsing nvsm devices:{}".format(e))
                pass

        self.addCheckMessage(
            "Checking output of 'nvme smart-log' for each NVMe device")
        nvme_smart_log = self.__nvme_smart_log_task.getResult()
        if not nvme_smart_log:
            self.addUnknown()
            return

        healthy = True
        for name, device in nvme_smart_log.items():
            # Check for critical warnings, which indicate drive is in error state
            critical_warnings = device.get('critical_warning', 0)
            if critical_warnings != '0':
                self.addCheckMessage('Found {0} critical warning(s) on NVMe drive "{1}".'.format(
                    critical_warnings, name))
                healthy = False

            # Check that remaining spare capacity is above the threshold
            available_spare = device.get('available_spare', 1.0)
            available_spare_threshold = device.get(
                'available_spare_threshold', 0.1)
            if available_spare < available_spare_threshold:
                self.addCheckMessage('Remaining spare capacity of {remaining}% on NVMe drive "{drive}" fails to meet threshold of {threshold}%.'.format(
                    drive=name,
                    remaining=int(available_spare * 1e2),
                    threshold=int(available_spare_threshold * 1e2)))
                healthy = False

            # Check that vendor estimate of percentage used is below 90%
            used = device.get('percentage_used', 0.0)
            if used > 0.9:
                self.addCheckMessage('Over {used}% expected life used on NVMe drive "{drive}".'.format(
                    used=int(used * 1e2),
                    drive=name))
                healthy = False

            # Skip media error check for locked devices
            if nvme_drive_map[name] == skipDriveModel:
                continue
            # Check for media errors, which occur when the controller detects
            # unrecovered data integrity errors
            media_errors = device.get('media_errors', 0)
            if media_errors != '0':
                self.addCheckMessage('Found {0} media error(s) on NVMe drive "{1}".'.format(
                    media_errors, name))
                healthy = False

        if healthy:
            self.addHealthy()
        else:
            self.addUnHealthy()


check_nvme_smart_log = CheckNvmeSmartLog() \
    .title('Check SMART status of NVMe devices')
