# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import sys
import json
import os
import re
import subprocess
import importlib
import logging
from nvsmhealth import tasks, config, platforms

__all__ = [
    'determine_platform'
]

class PlatformNotFound(Exception):
    """Raised when platfrom not found"""
    pass

class DeterminePlatform(tasks.ITask):

    def __init__(self):
        super(DeterminePlatform, self).__init__()

        # Construct IPlatform objects for all known platform configurations
        self.__platforms = self.discover_platforms()

        # Forward any task dependencies declared by IPlatform objects
        # NOTE: These are the tasks used by IPlatform objects to perform
        #       platform determination.
        __task = []
        for p in self.__platforms:
            for task in p.getDependencies():
                if task not in __task:
                    __task.append(task)
        for t in __task:
            self.depends(t)

        self.__platform = None

    def getPlatform(self):
        return self.__platform

    def discover_platforms(self):

        result = []

        module_names = platforms.__all__
        for module_name in module_names:
            module = importlib.import_module('nvsmhealth.platforms.{}'.format(module_name))
            logging.info('Found NVSMHealth module: {}'.format(module_name))
            for member_name in module.__all__:
                if member_name not in module.__dict__:
                    logging.warning(
                        'Missing member "{member}" in NVSMHealth module "{module}"'.format(
                        member=member_name,
                        module=module_name))
                    continue
                member = module.__dict__[member_name]
                _platforms = member()
                if _platforms:
                    result.extend(_platforms)

        return result

    def run(self):
        # Loop all over IPlatform objects to find correct variant
        matching_platform = None
        for platform in self.__platforms:
            if not platform.matches():
                continue
            if matching_platform != None and matching_platform.getName() != platform.getName():
                # IPlatform objects *must* be mutually exclusive
                logging.error('ERROR: Multiple matching platform configs found')
                raise Exception('Multiple matching platform configs found')

            matching_platform = platform

        if matching_platform == None:
            raise PlatformNotFound

        logging.debug("Platform Selection Complete: {} {} ".format(matching_platform.getName(),
            matching_platform.getPrettyConfig()))
        self.__platform = matching_platform

determine_platform = DeterminePlatform() \
    .title('''Determine platform using IPlatform objects''')

