# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from nvsmhealth import modules, tasks

__all__ = [
    'collect_psu_info',
    'check_psu_info'
]

class CollectPsuInfo(tasks.ITask):
    def __init__(self, ipmi_info_task):
        super(CollectPsuInfo, self).__init__()
        self.__parameters_task = modules.parameter.parameters
        self.__ipmi_info_task = ipmi_info_task
        self.depends(self.__parameters_task)
        self.parameters('psu_info')
        self.__result = None
        self.__retCode = 1

    def getResult(self):
        return self.__result

    def getReturnCode(self):
        return self.__retCode

    def get_data(self, cmd_out):
        out = cmd_out.replace("\n","").strip().split(" ")
        # Get the length from second byte
        l = int(out[1], 16)
        # Get data from the length
        data = out[2:(l+2)]
        # Convert hex into ascii string
        return bytes.fromhex(" ".join(data)).decode('ascii')

    def run(self):
        if self.__ipmi_info_task.getReturnCode():
            # ipmitool isn't working so return
            return
        
        psu_vendor_info = None
        psu_model_info = None

        # Get psu_info_cmds from parameters if not return None
        try:
            psu_vendor_info = self.__parameters_task.getResult()['psu_vendor_info']
            psu_model_info = self.__parameters_task.getResult()['psu_model_info']
        except:
            return None

        psu_vendor_res = {}
        psu_model_res = {}

        '''
        psu_vendor_info looks like
        {'PSU0': '0x30 0x0e 0x00 0x02', .. } 

        psu_model_info looks like
        {'PSU0': '0x30 0x0e 0x00 0x01', .. } 
        '''

        try:
            ## Run all PSU Vendor info cmds and fill psu_vendor_res
            for i in psu_vendor_info.keys():
                cmd = 'ipmitool raw {}'.format(psu_vendor_info[i])
                cmd_task = tasks.RunCommand(shell=True, args=[cmd])
                cmd_task.run()
                if cmd_task.getReturnCode() != 0:
                    return None
                psu_vendor_res[i] = self.get_data(cmd_task.getOutput())
        except:
            self.__result = None
            self.__retCode = 1
            return

        try:
            ## Run all PSU Model info cmds and fill psu_vendor_res
            for i in psu_model_info.keys():
                cmd = 'ipmitool raw {}'.format(psu_model_info[i])
                cmd_task = tasks.RunCommand(shell=True, args=[cmd])
                cmd_task.run()
                if cmd_task.getReturnCode() != 0:
                    return None
                psu_model_res[i] = self.get_data(cmd_task.getOutput())
        except:
            self.__result = None
            self.__retCode = 1
            return

        # Return result is a list 
        self.__result = []
        # Append all vendor information
        self.__result.append(psu_vendor_res)
        # Append all model information 
        self.__result.append(psu_model_res)
        # Set retCode as 0
        self.__retCode = 0

collect_psu_info = CollectPsuInfo(modules.ipmitool_cmd.run_ipmi_info) \
    .title(''' Collect PSU Information ''') \
    .depends(modules.common.check_superuser_privileges) \
    .depends(modules.ipmitool_cmd.run_ipmi_info) \
    .parameters('psu_vendor_info') \
    .parameters('psu_model_info') \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_psu_consistency_check'))

class CheckPsuInfo(tasks.ICheck):
    def __init__(self, collect_psu_info_task):
        super(CheckPsuInfo, self).__init__()
        # Declare dependency on parse lspci task using the ITask interface
        self.__collect_psu_info_task = collect_psu_info_task

    def run(self):
        # print(self.__collect_psu_info_task.getReturnCode(), self.__collect_psu_info_task.getResult())

        if self.__collect_psu_info_task.getReturnCode():
            self.addCheckMessage("Unable to collect PSU (Vendor, Model) Information.")
            self.addUnknown()
            return
    
        psu_info = self.__collect_psu_info_task.getResult()

        for item in psu_info:
            s = set()
            for i in item.keys():
                k = i.split("_")[0]
                s.add(item[i])
            if len(s) > 1:
                self.addCheckMessage("Multiple PSU {}s found. {}".format(k,[i for i in s]))
                self.addUnHealthy()
                return
                
        self.addHealthy()


check_psu_info = CheckPsuInfo(collect_psu_info) \
    .title('''Check PSU Info (Vendor, Model) for Consistency''') \
    .depends(collect_psu_info) \
    .predicate(lambda: modules.parameter.parameters.getResult().get('enable_psu_consistency_check'))
