# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import json
import os
import re
import sys

from nvsmhealth import modules, tasks

__all__ = [
    'run_smartctl_scan',
    'parse_smartctl_scan',
    'parse_smartctl_device_info',
    'parse_smartctl_system_disk_info',
]

run_smartctl_scan = tasks.RunCommand(
        args=['smartctl', '--scan']) \
    .title('''Run the smartctl utility''') \
    .describe('''
            Run the "smartctl" utility to scan for devices.
            Some flags are added to output in a machine-readable format.
            This command requires superuser privileges.''') \
    .depends(modules.common.check_superuser_privileges)

class ParseSmartctlScan(tasks.ITask):
    def __init__(self, smartctl_task, lsblk_task):
        super(ParseSmartctlScan, self).__init__()

        # Declare dependency on smartctl using the ITask interface
        self.__smartctl_task = smartctl_task
        self.__lsblk_task = lsblk_task
        self.depends(run_smartctl_scan)
        self.depends(lsblk_task)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        # Avoid locale differences in numbers printed by smartctl
        os.environ['LC_NUMERIC'] = 'POSIX'

        # Parse the result as a JSON object from parent task
        devices = []
        comment_pattern = re.compile(r'^(?P<value>[^#]*)(#(?P<comment>.*))?$')
        space_pattern = re.compile('\s+')
        if self.__smartctl_task.getReturnCode() == 0:
            out = self.__smartctl_task.getOutput().strip()
            lines = out.split('\n')
            for l in lines:
                l = l.strip()
                if not l:
                    # Ignore blank lines
                    continue
                m = comment_pattern.match(l)
                if m:
                    # Ignore comments
                    l = m.group('value').strip()
                # Split the string into arguments
                args = space_pattern.split(l)
                devices.append(args)

        # NVBugs-200587168: NGC clusters contains scsi VDisks
        # These disks needs to filtered to avoid smartctl info hang failures
        # Perform smartctl operation on physical devices only filtered from lsblk tool
        if not self.__lsblk_task.getResult():
            return

        lsblk_devices = self.__lsblk_task.getResult()

        # Backup megaraid devices
        megaraid_devices = []
        for device in devices:
            for d in device:
                if "megaraid" in d:
                    megaraid_devices.append(device)

        filter_devices = []
        for lsblk_device in lsblk_devices:
            device_name = "/dev/{}".format(lsblk_device['name'])
            # Filter smart devices
            for smart_device in devices:
                if device_name in smart_device:
                    filter_devices.append(smart_device)

        # Append megaraid devices, As lsblk doesn't provide its information
        filter_devices.extend(megaraid_devices)
        self.__result = filter_devices

parse_smartctl_scan = ParseSmartctlScan(run_smartctl_scan, modules.lsblk.parse_lsblk_scsi_device_info) \
    .title('''Parse smartctl scan devices information gathered from smartctl tool''') \


class ParseSmartctlDeviceInfo(tasks.ITask):
    def __init__(self, smartctl_scan_task):
        super(ParseSmartctlDeviceInfo, self).__init__()

        # Declare dependency on smartctl using the ITask interface
        self.__smartctl_scan_task = smartctl_scan_task
        self.depends(parse_smartctl_scan)
        self.depends(modules.common.check_superuser_privileges)
        self.__result = None
        self.__bom_list = None
        self.__disks_by_size = None

    def getResult(self):
        return self.__result

    def getBomList(self):
        return self.__bom_list

    def getDisksBySize(self):
        return self.__disks_by_size

    def run(self):
        # Parse the result as a JSON object from parent task
        smartctl_field_pattern = re.compile(r'^\s*(?P<field>[^:]+):\s*(?P<value>.*)\s*$')
        space_pattern = re.compile('\s+')

        devices = self.__smartctl_scan_task.getResult()
        if not devices:
            return
        bom_list = {}

        for d in devices:
            # Invoke smartctl for this device
            args = ['smartctl', '--info']
            args.extend(d)
            info_task = tasks.RunCommand(args).depends(modules.common.check_superuser_privileges)
            info_task.run()
            if info_task.getReturnCode():
                # could not run the info command
                continue
            out = info_task.getOutput().strip()

            # Extract BOM fields for this device
            bom = {}
            lines = out.split('\n')
            for l in lines:
                l = l.strip()
                m = smartctl_field_pattern.match(l)
                if not m:
                    continue

                field = m.group('field').strip().lower()
                field = space_pattern.sub('_', field)
                value = m.group('value').strip()
                bom[field] = value
            bom_list[' '.join(d)] = bom
        self.__bom_list = bom_list

        # Count the disks by capacity
        disk_count = {}
        capacity_pattern = re.compile(r'^.*\[(?P<human_readable>.*)\]$')
        megaraid_pattern = re.compile(r'^.*megaraid.*$')
        for key, value in bom_list.items():
            m = megaraid_pattern.match(key)
            if not m:
                # Skip disks not in MegaRAID
                continue
            capacity = value.get('user_capacity')
            if not capacity:
                continue
            # Get the human-readable disk capacity
            m = capacity_pattern.match(capacity)
            if not m:
                continue
            capacity = m.group('human_readable')
            # Increment the disk count for this capacity
            if capacity not in disk_count:
                disk_count[capacity] = 0
            disk_count[capacity] += 1
        self.__result = disk_count

        # Quantize disk
        def quantize_bytes(num_bytes):
            # Quantize bytes into 100GB increments, expressed in TB
            num_gigabytes = float(num_bytes) / float(1e12)
            return '{0:.1f}TB'.format(num_gigabytes)

        disks_by_size = {}
        # Attempt to quantize disk counts in 100GB increments
        capacity_pattern = re.compile(r'^(?P<bytes>[\d,]+)\s*bytes')
        try:
            for key, value in bom_list.items():
                m = capacity_pattern.match(value['user_capacity'])
                num_bytes = int(m.group('bytes').replace(',', ''))
                quantized = quantize_bytes(num_bytes)
                if quantized not in disks_by_size:
                    disks_by_size[quantized] = 0
                disks_by_size[quantized] += 1
        except:
            pass
        self.__disks_by_size = disks_by_size

parse_smartctl_device_info = ParseSmartctlDeviceInfo(parse_smartctl_scan) \
    .title('''Parse smartctl device information gathered from parse smartctl scan tool''') \

class ParseSmartctlSystemDiskInfo(tasks.ITask):
    def __init__(self, smartctl_scan_task):
        super().__init__()

        # Declare dependency on smartctl using the ITask interface
        self.__smartctl_scan_task = smartctl_scan_task
        self.depends(parse_smartctl_scan)
        self.depends(modules.common.check_superuser_privileges)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        devices = self.__smartctl_scan_task.getResult()
        if not devices:
            return

        disk_info = {}

        for d in devices:
            pattern = re.compile(r'megaraid,[0-9]+')
            match = pattern.search(' '.join(d))
            if match:
                disk = d[2]
            else:
                disk = d[0]

            # Invoke smartctl for this device
            args = ['smartctl', '-a']
            args.extend(d)
            info_task = tasks.RunCommand(args).depends(modules.common.check_superuser_privileges)
            info_task.run()
            if info_task.getReturnCode():
                # could not run the info command
                continue
            out = info_task.getOutput().strip()

            keys = [('rotation_rate', "Rotation Rate:\s+(?P<{}>[\w ]+)"),
                    ('wear_leveling_count', '(?:177 Wear_Leveling_Count|233 Media_Wearout_Indicator)\\s+0[xX][0-9a-fA-F]+\\s+(?P<{}>\\d+)'),
                    ]

            disk_info[disk] = {}
            for key, pattern in keys:
                match_string = pattern.format(key)
                p = re.compile(r"{}".format(match_string))
                for m in p.finditer(out):
                    if key in  disk_info[disk]:
                        # consider first match only
                        continue
                    disk_info[disk][key] = m.group(key).strip()
                if key == 'wear_leveling_count':
                    try:
                        disk_info[disk][key] = int(disk_info[disk][key])
                    except:
                        pass
        self.__result = disk_info

parse_smartctl_system_disk_info = ParseSmartctlSystemDiskInfo(parse_smartctl_scan) \
    .title('''Verify disk information using the smartctl system command''')
