# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from nvsmhealth import tasks

import fnmatch
import logging
import os
import subprocess
import tarfile

__all__ = [
    "sosreport"
]


class SosreportLogs(tasks.ILogs):
    def __init__(self):
        super(SosreportLogs, self).__init__()
        # Give sosreport a generous 10 minutes to run
        timeout = os.getenv('SOSREPORT_TIMEOUT')
        try:
            timeout = int(timeout) * 60
        except:
            # default plugin timeout of sosreport is 5mins,
            # so we will timeout after 6mins
            timeout = 6 * 60
        self.__timeout = timeout

    def getTimeout(self):
        return self.__timeout

    def run(self):
        pass

    def collect(self, context):
        # Start sosreport with non-interactive flags
        sosreport_args = ["sosreport",
                          "--batch",
                          "--alloptions",
                          "--tmp-dir", context.tempdir,
                          "--name", "sosreport",
                          "--compression-type", "gzip"]

        # Check if we are running on a redhat system
        # Since ubuntu does not support "-n jars" argument,
        # we only provide this argument for redhat
        # http://nvbugs/200682309 - [RedOctober][NVSM 20.09.10] nvsm dump health does not collect sosreport
        if os.path.isfile("/etc/redhat-release"):
            # Skipping jars plugin in sosreport -> NvBug 200679241
            sosreport_args.extend(["-n", "jars"])
        logging.info('Running sosreport with command "{0}"'.format(
            ' '.join(sosreport_args)))
        try:
            import pty
            # open pseudo-terminal to interact with subprocess
            master_fd, slave_fd = pty.openpty()
            p = subprocess.Popen(
                args=sosreport_args,
                cwd=context.tempdir,
                stdout=subprocess.DEVNULL,
                stdin=slave_fd,
                stderr=subprocess.PIPE,
                preexec_fn=os.setpgrp)
        except Exception as e:
            logging.error(f'Exception: {e}')
            return
        done = False

        sosreport_tmp_dir = None

        def get_sosreport_size():
            nonlocal sosreport_tmp_dir
            try:
                if not sosreport_tmp_dir:
                    for f in os.listdir(context.tempdir):
                        if fnmatch.fnmatch(f, 'sos.*'):
                            sosreport_tmp_dir = f'{context.tempdir}{os.path.sep}{f}'
                if sosreport_tmp_dir:
                    from pathlib import Path
                    sum = 0
                    for p in Path(sosreport_tmp_dir).rglob('*'):
                        try:
                            sum += p.stat().st_size
                        except Exception as e:
                            pass
                    if sum == 0:
                        # possibly the sosreport has completed
                        logging.info(
                            f'sos report {sosreport_tmp_dir} is empty')
                        return 0
                    logging.info(f'sos report {sosreport_tmp_dir} size: {sum}')
                    return sum
            except Exception as e:
                logging.debug(f'Exception: {e}')
                pass
            logging.debug('sosreport temp file not found')
            return 0

        import time
        # give some time for the sos report archive to get populated initially
        time.sleep(10)
        sosreport_size = get_sosreport_size()
        while not done:
            try:
                # Wait for sosreport process to finish
                stdout_data, stderr_data = p.communicate(
                    timeout=self.__timeout)
                done = True
            except subprocess.TimeoutExpired as e:
                logging.info(f'{e}')

                cur_sosreport_size = get_sosreport_size()
                if cur_sosreport_size == 0:
                    # sosreport is most likely completed.
                    logging.info(
                        f'sosreport is most likely going to complete, wait for archive creation')
                elif cur_sosreport_size > sosreport_size:
                    # if the size is growing, wait for another timeout
                    logging.info(
                        'sosreport growing in size, not killing on timeout.')
                    sosreport_size = cur_sosreport_size
                else:
                    # Kill the entire sosreport process group on timeout
                    logging.info("Killing sosreport on timeout.")
                    import signal
                    os.killpg(p.pid, signal.SIGKILL)
                    stdout_data, stderr_data = p.communicate()
                    done = True
            except Exception as e:
                done = True
                logging.error(f'Exception: {e}')
        logging.info('Received exit code {0} from sosreport with PID {1}'.format(
            p.returncode, p.pid))
        logging.info('stdout from sosreport:\n{0}'.format(stdout_data))
        logging.info('stderr from sosreport:\n{0}'.format(stderr_data))
        # Look for resulting sosreport output
        sosreport_output = None
        for f in os.listdir(context.tempdir):
            if fnmatch.fnmatch(f, 'sosreport*.tar.gz'):
                sosreport_output = f
                logging.debug('Found sosreport output file at "{0}"'.format(
                    sosreport_output))
                break
        if not sosreport_output:
            logging.info('Could not find sosreport output file')
            return
        sosreport_output = os.path.join(context.tempdir, sosreport_output)
        # Add all sosreport files to the collection context
        with tarfile.open(
                name=sosreport_output,
                mode='r') as sosreport_archive:
            info = sosreport_archive.next()
            while info:
                if info.isfile():
                    logging.info('Adding file "{0}" from sosreport'.format(
                        info.name))
                    buff = sosreport_archive.extractfile(info)
                    path = os.path.join(*(info.name.split(os.path.sep)[1:]))
                    context.add_file(fileobj=buff, tarinfo=info, path=path)
                elif info.issym():
                    logging.info('Adding symbolic link "{0} -> {1}" from sosreport'.format(
                        info.name, info.linkname))
                    path = os.path.join(*(info.name.split(os.path.sep)[1:]))
                    context.add_file(fileobj=None, tarinfo=info, path=path)
                info = sosreport_archive.next()


sosreport = SosreportLogs()
