# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import re
import json
from nvsmhealth import tasks, modules

__all__ = [
    'export_storcli_path',
    'run_storcli_pall',
    'run_storcli_vall',
    'run_storcli_version',
    'check_storcli_sanity_supported',
    'check_storcli_sanity_installed',
    'check_storcli_phy_links',
    'check_storcli_disk_state'
]

class ExportStorcliPath(tasks.ITask):
    def run(self):
        import os
        os.environ["PATH"] += os.pathsep + '/opt/MegaRAID/storcli'

export_storcli_path = ExportStorcliPath() \
    .title('''Export storcli path''')

run_storcli_version = tasks.RunCommand(
        args=['storcli64', '-v', '-NoLog']) \
    .title('''Run the storcli command''') \
    .depends(export_storcli_path)

run_storcli_pall = tasks.RunCommand(
        args=['storcli64', '/c0/pall', 'show', 'all', 'J']) \
    .title('''Run the storcli command''') \
    .depends(export_storcli_path) \
    .depends(modules.common.check_superuser_privileges)

run_storcli_vall = tasks.RunCommand(
        args=['storcli64', '/c0/vall', 'show', 'all', 'J']) \
    .title('''Run the storcli command''') \
    .depends(export_storcli_path) \
    .depends(modules.common.check_superuser_privileges)

class Version:
    def __init__(self, value, base=10):
        self._value = value.strip().split('-')[0]
        self._base = base
        self._validate()

    def _validate(self):
        fields = self._value.split('.')
        try:
            for f in fields:
                int(f, self._base)
        except ValueError:
            raise ValueError('Invalid version string: {0}'.format(self._value))

    def __ge__(self, other):
        fields_self = self._value.split('.')
        fields_other = other._value.split('.')
        while len(fields_self) < len(fields_other):
            fields_self.append('0')
        while len(fields_other) < len(fields_self):
            fields_other.append('0')
        for i in range(len(fields_self)):
            a = int(fields_self[i], self._base)
            b = int(fields_other[i], self._base)
            if a < b:
                return False
            if a > b:
                return True
        return True

    def __str__(self):
        return self._value

class CheckStorCliSanitySupported(tasks.ICheck):
    def __init__(self):
        super().__init__()
        self.__base_os_version_task = modules.system.base_os_version
        self.depends(self.__base_os_version_task)
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('storcli_platfrom_string')

    def run(self):
        try:
            storcli_platform_string = self.__parameters_task.getResult()['storcli_platform_string']
        except:
            # paramter not found
            return

        self.title(self.getTitle().format(storcli_platform_string))

        baseos_version = self.__base_os_version_task.getResult()
        if not baseos_version or 'sw_version' not in baseos_version:
            self.addCheckMessage("Error checking {} BaseOS version".format( storcli_platform_string))
            self.addUnknown()
            return

        # Check DGX BaseOS version for storcli support
        message = 'Installed {} BaseOS version "{}" '.format(storcli_platform_string, baseos_version)
        if Version(baseos_version['sw_version']) >= Version('3.1.6'):
            # DGX BaseOS 3.1.6 introduces support for the storcli64 utility
            message += 'should support storcli'
            self.addHealthy()
        else:
            message += 'does not support storcli'
            self.addUnHealthy()
        self.addCheckMessage(message)

check_storcli_sanity_supported = CheckStorCliSanitySupported() \
        .title('[sanity] {} BaseOS support for storcli utility')

class CheckStorCliSanityInstalled(tasks.ICheck):
    def __init__(self):
        super().__init__()
        self.__run_command = run_storcli_version
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__run_command)
        self.parameters('storcli_platfrom_string')

    def run(self):
        try:
            storcli_platform_string = self.__parameters_task.getResult()['storcli_platform_string']
        except:
            # paramter not found
            return
        if self.__run_command.getReturnCode():
            self.addUnHealthy()
            self.addCheckMessage('The storcli utility does not appear to be installed')
            self.addCheckMessage('Please ensure storcli64 is installed in the /opt/MegaRAID/storcli/ directory')
        else:
            self.addHealthy()

check_storcli_sanity_installed = CheckStorCliSanityInstalled() \
        .title('[sanity] MegaRAID storcli utility installed')

class CheckPhy(tasks.ICheck):
    def __init__(self, storcli_out, check_link, check_type):
        super().__init__()
        self.__storcli_out = storcli_out
        self.__check_link = check_link
        self.__check_type = check_type
        super().name("CheckPhy")

    def parameters(self, param, required=False):
        assert False, "Do not add parameter dependency on this class"

    def depends(self, task):
        assert False, "Do not add task dependency on this class"

    def run(self):
        try:
            if self.__check_type == "LINK_SPEED":
                pattern = re.compile(r'^\s*(?:(?:[Nn]o\s+[Ll]imit)|(?:12Gb/s))\s*$', flags=re.MULTILINE)
                message = "Observed PHY link speed \"{value}\" when \"No limit\" was expected"
                title = "MegaRAID PHY link speed [ /c0/p{0}  ][ Speed: {1}  ]".format(self.__check_link, "{}")
                val = self.__storcli_out['Controllers'][0]['Response Data']['PhyInfo'][self.__check_link]['Link_Speed']
                title = title.format(val)
            elif self.__check_type == "SAS_ADDRESS":
                pattern = re.compile(r'^0x(?!0{16})([a-fA-F0-9]{16})$', flags=re.MULTILINE)
                title = "MegaRAID SAS address [ /c0/p{0}  ][ {1}  ]".format(self.__check_link, "{}")
                val = self.__storcli_out['Controllers'][0]['Response Data']['PhyInfo'][self.__check_link]['SAS_Addr']
                title = title.format(val)
                message = "Observed invalid SAS address \"{}\"".format(val)
            elif self.__check_type == "SAS_PORT":
                pattern = re.compile(r'1', flags=re.MULTILINE)
                message = "SAS port at /c0/p{0} does not appear to be valid/working".format(self.__check_link)
                title = "MegaRAID SAS port valid [ /c0/p{0}  ]".format(self.__check_link)
                val = str(self.__storcli_out['Controllers'][0]['Response Data']['PhyInfo'][self.__check_link]['Port_valid'])
        except:
            self.addUnknown()
            self.title(title)
            return
        match  = pattern.search(val)
        if not match:
            self.addCheckMessage(message.format(value=val))
            self.addUnHealthy()
        else:
            self.addHealthy()
        self.title(title)

class CheckPhyLinks(tasks.ICheck):
    def __init__(self):
        super().__init__()

        self.__run_command = run_storcli_pall
        self.depends(self.__run_command)
        self.__parameters_task = modules.parameter.parameters
        self.parameters('storcli_phy_links')
        self.depends(self.__parameters_task)
        self.depends(check_storcli_sanity_supported)
        self.depends(check_storcli_sanity_installed)

    def run(self):
        if self.__run_command.getReturnCode():
            # command failed to execute
            return
        try:
            devices = self.__parameters_task.getResult()['storcli_phy_links']
        except:
            return
        output = json.loads(self.__run_command.getOutput().strip())

        for check in ["LINK_SPEED", "SAS_ADDRESS", "SAS_PORT"]:
            for device in devices:
                check_cls = CheckPhy(output, device, check)
                check_cls.setCallback(self.getCallback())
                check_cls.run()
                self.addHealthy(count=check_cls.getResult()['healthy'])
                self.addUnHealthy(count=check_cls.getResult()['unhealthy'])
                self.addUnknown(count=check_cls.getResult()['unknown'])
                check_cls.sendComplete()
                super().addMessages(check_cls.getMessages())

        # clear message as this task doesnt print anything
        self.title('')

check_storcli_phy_links = CheckPhyLinks()

class CheckDisk(tasks.ICheck):
    def __init__(self, storcli_out, name, key, idx, expected_val):
        super().__init__()
        self.__storcli_out = storcli_out
        self.__name = name
        self.__key = key
        self.__idx = idx
        self.__expected_val = expected_val
        super().name("CheckDisk")

    def parameters(self, param, required=False):
        assert False, "Do not add parameter dependency on this class"

    def depends(self, task):
        assert False, "Do not add task dependency on this class"

    def run(self):
        try:
            disk_title = "virtual"
            disk_message = "volume"
            title = ""
            if "PD" in self.__name:
                disk_title = "physical"
                disk_message = "physical disk"
            title = "MegaRAID {disk} disk state [ {name}  ][ State: {val}  ]".format(
                    disk = disk_title,
                    name = self.__name,
                    val = "{}"
                    )
            message = "Observed {disk_msg} state \"{value}\" when \"{expected}\" was expected".format(
                    disk_msg = disk_message,
                    value = "{}",
                    expected = self.__expected_val
                    )
            val = self.__storcli_out['Controllers'][0]['Response Data'][self.__key][self.__idx]['State']
            title = title.format(val)
            message = message.format(val)

        except:
            self.addUnknown()
            self.title(title)
            return
        if val != self.__expected_val:
            self.addCheckMessage(message)
            self.addUnHealthy()
        else:
            self.addHealthy()
        self.title(title)

class CheckDiskState(tasks.ICheck):
    def __init__(self):
        super().__init__()

        self.__run_command = run_storcli_vall
        self.depends(self.__run_command)
        self.__parameters_task = modules.parameter.parameters
        self.parameters('storcli_disk_stats')
        self.depends(self.__parameters_task)
        self.depends(check_storcli_sanity_supported)
        self.depends(check_storcli_sanity_installed)

    def run(self):
        if self.__run_command.getReturnCode():
            # command failed to execute
            return
        try:
            devices = self.__parameters_task.getResult()['storcli_disk_stats']
        except:
            return
        output = json.loads(self.__run_command.getOutput().strip())

        for name, key, idx, exp_val in devices:
            check_cls = CheckDisk(output, name, key, idx, exp_val)
            check_cls.setCallback(self.getCallback())
            check_cls.run()
            self.addHealthy(count=check_cls.getResult()['healthy'])
            self.addUnHealthy(count=check_cls.getResult()['unhealthy'])
            self.addUnknown(count=check_cls.getResult()['unknown'])
            check_cls.sendComplete()
            super().addMessages(check_cls.getMessages())

        # clear message as this task doesnt print anything
        self.title('')

check_storcli_disk_state = CheckDiskState()
