import re
from nvsmhealth import modules, tasks, config

__all__ = [
    "base_os_version",
    "uptime",
    "system_summary",
    "print_system_info"
]

uptime = tasks.RunCommand(
    args=['uptime', '-p']) \
    .title('Run uptime command') \
    .describe('''
                Check system uptime with the uptime utility''')

class BaseOsVersion(tasks.ITask):
    def __init__(self):
        super().__init__()
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.parameters('sw_rel_file')
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        try:
            rel_file = self.__parameters_task.getResult()['sw_rel_file']
        except:
            return
        rel_file = config.etc_rel_file.format(rel_file)
        with open(rel_file, encoding="utf-8") as f:
            output = f.read()

        res = {}
        # Get the latest OTA version, always appended to release file
        for m in re.compile(r"_OTA_VERSION\=(?P<ver>.*)").finditer(output):
            res['sw_version'] = m.group('ver').strip().strip('"')

        # Look for build version if OTA version not found
        if not res:
            keys = [('sw_version', "_SWBUILD_VERSION\=(?P<{}>.*)"),
                    ('sw_version', "VERSION\=(?P<{}>.*)")]

            for key, pattern in keys:
                match_string = pattern.format(key)
                p = re.compile(r"{}".format(match_string))
                for m in p.finditer(output):
                    if key in res.keys():
                        # consider first match only
                        continue
                    res[key] = m.group(key).strip().strip('"')

        self.__result = res


base_os_version = BaseOsVersion() \
    .title('Get base OS version')


class SystemSummary(tasks.ITask):
    def __init__(self):
        super().__init__()
        self.__parse_uname_task = modules.uname.parse_uname
        self.depends(self.__parse_uname_task)
        self.__base_os_version_task = base_os_version
        self.depends(self.__base_os_version_task)
        self.__uptime_task = uptime
        self.depends(self.__uptime_task)
        self.__ipmi_fru_task = modules.ipmitool.parse_ipmi_fru
        self.depends(self.__ipmi_fru_task)
        self.__dmidecode_task = modules.dmidecode.parse_dmidecode
        self.depends(self.__dmidecode_task)
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.__ipmi_info_task = modules.ipmitool_cmd.parse_ipmi_info
        self.depends(self.__ipmi_info_task)
        self.__firmware_revision_task = modules.ipmitool.parse_firmware_revision
        self.depends(self.__firmware_revision_task)
        self.__nvidia_driver_version_task = modules.nvidia_smi.nvidia_driver_version
        self.depends(self.__nvidia_driver_version_task)
        self.__nvidia_vbios_version_task = modules.nvidia_smi.nvidia_vbios_version
        self.depends(self.__nvidia_vbios_version_task)
        self.__check_gpus_task = modules.nvidia_smi.check_gpus
        self.depends(self.__check_gpus_task)
        self.__cec_data_task =  modules.cec_info.parse_cec_info
        self.depends(self.__cec_data_task)
        self.__result = None

        # Prepare default values for the system summary
        self.__system_summary = {
            'chassis': {
                'serial_number': 'Unknown',
            },
            'system': {
                'product_name': 'Unknown',
                'manufacturer': 'Unknown',
                'serial_number': 'Unknown',
                'uptime': 'Unknown',
            },
            'motherboard': {
                'firmware_version': 'Unknown',
                'bios_revision': 'Unknown',
                'serial_number': 'Unknown'
            },
            'bmc': {
                'firmware_version': 'Unknown',
                'ipmi_version': 'Unknown'
            },
            'gpu': {
                'product_name': 'Unknown',
                'driver_version': 'Unknown',
                'vbios_version': 'Unknown'
            },
            'software': {
                'baseos_version': 'Unknown',
                'kernel_version': 'Unknown'
            },
            'CEC': {
                'cec_version': 'Unknown',
                'EC_FW_TAG0': 'Unknown',
                'EC_FW_TAG1': 'Unknown',
                'bmc_boot_slot': 'Unknown',
            }
        }

    def getResult(self):
        return self.__result

    def getSummary(self):
        return self.__system_summary

    def run(self):
        fmt = ""
        # Get the system uptime
        uptime_output = self.__uptime_task.getOutput().strip()
        if uptime_output:
            if uptime_output.lower() == 'up':
                self.__system_summary['system']['uptime'] = 'up less than 1 minute'
            else:
                self.__system_summary['system']['uptime'] = uptime_output

        # Get info from impitool FRU
        fru_output = self.__ipmi_fru_task.getResult()

        # Get the system serial number from ipmitool FRU
        chassis_serial_number = None
        try:
            index = self.__parameters_task.getResult()[
                'chassis_serial_number_index']
            fru_task = tasks.RunCommand(args=['ipmitool', 'fru', 'print', index]) \
                .depends(modules.common.check_superuser_privileges)
            fru_task.run()
            output = fru_task.getOutput().strip()
            for line in output.split('\n'):
                line = line.strip()
                if not line:
                    continue
                field = line.split(':', 1)
                if len(field) != 2:
                    continue
                field_name = field[0].strip().lower().replace(' ', '_')
                if 'chassis_serial' == field_name:
                    chassis_serial_number = field[1].strip()
        except:
            pass

        # Get info from DMI table
        dmi_output = self.__dmidecode_task.getResult()
        if dmi_output:
            chassis_info = None
            for k,v in dmi_output.items():
                if 'Chassis Information' in k:
                    chassis_info = v
                    break
            if chassis_info:
                for key in [ 'manufacturer', 'serial_number' ]:
                    if key in chassis_info:
                        self.__system_summary['chassis'][key] = chassis_info[key]

            system_info = None
            for k, v in dmi_output.items():
                if 'System Information' in k:
                    system_info = v
                    break
            if system_info:
                for key in ['manufacturer', 'product_name', 'serial_number']:
                    if key in system_info:
                        self.__system_summary['system'][key] = system_info[key]
            if 'DGX-2' in self.__system_summary['system']['product_name']:
                # Obtain system serial number from mid-plane
                try:
                    self.__system_summary['system']['serial_number'] = fru_output['Mid_FRU']['chassis_serial']
                except:
                    self.__system_summary['system']['serial_number'] = 'Unknown'

            bios_info = None
            for k, v in dmi_output.items():
                if 'BIOS Information' in k:
                    bios_info = v
                    break
            if bios_info:
                if 'version' in bios_info:
                    self.__system_summary['motherboard']['firmware_version'] = bios_info['version']
                if 'bios_revision' in bios_info:
                    self.__system_summary['motherboard']['bios_revision'] = bios_info['bios_revision']
            base_board_info = None
            for k, v in dmi_output.items():
                if 'Base Board Information' in k:
                    base_board_info = v
                    break
            if base_board_info:
                if 'serial_number' in base_board_info:
                    self.__system_summary['motherboard']['serial_number'] = base_board_info['serial_number']

        # Overwrite serial number if found in FRU
        if chassis_serial_number:
            self.__system_summary['system']['serial_number'] = chassis_serial_number
        # Look for BaseOS version in /etc/<platform>-release
        sw_rel_output = self.__base_os_version_task.getResult()
        key = 'sw_version'
        if sw_rel_output and key in sw_rel_output:
            self.__system_summary['software']['baseos_version'] = sw_rel_output[key]

        # Look for kernel version from uname
        uname_output = self.__parse_uname_task.getResult()
        key = 'kernel-release'
        if uname_output and key in uname_output:
            self.__system_summary['software']['kernel_version'] = uname_output[key]

        # Look for GPU information from nvidia-smi
        driver_version = self.__nvidia_driver_version_task.getResult()
        if driver_version != None:
            self.__system_summary['gpu']['driver_version'] = driver_version

        gpus_dict = self.__check_gpus_task.getGPUResult()
        if gpus_dict != None:
            vbios_versions = set()
            product_names = set()
            for gpu, info in gpus_dict.items():
                try:
                    # Create list of unique VBIOS versions
                    vbios_versions.add(info['vbios_version'])
                    vbios_version = ' ; '.join(vbios_versions)
                    # Create list of unique product names
                    product_names.add(info['product_name'])
                    product_name = ' ; '.join(product_names)
                except:
                    pass

            self.__system_summary['gpu']['vbios_version'] = vbios_version
            self.__system_summary['gpu']['product_name'] = product_name

        # Look for the BMC firmware version
        bmc_output = self.__ipmi_info_task.getResult()
        if bmc_output:
            if 'firmware_revision' in bmc_output:
                self.__system_summary['bmc']['firmware_version'] = self.__firmware_revision_task.getResult()
            if 'ipmi_version' in bmc_output:
                self.__system_summary['bmc']['ipmi_version'] = bmc_output['ipmi_version']

        # Look for the CEC Info
        cec_output = self.__cec_data_task.getResult()
        if cec_output:
            # Caller will take care of filling all values
            self.__system_summary['CEC'] = cec_output
            # for key, value in cec_output.items():
            #     if 'cec_version' == key:
            #         self.__system_summary['CEC']['cec_version'] = value
            #     if 'EC_FW_TAG0' == key:
            #         self.__system_summary['CEC']['EC_FW_TAG0'] = value
            #     if 'EC_FW_TAG1' == key:
            #         self.__system_summary['CEC']['EC_FW_TAG1'] = value
            #     if 'bmc_boot_slot' == key:
            #         self.__system_summary['CEC']['bmc_boot_slot'] = value

        try:
            summary_serial_num_string = self.__parameters_task.getResult()[
                'summary_serial_num_string']
            summary_base_os_string = self.__parameters_task.getResult()[
                'summary_base_os_string']
        except:
            return

        if 'RTX' in self.__system_summary['system']['product_name']:
            fmt += '\n' \
                  'System Summary\n' \
                  '--------------\n' \
                  '    Product Name: {system[product_name]}\n' \
                  '    Manufacturer: {system[manufacturer]}\n' \
                  '    Chassis Serial Number: {chassis[serial_number]}\n' \
                  '    {0}: {system[serial_number]}\n' \
                  '    Uptime: {system[uptime]}\n' \
                  'Motherboard:\n' \
                  '    BIOS Version: {motherboard[firmware_version]}\n' \
                  '    Serial Number: {motherboard[serial_number]}'.format(summary_serial_num_string, **self.__system_summary)
        else:
            fmt += '\n' \
                  'System Summary\n' \
                  '--------------\n' \
                  '    Product Name: {system[product_name]}\n' \
                  '    Manufacturer: {system[manufacturer]}\n' \
                  '    {0}: {system[serial_number]}\n' \
                  '    Uptime: {system[uptime]}\n' \
                  'Motherboard:\n' \
                  '    BIOS Version: {motherboard[firmware_version]}\n' \
                  '    Serial Number: {motherboard[serial_number]}'.format(summary_serial_num_string, **self.__system_summary)

        # Only print the BMC section if the system appears to have a BMC
        if bmc_output:
            fmt += '\nBMC:\n' \
                '    Firmware Version: {bmc[firmware_version]}\n' \
                '    IPMI Version: {bmc[ipmi_version]}'.format(
                    **self.__system_summary)

        fmt += '\nGPU:\n' \
            '    NVIDIA Driver Version: {gpu[driver_version]}\n' \
            '    Product Name(s): {gpu[product_name]}\n' \
            '    VBIOS Version(s): {gpu[vbios_version]}\n' \
            'Software:\n' \
            '    {0}: {software[baseos_version]}\n' \
            '    Kernel Version: {software[kernel_version]}'.format(
                summary_base_os_string, **self.__system_summary)

        if cec_output:
            fmt += '\nCEC:\n' \
                '    CEC Version: {CEC[cec_version]}\n' \
                '    EC_FW_TAG0: {CEC[EC_FW_TAG0]}\n' \
                '    EC_FW_TAG1: {CEC[EC_FW_TAG1]}\n' \
                '    BMC FW authentication state: {CEC[bmc_fw_auth_state]}\n' \
                '    Currently running BMC Slot : {CEC[bmc_boot_slot]}'.format(
                    **self.__system_summary)

        self.__result = fmt


system_summary = SystemSummary() \
    .title('Prepare System Summary') \
    .parameters('summary_base_os_string') \
    .parameters('summary_serial_num_string')


class PrintSystemInfo(tasks.ITask):
    def __init__(self):
        super().__init__()
        self.__system_summary_task = system_summary
        self.depends(self.__system_summary_task)
        self.__parameters_task = modules.parameter.parameters
        self.depends(self.__parameters_task)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        try:
            summary_base_os_string = self.__parameters_task.getResult()[
                'summary_base_os_string']
            summary_serial_num_string = self.__parameters_task.getResult()[
                'summary_serial_num_string']
        except:
            return
        self.title(summary_base_os_string)
        self.send(msg=self.__system_summary_task.getSummary()
                  ['software']['baseos_version'])

        self.title("BIOS Version")
        self.send(msg=self.__system_summary_task.getSummary()
                  ['motherboard']['firmware_version'])

        self.title("Linux kernel version")
        self.send(msg=self.__system_summary_task.getSummary()
                  ['software']['kernel_version'])

        self.title("System Uptime")
        self.send(msg=self.__system_summary_task.getSummary()
                  ['system']['uptime'])

        self.title(summary_serial_num_string)
        self.send(msg=self.__system_summary_task.getSummary()
                  ['system']['serial_number'])


print_system_info = PrintSystemInfo() \
    .title('System Information') \
    .parameters('summary_base_os_string')
