# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import re
from nvsmhealth import modules, tasks

__all__ = [
    'run_xl_info',
    'parse_xl_info',
    'check_xenserver_logical_core_count',
]

run_xl_info = tasks.RunCommand(
    args=['xl', 'info']) \
    .title('''Run the "xl info" command for XenServer information''') \
    .describe('''The "xl info" command prints basic information about the
                 running XenServer hypervisor.''') \
    .depends(modules.common.check_superuser_privileges)

class ParseXlInfo(tasks.ITask):
    def __init__(self):
        super().__init__()
        self.depends(run_xl_info)
        self.__result = None

    def getResult(self):
        return self.__result

    def run(self):
        self.__result = {}
        if run_xl_info.getReturnCode() != 0:
            return
        output = run_xl_info.getOutput()
        item_pattern = re.compile(r'^(?P<name>\S+)\s+:\s+(?P<value>.*)$', flags=re.MULTILINE)
        for m in item_pattern.finditer(output):
            try:
                self.__result[m.group('name')] = int(m.group('value').strip())
            except:
                self.__result[m.group('name')] = m.group('value').strip()

parse_xl_info = ParseXlInfo() \
    .title('''Parse output from the "xl info" command''') \
    .describe('''The output from "xl info" is in a simple colon-separated
                 format that this task parses into a Python dict.''') \

class CheckXenServerLogicalCoreCount(tasks.ICheck):
    def __init__(self):
        super().__init__()
        self.depends(modules.parameter.parameters)
        self.depends(parse_xl_info);
        self.parameters('xenserver_number_of_cores');

    def run(self):
        try:
            expected_core_count = modules.parameter.parameters.getResult()['xenserver_number_of_cores']
        except:
            return

        xlinfo_result = parse_xl_info.getResult()
        if not xlinfo_result:
            self.addUnknown()
            return

        observed_core_count = xlinfo_result['nr_cpus']
        self.title(self.getTitle().format(observed_core_count))

        if observed_core_count == expected_core_count:
            self.addCheckMessage('Observed {0} logical CPU cores, matching expectations'.format(
                        observed_core_count))
            self.addHealthy()
            return

        if observed_core_count * 2 == expected_core_count:
            # When only half of the expected logical cores are observed, we
            # suspect hyperthreading might be disabled

            # Look for the hyperthreading flag
            hyperthreading_enabled = xlinfo_result['threads_per_core'] == 2

            if not hyperthreading_enabled:
                self.addCheckMessage('It appears that Hyper-Threading is disabled.' \
                      ' Some customers choose to disable Hyper-Threading in' \
                  ' order to improve the performance of certain' \
                  ' workloads. If Hyper-Threading was intentionally' \
                  ' disabled, please ignore this message.')
                self.addUnHealthy()
                return

        self.addCheckMessage('Observed {0} logical CPU cores when {1} cores were expected'.format(
                    observed_core_count, expected_core_count))
        self.addUnHealthy()

check_xenserver_logical_core_count = CheckXenServerLogicalCoreCount() \
        .title('Number of logical CPU cores [{0}]')
