# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import sys

import nvsmhealth.exceptions
import inspect
import logging
from nvsmhealth import tasks

class IPlatform(object):
    ''' Class representing a generic platform within the NVSM Health

        IPlatform defines the interface needed for creating a platform object.
        This object will contain platform parameters, platform name and other
        metadata.

        On identifying the platform where nvsm health is running, platform
        object can be fetched using global list of platform_name.
    '''
    def __init__(self, name, config, columns=None, pretty=None):
        super(IPlatform, self).__init__()
        self.__name = name
        self.__config = config
        self.__columns = columns
        if not pretty:
            self.__pretty = [{}] * len(config)
        else:
            self.__pretty = pretty
        self.__parameters = {}
        self.__depends = []
        self.__conditions = []

        # Annotation members for generating documentation
        self.__title = None
        self.__description = None

    def getName(self):
        return self.__name

    def getConfig(self):
        return self.__config

    def getColumns(self):
        return self.__columns

    def getConfigName(self):
        return '_'.join([self.__name, '_'.join(filter(lambda x: x, self.__config))])

    def getPrettyConfig(self):
        return tuple(map(lambda i: self.__pretty[i].get(self.__config[i], self.__config[i]), range(len(self.__config))))

    def title(self, text):
        self.__title = text
        return self

    def getTitle(self):
        return self.__title

    def describe(self, text):
        self.__description = text
        return self

    def getDescription(self):
        return self.__description

    def setParameter(self, name, value):
        self.__parameters[name] = value
        return self

    def getParameters(self):
        return self.__parameters

    def getDependencies(self):
        return self.__depends

    def depends(self, task):
        assert isinstance(task, tasks.ITask)
        if task not in self.__depends:
            self.__depends.append(task)
        return self

    def qualify(self, condition_callback):
        self.__conditions.append(condition_callback)
        return self

    def matches(self):
        for condition_callback in self.__conditions:
            check = condition_callback()
            logging.debug(f'{inspect.getfile(condition_callback).split("/")[-1]} {self.getConfigName()} {condition_callback.__name__} {check}')
            if check == False:
                return False
        return True

