# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from nvsmhealth.modules import lspci, smartctl, fru, vgpu
from nvsmhealth.platform import IPlatform
import itertools, json

__all__ = [
    'generate_bladerunner_platform_objects'
]

platform_type = 'bladerunner'

def check_product_name(platform, config):
    platform.depends(fru.parse_fru)
    def check_product():
        try:
            return 'rtx' in fru.parse_fru.getResult()['Builtin FRU Device']['product_name'].lower()
        except:
            pass
        return False
    platform.qualify(check_product)

def define_gpu_parameters(platform, config):
    if 'vgpu' in config:
        gpumonitor_active = True
        gpu_config = True
    elif 'passthrough' in config:
        gpumonitor_active = False
        gpu_config = False

    platform.depends(vgpu.check_gpu_monitor_active)
    def gpu_monitor_active():
        return vgpu.check_gpu_monitor_active.getResult() == gpumonitor_active
    platform.qualify(gpu_monitor_active)

    if 'tu102' not in config:
        raise NotImplementedError
    gpu_name = 'tu102'
    device_id = '1e37'

    # Define bill-of-materials for GPUs in this configuration
    gpu_bom = {}
    gpu_bdfs = [ '01', '02' ]
    for bdf in gpu_bdfs:
        # gpu_bom follows "lspci" format
        gpu_bom['{}:00.0'.format(bdf)] = {
            'class': '3D controller [0302]',
            'device_id': device_id
        }
    if gpu_config:
        platform.setParameter('gpu_bom', gpu_bom)
    else:
        platform.setParameter('gpu_bom', None)

    # Define bill-of-materials for gpu link width and speed
    device_ids = [ '01', '02' ]
    gpu_link_device_ids = set(map(lambda x: '0000:{}:00.0'.format(x), device_ids))
    gpu_link_info = []
    for device in gpu_link_device_ids:
        gpu_link = {}
        gpu_link['bdf'] = device
        gpu_link['speed'] = {'P8':'2.5GT/s', 'P0':'8GT/s'}
        gpu_link['width'] = {'P8': 8, 'P0': 8}
        gpu_link_info.append(gpu_link)

    if gpu_config:
        platform.setParameter('gpu_link_info', gpu_link_info)
    else:
        platform.setParameter('gpu_link_info', None)


    # Define DCGM GPU blacklist recommendations watch type
    # On bladerunner platform, NVLINK is not supported, Add args to watch for 'p'=PCIE, 'm'=Memory, 'i'=Inforom, 't'=Thermal and Power
    platform.setParameter('dcgm_instant_watch_type', ['pmit', '--instantaneous'])
    # Enable only for longer tests
    #platform.setParameter('dcgm_watch_type', ['pmit'])

    # Enable GPU VBIOS version consistency check
    platform.setParameter('enable_vbios_version_check', gpu_config)

def define_vga_parameters(platform, config):
    device_id = '2000'

    # Define bill-of-materials for VGA Controller in this configuration
    vga_bom = {}
    vga_bdfs = [ '21', ]
    for bdf in vga_bdfs:
        # vga_bom follows "lspci" format
        vga_bom['{}:00.0'.format(bdf)] = {
            'class': 'VGA compatible controller [0300]',
            'device_id': device_id
        }
    platform.setParameter('vga_bom', vga_bom)

def define_cpu_parameters(platform, config):
    platform.setParameter('xenserver_number_of_cores', 16)

def define_dimm_parameters(platform, config):
    # Configuration has each DIMM of 16 GB, and total of 64 GB
    dimm_size = '16384 MB'

    # Define bill-of-materials for DIMMs in this configuration
    dimm_bom = {}
    dimm_slots = ['DIMM A0', 'DIMM B0']

    for slot in dimm_slots:
        # dimm_bom follows "dmidecode" format
        dimm_bom['Memory Device ({})'.format(slot)] = {
            'locator': slot,
            'size': dimm_size
        }
    platform.setParameter('dimm_bom', dimm_bom)

    # Define DIMM vendors
    platform.setParameter('dmidecode_dimm_vendors', ['Kingston', 'SK Hynix', 'Micron'])
    # Define DIMM part number for GFN performance
    platform.setParameter('dimm_part_number', ['NV26D4U16D8ME-16NX', 'HX426C16FB/16'])

def define_network_parameters(platform, config):
    # Define bill-of-materials for ethernet controllers
    ethernet_controllers_bom = {}
    eth_bdfs = [ '0', '1']
    for bdf in eth_bdfs:
        # ethernet_controllers_bom follows "lspci" format
        ethernet_controllers_bom['10:00.{}'.format(bdf)] = {
            'class': 'Ethernet controller [0200]',
            'device_id': '1581',
            'vendor': 'Intel Corporation [8086]'
        }

    platform.setParameter('ethernet_controllers_bom', ethernet_controllers_bom)

    # Define bill-of-materials ethernet controller link speeds and widths
    intel_eth_bdfs = [ ('10',), (0, 1)]
    intel_eth_bdfs = itertools.product(*intel_eth_bdfs)
    eth_bdfs = map(lambda x: '{}:00.{}'.format(x[0], x[1]), intel_eth_bdfs)
    name = "10G"
    speed = "8GT/s"
    width = 4
    ethernet_controller_info = []
    for device in eth_bdfs:
        ethernet_controller_id_info = {}
        ethernet_controller_id_info['bdf'] = '0000:{}'.format(device)
        ethernet_controller_id_info['name'] = name
        ethernet_controller_id_info['speed'] = speed
        ethernet_controller_id_info['width'] = width
        ethernet_controller_info.append(ethernet_controller_id_info)

    platform.setParameter('ethernet_controller_info', ethernet_controller_info)

def define_generic_parameters(platform, config):
    platform.setParameter('sw_rel_file', "os-release")
    platform.setParameter('summary_serial_num_string', "Serial Number")
    platform.setParameter('summary_base_os_string', "OS Version")

def define_nvme_parameters(platform, config):
    # Set config parameter for nvme
    nvme_config = [{"0.5TB": 1}, {"1.0TB": 1}]
    platform.setParameter('nvme', nvme_config)
    platform.setParameter('nvme_check_smart_log', True)

    # Define bill-of-materials for nvme link width and speed
    nvme_link_info = []
    nvme_link = {}
    nvme_link['bdf'] = '0000:30:00.0'
    nvme_link['speed'] = '8GT/s'
    nvme_link['width'] = 4
    nvme_link_info.append(nvme_link)
    platform.setParameter('nvme_link_info', nvme_link_info)

def __generate_bladerunner_platform_objects(configs, columns=None, pretty=None):
    # Generate an IPlatform object for each valid Bladerunner configuration
    platforms = []
    for config in configs:
        config = list(config)
        platform = IPlatform(name=platform_type, config=config, columns=columns, pretty=pretty)
        check_product_name(platform, config)

        # Define parameters for each platform
        define_gpu_parameters(platform, config)
        define_vga_parameters(platform, config)
        define_dimm_parameters(platform, config)
        define_network_parameters(platform, config)
        define_cpu_parameters(platform, config)
        define_nvme_parameters(platform, config)
        define_generic_parameters(platform, config)

        platforms.append(platform)

    return platforms

def generate_bladerunner_platform_objects():
    # Matrix (Cartesian product) of all possible Bladerunner configurations
    variants = [
        ('tu102',),  # GPUs
        ('vgpu','passthrough')  # VGPU or Passthrough mode
    ]

    configs = list(itertools.product(*variants))
    columns = ["GPU", "GPU Technology"]
    pretty = [
        {
            'tu102': 'TU102',
        },
        {
            'vgpu': 'vGPU',
            'passthrough': 'GPU-Passthrough',
        },
    ]
    return __generate_bladerunner_platform_objects(configs=configs, columns=columns, pretty=pretty)
