# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from nvsmhealth.modules import lspci, dmidecode, smartctl, sysfs
from nvsmhealth.platform import IPlatform
import itertools, json

__all__ = [
    'generate_cgx_platform_objects'
]

platform_type = 'cgx'

#~ Based on current 2.04 Version FW Release [3/28]
dcv_bmc_config = {
        'ip': '192.168.1.42',
        'user': 'nvsm-admin'
    }

def check_system_vendor(platform):
    platform.depends(sysfs.sysfs_dmi_system_vendor)
    def check_vendor():
        system_vendor = sysfs.sysfs_dmi_system_vendor.getOutput().strip()
        if system_vendor != 'NVIDIA':
            return False
        return True
    platform.qualify(check_vendor)

def check_product_name(platform):
    platform.depends(sysfs.sysfs_dmi_product_name)
    def check_product():
        product_name = sysfs.sysfs_dmi_product_name.getOutput()
        if 'drive constellation simulator' in product_name.lower():
            return True
        return False
    platform.qualify(check_product)

def define_gpu_parameters(platform, config):
    # Define GPUs config for DCS and DCC platform
    gpu_name = 't40'
    device_id = '1e38'
    sm_count = 72
    drive_gpu_name = 'p4000'
    drive_device_id = '1bb1'
    drive_sm_count = 14
    # Define bill-of-materials for GPUs in this configuration
    gpu_bom = {}
    gpu_bdfs = [ '1d', '23', '43', '49', 'b4', 'ba', 'e0', 'e6' ]
    drive_gpu_bdfs = [ '6d', '92' ]
    for bdf in gpu_bdfs:
        # gpu_bom follows "lspci" format
        gpu_bom['{}:00.0'.format(bdf)] = {
            'class': '3D controller [0302]',
            'device_id': device_id
        }
    for bdf in drive_gpu_bdfs:
        # gpu_bom follows "lspci" format
        gpu_bom['{}:00.0'.format(bdf)] = {
            'class': 'VGA compatible controller [0300]',
            'device_id': drive_device_id
        }
    platform.setParameter('gpu_bom', gpu_bom)
    # Define SM count for this configuration
    platform.setParameter('gpu_sm_count', sm_count)
    platform.setParameter('drive_gpu_sm_count', drive_sm_count)
    # Define GPU total retired page count threshold
    # Refer https://docs.nvidia.com/deploy/dynamic-page-retirement/index.html#faq for retired pages limit
    gpu_total_retired_page_count = 60
    platform.setParameter('gpu_total_retired_page_count', gpu_total_retired_page_count)

    # Define bill-of-materials for gpu link width and speed
    device_ids = gpu_bdfs + drive_gpu_bdfs
    gpu_link_device_ids = set(map(lambda x: '0000:{}:00.0'.format(x), device_ids))
    gpu_link_info = []
    for device in gpu_link_device_ids:
        gpu_link = {}
        gpu_link['bdf'] = device
        # pstates as defined -
        # https://docs.google.com/document/d/1u9iOpsyw2RlK7pGt0oCjllBB4n-QRJ9FigH_c09H7u4/edit#heading=h.99bmm29mseo5
        # TODO: should we check for user defined - PCIe throttling
        gpu_link['speed'] = {'P8':'2.5GT/s', 'P0':'8GT/s'}
        gpu_link['width'] = {'P8': 16, 'P0': 16}
        gpu_link_info.append(gpu_link)
    platform.setParameter('gpu_link_info', gpu_link_info)

    # Define DCGM GPU blacklist recommendations watch type
    # On bladerunner platform, NVLINK is not supported, Add args to watch for 'p'=PCIE, 'm'=Memory, 'i'=Inforom, 't'=Thermal and Power
    platform.setParameter('dcgm_instant_watch_type', ['pmit', '--instantaneous'])
    # Enable only for longer tests
    #platform.setParameter('dcgm_watch_type', ['pmit'])

def define_dimm_parameters(platform, config):
    # DCS has 12x32GB DIMMs config
    # Define bill-of-materials for DIMMs in this configuration
    dimm_slots = [('0', '1'), ('A', 'B', 'C', 'D', 'E', 'F'), (0, )]
    dimm_size = '32 GB'
    dimm_bom = {}
    dimm_slots = itertools.product(*dimm_slots)
    dimm_slots = map(lambda x: 'CPU{}_{}{}'.format(x[0], x[1], x[2]), dimm_slots)
    for slot in dimm_slots:
        # dimm_bom follows "dmidecode" format
        dimm_bom['Memory Device ({})'.format(slot)] = {
            'locator': slot,
            'size': dimm_size
        }
    platform.setParameter('dimm_bom', dimm_bom)
    # Define DIMM vendor
    platform.setParameter('dmidecode_dimm_vendors', ['Samsung'])

def define_network_parameters(platform, config):
    # Ethernet controllers bdfs
    mlnx_eth_bdfs = [ '5e']
    intel_1G_eth_bdfs = [ ('70', '95'), (0, 1, 2, 3)] # {}.00.{} format
    intel_10G_eth_bdfs = [ ('da', '3d'), (0, 1)] # {}.00.{} format
    # Define bill-of-materials for ethernet controllers
    ethernet_controllers_bom = {}
    # ethernet_controllers_bom follows "lspci" format
    ethernet_controllers_config = {
        'class': 'Ethernet controller [0200]',
        'device_id': '1015',
        'vendor': 'Mellanox Technologies [15b3]'
    }
    for bdf in mlnx_eth_bdfs:
        ethernet_controllers_bom['{}:00.0'.format(bdf)] = ethernet_controllers_config

    intel_ethernet_controllers_config = {
        'class': 'Ethernet controller [0200]',
        'device_id': '1521',
        'vendor': 'Intel Corporation [8086]'
    }
    intel_eth_bdfs = itertools.product(*intel_1G_eth_bdfs)
    eth_bdfs = map(lambda x: '{}:00.{}'.format(x[0], x[1]), intel_eth_bdfs)
    for bdf in eth_bdfs:
        ethernet_controllers_bom['{}'.format(bdf)] = intel_ethernet_controllers_config

    intel_ethernet_controllers_config = {
        'class': 'Ethernet controller [0200]',
        'device_id': '1563',
        'vendor': 'Intel Corporation [8086]'
    }
    intel_eth_bdfs = itertools.product(*intel_10G_eth_bdfs)
    eth_bdfs = map(lambda x: '{}:00.{}'.format(x[0], x[1]), intel_eth_bdfs)
    for bdf in eth_bdfs:
        ethernet_controllers_bom['{}'.format(bdf)] = intel_ethernet_controllers_config

    platform.setParameter('ethernet_controllers_bom', ethernet_controllers_bom)

    # Define bill-of-materials ethernet controller link speeds and widths
    device_ids = [('5e', '10G', '8GT/s', 8),]
    ethernet_controller_info = []
    for device_id, name, speed, width in device_ids:
        ethernet_controller_id_info = {}
        ethernet_controller_id_info['bdf'] = '0000:{}:00.0'.format(device_id)
        ethernet_controller_id_info['name'] = name
        ethernet_controller_id_info['speed'] = speed
        ethernet_controller_id_info['width'] = width
        ethernet_controller_info.append(ethernet_controller_id_info)

    intel_eth_bdfs = itertools.product(*intel_1G_eth_bdfs)
    eth_bdfs = map(lambda x: '{}:00.{}'.format(x[0], x[1]), intel_eth_bdfs)
    name = "1G"
    speed = "5GT/s"
    width = 4
    for device in eth_bdfs:
        ethernet_controller_id_info = {}
        ethernet_controller_id_info['bdf'] = '0000:{}'.format(device)
        ethernet_controller_id_info['name'] = name
        ethernet_controller_id_info['speed'] = speed
        ethernet_controller_id_info['width'] = width
        ethernet_controller_info.append(ethernet_controller_id_info)

    intel_eth_bdfs = itertools.product(*intel_10G_eth_bdfs)
    eth_bdfs = map(lambda x: '{}:00.{}'.format(x[0], x[1]), intel_eth_bdfs)
    name = "10G"
    speed = "8GT/s"
    width = 4
    for device in eth_bdfs:
        ethernet_controller_id_info = {}
        ethernet_controller_id_info['bdf'] = '0000:{}'.format(device)
        ethernet_controller_id_info['name'] = name
        ethernet_controller_id_info['speed'] = speed
        ethernet_controller_id_info['width'] = width
        ethernet_controller_info.append(ethernet_controller_id_info)
    platform.setParameter('ethernet_controller_info', ethernet_controller_info)

    # Define bill-of-materials for pcie switches
    pcie_switches_bom = {}
    pcie_bdfs = ['00:1b.0', '00:1c.0', '00:1d.0', '00:1d.4', '17:00.0', '3a:00.0',
            '5d:00.0', '5d:02.0', '85:00.0', '85:02.0', '85:03.0', 'ae:00.0',
            'd7:00.0']
    for bdf in pcie_bdfs:
        # pcie_switches_bom follows "lspci" format
        pcie_switches_bom['{}'.format(bdf)] = {
            'class': '0604',
            'vendor': '8086'
        }
    pcie_bdfs = ['04:00.0',]
    for bdf in pcie_bdfs:
        # pcie_switches_bom follows "lspci" format
        pcie_switches_bom['{}'.format(bdf)] = {
            'class': '0604',
            'vendor': '1a03'
        }
    pcie_bdfs = ['18:00.0', '19:04.0', '19:08.0', '19:0c.0', '19:10.0', '19:14.0',
                 '3b:00.0', '3c:04.0', '3c:08.0', '3c:0c.0', '3c:10.0', '3c:14.0',
                 '61:00.0', '62:04.0', '63:00.0', '64:00.0',
                 '65:00.0', '66:04.0', '66:05.0', '66:08.0', '66:0c.0', '66:0d.0', '66:10.0',
                 '86:00.0', '87:04.0', '88:00.0', '89:00.0', '8a:00.0', '8b:04.0',
                 '8b:05.0', '8b:08.0', '8b:0c.0', '8b:0d.0', '8b:10.0',
                 'af:00.0', 'b0:04.0', 'b0:08.0', 'b0:0c.0', 'b0:10.0', 'b0:14.0',
                 'd8:00.0', 'd9:04.0', 'd9:08.0', 'd9:0c.0', 'd9:10.0', 'd9:14.0']
    for bdf in pcie_bdfs:
        # pcie_switches_bom follows "lspci" format
        pcie_switches_bom['{}'.format(bdf)] = {
            'class': '0604',
            'vendor': '10b5'
        }
    platform.setParameter('pcie_switches_bom', pcie_switches_bom)

def define_generic_parameters(platform, config):
    platform.setParameter('sw_rel_file', "dcs-release")
    platform.setParameter('summary_serial_num_string', "DCS Serial Number")
    platform.setParameter('summary_base_os_string', "DCS BaseOS Version")
    platform.setParameter('chassis_serial_number_index', '0')

def define_fan_parameters(platform, config):
    fan_bom = {
                    "Fan_SYS0_0": True,
                    "Fan_SYS0_1": True,
                    "Fan_SYS1_0": True,
                    "Fan_SYS1_1": True,
                    "Fan_SYS2_0": True,
                    "Fan_SYS2_1": True,
                    "Fan_SYS3_0": True,
                    "Fan_SYS3_1": True,
                    "Fan_PSU0": True,
                    "Fan_PSU1": True,
                    "Fan_PSU2": True,
                    "Fan_PSU3": True,
                    "Fan_SYS0_Status": True,
                    "Fan_SYS1_Status": True,
                    "Fan_SYS2_Status": True,
                    "Fan_SYS3_Status": True,
                  }

    platform.setParameter('fan_bom', fan_bom)

def define_psu_parameters(platform, config):
    psu_bom = {
            "Temp_PSU0": True,
            "PSU0_Status": True,
            "PSU0_Input": True,
            "Temp_PSU1": True,
            "PSU1_Status": True,
            "PSU1_Input": True,
            "Temp_PSU2": True,
            "PSU2_Status": True,
            "PSU2_Input": True,
            "Temp_PSU3": True,
            "PSU3_Status": True,
            "PSU3_Input": True,
            "PSU Redundancy": True,
            "T_PSU_Highest": True,
                  }

    platform.setParameter('psu_bom', psu_bom)

def define_cpu_parameters(platform, config):
    lscpu_number_of_cores = 80

    platform.setParameter('lscpu_number_of_cores', lscpu_number_of_cores)

def define_nvme_parameters(platform, config):
    # List of supported nvme drives configurations
    nvme_config = [{"1.0TB": 2, "3.8TB": 1}, {"1.0TB": 2, "3.8TB": 2}]
    # Set config parameter for nvme
    platform.setParameter('nvme', nvme_config)

def define_dcv_parameters(platform, config):
    platform.setParameter('enable_dcv_checks', True)

    # Define USB Parameters
    num_usb_devices = 19
    usb_list = [{'class':'USB Device','product_id':'','product_string':'',
                    'speed':'', 'vendor_id':'','vendor_string':'',
                    'version':''} for k in range(num_usb_devices)]
    port_list = ["1-2.5.1", "1-2.5.2", "1-2.5.3", "1-2.5.4", "1-2.6.1", "1-2.6.2",
                 "1-2.6.3", "1-2.6.4", "1-2.1", "1-2.2.4", "1-2.2.5", "1-2.2.6",
                 "1-2.2.7", "1-2", "1-2.2", "1-2.3", "1-2.4", "1-3", "1-4"]

    product_id_list = ["6015" for x in range(8)] + ["6011" for x in range(5)] + \
                    ["2807","2807","9204","9204","6570","6570"]

    product_string_list = ["FT200X USB I2C", "FT200X USB I2C", "FT200X USB I2C",
                           "FT200X USB I2C", "FT200X USB I2C", "FT200X USB I2C",
                           "FT200X USB I2C", "FT200X USB I2C"] +  \
                          ["Quad RS232-HS" for x in range(5)] +  \
                          ["USB2807 Hub", "USB2807 Hub", "SparkFun Pro M",
                          "SparkFun Pro M", "USB2.0 Hub", "USB2.0 Hub"]

    speed_list = ["12" for x in range(8)] + ["480" for x in range(5)] + \
                    ["480", "480", "12", "12", "480", "480"]

    vendor_id_list = ["0403" for x in range(13)] +  ["0424", "0424", "1b4f", "1b4f", "04b4", "04b4"]
    vendor_string_list = ["FTDI" for x in range(13)] + ["Microchip", "Microchip", "SparkFun", "SparkFun"]
    version_list = [" 2.10", " 2.00", " 2.10"] + [" 2.00" for i in range (16)]
    version_list = [" 2.00" for x in range(13)] + [" 2.10", " 2.10"] + [" 2.00" for x in range(4)]

    # Make USB Tree
    dcs_usb_tree = {}

    for x in range(num_usb_devices):
        usb_list[x]['product_id'] = product_id_list[x]
        usb_list[x]['product_string'] = product_string_list[x]
        usb_list[x]['speed'] = speed_list[x]
        usb_list[x]['vendor_id'] = vendor_id_list[x]
        usb_list[x]['version'] = version_list[x]
        if (x > 16): # Last 2 Hub devices don't have vendor_string
            del usb_list[x]['vendor_string']
        else:
            usb_list[x]['vendor_string'] = vendor_string_list[x]
        dcs_usb_tree [port_list[x]] = usb_list[x]

    platform.setParameter('dcs_usb_tree', dcs_usb_tree)

    net_link = {}
    net_interfaces = ['DCC_BMC', 'HCL_Radar', 'HCR_Radar', \
        'HLC_Radar', 'HXL_Radar', 'HXR_Radar', 'TCL_Radar', \
        'TCR_Radar', 'TEGRA_A', 'TEGRA_B', 'TLC_Radar']

    for interface in net_interfaces:
        net_link[interface] = {'link': True}

    platform.setParameter('net_link', net_link)

    platform.setParameter('dcs_grid_license', True)

    platform.setParameter('dcs_dcv_bmc', True)

    platform.setParameter('dcs_dcv_sensor_threshold', True)

    dcv_psu_bom = {
        "PSU0_Status": True,
        "PSU0_POUT": True,
        "PSU0_POUT_MAX": True,
        "Temp_PSU0": True,
        "PSU1_Status": True,
        "PSU1_POUT": True,
        "PSU1_POUT_MAX": True,
        "Temp_PSU1": True
    }
    platform.setParameter('dcv_psu_bom', dcv_psu_bom)

    dcv_fan_bom = {
        "FAN1": True,
        "FAN2": True,
        "FAN3": True,
        "FAN4": True
    }
    platform.setParameter('dcv_fan_bom', dcv_fan_bom)

    platform.setParameter('dcs_camera', True)

    # nvidia-settings tool configuration
    dcs_nv_settings_cfg = {
       "Quadro P4000": {
          "framelock": 2
       }
    }
    platform.setParameter('dcs_nv_settings_cfg', dcs_nv_settings_cfg)

    # Expected CAN  interface configuration
    dcs_can_cfg = {
        "can0": {
            "valid": True,
        }
    }
    platform.setParameter('dcs_can_cfg', dcs_can_cfg)

    # Valid values for PSU attributes
    dcs_psu_attrib_values = {
        'Vendor': ['Liteon Power', 'DELTA'],
        'Model': ['PS-2202-3L', 'DPS-2000AB-2 D']
    }
    platform.setParameter('dcs_psu_attrib_values', dcs_psu_attrib_values)

    # PSU Info commands
    dcs_psu_info_commands = {
        # PSU Vendor, Model and Serial Number commands are currently common across supported vendors.
        # If this changes in the future, the command strings should be converted to directories of
        # command strings as is the case with PSU FW Version Info commands

        # PSU Vendor information commands
        'psu0_vendor': '0x30 0x0a 0x00 0x01',
        'psu1_vendor': '0x30 0x0a 0x01 0x01',

        # PSU Model information commands
        'psu0_model': '0x30 0x0a 0x00 0x02',
        'psu1_model': '0x30 0x0a 0x01 0x02',

        # PSU Serial Number information commands
        'psu0_serial_number': '0x30 0x0a 0x00 0x00',
        'psu1_serial_number': '0x30 0x0a 0x01 0x00',

        # PSU FW Version information commands
        'psu0_fw_version': {
            #'Liteon Power':'0x06 0x52 0x09 0xb0 0x08 0xd5',
            'Liteon Power':'0x30 0x0f 0x00',
            'DELTA':'0x30 0x0f 0x00'
        },
        'psu1_fw_version': {
            #'Liteon Power':'0x06 0x52 0x09 0xb2 0x08 0xd5',
            'Liteon Power':'0x30 0x0f 0x01',
            'DELTA':'0x30 0x0f 0x01'
        }
    }
    platform.setParameter('dcs_psu_info_commands', dcs_psu_info_commands)

    pegasus_storage_config_list = {
        'Tegra_A':{
            '/dev/vblkdev0p1':{'threshold':95.0},
        },
        'Tegra_B': {
            '/dev/vblkdev0p1':{'threshold':95.0},
        },
    }
    platform.setParameter('pegasus_storage_config_list', pegasus_storage_config_list)

def __generate_cgx_platform_objects(configs, columns=None, pretty=None):
    # Generate an IPlatform object for each valid CGX configuration
    platforms = []
    for config in configs:
        config = list(config)
        platform = IPlatform(name=platform_type, config=config, columns=columns, pretty=pretty)
        check_system_vendor(platform)
        check_product_name(platform)
        # Define parameters for each platform
        define_gpu_parameters(platform, config)
        define_dimm_parameters(platform, config)
        define_network_parameters(platform, config)
        define_fan_parameters(platform, config)
        define_psu_parameters(platform, config)
        define_cpu_parameters(platform, config)
        define_nvme_parameters(platform, config)
        define_generic_parameters(platform, config)
        define_dcv_parameters(platform, config)

        platforms.append(platform)

    return platforms

def generate_cgx_platform_objects():
    # Matrix (Cartesian product) of all possible CGX configurations
    variants = [
        ('t40-p4000',),  # GPUs
    ]

    configs = list(itertools.product(*variants))
    columns = ["GPU"]
    pretty = [
        {
            't40-p4000': 'Tesla T40 with Quadro P4000',
        },
    ]
    return __generate_cgx_platform_objects(configs=configs, columns=columns, pretty=pretty)
