# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

""" Support for DGX-1 systems in any configuration.
"""

from nvsmhealth.modules import lspci, dmidecode, smartctl, sysfs
from nvsmhealth.platform import IPlatform
import itertools, json

__all__ = [
    'generate_dgx1_platform_objects'
]

platform_type = 'dgx-1'

def check_system_vendor(platform, config):
    platform.depends(sysfs.sysfs_dmi_system_vendor)
    def check_vendor():
        system_vendor = sysfs.sysfs_dmi_system_vendor.getOutput().strip()
        if system_vendor != 'NVIDIA':
            return False
        return True
    platform.qualify(check_vendor)

def check_product_name(platform, config):
    platform.depends(sysfs.sysfs_dmi_product_name)
    def check_product():
        product_name = sysfs.sysfs_dmi_product_name.getOutput().lower()
        if 'dgx-1' in product_name:
            return True
        # TODO: Check specific products, like NGN, MaxQ, and 32G
        return False
    platform.qualify(check_product)

def get_p100_gpu_direct_topology():
    return {
        'GPU0': {
            'GPU0': 'X',
            'GPU1': 'NV1',
            'GPU2': 'NV1',
            'GPU3': 'NV1',
            'GPU4': 'NV1',
            'GPU5': 'SYS',
            'GPU6': 'SYS',
            'GPU7': 'SYS'
        },
        'GPU1': {
            'GPU0': 'NV1',
            'GPU1': 'X',
            'GPU2': 'NV1',
            'GPU3': 'NV1',
            'GPU4': 'SYS',
            'GPU5': 'NV1',
            'GPU6': 'SYS',
            'GPU7': 'SYS'
        },
        'GPU2': {
            'GPU0': 'NV1',
            'GPU1': 'NV1',
            'GPU2': 'X',
            'GPU3': 'NV1',
            'GPU4': 'SYS',
            'GPU5': 'SYS',
            'GPU6': 'NV1',
            'GPU7': 'SYS'
        },
        'GPU3': {
            'GPU0': 'NV1',
            'GPU1': 'NV1',
            'GPU2': 'NV1',
            'GPU3': 'X',
            'GPU4': 'SYS',
            'GPU5': 'SYS',
            'GPU6': 'SYS',
            'GPU7': 'NV1'
        },
        'GPU4': {
            'GPU0': 'NV1',
            'GPU1': 'SYS',
            'GPU2': 'SYS',
            'GPU3': 'SYS',
            'GPU4': 'X',
            'GPU5': 'NV1',
            'GPU6': 'NV1',
            'GPU7': 'NV1'
        },
        'GPU5': {
            'GPU0': 'SYS',
            'GPU1': 'NV1',
            'GPU2': 'SYS',
            'GPU3': 'SYS',
            'GPU4': 'NV1',
            'GPU5': 'X',
            'GPU6': 'NV1',
            'GPU7': 'NV1'
        },
        'GPU6': {
            'GPU0': 'SYS',
            'GPU1': 'SYS',
            'GPU2': 'NV1',
            'GPU3': 'SYS',
            'GPU4': 'NV1',
            'GPU5': 'NV1',
            'GPU6': 'X',
            'GPU7': 'NV1'
        },
        'GPU7': {
            'GPU0': 'SYS',
            'GPU1': 'SYS',
            'GPU2': 'SYS',
            'GPU3': 'NV1',
            'GPU4': 'NV1',
            'GPU5': 'NV1',
            'GPU6': 'NV1',
            'GPU7': 'X'
        }
    }

def get_v100_gpu_direct_topology():
    return {
        'GPU0': {
            'GPU0': 'X',
            'GPU1': 'NV1',
            'GPU2': 'NV1',
            'GPU3': 'NV2',
            'GPU4': 'NV2',
            'GPU5': 'SYS',
            'GPU6': 'SYS',
            'GPU7': 'SYS'
        },
        'GPU1': {
            'GPU0': 'NV1',
            'GPU1': 'X',
            'GPU2': 'NV2',
            'GPU3': 'NV1',
            'GPU4': 'SYS',
            'GPU5': 'NV2',
            'GPU6': 'SYS',
            'GPU7': 'SYS'
        },
        'GPU2': {
            'GPU0': 'NV1',
            'GPU1': 'NV2',
            'GPU2': 'X',
            'GPU3': 'NV2',
            'GPU4': 'SYS',
            'GPU5': 'SYS',
            'GPU6': 'NV1',
            'GPU7': 'SYS'
        },
        'GPU3': {
            'GPU0': 'NV2',
            'GPU1': 'NV1',
            'GPU2': 'NV2',
            'GPU3': 'X',
            'GPU4': 'SYS',
            'GPU5': 'SYS',
            'GPU6': 'SYS',
            'GPU7': 'NV1'
        },
        'GPU4': {
            'GPU0': 'NV2',
            'GPU1': 'SYS',
            'GPU2': 'SYS',
            'GPU3': 'SYS',
            'GPU4': 'X',
            'GPU5': 'NV1',
            'GPU6': 'NV1',
            'GPU7': 'NV2'
        },
        'GPU5': {
            'GPU0': 'SYS',
            'GPU1': 'NV2',
            'GPU2': 'SYS',
            'GPU3': 'SYS',
            'GPU4': 'NV1',
            'GPU5': 'X',
            'GPU6': 'NV2',
            'GPU7': 'NV1'
        },
        'GPU6': {
            'GPU0': 'SYS',
            'GPU1': 'SYS',
            'GPU2': 'NV1',
            'GPU3': 'SYS',
            'GPU4': 'NV1',
            'GPU5': 'NV2',
            'GPU6': 'X',
            'GPU7': 'NV2'
        },
        'GPU7': {
            'GPU0': 'SYS',
            'GPU1': 'SYS',
            'GPU2': 'SYS',
            'GPU3': 'NV1',
            'GPU4': 'NV2',
            'GPU5': 'NV1',
            'GPU6': 'NV2',
            'GPU7': 'X'
        }
    }

def check_lspci_nvidia_gpu(parse_lspci_task):
    pci_devices = json.loads(parse_lspci_task.getResult())
    nvidia_devices = list(filter(lambda d: d.get('vendor_id') == '10de', pci_devices.values()))
    known_gpus = [
        ('v100-32', ['1db5', '1db7']),
        ('v100', ['1db1', '1db2']),
        ('p100', ['15f9'])
    ]
    # Guess the GPU configuration based on what is installed on the system
    for name, device_ids in known_gpus:
        if any(map(lambda d: d.get('device_id') in device_ids, nvidia_devices)):
            return name
    return 'unknown'

def define_gpu_parameters(platform, config):
    # TODO: Add SSID variations for NGN SKUs?
    if 'p100' in config:
        gpu_name = 'p100'
        device_id = '15f9'
        sm_count = 56
        gpu_nvlink_speed = '20 GB/s'
        gpu_direct_topology = get_p100_gpu_direct_topology()
    elif 'v100' in config:
        gpu_name = 'v100'
        device_id = '1db1'
        sm_count = 80
        gpu_nvlink_speed = '25.781 GB/s'
        gpu_direct_topology = get_v100_gpu_direct_topology()
    elif 'v100-32' in config:
        gpu_name = 'v100-32'
        device_id = '1db5'
        sm_count = 80
        gpu_nvlink_speed = '25.781 GB/s'
        gpu_direct_topology = get_v100_gpu_direct_topology()
    # Check lspci output for GPU configuration
    parse_lspci_task = lspci.parse_lspci
    platform.depends(parse_lspci_task)
    def check_gpu():
        return check_lspci_nvidia_gpu(parse_lspci_task) == gpu_name
    platform.qualify(check_gpu)
    # Define bill-of-materials for GPUs in this configuration
    gpu_bom = {}
    gpu_bdfs = [ '06', '07', '0a', '0b', '85', '86', '89', '8a' ]
    for bdf in gpu_bdfs:
        # gpu_bom follows "lspci" format
        gpu_bom['{}:00.0'.format(bdf)] = {
            'class': '3D controller [0302]',
            'device_id': device_id
        }
    platform.setParameter('gpu_bom', gpu_bom)
    # Define SM count for this configuration
    platform.setParameter('gpu_sm_count', sm_count)

    # Define bill-of-materials for gpu link width and speed
    gpu_link_device_ids = set(map(lambda x: '0000:{}:00.0'.format(x), gpu_bdfs))
    gpu_link_info = []
    for device in gpu_link_device_ids:
        gpu_link = {}
        gpu_link['bdf'] = device
        gpu_link['speed'] = {'P0':'8GT/s'}
        gpu_link['width'] = {'P0': 16}
        gpu_link_info.append(gpu_link)

    platform.setParameter('gpu_link_info', gpu_link_info)
    # Define DCGM GPU blacklist recommendations watch type
    # 'a' represents 'All watches' including PCIE, Memory, Inforom, NVLINK, Thermal and Power
    platform.setParameter('dcgm_instant_watch_type', ['a', '--instantaneous'])
    # Enable only for longer tests
    #platform.setParameter('dcgm_watch_type', ['a'])

    # Define GPU NVLink speed
    platform.setParameter('gpu_nvlink_speed', gpu_nvlink_speed)

    # Define GPUDirect topology
    platform.setParameter('gpu_direct_topology', gpu_direct_topology)

    # Define GPU total retired page count threshold
    # Refer https://docs.nvidia.com/deploy/dynamic-page-retirement/index.html#faq for retired pages limit
    gpu_total_retired_page_count = 60
    platform.setParameter('gpu_total_retired_page_count', gpu_total_retired_page_count)

    # Enable GPU VBIOS version consistency check
    platform.setParameter('enable_vbios_version_check', True)


def define_dimm_parameters(platform, config):
    # Check DMI table for DIMM configuration
    parse_dimms_task = dmidecode.parse_memory_dimms
    platform.depends(parse_dimms_task)
    # Configuration has 1TB RAM total (each DIMM is 64 GB)
    def check_1tb_ram():
        # If there is at least one '64 GB' DIMM installed, then assume this
        # system must be a 1TB configuration
        observed_dimms = json.loads(parse_dimms_task.getResult())
        if '64 GB' in observed_dimms:
            return True
        return False
    def check_not_1tb_ram():
        return not check_1tb_ram()
    if '1tb' in config:
        platform.qualify(check_1tb_ram)
        dimm_size = '64 GB'
        memory_size = 1056556760
    else:
        # Configuration has 512GB RAM total (each DIMM is 32 GB)
        dimm_size = '32 GB'
        platform.qualify(check_not_1tb_ram)
        memory_size = 528278380
    # Define bill-of-materials for DIMMs in this configuration
    dimm_bom = {}
    dimm_slots = [('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H'), (0, 1)]
    dimm_slots = itertools.product(*dimm_slots)
    dimm_slots = map(lambda x: 'DIMM_{}{}'.format(x[0], x[1]), dimm_slots)
    for slot in dimm_slots:
        # dimm_bom follows "dmidecode" format
        dimm_bom['Memory Device ({})'.format(slot)] = {
            'locator': slot,
            'size': dimm_size
        }
    platform.setParameter('dimm_bom', dimm_bom)
    # Define DIMM vendors
    dimm_vendors = set(['Samsung', 'Micron', 'Hynix Semiconductor'])
    platform.setParameter('dmidecode_dimm_vendors', dimm_vendors)

    platform.setParameter('meminfo_memory_size', memory_size)

def check_lspci_mlnx(parse_lspci_task):
    known_mlnx_bdfs = ['05:00.0', '0c:00.0', '84:00.0', '8b:00.0']

    pci_devices = json.loads(parse_lspci_task.getResult())
    mlnx_cards = { k: pci_devices.get(k, {}) for k in known_mlnx_bdfs }
    m_list = list(map(lambda v: v.get('device_id') == '1017', mlnx_cards.values()))
    if any(m_list):
        if sum(m_list) == 2:
            return '2x_cx5'
        return 'cx5'
    if any(map(lambda v: v.get('device_id') == '1013', mlnx_cards.values())):
        return 'cx4'
    m_list = list(map(lambda v: v.get('device_id') == '101b', mlnx_cards.values()))
    if any(m_list):
        if sum(m_list) == 2:
            return '2x_cx6'
        return 'cx6'
    return ''

def define_mlnx_parameters(platform, config):
    platform.setParameter('mlnx_firmware_check', 1)

def check_lspci_mode(parse_lspci_task):
    known_mlnx_bdfs = ['05:00.0', '0c:00.0', '84:00.0', '8b:00.0']

    pci_devices = json.loads(parse_lspci_task.getResult())
    mlnx_cards = { k: pci_devices.get(k, {}) for k in known_mlnx_bdfs }
    if any(map(lambda v: v.get('class_id') == '0200', mlnx_cards.values())):
        return 'eth'
    return 'ib'

def define_network_parameters(platform, config):
    if 'ib' in config:
        mode = 'ib'
    elif 'eth' in config:
        mode = 'eth'
    # Check lspci output for GPU configuration
    parse_lspci_task = lspci.parse_lspci
    platform.depends(parse_lspci_task)
    def check_mode():
        return check_lspci_mode(parse_lspci_task) == mode
    platform.qualify(check_mode)

    if 'cx4' in config:
        mlnx_card = 'cx4'
    elif 'cx5' in config:
        mlnx_card = 'cx5'
    elif '2x_cx5' in config:
        mlnx_card = '2x_cx5'
    elif 'cx6' in config:
        mlnx_card = 'cx6'
    elif '2x_cx6' in config:
        mlnx_card = '2x_cx6'
    # Check lspci output for GPU configuration
    parse_lspci_task = lspci.parse_lspci
    platform.depends(parse_lspci_task)
    def check_mlnx():
        return check_lspci_mlnx(parse_lspci_task) == mlnx_card
    platform.qualify(check_mlnx)

    # Define bill-of-materials for ethernet controllers
    platform.depends(lspci.parse_lspci)

    def check_qct_mezzanine_fiber():
        pci_devices = json.loads(lspci.parse_lspci.getResult())
        return pci_devices.get('01:00.0', {}).get('device_id') == '10fb'

    def check_ocp_mezzanine_card():
        pci_devices = json.loads(lspci.parse_lspci.getResult())
        return pci_devices.get('01:00.0', {}).get('device_id') == '1007'

    if 'ocp_mezzanine_card' in config:
        # Optional OCP mezzanine fiber card
        platform.qualify(lambda: check_ocp_mezzanine_card())
        device_ids = ['0']
        class_str = 'Network controller [0280]'
        device_id = '1007'
        vendor = 'Mellanox Technologies [15b3]'
        speed = '8GT/s'
        width = 8
        name = '10G'
    else:
        if 'qct_mezzanine_fiber' in config:
            # Optional QCT mezzanine fiber card
            platform.qualify(lambda: check_qct_mezzanine_fiber())
            device_id = '10fb'
        else:
            platform.qualify(lambda: not check_qct_mezzanine_fiber()
                                 and not check_ocp_mezzanine_card())
            device_id = '1528'
        # Common params
        device_ids = ['0', '1']
        class_str = 'Ethernet controller [0200]'
        vendor = 'Intel Corporation [8086]'
        speed = '5GT/s'
        width = 8
        name = '1G'

    ethernet_controllers_bom = {}
    for bdf in device_ids:
        # ethernet_controllers_bom follows "lspci" format
        ethernet_controllers_bom['01:00.{}'.format(bdf)] = {
            'class': class_str,
            'device_id': device_id,
            'vendor': vendor
        }

    # Define bill-of-materials ethernet controller link speed 5gts and width x8
    ethernet_controller_device_ids = set(map(lambda x: '0000:01:00.{}'.format(x), device_ids))
    ethernet_controller_info = []
    for device_id in ethernet_controller_device_ids:
        ethernet_controller_id_info = {}
        ethernet_controller_id_info['bdf'] = device_id
        # ethernet_controller_id_info['name'] = name
        ethernet_controller_id_info['speed'] = speed
        ethernet_controller_id_info['width'] = width
        ethernet_controller_info.append(ethernet_controller_id_info)

    # Define bill-of-materials for ethernet controller link speed 8gts and width x16
    device_ids = ['05', '0c', '84', '8b']
    vendor_device_id = ''
    if mlnx_card in ['2x_cx5', 'cx5']:
        device_ids = ['05', '84']
        vendor_device_id = '1017'
    if mlnx_card == 'cx4':
        vendor_device_id = '1013'
    if mlnx_card in ['2x_cx6', 'cx6']:
        vendor_device_id = '101b'
    if mode == 'eth':
        ethernet_controller_device_ids = set(map(lambda x: '0000:{}:00.0'.format(x), device_ids))
        for device_id in ethernet_controller_device_ids:
            ethernet_controller_id_info = {}
            ethernet_controller_id_info['bdf'] = device_id
            # ethernet_controller_id_info['name'] = '10G'
            ethernet_controller_id_info['speed'] = '8GT/s'
            ethernet_controller_id_info['width'] = 16
            ethernet_controller_info.append(ethernet_controller_id_info)

        # Add bill-of-materials for ethernet controllers on ethernet mode
        for bdf in device_ids:
            # ethernet_controllers_bom follows "lspci" format
            ethernet_controllers_bom['{}:00.0'.format(bdf)] = {
                'class': 'Ethernet controller [0200]',
                'device_id': vendor_device_id,
                'vendor': 'Mellanox Technologies [15b3]'
        }

    platform.setParameter('ethernet_controllers_bom', ethernet_controllers_bom)
    platform.setParameter('ethernet_controller_info', ethernet_controller_info)

    if mode == 'ib':
        # Add bill-of-materials for IB controllers
        ib_controllers_bom = {}
        for bdf in device_ids:
            # ib_controllers_bom follows "lspci" format
            ib_controllers_bom['{}:00.0'.format(bdf)] = {
                'class': 'Infiniband controller [0207]',
                'device_id': vendor_device_id,
                'vendor': 'Mellanox Technologies [15b3]'
            }
        platform.setParameter('ib_controllers_bom', ib_controllers_bom)

        # Define bill-of-materials for ib controller link speed and width
        ib_controller_link_info = []
        for bdf in device_ids:
            ib_controller_id_info = {}
            ib_controller_id_info['bdf'] = '0000:{}:00.0'.format(bdf)
            ib_controller_id_info['speed'] = '8GT/s'
            ib_controller_id_info['width'] = 16
            ib_controller_link_info.append(ib_controller_id_info)

        platform.setParameter('ib_controller_link_info', ib_controller_link_info)

    # Define bill-of-materials for pcie switches
    pcie_switches_bom = {}
    pcie_bdfs = [ '03', '08', '82', '87']
    for bdf in pcie_bdfs:
        device_id = '8764'
        if bdf in ['08', '87']:
            device_id = '8780'
        # pcie_switches_bom follows "lspci" format
        # device is taken directly from "lspci" output, without appenidng "_id"
        pcie_switches_bom['{}:00.0'.format(bdf)] = {
            'class': '0604',
            'device': device_id,
            'vendor': '10b5'
        }
    platform.setParameter('pcie_switches_bom', pcie_switches_bom)

def check_smartctl_scan(parse_smartctl_task):
    disks = parse_smartctl_task.getResult()
    if disks and disks.get('480 GB') == 2:
        return 'raid1'
    return None

def define_raid_parameters(platform, config):
    raid = None
    if 'raid1' in config:
        raid = 'raid1'
        drives = 2
    else:
        drives = 1

    # Check smart scan output for raid1 configuration
    parse_smartctl_task = smartctl.parse_smartctl_device_info
    platform.depends(parse_smartctl_task)
    def check_raid1():
        return check_smartctl_scan(parse_smartctl_task) == raid
    platform.qualify(check_raid1)
    # Define disk controllers bom
    disk_controllers_bom = {}
    disk_controllers_bom['81:00.0'] = {
        'class': 'RAID bus controller [0104]',
        'device_id': '005d',
        'vendor_id': '1000'
    }
    platform.setParameter('disk_controllers_bom', disk_controllers_bom)

    # Define smartctl megaraid disk count
    smartctl_megaraid_disk_count = {
        '1.92 TB': 4,
        '480 GB': drives
    }
    platform.setParameter('smartctl_megaraid_disk_count', smartctl_megaraid_disk_count)
    platform.setParameter('smartctl_check_ssd_brick', True)

def define_stor_cli_parameters(platform, config):
    platform.setParameter('storcli_platform_string', "DGX")
    # p0..p7
    platform.setParameter('storcli_phy_links', range(0, 8))
    # disk status   - name          key             id, expected_val
    disk_stats = [("/c0/v0",        "/c0/v0",       0,  "Optl"),
                  ("/c0/v1",        "/c0/v1",       0,  "Optl"),
                  ("/c0/v0 PD 0",   "PDs for VD 0", 0,  "Onln"),
                  ("/c0/v1 PD 0",   "PDs for VD 1", 0,  "Onln"),
                  ("/c0/v1 PD 1",   "PDs for VD 1", 1,  "Onln"),
                  ("/c0/v1 PD 2",   "PDs for VD 1", 2,  "Onln"),
                  ("/c0/v1 PD 3",   "PDs for VD 1", 3,  "Onln")]
    platform.setParameter("storcli_disk_stats", disk_stats)

def define_generic_parameters(platform, config):
    # Define bill-of-materials for FAN
    fan_bom = {}
    fans = [(0, 1, 2, 3), # Fan_SYS0_1, Fan_SYS1_1
            (1, 2)]       # Fan_SYS0_1, Fan_SYS0_2
    fans = itertools.product(*fans)
    for x in fans:
        fan_bom[f'Fan_SYS{x[0]}_{x[1]}'] = True
    platform.setParameter('fan_bom', fan_bom)

    # Define bill-of-materials for PSU
    psu_bom = {}
    for i in range(1, 5):
        psu_bom['PSU{} Status'.format(i)] = True
    platform.setParameter('psu_bom', psu_bom)

    # Define lscpu number of cores
    lscpu_number_of_cores = 80
    platform.setParameter('lscpu_number_of_cores', lscpu_number_of_cores)
    platform.setParameter('sw_rel_file', "dgx-release")
    platform.setParameter('summary_serial_num_string', "DGX Serial Number")
    platform.setParameter('summary_base_os_string', "DGX BaseOS Version")
    platform.setParameter('chassis_serial_number_index', '1')
    # Define FRU device details
    platform.setParameter('fru_devices', ['Builtin FRU Device', 'BB_FRU', 'EXPB_FRU', 'HSBP_FRU', 'PDB_FRU'])

def __generate_dgx1_platform_objects(configs, columns=None, pretty=None):
    # Generate an IPlatform object for each valid DGX-1 configuration
    platforms = []
    for config in configs:
        config = list(config)
        platform = IPlatform(name=platform_type, config=config, columns=columns, pretty=pretty)
        check_system_vendor(platform, config)
        check_product_name(platform, config)
        # Define parameters for each platform
        define_gpu_parameters(platform, config)
        define_dimm_parameters(platform, config)
        define_mlnx_parameters(platform, config)
        define_network_parameters(platform, config)
        define_raid_parameters(platform, config)
        define_stor_cli_parameters(platform, config)
        define_generic_parameters(platform, config)

        platforms.append(platform)

    return platforms

def generate_dgx1_platform_objects():
    # Matrix (Cartesian product) of all possible DGX-1 configurations
    variants = [
        ('p100', 'v100', 'v100-32'),  # GPUs
        #TODO Update ngn variant config and check
        #('', 'ngn'),  # NGN variant
        ('', '1tb'),  # Optional 1TB system memory
        ('cx4', 'cx5', '2x_cx5', 'cx6', '2x_cx6'),  # Optional MLNX CX-5 cards
        ('ib', 'eth'),  # InfiniBand or Ethernet mode
        ('', 'raid1'),  # Optional RAID-1 for OS drive
        ('', 'qct_mezzanine_fiber', 'ocp_mezzanine_card'),  # Optional QCT Mezzanine fiber/OCP mezzaine card
    ]

    configs = list(itertools.product(*variants))
    columns = ["GPU", "System Memory", "Network Card", "Network Mode", "Storage", "Optional Mezzanine Card"]
    pretty = [
        {
            'p100': 'Tesla P100 w/ 16GB',
            'v100': 'Tesla V100 w/ 16GB',
            'v100-32': 'Tesla V100 w/ 32GB',
        },
        {
            '': '512GB',
            '1tb': '1TB',
        },
        {
            'cx4': '(x4) Mellanox ConnectX-4',
            'cx5': '(x4) Mellanox ConnectX-5',
            '2x_cx5': '(x2) Mellanox ConnectX-5',
            'cx6': '(x4) Mellanox ConnectX-6',
            '2x_cx6': '(x2) Mellanox ConnectX-6',
        },
        {
            'ib': 'InfiniBand',
            'eth': 'Ethernet',
        },
        {
            '': 'N/A',
            'raid1': 'RAID-1 OS Volume',
        },
        {
            '': 'N/A',
            'QCT_mezzanine_fiber': 'QCT Fiber',
            'OCP_mezzanine_card': 'Mellanox ConnectX-3',
        },
    ]
    return __generate_dgx1_platform_objects(configs=configs, columns=columns, pretty=pretty)

