# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from nvsmhealth.modules import lspci, dmidecode, smartctl, sysfs, ipmitool_cmd
from nvsmhealth.platform import IPlatform
import itertools
import json
import logging
from distutils.version import LooseVersion

__all__ = [
    'generate_dgx2_platform_objects'
]

platform_type = 'dgx-2'

# Define DGX-2H v100 GPUs SSID='131d' v1
dgx2h_gpu_ssid_v1 = '131d'
# Define DGX-2H v100 GPUs SSID='1353' v2
dgx2h_gpu_ssid_v2 = '1353'

def check_system_vendor(platform, config):
    platform.depends(sysfs.sysfs_dmi_system_vendor)

    def check_vendor():
        system_vendor = sysfs.sysfs_dmi_system_vendor.getOutput().strip()
        if system_vendor != 'NVIDIA':
            return False
        return True
    platform.qualify(check_vendor)


def check_product_name(platform, config):
    platform.depends(sysfs.sysfs_dmi_product_name)

    def check_product():
        product_name = sysfs.sysfs_dmi_product_name.getOutput().lower()
        if 'dgx-2' in product_name or 'dgx-2h' in product_name:
            return True
        return False
    platform.qualify(check_product)


def get_gpu_direct_topology():
    # DGX-2 with 16 GPUs has full mesh topology with two NVSwitches (six NVLinks) between each GPU
    gpu_direct_topology = {}
    for device in range(16):
        key = 'GPU{}'.format(device)
        sub_topology = {}
        for sub_device in range(16):
            sub_key = 'GPU{}'.format(sub_device)
            if key == sub_key:
                sub_topology[sub_key] = 'X'
            else:
                sub_topology[sub_key] = 'NV6'
        gpu_direct_topology[key] = sub_topology

    return gpu_direct_topology


def get_gpu_p2p_topology():
    gpu_p2p_topology = {}
    for device in range(16):
        key = 'GPU{}'.format(device)
        sub_topology = {}
        for sub_device in range(16):
            sub_key = 'GPU{}'.format(sub_device)
            if key == sub_key:
                sub_topology[sub_key] = 'X'
            else:
                sub_topology[sub_key] = 'OK'
        gpu_p2p_topology[key] = sub_topology

    return gpu_p2p_topology


def check_lspci_nvidia_gpu(parse_lspci_task):
    pci_devices = json.loads(parse_lspci_task.getResult())
    nvidia_devices = list(filter(lambda d: d.get(
        'vendor_id') == '10de', pci_devices.values()))
    known_gpus = [
        ('v100', ['1db8'])
    ]
    # Guess the GPU configuration based on what is installed on the system
    for name, device_ids in known_gpus:
        if any(map(lambda d: d.get('device_id') in device_ids, nvidia_devices)):
            return name
    return 'unknown'


def check_lspci_nvidia_gpu_ssid(parse_lspci_task, sysfs_task):
    pci_devices = json.loads(parse_lspci_task.getResult())
    nvidia_devices = list(filter(lambda d: d.get(
        'vendor_id') == '10de', pci_devices.values()))

    # Get the GPU sdevice_id configuration based on what is present on the system
    # Check ssid v1 and v2 only for dgx-2h platform
    product_name = sysfs_task.getOutput().lower()

    if 'dgx-2h' in product_name:
        try:
            if any(map(lambda d: d.get('sdevice_id') == dgx2h_gpu_ssid_v1, nvidia_devices)):
                return dgx2h_gpu_ssid_v1
            elif any(map(lambda d: d.get('sdevice_id') == dgx2h_gpu_ssid_v2, nvidia_devices)):
                return dgx2h_gpu_ssid_v2
        except:
            pass
    # Default config is DGX-2 GPUs
    return '12ab'


def define_gpu_parameters(platform, config):

    gpu_name = 'v100'
    device_id = '1db8'
    sm_count = 80

    # Set SSID for DGX-2 and DGX-2H GPUs
    if 'v100' in config:
        sdevice_id = '12ab'
    elif 'v100_2h_131d' in config:
        sdevice_id = dgx2h_gpu_ssid_v1
    elif 'v100_2h_1353' in config:
        sdevice_id = dgx2h_gpu_ssid_v2
    else:
        gpu_name = device_id = sm_count = sdevice_id = None

    # Check lspci output for GPU configuration
    parse_lspci_task = lspci.parse_lspci
    platform.depends(parse_lspci_task)

    # Check product name to differentiate DGX-2 and DGX-2H GPUs
    sysfs_task = sysfs.sysfs_dmi_product_name
    platform.depends(sysfs_task)

    def check_gpu():
        return check_lspci_nvidia_gpu(parse_lspci_task) == gpu_name
    platform.qualify(check_gpu)

    # Check lspci output for GPU SSID configuration
    def check_gpu_ssid():
        return check_lspci_nvidia_gpu_ssid(parse_lspci_task,
            sysfs_task) == sdevice_id
    platform.qualify(check_gpu_ssid)

    # Define bill-of-materials for GPUs in this configuration
    gpu_bom = {}
    gpu_bdfs = ['34', '36', '39', '3b', '57', '59', '5c',
                '5e', 'b7', 'b9', 'bc', 'be', 'e0', 'e2', 'e5', 'e7']
    for bdf in gpu_bdfs:
        # gpu_bom follows "lspci" format
        gpu_bom['{}:00.0'.format(bdf)] = {
            'class': '3D controller [0302]',
            'device_id': device_id,
            'sdevice_id': sdevice_id,
            'vendor': 'NVIDIA Corporation [10de]'
        }
    platform.setParameter('gpu_bom', gpu_bom)
    # Define SM count for this configuration
    platform.setParameter('gpu_sm_count', sm_count)
    # Define GPUDirect topology for this configuration
    gpu_direct_topology = get_gpu_direct_topology()
    platform.setParameter('gpu_direct_topology', gpu_direct_topology)

    # Define GPUs p2p topology for this configuration
    gpu_p2p_topology = get_gpu_p2p_topology()
    platform.setParameter('gpu_p2p_topology', gpu_p2p_topology)
    get_gpu_p2p_topology

    # Define GPU NVLink speed
    gpu_nvlink_speed = '25.781 GB/s'
    platform.setParameter('gpu_nvlink_speed', gpu_nvlink_speed)

    # Define bill-of-materials for gpu link width and speed
    gpu_link_device_ids = set(
        map(lambda x: '0000:{}:00.0'.format(x), gpu_bdfs))
    gpu_link_info = []
    for device in gpu_link_device_ids:
        gpu_link = {}
        gpu_link['bdf'] = device
        gpu_link['speed'] = {'P0': '8GT/s'}
        gpu_link['width'] = {'P0': 16}
        gpu_link_info.append(gpu_link)
    platform.setParameter('gpu_link_info', gpu_link_info)

    # Define GPU total retired page count threshold
    # Refer https://docs.nvidia.com/deploy/dynamic-page-retirement/index.html#faq for retired pages limit
    gpu_total_retired_page_count = 60
    platform.setParameter('gpu_total_retired_page_count', gpu_total_retired_page_count)

    # Define DCGM GPU blacklist recommendations watch type
    # 'a' represents 'All watches' including PCIE, Memory, Inforom, NVLINK, Thermal and Power
    platform.setParameter('dcgm_instant_watch_type', ['a', '--instantaneous'])
    # Enable only for longer tests
    #platform.setParameter('dcgm_watch_type', ['a'])

    # Enable GPU VBIOS version consistency check
    platform.setParameter('enable_vbios_version_check', True)


def define_dimm_parameters(platform, config):
    # Configuration has each DIMM of 64 GB
    dimm_size = '64 GB'

    # Define bill-of-materials for DIMMs in this configuration
    dimm_bom = {}
    dimm_slots1 = [('A', 'B', 'C', 'D', 'E', 'F'), (0, 1)]
    dimm_slots1 = itertools.product(*dimm_slots1)
    dimm_slots1 = map(lambda x: 'CPU0_DIMM_{}{}'.format(
        x[0], x[1]), dimm_slots1)
    # Segregation based on CPU0 or CPU1 slot config
    dimm_slots2 = [('G', 'H', 'I', 'J', 'K', 'L'), (0, 1)]
    dimm_slots2 = itertools.product(*dimm_slots2)
    dimm_slots2 = map(lambda x: 'CPU1_DIMM_{}{}'.format(
        x[0], x[1]), dimm_slots2)

    dimm_slots = list(dimm_slots1) + list(dimm_slots2)
    for slot in dimm_slots:
        # dimm_bom follows "dmidecode" format
        dimm_bom['Memory Device ({})'.format(slot)] = {
            'locator': slot,
            'size': dimm_size
        }
    platform.setParameter('dimm_bom', dimm_bom)

    # Define DIMM vendors
    dimm_vendors = set(['Samsung', 'Micron'])
    platform.setParameter('dmidecode_dimm_vendors', dimm_vendors)


def check_lspci_mlnx(known_mlnx_bdfs, parse_lspci_task):
    pci_devices = json.loads(parse_lspci_task.getResult())
    mlnx_cards = {k: pci_devices.get(k, {}) for k in known_mlnx_bdfs}
    if any(map(lambda v: v.get('device_id') == '1013', mlnx_cards.values())):
        return 'cx4'
    if any(map(lambda v: v.get('device_id') == '1017', mlnx_cards.values())):
        return 'cx5'
    if any(map(lambda v: v.get('device_id') == '101b', mlnx_cards.values())):
        return 'cx6'
    # NVbug: 200682445 - If configuration not matching either of the above, then either 
    # a. All devices in known_mlnx_bds aren't present on the host. 
    # b. A new device ID has been introduced.
    # If a new device ID has been introduced, needs to be added to product cartesian product.
    # If no configuration matched, assume that CX6 is expected on the current platform. 
    return 'cx6'


def define_mlnx_parameters(platform, config):
    mlnx_card = ''
    if '8x_cx4' in config:
        mlnx_card = 'cx4'
    elif '8x_cx5' in config:
        mlnx_card = 'cx5'
    elif '8x_cx6' in config:
        mlnx_card = 'cx6'
    # Check lspci output for GPU configuration
    parse_lspci_task = lspci.parse_lspci
    platform.depends(parse_lspci_task)

    def check_mlnx():
        known_mlnx_bdfs = ['35:00.0', '3a:00.0', '58:00.0', '5d:00.0',
                           'b8:00.0', 'bd:00.0', 'e1:00.0', 'e6:00.0']
        return check_lspci_mlnx(known_mlnx_bdfs, parse_lspci_task) == mlnx_card
    platform.qualify(check_mlnx)

    mlnx_dual_card = 'cx5'
    if 'dual_cx6' in config:
        mlnx_dual_card = 'cx6'

    def check_mlnx_dual_port():
        known_mlnx_bdfs = ['86:00.0', '86:00.1']
        return check_lspci_mlnx(known_mlnx_bdfs, parse_lspci_task) == mlnx_dual_card

    platform.qualify(check_mlnx_dual_port)

    firmware_versions = 1
    if mlnx_card != mlnx_dual_card:
        firmware_versions = 2

    platform.setParameter('mlnx_firmware_check', firmware_versions)


def check_lspci_mode(parse_lspci_task):
    # Make return variable, first item is Cluster NIC Mode, second item is Storage NIC Mode
    retRes = {}

    cluster_known_mlnx_bdfs = ['35:00.0', '3a:00.0', '58:00.0', '5d:00.0',
                       'b8:00.0', 'bd:00.0', 'e1:00.0', 'e6:00.0']

    pci_devices = json.loads(parse_lspci_task.getResult())
    clstr_mlnx_cards = {k: pci_devices.get(
        k, {}) for k in cluster_known_mlnx_bdfs}
    if any(map(lambda v: v.get('class_id') == '0200', clstr_mlnx_cards.values())):
        retRes['cluster_nic'] = 'eth'
    else:
        retRes['cluster_nic'] = 'ib'

    storage_known_mlnx_bdfs = ['86:00.0', '86:00.1']

    strge_mlnx_cards = {k: pci_devices.get(k, {}) for k in storage_known_mlnx_bdfs}
    if any(map(lambda v: v.get('class_id') == '0200', strge_mlnx_cards.values())):
        retRes['storage_nic'] = 'eth'
    else:
        retRes['storage_nic'] = 'ib'

    return retRes

def define_network_parameters(platform, config):
    # Define ethernet controller device ids
    device_ids = ['0', '1']
    default_device_ids = set(
        map(lambda x: '0000:86:00.{}'.format(x), device_ids))
    ethernet_controller_device_ids = set()

    if 'strge_ib' in config:
        storage_nic_mode = 'ib'
    
    if 'strge_eth' in config:
        storage_nic_mode = 'eth'
        ethernet_controller_device_ids.update(default_device_ids)

    if 'clstr_ib' in config:
        cluster_nic_mode = 'ib'
    elif 'clstr_eth' in config:
        cluster_nic_mode = 'eth'
        device_ids = ['35', '3a',  '58', '5d', 'b8', 'bd', 'e1', 'e6']
        ethernet_controller_device_ids.update(set(
            map(lambda x: '0000:{}:00.0'.format(x), device_ids)))

    # Check lspci output for GPU configuration
    parse_lspci_task = lspci.parse_lspci
    platform.depends(parse_lspci_task)

    def check_clstr_nic_mode():
        return check_lspci_mode(parse_lspci_task)['cluster_nic'] == cluster_nic_mode
    platform.qualify(check_clstr_nic_mode)

    def check_strge_nic_mode():
        return check_lspci_mode(parse_lspci_task)['storage_nic'] == storage_nic_mode
    platform.qualify(check_strge_nic_mode)

    mlnx_card = ''
    if '8x_cx4' in config:
        mlnx_card = 'cx4'
    elif '8x_cx5' in config:
        mlnx_card = 'cx5'
    elif '8x_cx6' in config:
        mlnx_card = 'cx6'
    
    # Define bill-of-materials for ethernet controller 8gts link speed and width x16
    ethernet_controller_info = []
    for device_id in ethernet_controller_device_ids:
        ethernet_controller_id_info = {}
        ethernet_controller_id_info['bdf'] = device_id
        # ethernet_controller_id_info['name'] = '10G'
        ethernet_controller_id_info['speed'] = '8GT/s'
        ethernet_controller_id_info['width'] = 16
        ethernet_controller_info.append(ethernet_controller_id_info)

    # Intel NIC Addition
    intel_nic_bdf = ['06']
    for bdf in intel_nic_bdf:
        ethernet_controller_id_info = {}
        ethernet_controller_id_info['bdf'] = '0000:{}:00.0'.format(bdf)
        ethernet_controller_id_info['speed'] = '2.5GT/s'
        ethernet_controller_id_info['width'] = 1
        ethernet_controller_info.append(ethernet_controller_id_info)

    platform.setParameter('ethernet_controller_info', ethernet_controller_info)

    vendor_device_id = ''
    if mlnx_card in ['cx5']:
        device_ids = ['05', '84']
        vendor_device_id = '1017'
    if mlnx_card == 'cx4':
        vendor_device_id = '1013'
    if mlnx_card in ['cx6']:
        vendor_device_id = '101b'

    # Define bill-of-materials for ethernet controllers
    ethernet_controllers_bom = {}
    # ethernet_controllers_bom follows "lspci" format
    intel_nic_bdf = ['06']
    for bdf in intel_nic_bdf:
        ethernet_controllers_bom['{}:00.0'.format(
            bdf)] = {
            'class': 'Ethernet controller [0200]',
            'device_id': '1533',
            'vendor': 'Intel Corporation [8086]'
        }


    dual_port_device_id = '1017'
    if 'dual_cx6' in config:
        dual_port_device_id = '101b'

    if storage_nic_mode == 'eth':
        eth_bdfs = ['0', '1']
        for bdf in eth_bdfs:
            ethernet_controllers_bom['86:00.{}'.format(
                bdf)] = {
                'class': 'Ethernet controller [0200]',
                'device_id': dual_port_device_id,
                'vendor': 'Mellanox Technologies [15b3]'
            }

    if cluster_nic_mode == 'eth':
        # Define bill-of-materials for additional ethernet controllers
        eth_bdfs = ['35', '3a', '58', '5d', 'b8', 'bd', 'e1', 'e6']
        for bdf in eth_bdfs:
            ethernet_controllers_bom['{}:00.0'.format(
                bdf)] = {
                'class': 'Ethernet controller [0200]',
                'device_id': vendor_device_id,
                'vendor': 'Mellanox Technologies [15b3]'
            }

    platform.setParameter('ethernet_controllers_bom', ethernet_controllers_bom)

    ib_controllers_bom = {}
    ib_controller_link_info = []

    if storage_nic_mode == 'ib':
        ib_bdfs = ['0', '1']
        for bdf in ib_bdfs:
            ib_controllers_bom['86:00.{}'.format(
                bdf)] = {
                'class': 'Infiniband controller [0207]',
                'device_id': dual_port_device_id,
                'vendor': 'Mellanox Technologies [15b3]'
            }
        
        for bdf in ib_bdfs:
            ib_controller_id_info = {}
            ib_controller_id_info['bdf'] = '0000:86:00.{}'.format(bdf)
            ib_controller_id_info['speed'] = '8GT/s'
            ib_controller_id_info['width'] = 16
            ib_controller_link_info.append(ib_controller_id_info)

    if cluster_nic_mode == 'ib':
        # Add bill-of-materials for IB controllers
        ib_bdfs = ['35', '3a', '58', '5d', 'b8', 'bd', 'e1', 'e6']
        for bdf in ib_bdfs:
            # ib_controllers_bom follows "lspci" format
            ib_controllers_bom['{}:00.0'.format(bdf)] = {
                'class': 'Infiniband controller [0207]',
                'device_id': vendor_device_id,
                'vendor': 'Mellanox Technologies [15b3]'
            }

        # Define bill-of-materials for ib controller link speed and width
        device_ids = ['35', '3a', '58', '5d', 'b8', 'bd', 'e1', 'e6']
        ib_controller_link_device_ids = set(
            map(lambda x: '0000:{}:00.0'.format(x), device_ids))
        # Define bill-of-materials for ib controller link speed and width
        
        for device_id in ib_controller_link_device_ids:
            ib_controller_id_info = {}
            ib_controller_id_info['bdf'] = device_id
            ib_controller_id_info['speed'] = '8GT/s'
            ib_controller_id_info['width'] = 16
            ib_controller_link_info.append(ib_controller_id_info)

    if len(ib_controllers_bom) != 0:
        platform.setParameter('ib_controllers_bom', ib_controllers_bom)

    if len(ib_controller_link_info) != 0:
        platform.setParameter('ib_controller_link_info',
                              ib_controller_link_info)

    # Define bill-of-materials for pcie switches
    pcie_switches_bom = {}
    pcie_bdfs = ['2c', '32', '37', '4f', '55', '5a',
                 '5f', 'af', 'b5', 'ba', 'bf', 'd8', 'de', 'e3']
    for bdf in pcie_bdfs:
        # pcie_switches_bom follows "lspci" format
        pcie_switches_bom['{}:00.0'.format(bdf)] = {
            'class': '0604',
            'vendor': '10b5'
        }
    platform.setParameter('pcie_switches_bom', pcie_switches_bom)

    # Define bill-of-materials for nvswitch
    nvswitch_bom = {}
    nvswitch_bdfs = ['61', '62', '63', '65', '66',
                     '67', 'c1', 'c2', 'c3', 'c5', 'c6', 'c7']
    for bdf in nvswitch_bdfs:
        # nvswitch_bom follows "lspci" format
        nvswitch_bom['{}:00.0'.format(bdf)] = {
            'class_id': '0680',
            'device_id': '1ac2',
            'vendor': 'NVIDIA Corporation [10de]'
        }
    platform.setParameter('nvswitch_bom', nvswitch_bom)


def check_lspci_nvme(parse_lspci_task):
    known_nvme_bdfs = ['2e:00.0', '2f:00.0', '30:00.0', '31:00.0',
                       '51:00.0', '52:00.0', '53:00.0', '54:00.0',
                       'b1:00.0', 'b2:00.0', 'b3:00.0', 'b4:00.0',
                       'da:00.0', 'db:00.0', 'dc:00.0', 'dd:00.0']

    lspci = json.loads(parse_lspci_task.getResult())
    nvme_drives = {k: v for k, v in lspci.items() if v['class_id'] == '0108'}
    drives = {k: v for k, v in nvme_drives.items() if k in known_nvme_bdfs}

    if len(drives) > 8:
        # Assume 16 NVMe drive configuration
        return '16nvme'
    return '8nvme'


def define_nvme_parameters(platform, config):
    if '8nvme' in config:
        drives = '8nvme'
        nvme_config = [{'1.0TB': 2, '3.8TB': 8}]
    elif '16nvme' in config:
        drives = '16nvme'
        nvme_config = [{'1.0TB': 2, '3.8TB': 16}]
    # Check lspci output for GPU configuration
    parse_lspci_task = lspci.parse_lspci
    platform.depends(parse_lspci_task)

    def check_nvme():
        return check_lspci_nvme(parse_lspci_task) == drives
    platform.qualify(check_nvme)
    # Set config parameter for nvme
    platform.setParameter('nvme', nvme_config)
    platform.setParameter('nvme_check_smart_log', True)
    platform.setParameter('mdadm_disk_status', True)
    platform.setParameter('mdadm_volume_status', True)
    platform.setParameter('skip_nvme_drive_model', 'KCM5DRUG3T84')

def define_generic_parameters(platform, config):
    def bmc_firmware_gt_1_05_05():
        try:
            platform.depends(ipmitool_cmd.parse_ipmi_info)
            version = ipmitool_cmd.parse_ipmi_info.getResult()[
                'firmware_revision']
            # BUG-200566446: Aux version is in hexadecimal
            ver = version.split(".")
            try:
                ver[2] = "{}".format(int(ver[2], 16))
                ver = ".".join(ver)
            except IndexError:
                pass

            if LooseVersion(ver) >= LooseVersion('1.05.05'):
                return True
        except Exception as e:
            logging.error(e)
        return False
    if 'bmc_firmware_gt_1_05_05' in config:
        platform.qualify(bmc_firmware_gt_1_05_05)
    else:
        platform.qualify(lambda: not bmc_firmware_gt_1_05_05())

    # Define bill-of-materials for FAN
    fan_bom = {}
    for i in range(10):
        fan_bom['FAN{}_R'.format(i)] = True
        fan_bom['FAN{}_F'.format(i)] = True
    if 'bmc_firmware_gt_1_05_05' in config:
        for i in range(4):
            fan_bom['FAN{}_PDB'.format(i)] = True
    else:
        for i in range(4):
            fan_bom['PDB_FAN{}'.format(i)] = True
    platform.setParameter('fan_bom', fan_bom)

    # Define bill-of-materials for PSU
    psu_bom = {}
    if 'bmc_firmware_gt_1_05_05' in config:
        for i in range(6):
            psu_bom['POWER_PSU{}'.format(i)] = True
            psu_bom['STATUS_PSU{}'.format(i)] = True
    else:
        for i in range(6):
            psu_bom['PSU{} POWER'.format(i)] = True
            psu_bom['PSU{} STATUS'.format(i)] = True

    platform.setParameter('psu_bom', psu_bom)

    # Define lscpu number of cores
    lscpu_number_of_cores = 96
    platform.setParameter('lscpu_number_of_cores', lscpu_number_of_cores)
    platform.setParameter('sw_rel_file', "dgx-release")
    platform.setParameter('summary_serial_num_string', "DGX Serial Number")
    platform.setParameter('summary_base_os_string', "DGX BaseOS Version")
    platform.setParameter('chassis_serial_number_index', '1')
    # Define FRU device details
    platform.setParameter('fru_devices', ['MID_FRU'])
    # Define firmware version conversion from hex to decimal
    platform.setParameter('aux_firmware_version_type', 'hex')


def __generate_dgx2_platform_objects(configs, columns=None, pretty=None):
    # Generate an IPlatform object for each valid DGX-2 configuration
    platforms = []
    for config in configs:
        config = list(config)
        platform = IPlatform(name=platform_type, config=config,
                             columns=columns, pretty=pretty)
        check_system_vendor(platform, config)
        check_product_name(platform, config)
        # Define parameters for each platform
        define_gpu_parameters(platform, config)
        define_dimm_parameters(platform, config)
        define_mlnx_parameters(platform, config)
        define_network_parameters(platform, config)
        define_nvme_parameters(platform, config)
        define_generic_parameters(platform, config)

        platforms.append(platform)

    return platforms


def generate_dgx2_platform_objects():
    # Matrix (Cartesian product) of all possible DGX-2 configurations
    variants = [
        # GPUs 'Tesla V100-SXM3-32GB' for DGX-2 and 'Tesla V100-SXM3-32GB-H' for DGX-2H with SSIDs
        ('v100', 'v100_2h_131d', 'v100_2h_1353'),
        ('8x_cx4', '8x_cx5', '8x_cx6'), # Clustering NIC Variations: cx4, cx5, cx6
        ('dual_cx5', 'dual_cx6'),       # Storage NIC Variations: cx5, cx6
        ('clstr_ib', 'clstr_eth'),      # Clustering NICs Modes: InfiniBand or Ethernet mode
        ('strge_ib', 'strge_eth'),      # Storage NICs Modes: InfiniBand or Ethernet mode
        ('8nvme', '16nvme'),            # NVME drive count: 8 or 16
        ('', 'bmc_firmware_gt_1_05_05')
    ]

    configs = list(itertools.product(*variants))
    columns = ["GPU", "Network Card", "Network Mode", "Storage"]
    pretty = [
        {
            'v100': 'Tesla V100 w/ 32GB',
            'v100_2h_131d': 'Tesla V100 w/ 32GB for DGX-2H (SSID 131d)',
            'v100_2h_1353': 'Tesla V100 w/ 32GB for DGX-2H (SSID 1353)',
        },
        {
            '8x_cx4': 'Mellanox ConnectX-4',
            '8x_cx5': 'Mellanox ConnectX-5',
            '8x_cx6': 'Mellanox ConnectX-6',
        },
        {
            'dual_cx5': 'Mellanox ConnectX-5',
            'dual_cx6': 'Mellanox ConnectX-6',
        },
        {
            'clstr_ib': 'Clustering NICs IB Mode',
            'clstr_eth': 'Clustering NICs Eth Mode',
        },
        {
            'strge_ib': 'Storage NICs IB Mode',
            'strge_eth': 'Storage NICs Eth Mode',
        },
        {
            '8nvme': '(x8) NVMe drives',
            '16nvme': '(x16) NVMe drives',
        },
        {
            '': 'N/A',
            'bmc_firmware_gt_1_05_05': 'BMC firmware version greater than 1.05.05',
        },
    ]
    return __generate_dgx2_platform_objects(configs=configs, columns=columns, pretty=pretty)
