# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

""" Support for DGX-1 systems in any configuration.
"""

from nvsmhealth.modules import lspci, dmidecode, smartctl, sysfs
from nvsmhealth.platform import IPlatform
import itertools, json

__all__ = [
    'generate_dgxStation_platform_objects'
]

platform_type = 'dgxStation'

def check_system_vendor(platform, config):
    platform.depends(sysfs.sysfs_dmi_system_vendor)
    def check_vendor():
        system_vendor = sysfs.sysfs_dmi_system_vendor.getOutput().strip()
        if system_vendor != 'NVIDIA':
            return False
        return True
    platform.qualify(check_vendor)

def check_product_name(platform, config):
    platform.depends(sysfs.sysfs_dmi_product_name)
    def check_product():
        product_name = sysfs.sysfs_dmi_product_name.getOutput().lower()
        if 'dgx station' in product_name and '920-23487' not in product_name:
            return True
        # TODO: Check specific products, like NGN, MaxQ, and 32G
        return False
    platform.qualify(check_product)


def check_lspci_nvidia_gpu(parse_lspci_task):
    pci_devices = json.loads(parse_lspci_task.getResult())
    nvidia_devices = list(filter(lambda d: d.get('vendor_id') == '10de', pci_devices.values()))
    known_gpus = [
        ('v100', ['1db2']),
        ('p100', ['15f9']),
        ('v100_32', ['1db7']),
    ]
    # Guess the GPU configuration based on what is installed on the system
    for name, device_ids in known_gpus:
        if any(map(lambda d: d.get('device_id') in device_ids, nvidia_devices)):
            return name
    return 'unknown'

def define_gpu_parameters(platform, config):
    # TODO: Add SSID variations for NGN SKUs?
    if 'p100' in config:
        gpu_name = 'p100'
        device_id = '15f9'
        sm_count = 56
        gpu_nvlink_speed = '20 GB/s'
    elif 'v100_32' in config:
        gpu_name = 'v100_32'
        device_id = '1db7'
        sm_count = 80
        gpu_nvlink_speed = '25.781 GB/s'
    else:
        gpu_name = 'v100'
        device_id = '1db2'
        sm_count = 80
        gpu_nvlink_speed = '25.781 GB/s'
    # Check lspci output for GPU configuration
    parse_lspci_task = lspci.parse_lspci
    platform.depends(parse_lspci_task)
    def check_gpu():
        return check_lspci_nvidia_gpu(parse_lspci_task) == gpu_name
    platform.qualify(check_gpu)
    # Define bill-of-materials for GPUs in this configuration
    gpu_bom = {}
    gpu_bdfs = [ '07', '08', '0e', '0f' ]
    for bdf in gpu_bdfs:
        # gpu_bom follows "lspci" format
        gpu_bom['{}:00.0'.format(bdf)] = {
            'class': 'VGA compatible controller [0300]',
            'device_id': device_id
        }
    platform.setParameter('gpu_bom', gpu_bom)
    # Define SM count for this configuration
    platform.setParameter('gpu_sm_count', sm_count)

    # Define bill-of-materials for gpu link width and speed
    gpu_link_device_ids = set(map(lambda x: '0000:{}:00.0'.format(x), gpu_bdfs))
    gpu_link_info = []
    for device in gpu_link_device_ids:
        gpu_link = {}
        gpu_link['bdf'] = device
        gpu_link['speed'] = {'P0':'8GT/s'}
        gpu_link['width'] = {'P0': 16}
        gpu_link_info.append(gpu_link)

    platform.setParameter('gpu_link_info', gpu_link_info)
    # Define DCGM GPU blacklist recommendations watch type
    # 'a' represents 'All watches' including PCIE, Memory, Inforom, NVLINK, Thermal and Power
    platform.setParameter('dcgm_instant_watch_type', ['a', '--instantaneous'])
    # Enable only for longer tests
    #platform.setParameter('dcgm_watch_type', ['a'])

    # Define GPU NVLink speed
    platform.setParameter('gpu_nvlink_speed', gpu_nvlink_speed)

    # Define GPU total retired page count threshold
    # Refer https://docs.nvidia.com/deploy/dynamic-page-retirement/index.html#faq for retired pages limit
    gpu_total_retired_page_count = 60
    platform.setParameter('gpu_total_retired_page_count', gpu_total_retired_page_count)

    # Enable GPU VBIOS version consistency check
    platform.setParameter('enable_vbios_version_check', True)


def define_dimm_parameters(platform, config):
    # Check DMI table for DIMM configuration
    parse_dimms_task = dmidecode.parse_memory_dimms
    platform.depends(parse_dimms_task)
    # Configuration has 1TB RAM total (each DIMM is 64 GB)
    def check_512gb_ram():
        # If there is at least one '64 GB' DIMM installed, then assume this
        # system must be a 1TB configuration
        observed_dimms = json.loads(parse_dimms_task.getResult())
        if '64 GB' in observed_dimms:
            return True
        return False
    def check_not_512gb_ram():
        return not check_512gb_ram()
    if '512gb' in config:
        platform.qualify(check_512gb_ram)
        dimm_size = '64 GB'
        mem_size = 528294240
    else:
        # Configuration has 512GB RAM total (each DIMM is 32 GB)
        dimm_size = '32 GB'
        mem_size = 264057072
        platform.qualify(check_not_512gb_ram)
    # Define bill-of-materials for DIMMs in this configuration
    dimm_bom = {}
    dimm_slots = [('A', 'B', 'C', 'D'), (1, 2)]
    dimm_slots = itertools.product(*dimm_slots)
    dimm_slots = map(lambda x: 'DIMM_{}{}'.format(x[0], x[1]), dimm_slots)
    for slot in dimm_slots:
        # dimm_bom follows "dmidecode" format
        dimm_bom['Memory Device ({})'.format(slot)] = {
            'locator': slot,
            'size': dimm_size
        }
    platform.setParameter('dimm_bom', dimm_bom)
    # Define DIMM vendors
    dimm_vendors = set(['Samsung', 'Micron'])
    platform.setParameter('dmidecode_dimm_vendors', dimm_vendors)

    platform.setParameter('meminfo_memory_size', mem_size)

def check_lspci_mode(parse_lspci_task):
    known_eth_bdfs = ['02:00.0', '02:00.1']

    pci_devices = json.loads(parse_lspci_task.getResult())
    eth_cards = { k: pci_devices.get(k, {}) for k in known_eth_bdfs }
    if any(map(lambda v: v.get('class_id') == '0200', eth_cards.values())):
        return 'eth'
    return 'unknown'

def define_network_parameters(platform, config):
    mode='eth'
    # Check lspci output for GPU configuration
    parse_lspci_task = lspci.parse_lspci
    platform.depends(parse_lspci_task)
    def check_mode():
        return check_lspci_mode(parse_lspci_task) == mode
    platform.qualify(check_mode)

    # Check lspci output for GPU configuration
    parse_lspci_task = lspci.parse_lspci
    platform.depends(parse_lspci_task)

    # Define bill-of-materials for ethernet controllers
    platform.depends(lspci.parse_lspci)

    def check_qct_mezzanine_fiber():
        pci_devices = json.loads(lspci.parse_lspci.getResult())
        return pci_devices.get('02:00.0', {}).get('device_id') == '10fb'

    def check_ocp_mezzanine_card():
        pci_devices = json.loads(lspci.parse_lspci.getResult())
        return pci_devices.get('02:00.0', {}).get('device_id') == '1007'

    platform.qualify(lambda: not check_qct_mezzanine_fiber()
                                 and not check_ocp_mezzanine_card())
    device_id = '1563'
    # Common params
    device_ids = ['0', '1']
    class_str = 'Ethernet controller [0200]'
    vendor = 'Intel Corporation [8086]'
    speed = '8GT/s'
    width = 4
    name = '1G'

    ethernet_controllers_bom = {}
    for bdf in device_ids:
        # ethernet_controllers_bom follows "lspci" format
        ethernet_controllers_bom['02:00.{}'.format(bdf)] = {
            'class': class_str,
            'device_id': device_id,
            'vendor': vendor
        }

    # Define bill-of-materials ethernet controller link speed 5gts and width x8
    ethernet_controller_device_ids = set(map(lambda x: '0000:02:00.{}'.format(x), device_ids))
    ethernet_controller_info = []
    for device_id in ethernet_controller_device_ids:
        ethernet_controller_id_info = {}
        ethernet_controller_id_info['bdf'] = device_id
        # ethernet_controller_id_info['name'] = name
        ethernet_controller_id_info['speed'] = speed
        ethernet_controller_id_info['width'] = width
        ethernet_controller_info.append(ethernet_controller_id_info)

    platform.setParameter('ethernet_controllers_bom', ethernet_controllers_bom)
    platform.setParameter('ethernet_controller_info', ethernet_controller_info)


    # Define bill-of-materials for pcie switches
    pcie_switches_bom = {}
    pcie_bdfs = [ '05', '0c']
    for bdf in pcie_bdfs:
        device_id = '8747'
        # pcie_switches_bom follows "lspci" format
        # device is taken directly from "lspci" output, without appenidng "_id"
        pcie_switches_bom['{}:00.0'.format(bdf)] = {
            'class': '0604',
            'device': device_id,
            'vendor': '10b5'
        }
    platform.setParameter('pcie_switches_bom', pcie_switches_bom)

def check_smartctl_scan(parse_smartctl_task):
    disks = parse_smartctl_task.getResult()
    if disks and disks.get('1.92 TB') > 4:
        return 8
    return 4

def define_raid_parameters(platform, config):

    # Check smart scan output for raid1 configuration
    parse_smartctl_task = smartctl.parse_smartctl_device_info
    platform.depends(parse_smartctl_task)
    def check_drives():
        return check_smartctl_scan(parse_smartctl_task) == drives

    if '4ssd' in config:
        drives = 4
    else:
        drives = 8

    platform.qualify(check_drives)

    # Define disk controllers bom
    disk_controllers_bom = {}
    disk_controllers_bom['00:1f.2'] = {
        'class': 'SATA controller [0106]',
        'device_id': '8d02',
        'vendor_id': '8086'
    }
    platform.setParameter('disk_controllers_bom', disk_controllers_bom)

    platform.setParameter('smartctl_check_ssd_brick', True)

    platform.setParameter('check_smartctl_disk_count', drives)

    platform.setParameter('mdadm_volume_status', True)

    platform.setParameter('mdadm_disk_status', True)

def define_generic_parameters(platform, config):

    # Define lscpu number of cores
    lscpu_number_of_cores = 40
    platform.setParameter('lscpu_number_of_cores', lscpu_number_of_cores)
    platform.setParameter('sw_rel_file', "dgx-release")
    platform.setParameter('summary_serial_num_string', "DGX Serial Number")
    platform.setParameter('summary_base_os_string', "DGX BaseOS Version")
    platform.setParameter('chassis_serial_number_index', '1')
    platform.setParameter('skip_bmc_revision_check', True)

def __generate_dgxStation_platform_objects(configs, columns=None, pretty=None):
    # Generate an IPlatform object for each valid DGX-1 configuration
    platforms = []
    for config in configs:
        config = list(config)
        platform = IPlatform(name=platform_type, config=config, columns=columns, pretty=pretty)
        check_system_vendor(platform, config)
        check_product_name(platform, config)
        # Define parameters for each platform
        define_gpu_parameters(platform, config)
        define_dimm_parameters(platform, config)
        define_network_parameters(platform, config)
        define_raid_parameters(platform, config)
        define_generic_parameters(platform, config)

        platforms.append(platform)

    return platforms

def generate_dgxStation_platform_objects():
    # Matrix (Cartesian product) of all possible DGX-1 configurations
    variants = [
        ('p100', 'v100', 'v100_32'),  # GPUs
        ('256gb', '512gb'),  # Optional 1TB system memory
        ('eth'),  # InfiniBand or Ethernet mode
        ('4ssd', '8ssd'), # 4 or 8 SSDs
    ]

    configs = list(itertools.product(*variants))
    columns = ["GPU", "System Memory", "Network Card", "Storage Drives", "Storage Raid"]
    pretty = [
        {
            'p100': 'Tesla P100 w/ 16GB',
            'v100': 'Tesla V100 w/ 16GB',
            'v100_32': 'Tesla V100 DGXS 32GB'
        },
        {
            '256gb': '256GB',
            '512gb': '512GB',
        },
        {
            'eth': 'Ethernet',
        },
        {
            '4ssd': '4 SSDs',
            '8ssd': '8 SSDs',
        },
    ]
    return __generate_dgxStation_platform_objects(configs=configs, columns=columns, pretty=pretty)
