# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from nvsmhealth.modules import lspci, dmidecode, smartctl, sysfs
from nvsmhealth.platform import IPlatform
import itertools, json
import logging

__all__ = [
    'generate_dgxa100_platform_objects'
]

platform_type = 'dgxa100'

def check_system_vendor(platform, config):
    platform.depends(sysfs.sysfs_dmi_system_vendor)
    def check_vendor():
        system_vendor = sysfs.sysfs_dmi_system_vendor.getOutput().strip()
        if 'NVIDIA' in system_vendor:
            return True
        return False
    platform.qualify(check_vendor)

def check_product_name(platform, config):
    platform.depends(sysfs.sysfs_dmi_product_name)
    def check_product():
        product_name = sysfs.sysfs_dmi_product_name.getOutput().lower()
        if 'dgxa100' in product_name and '920-23687' in product_name:
            return True
        return False
    platform.qualify(check_product)

def get_gpu_direct_topology():
    gpus = 8
    gpu_direct_topology = {}
    for device in range(gpus):
        key = 'GPU{}'.format(device)
        sub_topology = {}
        for sub_device in range(gpus):
            sub_key = 'GPU{}'.format(sub_device)
            if key == sub_key:
                sub_topology[sub_key] = 'X'
            else:
                sub_topology[sub_key] = 'NV12'
        gpu_direct_topology[key] = sub_topology

    return gpu_direct_topology

def check_lspci_gpu(platform, config, parse_lspci_task):
    lspci = json.loads(parse_lspci_task.getResult())
    # PCI Device ID used in NVIDIA GPU for A100 40GB is 20b0h and 80GB is 20b2h.
    gpu_40GB_devices = {k: v for k, v in lspci.items() if v['device_id'] ==
            '20b0'}
    gpu_80GB_devices = {k: v for k, v in lspci.items() if v['device_id'] ==
            '20b2'}

    if len(gpu_40GB_devices) > len(gpu_80GB_devices) :
        return '40gb_gpu'
    return '80gb_gpu'

def define_gpu_parameters(platform, config):
    # PCI Device ID used in NVIDIA GPU for A100 40GB is 20b0h and 80GB is 20b2h.
    if '40gb_gpu' in config:
        device_id = '20b0'
        sdevice_id = '134f'
        gpu_sku = '40gb_gpu'
    elif '80gb_gpu' in config:
        device_id = '20b2'
        sdevice_id = '1463'
        gpu_sku = '80gb_gpu'

    # Check lspci output for GPU configuration
    parse_lspci_task = lspci.parse_lspci
    platform.depends(parse_lspci_task)

    def check_gpu():
        return check_lspci_gpu(platform, config, parse_lspci_task) == gpu_sku
    platform.qualify(check_gpu)

    # Set common properties for both 40GB and 80GB GPU SKU.
    gpu_name = 'ampere'
    sm_count = 80

    # Define bill-of-materials for GPUs in this configuration
    gpu_bom = {}
    gpu_bdfs = [ '07', '0f', '47', '4e', '87', '90', 'b7', 'bd']
    for bdf in gpu_bdfs:
        # gpu_bom follows "lspci" format
        gpu_bom['{}:00.0'.format(bdf)] = {
            'class': '3D controller [0302]',
            'device_id': device_id,
            'sdevice_id': sdevice_id,
            'vendor': 'NVIDIA Corporation [10de]'
        }
    platform.setParameter('gpu_bom', gpu_bom)

    # Define SM count for this configuration
    platform.setParameter('gpu_sm_count', sm_count)

    # Enable GPU MIG before Topology checks
    platform.setParameter('enable_gpu_mig', True)

    # Define GPUDirect topology for this configuration
    gpu_direct_topology = get_gpu_direct_topology()
    platform.setParameter('gpu_direct_topology', gpu_direct_topology)

    # Define GPU NVLink speed
    gpu_nvlink_speed = '25 GB/s'
    platform.setParameter('gpu_nvlink_speed', gpu_nvlink_speed)

    # Define bill-of-materials for gpu link width and speed
    gpu_link_device_ids = set(map(lambda x: '0000:{}:00.0'.format(x), gpu_bdfs))
    gpu_link_info = []
    for device in gpu_link_device_ids:
        gpu_link = {}
        gpu_link['bdf'] = device
        gpu_link['speed'] = {'P0':'16GT/s'}
        gpu_link['width'] = {'P0': 16} #todo
        gpu_link_info.append(gpu_link)
    platform.setParameter('gpu_link_info', gpu_link_info)

    # Define GPU Blacklist Recommendations
    # platform.setParameter('blacklist_recommendations_instant', True)
    # platform.setParameter('blacklist_recommendations', False)

    platform.setParameter('dcgm_instant_watch_type', ['a', '--instantaneous'])


    # Enable GPU VBIOS version consistency check
    platform.setParameter('enable_vbios_version_check', True)


def define_dimm_parameters(platform, config):
    # Configuration has each DIMM of 64 GB
    dimm_size = '64 GB'

    # Check DMI table for DIMM configuration
    parse_dimms_task = dmidecode.parse_memory_dimms
    platform.depends(parse_dimms_task)

    def check_2tb_ram():
        # Check if number of DIMMs is 16 or more
        observed_dimms = json.loads(parse_dimms_task.getResult())
        populated_dimms = [i for i in observed_dimms if i == dimm_size]
        if len(populated_dimms)>16:
            return True
        return False

    def check_not_2tb_ram():
        return not check_2tb_ram()

    dimm_slots = []

    if '2tb' in config:
        platform.qualify(check_2tb_ram)
        dimm_slots1 = [('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H'), (0, 1)]
        dimm_slots1 = itertools.product(*dimm_slots1)
        dimm_slots1 = map(lambda x: 'CPU0_DIMM_{}{}'.format(
            x[0], x[1]), dimm_slots1)

        dimm_slots2 = [('I', 'J', 'K', 'L', 'M', 'N', 'O', 'P'), (0, 1)]
        dimm_slots2 = itertools.product(*dimm_slots2)
        dimm_slots2 = map(lambda x: 'CPU1_DIMM_{}{}'.format(
            x[0], x[1]), dimm_slots2)
        dimm_slots = list(dimm_slots1) + list(dimm_slots2)

    else:
        dimm_slots1 = [('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H'), (1, )]
        dimm_slots1 = itertools.product(*dimm_slots1)
        dimm_slots1 = map(lambda x: 'CPU0_DIMM_{}{}'.format(
            x[0], x[1]), dimm_slots1)

        dimm_slots2 = [('I', 'J', 'K', 'L', 'M', 'N', 'O', 'P'), (1, )]
        dimm_slots2 = itertools.product(*dimm_slots2)
        dimm_slots2 = map(lambda x: 'CPU1_DIMM_{}{}'.format(
            x[0], x[1]), dimm_slots2)
        platform.qualify(check_not_2tb_ram)

        dimm_slots = list(dimm_slots1) + list(dimm_slots2)

    # Define bill-of-materials for DIMMs in this configuration
    # Segregation based on CPU0 or CPU1 slot config

    dimm_bom = {}
    for slot in dimm_slots:
        # dimm_bom follows "dmidecode" format
        dimm_bom['Memory Device ({})'.format(slot)] = {
            'locator': slot,
            'size': dimm_size
        }
    platform.setParameter('dimm_bom', dimm_bom)

    # Define DIMM vendors
    dimm_vendors = set(['Micron Technology', 'SK Hynix'])
    platform.setParameter('dmidecode_dimm_vendors', dimm_vendors)

    # Define DIMM part number
    platform.setParameter('dimm_part_number', ['36ASF8G72PZ-3G2B2','HMAA8GR7AJR4N-XN'])

def define_mlnx_parameters(platform, config):
    # Only 'cx-6' support for now
    # will not work without mellanox firmware manager
    # Currently disabled since we don't know if different FW or not
    platform.setParameter('mlnx_firmware_check', True)

def check_lspci_ib(parse_lspci_task):
    known_strg_ib_bdfs = ['61:00.0', 'e1:00.0']

    lspci = json.loads(parse_lspci_task.getResult())
    strg_ib_devs = {k: v for k, v in lspci.items() if v['device_id'] == '101b'}
    strg_ib = {k: v for k, v in strg_ib_devs.items() if k in known_strg_ib_bdfs}

    if len(strg_ib) > 1:
        return 'strg_nic_qty_2'
    return 'strg_nic_qty_1'

def check_lspci_mode(platform, config, parse_lspci_task):
    # Make retRes, first item is clustering NIC mode, second item is storage NIC mode
    retRes ={}
    pci_devices = json.loads(parse_lspci_task.getResult())

    # Find Mode of Clustering NIC (IB/Eth), append as first item in retRes
    if 'bios_firmware_gt_0_25' in config:
        known_clstr_mlnx_bdfs = ['0c:00.0', '12:00.0', '4b:00.0', '54:00.0',
                '8d:00.0', '94:00.0', 'ba:00.0', 'cc:00.0']
    else:
        known_clstr_mlnx_bdfs = ['0c:00.0', '12:00.0', '4b:00.0', '54:00.0',
                '8d:00.0', '94:00.0', 'ba:00.0', 'ca:00.0']

    clstr_mlnx_cards = {k: pci_devices.get(k, {}) for k in known_clstr_mlnx_bdfs}
    if any(map(lambda v: v.get('class_id') == '0200', clstr_mlnx_cards.values())):
        retRes['cluster_nic'] = 'eth'
    else:
        retRes['cluster_nic'] = 'ib'

    # Find Mode of Storage NIC (IB/Eth), append as second item in retRes
    known_strg_mlnx_bdfs = ['61:00.0', '61:00.1', 'e1:00.0', 'e1:00.1']
    strg_mlnx_cards = {k: pci_devices.get(
        k, {}) for k in known_strg_mlnx_bdfs}
    if any(map(lambda v: v.get('class_id') == '0200', strg_mlnx_cards.values())):
        retRes['storage_nic'] = 'eth'
    else:
        retRes['storage_nic'] = 'ib'

    return retRes


def define_network_parameters(platform, config):
    # Section 1: Define dependencies and figure out modes and configs
    # Check lspci output for Mlnx
    parse_lspci_task = lspci.parse_lspci
    platform.depends(parse_lspci_task)

    if 'clstr_nic_mode_eth' in config:
        cluster_nic_mode = 'eth'
    elif 'clstr_nic_mode_ib' in config:
        cluster_nic_mode = 'ib'

    if 'strg_nic_mode_eth' in config:
        storage_nic_mode = 'eth'
    elif 'strg_nic_mode_ib' in config:
        storage_nic_mode = 'ib'

    def check_clstr_nic_mode():
        #return check_lspci_mode(parse_lspci_task)['cluster_nic'] == cluster_nic_mode
        return check_lspci_mode(platform, config, parse_lspci_task)['cluster_nic'] == cluster_nic_mode
    platform.qualify(check_clstr_nic_mode)

    def check_storage_nic_mode():
        #return check_lspci_mode(parse_lspci_task)['storage_nic'] == storage_nic_mode
        return check_lspci_mode(platform, config, parse_lspci_task)['storage_nic'] == storage_nic_mode
    platform.qualify(check_storage_nic_mode)

    def check_2_strg_ib():
        return check_lspci_ib(parse_lspci_task) == 'strg_nic_qty_2'
    def check_not_2_strg_ib():
        return check_lspci_ib(parse_lspci_task) == 'strg_nic_qty_1'

    strg_ib_bdfs = []
    if 'strg_nic_qty_2' in config:
        platform.qualify(check_2_strg_ib)
        strg_ib_bdfs = ['61', 'e1']
    else:
        platform.qualify(check_not_2_strg_ib)
        strg_ib_bdfs = ['e1']

    # Section 2: Do Ethernet First (Start with any default devices), incorporate other devices based on mode
    # Define ethernet controller device ids
    device_ids = [ 'e2' ]
    ethernet_controller_device_ids = set(map(lambda x: '0000:{}:00.0'.format(x), device_ids))

    # Define bill-of-materials for ethernet controller
    # Add onboards LAN devices to Ethernet Info
    ethernet_controller_info = []
    for device_id in ethernet_controller_device_ids:
        ethernet_controller_id_info = {}
        ethernet_controller_id_info['bdf'] = device_id
        ethernet_controller_id_info['speed'] = '2.5GT/s'
        ethernet_controller_id_info['width'] = 1
        ethernet_controller_info.append(ethernet_controller_id_info)

    # Add LAN devices in Eth mode to Ethernet BOM
    ethernet_controllers_bom = {}
    # ethernet_controllers_bom follows "lspci" format
    ethernet_controllers_config = {
        'class': 'Ethernet controller [0200]',
        'device_id': '1533',
        'vendor': 'Intel Corporation [8086]'
    }
    ethernet_controllers_bom['e2:00.0'] = ethernet_controllers_config

    if storage_nic_mode == 'eth':
        # Add Storage IB devices to Ethernet Info
        strg_ib_controller_link_device_ids = []
        for i in range(2):
            strg_ib_controller_link_device_ids += list(map(lambda x: '0000:{}:00.{}'.format(x, i), strg_ib_bdfs))

        for device_id in strg_ib_controller_link_device_ids:
            ethernet_controller_id_info = {}
            ethernet_controller_id_info['bdf'] = device_id
            ethernet_controller_id_info['speed'] = '16GT/s'
            ethernet_controller_id_info['width'] = 16
            ethernet_controller_info.append(ethernet_controller_id_info)

        # Add storage IB devices to Ethernet BOM
        for bdf in strg_ib_bdfs:
            for x in range(2):
                ethernet_controllers_bom['{}:00.{}'.format(bdf, x)] = {
                    'class': 'Ethernet controller [0200]',
                    'device_id': '101b',
                    'sdevice_id': '0009',
                    'vendor': 'Mellanox Technologies [15b3]'
                }

    if cluster_nic_mode == 'eth':
        # Add IB devices in Eth mode to Ethernet BOM
        if 'bios_firmware_gt_0_25' in config:
            net_eth_bdfs = [ '0c', '12',  '4b', '54', '8d', '94', 'ba', 'cc']
        else:
            net_eth_bdfs = [ '0c', '12',  '4b', '54', '8d', '94', 'ba', 'ca']
        for bdf in net_eth_bdfs:
            ethernet_controllers_bom['{}:00.0'.format(bdf)] = {
                'class': 'Ethernet controller [0200]',
                'device_id': '101b',
                'sdevice_id': '0007',
                'vendor': 'Mellanox Technologies [15b3]'
            }

        # Add IB devices in Eth mode to Ethernet Info (link speed  and width)
        eth_controller_link_device_ids = list(map(lambda x: '0000:{}:00.0'.format(x), net_eth_bdfs))
        for device_id in eth_controller_link_device_ids:
            eth_controller_id_info = {}
            eth_controller_id_info['bdf'] = device_id
            eth_controller_id_info['speed'] = '16GT/s'
            eth_controller_id_info['width'] = 16
            ethernet_controller_info.append(eth_controller_id_info)

    # Add Eth BOM and Info Parameters
    platform.setParameter('ethernet_controllers_bom', ethernet_controllers_bom)
    platform.setParameter('ethernet_controller_info', ethernet_controller_info)

    # Section 3: Do IB Next (Start with any default devices), incorporate other devices based on mode
    # No default devices present in case of IB
    ib_controllers_bom = {}
    ib_controller_link_info = []

    if storage_nic_mode == 'ib':
        storage_nic_bdfs = strg_ib_bdfs
        for bdf in storage_nic_bdfs:
            for i in range (2):
                # ib_controllers_bom follows "lspci" format
                device_bdf = '{}:00.{}'.format(bdf, i)
                ib_controllers_bom[device_bdf] = {
                    'class': 'Infiniband controller [0207]',
                    'device_id': '101b',
                    'sdevice_id': '0009',
                    'vendor': 'Mellanox Technologies [15b3]'
                }
                ib_controller_id_info = {}
                ib_controller_id_info['bdf'] = '0000:{}'.format(device_bdf)
                ib_controller_id_info['speed'] = '16GT/s'
                ib_controller_id_info['width'] = 16
                ib_controller_link_info.append(ib_controller_id_info)

    if cluster_nic_mode == 'ib':
        # Add IB devices in IB mode to IB BOM
        if 'bios_firmware_gt_0_25' in config:
            net_ib_bdfs = [ '0c', '12',  '4b', '54', '8d', '94', 'ba', 'cc']
        else:
            net_ib_bdfs = [ '0c', '12',  '4b', '54', '8d', '94', 'ba', 'ca']
        for bdf in net_ib_bdfs:
            # ib_controllers_bom follows "lspci" format
            ib_controllers_bom['{}:00.0'.format(bdf)] = {
                'class': 'Infiniband controller [0207]',
                'device_id': '101b',
                'sdevice_id': '0007',
                'vendor': 'Mellanox Technologies [15b3]'
            }

        # Add IB devices in IB mode to IB Info (link speed  and width)
        ib_controller_link_device_ids = list(map(lambda x: '0000:{}:00.0'.format(x), net_ib_bdfs))
        for device_id in ib_controller_link_device_ids:
            ib_controller_id_info = {}
            ib_controller_id_info['bdf'] = device_id
            ib_controller_id_info['speed'] = '16GT/s'
            ib_controller_id_info['width'] = 16
            ib_controller_link_info.append(ib_controller_id_info)

    if len(ib_controllers_bom)!=0:
        platform.setParameter('ib_controllers_bom', ib_controllers_bom)

    if len(ib_controller_link_info)!=0:
        platform.setParameter('ib_controller_link_info', ib_controller_link_info)

    # Define bill-of-materials for pcie switches
    pcie_switches_bom = {}

    if 'bios_firmware_gt_0_25' in config:
        pcie_bdfs = ['01', '41', '81', 'b1', '05', '0d', '45',
                    '4c', '85', '8e', 'b5', 'bb', 'c0']
    else:
        pcie_bdfs = ['01', '41', '81', 'b1', '05', '0d', '45',
                     '4c', '85', '8e', 'b5', 'bb', 'c0', 'c1']

    for bdf in pcie_bdfs:
        # pcie_switches_bom follows "lspci" format
        pcie_switches_bom['{}:00.0'.format(bdf)] = {
            'class': '0604',
            'vendor': '1000'
        }

    platform.setParameter('pcie_switches_bom', pcie_switches_bom)

    # Define bill-of-materials for nvswitch
    nvswitch_bom = {}
    if 'bios_firmware_gt_0_25' in config:
        nvswitch_bdfs = [ 'c4', 'c5', 'c6', 'c7', 'c8', 'c9']
    else:
        nvswitch_bdfs = [ 'c2', 'c3', 'c4', 'c5', 'c6', 'c7']
    for bdf in nvswitch_bdfs:
        # nvswitch_bom follows "lspci" format
        nvswitch_bom['{}:00.0'.format(bdf)] = {
            'class_id': '0680',
            'device_id': '1af1',
            'vendor': 'NVIDIA Corporation [10de]'
        }
    platform.setParameter('nvswitch_bom', nvswitch_bom)

def check_lspci_nvme(platform, config, parse_lspci_task):
    if 'bios_firmware_gt_0_25' in config:
        known_u2_nvme_bdfs = ['08:00.0', '09:00.0', '52:00.0', '53:00.0',
                '89:00.0', '8a:00.0', 'ca:00.0', 'cb:00.0']
    else:
        known_u2_nvme_bdfs = ['08:00.0', '09:00.0', '52:00.0', '53:00.0',
                '89:00.0', '8a:00.0', 'c8:00.0', 'c9:00.0']

    lspci = json.loads(parse_lspci_task.getResult())
    u2_nvme_drives = {k: v for k, v in lspci.items() if v['class_id'] == '0108'}
    drives = {k: v for k, v in u2_nvme_drives.items() if k in known_u2_nvme_bdfs}

    if len(drives) > 4:
        # Assume 16 NVMe drive configuration
        return '8u2_nvme'
    return '4u2_nvme'

def define_nvme_parameters(platform, config):
    # Support for multiple U.2 SKUs
    if '4u2_nvme' in config:
        drives = '4u2_nvme'
        nvme_config = [{"1.9TB": 2, "3.8TB": 4}, {"1.9TB": 2, "7.7TB": 4}]
        if 'bios_firmware_gt_0_25' in config:
            known_nvme_bdfs = ['09', '52', '22', '23', '8a' , 'ca']
        else:
            known_nvme_bdfs = ['09', '52', '22', '23', '8a' , 'c8']
    elif '8u2_nvme' in config:
        drives = '8u2_nvme'
        nvme_config = [{"1.9TB": 2, "3.8TB": 8}, {"1.9TB": 2, "7.7TB": 8}]
        if 'bios_firmware_gt_0_25' in config:
            known_nvme_bdfs = ['08', '09', '52', '53', '22', '23', '89', '8a',
                    'ca', 'cb']
        else:
            known_nvme_bdfs = ['08', '09', '52', '53', '22', '23', '89', '8a',
                    'c8', 'c9']


    # Check lspci output for NVMe configuration
    parse_lspci_task = lspci.parse_lspci
    platform.depends(parse_lspci_task)

    def check_nvme():
        #return check_lspci_nvme(parse_lspci_task) == drives
        return check_lspci_nvme(platform, config, parse_lspci_task) == drives
    platform.qualify(check_nvme)

    # Set config parameter for nvme
    platform.setParameter('nvme', nvme_config)

    # Define bill-of-materials for nvme link width and speed
    nvme_link_info = []

    nvme_bdfs = set(map(lambda x: '0000:{}:00.0'.format(x), known_nvme_bdfs))

    for bdf in nvme_bdfs:
        nvme_link = {}
        nvme_link['bdf'] = bdf
        if bdf in ["0000:22:00.0", "0000:23:00.0"]:
            nvme_link['speed'] = '8GT/s'
        else:
            nvme_link['speed'] = '16GT/s'
        nvme_link['width'] = 4
        nvme_link_info.append(nvme_link)
    platform.setParameter('nvme_link_info', nvme_link_info)
    platform.setParameter('nvme_check_smart_log', True)
    platform.setParameter('mdadm_disk_status', True)
    platform.setParameter('mdadm_volume_status', True)

def define_generic_parameters(platform, config):
    def bios_firmware_gt_0_25():
        try:
            platform.depends(sysfs.sysfs_dmi_bios_version)
            bios_version = sysfs.sysfs_dmi_bios_version.getOutput().lower()
            major, minor = (int(x) for x in bios_version.split('.'))
            # Check if BIOS version is greater than 0.25.
            if major > 0 or minor > 25:
                return True
            return False
        except Exception as e:
            logging.error(e)
    if 'bios_firmware_gt_0_25' in config:
        platform.qualify(bios_firmware_gt_0_25)
    else:
        platform.qualify(lambda: not bios_firmware_gt_0_25())

    # Define bill-of-materials for FAN
    fan_bom = {}
    fan_list = ["SPD_FAN_SYS0_R", "SPD_FAN_SYS0_F", "SPD_FAN_SYS1_R",
    "SPD_FAN_SYS1_F", "SPD_FAN_SYS2_R", "SPD_FAN_SYS2_F", "SPD_FAN_SYS3_R",
    "SPD_FAN_SYS3_F", "SPD_FAN_SYS4_R", "SPD_FAN_SYS4_F", "SPD_FAN_SYS5_R",
    "SPD_FAN_SYS5_F", "SPD_FAN_SYS6_R", "SPD_FAN_SYS6_F", "SPD_FAN_SYS7_R",
    "SPD_FAN_SYS7_F", "SPD_FAN_PSU0", "SPD_FAN_PSU1", "SPD_FAN_PSU2",
    "SPD_FAN_PSU3", "SPD_FAN_PSU4", "SPD_FAN_PSU5"]

    for fan in fan_list:
        fan_bom[fan] = True

    platform.setParameter('fan_bom', fan_bom)

    # Define bill-of-materials for PSU
    psu_bom = {}
    # Predicate for DGXA100 for PSU Consistency Checks
    platform.setParameter('enable_psu_consistency_check', True)
    # DGXPM-506: PSU homogeneity
    psu_vendor_info = {}
    psu_model_info = {}
    for i in range(6):
        psu_bom['PWR_PSU{}'.format(i)] = True
        psu_bom['TEMP_PSU{}'.format(i)] = True
        psu_bom['STATUS_PSU{}'.format(i)] = True
        psu_vendor_info['VENDOR_PSU{}'.format(i)] = '0x3C 0x0e 0x0{} 0x02'.format(i)
        psu_model_info['MODEL_PSU{}'.format(i)] = '0x3C 0x0e 0x0{} 0x01'.format(i)
    platform.setParameter('psu_bom', psu_bom)
    platform.setParameter('psu_vendor_info', psu_vendor_info)
    platform.setParameter('psu_model_info', psu_model_info)
    
    # Define lscpu number of cores
    lscpu_number_of_cores = 256
    platform.setParameter('lscpu_number_of_cores', lscpu_number_of_cores)
    platform.setParameter('sw_rel_file', "dgx-release")
    platform.setParameter('summary_serial_num_string', "Serial Number")
    platform.setParameter('summary_base_os_string', "BaseOS Version")
    platform.setParameter('chassis_serial_number_index', '1')
    # Define firmware version conversion from hex to decimal
    platform.setParameter('aux_firmware_version_type', 'hex')
    # Define FRU device details

    # 11/7/19: FRU device tests disabled for now
    # EEPROM has non-unicode contents (see ipmitool fru print)
    # platform.setParameter('fru_devices', ['MID_FRU', 'IOE0_FRU', 'IOE1_FRU', 'PDB_FRU', 'GB_FRU', 'M.2_FRU'])

def __generate_dgxa100_platform_objects(configs, columns=None, pretty=None):
    # Generate an IPlatform object for each valid DGX-2 configuration
    platforms = []
    for config in configs:
        config = list(config)
        platform = IPlatform(name=platform_type, config=config, columns=columns, pretty=pretty)
        check_system_vendor(platform, config)
        check_product_name(platform, config)
        # Define parameters for each platform
        define_gpu_parameters(platform, config)
        define_dimm_parameters(platform, config)
        define_mlnx_parameters(platform, config)
        define_network_parameters(platform, config)
        define_nvme_parameters(platform, config)
        define_generic_parameters(platform, config)

        platforms.append(platform)

    return platforms

def generate_dgxa100_platform_objects():
    # Matrix (Cartesian product) of all possible dgx3 configurations
    variants = [
        ('GA100', ),                                # GPU
        ('40gb_gpu', '80gb_gpu'),                   # SKUs of GPU (40GB or 80GB)
        ('4u2_nvme', '8u2_nvme'),                   # Quantity of U.2 NVMes (4x or 8x)
        ('1tb', '2tb'),                             # DIMMs 1TB supported or 2TB
        ('strg_nic_qty_1', 'strg_nic_qty_2'),       # 1x or 2x Storage NICs present
        ('strg_nic_mode_ib', 'strg_nic_mode_eth'),  # All storage NICs in IB or Eth mode together
        ('clstr_nic_mode_ib', 'clstr_nic_mode_eth'), # All clustering NICs in IB or Eth mode together
        ('', 'bios_firmware_gt_0_25')
    ]

    configs = list(itertools.product(*variants))
    columns = ["GPU"]
    pretty = [
        {
            'GA100': 'Ampere based GPU',
        },
        {
            '40gb_gpu': '40GB A100 GPU',
            '80gb_gpu': '80GB A100 GPU',
        },
        {
            '4u2_nvme': '(x4) U.2 NVMe drives',
            '8u2_nvme': '(x8) U.2 NVMe drives',
        },
        {
            '1tb': '1TB System Memory',
            '2tb': '2TB System Memory',
        },
        {
            'strg_nic_qty_1': '1 Storage Infiniband Controller',
            'strg_nic_qty_2': '2 Storage Infiniband Controller',
        },
        {
            'strg_nic_mode_ib': 'Storage NICs in InfiniBand Mode',
            'strg_nic_mode_eth': 'Storage NICs in Ethernet Mode',
        },
        {
            'clstr_nic_mode_ib': 'Clustering NICs in InfiniBand Mode',
            'clstr_nic_mode_eth': 'Clustering NICs in Ethernet Mode',
        },
        {
            '': 'N/A',
            'bios_firmware_gt_0_25': 'BIOS firmware version greater than 0.25',
        },
    ]
    return __generate_dgxa100_platform_objects(configs=configs, columns=columns, pretty=pretty)
