# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from .task import ITask
import textwrap
from nvsmhealth import config

class ICheck(ITask):
    ''' Class representing a generic NVSMHealth health check task

        This class represents a generic "health check" task appearing in the
        NVSMHealth task dependency graph.

        Typically, a health check task will depend on the output of one or more
        other tasks in the NVSMHealth task dependency graph. These dependencies
        can include system logs, command output, or even the results of other
        health checks. It is discouraged for health checks to perform their own
        data collection.

        Implementing health check classes have the flexibility to define
        "healthy" or "unhealthy" status with whatever metric is most
        appropriate. If the results of other NVSMHealth tasks are used to
        perform health checks, implementing classes are responsible for
        declaring task dependencies via the ITask interface.
    '''
    def __init__(self):
        super(ICheck, self).__init__()
        # __check_result will communicate ICheck task health check results to the caller
        # Increase the healthy count for positive health check
        # Increase the unhealthy count for negative health check
        # Increase the unknown count for unknown health check status
        # TODO Update inheriting class to use this below interface
        self.__check_result     = {'healthy':0, 'unhealthy':0, 'unknown':0, 'informational':0}
        self.__check_messages   = []
        self.__messages         = []

    def getResult(self):
        return self.__check_result

    def getCheckMessages(self):
        return self.__check_messages

    def addMessages(self, message):
        if message not in self.__messages:
            self.__messages.append(message.strip('\n'))

    def getMessages(self):
        return "\n".join(self.__messages)

    def addHealthy(self, count=1):
        self.__check_result['healthy'] += count

    def addUnHealthy(self, count=1):
        self.__check_result['unhealthy'] += count

    def addUnknown(self, count=1):
        self.__check_result['unknown'] += count

    def addInformational(self, count=1):
        self.__check_result['informational'] += count

    def addCheckMessage(self, msg):
        if not msg:
            return
        lines = msg.split('\n')
        indent = ' '*4
        if len(lines) == 1:
            self.__check_messages.append(textwrap.fill(lines[0], initial_indent=indent,
                                subsequent_indent=indent, width=config.display_width))
        else:
            for line in lines:
                self.__check_messages.append(textwrap.fill(line, initial_indent=indent,
                                    subsequent_indent=indent, width=config.display_width))

    '''
    Send completion message after the task is run,
    calculating the overall status of the check
    '''
    def sendComplete(self):
        if (self.__check_result['healthy'] or \
           self.__check_result['unhealthy'] or \
           self.__check_result['unknown'] or \
           self.__check_result['informational']) and \
           (self.getTitle() != ''):
            if self.__check_result['unhealthy']:
                check_result = 'Unhealthy'
            elif self.__check_result['unknown']:
                check_result = 'Unknown'
            elif self.__check_result['informational']:
                check_result = 'Informational'
            else:
                check_result = 'Healthy'
                self.__check_messages = []
        else:
            '''
            This case is useful for classes which inherit
            from ICheck but not necessarily implement any results.
            Eg CheckSuperuserPrivileges
            '''
            check_result = ''
        self.send(check_msgs=self.__check_messages,
                  check_result=check_result,
                  progress=100.0,
                  result_obj=self.getResult())
        # clear results for next run
        self.__check_result     = {'healthy':0, 'unhealthy':0, 'unknown':0, 'informational':0}
        self.__check_messages   = []

    '''
    This function is supposed to be hidden and not used directly
    unless a usecase where one needs to specially prepare and send
    info messages or multiple check results from a single check task

    If `check_result` is set, output is formatted as per nvhealth
    else `info` is printed as is without formatting
    '''
    def send(self, info='',
             check_msgs=[],
             check_result='',
             result_obj=None,
             progress=0.0):
        if check_result != '':
            # Print the description and status. Pad the description with dots. If the
            # description exceeds the allowed length, then wrap it.
            lines = textwrap.wrap(super().getTitle(), width=config.display_width)
            lines[len(lines)-1] = (
                    '{last_line:{fill}{align}{width}} {status}'.format(
                    last_line=lines[len(lines)-1], fill='.', align='<',
                    width=config.display_width, status=check_result))
            print_msg = '\n'.join(lines)
            super().send(info=print_msg)
            super().addMessages(print_msg)
            for msg in check_msgs:
                # TODO formatting
                super().send(info=msg)
                super().addMessages(msg)
            super().send(result_obj=result_obj, progress=progress)
        else:
            super().addMessages(info)
            super().send(info=info, result_obj=result_obj, progress=progress)


class CheckExitStatus(ICheck):
    ''' Class representing a NVSMHealth task to check the exit status of a command

        This class represents the most common type of health check, which is to
        simply check the exit status of a UNIX command. Typically, a zero exit
        status indicates success.

        The parameters for constructing this class are similar to the
        parameters passed to subprocess.run().
    '''
    def __init__(self, command_task, expected_status=0):
        super(CheckExitStatus, self).__init__()
        self.__command_task = command_task

        # Declare dependency on the given command task
        self.depends(command_task)
        # TODO Use ICheck check_result attribute
        self.__check_result = {'healthy':0, 'unhealthy':0, 'unknown':0, 'informational':0}

    # TODO Use ICheck getResult method
    def getResult(self):
        return self.__check_result

    def run(self):
        pass
        #if self.__command_task.exit_status != self.__expected_status:
        #    pass  # TODO: Indicate unhealthy result
        # TODO: Indicate healthy result
