# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import logging
from nvsmhealth.tasks.task import ITask
from nvsmhealth import config

class ICommand(ITask):
    ''' Class representing a generic NVSMHealth command execution task

        This class represents a generic command execution task appearing in the
        NVSMHealth task dependency graph.

        Command execution tasks are typically used to collect command output
        for subsequent analysis by health check tasks. For example, NVSMHealth
        might have a task to run the "lspci" command, and a subsequent "PCIe
        hierarchy check" would examine the output of that command in order to
        determine a health status.

        Because NVSMHealth is strictly a health-check tool and not an
        orchestration tool, it is critical that command execution tasks have
        negligible side-effects on the system state. It is also important that
        temporary files be cleaned up after command execution.
    '''
    def __init__(self):
        super(ICommand, self).__init__()

    # TODO: Define interface used by all NVSMHealth command execution tasks

import subprocess

class RunCommand(ICommand):
    ''' Class representing a NVSMHealth task to run a command

        This class represents the most common type of command execution task,
        which is to simply run a command on the local system (optionally in a
        shell).

        The parameters for constructing this class are similar to the
        parameters passed to subprocess.run().
    '''
    def __init__(self, args, shell=False, timeout=config.default_timeout, binary_output=False, file_name=None):
        super(RunCommand, self).__init__()
        self.__args = args
        self.__shell = shell
        self.__timeout = timeout
        self.__binary_output = binary_output
        self.__output = None
        self.__error = None
        self.__ret = None
        self.__file_name = file_name

    def getArgs(self):
        return self.__args

    def getShell(self):
        return self.__shell

    def getTimeout(self):
        return self.__timeout

    def getOutput(self):
        if self.__output and not self.__binary_output:
            return self.__output.decode('utf-8')
        return self.__output

    def getError(self):
        if self.__error and not self.__binary_output:
            return self.__error.decode('utf-8')
        return self.__error

    def getReturnCode(self):
        return self.__ret

    def run(self, timeout=None):
        if not timeout:
            timeout = self.__timeout
        # Run the command with subprocess.run() and collect the output
        try:
            import signal
            def preexec_function():
                signal.signal(signal.SIGINT, signal.SIG_IGN)
            result = subprocess.run(self.__args, shell=self.__shell, timeout=timeout,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=False,
                encoding=None, universal_newlines=False,
                preexec_fn=preexec_function)
            self.__output = result.stdout
            self.__error = result.stderr
            self.__ret = result.returncode
        except subprocess.TimeoutExpired as ex:
            self.__output = ex.stdout
            self.__error = ex.stderr
            logging.warning('Command "{args}" reached {seconds}secs timeout.'
                .format(args=' '.join(self.__args), seconds=timeout))
        except FileNotFoundError as ex:
            self.__ret = ex.errno
            logging.info('Failed to run command "{args}" with error "{error}"'
                .format(args=' '.join(self.__args), error=ex))
        except UnicodeDecodeError as ex:
            self.__ret = ex
            logging.warning('Failed to run command "{args}" with error "{error}"'
                .format(args=' '.join(self.__args), error=ex))
        else:
            logging.info('Command "{args}" finished with exit code: {ret}'
                .format(args=' '.join(self.__args), ret=self.__ret))

    def __pathify_args(self):
        path = '_'.join(self.__args)
        path = path.replace(' ', '_')
        path = path.replace('/', '_')
        path = path.replace('\\', '')
        path = path.replace('$', '')
        path = path.replace('{', '')
        path = path.replace('}', '')
        path = path.replace('\'', '')
        path = path.replace('"', '')
        logging.debug('Pathify args "{0}" become "{1}"'.format(
                    ' '.join(self.__args), path))
        return path

    def collect(self, context):
        import io, os, tarfile, time
        if self.__file_name:
            path = self.__file_name
        else:
            path = self.__pathify_args()
        # Store stdout and stderr in collection context
        path = os.path.join('nvsmhealth_commands', path)
        stdout_path = os.path.join(path, 'stdout.txt')
        stderr_path = os.path.join(path, 'stderr.txt')
        for data, path in [ (self.__output, stdout_path),
                            (self.__error, stderr_path) ]:
            if not data:
                continue
            with io.BytesIO(data) as buff:
                info = tarfile.TarInfo()
                info.name = path
                info.type = tarfile.REGTYPE
                info.mode = 0o644
                info.uid = 0
                info.gid = 0
                info.uname = 'root'
                info.gname = 'root'
                info.mtime = time.time()
                buff.seek(0, io.SEEK_END)
                info.size = buff.tell()
                buff.seek(0, io.SEEK_SET)
                context.add_file(fileobj=buff, tarinfo=info, path=path)
