# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import glob
import logging
import re

from nvsmhealth.tasks.task import ITask

class ILogs(ITask):
    ''' Class representing a generic NVSMHealth log collection task

        This class represents a generic log collection task appearing in the
        NVSMHealth task dependency graph.

        Log collection is used to gather evidence to help diagnose system
        problems. As a general rule, NVSMHealth will try to gather any and all
        logs that have the potential to corroborate a system problem, even if
        not all logs prove be useful in practice.

        Sometimes, NVSMHealth health check tasks will depend on the contents of
        a log. For example, a given health check task might examine the
        contents of syslog.

        Note that NVSMHealth considers sosreport as a special-case of log
        collection, since sosreport is an external tool over which NVSMHealth
        exerts only limited control.
    '''
    def __init__(self):
        super(ILogs, self).__init__()

    # TODO: Define interface used by all NVSMHealth log collection tasks

class CollectLogFiles(ILogs):
    ''' Class representing a NVSMHealth task to collect logs from file

        This class represents the most common type of log collection. Log files
        are collected from the local filesystem based on the given glob path.
    '''
    def __init__(self, path, exclude=None):
        super(CollectLogFiles, self).__init__()
        self.__path = path
        self.__exclude = exclude

    def getPath(self):
        return self.__path

    def getExclude(self):
        return self.__exclude

    def run(self):
        pass

    def collect(self, context):
        file_paths = sorted(glob.glob(self.__path))
        file_count = 0
        for path in file_paths:
            # Skip files matching the "exclude" pattern
            if self.__exclude and re.search(self.__exclude, path):
                continue
            file_count += 1
            # Add file directly to the collection context
            with open(path, 'rb') as f:
                context.add_file(fileobj=f)
        if file_count == 0:
            logging.info('Found empty path glob "{0}"'.format(self.__path))
