# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import sys
import textwrap

import nvsmhealth.exceptions
from nvsmhealth.tasks.stream import HealthTaskStream
from nvsmhealth import config

class ITask(object):
    ''' Class representing a generic task within the NVSMHealth task dependency graph

        This class represents a generic NVSMHealth task appearing in the
        dependency graph that controls the behavior of NVSMHealth task
        execution. ITask defines the interface that implementing task classes
        use to declare dependencies on other tasks.

        The task dependency graph is necessarily a directed acyclic graph.
        Before executing the task dependency graph, NVSMHealth processes the
        declared dependencies to check for missing dependencies or cyclic
        dependencies. Once the task dependency graph has been validated,
        NVSMHealth executes the task graph by simple graph traversal (finding
        a spanning tree for DAG is trival).

        Tasks in this graph include (but are not limited to) log collection
        (including sosreport), command execution (e.g. lspci, dmidecode), SKU
        determination, and health checks.

        Note that for most tasks, the dependency graph is mostly flat. However,
        the dependency graph can become complicated in certain situations
        (especially with SKU determination and special-cases).
    '''
    def __init__(self):
        super(ITask, self).__init__()
        self.__parents = set()
        self.__children = set()

        # callback for publishing results if installed
        self.__publish_cb = None

        # Annotation members for generating documentation
        self.__name = None
        self.__module = None
        self.__path = None
        self.__title = None
        self.__description = None
        self.__messages = []
        self.__parameters = {}
        self.__predicate = []

    def __addParent(self, task):
        self.__parents.add(task)

    def __addChild(self, task):
        self.__children.add(task)

    def isRoot(self):
        return len(self.__parents) == 0

    def printGraphviz(self, file=sys.stdout, visited=set()):
        if self in visited:
            return
        node_url = 'tasks.html#{}'.format(self.getName())
        visited.add(self)
        print('  "{self}" [URL="{url}"];'.format(
                    self=self.getName(), url=node_url))
        for child in self.__children:
            print('  "{child}" -> "{self}";'.format(
                  self=self.getName(), child=child.getName()),
                  file=file)
            child.printGraphviz(file=file, visited=visited)

    def __checkCycles(self, cumulative_visited=set(), visited=set()):
        if self in visited:
            raise nvsmhealth.exceptions.CyclicTaskDependency(
                    'Cycle detected in task dependency graph')
        cumulative_visited.add(self)
        for child in self.__children:
            child.__checkCycles(cumulative_visited, visited | set([self]))

    def checkCycles(self, cumulative_visited=set()):
        # Traverse the dependency graph, looking for cycles
        self.__checkCycles(cumulative_visited)

    @property
    def parents(self):
        import copy
        return copy.copy(self.__parents)

    @property
    def children(self):
        import copy
        return copy.copy(self.__children)

    # TODO: Define interface used by implementing task classes to declare dependencies
    # TODO: Define interface used by NVSMHealth to execute the dependency graph

    def name(self, text):
        self.__name = text
        return self

    def getName(self):
        return self.__name

    def setModule(self, module):
        self.__module = module

    def getModule(self):
        return self.__module

    def path(self, text):
        self.__path = text
        return self

    def getPath(self):
        return self.__path

    def title(self, text):
        self.__title = text
        return self

    def getTitle(self):
        return self.__title

    def describe(self, text):
        self.__description = text
        return self

    def getDescription(self):
        return self.__description

    def addMessages(self, message):
        if message not in self.__messages:
            self.__messages.append(message.strip('\n'))

    def getMessages(self):
        return "\n".join(self.__messages)

    def parameters(self, param, required=False):
        self.__parameters[param] = required
        return self

    def getParameters(self):
        return self.__parameters

    def depends(self, task):
        assert isinstance(task, ITask)
        # Store dependency information that this task is declaring
        # NOTE: Because inter-module dependencies must be resolved first, the
        #       task.__addChild() method is not called at this time. Instead,
        #       the self._registerWithParents() method is called in a later
        #       pass after inter-module dependencies are resolved.
        self.__addParent(task)
        return self

    def predicate(self, cb):
        self.__predicate.append(cb)
        return self

    def checkPredicate(self):
        return all(map(lambda cb: cb(), self.__predicate))

    def setCallback(self, cb):
        self.__publish_cb = cb
        return self

    def getCallback(self):
        return self.__publish_cb

    def _registerWithParents(self):
        for task in self.__parents:
            task.__addChild(self)

    '''
    Send completion message after the task is run
    Ideally useful to stream the progress to HTTP Client
    '''
    def sendComplete(self):
        result_obj = None
        self.send(result_obj=result_obj,
                  progress=100.0)

    '''
    Send a message either to stdout or http stream

    if `msg` is set, it is formatted as per nvhealth info messages
    else `info` is printed as is without formatting
    '''
    def send(self, info='', msg='', result_obj=None, progress=0.0):
        if msg != '':
            # Print the description and status. Pad the description with dots. If the
            # description exceeds the allowed length, then wrap it.
            fmt = "{0} [{1}]"
            print_msg = fmt.format(self.__title, msg)
            lines = textwrap.wrap(print_msg, width=config.display_width)
            lines[len(lines)-1] = (
                    '{last_line:{fill}{align}{width}} '.format(
                    last_line=lines[len(lines)-1], fill='.', align='<',
                    width=config.display_width))
            print_msg = '\n'.join(lines)
        else:
            print_msg = info
        if self.__publish_cb:
            stream = HealthTaskStream(self.__name, self.__publish_cb)
            self.addMessages(print_msg)
            if result_obj:
                stream.stream_obj.result = result_obj
            stream.stream(message=print_msg, progress=progress)
        else:
            if print_msg != '':
                self.addMessages(print_msg)
                print(print_msg)

