#!/usr/bin/env bash

# Return Code Map
# 0  - Script ran successfully
# User defined exit codes
# 71 - Grafana Config is not present
# 72 - Problem with grafana installation
# 81 - Prometheus Config is not present
# 82 - Problem with Prometheus installation
# 91 - NVSM Exporter Service not found

# Make sure root permissions
if [ "$EUID" -ne 0 ]
  then echo "Please run as root"
  exit
fi


if [ -f /etc/os-release ]; then
    # freedesktop.org and systemd
    . /etc/os-release
    OS=$NAME
    VER=$VERSION_ID
fi

help_text() {
        echo -e "\n Welcome to the NVSM Enable Dashboards Utility"
        echo -e "\n -h or --help for help text"
        echo -e " --logfile to omit output from STDOUT and log at /var/log/nvsm/nvsm_dashboard.log"
        exit 0
}

# Help Text
if [[ $# -eq 1 ]] && [[ $1 == "--help" ]] || [[ $1 == "-h" ]]; then
        help_text
fi


#Enable logging
if [[ $# -eq 1 ]] && [[ $1 == "--logfile" ]]; then
        logfile="/var/log/nvsm/nvsm_dashboard.log"
        # echo "Logging in $logfile"
        exec &>> ${logfile}
else
        help_text
fi

# if no arguments then print on console
# if argument --log-file then log into file

test_output() {
	echo "Running CMD: $1"
	if eval $1 > /dev/null 2>&1
	then 
		echo -e "Return Command 0 \n"
		return 0
	else
		echo -e "Return Command Non-Zero \n"
		return $2
	fi
}

echo -e "\n========================"
echo -e "Enabling NVSM Dashboards"
echo -e "========================\n"
eval date

# Check for nvsm-exporter service
test_output "find /usr/lib/systemd/system/nvsm-exporter.service" "91"
rc=$?
if [ $rc -ne 0 ] ; then
	echo "NVSM Exporter Service not found. Exiting.."
	exit $rc
fi


# Check for Grafana Config
test_output "find /etc/nvsm/nvsm-exporter/grafana_env.config" "71"
rc=$?
if [ $rc -ne 0 ] ; then
	echo "Grafana config not found. Exiting.."
	exit $rc
fi

# Check for Prometheus Config
test_output "find /etc/nvsm/nvsm-exporter/prometheus_env.config" "81"
rc=$?
if [ $rc -ne 0 ]; then
	echo "Prometheus config not found. Exiting.."
	exit $rc
fi

# Check if Grafana is installed if not install
echo "Checking Grafana Installation"
if [[ $OS == "Ubuntu" ]]; then
	test_output "apt list --installed  | grep -i "grafana"" "72"
	rc=$?
	if [ $rc -ne 0 ]; then
		echo "Grafana OSS Package not found. Installing.."
		wget -q -O - https://packages.grafana.com/gpg.key | sudo apt-key add -
		cat /etc/apt/sources.list.d/grafana.list | grep -i "deb https://packages.grafana.com/oss/deb stable main"
		if [ $? -ne 0 ]; then
			echo "deb https://packages.grafana.com/oss/deb stable main" | sudo tee -a /etc/apt/sources.list.d/grafana.list 
		fi
		export DEBIAN_FRONTEND=noninteractive \
		&& apt-get update -y && apt-get -y install \
		apt-transport-https \
		software-properties-common wget \
		grafana
		if [ $? -ne 0 ]; then
			exit $rc
		fi
	fi
else
	test_output "yum list installed | grep -i "grafana.x86_64"" "72"
	rc=$?
	if [ $rc -ne 0 ]; then
	echo "RHEL Grafana Package not found. Installing.."
cat <<EOF | sudo tee /etc/yum.repos.d/grafana.repo
[grafana]
name=grafana
baseurl=https://packages.grafana.com/oss/rpm
repo_gpgcheck=1
enabled=1
gpgcheck=1
gpgkey=https://packages.grafana.com/gpg.key
sslverify=1
sslcacert=/etc/pki/tls/certs/ca-bundle.crt
EOF
	sudo yum check-update
	sudo yum install -y grafana
	if [ $? -ne 0 ]; then
		exit $rc
	fi	
	fi
fi

sudo groupadd -f prometheus
sudo useradd --no-create-home --shell /bin/false prometheus -g prometheus
sudo mkdir -p /etc/prometheus
sudo mkdir -p /var/lib/prometheus/data/
sudo chown prometheus:prometheus /etc/prometheus
sudo chown -R prometheus:prometheus /var/lib/prometheus/
sudo chown -R prometheus:prometheus /var/lib/prometheus/data/
cd /tmp/
curl -LO https://github.com/prometheus/prometheus/releases/download/v2.21.0/prometheus-2.21.0.linux-amd64.tar.gz
tar xvf prometheus-2.21.0.linux-amd64.tar.gz
sudo cp prometheus-2.21.0.linux-amd64/prometheus /usr/local/bin/
sudo cp prometheus-2.21.0.linux-amd64/promtool /usr/local/bin/
sudo chown prometheus:prometheus /usr/local/bin/prometheus
sudo chown prometheus:prometheus /usr/local/bin/promtool
sudo cp -r prometheus-2.21.0.linux-amd64/consoles /etc/prometheus
sudo cp -r prometheus-2.21.0.linux-amd64/console_libraries /etc/prometheus
sudo chown -R prometheus:prometheus /etc/prometheus/consoles
sudo chown -R prometheus:prometheus /etc/prometheus/console_libraries
sudo rm -rf /tmp/prometheus-2.21.0.linux-amd64.tar.gz

cat <<EOF | sudo tee /etc/systemd/system/prometheus.service
[Unit]
Description=Prometheus
Wants=network-online.target
After=network-online.target

[Service]
Type=simple
ExecStart=/usr/local/bin/prometheus \
    --config.file /etc/nvsm/prometheus/prometheus.yml \
    --storage.tsdb.path /var/lib/prometheus/ \
    --web.console.templates=/etc/prometheus/consoles \
    --web.console.libraries=/etc/prometheus/console_libraries
[Install]
WantedBy=multi-user.target
EOF

# Any command which fails after this should exit
set -e

# Configurre Grafana and Prometheus 
systemctl daemon-reload

# Currently enable all services
systemctl enable nvsm-exporter.service
systemctl enable grafana-server
systemctl enable prometheus.service

# Stop nvsm-exporter
systemctl stop nvsm-exporter.service

# Stop Grafana
echo "Stopping Grafana.service"
systemctl stop grafana-server.service

# Stop Prometheus
echo "Stopping Prometheus.service"
systemctl stop prometheus.service

# Configure Grafana
mkdir -p /etc/systemd/system/grafana-server.service.d/
{ echo "[Service]"; 
	echo "EnvironmentFile=/etc/nvsm/nvsm-exporter/grafana_env.config";
} | tee /etc/systemd/system/grafana-server.service.d/nvsm.conf

# Configure Prometheus
mkdir -p /etc/systemd/system/prometheus.service.d/
{ echo "[Service]"; 
	echo "EnvironmentFile=/etc/nvsm/nvsm-exporter/prometheus_env.config";
} | tee /etc/systemd/system/prometheus.service.d/nvsm.conf

systemctl daemon-reload
systemctl start nvsm-exporter.service
systemctl start grafana-server.service
systemctl start prometheus.service

exit 0